/* Evlan Client - Command shell for the Evlan programming language.
 * Copyright (C) 2004-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

import java.io.*;
import java.util.*;

public class TransferManager
{
   // public interface ===============================================
   public TransferManager(CommandStream stream)
   {
      mUploads = new Hashtable();
      mDownloads = new Hashtable();
      mStream = stream;
   }

   public ServerCommand read() throws IOException
   {
      while(true)
      {
         String text = mStream.read();
         if(text == null)
            return null;
         ServerCommand command = new ServerCommand(text);

         if(command.type == "DATA")
         {
            Integer idObject = new Integer(command.id);
            OutputStream out = (OutputStream)mDownloads.get(idObject);

            if(command.size == 0)
            {
               if(out != null)
                  out.close();
               mDownloads.remove(idObject);
            }
            else
            {
               byte[] data = mStream.readData(command.size);
               if(out != null)
                  out.write(data);
            }
         }
         else if(command.type == "RECEIVED")
         {
            UploadThread thread = (UploadThread)mUploads.get(new Integer(command.id));
            if(thread != null)
               thread.receiveNotify();
         }
         else if(command.type == "BROKEN")
         {
            UploadThread thread = (UploadThread)mUploads.get(new Integer(command.id));
            if(thread != null)
               thread.brokenNotify();
            return command;
         }
         else
            return command;
      }
   }

   public void write(String command) throws IOException
   {
      synchronized(mStream)
      {
         mStream.write(command);
      }
   }

   public void writeCommandWithData(String command, byte[] data) throws IOException
   {
      synchronized(mStream)
      {
         mStream.write(command);
         mStream.writeData(data, data.length);
      }
   }

   public void close() throws IOException
   {
      mStream.close();
   }

   public void upload(int id, InputStream input)
   {
      UploadThread thread = new UploadThread(id, input);
      thread.start();
      mUploads.put(new Integer(id), thread);
   }

   public void download(int id, OutputStream output)
   {
      mDownloads.put(new Integer(id), output);
   }

   // internal =======================================================
   private Map mUploads;
   private Map mDownloads;
   private CommandStream mStream;

   private class UploadThread extends Thread
   {
      private InputStream mInput;
      private int mId;
      private int mWatermark;

      private static final int LOW_WATERMARK  = 8;
      private static final int HIGH_WATERMARK = 16;

      public UploadThread(int id, InputStream input)
      {
         mInput = input;
         mId = id;
         mWatermark = 0;
      }

      public synchronized void receiveNotify()
      {
         if(--mWatermark <= LOW_WATERMARK)
            notify();
      }

      public synchronized void brokenNotify()
      {
         mWatermark = -1;
         notify();
      }

      public void run()
      {
         String prefix = "/data " + mId + " ";
         try
         {
            byte[] buffer = new byte[8192];

            while(true)
            {
               int size = mInput.read(buffer);
               if(size == -1)
                  break;
               else if(size == 0)
                  continue;

               synchronized(mStream)
               {
                  mStream.write(prefix + size);
                  mStream.writeData(buffer, size);
               }

               synchronized(this)
               {
                  if(mWatermark == -1)
                     break;
                  if(++mWatermark >= HIGH_WATERMARK)
                     try{wait();}catch(Exception e){}
                  if(mWatermark == -1)
                     break;
               }
            }
         } catch(IOException e) {}

         if(mWatermark != -1)
         {
            try {mStream.write(prefix + "0");}
               catch(IOException e) {}
         }

         mUploads.remove(new Integer(mId));
      }
   }
}
