#! /bin/bash

if [ $# -lt 2 -o $# -gt 3 ]; then
   cat << __EOF__
USAGE: mkmodule.sh TYPE NAME [IMPLEMENTS]
Creates a new module in the source tree rooted at "src".  Produces both the cpp
and header file.
   TYPE          Indicates what type of module to create.  One of the
                 following:
                    class           A generic class.
                    interface       An abstract interface.
                    implementation  An insulated implementation.
                    frontend        An executable front end.
                    function        A simple function.
   NAME          The fully-qualified module name (including scope).
                 Example:
                    framework::common::String
   IMPLEMENTS    For interfaces and implementations, specifies an interface to
                 inherit.  Default is "Object".  Note that you should NOT
                 specify the namespace of the interface.
mkmodule.sh requires that the environment variable \$INITIALS be set to the
user's initials, such that the user may be credited with the ceration of the
module.
__EOF__
   exit 1
fi

if [ "$INITIALS" == "" ]; then
   echo "You have not set your initials.  Set the \$INITIALS environment variable"
   echo "to your initials.  Example:"
   echo "   export INITIALS=KV"
   exit 1
fi

TYPE=$1
CODEPATH=$2
COMPONENTS=`echo "$CODEPATH" | sed -e "s/::/ /g"`
UNDERSCORES=`echo "$CODEPATH" | sed -e "s/::/_/g"`

DATE=`date +%Y/%m/%d`

NAMESPACES=""
NAME=""
for DIR in $COMPONENTS; do
   NAMESPACES="$NAMESPACES $NAME"
   NAME=$DIR
done

PACKAGEDIR=src
HEADER=src/header.txt
for NAMESPACE in $NAMESPACES; do
   PACKAGEDIR=$PACKAGEDIR/$NAMESPACE
   if [ -e $PACKAGEDIR/header.txt ]; then
      HEADER=$PACKAGEDIR/header.txt
   fi
done

mkdir -p $PACKAGEDIR

if [ $# -ge 3 ]; then
   INTERFACE=$3
else
   INTERFACE=Object
fi

case $TYPE in
   class | interface | implementation | frontend | function )
      ;;
   * )
      echo "Unknown module type: $TYPE"
      exit 1
      ;;
esac

if [ -e $PACKAGEDIR/$NAME.h -o -e $PACKAGEDIR/$NAME.cpp ]; then
   echo "Module already exists!"
   exit 1
fi

exec 3>$PACKAGEDIR/$NAME.h
exec 4>$PACKAGEDIR/$NAME.cpp

######################################################################

if [ -e $HEADER ]; then
   cat >&3 < $HEADER
   cat >&4 < $HEADER
fi

cat >&3 << __EOF__

// ===================================================================

/* Revision History:
 * $DATE ($INITIALS)
 *  + First Version
 */

// ===================================================================

#ifndef ${UNDERSCORES}_h__INCLUDED__
#define ${UNDERSCORES}_h__INCLUDED__

__EOF__

cat >&4 << __EOF__

// ===================================================================

/* Revision History:
 * $DATE ($INITIALS)
 *  + First Version
 */

// ===================================================================

__EOF__

######################################################################

case $TYPE in
   frontend )
      echo "#include <framework/FrontEnd.h>" >&3
      ;;
   * )
      if [ "$INTERFACE" != "Object" ]; then
         echo "#include \"$INTERFACE.h\"" >&3
      else
         echo "#include <framework/common/all.h>" >&3
      fi
      ;;
esac

echo "#include \"$NAME.h\"" >&4
echo >&3
echo >&4

for NAMESPACE in $NAMESPACES; do
   echo "namespace $NAMESPACE {" >&3
   echo "namespace $NAMESPACE {" >&4
   BACKWARDS="$NAMESPACE $BACKWARDS"
done
echo "using namespace ::framework::common;" >&3

if [ "$TYPE" == "frontend" ]; then
   echo "using ::framework::FrontEnd;" >&4
   echo "using ::framework::FrontEndPtr;" >&4
   echo "using ::framework::messages;" >&4
fi

cat >&3 << __EOF__

// ===================================================================
// AUTHOR ($INITIALS): Created module.
// TODO ($INITIALS): Document $CODEPATH
// TODO ($INITIALS): Test $CODEPATH

__EOF__

cat >&4 << __EOF__

// ===================================================================
// AUTHOR ($INITIALS): Created module.
// TODO ($INITIALS): Implement $CODEPATH
// TODO ($INITIALS): Review $CODEPATH

__EOF__

######################################################################

case $TYPE in
   class )
########################################
cat >&3 << __EOF__
class $NAME;
typedef SmartPointer<$NAME> ${NAME}Ptr;

class $NAME
{
public:
   $NAME();
   ~$NAME();

private:
   //disable copying
   $NAME(const $NAME&);
   $NAME& operator=(const $NAME&);
};
__EOF__
####################
cat >&4 << __EOF__
$NAME::$NAME()
{
}

$NAME::~$NAME()
{
}
__EOF__
########################################
      ;;
   interface )
########################################
cat >&3 << __EOF__
interface $NAME;
typedef SmartPointer<$NAME> ${NAME}Ptr;

interface $NAME: implements $INTERFACE
{
   virtual ~$NAME();
};
__EOF__
####################
cat >&4 << __EOF__
$NAME::~$NAME()
{
}
__EOF__
########################################
      ;;
   implementation )
########################################
cat >&3 << __EOF__
${INTERFACE}Ptr make$NAME();
__EOF__
####################
cat >&4 << __EOF__
implementation $NAME: implements $INTERFACE
{
   $NAME();
   ~$NAME();

   // $INTERFACE -----------------------------------------------------

   // internal -------------------------------------------------------
};

$NAME::$NAME()
{
}

$NAME::~$NAME()
{
}

// factory function ==================================================
${INTERFACE}Ptr make$NAME()
{
   return spnew $NAME;
}
__EOF__
########################################
      ;;
   frontend )
########################################
cat >&3 << __EOF__
framework::FrontEndPtr make$NAME();
__EOF__
####################
cat >&4 << __EOF__
implementation $NAME: implements FrontEnd,
                      implements messages::MessageReceiver
{
   $NAME();
   ~$NAME();

   // FrontEnd -------------------------------------------------------
   ProgramInfo getProgramInfo();
   void run(StartupInfo startupInfo);

   // MessageReceiver ------------------------------------------------
   void receive(messages::MessagePtr message);

   // internal -------------------------------------------------------
   enum Switches
   {
//      SWITCH_,
      SWITCH_COUNT
   };

   static const SwitchDescriptor SYNTAX[SWITCH_COUNT];
};

const FrontEnd::SwitchDescriptor $NAME::SYNTAX[SWITCH_COUNT] =
{
//   {SWITCH_, \$(""), ' ', 0},
};

$NAME::$NAME()
{
}

$NAME::~$NAME()
{
}

// FrontEnd ----------------------------------------------------------
FrontEnd::ProgramInfo $NAME::getProgramInfo()
{
   ProgramInfo info;
   info.name = \$();   //TODO
   info.switchSyntax = Array<const SwitchDescriptor>::makeStatic(SYNTAX, SWITCH_COUNT);
   info.helpMessage = \$("This program has not provided usage information.");
   return info;
}

void $NAME::run(StartupInfo startupInfo)
{
}

// MessageReceiver ---------------------------------------------------
void $NAME::receive(messages::MessagePtr message)
{
}

// factory function ==================================================
FrontEndPtr make$NAME()
{
   return spnew $NAME;
}
__EOF__
########################################
      ;;
   function )
########################################
cat >&3 << __EOF__
${INTERFACE} $NAME();
__EOF__
####################
cat >&4 << __EOF__
${INTERFACE} $NAME()
{
}
__EOF__
########################################
      ;;
esac

######################################################################

cat >&3 << __EOF__

// ===================================================================

__EOF__

cat >&4 << __EOF__

// ===================================================================

__EOF__

for NAMESPACE in $BACKWARDS; do
   echo "} //namespace $NAMESPACE" >&3
   echo "} //namespace $NAMESPACE" >&4
done

cat >&3 << __EOF__

#endif //included
__EOF__
