/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/12/15 (KV)
 *  + First Version
 */

// ===================================================================

#include "EvlanService.h"
#include <evlan/vm/init.h>
#include <evlan/vm/arrays.h>
#include <evlan/vm/native.h>
#include <evlan/vm/loader.h>
#include <evlan/vm/persistence.h>
#include <evlan/vm/shortcuts.h>
#include <evlan/api/directory.h>

namespace evlan {
using ::framework::FrontEnd;
using ::framework::FrontEndPtr;
using namespace vm::shortcuts;

//Fuck the C++ standard for disallowing this, and shame on GCC for
//allowing it in previous versions then spontaneously breaking it.
//using ::framework::messages;
namespace messages {using namespace ::framework::messages;}

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::EvlanService
// TODO (KV): Review evlan::EvlanService

implementation EvlanService: implements FrontEnd,
                             implements messages::MessageReceiver
{
   EvlanService();
   ~EvlanService();

   // FrontEnd -------------------------------------------------------
   ProgramInfo getProgramInfo();
   void run(StartupInfo startupInfo);

   // MessageReceiver ------------------------------------------------
   void receive(messages::MessagePtr message);

   // internal -------------------------------------------------------
   enum Switches
   {
//      SWITCH_,
      SWITCH_CREATE,
      SWITCH_STATEDIR,
      SWITCH_COUNT
   };

   static const SwitchDescriptor SYNTAX[SWITCH_COUNT];

   bool mIsRunning;
   String mAppName;
   int mFrameNumber;

   framework::SystemPtr mSystem;
   String mStateDir;
   String mModulesDir;
   compiler::ErrorCollectorPtr mErrorCollector;

   messages::MessagePtr mQuitMessage;
   messages::MessagePtr mSaveMessage;
   messages::MessagePtr mRestoreMessage;

   static EvlanService* mMainInstance;

   implementation ErrorCollectorImpl: implements compiler::ErrorCollector
   {
      ErrorCollectorImpl();
      ~ErrorCollectorImpl();

      String setModule(String name);
      int getErrorCount();

      void addError  (int line, int column, String description);
      void addError  (String description);
      void addWarning(int line, int column, String description);
      void addWarning(String description);

      int mErrorCount;
      String mModule;
   };
};

const FrontEnd::SwitchDescriptor EvlanService::SYNTAX[SWITCH_COUNT] =
{
//   {SWITCH_, $(""), ' ', 0},
   {SWITCH_CREATE, $("create"), 'c', 0},
   {SWITCH_STATEDIR, $("statedir"), 's', 1},
};

EvlanService* EvlanService::mMainInstance = null;

EvlanService::EvlanService()
   : mIsRunning(false), mAppName($("evlan"))
{
   mQuitMessage = makeMessage(this, 0, null);
   mSaveMessage = makeMessage(this, 0, null);
   mRestoreMessage = makeMessage(this, 0, null);

   mMainInstance = this;
}

EvlanService::~EvlanService()
{
   /* shutdown() crashes at the moment, and I don't care to fix it
    * because it has no practical purpose when the process just exits
    * immediately afterwards anyway.  shutdown() especially needs to
    * NOT be called when exceptions are thrown since this would
    * prevent the exception report from being seen. */
//   if(mIsRunning)
//      vm::init::shutdown();
}

// FrontEnd ----------------------------------------------------------
FrontEnd::ProgramInfo EvlanService::getProgramInfo()
{
   ProgramInfo info;
   info.name = $("evlan");   //TODO
   info.switchSyntax = Array<const SwitchDescriptor>::makeStatic(SYNTAX, SWITCH_COUNT);
   info.helpMessage = $(
      "evlan [-c|--create] [-s|--statedir STATEDIR] [PROGRAMNAME] [FRAMENUM]\n"
      "   Runs the evlan service.  PROGRAMNAME identifies the saved program\n"
      "   state to run.  If omitted, \"evlan\" is used.  If -c or --create\n"
      "   is specified, then a new state by this name is created.  Otherwise,\n"
      "   the previously saved state is run.  FRAMENUM specifies the frame\n"
      "   number within the saved state file to load.  If omitted, the last\n"
      "   frame is loaded by default.  Persistent state information is placed\n"
      "   in and read from the directory STATEDIR, which must be an absolute\n"
      "   path.  If STATEDIR is not specified, the current directory is used.\n");
   return info;
}

void EvlanService::run(StartupInfo startupInfo)
{
   mSystem = startupInfo.system;

   bool creating = false;

   mStateDir = mSystem->getCurrentDirectory();

   for(int i = 0; i < startupInfo.switches.getSize(); i++)
   {
      if(startupInfo.switches[i].identifier == SWITCH_CREATE)
         creating = true;
      else if(startupInfo.switches[i].identifier == SWITCH_STATEDIR)
         mStateDir = startupInfo.switches[i].arguments[0];
   }

   if(startupInfo.arguments.getSize() > 0)
      mAppName = startupInfo.arguments[0];
   mFrameNumber = -1;
   if(startupInfo.arguments.getSize() > 1)
   {
      int endPoint = 0;
      mFrameNumber = startupInfo.arguments[1].toInt(&endPoint);
      if(endPoint != startupInfo.arguments[1].getLength())
         FW_THROW(spnew UsageError($("Frame number was not an integer.")));
   }

   mErrorCollector = spnew ErrorCollectorImpl;

   mModulesDir = mSystem->createDirectory(mSystem->getDataFileDirectory(), $("modules"));

   String currentDir = mSystem->getCurrentDirectory();
//   String configDir  = mSystem->createDirectory(mSystem->getSystemConfigDirectory(), mAppName);
   String userDir    = mSystem->createDirectory(mStateDir, mAppName);

   api::directory::addRoot($("current"), currentDir );
//   api::directory::addRoot($("config") , configDir  );
   api::directory::addRoot($("user")   , userDir    );
   api::directory::addRoot($("state")  , mStateDir  );
   api::directory::addRoot($("modules"), mModulesDir);

   if(creating)
   {
      vm::init::init(mSystem, mStateDir, mModulesDir, mErrorCollector);
      vm::persistence::create(mAppName);

      simPush(Value::makeContext(Value::STOP));
      vm::loader::loadModule($("/evlan.org/internal/InitService"));
      vm::native::makeSystemProcess(mSystem);
      vm::arrays::makeString(mAppName);
      simApply(2);
      simQueue();
   }
   else
   {
      Debug::log() << "Restoring state...";
      vm::init::restore(mSystem, mStateDir, mModulesDir, mErrorCollector, mAppName, mFrameNumber);
   }

   Debug::log() << "Running...";
   mIsRunning = true;
}

// MessageReceiver ---------------------------------------------------
void EvlanService::receive(messages::MessagePtr message)
{
   if(message == mQuitMessage)
   {
      vm::init::shutdown();
      mIsRunning = false;
   }
   else if(message == mSaveMessage)
   {
      Debug::log() << "Saving state...";
      vm::persistence::update();
   }
   else if(message == mRestoreMessage)
   {
      vm::init::shutdown();
      vm::init::restore(mSystem, mStateDir, mModulesDir, mErrorCollector, mAppName, mFrameNumber);
   }
}

// EvlanService::ErrorCollectorImpl ==================================
EvlanService::ErrorCollectorImpl::ErrorCollectorImpl(): mErrorCount(0) {}
EvlanService::ErrorCollectorImpl::~ErrorCollectorImpl() {}

String EvlanService::ErrorCollectorImpl::setModule(String name)
{
   String oldName = mModule;
   mModule = name;
   return oldName;
}

int EvlanService::ErrorCollectorImpl::getErrorCount()
   {return mErrorCount;}

void EvlanService::ErrorCollectorImpl::addError(int line, int column, String description)
   {++mErrorCount; Debug::log(Debug::ERROR) << mModule << ":" << (line+1) << ":" << (column+1) << ": " << description;}
void EvlanService::ErrorCollectorImpl::addError(String description)
   {++mErrorCount; Debug::log(Debug::ERROR) << mModule << ": " << description;}
void EvlanService::ErrorCollectorImpl::addWarning(int line, int column, String description)
   {Debug::log(Debug::WARNING) << mModule << ":" << (line+1) << ":" << (column+1) << ": " << description;}
void EvlanService::ErrorCollectorImpl::addWarning(String description)
   {Debug::log(Debug::WARNING) << mModule << ": " << description;}

// glue code =========================================================
namespace
{
   EVLAN_NATIVE_FUNCTION_0(serviceShutdown)
   {
      EvlanService::mMainInstance->mQuitMessage->send();
      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_0(serviceSave)
   {
      EvlanService::mMainInstance->mSaveMessage->send();
      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(serviceRestore, frameNumber)
   {
      if(!frameNumber.isInteger())
         return simReturn();

      EvlanService::mMainInstance->mFrameNumber = (int)frameNumber.builtin.value.scalar;
      EvlanService::mMainInstance->mRestoreMessage->send();
      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_0(serviceDestroyState)
   {
      vm::persistence::create(EvlanService::mMainInstance->mAppName);
      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(serviceDie, message)
   {
      FW_THROW(spnew FatalError(vm::arrays::getString(message)));
   }
}

// factory function ==================================================
FrontEndPtr makeEvlanService()
{
   return spnew EvlanService;
}

// ===================================================================

} //namespace evlan
