/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/04/30 (KV)
 *  + First Version
 */

// ===================================================================

#include "clock.h"
#include <ctime>
#include <framework/Clock.h>
#include <evlan/vm/native.h>
#include <evlan/vm/memory.h>
#include <evlan/vm/simulator.h>
#include <evlan/vm/operations.h>
#include <evlan/vm/messages.h>
#include <evlan/vm/loader.h>
#include <evlan/vm/shortcuts.h>

namespace evlan {
namespace api {
using namespace evlan::vm;
using namespace evlan::vm::shortcuts;
using framework::messages::MessageReceiver;
using framework::messages::MessagePtr;
using framework::messages::makeMessage;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::api::clock
// TODO (KV): Review evlan::api::clock

namespace clock
{
   // process info ===================================================
   implementation TimerInfo: implements MessageReceiver
   {
      TimerInfo(core::Process* process, framework::SystemPtr system);
      TimerInfo(core::Process* process, framework::ClockPtr clock, double time);
      ~TimerInfo();

      void receive(MessagePtr message);

      framework::ClockPtr mClock;
      core::Process* mProcess;
      MessagePtr mMessage;
   };
   typedef SmartPointer<TimerInfo> TimerInfoPtr;

   TimerInfo::TimerInfo(core::Process* process, framework::SystemPtr system)
   {
      mProcess = process;

      if(process->state.isBoolean())
      {
         //already finished
      }
      else if(!process->state.isArray()
           || process->state.array.size != 2
           || !process->state.array.elements[0].isScalar())
         FW_THROW(spnew FatalError($("Type error in timer process state.")));
      else
      {
         process->keepAlive = true;

         mClock = system->getClock();

         double time = process->state.array.elements[1].builtin.value.scalar;

         mMessage = makeMessage(this, 0, null);
         mClock->notifyAtTime(time, mMessage);
      }
   }

   TimerInfo::TimerInfo(core::Process* process, framework::ClockPtr clock, double time)
   {
      mProcess = process;
      process->keepAlive = true;

      mClock = clock;

      mMessage = makeMessage(this, 0, null);
      mClock->notifyAtTime(time, mMessage);
   }

   TimerInfo::~TimerInfo()
   {}

   void TimerInfo::receive(MessagePtr message)
   {
      if(message == mMessage)
      {
         simPush(mProcess->state.array.elements[1]);

         while(!gStackTop->isEmpty())
         {
            simPush(gStackTop->array.elements[0]);
            simPush(simReturn(Value::makeBoolean(true)));
            simQueue();
            *gStackTop = gStackTop->array.elements[1];
         }

         simPop();

         simPush(Value::makeBoolean(true));
         memory::modifyProcessState(mProcess);
         mProcess->keepAlive = false;
         mMessage->disable();
         mMessage = null;
      }
   }

   EVLAN_NATIVE_PROCESS(timer, TimerInfo)

   // glue code ======================================================
   EVLAN_NATIVE_FUNCTION_1(clockGetTime, systemProcess)
   {
      if(!systemProcess.isProcess())
         return simTypeError(systemProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(system, systemProcess, framework::SystemPtr);

      return simReturn(Value::makeScalar(system->getClock()->getTime()));
   }

   EVLAN_NATIVE_FUNCTION_1(clockGetRealTime, systemProcess)
   {
      if(!systemProcess.isProcess())
         return simTypeError(systemProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(system, systemProcess, framework::SystemPtr);

      return simReturn(Value::makeScalar(system->getClock()->getRealTime()));
   }

   EVLAN_NATIVE_FUNCTION_2(timerMake, systemProcess, endTime)
   {
      if(!systemProcess.isProcess() || !endTime.isScalar())
         return simTypeError(systemProcess, endTime);
      EVLAN_NATIVE_GET_PROCESS_INFO(system, systemProcess, framework::SystemPtr);

      Value* elements = memory::allocateValues(2);
      elements[0] = endTime;
      elements[1] = Value::makeEmpty();

      simPush(Value::makeArray(2, elements));
      memory::makeProcess($("timer"));

      gStackTop->process.process->nativeInfo =
         spnew TimerInfo(gStackTop->process.process, system->getClock(),
                         endTime.builtin.value.scalar);

      return simReturn(*gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_1(timerWait, timerProcess)
   {
      if(!timerProcess.isProcess())
         return simTypeError(timerProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(timerInfo, timerProcess, TimerInfoPtr);

      if(timerInfo->mProcess->state.isArray())
      {
         messages::getCurrentContext();
         Value* elements = memory::allocateValues(2);
         elements[0] = *gStackTop;
         elements[1] = timerInfo->mProcess->state.array.elements[1];
         *gStackTop = Value::makeArray(2, elements);

         elements = memory::allocateValues(2);
         elements[0] = timerInfo->mProcess->state.array.elements[0];
         elements[1] = *gStackTop;
         *gStackTop = Value::makeArray(2, elements);

         memory::modifyProcessState(timerInfo->mProcess);
         timerInfo->mProcess->keepAlive = true;
         return Value::makeTask(Value::STOP);
      }
      else
         return simReturn(timerInfo->mProcess->state);
   }

   EVLAN_NATIVE_FUNCTION_1(timerCancel, timerProcess)
   {
      if(!timerProcess.isProcess())
         return simTypeError(timerProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(timerInfo, timerProcess, TimerInfoPtr);

      if(timerInfo->mProcess->state.isArray())
      {
         simPush(timerInfo->mProcess->state.array.elements[1]);

         while(!gStackTop->isEmpty())
         {
            simPush(gStackTop->array.elements[0]);
            simPush(simReturn(Value::makeBoolean(false)));
            simQueue();
            *gStackTop = gStackTop->array.elements[1];
         }
      }

      simPush(Value::makeBoolean(false));
      memory::modifyProcessState(timerInfo->mProcess);
      if(timerInfo->mMessage != null)
      {
         timerInfo->mMessage->disable();
         timerInfo->mMessage = null;
      }
      timerInfo->mProcess->keepAlive = false;

      return simReturn();
   }
}

// ===================================================================

} //namespace api
} //namespace evlan
