/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/12/12 (KV)
 *  + First Version
 */

// ===================================================================

#include "crypto.h"
#include "sha1.h"
#include <evlan/vm/native.h>
#include <evlan/vm/memory.h>
#include <evlan/vm/shortcuts.h>

#ifdef PLATFORM_WIN32
#include <windows.h>
#include <wincrypt.h>
#elif defined(PLATFORM_POSIX)
#include <unistd.h>
#include <fcntl.h>
#endif

namespace evlan {
namespace api {
using namespace evlan::vm;
using namespace evlan::vm::shortcuts;
using framework::common::byte;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::api::crypto
// TODO (KV): Review evlan::api::crypto

namespace crypto
{
   inline int bytesToWords(int i) {return (i + sizeof(word)-1) / sizeof(word);}

#ifdef PLATFORM_WIN32
   HCRYPTPROV acquireWincryptContext()
   {
      HCRYPTPROV handle;
      if(!CryptAcquireContext(&handle, NULL, NULL, PROV_RSA_FULL, 0))
         FW_THROW(spnew Win32Error($("CryptAcquireContext")));
      return handle;
   }

   EVLAN_NATIVE_FUNCTION_1(cryptoGetSecureRandom, size)
   {
      static HCRYPTPROV handle = acquireWincryptContext();

      if(!size.isInteger())
         return simTypeError(size);

      int byteCount = (int)size.builtin.value.scalar;
      byte* target = (byte*)memory::allocateWords(bytesToWords(byteCount));
      CryptGenRandom(handle, byteCount, target);

      return simReturn(Value::makeByteArray(byteCount, target));
   }

#elif defined(PLATFORM_POSIX)
   /* FIXME:  Should we just use urandom to seed some SPRNG and stir every
    *         now and then? */
   EVLAN_NATIVE_FUNCTION_1(cryptoGetSecureRandom, size)
   {
      static int urandom = open("/dev/urandom", O_RDONLY);

      if(!size.isInteger())
         return simTypeError(size);

      int byteCount = (int)size.builtin.value.scalar;
      byte* target = (byte*)memory::allocateWords(bytesToWords(byteCount));
      int count = 0;
      while(count < byteCount)
      {
         int result = read(urandom, target + count, byteCount - count);
         if(result < 0)
            FW_THROW(spnew ErrnoError($("/dev/urandom")));
         count += result;
      }

      return simReturn(Value::makeByteArray(byteCount, target));
   }

#else
# error "Port me."
#endif

   const int SHA1_CONTEXT_SIZE = bytesToWords(sizeof(SHA1Context));

   EVLAN_NATIVE_FUNCTION_0(cryptoSha1Start)
   {
      word* target = memory::allocateWords(SHA1_CONTEXT_SIZE);

      SHA1Reset((SHA1Context*)target);

      return Value::makePackedArray(
         Value::PACKED_DATA, core::LOG2_WORD_SIZE,
         SHA1_CONTEXT_SIZE, target);
   }

   EVLAN_NATIVE_FUNCTION_2(cryptoSha1Add, context, data)
   {
      if(!data.isByteArray()
      || !context.isPackedArray()
      || context.packedArray.packType != Value::PACKED_DATA
      || context.packedArray.elementSize != core::LOG2_WORD_SIZE
      || context.packedArray.offset != 0
      || context.packedArray.size != SHA1_CONTEXT_SIZE)
         return simTypeError(context);

      byte* bytes = (byte*)data.packedArray.elements + data.packedArray.offset;
      int size = data.packedArray.size;

      word* target = memory::allocateWords(SHA1_CONTEXT_SIZE);
      memcpy(target, context.packedArray.elements, sizeof(SHA1Context));

      SHA1Input((SHA1Context*)target, bytes, size);

      return Value::makePackedArray(
         Value::PACKED_DATA, core::LOG2_WORD_SIZE,
         SHA1_CONTEXT_SIZE, target);
   }

   EVLAN_NATIVE_FUNCTION_1(cryptoSha1Finish, context)
   {
      if(!context.isPackedArray()
      || context.packedArray.packType != Value::PACKED_DATA
      || context.packedArray.elementSize != core::LOG2_WORD_SIZE
      || context.packedArray.offset != 0
      || context.packedArray.size != SHA1_CONTEXT_SIZE)
         return simTypeError(context);

      byte* target = (byte*)memory::allocateWords(bytesToWords(SHA1HashSize));

      SHA1Context sha1Context;
      memcpy(&sha1Context, context.packedArray.elements, sizeof(SHA1Context));
      SHA1Result(&sha1Context, target);

      return Value::makeByteArray(SHA1HashSize, target);
   }
}

// ===================================================================

} //namespace api
} //namespace evlan
