/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/03/18 (KV)
 *  + First Version
 */

// ===================================================================

#include "internal.h"
#include <evlan/utf8.h>
#include <evlan/vm/memory.h>
#include <evlan/vm/operations.h>
#include <evlan/vm/arrays.h>
#include <evlan/vm/simulator.h>
#include <evlan/vm/messages.h>
#include <evlan/vm/native.h>
#include <evlan/vm/loader.h>
#include <evlan/vm/shortcuts.h>
#include <math.h>       //don't use <cmath>; it's broken on several platforms
#include <time.h>

namespace evlan {
namespace api {
using namespace vm;
using namespace vm::shortcuts;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::vm::internal
// TODO (KV): Review evlan::vm::internal

/* I appologize for this module.  It really should be split into a
 * number of smaller modules, but unfortunately it didn't evolve that
 * way, and I'm too lazy to split it up now. */
namespace internal
{
   core::Value runtimeError(core::Value::ErrorCode code, core::Value& param1)
   {
      if(param1.isError())
         return param1;
      return core::Value::makeError(code, null);
   }
   core::Value runtimeError(core::Value::ErrorCode code, core::Value& param1, core::Value& param2)
   {
      if(param1.isError())
         return param1;
      if(param2.isError())
         return param2;
      return core::Value::makeError(code, null);
   }
   core::Value runtimeError(core::Value::ErrorCode code, core::Value& param1, core::Value& param2, core::Value& param3)
   {
      if(param1.isError())
         return param1;
      if(param2.isError())
         return param2;
      if(param3.isError())
         return param3;
      return core::Value::makeError(code, null);
   }
   core::Value runtimeError(core::Value::ErrorCode code, core::Value& param1, core::Value& param2, core::Value& param3, core::Value& param4)
   {
      if(param1.isError())
         return param1;
      if(param2.isError())
         return param2;
      if(param3.isError())
         return param3;
      if(param4.isError())
         return param4;
      return core::Value::makeError(code, null);
   }

#define SIMPLE_MATH_FUNCTION(NAME, FUNCTION)                               \
   EVLAN_NATIVE_FUNCTION_1(NAME, x)                                        \
   {                                                                       \
      if(x.isScalar())                                                     \
         return core::Value::makeScalar(FUNCTION(x.builtin.value.scalar)); \
      else                                                                 \
         return runtimeError(core::Value::TYPE_ERROR, x);                  \
   }

   // fp impl ========================================================
   EVLAN_NATIVE_FUNCTION_1(isNan, x)
   {
      if(x.isScalar())
         return core::Value::makeBoolean(isnan(x.builtin.value.scalar));
      else
         return runtimeError(core::Value::TYPE_ERROR, x);
   }

   EVLAN_NATIVE_FUNCTION_1(getMantissaAndExponent, x)
   {
      if(x.isScalar())
      {
         int exponent;
         double mantissa = ::frexp(x.builtin.value.scalar, &exponent);

         core::Value* members = memory::allocateValues(2);
         members[0] = core::Value::makeScalar(mantissa);
         members[1] = core::Value::makeScalar(exponent);

         return core::Value::makeArray(2, members);
      }
      else
         return runtimeError(core::Value::TYPE_ERROR, x);
   }

   // rounding =======================================================
   SIMPLE_MATH_FUNCTION(floor, ::floor)
   SIMPLE_MATH_FUNCTION(ceil, ::ceil)

   EVLAN_NATIVE_FUNCTION_1(round, x)
   {
      if(x.isScalar())
         return core::Value::makeScalar(::floor(x.builtin.value.scalar + 0.5));
      else
         return runtimeError(core::Value::TYPE_ERROR, x);
   }

   // exponents ======================================================
   SIMPLE_MATH_FUNCTION(ln, ::log)
   SIMPLE_MATH_FUNCTION(exp, ::exp)
   SIMPLE_MATH_FUNCTION(sqrt, ::sqrt)

   EVLAN_NATIVE_FUNCTION_1(lg, x)
   {
      if(x.isScalar())
      {
         int intval = (int)x.builtin.value.scalar;

         if(x.builtin.value.scalar == (double)intval
         && intval > 0)
         {
            int result = 0;

            while(intval % 2 == 0)
            {
               intval >>= 1;
               ++result;
            }

            if(intval == 1)
               return core::Value::makeScalar(result);
         }

         return core::Value::makeScalar(
            ::log(x.builtin.value.scalar) *
            1.4426950408889634073599246810019);
         //That number = 1 / ln(2.0)
      }
      else
         return runtimeError(core::Value::TYPE_ERROR, x);
   }

   EVLAN_NATIVE_FUNCTION_2(log, x, base)
   {
      if(x.isScalar() && base.isScalar())
      {
         int intval1 = (int)x.builtin.value.scalar;
         int intval0 = (int)base.builtin.value.scalar;

         if(x.builtin.value.scalar == (double)intval1
         && intval1 > 0
         && base.builtin.value.scalar == (double)intval0
         && intval0 > 0)
         {
            //for the sake of exactness, and possibly optimization...
            int base = intval0;
            if(base == 2)
            {
               int result = 0;

               while(intval1 % 2 == 0)
               {
                  intval1 >>= 1;
                  ++result;
               }

               if(intval1 == 1)
                  return core::Value::makeScalar(result);
            }
            else
            {
               int i = intval1;
               int result = 0;

               while(i % base == 0)
               {
                  i /= base;
                  ++result;
               }

               if(i == 1)
                  return core::Value::makeScalar(result);
            }
         }

         return core::Value::makeScalar(
            ::log(x.builtin.value.scalar) /
            ::log(base.builtin.value.scalar));
      }
      else
         return runtimeError(core::Value::TYPE_ERROR, x, base);
   }

   // trigenometry ===================================================
   SIMPLE_MATH_FUNCTION(sin, ::sin)
   SIMPLE_MATH_FUNCTION(cos, ::cos)
   SIMPLE_MATH_FUNCTION(tan, ::tan)
   SIMPLE_MATH_FUNCTION(arcsin, ::asin)
   SIMPLE_MATH_FUNCTION(arccos, ::acos)
   SIMPLE_MATH_FUNCTION(arctan, ::atan)
   SIMPLE_MATH_FUNCTION(sinh, ::sinh)
   SIMPLE_MATH_FUNCTION(cosh, ::cosh)
   SIMPLE_MATH_FUNCTION(tanh, ::tanh)

   EVLAN_NATIVE_FUNCTION_2(coordinatesToAngle, x, y)
   {
      if(x.isScalar() && y.isScalar())
      {
         return core::Value::makeScalar(atan2(
            y.builtin.value.scalar,
            x.builtin.value.scalar));
      }
      else
         return runtimeError(core::Value::TYPE_ERROR, x, y);
   }

   // array ops ======================================================
   const int ARRAY_BLOCK_SIZE = 65536 / sizeof(core::Value);

   /* I asked Kass whether I should use "first"/"rest" or
    * "head"/"tail" for the names of the fields of a list node, and
    * she chose the former. */
   EVLAN_NATIVE_IDENTIFIER(gIsEmptyId, isEmpty);
   EVLAN_NATIVE_IDENTIFIER(gFirstId,   first);
   EVLAN_NATIVE_IDENTIFIER(gRestId,    rest);

   EVLAN_NATIVE_FUNCTION_2(makeArray, sizeParam, makeElement)
   {
      if(sizeParam.isScalar() && makeElement.isFunction())
      {
         int size = (int)sizeParam.builtin.value.scalar;

         if(sizeParam.builtin.value.scalar == (double)size
         && size >= 0)
         {
            if(size == 0)
               return core::Value::makeArray(0, null);

            int blockSize = 0;
            for(int i = 0; i < size; i++, blockSize++)
            {
               if(blockSize == ARRAY_BLOCK_SIZE)
               {
                  core::Value* elements = memory::allocateValues(ARRAY_BLOCK_SIZE);

                  for(int j = 0; j < ARRAY_BLOCK_SIZE; j++)
                     elements[j] = core::gStackTop[ARRAY_BLOCK_SIZE - 1 - j];

                  core::gStackTop += ARRAY_BLOCK_SIZE - 1;
                  *core::gStackTop = core::Value::makeArray(ARRAY_BLOCK_SIZE, elements);

                  blockSize = 1;
               }

               core::gStackTop -= 2;
               core::gStackTop[1] = makeElement;
               core::gStackTop[0] = core::Value::makeScalar(i);
               simulator::apply(1);
            }

            core::Value* elements = memory::allocateValues(size);

            for(int j = 1; j < blockSize; j++)
               elements[size - j] = core::gStackTop[j - 1];

            core::gStackTop += blockSize - 1;
            int remaining = size - blockSize + 1;

            while(remaining > 1)
            {
               remaining -= ARRAY_BLOCK_SIZE - 1;
               memcpy(elements + remaining,
                      core::gStackTop->array.elements + 1,
                      (ARRAY_BLOCK_SIZE - 1) * sizeof(core::Value));
               *core::gStackTop = core::gStackTop->array.elements[0];
            }

            elements[0] = *core::gStackTop;

            *core::gStackTop = core::Value::makeArray(size, elements);

            arrays::tryPackArray();

            return *core::gStackTop;
         }
      }

      return runtimeError(core::Value::TYPE_ERROR, sizeParam, makeElement);
   }

   EVLAN_NATIVE_FUNCTION_1(reverseArray, array)
   {
      if(array.isArray())
      {
         int size = array.array.size;
         core::Value* newElements = memory::allocateValues(size);
         core::Value* oldElements = array.array.elements;

         for(int i = 0; i < size; i++)
            newElements[size - i - 1] = oldElements[i];

         return core::Value::makeArray(size, newElements);
      }
      else if(array.isPackedArray())
      {
         int size = array.packedArray.size;
         int elementSize = array.packedArray.elementSize;
         int offset = array.packedArray.offset;

         int wordSize;
         if(elementSize > core::LOG2_WORD_SIZE)
            wordSize = size << (elementSize - core::LOG2_WORD_SIZE);
         else if(elementSize == core::LOG2_WORD_SIZE)
            wordSize = size;
         else
         {
            int diff = (core::LOG2_WORD_SIZE - elementSize);
            wordSize = size >> diff;
            if((size & ((1 << diff) - 1)) != 0)
               ++wordSize;
         }

         word* newElements = memory::allocateWords(wordSize);
         word* oldElements = array.packedArray.elements;

         switch(elementSize)
         {
         case 0:
            for(int i = 0; i < size; i++)
            {
               int j = size-i-1;
               int k = i+offset;
               ((byte*)newElements)[j/8] |= ((((byte*)oldElements)[k/8] >> (7-k%8)) & 1) << (7-j%8);
            }
            break;
         case 1:
            for(int i = 0; i < size; i++)
            {
               int j = size-i-1;
               int k = i+offset;
               ((byte*)newElements)[j/4] |= ((((byte*)oldElements)[k/4] >> ((3-k%4)*2)) & 3) << ((3-j%4)*2);
            }
            break;
         case 2:
            for(int i = 0; i < size; i++)
            {
               int j = size-i-1;
               int k = i+offset;
               ((byte*)newElements)[j/2] |= ((((byte*)oldElements)[k/2] >> ((1-k%2)*4)) & 15) << ((1-j%2)*4);
            }
            break;
         case 3:
            for(int i = 0; i < size; i++)
               ((byte*)newElements)[size-i-1] = ((byte*)oldElements)[i+offset];
            break;
         case 4:
            for(int i = 0; i < size; i++)
               ((uint16*)newElements)[size-i-1] = ((uint16*)oldElements)[i+offset];
            break;
         case 5:
            for(int i = 0; i < size; i++)
               ((uint32*)newElements)[size-i-1] = ((uint32*)oldElements)[i];
            break;
         case 6:
            for(int i = 0; i < size; i++)
               ((uint64*)newElements)[size-i-1] = ((uint64*)oldElements)[i];
            break;
         }

         core::Value result = array;
         result.packedArray.elements = newElements;
         result.packedArray.offset = 0;
         return result;
      }
      else
         return runtimeError(core::Value::TYPE_ERROR, array);
   }

   EVLAN_NATIVE_FUNCTION_2(makeArrayFromList, list, reverseParam)
   {
      if(!reverseParam.isBoolean())
         return runtimeError(core::Value::TYPE_ERROR, list, reverseParam);

      bool reverse = reverseParam.builtin.value.boolean;

      --core::gStackTop;
      *core::gStackTop = list;

      int size = 0;
      while(true)
      {
         //get isEmpty
         --core::gStackTop;
         *core::gStackTop = core::gStackTop[1];
         operations::getMember(gIsEmptyId);
         if(core::gStackTop->isError())
            return *core::gStackTop;
         else if(!core::gStackTop->isBoolean())
            return core::Value::makeError(core::Value::TYPE_ERROR, null);

         //at end?
         if(core::gStackTop->builtin.value.boolean)
            break;
         ++size;
         ++core::gStackTop;

         //get tail
         operations::getMember(gRestId);
      }

      ++core::gStackTop;

      core::Value* elements = memory::allocateValues(size);

      *core::gStackTop = list;
      for(int i = 0; i < size; i++)
      {
         //get head
         --core::gStackTop;
         *core::gStackTop = core::gStackTop[1];
         operations::getMember(gFirstId);

         //store it to array
         if(reverse)
            elements[size - i - 1] = *core::gStackTop;
         else
            elements[i] = *core::gStackTop;
         ++core::gStackTop;

         //get tail
         operations::getMember(gRestId);
      }

      *core::gStackTop = core::Value::makeArray(size, elements);
      arrays::tryPackArray();
      return *core::gStackTop;
   }

   EVLAN_NATIVE_FUNCTION_3(getInterval, array, startParam, endParam)
   {
      if((!array.isArray()
       && !array.isPackedArray())
      || !startParam.isScalar()
      || !endParam.isScalar())
         return runtimeError(core::Value::TYPE_ERROR, array, startParam, endParam);

      int start = (int)startParam.builtin.value.scalar;
      int end = (int)endParam.builtin.value.scalar;

      if(startParam.builtin.value.scalar != (double)start
      || endParam.builtin.value.scalar != (double)end)
         return runtimeError(core::Value::TYPE_ERROR, array, startParam, endParam);

      int size;
      if(array.isArray())
         size = array.array.size;
      else
         size = array.packedArray.size;

      if(start < 0 || start > size
      || end < 0   || end > size
      || end < start)
         return core::Value::makeError(core::Value::OUT_OF_BOUNDS, null);

      core::Value result = array;
      if(result.isArray())
      {
         result.array.size = end - start;
         result.array.elements += start;
      }
      else
      {
         result.packedArray.size = end - start;

         if(result.packedArray.elementSize < core::LOG2_WORD_SIZE)
         {
            start += result.packedArray.offset;
            int diff = core::LOG2_WORD_SIZE - result.packedArray.elementSize;
            result.packedArray.elements += start >> diff;
            result.packedArray.offset = start & ((1 << diff) - 1);
         }
         else if(result.packedArray.elementSize == core::LOG2_WORD_SIZE)
            result.packedArray.elements += start;
         else
         {
            result.packedArray.elements +=
               start << (result.packedArray.elementSize - core::LOG2_WORD_SIZE);
         }
      }

      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(joinArrays, arraysParam)
   {
      if(!arraysParam.isArray())
         return runtimeError(core::Value::TYPE_ERROR, arraysParam);

      int totalSize = 0;
      int packType = -1;
      int elementSize = 0;
      core::Value* arrays = arraysParam.array.elements;
      int count = arraysParam.array.size;
      int nonEmptyIndex = -1;
      for(int i = 0; i < count; i++)
      {
         if(arrays[i].isArray())
         {
            if(arrays[i].array.size > 0)
            {
               if(nonEmptyIndex == -1)
                  nonEmptyIndex = i;
               else
                  nonEmptyIndex = -2;
               totalSize += arrays[i].array.size;
            }
            packType = -2;
         }
         else if(arrays[i].isPackedArray())
         {
            if(arrays[i].packedArray.size > 0)
            {
               if(nonEmptyIndex == -1)
                  nonEmptyIndex = i;
               else
                  nonEmptyIndex = -2;
               totalSize += arrays[i].packedArray.size;
            }

            if(packType == -1)
            {
               packType = arrays[i].packedArray.packType;
               elementSize = arrays[i].packedArray.elementSize;
            }
            else if(packType != arrays[i].packedArray.packType
                 || elementSize != arrays[i].packedArray.elementSize)
               packType = -2;
         }
         else if(arrays[i].isError())
            return arrays[i];
         else
            return core::Value::makeError(core::Value::TYPE_ERROR, null);
      }

      //avoid copies when unnecessary
      if(nonEmptyIndex == -1)
         return core::Value::makeArray(0, null);
      else if(nonEmptyIndex != -2)
         return arraysParam.array.elements[nonEmptyIndex];

      if(packType >= 0 && elementSize >= 3)
      {
         int wordSize;
         if(elementSize > core::LOG2_WORD_SIZE)
            wordSize = totalSize << (elementSize - core::LOG2_WORD_SIZE);
         else if(elementSize == core::LOG2_WORD_SIZE)
            wordSize = totalSize;
         else
         {
            int diff = (core::LOG2_WORD_SIZE - elementSize);
            wordSize = totalSize >> diff;
            if((totalSize & ((1 << diff) - 1)) != 0)
               ++wordSize;
         }

         word* target = memory::allocateWords(wordSize);
         totalSize = 0;
         arrays = arraysParam.array.elements; //may have changed

         switch(elementSize)
         {
         case 3:
            for(int i = 0; i < count; i++)
            {
               memcpy((byte*)target + totalSize,
                      (byte*)arrays[i].packedArray.elements + arrays[i].packedArray.offset,
                      arrays[i].packedArray.size);
               totalSize += arrays[i].packedArray.size;
            }
            break;
         case 4:
            for(int i = 0; i < count; i++)
            {
               memcpy((uint16*)target + totalSize,
                      (uint16*)arrays[i].packedArray.elements + arrays[i].packedArray.offset,
                      arrays[i].packedArray.size * sizeof(uint16));
               totalSize += arrays[i].packedArray.size;
            }
            break;
         case 5:
            for(int i = 0; i < count; i++)
            {
               memcpy((uint32*)target + totalSize,
                      (uint32*)arrays[i].packedArray.elements + arrays[i].packedArray.offset,
                      arrays[i].packedArray.size * sizeof(uint32));
               totalSize += arrays[i].packedArray.size;
            }
            break;
         case 6:
            for(int i = 0; i < count; i++)
            {
               memcpy((uint64*)target + totalSize,
                      (uint64*)arrays[i].packedArray.elements + arrays[i].packedArray.offset,
                      arrays[i].packedArray.size * sizeof(uint64));
               totalSize += arrays[i].packedArray.size;
            }
            break;
         }

         core::Value result;
         result.type = core::Value::PACKED_ARRAY;
         result.packedArray.packType = packType;
         result.packedArray.elementSize = elementSize;
         result.packedArray.offset = 0;
         result.packedArray.size = totalSize;
         result.packedArray.elements = target;
         return result;
      }
      else
      {
         core::Value* target = memory::allocateValues(totalSize);
         totalSize = 0;
         arrays = arraysParam.array.elements; //may have changed

         --core::gStackTop;
         for(int i = 0; i < count; i++)
         {
            int subSize;
            if(arrays[i].isArray())
               subSize = arrays[i].array.size;
            else
               subSize = arrays[i].packedArray.size;

            for(int j = 0; j < subSize; j++)
            {
               *core::gStackTop = arrays[i];
               arrays::getElement(j);
               target[totalSize++] = *core::gStackTop;
            }
         }

         *core::gStackTop = core::Value::makeArray(totalSize, target);
         arrays::tryPackArray();
         return *core::gStackTop;
      }
   }

   EVLAN_NATIVE_FUNCTION_3(pivotArray, array, groupCountParam, groupFunc)
   {
      if(!groupCountParam.isScalar()
      || !groupFunc.isFunction())
         return runtimeError(core::Value::TYPE_ERROR, array, groupCountParam, groupFunc);

      int groupCount = (int)groupCountParam.builtin.value.scalar;
      if(groupCountParam.builtin.value.scalar != (double)groupCount
      || groupCount <= 0)
         return runtimeError(core::Value::TYPE_ERROR, array, groupCountParam, groupFunc);

      int size;
      if(array.isArray())
         size = array.array.size;
      else if(array.isPackedArray())
         size = array.packedArray.size;
      else
         return runtimeError(core::Value::TYPE_ERROR, array, groupCountParam, groupFunc);

      Array<int> groups(size);
      int groupCounts[groupCount];
      for(int i = 0; i < groupCount; i++)
         groupCounts[i] = 0;

      --core::gStackTop;
      for(int i = 0; i < size; i++)
      {
         *core::gStackTop = groupFunc;
         --core::gStackTop;
         *core::gStackTop = Value::makeScalar(i);
         simulator::apply(1);
         if(!core::gStackTop->isScalar())
         {
            if(core::gStackTop->isError())
               return *core::gStackTop;
            else
               return core::Value::makeError(core::Value::TYPE_ERROR, null);
         }
         int group = (int)core::gStackTop->builtin.value.scalar;
         if(core::gStackTop->builtin.value.scalar != (double)group
         || group < 0 || group >= groupCount)
            return core::Value::makeError(core::Value::OUT_OF_BOUNDS, null);
         groups[i] = group;
         ++groupCounts[group];
      }

      int groupOffsets[groupCount];
      groupOffsets[0] = 0;
      for(int i = 1; i < groupCount; i++)
         groupOffsets[i] = groupOffsets[i-1] + groupCounts[i-1];

      core::Value* elements = memory::allocateValues(size + groupCount);

      for(int i = 0; i < size; i++)
      {
         *core::gStackTop = array;
         arrays::getElement(i);
         elements[groupOffsets[groups[i]]++] = *core::gStackTop;
      }


      simPush(Value::makeArray(groupCounts[0], elements));
      arrays::tryPackArray();
      elements[size] = *gStackTop;
      simPop();

      for(int i = 1; i < groupCount; i++)
      {
         simPush(Value::makeArray(groupCounts[i], elements + groupOffsets[i-1]));
         arrays::tryPackArray();
         elements[size+i] = *gStackTop;
         simPop();
      }

      return core::Value::makeArray(groupCount, elements + size);
   }

   // data ops =======================================================
   EVLAN_NATIVE_FUNCTION_2(bitwiseAnd, a, b)
   {
      if(a.isData()
      && b.isBuiltin()
      && a.builtin.builtinType == b.builtin.builtinType)
      {
         core::Value result = a;
         result.builtin.value.data64 &= b.builtin.value.data64;
         return result;
      }
      else if(a.isLargeData()
           && b.isLargeData()
           && a.largeData.size == b.largeData.size)
      {
         FW_THROW(spnew FatalError($("large data not implemented")));
      }
      else
         return runtimeError(core::Value::TYPE_ERROR, a, b);
   }

   EVLAN_NATIVE_FUNCTION_2(bitwiseOr, a, b)
   {
      if(a.isData()
      && b.isBuiltin()
      && a.builtin.builtinType == b.builtin.builtinType)
      {
         core::Value result = a;
         result.builtin.value.data64 |= b.builtin.value.data64;
         return result;
      }
      else if(a.isLargeData()
           && b.isLargeData()
           && a.largeData.size == b.largeData.size)
      {
         FW_THROW(spnew FatalError($("large data not implemented")));
      }
      else
         return runtimeError(core::Value::TYPE_ERROR, a, b);
   }

   EVLAN_NATIVE_FUNCTION_2(bitwiseXor, a, b)
   {
      if(a.isData()
      && b.isBuiltin()
      && a.builtin.builtinType == b.builtin.builtinType)
      {
         core::Value result = a;
         result.builtin.value.data64 ^= b.builtin.value.data64;
         return result;
      }
      else if(a.isLargeData()
           && b.isLargeData()
           && a.largeData.size == b.largeData.size)
      {
         FW_THROW(spnew FatalError($("large data not implemented")));
      }
      else
         return runtimeError(core::Value::TYPE_ERROR, a, b);
   }

   EVLAN_NATIVE_FUNCTION_1(bitwiseNot, data)
   {
      if(data.isData())
      {
         core::Value result = data;
         result.builtin.value.data64 = ~result.builtin.value.data64;
         return result;
      }
      else if(data.isLargeData())
         FW_THROW(spnew FatalError($("large data not implemented")));
      else
         return runtimeError(core::Value::TYPE_ERROR, data);
   }

   EVLAN_NATIVE_FUNCTION_2(shiftBits, data, offsetParam)
   {
      if(data.isData()
      && offsetParam.isScalar())
      {
         int offset = (int)offsetParam.builtin.value.scalar;
         if(offsetParam.builtin.value.scalar != (double)offset)
            return core::Value::makeError(core::Value::TYPE_ERROR, null);

         core::Value result = data;
         int size = result.builtin.builtinType - core::Value::DATA1;
         int bitSize = 1 << size;
         if(offset >= bitSize || offset <= -bitSize)
            result.builtin.value.data64 = 0;
         else if(size < 4)
         {
            uint32 mask = (1 << bitSize) - 1;
            result.builtin.value.data32 &= mask;
            if(offset > 0)
               result.builtin.value.data32 <<= offset;
            else
               result.builtin.value.data32 >>= -offset;
         }
         else if(size == 4)
         {
            if(offset > 0)
               result.builtin.value.data32 <<= offset;
            else
               result.builtin.value.data32 >>= -offset;
         }
         else
         {
            if(offset > 0)
               result.builtin.value.data64 <<= offset;
            else
               result.builtin.value.data64 >>= -offset;
         }

         return result;
      }
      else if(data.isLargeData())
         FW_THROW(spnew FatalError($("large data not implemented")));
      else
         return runtimeError(core::Value::TYPE_ERROR, data, offsetParam);
   }

   EVLAN_NATIVE_FUNCTION_3(formatData, dataArray, newElementSizeParam, littleEndian)
   {
      //Good luck figuring out how this one works!

      if(dataArray.isArray() && dataArray.array.size == 0)
         return dataArray;

      if(!dataArray.isPackedArray()
      || dataArray.packedArray.packType != core::Value::PACKED_DATA
      || !newElementSizeParam.isScalar()
      || !littleEndian.isBoolean())
         return runtimeError(core::Value::TYPE_ERROR, dataArray, newElementSizeParam);

      int newElementBits = (int)newElementSizeParam.builtin.value.scalar;
      if(newElementSizeParam.builtin.value.scalar != (double)newElementBits)
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      int newElementSize = 0;
      while(newElementBits > (1 << newElementSize))
         ++newElementSize;
      if(newElementBits != (1 << newElementSize))
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      int size = dataArray.packedArray.size;
      int elementSize = dataArray.packedArray.elementSize;

      if(elementSize == newElementSize)
         return dataArray;

      int low, high, newSize;
      if(elementSize < newElementSize)
      {
         low = elementSize;
         high = newElementSize;
         newSize = size >> (newElementSize - elementSize);

         //size evenly divisible into new elements?
         if((size & ((1 << (newElementSize - elementSize)) - 1)) != 0)
            return core::Value::makeError(core::Value::OUT_OF_BOUNDS, null);
      }
      else
      {
         low = newElementSize;
         high = elementSize;
         newSize = size << (elementSize - newElementSize);
      }
      if(low < 3)
         low = 3;
      if(low > 6)
         low = 6;
      if(high < 3)
         high = 3;
      if(high > 6)
         high = 6;

      int swapBits = 0;
#ifdef PLATFORM_BIG_ENDIAN
      if(littleEndian.builtin.value.boolean)
      {
         for(int i = low; i < high; i++)
            swapBits = swapBits | (1 << (i - 3));
      }
#else
      if(!littleEndian.builtin.value.boolean)
      {
         for(int i = low; i < high; i++)
            swapBits = swapBits | (1 << (i - 3));
      }
#endif

      byte* source = (byte*)dataArray.packedArray.elements;
      int offset = dataArray.packedArray.offset;
      if(elementSize < 3)
      {
         source += offset >> (3 - elementSize);
         offset = (offset << elementSize) & 7;
      }
      else
      {
         source += offset << (elementSize - 3);
         offset = 0;
      }

      if(offset == 0 && swapBits == 0)
      {
         core::Value result = dataArray;
         result.packedArray.elementSize = newElementSize;
         result.packedArray.size = newSize;
         return result;
      }

      int wordSize;
      if(elementSize > core::LOG2_WORD_SIZE)
         wordSize = size << (elementSize - core::LOG2_WORD_SIZE);
      else if(elementSize == core::LOG2_WORD_SIZE)
         wordSize = size;
      else
      {
         int diff = (core::LOG2_WORD_SIZE - elementSize);
         wordSize = size >> diff;
         if((size & ((1 << diff) - 1)) != 0)
            ++wordSize;
      }

      byte* target = (byte*)memory::allocateWords(wordSize);
      source = (byte*)dataArray.packedArray.elements;
      int byteSize = wordSize * sizeof(word);

      if(offset == 0)
      {
         for(int i = 0; i < byteSize; i++)
            target[i^swapBits] = source[i];
      }
      else
      {
         int mask = (1 << offset) - 1;
         for(int i = 0; i < byteSize; i++)
         {
            target[i^swapBits] =
               ((source[i]   << offset      ) & ~mask) |
               ((source[i+1] >> (8 - offset)) &  mask);
         }
      }

      core::Value result;
      result.type = core::Value::PACKED_ARRAY;
      result.packedArray.packType = core::Value::PACKED_DATA;
      result.packedArray.elementSize = newElementSize;
      result.packedArray.offset = 0;
      result.packedArray.size = newSize;
      result.packedArray.elements = (word*)target;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_2(repackData, dataArray, newElementSizeParam)
   {
      if(dataArray.isArray() && dataArray.array.size == 0)
         return dataArray;

      if(!dataArray.isPackedArray()
      || !newElementSizeParam.isScalar())
         return runtimeError(core::Value::TYPE_ERROR, dataArray, newElementSizeParam);

      int newElementBits = (int)newElementSizeParam.builtin.value.scalar;
      if(newElementSizeParam.builtin.value.scalar != (double)newElementBits)
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      int newElementSize = 0;
      while(newElementBits > (1 << newElementSize))
         ++newElementSize;
      if(newElementBits != (1 << newElementSize))
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      int size = dataArray.packedArray.size;
      int elementSize = dataArray.packedArray.elementSize;
      int offset = dataArray.packedArray.offset;

      if(newElementSize == elementSize)
         return dataArray;

      int wordSize;
      if(newElementSize < core::LOG2_WORD_SIZE)
      {
         int diff = core::LOG2_WORD_SIZE - newElementSize;
         wordSize = size >> diff;
         if((size & ((1 << diff) - 1)) != 0)
            ++wordSize;
      }
      else if(newElementSize == core::LOG2_WORD_SIZE)
         wordSize = size;
      else
         wordSize = size << (newElementSize - core::LOG2_WORD_SIZE);

      word* target = memory::allocateWords(wordSize);
      word* source = dataArray.packedArray.elements;

      switch(elementSize | (newElementSize << 4))
      {
      case (3 | (4 << 4)):
         for(int i = 0; i < size; i++)
            ((uint16*)target)[i] = ((uint8*)source)[i + offset];
         break;
      case (3 | (5 << 4)):
         for(int i = 0; i < size; i++)
            ((uint32*)target)[i] = ((uint8*)source)[i + offset];
         break;
      case (3 | (6 << 4)):
         for(int i = 0; i < size; i++)
            ((uint64*)target)[i] = ((uint8*)source)[i + offset];
         break;
      case (4 | (3 << 4)):
         for(int i = 0; i < size; i++)
            ((uint8*)target)[i] = ((uint16*)source)[i + offset];
         break;
      case (4 | (5 << 4)):
         for(int i = 0; i < size; i++)
            ((uint32*)target)[i] = ((uint16*)source)[i + offset];
         break;
      case (4 | (6 << 4)):
         for(int i = 0; i < size; i++)
            ((uint64*)target)[i] = ((uint16*)source)[i + offset];
         break;
      case (5 | (3 << 4)):
         for(int i = 0; i < size; i++)
            ((uint8*)target)[i] = ((uint32*)source)[i + offset];
         break;
      case (5 | (4 << 4)):
         for(int i = 0; i < size; i++)
            ((uint16*)target)[i] = ((uint32*)source)[i + offset];
         break;
      case (5 | (6 << 4)):
         for(int i = 0; i < size; i++)
            ((uint64*)target)[i] = ((uint32*)source)[i + offset];
         break;
      case (6 | (3 << 4)):
         for(int i = 0; i < size; i++)
            ((uint8*)target)[i] = ((uint64*)source)[i + offset];
         break;
      case (6 | (4 << 4)):
         for(int i = 0; i < size; i++)
            ((uint16*)target)[i] = ((uint64*)source)[i + offset];
         break;
      case (6 | (5 << 4)):
         for(int i = 0; i < size; i++)
            ((uint32*)target)[i] = ((uint64*)source)[i + offset];
         break;
      default:
         return core::Value::makeError(core::Value::TYPE_ERROR, null);
      }

      core::Value result = dataArray;
      result.packedArray.offset = 0;
      result.packedArray.elementSize = newElementSize;
      result.packedArray.elements = target;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(isPacked, array)
   {
      if(array.isPackedArray())
         return core::Value::makeBoolean(true);
      else if(array.isArray())
         return core::Value::makeBoolean(false);
      else
         return runtimeError(core::Value::TYPE_ERROR, array);
   }

   EVLAN_NATIVE_FUNCTION_1(getPackedElementSize, array)
   {
      if(array.isPackedArray())
         return core::Value::makeScalar(1 << array.packedArray.elementSize);
      else
         return runtimeError(core::Value::TYPE_ERROR, array);
   }

   EVLAN_NATIVE_FUNCTION_1(isPackedAsInteger, array)
   {
      if(array.isPackedArray())
         return core::Value::makeBoolean(array.packedArray.packType == core::Value::PACKED_INTEGER);
      else
         return runtimeError(core::Value::TYPE_ERROR, array);
   }

   EVLAN_NATIVE_IDENTIFIER(gBooleanAtomId,   boolean);
   EVLAN_NATIVE_IDENTIFIER(gScalarAtomId,    scalar);
   EVLAN_NATIVE_IDENTIFIER(gIntegerAtomId,   integer);
   EVLAN_NATIVE_IDENTIFIER(gCharacterAtomId, character);
   EVLAN_NATIVE_IDENTIFIER(gAtomAtomId,      atom);
   EVLAN_NATIVE_IDENTIFIER(gDataAtomId,      data);

   EVLAN_NATIVE_FUNCTION_2(setPackedDataType, array, type)
   {
      if(array.isArray() && array.array.size == 0)
         return array;

      if(!array.isPackedArray()
      || !type.isAtom())
         return runtimeError(core::Value::TYPE_ERROR, array, type);
      else
      {
         core::Value result = array;
         if(type.builtin.value.data32 == gBooleanAtomId)
            result.packedArray.packType = core::Value::PACKED_BOOLEAN;
         else if(type.builtin.value.data32 == gScalarAtomId)
            result.packedArray.packType = core::Value::PACKED_SCALAR;
         else if(type.builtin.value.data32 == gIntegerAtomId)
            result.packedArray.packType = core::Value::PACKED_INTEGER;
         else if(type.builtin.value.data32 == gCharacterAtomId)
            result.packedArray.packType = core::Value::PACKED_CHARACTER;
         else if(type.builtin.value.data32 == gAtomAtomId)
            result.packedArray.packType = core::Value::PACKED_ATOM;
         else if(type.builtin.value.data32 == gDataAtomId)
            result.packedArray.packType = core::Value::PACKED_DATA;
         else
            return core::Value::makeError(core::Value::TYPE_ERROR, null);
         return result;
      }
   }

   EVLAN_NATIVE_FUNCTION_1(getPackedDataType, array)
   {
      if(!array.isPackedArray())
         return runtimeError(core::Value::TYPE_ERROR, array);
      else
      {
         switch(array.packedArray.packType)
         {
         case core::Value::PACKED_BOOLEAN:
            return core::Value::makeAtom(gBooleanAtomId);
         case core::Value::PACKED_SCALAR:
            return core::Value::makeAtom(gScalarAtomId);
         case core::Value::PACKED_INTEGER:
            return core::Value::makeAtom(gIntegerAtomId);
         case core::Value::PACKED_CHARACTER:
            return core::Value::makeAtom(gCharacterAtomId);
         case core::Value::PACKED_ATOM:
            return core::Value::makeAtom(gAtomAtomId);
         default:
            return core::Value::makeAtom(gDataAtomId);
         }
      }
   }

   EVLAN_NATIVE_FUNCTION_2(setBuiltinType, value, type)
   {
      if(!value.isBuiltin()
      || !type.isAtom())
         return runtimeError(core::Value::TYPE_ERROR, value, type);
      else
      {
         core::Value result = value;
         if(result.builtin.builtinType >= core::Value::DATA1
         && result.builtin.builtinType < core::Value::DATA32)
         {
            int elementSize = result.builtin.builtinType - core::Value::DATA1;
            result.builtin.value.data32 &= (1 << (1 << elementSize)) - 1;
         }

         if(type.builtin.value.data32 == gBooleanAtomId)
            result.builtin.builtinType = core::Value::BOOLEAN;
         else if(type.builtin.value.data32 == gScalarAtomId)
            result.builtin.builtinType = core::Value::SCALAR;
         else if(type.builtin.value.data32 == gCharacterAtomId)
            result.builtin.builtinType = core::Value::CHARACTER;
         else if(type.builtin.value.data32 == gAtomAtomId)
            result.builtin.builtinType = core::Value::ATOM;
         else
            return core::Value::makeError(core::Value::TYPE_ERROR, null);
         return result;
      }
   }

   EVLAN_NATIVE_FUNCTION_1(getBuiltinType, value)
   {
      if(!value.isBuiltin())
         return runtimeError(core::Value::TYPE_ERROR, value);
      else
      {
         switch(value.builtin.builtinType)
         {
         case core::Value::BOOLEAN:
            return core::Value::makeAtom(gBooleanAtomId);
         case core::Value::SCALAR:
            return core::Value::makeAtom(gScalarAtomId);
         case core::Value::CHARACTER:
            return core::Value::makeAtom(gCharacterAtomId);
         case core::Value::ATOM:
            return core::Value::makeAtom(gAtomAtomId);
         default:
            return core::Value::makeAtom(gDataAtomId);
         }
      }
   }

   EVLAN_NATIVE_FUNCTION_1(characterToInteger, c)
   {
      if(!c.isCharacter())
         return runtimeError(core::Value::TYPE_ERROR, c);
      else
         return core::Value::makeScalar((double)c.builtin.value.data32);
   }

   EVLAN_NATIVE_FUNCTION_1(integerToCharacter, i)
   {
      if(!i.isScalar())
         return runtimeError(core::Value::TYPE_ERROR, i);
      int intval = (int)i.builtin.value.scalar;
      if(i.builtin.value.scalar != (double)intval)
         return runtimeError(core::Value::TYPE_ERROR, i);
      return core::Value::makeCharacter(intval);
   }

   EVLAN_NATIVE_FUNCTION_2(setToData, data, elementSize)
   {
      if(!data.isBuiltin()
      || !elementSize.isScalar())
         return runtimeError(core::Value::TYPE_ERROR, data, elementSize);

      int size = (int)elementSize.builtin.value.scalar;
      if(elementSize.builtin.value.scalar != (double)size)
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      int lg2Size = 0;
      while(size > (1 << lg2Size))
         ++lg2Size;
      if(size != (1 << lg2Size))
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      core::Value result = data;

      if(lg2Size == 6)
      {
         if(result.builtin.builtinType < Value::DATA64)
            result.builtin.value.data64 = result.builtin.value.data32;
      }
      else if(result.builtin.builtinType == Value::DATA64)
         result.builtin.value.data32 = result.builtin.value.data64;

      if(lg2Size < 5)
         result.builtin.value.data32 &= (1 << (1 << lg2Size)) - 1;

      result.builtin.builtinType = core::Value::DATA1 + lg2Size;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(encodeInteger, value)
   {
      if(!value.isScalar())
         return runtimeError(core::Value::TYPE_ERROR, value);

      core::Value result = value;

      int i = (int)result.builtin.value.scalar;
      if(result.builtin.value.scalar != (double)i)
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      result.builtin.value.data32 = i;
      result.builtin.builtinType = core::Value::DATA32;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(encodeInteger64, value)
   {
      if(!value.isScalar())
         return runtimeError(core::Value::TYPE_ERROR, value);

      core::Value result = value;

      int64 i = (int64)result.builtin.value.scalar;
      if(result.builtin.value.scalar != (double)i)
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      result.builtin.value.data64 = i;
      result.builtin.builtinType = core::Value::DATA64;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(decodeInteger, data)
   {
      if(!data.isBuiltin()
      || data.builtin.builtinType < core::Value::DATA2)
         return runtimeError(core::Value::TYPE_ERROR, data);

      core::Value result = data;
      if(data.builtin.builtinType == core::Value::DATA32)
         result.builtin.value.scalar = (int32)result.builtin.value.data32;
      else if(data.builtin.builtinType == core::Value::DATA64)
         result.builtin.value.scalar = (int64)result.builtin.value.data64;
      else
      {
         int bits = 1 << (data.builtin.builtinType - core::Value::DATA1);
         int mask = (1 << (bits - 1)) - 1;
         int signBit = 1 << (bits - 1);
         int32 i = result.builtin.value.data32 & mask;
         if(result.builtin.value.data32 & signBit)
            i = -i;
         result.builtin.value.scalar = i;
      }

      result.builtin.builtinType = core::Value::SCALAR;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(encodeUnsigned, value)
   {
      if(!value.isScalar())
         return runtimeError(core::Value::TYPE_ERROR, value);

      core::Value result = value;

      uint i = (uint)result.builtin.value.scalar;
      if(result.builtin.value.scalar != (double)i)
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      result.builtin.value.data32 = i;
      result.builtin.builtinType = core::Value::DATA32;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(encodeUnsigned64, value)
   {
      if(!value.isScalar())
         return runtimeError(core::Value::TYPE_ERROR, value);

      core::Value result = value;

      uint64 i = (uint64)result.builtin.value.scalar;
      if(result.builtin.value.scalar != (double)i)
         return core::Value::makeError(core::Value::TYPE_ERROR, null);

      result.builtin.value.data64 = i;
      result.builtin.builtinType = core::Value::DATA64;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(decodeUnsigned, data)
   {
      if(!data.isBuiltin()
      || data.builtin.builtinType < core::Value::DATA1)
         return runtimeError(core::Value::TYPE_ERROR, data);

      core::Value result = data;
      if(data.builtin.builtinType == core::Value::DATA32)
         result.builtin.value.scalar = (uint32)result.builtin.value.data32;
      else if(data.builtin.builtinType == core::Value::DATA64)
         result.builtin.value.scalar = (uint64)result.builtin.value.data64;
      else
      {
         int bits = 1 << (data.builtin.builtinType - core::Value::DATA1);
         int mask = (1 << bits) - 1;
         uint32 i = result.builtin.value.data32 & mask;
         result.builtin.value.scalar = i;
      }

      result.builtin.builtinType = core::Value::SCALAR;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(encodeFloat, value)
   {
      if(!value.isScalar())
         return runtimeError(core::Value::TYPE_ERROR, value);

      union {float f; uint32 i;} u;

      core::Value result = value;
      u.f = result.builtin.value.scalar;
      result.builtin.value.data32 = u.i;
      result.builtin.builtinType = core::Value::DATA32;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(decodeFloat, data)
   {
      if(!data.isBuiltin()
      || data.builtin.builtinType != core::Value::DATA32)
         return runtimeError(core::Value::TYPE_ERROR, data);

      union {float f; uint32 i;} u;

      core::Value result = data;
      u.i = result.builtin.value.data32;
      result.builtin.value.scalar = u.f;
      result.builtin.builtinType = core::Value::SCALAR;
      return result;
   }

   // benchmarking ===================================================
   EVLAN_NATIVE_FUNCTION_2(benchmark, function, parameter)
   {
      --core::gStackTop;
      *core::gStackTop = function;
      --core::gStackTop;
      *core::gStackTop = parameter;
      clock_t start = ::clock();
      simulator::apply(1);
      clock_t end = ::clock();
      return core::Value::makeScalar((end - start) / (double)CLOCKS_PER_SEC);
   }

   // process ========================================================
   implementation StateInfo: implements Object
   {
      StateInfo(core::Process* process, framework::SystemPtr system);
      ~StateInfo();
   };
   typedef SmartPointer<StateInfo> StateInfoPtr;

   StateInfo::StateInfo(core::Process* process, framework::SystemPtr system) {}
   StateInfo::~StateInfo() {}

   EVLAN_NATIVE_PROCESS(state, StateInfo)

   EVLAN_NATIVE_IDENTIFIER(gNewStateId, newState);
   EVLAN_NATIVE_IDENTIFIER(gResultId,   result);

   EVLAN_NATIVE_FUNCTION_2(processSet, process, newState)
   {
      if(!process.isProcess())
         return simTypeError(process);

      simPush(newState);
      memory::modifyProcessState(process.process.process);

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(processGet, process)
   {
      if(!process.isProcess())
         return simTypeError(process);
      return simReturn(process.process.process->state);
   }

   EVLAN_NATIVE_FUNCTION_2(processUpdate, process, updateFunction)
   {
      if(!process.isProcess() || !updateFunction.isFunction())
         return simTypeError(process, updateFunction);
      simPush(updateFunction);
      simPush(process.process.process->state);
      simApply(1);

      simPush(*gStackTop);
      operations::getMember(gNewStateId);
      memory::modifyProcessState(process.process.process);

      operations::getMember(gResultId);
      return simReturn(*gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_1(processMake, initialState)
   {
      simPush(initialState);
      memory::makeProcess($("state"));

      return simReturn(*core::gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_1(fork, tasks)
   {
      if(!tasks.isArray())
         return runtimeError(core::Value::TYPE_ERROR, tasks);

      for(int i = 0; i < tasks.array.size; i++)
      {
         simPush(Value::makeContext(Value::STOP));
         simPush(tasks.array.elements[i]);
         simQueue();
      }

      return simReturn();
   }

   EVLAN_NATIVE_IDENTIFIER(gMessageId, message);
   EVLAN_NATIVE_IDENTIFIER(gWarningId, warning);
   EVLAN_NATIVE_IDENTIFIER(gErrorId, error);

   EVLAN_NATIVE_FUNCTION_2(logMessage, type, message)
   {
      if(!type.isAtom() || !message.isString())
         return simTypeError(type, message);

      Debug::WarningLevel level;
      if(type.builtin.value.data32 == gMessageId)
         level = Debug::MESSAGE;
      else if(type.builtin.value.data32 == gWarningId)
         level = Debug::WARNING;
      else if(type.builtin.value.data32 == gErrorId)
         level = Debug::ERROR;
      else
         return simTypeError();

      Debug::log(level) << arrays::getString(message);

      return simReturn();
   }

   // atoms ==========================================================
   EVLAN_NATIVE_FUNCTION_1(atomToString, atom)
   {
      if(!atom.isAtom())
         return runtimeError(core::Value::TYPE_ERROR, atom);

      arrays::makeString(operations::getAtomName(atom.builtin.value.data32));
      return *core::gStackTop;
   }

   EVLAN_NATIVE_FUNCTION_1(stringToAtom, name)
   {
      String nameString = arrays::getString(name);
      if(nameString == null)
         return runtimeError(core::Value::TYPE_ERROR, name);

      for(int i = 0; i < nameString.getLength(); i++)
      {
         if((nameString[i] < 'a' || nameString[i] > 'z')
         && (nameString[i] < 'A' || nameString[i] > 'Z')
         && (nameString[i] < '0' || nameString[i] > '9'))
            return core::Value::makeError(core::Value::TYPE_ERROR, null);
      }

      return core::Value::makeAtom(operations::getAtomId(nameString));
   }

   // inspection =====================================================
   EVLAN_NATIVE_IDENTIFIER(gEmptyId,        empty);
   EVLAN_NATIVE_IDENTIFIER(gBuiltinId,      builtin);
   EVLAN_NATIVE_IDENTIFIER(gObjectId,       object);
   EVLAN_NATIVE_IDENTIFIER(gNativeId,       native);
   EVLAN_NATIVE_IDENTIFIER(gFunctionId,     function);
   EVLAN_NATIVE_IDENTIFIER(gRuntimeErrorId, runtimeError);
   EVLAN_NATIVE_IDENTIFIER(gArrayId,        array);
   EVLAN_NATIVE_IDENTIFIER(gStringId,       string);
   EVLAN_NATIVE_IDENTIFIER(gLargeDataId,    largeData);
   EVLAN_NATIVE_IDENTIFIER(gTaskId,         task);
   EVLAN_NATIVE_IDENTIFIER(gContextId,      context);
   EVLAN_NATIVE_IDENTIFIER(gModuleId,       module);
   EVLAN_NATIVE_IDENTIFIER(gProcessId,      process);
   EVLAN_NATIVE_IDENTIFIER(gUnknownId,      unknown);

   EVLAN_NATIVE_FUNCTION_1(getType, value)
   {
      switch(value.type)
      {
      case Value::EMPTY:
         return Value::makeAtom(gEmptyId);
      case Value::BUILTIN_OBJECT:
         return Value::makeAtom(gBuiltinId);
      case Value::NATIVE_CONTAINER:
         return Value::makeAtom(gNativeId);
      case Value::NATIVE_FUNCTION:
         return Value::makeAtom(gFunctionId);
      case Value::RUNTIME_ERROR:
         return Value::makeAtom(gRuntimeErrorId);
      case Value::FUNCTION:
         return Value::makeAtom(gFunctionId);
      case Value::OBJECT:
         return Value::makeAtom(gObjectId);
      case Value::ARRAY:
         return Value::makeAtom(gArrayId);
      case Value::PACKED_ARRAY:
         if(arrays::getStringLength(value) > 0)
            return Value::makeAtom(gStringId);
         else
            return Value::makeAtom(gArrayId);
      case Value::LARGE_DATA:
         return Value::makeAtom(gLargeDataId);
      case Value::TASK:
         return Value::makeAtom(gTaskId);
      case Value::CONTEXT:
         return Value::makeAtom(gContextId);
      case Value::MODULE:
         return Value::makeAtom(gModuleId);
      case Value::PROCESS:
         return Value::makeAtom(gProcessId);
      default:
         return Value::makeAtom(gUnknownId);
      }
   }

   EVLAN_NATIVE_IDENTIFIER(gNotAFunctionId,        notAFunction);
   EVLAN_NATIVE_IDENTIFIER(gArityMismatchId,       arityMismatch);
   EVLAN_NATIVE_IDENTIFIER(gNotAnObjectId,         notAnObject);
   EVLAN_NATIVE_IDENTIFIER(gNoSuchMemberId,        noSuchMember);
   EVLAN_NATIVE_IDENTIFIER(gConditionNotBooleanId, conditionNotBoolean);
   EVLAN_NATIVE_IDENTIFIER(gWrongTypeId,           wrongType);
   EVLAN_NATIVE_IDENTIFIER(gOutOfBoundsId,         outOfBounds);

   EVLAN_NATIVE_FUNCTION_1(makeError, type)
   {
      if(!type.isAtom())
         return simTypeError(type);

      if(type.builtin.value.data32 == gNotAFunctionId)
         return Value::makeError(Value::NOT_A_FUNCTION);
      else if(type.builtin.value.data32 == gArityMismatchId)
         return Value::makeError(Value::ARITY_MISMATCH);
      else if(type.builtin.value.data32 == gNotAnObjectId)
         return Value::makeError(Value::NOT_AN_OBJECT);
      else if(type.builtin.value.data32 == gNoSuchMemberId)
         return Value::makeError(Value::NO_SUCH_MEMBER);
      else if(type.builtin.value.data32 == gConditionNotBooleanId)
         return Value::makeError(Value::CONDITION_NOT_BOOLEAN);
      else if(type.builtin.value.data32 == gWrongTypeId)
         return Value::makeError(Value::TYPE_ERROR);
      else if(type.builtin.value.data32 == gOutOfBoundsId)
         return Value::makeError(Value::OUT_OF_BOUNDS);
      else
         return Value::makeError(Value::TYPE_ERROR);
   }

   EVLAN_NATIVE_FUNCTION_1(describeError, error)
   {
      if(!error.isError())
         return simTypeError(error);

      switch(error.runtimeError.errorCode)
      {
      case Value::NOT_A_FUNCTION:
         simPush(Value::makeAtom(gNotAFunctionId));
         break;
      case Value::ARITY_MISMATCH:
         simPush(Value::makeAtom(gArityMismatchId));
         break;
      case Value::NOT_AN_OBJECT:
         simPush(Value::makeAtom(gNotAnObjectId));
         break;
      case Value::NO_SUCH_MEMBER:
         simPush(Value::makeAtom(gNoSuchMemberId));
         break;
      case Value::CONDITION_NOT_BOOLEAN:
         simPush(Value::makeAtom(gConditionNotBooleanId));
         break;
      case Value::TYPE_ERROR:
         simPush(Value::makeAtom(gWrongTypeId));
         break;
      case Value::OUT_OF_BOUNDS:
         simPush(Value::makeAtom(gOutOfBoundsId));
         break;
      default:
         simPush(Value::makeAtom(gUnknownId));
         break;
      }

      core::Instruction* instruction = error.runtimeError.location;
      core::FunctionTag* function = null;
      core::Module* module = null;
      core::SourceLocation* location = null;

      if(instruction != null)
         function = memory::findFunction(instruction);
      if(function != null)
      {
         module = function->module;
         if(function->sourceLocations != null)
         {
            location = function->sourceLocations
                     + (instruction - function->instructions);
         }
      }

      if(module != null)
      {
         loader::ModuleInfoPtr info = loader::getModuleInfo(module);
         if(info == null)
            arrays::makeString($("(unknown module)"));
         else if(!info->fromFile)
            arrays::makeString($("*") + info->name);
         else
            arrays::makeString(info->name);
      }
      else
         arrays::makeString($("(unknown module)"));

      if(location != null)
      {
         simPush(Value::makeScalar((int)location->line));
         simPush(Value::makeScalar((int)location->column));
      }
      else
      {
         simPush(Value::makeScalar(-1));
         simPush(Value::makeScalar(-1));
      }

      Value* elements = memory::allocateValues(4);
      elements[0] = gStackTop[3];
      elements[1] = gStackTop[2];
      elements[2] = gStackTop[1];
      elements[3] = gStackTop[0];
      simPop(4);
      return Value::makeArray(4, elements);
   }

   EVLAN_NATIVE_FUNCTION_1(getMemberIds, value)
   {
      switch(value.type)
      {
      case Value::OBJECT:
         {
            core::ObjectTag* tag = value.object.tag;

            Value* elements = memory::allocateValues(tag->memberCount);
            for(int i = 0; i < tag->memberCount; i++)
               elements[i] = Value::makeAtom(tag->members[i].nameId);
            simPush(Value::makeArray(tag->memberCount, elements));
            arrays::tryPackArray();
            return *gStackTop;
         }

      case Value::NATIVE_CONTAINER:
         {
            Array<const int> list = native::getIdList();

            Value* elements = memory::allocateValues(list.getSize());
            for(int i = 0; i < list.getSize(); i++)
               elements[i] = Value::makeAtom(list[i]);
            simPush(Value::makeArray(list.getSize(), elements));
            arrays::tryPackArray();
            return *gStackTop;
         }

      default:
         return simTypeError(value);
      }
   }

   EVLAN_NATIVE_FUNCTION_2(getMember, object, id)
   {
      if(!id.isAtom())
         return simTypeError(object, id);

      simPush(object);
      operations::getMember(id.builtin.value.data32);
      return *gStackTop;
   }

   EVLAN_NATIVE_FUNCTION_2(setMembers, object, newMembers)
   {
      if(!object.isObject() || !newMembers.isObject())
         return simTypeError(object, newMembers);

      /* We can do this in O(n) because object member tables are
       * sorted by nameId.  Think like a merge from mergeSort. */
      Value* members = memory::allocateValues(object.object.tag->memberCount);
      int pos = 0;
      for(int i = 0; i < object.object.tag->memberCount; i++)
      {
         int offset = object.object.tag->members[i].offset;
         if(pos < newMembers.object.tag->memberCount
         && object.object.tag->members[i].nameId == newMembers.object.tag->members[pos].nameId)
         {
            members[offset] = newMembers.object.members[newMembers.object.tag->members[pos].offset];
            ++pos;
         }
         else
            members[offset] = object.object.members[offset];
      }
      //missed any?
      if(pos < newMembers.object.tag->memberCount)
         return Value::makeError(Value::NO_SUCH_MEMBER, null);

      return Value::makeObject(object.object.tag, members);
   }

   EVLAN_NATIVE_FUNCTION_2(mutateMembers, object, newMembers)
   {
      if(!object.isObject() || !newMembers.isObject())
         return simTypeError(object, newMembers);

      /* We can do this in O(n) because object member tables are
       * sorted by nameId.  Think like a merge from mergeSort. */
      int pos = 0;
      for(int i = 0; i < object.object.tag->memberCount; i++)
      {
         int offset = object.object.tag->members[i].offset;
         if(pos < newMembers.object.tag->memberCount
         && object.object.tag->members[i].nameId == newMembers.object.tag->members[pos].nameId)
         {
            simPush(newMembers.object.members[newMembers.object.tag->members[pos].offset]);
            simPush(object.object.members[offset]);
            simApply(1);
            ++pos;
         }
         else
            simPush(object.object.members[offset]);
      }
      //missed any?
      if(pos < newMembers.object.tag->memberCount)
         return Value::makeError(Value::NO_SUCH_MEMBER, null);

      Value* members = memory::allocateValues(object.object.tag->memberCount);
      for(int i = object.object.tag->memberCount - 1; i >= 0; i--)
      {
         int offset = object.object.tag->members[i].offset;
         members[offset] = *gStackTop;
         simPop();
      }

      return Value::makeObject(object.object.tag, members);
   }

   EVLAN_NATIVE_FUNCTION_1(getParameterNames, function)
   {
      switch(function.type)
      {
      case Value::FUNCTION:
         {
            core::FunctionTag* tag = function.function.tag;

            Value* elements = memory::allocateValues(tag->arity);
            for(int i = 0; i < tag->arity; i++)
               elements[i] = Value::makeAtom(tag->parameterNames[i]);
            simPush(Value::makeArray(tag->arity, elements));
            arrays::tryPackArray();
            return *gStackTop;
         }

      case Value::NATIVE_FUNCTION:
         {
            core::NativeFunctionTag* tag = function.nativeFunction.tag;

            Value* elements = memory::allocateValues(tag->parameterNames.getSize());
            for(int i = 0; i < tag->parameterNames.getSize(); i++)
               elements[i] = Value::makeAtom(operations::getIdentifierId(tag->parameterNames[i]));
            simPush(Value::makeArray(tag->parameterNames.getSize(), elements));
            arrays::tryPackArray();
            return *gStackTop;
         }

      default:
         return simTypeError(function);
      }
   }

   EVLAN_NATIVE_FUNCTION_1(getEnvironment, function)
   {
      if(!function.isFunction())
         return runtimeError(core::Value::TYPE_ERROR, function);

      if(function.isNativeFunction())
         return core::Value::makeArray(0, null);

      return core::Value::makeArray(
         function.function.tag->environmentSize,
         function.function.environment);
   }

   EVLAN_NATIVE_FUNCTION_1(getSourceLocation, function)
   {
      if(!function.isFunction())
         return simTypeError(function);

      if(function.isNativeFunction())
      {
         arrays::makeString($("native:") + function.nativeFunction.tag->identifier);
         return *gStackTop;
      }
      else
      {
         loader::ModuleInfoPtr info = loader::getModuleInfo(function.function.tag->module);
         String moduleName;
         if(info == null)
            moduleName = $("(unknown module)");
         else if(!info->fromFile)
            moduleName = $("*") + info->name;
         else
            moduleName = info->name;
         arrays::makeString(moduleName + $(":") + String(function.function.tag->line+1) + $(":") + String(function.function.tag->column+1));
         return *gStackTop;
      }
   }

   EVLAN_NATIVE_FUNCTION_1(getProcessState, process)
   {
      if(!process.isProcess())
         return simTypeError(process);

      return process.process.process->state;
   }

   EVLAN_NATIVE_FUNCTION_1(getProcessType, process)
   {
      if(!process.isProcess())
         return simTypeError(process);

      return Value::makeAtom(process.process.process->type);
   }

   EVLAN_NATIVE_FUNCTION_1(getModuleRoot, module)
   {
      if(!module.isModule())
         return simTypeError(module);

      return module.module.module->root;
   }

   EVLAN_NATIVE_FUNCTION_1(getModuleName, module)
   {
      if(!module.isModule())
         return simTypeError(module);

      loader::ModuleInfoPtr info = loader::getModuleInfo(module.module.module);
      if(info == null)
         arrays::makeString($("(unknown module)"));
      else if(!info->fromFile)
         arrays::makeString($("*") + info->name);
      else
         arrays::makeString(info->name);

      return *gStackTop;
   }

   EVLAN_NATIVE_FUNCTION_1(getDocumentation, value)
   {
      if(value.isObject())
      {
         arrays::makeString(value.object.tag->documentation,
                            value.object.tag->documentationLength);
      }
      else if(value.isEvlanFunction())
      {
         arrays::makeString(value.function.tag->documentation,
                            value.function.tag->documentationLength);
      }
      else
         return Value::makeArray(0, null);

      return *gStackTop;
   }

   // UTF-8! =========================================================
   inline int divideRoundUp(int i, int j)
   {
      return (i+j-1)/j;
   }

   EVLAN_NATIVE_FUNCTION_1(decodeUtf8, data)
   {
      if(data.isArray() && data.array.size == 0)
         return data;  //empty
      if(!data.isByteArray())
         return simTypeError(data);

      byte* source = ((byte*)data.packedArray.elements)
                   + data.packedArray.offset;
      utf8::DecodeInfo info = utf8::getDecodeInfo(source, data.packedArray.size);

      switch(info.bitCount)
      {
      default:
      case 7:
         //All characters are 7-bit.  No decoding needed.
         data.packedArray.packType = Value::PACKED_CHARACTER;
         return data;
      case 8:
         {
            int wordSize = divideRoundUp(info.size, sizeof(word) / sizeof(char8));
            char8* target = (char8*)memory::allocateWords(wordSize);
            utf8::decode(target, source, data.packedArray.size);

            return Value::makePackedArray(Value::PACKED_CHARACTER, 3, info.size, (word*)target);
         }
      case 16:
         {
            int wordSize = divideRoundUp(info.size, sizeof(word) / sizeof(char16));
            char16* target = (char16*)memory::allocateWords(wordSize);
            utf8::decode(target, source, data.packedArray.size);

            return Value::makePackedArray(Value::PACKED_CHARACTER, 4, info.size, (word*)target);
         }
      case 32:
         {
            int wordSize = divideRoundUp(info.size, sizeof(word) / sizeof(char32));
            char32* target = (char32*)memory::allocateWords(wordSize);
            utf8::decode(target, source, data.packedArray.size);

            return Value::makePackedArray(Value::PACKED_CHARACTER, 5, info.size, (word*)target);
         }
      }
   }

   EVLAN_NATIVE_FUNCTION_1(encodeUtf8, text)
   {
      if(text.isArray() && text.array.size == 0)
         return text;  //empty
      if(!text.isString())
         return simTypeError(text);

      switch(text.packedArray.elementSize)
      {
      default:
      case 3:
         {
            char8* source = ((char8*)text.packedArray.elements)
                          + text.packedArray.offset;

            int outputSize = utf8::getEncodedSize(source, text.packedArray.size);
            if(outputSize == text.packedArray.size)
            {
               //no decoding needed!
               text.packedArray.packType = Value::PACKED_DATA;
               return text;
            }
            int wordSize = divideRoundUp(outputSize, sizeof(word)/sizeof(char8));
            byte* target = (byte*)memory::allocateWords(wordSize);
            source = ((char8*)text.packedArray.elements) + text.packedArray.offset;
            utf8::encode(target, source, text.packedArray.size);
            return Value::makeByteArray(outputSize, target);
         }
      case 4:
         {
            char16* source = ((char16*)text.packedArray.elements)
                           + text.packedArray.offset;

            int outputSize = utf8::getEncodedSize(source, text.packedArray.size);
            int wordSize = divideRoundUp(outputSize, sizeof(word)/sizeof(char16));
            byte* target = (byte*)memory::allocateWords(wordSize);
            source = ((char16*)text.packedArray.elements) + text.packedArray.offset;
            utf8::encode(target, source, text.packedArray.size);
            return Value::makeByteArray(outputSize, target);
         }
      case 5:
         {
            char32* source = ((char32*)text.packedArray.elements)
                           + text.packedArray.offset;

            int outputSize = utf8::getEncodedSize(source, text.packedArray.size);
            int wordSize = divideRoundUp(outputSize, sizeof(word)/sizeof(char32));
            byte* target = (byte*)memory::allocateWords(wordSize);
            source = ((char32*)text.packedArray.elements) + text.packedArray.offset;
            utf8::encode(target, source, text.packedArray.size);
            return Value::makeByteArray(outputSize, target);
         }
      }
   }

   // text newline conversions =======================================
   EVLAN_NATIVE_FUNCTION_1(newlinesToUnix, data)
   {
      if(data.isArray() && data.array.size == 0)
         return data;
      if(!data.isPackedArray()
      || data.packedArray.packType != core::Value::PACKED_DATA
      || data.packedArray.elementSize != 3)
         return runtimeError(core::Value::TYPE_ERROR, data);

      byte* source = (byte*)data.packedArray.elements + data.packedArray.offset;
      int size = data.packedArray.size;

      int outputSize = size;

      for(int i = 0; i < size; i++)
      {
         if(source[i] == '\n' && i != 0 && source[i-1] == '\r')
            --outputSize;
      }

      int wordSize = outputSize / sizeof(word);
      if(outputSize % sizeof(word) > 0)
         ++wordSize;
      byte* target = (byte*)memory::allocateWords(wordSize);
      source = (byte*)data.packedArray.elements + data.packedArray.offset;

      int pos = 0;
      for(int i = 0; i < size; i++)
      {
         if(source[i] == '\r')
            target[pos++] = '\n';
         else if(source[i] != '\n' || i == 0 || source[i-1] != '\r')
            target[pos++] = source[i];
      }

      core::Value result;
      result.type = core::Value::PACKED_ARRAY;
      result.packedArray.packType = core::Value::PACKED_DATA;
      result.packedArray.elementSize = 3;
      result.packedArray.offset = 0;
      result.packedArray.size = outputSize;
      result.packedArray.elements = (word*)target;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(newlinesToDos, data)
   {
      if(data.isArray() && data.array.size == 0)
         return data;
      if(!data.isPackedArray()
      || data.packedArray.packType != core::Value::PACKED_DATA
      || data.packedArray.elementSize != 3)
         return runtimeError(core::Value::TYPE_ERROR, data);

      byte* source = (byte*)data.packedArray.elements + data.packedArray.offset;
      int size = data.packedArray.size;

      int outputSize = size;

      for(int i = 0; i < size; i++)
      {
         if(source[i] == '\r')
            ++outputSize;
         else if(source[i] == '\n')
         {
            if(i == 0 || source[i-1] != '\r')
               ++outputSize;
            else
               --outputSize;
         }
      }

      int wordSize = outputSize / sizeof(word);
      if(outputSize % sizeof(word) > 0)
         ++wordSize;
      byte* target = (byte*)memory::allocateWords(wordSize);
      source = (byte*)data.packedArray.elements + data.packedArray.offset;

      int pos = 0;
      for(int i = 0; i < size; i++)
      {
         if(source[i] == '\r')
         {
            target[pos++] = '\r';
            target[pos++] = '\n';
         }
         else if(source[i] == '\n')
         {
            if(i == 0 || source[i-1] != '\r')
            {
               target[pos++] = '\r';
               target[pos++] = '\n';
            }
         }
         else
            target[pos++] = source[i];
      }

      core::Value result;
      result.type = core::Value::PACKED_ARRAY;
      result.packedArray.packType = core::Value::PACKED_DATA;
      result.packedArray.elementSize = 3;
      result.packedArray.offset = 0;
      result.packedArray.size = outputSize;
      result.packedArray.elements = (word*)target;
      return result;
   }

   EVLAN_NATIVE_FUNCTION_1(newlinesToMac, data)
   {
      if(data.isArray() && data.array.size == 0)
         return data;
      if(!data.isPackedArray()
      || data.packedArray.packType != core::Value::PACKED_DATA
      || data.packedArray.elementSize != 3)
         return runtimeError(core::Value::TYPE_ERROR, data);

      byte* source = (byte*)data.packedArray.elements + data.packedArray.offset;
      int size = data.packedArray.size;

      int outputSize = size;

      for(int i = 0; i < size; i++)
      {
         if(source[i] == '\n' && i != 0 && source[i-1] == '\r')
            --outputSize;
      }

      int wordSize = outputSize / sizeof(word);
      if(outputSize % sizeof(word) > 0)
         ++wordSize;
      byte* target = (byte*)memory::allocateWords(wordSize);
      source = (byte*)data.packedArray.elements + data.packedArray.offset;

      int pos = 0;
      for(int i = 0; i < size; i++)
      {
         if(source[i] != '\n')
            target[pos++] = source[i];
         else if(i == 0 || source[i-1] != '\r')
            target[pos++] = '\r';
      }

      core::Value result;
      result.type = core::Value::PACKED_ARRAY;
      result.packedArray.packType = core::Value::PACKED_DATA;
      result.packedArray.elementSize = 3;
      result.packedArray.offset = 0;
      result.packedArray.size = outputSize;
      result.packedArray.elements = (word*)target;
      return result;
   }

   // linear congruential PNRG =======================================
   //Had to implement this natively for the 64-bit integer math.
   EVLAN_NATIVE_FUNCTION_4(linearCongruentialRandom, x, a, c, m)
   {
      if(!x.isScalar() || !a.isScalar() || !c.isScalar() || !m.isScalar())
         return simTypeError(x, a, c, m);

      uint64 _x = (uint64)x.builtin.value.scalar;
      uint64 _a = (uint64)a.builtin.value.scalar;
      uint64 _c = (uint64)c.builtin.value.scalar;
      uint64 _m = (uint64)m.builtin.value.scalar;

      return Value::makeScalar((_a * _x + _c) % _m);
   }
}

// ===================================================================

} //namespace api
} //namespace evlan
