/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/05/21 (KV)
 *  + First Version
 */

// ===================================================================

#include "network.h"
#include <framework/System.h>
#include <evlan/vm/memory.h>
#include <evlan/vm/simulator.h>
#include <evlan/vm/operations.h>
#include <evlan/vm/messages.h>
#include <evlan/vm/loader.h>
#include <evlan/vm/arrays.h>
#include <evlan/vm/native.h>
#include <evlan/vm/shortcuts.h>
#include "error.h"

#ifdef PLATFORM_WIN32
#include <windows.h>
#include <winsock.h>
#include <ws2tcpip.h>
#undef ERROR
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#endif

namespace evlan {
namespace api {
using namespace evlan::vm;
using namespace evlan::vm::shortcuts;
using namespace error;
using ::framework::messages::EventManager;
using ::framework::messages::EventManagerPtr;
using ::framework::messages::MessagePtr;
using ::framework::messages::makeMessage;
using ::framework::messages::MessageReceiver;
using ::framework::common::byte;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::api::network
// TODO (KV): Review evlan::api::network

namespace network
{
#ifdef PLATFORM_WIN32
   typedef SOCKET Socket;
   inline ErrorCode getError() {return WSAGetLastError();}

   static const ErrorCode ERROR_CONNECT_IN_PROGRESS = WSAEWOULDBLOCK;
   static const ErrorCode ERROR_WOULD_BLOCK = WSAEWOULDBLOCK;

   static const int SHUT_RD = SD_RECEIVE;
   static const int SHUT_WR = SD_SEND;

   bool isAcceptErrorFatal(ErrorCode error)
   {
      /* Winsock does the Right Thing with accept() errors.  If the connection
       * breaks before accept() is called, it nevertheless returns a valid
       * socket in a broken state.  Any actual error from accept() (other than
       * WSAEWOULDBLOCK) is a fatal error indicating that the listening socket
       * is broken. */
      return true;
   }
#else
   typedef int Socket;
   inline ErrorCode getError() {return errno;}
   static const Socket INVALID_SOCKET = -1;
   inline int closesocket(int sock) {return close(sock);}

   static const ErrorCode ERROR_SUCCESS = 0;
   static const ErrorCode ERROR_CONNECT_IN_PROGRESS = EINPROGRESS;
   static const ErrorCode ERROR_WOULD_BLOCK = EAGAIN;

/*
   Linux man page for accept(2) says:
      ERROR HANDLING
         Linux accept passes already-pending network errors on the new socket as
         an error code from accept.   This  behaviour  differs  from  other  BSD
         socket  implementations.  For reliable operation the application should
         detect the network errors defined for the  protocol  after  accept  and
         treat  them  like EAGAIN by retrying. In case of TCP/IP these are ENET-
         DOWN, EPROTO, ENOPROTOOPT, EHOSTDOWN, ENONET, EHOSTUNREACH, EOPNOTSUPP,
         and ENETUNREACH.

   Note that the list of retry errors that they give is not complete!  And even
   if it were, it would only be valid for IPv4.  Therefore, we pretty much have
   to assume that all unrecognized errors are retry errors, otherwise the server's
   accept() loop will probably simply quit out when accept() returns an
   unrecognized error.

   This problem is not Linux-only.  FreeBSD's implementation of accept() will
   return ECONNABORTED if a connection is closed while waiting in the listen
   queue.  I honestly don't know if FreeBSD returns any other retry errors, as
   the FreeBSD man page for accept(2) does not even address this issue.

   This is HORRIBLE DESIGN.  The correct thing to do would be to return a valid
   socket in an error state, such that the first call to read() or write() on
   that socket would return an error.  I can't stress enough how much the people
   who are responsible for this are idiots.  How could you even write the above
   paragraph on error handling without realizing how ridiculously ugly it is?
*/

   bool isAcceptErrorFatal(ErrorCode error)
   {
      switch(error)
      {
      case EBADF:
      case EMFILE:
      case ENFILE:
      case ENOTSOCK:
      case EINVAL:
      case EFAULT:
         return true;
      default:
         return false;
      };
   }
#endif

   static const int BUFFER_SIZE = 8192;
   EVLAN_NATIVE_IDENTIFIER(gEndOfStreamId, endOfStream);
   EVLAN_NATIVE_IDENTIFIER(gOutputTerminatedId, outputTerminated);
   EVLAN_NATIVE_IDENTIFIER(gConnectFailedId, connectFailed);
   EVLAN_NATIVE_IDENTIFIER(gConnectionLostId, connectionLost);
   EVLAN_NATIVE_IDENTIFIER(gConnectionAbortedId, connectionAborted);
   EVLAN_NATIVE_IDENTIFIER(gListenerClosedId, listenerClosed);

   EVLAN_NATIVE_IDENTIFIER(gConnectionErrorId, connectionError);
   EVLAN_NATIVE_IDENTIFIER(gListenerErrorId, listenerError);

   // ConnectionInfo =================================================
   implementation ConnectionInfo: implements MessageReceiver
   {
      ConnectionInfo(core::Process* process, framework::SystemPtr system, ErrorCode error);
      ConnectionInfo(core::Process* process, Socket sock, framework::SystemPtr system);
      ConnectionInfo(core::Process* process, framework::SystemPtr system);
      ~ConnectionInfo();

      // MessageReceiver ---------------------------------------------
      void receive(MessagePtr message);

      // internal ----------------------------------------------------
      void close();  //close the socket and destroy messages

      void onConnectSuccess();
      void onConnectError(ErrorCode error);
      void onConnect(); //parameters must be on stack

      void onReadable();
      void onWritable();

      core::Process* mProcess;

      enum
      {
         AWAITING_CONNECT,
         CONNECTED,

         CONNECT_FAILED,
         BROKEN_AFTER_RESTORE,
         ABORTED
      } mState;

      //only valid in AWAITING_CONNECT and CONNECTED states
      Socket mSocket;
      EventManager::StreamWatcherPtr mWatcher;
      MessagePtr mReadableMessage;
      MessagePtr mWritableMessage;

      //only valid in CONNECTED state
      bool mIsReadable;
      bool mIsWritable;
      bool mIsReadShutdown;
      bool mIsWriteShutdown;

      //only valid in CONNECT_FAILED state
      int mConnectError;
   };
   typedef SmartPointer<ConnectionInfo> ConnectionInfoPtr;

   ConnectionInfo::ConnectionInfo(core::Process* process, framework::SystemPtr system, ErrorCode error)
      : mProcess(process), mState(CONNECT_FAILED), mSocket(INVALID_SOCKET), mConnectError(error)
   {
   }

   ConnectionInfo::ConnectionInfo(core::Process* process, Socket sock, framework::SystemPtr system)
      : mProcess(process), mState(AWAITING_CONNECT), mSocket(sock)
   {
      mWatcher = system->getEventManager()->makeStreamWatcher(mSocket);
      mReadableMessage = makeMessage(this, 0, null);
      mWritableMessage = makeMessage(this, 0, null);
   }

   ConnectionInfo::ConnectionInfo(core::Process* process, framework::SystemPtr system)
      : mProcess(process)
   {
      if(mProcess->state.isEmpty())
      {
         //Previous connection was already closed.  Nothing to see here.
         mState = ABORTED;
      }
      else if(mProcess->state.isData()
           && mProcess->state.builtin.builtinType == Value::DATA32)
      {
         /* connection had previously failed, either due to error or
          * restoration of state. */
         mConnectError = mProcess->state.builtin.value.data32;
         if(mConnectError == ERROR_SUCCESS)
            mState = BROKEN_AFTER_RESTORE;  //broken after a *previous* restore
         else
            mState = CONNECT_FAILED;
      }
      else if(mProcess->state.isArray()
           && mProcess->state.array.size >= 2
           && mProcess->state.array.size <= 3)
      {
         /* Connection was active or being connected.  There's no way
          * we can restore it.  Pretend it was lost or failed. */
         mState = BROKEN_AFTER_RESTORE;
         mReadableMessage = makeMessage(this, 0, null);
         mReadableMessage->send();
         mProcess->keepAlive = true;
      }
      else
         FW_THROW(spnew FatalError($("Type error in connection process state.")));
   }

   ConnectionInfo::~ConnectionInfo()
   {
      if(mState == AWAITING_CONNECT
      || (mState == CONNECTED && (!mIsWriteShutdown || !mIsReadShutdown)))
         close();
   }

   void ConnectionInfo::receive(MessagePtr message)
   {
      if(mState == BROKEN_AFTER_RESTORE)
      {
         //clean this stuff up...
         if(mProcess->state.array.size == 3)
         {
            mState = AWAITING_CONNECT;
            simPush(simThrow(gConnectFailedId,
               $("Connection failed due to restoration of saved state.")));
            onConnect();
         }
         else
         {
            mState = CONNECTED;
            onReadable();
            onWritable();
         }

         mState = BROKEN_AFTER_RESTORE;
         simPush(Value::makeBuiltin((uint32)ERROR_SUCCESS, Value::DATA32));
         memory::modifyProcessState(mProcess);

         mReadableMessage = null;
         mProcess->keepAlive = false;
      }
      else if(mState == AWAITING_CONNECT)
      {
         if(message == mWritableMessage)
         {
            int error;
            socklen_t length = sizeof(int);
            getsockopt(mSocket, SOL_SOCKET, SO_ERROR, (char*)&error, &length);

            if(error == ERROR_SUCCESS)
               onConnectSuccess();
            else
               onConnectError(error);

            mProcess->keepAlive = false;
         }
      }
      else if(mState == CONNECTED)
      {
         if(message == mReadableMessage)
         {
            onReadable();
            mProcess->keepAlive = !mProcess->state.array.elements[1].isEmpty();
         }
         else if(message == mWritableMessage)
         {
            onWritable();
            mProcess->keepAlive = !mProcess->state.array.elements[0].isEmpty();
         }
      }
   }

   void ConnectionInfo::close()
   {
      mWatcher = null;
      mReadableMessage->disable();
      mWritableMessage->disable();
      mReadableMessage = null;
      mWritableMessage = null;

      closesocket(mSocket);
      mSocket = INVALID_SOCKET;
   }

   void ConnectionInfo::onConnectSuccess()
   {
      simPush(simReturn());
      onConnect();
      mState = CONNECTED;
      mIsReadable = true;
      mIsWritable = true;
      mIsReadShutdown = false;
      mIsWriteShutdown = false;
   }

   void ConnectionInfo::onConnectError(ErrorCode error)
   {
      simPush(throwOsError(error, gConnectFailedId));
      onConnect();
      mState = CONNECT_FAILED;
      mConnectError = error;
      close();
      simPush(Value::makeBuiltin((uint32)error, Value::DATA32));
      memory::modifyProcessState(mProcess);
   }

   void ConnectionInfo::onConnect()
   {
      //fire "connected" events
      simPush(mProcess->state.array.elements[2]);

      simPush(mProcess->state);
      gStackTop->array.size = 2;
      memory::modifyProcessState(mProcess);

      while(!gStackTop->isEmpty())
      {
         simPush(gStackTop->array.elements[0]);
         simPush(gStackTop[2]);
         simQueue();
         *gStackTop = gStackTop->array.elements[1];
      }
      simPop(2);

      onReadable();
      onWritable();
   }

   void ConnectionInfo::onReadable()
   {
      mIsReadable = true;

      if(!mProcess->state.array.elements[0].isEmpty())
      {
         simPush(mProcess->state.array.elements[0]);

         Value* newElements = memory::allocateValues(2);
         newElements[0] = Value::makeEmpty();
         newElements[1] = mProcess->state.array.elements[1];
         simPush(Value::makeArray(2, newElements));
         memory::modifyProcessState(mProcess);

         while(!gStackTop->isEmpty())
         {
            simPush(gStackTop->array.elements[0]);
            simPush(simReturn());
            simQueue();
            *gStackTop = gStackTop->array.elements[1];
         }
         simPop();
      }
   }

   void ConnectionInfo::onWritable()
   {
      mIsWritable = true;

      if(!mProcess->state.array.elements[1].isEmpty())
      {
         simPush(mProcess->state.array.elements[1]);

         Value* newElements = memory::allocateValues(2);
         newElements[0] = mProcess->state.array.elements[0];
         newElements[1] = Value::makeEmpty();
         simPush(Value::makeArray(2, newElements));
         memory::modifyProcessState(mProcess);

         while(!gStackTop->isEmpty())
         {
            simPush(gStackTop->array.elements[0]);
            simPush(simReturn());
            simQueue();
            *gStackTop = gStackTop->array.elements[1];
         }
         simPop();
      }
   }

   EVLAN_NATIVE_PROCESS(connection, ConnectionInfo)

   // connection glue code ===========================================
   EVLAN_NATIVE_FUNCTION_2(connectionMake, systemProcess, address)
   {
      if(!systemProcess.isProcess() || !address.isString())
         return simTypeError(systemProcess, address);
      EVLAN_NATIVE_GET_PROCESS_INFO(system, systemProcess, framework::SystemPtr);

      EV_MAKE_CSTRING(addressString, address);
      char* portString = std::strrchr(addressString, ':');
      if(portString != null)
      {
         *portString = '\0';
         ++portString;
      }

      addrinfo hints;
      addrinfo* result = null;

      memset(&hints, 0, sizeof(addrinfo));
      hints.ai_flags = AI_NUMERICHOST;
      hints.ai_family = PF_UNSPEC;
      hints.ai_socktype = SOCK_STREAM;

      ErrorCode errorCode = getaddrinfo(addressString, portString, &hints, &result);

      if(errorCode == ERROR_SUCCESS && result != null)
      {
         Socket sock = socket(result->ai_family,
                              result->ai_socktype,
                              result->ai_protocol);

         if(sock == INVALID_SOCKET)
            errorCode = getError();
         else
         {
            Value* elements = memory::allocateValues(3);
            memset(elements, 0, 3 * sizeof(Value));
            simPush(Value::makeArray(3, elements));
            memory::makeProcess($("connection"));

            ConnectionInfoPtr connectionInfo =
               spnew ConnectionInfo(gStackTop->process.process, sock, system);
            gStackTop->process.process->nativeInfo = connectionInfo;

            if(::connect(sock, result->ai_addr, result->ai_addrlen) == 0)
               connectionInfo->onConnectSuccess();
            else
            {
               errorCode = getError();
               if(errorCode == ERROR_CONNECT_IN_PROGRESS)
                  connectionInfo->mWatcher->notifyWhenWritable(connectionInfo->mWritableMessage);
               else
                  connectionInfo->onConnectError(errorCode);
            }

            freeaddrinfo(result);

            return simReturn(*gStackTop);
         }
      }

      if(result != null)
         freeaddrinfo(result);

      simPush(Value::makeBuiltin((uint32)errorCode, Value::DATA32));
      memory::makeProcess($("connection"));

      ConnectionInfoPtr connectionInfo =
         spnew ConnectionInfo(gStackTop->process.process, system, errorCode);
      gStackTop->process.process->nativeInfo = connectionInfo;

      return simReturn(*gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_1(connectionRead, connectionProcess)
   {
      if(!connectionProcess.isProcess())
         return simTypeError(connectionProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(connectionInfo, connectionProcess, ConnectionInfoPtr);

      if(connectionInfo->mState != ConnectionInfo::CONNECTED
      && connectionInfo->mState != ConnectionInfo::AWAITING_CONNECT)
      {
         switch(connectionInfo->mState)
         {
         case ConnectionInfo::CONNECT_FAILED:
            return throwOsError(connectionInfo->mConnectError, gConnectFailedId);
         case ConnectionInfo::BROKEN_AFTER_RESTORE:
            return simThrow(gConnectionLostId, $("Tried to read from a network connection that was lost after loading from a persistent state."));
         default:
         case ConnectionInfo::ABORTED:
            return simThrow(gConnectionAbortedId, $("Tried to read from an aborted network connection."));
         }
      }

      if(connectionInfo->mState == ConnectionInfo::AWAITING_CONNECT
      || !connectionInfo->mIsReadable)
         return simReturn(Value::makeArray(0, null));

      if(connectionInfo->mIsReadShutdown)
         return simThrow(gEndOfStreamId, $("Tried to read past end of stream."));

      word* buffer = memory::allocateWords(BUFFER_SIZE / sizeof(word));

      bool hadError = false;
      int total = 0;
      while(true)
      {
         int count = recv(connectionInfo->mSocket, (char*)buffer + total, BUFFER_SIZE - total, 0);

         if(count < 0)
         {
            hadError = true;
            break;
         }

         if(count == 0)
            break;

         total += count;
         if(total >= BUFFER_SIZE)
            break;
      }

      if(total == 0)
      {
         ErrorCode errorCode = getError();
         memory::returnWords(buffer, BUFFER_SIZE / sizeof(word));

         if(hadError)
         {
            if(errorCode == ERROR_WOULD_BLOCK)
            {
               connectionInfo->mIsReadable = false;
               connectionInfo->mWatcher->notifyWhenReadable(connectionInfo->mReadableMessage);

               return simReturn(Value::makeArray(0, null));
            }
            else
               return throwOsError(errorCode, gConnectionLostId);
         }
         else
         {
            shutdown(connectionInfo->mSocket, SHUT_RD);
            connectionInfo->mIsReadShutdown = true;
            if(connectionInfo->mIsWriteShutdown)
               connectionInfo->close();
            return simThrow(gEndOfStreamId, $("Tried to read past end of stream."));
         }
      }
      else if(total < BUFFER_SIZE)
      {
         //funny math insures we round up to the nearest word
         int unused = (BUFFER_SIZE - total) / sizeof(word);
         memory::returnWords(buffer + (BUFFER_SIZE / sizeof(word) - unused), unused);
      }

      --gStackTop;
      gStackTop->type = Value::PACKED_ARRAY;
      gStackTop->packedArray.packType    = Value::PACKED_DATA;
      gStackTop->packedArray.elementSize = 3;
      gStackTop->packedArray.offset      = 0;
      gStackTop->packedArray.size        = total;
      gStackTop->packedArray.elements    = buffer;

      return simReturn(*gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_2(connectionWrite, connectionProcess, data)
   {
      if(!connectionProcess.isProcess())
         return simTypeError(connectionProcess, data);
      EVLAN_NATIVE_GET_PROCESS_INFO(connectionInfo, connectionProcess, ConnectionInfoPtr);

      if(connectionInfo->mState != ConnectionInfo::CONNECTED
      && connectionInfo->mState != ConnectionInfo::AWAITING_CONNECT)
      {
         switch(connectionInfo->mState)
         {
         case ConnectionInfo::CONNECT_FAILED:
            return throwOsError(connectionInfo->mConnectError, gConnectFailedId);
         case ConnectionInfo::BROKEN_AFTER_RESTORE:
            return simThrow(gConnectionLostId, $("Tried to write to a network connection that was lost after loading from a persistent state."));
         default:
         case ConnectionInfo::ABORTED:
            return simThrow(gConnectionAbortedId, $("Tried to write to a closed network connection."));
         }
      }

      if(connectionInfo->mState == ConnectionInfo::AWAITING_CONNECT
      || !connectionInfo->mIsWritable)
      {
         //not ready; return zero bytes written
         return simReturn(Value::makeScalar(0));
      }

      if(connectionInfo->mIsWriteShutdown)
         return simThrow(gOutputTerminatedId, $("Tried to write to a network connection that had already been terminated."));

      if((data.isArray() && data.array.size == 0)
      || (data.isPackedArray() && data.packedArray.size == 0))
      {
         //zero-size array; return zero bytes written
         return simReturn(Value::makeScalar(0));
      }

      if(!data.isPackedArray()
      || data.packedArray.packType != Value::PACKED_DATA
      || data.packedArray.elementSize != 3)
         return simTypeError(data);

      char* buffer = (char*)data.packedArray.elements + data.packedArray.offset;

      bool hadError = false;
      int bufferPos = 0;
      while(true)
      {
         int count =
            send(connectionInfo->mSocket,
                 buffer + bufferPos,
                 data.packedArray.size - bufferPos,
                 0);

         if(count < 0)
         {
            hadError = true;
            break;
         }

         bufferPos += count;
         if(bufferPos == data.packedArray.size)
            break;
      }

      if(hadError)
      {
         ErrorCode errorCode = getError();
         if(errorCode == ERROR_WOULD_BLOCK)
         {
            connectionInfo->mIsWritable = false;
            connectionInfo->mWatcher->notifyWhenWritable(connectionInfo->mWritableMessage);

            return simReturn(Value::makeScalar(bufferPos));
         }
         else
            return throwOsError(errorCode, gConnectionLostId);
      }
      else
         return simReturn(Value::makeScalar(bufferPos));
   }

   EVLAN_NATIVE_FUNCTION_1(connectionTerminate, connectionProcess)
   {
      if(!connectionProcess.isProcess())
         return simTypeError(connectionProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(connectionInfo, connectionProcess, ConnectionInfoPtr);

      if(connectionInfo->mState != ConnectionInfo::AWAITING_CONNECT
      && connectionInfo->mState != ConnectionInfo::CONNECTED)
      {
         switch(connectionInfo->mState)
         {
         case ConnectionInfo::CONNECT_FAILED:
            return throwOsError(connectionInfo->mConnectError, gConnectFailedId);
         case ConnectionInfo::BROKEN_AFTER_RESTORE:
            return simThrow(gConnectionLostId, $("Tried to terminate a network connection that was lost after loading from a persistent state."));
         default:
         case ConnectionInfo::ABORTED:
            return simThrow(gConnectionAbortedId, $("Tried to terminate a closed network connection."));
         }
      }

      if(connectionInfo->mState == ConnectionInfo::AWAITING_CONNECT
      || !connectionInfo->mIsWritable)
      {
         //not ready
         return simReturn(Value::makeBoolean(false));
      }
      else if(connectionInfo->mIsWriteShutdown)
         return simThrow(gOutputTerminatedId, $("Tried to terminate a network connection that had already been terminated."));
      else if(shutdown(connectionInfo->mSocket, SHUT_WR) == 0)
      {
         connectionInfo->mIsWriteShutdown = true;
         if(connectionInfo->mIsReadShutdown)
            connectionInfo->close();

         return simReturn(Value::makeBoolean(true));
      }
      else
      {
         ErrorCode errorCode = getError();
         return throwOsError(errorCode, gConnectionLostId);
      }
   }

   EVLAN_NATIVE_FUNCTION_1(connectionAbort, connectionProcess)
   {
      if(!connectionProcess.isProcess())
         return simTypeError(connectionProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(connectionInfo, connectionProcess, ConnectionInfoPtr);

      if(connectionInfo->mState == ConnectionInfo::AWAITING_CONNECT
      || connectionInfo->mState == ConnectionInfo::CONNECTED)
      {
         if(connectionInfo->mState == ConnectionInfo::AWAITING_CONNECT)
         {
            simPush(simThrow(gConnectionAbortedId, $("Connection attempt aborted.")));
            connectionInfo->onConnect();
         }
         else
         {
            connectionInfo->onReadable();
            connectionInfo->onWritable();
         }

         if(!connectionInfo->mIsReadShutdown || !connectionInfo->mIsWriteShutdown)
            connectionInfo->close();
         connectionInfo->mProcess->keepAlive = false;
      }

      simPush(Value::makeEmpty());
      memory::modifyProcessState(connectionInfo->mProcess);
      connectionInfo->mState = ConnectionInfo::ABORTED;

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(connectionWaitConnected, connectionProcess)
   {
      if(!connectionProcess.isProcess())
         return simTypeError(connectionProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(connectionInfo, connectionProcess, ConnectionInfoPtr);

      if(connectionInfo->mState == ConnectionInfo::AWAITING_CONNECT)
      {
         messages::getCurrentContext();

         Value* elements = memory::allocateValues(2);
         elements[0] = *gStackTop;
         elements[1] = connectionInfo->mProcess->state.array.elements[2];
         *gStackTop = Value::makeArray(2, elements);

         elements = memory::allocateValues(3);
         elements[0] = connectionInfo->mProcess->state.array.elements[0];
         elements[1] = connectionInfo->mProcess->state.array.elements[1];
         elements[2] = *gStackTop;
         *gStackTop = Value::makeArray(3, elements);

         memory::modifyProcessState(connectionInfo->mProcess);

         connectionInfo->mProcess->keepAlive = true;
         return Value::makeTask(Value::STOP);
      }
      else if(connectionInfo->mState == ConnectionInfo::CONNECTED)
         return simReturn();
      else
      {
         switch(connectionInfo->mState)
         {
         case ConnectionInfo::CONNECT_FAILED:
            return throwOsError(connectionInfo->mConnectError, gConnectFailedId);
         case ConnectionInfo::BROKEN_AFTER_RESTORE:
            return simThrow(gConnectionLostId, $("Connection attempt was interrupted by restoration of saved state."));
         default:
         case ConnectionInfo::ABORTED:
            return simThrow(gConnectionAbortedId, $("Connection attempt was aborted."));
         }
      }
   }

   EVLAN_NATIVE_FUNCTION_1(connectionWaitReadable, connectionProcess)
   {
      if(!connectionProcess.isProcess())
         return simTypeError(connectionProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(connectionInfo, connectionProcess, ConnectionInfoPtr);

      if(connectionInfo->mState == ConnectionInfo::AWAITING_CONNECT
      || (connectionInfo->mState == ConnectionInfo::CONNECTED
       && !connectionInfo->mIsReadable && !connectionInfo->mIsReadShutdown))
      {
         messages::getCurrentContext();

         Value* elements = memory::allocateValues(2);
         elements[0] = *gStackTop;
         elements[1] = connectionInfo->mProcess->state.array.elements[0];
         *gStackTop = Value::makeArray(2, elements);

         elements = memory::allocateValues(connectionInfo->mProcess->state.array.size);
         elements[0] = *gStackTop;
         elements[1] = connectionInfo->mProcess->state.array.elements[1];
         if(connectionInfo->mProcess->state.array.size == 3)
            elements[2] = connectionInfo->mProcess->state.array.elements[2];
         *gStackTop = Value::makeArray(connectionInfo->mProcess->state.array.size, elements);

         memory::modifyProcessState(connectionInfo->mProcess);

         connectionInfo->mProcess->keepAlive = true;
         return Value::makeTask(Value::STOP);
      }
      else
         return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(connectionWaitWritable, connectionProcess)
   {
      if(!connectionProcess.isProcess())
         return simTypeError(connectionProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(connectionInfo, connectionProcess, ConnectionInfoPtr);

      if(connectionInfo->mState == ConnectionInfo::AWAITING_CONNECT
      || (connectionInfo->mState == ConnectionInfo::CONNECTED
       && !connectionInfo->mIsWritable && !connectionInfo->mIsWriteShutdown))
      {
         messages::getCurrentContext();

         Value* elements = memory::allocateValues(2);
         elements[0] = *gStackTop;
         elements[1] = connectionInfo->mProcess->state.array.elements[1];
         *gStackTop = Value::makeArray(2, elements);

         elements = memory::allocateValues(connectionInfo->mProcess->state.array.size);
         elements[0] = connectionInfo->mProcess->state.array.elements[0];
         elements[1] = *gStackTop;
         if(connectionInfo->mProcess->state.array.size == 3)
            elements[2] = connectionInfo->mProcess->state.array.elements[2];
         *gStackTop = Value::makeArray(connectionInfo->mProcess->state.array.size, elements);

         memory::modifyProcessState(connectionInfo->mProcess);

         connectionInfo->mProcess->keepAlive = true;
         return Value::makeTask(Value::STOP);
      }
      else
         return simReturn();
   }

   // ListenerInfo ===================================================
   implementation ListenerInfo: implements MessageReceiver
   {
      ListenerInfo(core::Process* process, Socket sock, framework::SystemPtr system);
      ListenerInfo(core::Process* process, framework::SystemPtr system);
      ~ListenerInfo();

      // MessageReceiver ---------------------------------------------
      void receive(MessagePtr message);

      // internal ----------------------------------------------------
      void onAcceptable();

      framework::SystemPtr mSystem;
      core::Process* mProcess;

      enum
      {
         LISTENING,
         LIMBO,            //failed to re-enable after restore
         CLOSED
      } mState;

      //valid only in LISTENING state
      Socket mSocket;
      EventManager::StreamWatcherPtr mWatcher;
      MessagePtr mAcceptableMessage;
      bool mCanAccept;
   };
   typedef SmartPointer<ListenerInfo> ListenerInfoPtr;

   Socket makeListenerSocket(const Value& addressValue, ErrorCode* errorCode)
   {
      EV_MAKE_CSTRING(addressString, addressValue);

      char* address = addressString;
      char* port = std::strrchr(addressString, ':');
      if(port == null)
      {
         port = addressString;
         address = null;
      }
      else
      {
         *port = '\0';
         ++port;
      }

      addrinfo hints;
      addrinfo* result = null;

      memset(&hints, 0, sizeof(addrinfo));
      hints.ai_flags = AI_PASSIVE;
      if(address == null)
         hints.ai_family = PF_INET;
      else
         hints.ai_family = PF_UNSPEC;
      hints.ai_socktype = SOCK_STREAM;

      int error = getaddrinfo(address, port, &hints, &result);
      Socket sock = INVALID_SOCKET;

      if(error == 0 && result != null)
      {
         sock = socket(result->ai_family,
                       result->ai_socktype,
                       result->ai_protocol);

         if(sock == INVALID_SOCKET)
            error = getError();
         else if(bind(sock, result->ai_addr, result->ai_addrlen) != 0
              || listen(sock, SOMAXCONN) != 0)
         {
            error = getError();
            closesocket(sock);
            sock = INVALID_SOCKET;
         }
      }

      if(result != null)
         freeaddrinfo(result);

      if(errorCode != null)
         *errorCode = error;

      return sock;
   }

   ListenerInfo::ListenerInfo(core::Process* process, Socket sock, framework::SystemPtr system)
      : mSystem(system), mProcess(process), mState(LISTENING), mSocket(sock), mCanAccept(true)
   {
      mWatcher = system->getEventManager()->makeStreamWatcher(mSocket);
      mAcceptableMessage = makeMessage(this, 0, null);
   }

   ListenerInfo::ListenerInfo(core::Process* process, framework::SystemPtr system)
      : mSystem(system), mProcess(process)
   {
      if(process->state.type == Value::EMPTY)
         mState = CLOSED;
      else if(process->state.isArray()
           && process->state.array.size == 2)
      {
         mSocket = makeListenerSocket(process->state.array.elements[0], null);

         process->keepAlive = !process->state.array.elements[1].isEmpty();

         if(mSocket == INVALID_SOCKET)
         {
            mState = LIMBO;

            Debug::log(Debug::WARNING) << "Failed to restore listener from persistent state:  "
                                       << arrays::getString(process->state.array.elements[0])
                                       << ": " << strerror(errno);
         }
         else
         {
            mState = LISTENING;
            mCanAccept = true;
            mWatcher = system->getEventManager()->makeStreamWatcher(mSocket);
            mAcceptableMessage = makeMessage(this, 0, null);

            if(!process->state.array.elements[1].isEmpty())
               mAcceptableMessage->send();
         }
      }
      else
         FW_THROW(spnew FatalError($("Type error in listener process state.")));
   }

   ListenerInfo::~ListenerInfo()
   {
      mWatcher = null;
      if(mState == LISTENING)
         closesocket(mSocket);
   }

   void ListenerInfo::receive(MessagePtr message)
   {
      if(message == mAcceptableMessage)
      {
         onAcceptable();
         mProcess->keepAlive = false;
      }
   }

   void ListenerInfo::onAcceptable()
   {
      mCanAccept = true;

      if(!mProcess->state.array.elements[1].isEmpty())
      {
         simPush(mProcess->state.array.elements[1]);

         Value* newElements = memory::allocateValues(2);
         newElements[0] = mProcess->state.array.elements[0];
         newElements[1] = Value::makeEmpty();
         simPush(Value::makeArray(2, newElements));
         memory::modifyProcessState(mProcess);

         while(!gStackTop->isEmpty())
         {
            simPush(gStackTop->array.elements[0]);
            simPush(simReturn());
            simQueue();
            *gStackTop = gStackTop->array.elements[1];
         }
         simPop();
      }
   }

   EVLAN_NATIVE_PROCESS(listener, ListenerInfo)

   // listener glue code =============================================
   EVLAN_NATIVE_FUNCTION_2(listenerMake, systemProcess, port)
   {
      if(!systemProcess.isProcess() || !port.isString())
         return simTypeError(systemProcess, port);
      EVLAN_NATIVE_GET_PROCESS_INFO(system, systemProcess, framework::SystemPtr);

      ErrorCode error;
      Socket sock = makeListenerSocket(port, &error);

      if(sock == INVALID_SOCKET)
         return throwOsError(error);
      else
      {
         Value* elements = memory::allocateValues(2);
         elements[0] = port;
         elements[1].type = Value::EMPTY;
         simPush(Value::makeArray(2, elements));
         memory::makeProcess($("listener"));

         ListenerInfoPtr listenerInfo = spnew ListenerInfo(gStackTop->process.process, sock, system);
         listenerInfo->mProcess->nativeInfo = listenerInfo;

         return simReturn(*gStackTop);
      }
   }

   EVLAN_NATIVE_FUNCTION_1(listenerAccept, listenerProcess)
   {
      if(!listenerProcess.isProcess())
         return simTypeError(listenerProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(listenerInfo, listenerProcess, ListenerInfoPtr);

      if(listenerInfo->mState == ListenerInfo::CLOSED)
         return simThrow(gListenerClosedId, $("Tried to accept from a closed listener."));

      if(listenerInfo->mState == ListenerInfo::LIMBO
      || !listenerInfo->mCanAccept)
         return simReturn(Value::makeArray(0, null));

      word address[1024];
      socklen_t addressSize = 1024 * sizeof(word);
      Socket sock = INVALID_SOCKET;

      while(true)
      {
         sock = accept(listenerInfo->mSocket, (sockaddr*)address, &addressSize);

         if(sock != INVALID_SOCKET)
            break;

         ErrorCode errorCode = getError();

         if(errorCode == ERROR_WOULD_BLOCK)
         {
            listenerInfo->mCanAccept = false;
            listenerInfo->mWatcher->notifyWhenReadable(listenerInfo->mAcceptableMessage);

            return simReturn(Value::makeArray(0, null));
         }
         else if(isAcceptErrorFatal(errorCode))
            return throwOsError(errorCode);
      }

      char host[256];    //numeric addresses won't be more than 256 chars, right?

      if(getnameinfo((sockaddr*)address, addressSize,
                     host, 256, null, 0,
                     NI_NUMERICHOST | NI_NUMERICSERV) != 0)
         strcpy(host, "unknown");

      Value* elements = memory::allocateValues(3);
      memset(elements, 0, sizeof(Value) * 3);
      simPush(Value::makeArray(3, elements));
      memory::makeProcess($("connection"));
      ConnectionInfoPtr connectionInfo =
         spnew ConnectionInfo(gStackTop->process.process, sock, listenerInfo->mSystem);
      gStackTop->process.process->nativeInfo = connectionInfo;
      connectionInfo->onConnectSuccess();

      arrays::makeString(host);

      elements = memory::allocateValues(2);
      elements[0] = gStackTop[1];
      elements[1] = gStackTop[0];

      return simReturn(Value::makeArray(2, elements));
   }

   EVLAN_NATIVE_FUNCTION_1(listenerClose, listenerProcess)
   {
      if(!listenerProcess.isProcess())
         return simTypeError(listenerProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(listenerInfo, listenerProcess, ListenerInfoPtr);

      if(listenerInfo->mState != ListenerInfo::CLOSED)
      {
         if(listenerInfo->mState == ListenerInfo::LISTENING)
         {
            listenerInfo->mWatcher = null;
            listenerInfo->mAcceptableMessage->disable();
            listenerInfo->mAcceptableMessage = null;
            closesocket(listenerInfo->mSocket);
         }

         listenerInfo->onAcceptable();
         simPush(Value::makeEmpty());
         memory::modifyProcessState(listenerInfo->mProcess);
         listenerInfo->mState = ListenerInfo::CLOSED;
         listenerInfo->mProcess->keepAlive = false;
      }

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(listenerWaitAcceptable, listenerProcess)
   {
      if(!listenerProcess.isProcess())
         return simTypeError(listenerProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(listenerInfo, listenerProcess, ListenerInfoPtr);

      if((listenerInfo->mState == ListenerInfo::LISTENING && listenerInfo->mCanAccept)
       || listenerInfo->mState == ListenerInfo::CLOSED)
         return simReturn();
      else
      {
         messages::getCurrentContext();

         Value* elements = memory::allocateValues(2);
         elements[0] = *gStackTop;
         elements[1] = listenerInfo->mProcess->state.array.elements[1];
         *gStackTop = Value::makeArray(2, elements);

         elements = memory::allocateValues(2);
         elements[0] = listenerInfo->mProcess->state.array.elements[0];
         elements[1] = *gStackTop;
         *gStackTop = Value::makeArray(2, elements);

         memory::modifyProcessState(listenerInfo->mProcess);
         listenerInfo->mProcess->keepAlive = true;
         return Value::makeTask(Value::STOP);
      }
   }
}

// ===================================================================

} //namespace api
} //namespace evlan
