/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2005/02/17 (KV)
 *  + First Version
 */

// ===================================================================

#include "optimize.h"
#include <evlan/utf8.h>
#include <evlan/vm/native.h>
#include <evlan/vm/memory.h>
#include <evlan/vm/operations.h>
#include <evlan/vm/arrays.h>
#include <evlan/vm/shortcuts.h>
#include <stack>

namespace evlan {
namespace api {
using namespace vm;
using namespace vm::shortcuts;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::api::optimize
// TODO (KV): Review evlan::api::optimize

/* This module contains native code versions of a couple functions
 * which were taking the majority of my web server's CPU time. */
namespace optimize
{
   struct NodeStackItem
   {
      Value* branches;
      int count;

      inline NodeStackItem() {}
      inline NodeStackItem(Value* branches, int count)
         : branches(branches), count(count) {}
   };

   EVLAN_NATIVE_IDENTIFIER(gIsLeafId  , isLeaf  );
   EVLAN_NATIVE_IDENTIFIER(gValueId   , value   );
   EVLAN_NATIVE_IDENTIFIER(gBranchesId, branches);

   //state = array, Array(Array(Tree))
   EVLAN_NATIVE_FUNCTION_2(optimizeListTreeProgressive, state, thresholdParam)
   {
      if(!state.isArray() || !thresholdParam.isInteger())
         return simTypeError(state, thresholdParam);

      int count = 0, totalSize = 0;
      int threshold = (int)thresholdParam.builtin.value.scalar;

      std::stack<NodeStackItem> nodeStack;
      for(int i = 0; i < state.array.size; i++)
      {
         if(!state.array.elements[i].isArray())
            return simTypeError(state.array.elements[i]);

         nodeStack.push(NodeStackItem(
            state.array.elements[i].array.elements,
            state.array.elements[i].array.size));
      }

      while(!nodeStack.empty())
      {
         if(nodeStack.top().count == 0)
            nodeStack.pop();
         else
         {
            Value* branch = nodeStack.top().branches;
            ++nodeStack.top().branches;
            --nodeStack.top().count;

            simPush(*branch);
            operations::getMember(gIsLeafId);
            if(!gStackTop->isBoolean())
               return simTypeError(*gStackTop);
            if(gStackTop->builtin.value.boolean)
            {
               *gStackTop = *branch;
               operations::getMember(gValueId);

               ++count;
               if(gStackTop->isArray())
                  totalSize += gStackTop->array.size;
               else if(gStackTop->isPackedArray())
                  totalSize += gStackTop->packedArray.size;
               else
                  return simTypeError(*gStackTop);

               //leave on stack

               if(totalSize >= threshold)
                  break;
            }
            else
            {
               *gStackTop = *branch;
               operations::getMember(gBranchesId);
               if(!gStackTop->isArray())
                  return simTypeError(*gStackTop);
               nodeStack.push(NodeStackItem(
                  gStackTop->array.elements, gStackTop->array.size));
               simPop();
            }
         }
      }

      int finalStackSize = 0;
      while(!nodeStack.empty())
      {
         ++finalStackSize;
         simPush(Value::makeArray(nodeStack.top().count, nodeStack.top().branches));
         nodeStack.pop();
      }

      Value* elements = memory::allocateValues(finalStackSize + count + 2);
      Value* parts = elements + finalStackSize;
      Value* results = elements + finalStackSize + count;

      memcpy(elements, gStackTop, sizeof(Value) * finalStackSize);
      simPop(finalStackSize);

      for(int i = 0; i < count; i++)
         parts[i] = gStackTop[count-i-1];
      simPop(count);

      results[0] = Value::makeArray(finalStackSize, elements);
      results[1] = Value::makeArray(count, parts);

      return Value::makeArray(2, results);
   }


   template <typename Character>
   Value addEntities(const Value& text)
   {
      Character* source = (Character*)text.packedArray.elements
                        + text.packedArray.offset;
      int outSize = text.packedArray.size;
      for(int i = 0; i < text.packedArray.size; i++)
      {
         if(source[i] == '<' || source[i] == '>')
            outSize += 3;
         else if(source[i] == '&')
            outSize += 4;
         else if(source[i] == '\"')
            outSize += 5;
      }
      Character* target = (Character*)memory::allocate<Character>(outSize);
      source = (Character*)text.packedArray.elements + text.packedArray.offset;

      int pos = 0;
      for(int i = 0; i < text.packedArray.size; i++)
      {
         if(source[i] == '<')
         {
            target[pos++] = '&';
            target[pos++] = 'l';
            target[pos++] = 't';
            target[pos++] = ';';
         }
         else if(source[i] == '>')
         {
           target[pos++] = '&';
            target[pos++] = 'g';
            target[pos++] = 't';
            target[pos++] = ';';
         }
         else if(source[i] == '&')
         {
            target[pos++] = '&';
            target[pos++] = 'a';
            target[pos++] = 'm';
            target[pos++] = 'p';
            target[pos++] = ';';
         }
         else if(source[i] == '\"')
         {
            target[pos++] = '&';
            target[pos++] = 'q';
            target[pos++] = 'u';
            target[pos++] = 'o';
            target[pos++] = 't';
            target[pos++] = ';';
         }
         else
            target[pos++] = source[i];
      }

      return Value::makePackedArray(Value::PACKED_CHARACTER,
                                    text.packedArray.elementSize,
                                    outSize, (word*)target);
   }

   EVLAN_NATIVE_FUNCTION_1(optimizeHtmlAddEntities, text)
   {
      if(text.isArray() && text.array.size == 0)
         return text;
      if(!text.isPackedArray() || text.packedArray.packType != Value::PACKED_CHARACTER)
         return simTypeError(text);

      switch(text.packedArray.elementSize)
      {
      case 3:
         return addEntities<char>(text);
      case 4:
         return addEntities<char16>(text);
      case 5:
         return addEntities<char32>(text);
      default:
         return simTypeError(text);
      }
   }



   EVLAN_NATIVE_FUNCTION_2(optimizeBinarySearch, sortedArray, isBeforeKey)
   {
      if(sortedArray.isArray())
      {
         int start = 0, end = sortedArray.array.size;

         while(start < end)
         {
            int midpoint = (start + end) / 2;
            simPush(isBeforeKey);
            simPush(sortedArray.array.elements[midpoint]);
            simApply(1);
            if(!gStackTop->isBoolean())
               return simTypeError(*gStackTop);
            if(gStackTop->builtin.value.boolean)
               start = midpoint + 1;
            else
               end = midpoint;
            simPop();
         }

         return Value::makeScalar(start);
      }
      else if(sortedArray.isPackedArray())
      {
         int start = 0, end = sortedArray.packedArray.size;

         while(start < end)
         {
            int midpoint = (start + end) / 2;
            simPush(isBeforeKey);
            simPush(sortedArray);
            arrays::getElement(midpoint);
            simApply(1);
            if(!gStackTop->isBoolean())
               return simTypeError(*gStackTop);
            if(gStackTop->builtin.value.boolean)
               start = midpoint + 1;
            else
               end = midpoint;
            simPop();
         }

         return Value::makeScalar(start);
      }
      else
         return simTypeError(sortedArray, isBeforeKey);
   }
}

// ===================================================================

} //namespace api
} //namespace evlan
