/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2005/03/06 (KV)
 *  + First Version (copied from Win32 version)
 */

// ===================================================================

#include "directory.h"
#include <framework/util/Vector.h>
#include <framework/messages/EventManager.h>
#include <evlan/vm/native.h>
#include <evlan/vm/memory.h>
#include <evlan/vm/simulator.h>
#include <evlan/vm/operations.h>
#include <evlan/vm/arrays.h>
#include <evlan/vm/messages.h>
#include <evlan/vm/loader.h>
#include <evlan/vm/shortcuts.h>
#include <evlan/api/error.h>
#include "file.h"
#include <map>
#include <math.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>

namespace evlan {
namespace api {
namespace posix {
using namespace evlan::vm;
using namespace evlan::vm::shortcuts;
using namespace evlan::api::error;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::api::posix::directory
// TODO (KV): Review evlan::api::posix::directory

namespace directory
{
   /* Evlan programs are not given direct filesystem access.  Instead,
    * the VM provides a number of directories, identified by strings,
    * which may be accessed.  The user decides where these point.
    * This means that a saved Evlan program state can safely be moved
    * between two systems with different directory structures as long
    * as the same set of directories, with the same names, are
    * provided to the VM's. */
   std::map<String, String> gRoots;

   /* invalid:  ., ..
    * characters: /
    * ranges: 0x00 through 0x19, 0x7F through 0x9F, 0x100 and up. */
   struct ValidCharacterTable
   {
      bool table[256];
      inline bool operator[](unsigned char index) {return table[index];}

      ValidCharacterTable()
      {
         memset(table, 0, sizeof(table));
         for(int i = 0x20; i < 0x7f; i++) table[i] = true;
         for(int i = 0xa0; i < 0x9f; i++) table[i] = true;
         table['/'] = false;
      }
   };

   bool isValidPath(Value& path)
   {
      static ValidCharacterTable validTable;

      if(!path.isPackedArray()
      || path.packedArray.elementSize != 3   //disallow unicode for now...
      || path.packedArray.packType != Value::PACKED_CHARACTER
      || path.packedArray.size >= PATH_MAX)
         return false;

      char* buffer = (char*)path.packedArray.elements + path.packedArray.offset;

      int partStart = 0;
      for(int i = 0; ; i++)
      {
         switch(i-partStart)
         {
         case 0:
            //disallow empty components and leading or trailing slashes
            if(i == path.packedArray.size || buffer[i] == '/')
               return false;
            break;
         case 1:
            //disallow .
            if((i == path.packedArray.size || buffer[i] == '/')
            && buffer[i-1] == '.')
               return false;
            break;
         case 2:
            //disallow ..
            if((i == path.packedArray.size || buffer[i] == '/')
            && buffer[i-1] == '.' && buffer[i-2] == '.')
               return false;
            break;
         default:
            break;
         }

         if(i == path.packedArray.size)
            return true;

         if(buffer[i] == '/')
            partStart = i + 1;
         else if(!validTable[buffer[i]])
            return false;
      }
   }

   String makePath(String root, Value& subPath, bool addSlash)
   {
      int subPathLength = arrays::getStringLength(subPath);
      if(subPathLength == 0)
         addSlash = false;
      Array<char> buffer(root.getLength() + subPathLength + (addSlash?1:0));
      if(buffer != null)
      {
         root.copyToBuffer(&buffer[0], false);
         arrays::copyString(&buffer[0] + root.getLength(), subPath, false);
      }

      int totalLength = root.getLength() + subPathLength;
      if(addSlash)
         buffer[totalLength] = '/';

      return String(buffer.toConst());
   }

   void makePathCString(char* buffer, String root, Value& subPath)
   {
      root.copyToBuffer(buffer, false);
      arrays::copyString(buffer + root.getLength(), subPath, true);
   }

   #define PATH_CSTRING(CSTRING_NAME, ROOT, SUBPATH)                              \
      char CSTRING_NAME[ROOT.getLength() + arrays::getStringLength(SUBPATH) + 1]; \
      makePathCString(CSTRING_NAME, ROOT, SUBPATH)


   EVLAN_NATIVE_IDENTIFIER(gNameNotAllowedId, nameNotAllowed);
   EVLAN_NATIVE_IDENTIFIER(gWrongTypeId, wrongType);
   EVLAN_NATIVE_IDENTIFIER(gFileErrorId, fileError);

   EVLAN_NATIVE_IDENTIFIER(gNotFoundId, notFound);
   EVLAN_NATIVE_IDENTIFIER(gFileId, file);
   EVLAN_NATIVE_IDENTIFIER(gDirectoryId, directory);

   EVLAN_NATIVE_IDENTIFIER(gCurrentId, current);
   EVLAN_NATIVE_IDENTIFIER(gApplicationDataId, applicationData);

   EVLAN_NATIVE_IDENTIFIER(gTypeId, type);
   EVLAN_NATIVE_IDENTIFIER(gExtensionsId, extensions);

   inline Value throwNotAllowed(Value& path)
      {return simThrow(gNameNotAllowedId, gFileErrorId, path, $("The file or directory name is not allowed."));}

   // process info ===================================================
   implementation DirectoryInfo: implements Object
   {
      DirectoryInfo(String path, framework::SystemPtr system);
      DirectoryInfo(core::Process* process, framework::SystemPtr system);
      ~DirectoryInfo();

      String mPath;  //ends with '/'
      framework::SystemPtr mSystem;
   };
   typedef SmartPointer<DirectoryInfo> DirectoryInfoPtr;

   DirectoryInfo::DirectoryInfo(String path, framework::SystemPtr system)
      : mPath(path), mSystem(system) {}

   DirectoryInfo::DirectoryInfo(core::Process* process, framework::SystemPtr system)
      : mSystem(system)
   {
      if(!process->state.isString())
         FW_THROW(spnew FatalError($("Type error in directory process state.")));

      String rootName = arrays::getString(process->state);

      std::map<String, String>::iterator iter = gRoots.find(rootName);
      if(iter == gRoots.end())
         FW_THROW(spnew FatalError($("Open root directory is no longer mounted: ") + rootName));

      mPath = iter->second;
   }

   DirectoryInfo::~DirectoryInfo()
   {}

   EVLAN_NATIVE_PROCESS(directory, DirectoryInfo)

   // glue code ======================================================
   EVLAN_NATIVE_FUNCTION_1(directoryGetRoots, systemProcess)
   {
      if(!systemProcess.isProcess())
         simTypeError(systemProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(system, systemProcess, framework::SystemPtr);

      std::map<String, String>::iterator iter = gRoots.begin();
      int count = 0;
      for(; iter != gRoots.end(); ++iter, ++count)
      {
         arrays::makeString(iter->first);

         simPush(*gStackTop);
         memory::makeProcess($("directory"));
         gStackTop->process.process->nativeInfo = spnew DirectoryInfo(iter->second, system);

         Value* elements = memory::allocateValues(2);
         elements[0] = gStackTop[1];
         elements[1] = gStackTop[0];
         simPop(2);
         simPush(Value::makeArray(2, elements));
      }

      Value* elements = memory::allocateValues(count);
      for(int i = count-1; i >= 0; i--)
      {
         elements[i] = *gStackTop;
         simPop();
      }

      return simReturn(Value::makeArray(count, elements));
   }

   /* subPath is expected to end with a '/' and have already been checked
    * for validity previously. */
   EVLAN_NATIVE_FUNCTION_2(directoryList, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);

      int pathLength = directory->mPath.getLength()
                     + arrays::getStringLength(subPath) + 1;
      char path[pathLength];
      directory->mPath.copyToBuffer(path, false);
      arrays::copyString(path + directory->mPath.getLength(), subPath);

      int fileCount = 0;
      int directoryCount = 0;

      DIR* dir = opendir(path);
      if(dir != null)
      {
         struct dirent* entry;

         while((entry = readdir(dir)) != null)
         {
            if(std::strcmp(entry->d_name, "." ) != 0
            && std::strcmp(entry->d_name, "..") != 0)
            {
#ifdef PLATFORM_LINUX
               arrays::makeString(entry->d_name);
#else
               arrays::makeString(entry->d_name, entry->d_namlen);
#endif

               if(entry->d_type == DT_DIR)
               {
                  ++directoryCount;
                  if(fileCount > 0)
                  {
                     //swap to directory list
                     Value temp = gStackTop[fileCount];
                     gStackTop[fileCount] = *gStackTop;
                     *gStackTop = temp;
                  }
               }
               else
                  ++fileCount;
            }
         }

         closedir(dir);
      }

      Value* target = memory::allocateValues(fileCount + directoryCount + 2);
      memcpy(target, gStackTop, (fileCount + directoryCount) * sizeof(Value));
      simPop(fileCount + directoryCount);

      Value* elements = target + fileCount + directoryCount;
      elements[0] = Value::makeArray(fileCount, target);
      elements[1] = Value::makeArray(directoryCount, target + fileCount);
      return simReturn(Value::makeArray(2, elements));
   }

   EVLAN_NATIVE_FUNCTION_2(directoryGetType, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);
      PATH_CSTRING(path, directory->mPath, subPath);

      int id;

      struct stat stats;
      if(stat(path, &stats) < 0)
      {
         int errorCode = errno;
         if(errorCode == ENOENT
         || errorCode == ENOTDIR)
            id = gNotFoundId;
         else
            return throwOsError(errorCode, gFileErrorId, subPath);
      }
      else if(S_ISDIR(stats.st_mode))
         id = gDirectoryId;
      else
         id = gFileId;

      return simReturn(Value::makeAtom(id));
   }

   EVLAN_NATIVE_FUNCTION_2(directoryOpenFile, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);

      String path = makePath(directory->mPath, subPath, false);
      simPush(directoryProcess.process.process->state);
      simPush(subPath);
      int errorCode = file::make(path, false, directory->mSystem);

      if(errorCode == 0)
         return simReturn(*gStackTop);
      else if(errorCode == EISDIR)
         return simThrow(gWrongTypeId, gFileErrorId, subPath, $("Tried to open a directory as a file."));
      else
         return throwOsError(errorCode, gFileErrorId, subPath);
   }

   EVLAN_NATIVE_FUNCTION_2(directoryCreateFile, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);

      String path = makePath(directory->mPath, subPath, false);
      simPush(directoryProcess.process.process->state);
      simPush(subPath);
      int errorCode = file::make(path, true, directory->mSystem);

      if(errorCode == 0)
         return simReturn(*gStackTop);
      else if(errorCode == EISDIR)
         return simThrow(gWrongTypeId, gFileErrorId, subPath, $("Tried to open a directory as a file."));
      else
         return throwOsError(errorCode, gFileErrorId, subPath);
   }

   EVLAN_NATIVE_FUNCTION_2(directoryCreateDirectory, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);
      PATH_CSTRING(path, directory->mPath, subPath);

      if(mkdir(path, 0777) == 0)
         return simReturn();
      else
      {
         int error = errno;
         if(error == EEXIST)
            return simReturn();
         return throwOsError(errno, gFileErrorId, subPath);
      }
   }

   EVLAN_NATIVE_FUNCTION_2(directoryRemoveFile, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);
      PATH_CSTRING(path, directory->mPath, subPath);

      if(unlink(path) == 0)
         return simReturn();
      else
         return throwOsError(errno, gFileErrorId, subPath);
   }

   EVLAN_NATIVE_FUNCTION_2(directoryRemoveDirectory, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);
      PATH_CSTRING(path, directory->mPath, subPath);

      if(rmdir(path))
         return simReturn();
      else
         return throwOsError(errno, gFileErrorId, subPath);
   }

   /* subPath is expected to end with a '/' and have already been checked
    * for validity previously. */
   EVLAN_NATIVE_FUNCTION_2(directoryGetTimestamp, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);

      int pathLength = directory->mPath.getLength()
                     + arrays::getStringLength(subPath) + 1;
      char path[pathLength];
      directory->mPath.copyToBuffer(path, false);
      arrays::copyString(path + directory->mPath.getLength(), subPath, true);

      struct stat stats;
      if(stat(path, &stats) != 0)
         return throwOsError(errno, gFileErrorId, subPath);

      return simReturn(Value::makeScalar(stats.st_mtime));
   }

   EVLAN_NATIVE_FUNCTION_3(directorySetTimestamp, directoryProcess, subPath, newTime)
   {
      if(!directoryProcess.isProcess() || !subPath.isString() || !newTime.isScalar())
         simTypeError(directoryProcess, subPath, newTime);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);

      int pathLength = directory->mPath.getLength()
                     + arrays::getStringLength(subPath) + 1;
      char path[pathLength];
      directory->mPath.copyToBuffer(path, false);
      arrays::copyString(path + directory->mPath.getLength(), subPath, true);

      double timestamp = newTime.builtin.value.scalar;
      timeval times[2];
      times[0].tv_usec = (long)(modf(timestamp, &timestamp) * 1000000.0);
      times[0].tv_sec  = (time_t)timestamp;
      times[1] = times[0];

      if(utimes(path, times) != 0)
         return throwOsError(errno, gFileErrorId, subPath);

      return simReturn();
   }

   // interface ======================================================
   bool addRoot(String rootName, String path)
   {
      FW_MAKE_CSTRING(pathCString, path);

      struct stat stats;
      int statResult = stat(pathCString, &stats);

      if(statResult < 0 || !S_ISDIR(stats.st_mode))
         return false;

      if(path[path.getLength() - 1] != '/')
         path += $("/");

      gRoots[rootName] = path;

      return true;
   }

   String getRootPath(String rootName)
   {
      std::map<String, String>::iterator iter = gRoots.find(rootName);
      if(iter == gRoots.end())
         return null;
      else
         return iter->second;
   }
}

// ===================================================================

} //namespace posix
} //namespace api
} //namespace evlan
