/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2005/03/06 (KV)
 *  + First Version (copied from Win32 version)
 */

// ===================================================================

#include "file.h"
#include <map>
#include <evlan/vm/memory.h>
#include <evlan/vm/simulator.h>
#include <evlan/vm/operations.h>
#include <evlan/vm/arrays.h>
#include <evlan/vm/messages.h>
#include <evlan/vm/loader.h>
#include <evlan/vm/shortcuts.h>
#include <evlan/api/error.h>
#include <math.h>

#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <errno.h>

#include "directory.h"

namespace evlan {
namespace api {
namespace posix {
using namespace evlan::vm;
using namespace evlan::vm::shortcuts;
using ::framework::messages::EventManager;
using ::framework::messages::EventManagerPtr;
using ::framework::messages::MessagePtr;
using ::framework::messages::makeMessage;
using ::framework::messages::MessageReceiver;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::api::posix::file
// TODO (KV): Review evlan::api::posix::file

namespace file
{
   struct HandleInfo
   {
      String filename;
      int handle;
      int referenceCount;
      bool canRead;
      bool canWrite;
   };

   /* Multiple processes might refer to the same open file.  Since
    * there is no way to guarantee that one process will be restored
    * before another when restoring state, we need to have some way to
    * keep track of which files are currently open... */
   std::map<String, HandleInfo*> gOpenHandles;

   HandleInfo* openHandle(String filename, bool create)
   {
      std::map<String, HandleInfo*>::iterator iter = gOpenHandles.find(filename);
      if(iter != gOpenHandles.end())
      {
         ++iter->second->referenceCount;
         return iter->second;
      }

      FW_MAKE_CSTRING(filenameCString, filename);

      struct stat stats;
      int statResult = stat(filenameCString, &stats);

      if(statResult >= 0 && S_ISDIR(stats.st_mode))
      {
         errno = EISDIR;
         return null;
      }

      bool canRead  = access(filenameCString, R_OK) == 0;
      bool canWrite = access(filenameCString, W_OK) == 0;

      int fd;

      int flag = create?O_CREAT:0;

      fd = open(filenameCString, O_RDWR | flag, 0777);
      if(fd < 0)
         fd = open(filenameCString, O_RDONLY | flag, 0777);
      if(fd < 0)
         fd = open(filenameCString, O_WRONLY | flag, 0777);
      if(fd < 0)
         return null;

      HandleInfo* result = new HandleInfo;
      result->filename = filename;
      result->handle = fd;
      result->referenceCount = 1;
      result->canRead = canRead;
      result->canWrite = canWrite;
      gOpenHandles[filename] = result;
      return result;
   }

   void closeHandle(HandleInfo* info)
   {
      if(--info->referenceCount == 0)
      {
         close(info->handle);
         gOpenHandles.erase(info->filename);
         delete info;
      }
   }

   EVLAN_NATIVE_IDENTIFIER(gFileErrorId, fileError);

   Value throwError(String path, int errorCode)
   {
      //for security reasons, don't reveal full file path
      int lastSlash = path.findLastOf('/');
      if(lastSlash >= 0)
         path = path.getInterval(lastSlash + 1);
      arrays::makeString(path);
      Value result = error::throwOsError(errorCode, gFileErrorId, *gStackTop);
      simPop();
      return result;
   }

   // FileInfo =======================================================
   implementation FileInfo: implements MessageReceiver
   {
      FileInfo(core::Process* process, HandleInfo* handle, EventManagerPtr eventManager);
      FileInfo(core::Process* process, framework::SystemPtr system);
      ~FileInfo();

      // MessageReceiver ---------------------------------------------
      void receive(MessagePtr message);

      // internal ----------------------------------------------------
      bool doRead();
      bool doWrite();

      core::Process* mProcess;
      HandleInfo* mHandle;
      byte* mBuffer;
      EventManager::AioStreamPtr mAio;

      EventManagerPtr mEventManager;
      MessagePtr mMessage;
   };
   typedef SmartPointer<FileInfo> FileInfoPtr;

   FileInfo::FileInfo(core::Process* process, HandleInfo* handle, EventManagerPtr eventManager)
      : mProcess(process), mHandle(handle), mBuffer(null), mEventManager(eventManager)
   {
      mMessage = makeMessage(this, 0, null);

      mAio = mEventManager->makeAioStream(handle->handle);
   }
   FileInfo::FileInfo(core::Process* process, framework::SystemPtr system)
      : mProcess(process), mBuffer(null), mEventManager(system->getEventManager())
   {
      mMessage = makeMessage(this, 0, null);

      if(mProcess->state.isArray()
      && (mProcess->state.array.size == 2
       || (process->state.array.size == 5
        && mProcess->state.array.elements[3].isScalar()
        && (mProcess->state.array.elements[4].isScalar()
         || mProcess->state.array.elements[4].isPackedArray()))))
      {
         String rootPath = directory::getRootPath(arrays::getString(mProcess->state.array.elements[0]));
         String filename = rootPath + $("/") + arrays::getString(mProcess->state.array.elements[1]);
         mHandle = openHandle(filename, false);

         if(mHandle == null)
            Debug::log(Debug::ERROR) << "File that was open before persistent state restore could not be re-opened:  " << filename;

         if(mProcess->state.array.size == 5)
         {
            mProcess->keepAlive = true;
            mMessage->send();      //gotta init the outstanding I/O later...
         }
         else if(mHandle != null)
            mAio = mEventManager->makeAioStream(mHandle->handle);
      }
      else if(mProcess->state.isArray()
           && (mProcess->state.array.size == 1
            || (process->state.array.size == 4
             && mProcess->state.array.elements[2].isScalar()
             && (mProcess->state.array.elements[3].isScalar()
              || mProcess->state.array.elements[3].isPackedArray()))))
      {
         Debug::log(Debug::ERROR) << "Encountered file with obsolete state:  " << arrays::getString(mProcess->state.array.elements[0]);
         mHandle = null;

         if(mProcess->state.array.size == 4)
         {
            mProcess->keepAlive = true;
            mMessage->send();      //gotta init the outstanding I/O later...
         }
      }
      else
         FW_THROW(spnew FatalError($("File process had invalid state on restore.")));
   }
   FileInfo::~FileInfo()
   {
      mAio = null;
      if(mBuffer != null)
         free(mBuffer);
      if(mHandle != null)
         closeHandle(mHandle);
   }

   void FileInfo::receive(MessagePtr message)
   {
      if(mHandle == null)
      {
         if(mProcess->state.array.size == 4)
            simPush(mProcess->state.array.elements[1]);
         else
            simPush(mProcess->state.array.elements[2]);
         simPush(simThrow(gFileErrorId, $("File no longer exists.")));
         simQueue();

         simPush(mProcess->state);
         if(mProcess->state.array.size == 4)
            gStackTop->array.size = 1;
         else
            gStackTop->array.size = 2;
         memory::modifyProcessState(mProcess);
         mProcess->keepAlive = false;
         return;
      }
      else if(mAio == null)
      {
         mAio = mEventManager->makeAioStream(mHandle->handle);

         bool success;
         if(mProcess->state.array.elements[4].isScalar())
            success = doRead();
         else
            success = doWrite();

         if(!success)
         {
            simPush(mProcess->state.array.elements[2]);
            simPush(throwError(mHandle->filename, errno));
            simQueue();

            simPush(mProcess->state);
            gStackTop->array.size = 2;
            memory::modifyProcessState(mProcess);
         }

         return;
      }

      byte* buffer = mBuffer;
      mBuffer = null;

      simPush(mProcess->state.array.elements[2]);
      bool isRead = mProcess->state.array.elements[4].isScalar();

      simPush(mProcess->state);
      gStackTop->array.size = 2;
      memory::modifyProcessState(mProcess);

      int result;
      if(mAio->getResult(&result))
      {
         if(isRead)
         {
            if(result > 0)
            {
               byte* target = memory::allocate<byte>(result);
               memcpy(target, buffer, result);

               --gStackTop;
               gStackTop->type = Value::PACKED_ARRAY;
               gStackTop->packedArray.packType    = Value::PACKED_DATA;
               gStackTop->packedArray.elementSize = 3;
               gStackTop->packedArray.offset      = 0;
               gStackTop->packedArray.size        = result;
               gStackTop->packedArray.elements    = (word*)target;
            }
            else
               simPush(Value::makeArray(0, null));
         }
         else
            simPush(Value::makeScalar(result));

         *gStackTop = simReturn(*gStackTop);
      }
      else
         simPush(throwError(mHandle->filename, result));

      simQueue();
      free(buffer);
      mProcess->keepAlive = false;
   }

   bool FileInfo::doRead()
   {
      uint64 start = (uint64)mProcess->state.array.elements[3].builtin.value.scalar;
      uint64 end = (uint64)mProcess->state.array.elements[4].builtin.value.scalar;
      int size = end - start;

      mBuffer = (byte*)malloc(size);
      return mProcess->keepAlive = mAio->read(mBuffer, start, size, mMessage);
   }

   bool FileInfo::doWrite()
   {
      uint64 start = (uint64)mProcess->state.array.elements[3].builtin.value.scalar;
      int size = mProcess->state.array.elements[4].packedArray.size;
      byte* data = (byte*)mProcess->state.array.elements[4].packedArray.elements
                 + mProcess->state.array.elements[4].packedArray.offset;

      mBuffer = (byte*)malloc(size);
      memcpy(mBuffer, data, size);
      return mProcess->keepAlive = mAio->write(mBuffer, start, size, mMessage);
   }

   EVLAN_NATIVE_PROCESS(file, FileInfo)

   // interface ======================================================
   EVLAN_NATIVE_FUNCTION_3(fileRead, fileProcess, startPos, endPos)
   {
      if(!fileProcess.isProcess() || !startPos.isScalar() || !endPos.isScalar())
         return simTypeError(fileProcess, startPos, endPos);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));
      if(fileInfo->mBuffer != null)
         return simThrow(gFileErrorId, $("An I/O operation on this file is already in progress."));

      messages::getCurrentContext();

      Value* elements = memory::allocateValues(5);
      elements[0] = fileInfo->mProcess->state.array.elements[0];
      elements[1] = fileInfo->mProcess->state.array.elements[1];
      elements[2] = *gStackTop;
      elements[3] = startPos;
      elements[4] = endPos;
      simPush(Value::makeArray(5, elements));
      memory::modifyProcessState(fileInfo->mProcess);

      if(fileInfo->doRead())
         return Value::makeTask(Value::STOP);
      else
      {
         int error = errno;

         simPush(fileInfo->mProcess->state);
         gStackTop->array.size = 2;
         memory::modifyProcessState(fileInfo->mProcess);

         return throwError(fileInfo->mHandle->filename, error);
      }
   }

   EVLAN_NATIVE_FUNCTION_3(fileWrite, fileProcess, startPos, data)
   {
      if(!fileProcess.isProcess() || !startPos.isScalar())
         return simTypeError(fileProcess, startPos, data);
      if((!data.isArray() && !data.isPackedArray())
      || (data.isArray() && data.array.size != 0)
      || (data.isPackedArray()
       && (data.packedArray.packType != Value::PACKED_DATA
        || data.packedArray.elementSize != 3)))
         return simTypeError(data);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));
      if(fileInfo->mBuffer != null)
         return simThrow(gFileErrorId, $("An I/O operation on this file is already in progress."));

      messages::getCurrentContext();

      Value* elements = memory::allocateValues(5);
      elements[0] = fileInfo->mProcess->state.array.elements[0];
      elements[1] = fileInfo->mProcess->state.array.elements[1];
      elements[2] = *gStackTop;
      elements[3] = startPos;
      elements[4] = data;
      simPush(Value::makeArray(5, elements));
      memory::modifyProcessState(fileInfo->mProcess);

      if(fileInfo->doWrite())
         return Value::makeTask(Value::STOP);
      else
      {
         int error = errno;

         simPush(fileInfo->mProcess->state);
         gStackTop->array.size = 2;
         memory::modifyProcessState(fileInfo->mProcess);

         return throwError(fileInfo->mHandle->filename, error);
      }
   }

   EVLAN_NATIVE_FUNCTION_1(fileGetSize, fileProcess)
   {
      if(!fileProcess.isProcess())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      struct stat stats;
      if(fstat(fileInfo->mHandle->handle, &stats) < 0)
         return throwError(fileInfo->mHandle->filename, errno);

      return simReturn(core::Value::makeScalar(stats.st_size));
   }

   EVLAN_NATIVE_FUNCTION_2(fileSetSize, fileProcess, newSize)
   {
      if(!fileProcess.isProcess() || !newSize.isScalar())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      if(ftruncate(fileInfo->mHandle->handle, (uint64)newSize.builtin.value.scalar) < 0)
         return throwError(fileInfo->mHandle->filename, errno);

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(fileGetTimestamp, fileProcess)
   {
      if(!fileProcess.isProcess())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      struct stat stats;
      if(fstat(fileInfo->mHandle->handle, &stats) < 0)
         return throwError(fileInfo->mHandle->filename, errno);

      return simReturn(core::Value::makeScalar(stats.st_mtime));
   }

   EVLAN_NATIVE_FUNCTION_2(fileSetTimestamp, fileProcess, newTime)
   {
      if(!fileProcess.isProcess() || !newTime.isScalar())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      double timestamp = newTime.builtin.value.scalar;
      timeval times[2];
      times[0].tv_usec = (long)(modf(timestamp, &timestamp) * 1000000.0);
      times[0].tv_sec  = (time_t)timestamp;
      times[1] = times[0];

#ifdef PLATFORM_LINUX
      /* FIXME:  Linux doesn't seem to provide a way to set file times
       * on a descriptor (as opposed to a named file).  We'll silently
       * ignore this call on Linux for now. */
#else
      if(futimes(fileInfo->mHandle->handle, times) != 0)
         return throwError(fileInfo->mHandle->filename, errno);
#endif

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(fileClone, fileProcess)
   {
      if(!fileProcess.isProcess())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      simPush(Value::makeArray(2, fileInfo->mProcess->state.array.elements));
      memory::makeProcess($("file"));
      gStackTop->process.process->nativeInfo =
         spnew FileInfo(gStackTop->process.process, fileInfo->mHandle, fileInfo->mEventManager);
      ++fileInfo->mHandle->referenceCount;

      return simReturn(*gStackTop);
   }

   int make(String path, bool create, framework::SystemPtr system)
   {
      HandleInfo* handle = openHandle(path, create);
      if(handle == null)
         return errno;

      Value* elements = memory::allocateValues(2);
      elements[0] = gStackTop[1];
      elements[1] = gStackTop[0];
      simPop();
      *gStackTop = Value::makeArray(2, elements);
      memory::makeProcess($("file"));

      gStackTop->process.process->nativeInfo =
         spnew FileInfo(gStackTop->process.process, handle, system->getEventManager());

      return 0;
   }
}

// ===================================================================

} //namespace posix
} //namespace api
} //namespace evlan
