/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/05/06 (KV)
 *  + First Version
 */

// ===================================================================

#include "directory.h"
#include <framework/util/Vector.h>
#include <framework/messages/EventManager.h>
#include <evlan/vm/native.h>
#include <evlan/vm/memory.h>
#include <evlan/vm/simulator.h>
#include <evlan/vm/operations.h>
#include <evlan/vm/arrays.h>
#include <evlan/vm/messages.h>
#include <evlan/vm/loader.h>
#include <evlan/vm/shortcuts.h>
#include <evlan/api/error.h>
#include "file.h"

#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <shlobj.h>
#include <map>

namespace evlan {
namespace api {
namespace win32 {
using namespace evlan::vm;
using namespace evlan::vm::shortcuts;
using namespace evlan::api::error;
using framework::common::byte;  //WTF!?

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::api::win32::directory
// TODO (KV): Review evlan::api::win32::directory

namespace directory
{
   /* Evlan programs are not given direct filesystem access.  Instead,
    * the VM provides a number of directories, identified by strings,
    * which may be accessed.  The user decides where these point.
    * This means that a saved Evlan program state can safely be moved
    * between two systems with different directory structures as long
    * as the same set of directories, with the same names, are
    * provided to the VM's. */
   std::map<String, String> gRoots;

   /* invalid:  ., .., con, nul, prn, aux, lpt#, com#,
    * characters: \ / : * ? " < > |
    * ranges: 0x00 through 0x19, 0x7F through 0x9F, 0x100 and up. */
   struct ValidCharacterTable
   {
      bool table[256];
      inline bool operator[](unsigned char index) {return table[index];}

      ValidCharacterTable()
      {
         memset(table, 0, sizeof(table));
         for(int i = 0x20; i < 0x7f; i++) table[i] = true;
         for(int i = 0xa0; i < 0x9f; i++) table[i] = true;
         table['\\'] = false;
         table['/'] = false;
         table[':'] = false;
         table['*'] = false;
         table['?'] = false;
         table['\"'] = false;
         table['<'] = false;
         table['>'] = false;
         table['|'] = false;
      }
   };

   bool isValidPath(Value& path)
   {
      static ValidCharacterTable validTable;

      if(!path.isPackedArray()
      || path.packedArray.elementSize != 3   //disallow unicode for now...
      || path.packedArray.packType != Value::PACKED_CHARACTER
      || path.packedArray.size >= MAX_PATH)
         return false;

      char* buffer = (char*)path.packedArray.elements + path.packedArray.offset;

      int partStart = 0;
      for(int i = 0; ; i++)
      {
         switch(i-partStart)
         {
         case 0:
            //disallow empty components and leading or trailing slashes
            if(i == path.packedArray.size || buffer[i] == '/')
               return false;
            break;
         case 1:
            //disallow .
            if((i == path.packedArray.size || buffer[i] == '/')
            && buffer[i-1] == '.')
               return false;
            break;
         case 2:
            //disallow ..
            if((i == path.packedArray.size || buffer[i] == '/')
            && buffer[i-1] == '.' && buffer[i-2] == '.')
               return false;
            break;
         case 3:
            //disallow con, nul, prn, and aux (DOS device names)
            if(i == path.packedArray.size || buffer[i] == '/' || buffer[i] == '.')
            {
               if(_strnicmp(buffer + partStart, "con", 3) == 0
               || _strnicmp(buffer + partStart, "nul", 3) == 0
               || _strnicmp(buffer + partStart, "prn", 3) == 0
               || _strnicmp(buffer + partStart, "aux", 3) == 0)
                  return false;
            }
            break;
         case 4:
            //disallow com# and lpt# (DOS device names)
            if((i == path.packedArray.size || buffer[i] == '/' || buffer[i] == '.')
            && '0' <= buffer[i-1] && buffer[i-1] <= '9')
            {
               if(_strnicmp(buffer + partStart, "com", 3) == 0
               || _strnicmp(buffer + partStart, "lpt", 3) == 0)
                  return false;
            }
            break;
         default:
            break;
         }

         if(i == path.packedArray.size)
            return true;

         if(buffer[i] == '/')
            partStart = i + 1;
         else if(!validTable[buffer[i]])
            return false;
      }
   }

   String makePath(String root, Value& subPath, bool addSlash)
   {
      int subPathLength = arrays::getStringLength(subPath);
      if(subPathLength == 0)
         addSlash = false;
      Array<char> buffer(root.getLength() + subPathLength + (addSlash?1:0));
      if(buffer != null)
      {
         root.copyToBuffer(&buffer[0], false);
         arrays::copyString(&buffer[0] + root.getLength(), subPath, false);
      }

      int totalLength = root.getLength() + subPathLength;
      for(int i = root.getLength(); i < totalLength; i++)
      {
         if(buffer[i] == '/')
            buffer[i] = '\\';
      }

      if(addSlash)
         buffer[totalLength] = '\\';

      return String(buffer.toConst());
   }

   void makePathCString(char* buffer, String root, Value& subPath)
   {
      root.copyToBuffer(buffer, false);
      arrays::copyString(buffer + root.getLength(), subPath, true);

      int totalLength = root.getLength() + arrays::getStringLength(subPath);
      for(int i = root.getLength(); i < totalLength; i++)
      {
         if(buffer[i] == '/')
            buffer[i] = '\\';
      }
   }

   #define PATH_CSTRING(CSTRING_NAME, ROOT, SUBPATH)                              \
      char CSTRING_NAME[ROOT.getLength() + arrays::getStringLength(SUBPATH) + 1]; \
      makePathCString(CSTRING_NAME, ROOT, SUBPATH)


   EVLAN_NATIVE_IDENTIFIER(gNameNotAllowedId, nameNotAllowed);
   EVLAN_NATIVE_IDENTIFIER(gWrongTypeId, wrongType);
   EVLAN_NATIVE_IDENTIFIER(gFileErrorId, fileError);

   EVLAN_NATIVE_IDENTIFIER(gNotFoundId, notFound);
   EVLAN_NATIVE_IDENTIFIER(gFileId, file);
   EVLAN_NATIVE_IDENTIFIER(gDirectoryId, directory);

   EVLAN_NATIVE_IDENTIFIER(gCurrentId, current);
   EVLAN_NATIVE_IDENTIFIER(gApplicationDataId, applicationData);

   EVLAN_NATIVE_IDENTIFIER(gTypeId, type);
   EVLAN_NATIVE_IDENTIFIER(gExtensionsId, extensions);

   inline Value throwNotAllowed(Value& path)
      {return simThrow(gNameNotAllowedId, gFileErrorId, path, $("The file or directory name is not allowed."));}

   // process info ===================================================
   implementation DirectoryInfo: implements Object
   {
      DirectoryInfo(String path, framework::SystemPtr system);
      DirectoryInfo(core::Process* process, framework::SystemPtr system);
      ~DirectoryInfo();

      String mPath;  //ends with '\'
      framework::SystemPtr mSystem;
   };
   typedef SmartPointer<DirectoryInfo> DirectoryInfoPtr;

   DirectoryInfo::DirectoryInfo(String path, framework::SystemPtr system)
      : mPath(path), mSystem(system) {}

   DirectoryInfo::DirectoryInfo(core::Process* process, framework::SystemPtr system)
      : mSystem(system)
   {
      if(!process->state.isString())
         FW_THROW(spnew FatalError($("Type error in directory process state.")));

      String rootName = arrays::getString(process->state);

      std::map<String, String>::iterator iter = gRoots.find(rootName);
      if(iter == gRoots.end())
         FW_THROW(spnew FatalError($("Open root directory is no longer mounted: ") + rootName));

      mPath = iter->second;
   }

   DirectoryInfo::~DirectoryInfo()
   {}

   EVLAN_NATIVE_PROCESS(directory, DirectoryInfo)

   // glue code ======================================================
   EVLAN_NATIVE_FUNCTION_1(directoryGetRoots, systemProcess)
   {
      if(!systemProcess.isProcess())
         simTypeError(systemProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(system, systemProcess, framework::SystemPtr);

      std::map<String, String>::iterator iter = gRoots.begin();
      int count = 0;
      for(; iter != gRoots.end(); ++iter, ++count)
      {
         arrays::makeString(iter->first);

         simPush(*gStackTop);
         memory::makeProcess($("directory"));
         gStackTop->process.process->nativeInfo = spnew DirectoryInfo(iter->second, system);

         Value* elements = memory::allocateValues(2);
         elements[0] = gStackTop[1];
         elements[1] = gStackTop[0];
         simPop(2);
         simPush(Value::makeArray(2, elements));
      }

      Value* elements = memory::allocateValues(count);
      for(int i = count-1; i >= 0; i--)
      {
         elements[i] = *gStackTop;
         simPop();
      }

      return simReturn(Value::makeArray(count, elements));
   }

   /* subPath is expected to end with a '\' and have already been checked
    * for validity previously. */
   EVLAN_NATIVE_FUNCTION_2(directoryList, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);

      int pathLength = directory->mPath.getLength()
                     + arrays::getStringLength(subPath) + 2;
      char path[pathLength];
      directory->mPath.copyToBuffer(path, false);
      arrays::copyString(path + directory->mPath.getLength(), subPath, false);
      path[pathLength-2] = '*';
      path[pathLength-1] = '\0';

      int fileCount = 0;
      int directoryCount = 0;

      WIN32_FIND_DATA findData;

      HANDLE handle = FindFirstFile(path, &findData);
      if(handle != INVALID_HANDLE_VALUE)
      {
         do
         {
            if(std::strcmp(findData.cFileName, "." ) != 0
            && std::strcmp(findData.cFileName, "..") != 0)
            {
               arrays::makeString(findData.cFileName);

               if(findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
               {
                  ++directoryCount;
                  if(fileCount > 0)
                  {
                     //swap to directory list
                     Value temp = gStackTop[fileCount];
                     gStackTop[fileCount] = *gStackTop;
                     *gStackTop = temp;
                  }
               }
               else
                  ++fileCount;
            }
         } while(FindNextFile(handle, &findData));

         FindClose(handle);
      }

      Value* target = memory::allocateValues(fileCount + directoryCount + 2);
      memcpy(target, gStackTop, (fileCount + directoryCount) * sizeof(Value));
      simPop(fileCount + directoryCount);

      Value* elements = target + fileCount + directoryCount;
      elements[0] = Value::makeArray(fileCount, target);
      elements[1] = Value::makeArray(directoryCount, target + fileCount);
      return simReturn(Value::makeArray(2, elements));
   }

   EVLAN_NATIVE_FUNCTION_2(directoryGetType, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);
      PATH_CSTRING(path, directory->mPath, subPath);

      int id;

      DWORD attributes = GetFileAttributes(path);
      if(attributes == INVALID_FILE_ATTRIBUTES)
      {
         DWORD errorCode = GetLastError();
         if(errorCode == ERROR_FILE_NOT_FOUND
         || errorCode == ERROR_PATH_NOT_FOUND)
            id = gNotFoundId;
         else
            return throwOsError(errorCode, gFileErrorId, subPath);
      }
      else if(attributes & FILE_ATTRIBUTE_DIRECTORY)
         id = gDirectoryId;
      else
         id = gFileId;

      return simReturn(Value::makeAtom(id));
   }

   EVLAN_NATIVE_FUNCTION_2(directoryOpenFile, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);

      String path = makePath(directory->mPath, subPath, false);
      simPush(directoryProcess.process.process->state);
      simPush(subPath);
      DWORD errorCode = file::make(path, false, directory->mSystem);

      if(errorCode == ERROR_SUCCESS)
         return simReturn(*gStackTop);
      else if(errorCode == file::ERROR_IS_DIRECTORY)
         return simThrow(gWrongTypeId, gFileErrorId, subPath, $("Tried to open a directory as a file."));
      else
         return throwOsError(errorCode, gFileErrorId, subPath);
   }

   EVLAN_NATIVE_FUNCTION_2(directoryCreateFile, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);

      String path = makePath(directory->mPath, subPath, false);
      simPush(directoryProcess.process.process->state);
      simPush(subPath);
      DWORD errorCode = file::make(path, true, directory->mSystem);

      if(errorCode == ERROR_SUCCESS)
         return simReturn(*gStackTop);
      else if(errorCode == file::ERROR_IS_DIRECTORY)
         return simThrow(gWrongTypeId, gFileErrorId, subPath, $("Tried to open a directory as a file."));
      else
         return throwOsError(errorCode, gFileErrorId, subPath);
   }

   EVLAN_NATIVE_FUNCTION_2(directoryCreateDirectory, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);
      PATH_CSTRING(path, directory->mPath, subPath);

      if(CreateDirectory(path, null))
         return simReturn();
      else
      {
         DWORD error = GetLastError();
         if(error == ERROR_ALREADY_EXISTS)
            return simReturn();
         return throwOsError(GetLastError(), gFileErrorId, subPath);
      }
   }

   EVLAN_NATIVE_FUNCTION_2(directoryRemoveFile, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);
      PATH_CSTRING(path, directory->mPath, subPath);

      if(DeleteFile(path))
         return simReturn();
      else
         return throwOsError(GetLastError(), gFileErrorId, subPath);
   }

   EVLAN_NATIVE_FUNCTION_2(directoryRemoveDirectory, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);
      if(!isValidPath(subPath))
         return throwNotAllowed(subPath);
      PATH_CSTRING(path, directory->mPath, subPath);

      if(RemoveDirectory(path))
         return simReturn();
      else
         return throwOsError(GetLastError(), gFileErrorId, subPath);
   }

   /* subPath is expected to end with a '\' and have already been checked
    * for validity previously. */
   EVLAN_NATIVE_FUNCTION_2(directoryGetTimestamp, directoryProcess, subPath)
   {
      if(!directoryProcess.isProcess() || !subPath.isString())
         simTypeError(directoryProcess, subPath);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);

      int pathLength = directory->mPath.getLength()
                     + arrays::getStringLength(subPath) + 1;
      char path[pathLength];
      directory->mPath.copyToBuffer(path, false);
      arrays::copyString(path + directory->mPath.getLength(), subPath, true);

      int64 modified;

      HANDLE handle = CreateFile(path, GENERIC_READ,
         FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
         NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);

      if(handle == INVALID_HANDLE_VALUE)
         return throwOsError(GetLastError(), gFileErrorId, subPath);
      BOOL success = GetFileTime(handle, null, null, (FILETIME*)&modified);
      CloseHandle(handle);
      if(!success)
         return throwOsError(GetLastError(), gFileErrorId, subPath);

      modified -= 116444736000000000ll;   //1970

      return simReturn(Value::makeScalar((double)modified / 10000000.0));
   }

   EVLAN_NATIVE_FUNCTION_3(directorySetTimestamp, directoryProcess, subPath, timestamp)
   {
      if(!directoryProcess.isProcess() || !subPath.isString() || !timestamp.isScalar())
         simTypeError(directoryProcess, subPath, timestamp);
      EVLAN_NATIVE_GET_PROCESS_INFO(directory, directoryProcess, DirectoryInfoPtr);

      int pathLength = directory->mPath.getLength()
                     + arrays::getStringLength(subPath) + 1;
      char path[pathLength];
      directory->mPath.copyToBuffer(path, false);
      arrays::copyString(path + directory->mPath.getLength(), subPath, true);

      int64 modified = (int64)(timestamp.builtin.value.scalar * 10000000.0);
      modified += 116444736000000000ll;   //1970

      HANDLE handle = CreateFile(path, GENERIC_WRITE,
         FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
         NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);

      if(handle == INVALID_HANDLE_VALUE)
         return throwOsError(GetLastError(), gFileErrorId, subPath);
      BOOL success = GetFileTime(handle, null, null, (FILETIME*)&modified);
      CloseHandle(handle);
      if(!success)
         return throwOsError(GetLastError(), gFileErrorId, subPath);

      return simReturn();
   }

   // interface ======================================================
   bool addRoot(String rootName, String path)
   {
      FW_MAKE_CSTRING(pathCString, path);

      DWORD attributes = GetFileAttributes(pathCString);
      if(attributes == INVALID_FILE_ATTRIBUTES
      || !(attributes & FILE_ATTRIBUTE_DIRECTORY))
         return false;

      if(path[path.getLength() - 1] != '\\')
         path += $("\\");

      gRoots[rootName] = path;

      return true;
   }

   String getRootPath(String rootName)
   {
      std::map<String, String>::iterator iter = gRoots.find(rootName);
      if(iter == gRoots.end())
         return null;
      else
         return iter->second;
   }
}

// ===================================================================

} //namespace win32
} //namespace api
} //namespace evlan
