/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/05/05 (KV)
 *  + First Version
 */

// ===================================================================

#include "file.h"
#include <map>
#include <evlan/vm/memory.h>
#include <evlan/vm/simulator.h>
#include <evlan/vm/operations.h>
#include <evlan/vm/arrays.h>
#include <evlan/vm/messages.h>
#include <evlan/vm/loader.h>
#include <evlan/vm/shortcuts.h>
#include <evlan/api/error.h>

#include <windows.h>
#undef ERROR

#include "directory.h"

namespace evlan {
namespace api {
namespace win32 {
using namespace evlan::vm;
using namespace evlan::vm::shortcuts;
using ::framework::messages::EventManager;
using ::framework::messages::EventManagerPtr;
using ::framework::messages::MessagePtr;
using ::framework::messages::makeMessage;
using ::framework::messages::MessageReceiver;
using framework::common::byte;  //WTF!?

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::api::win32::file
// TODO (KV): Review evlan::api::win32::file

namespace file
{
   struct Buffer: public OVERLAPPED
   {
      enum
      {
         READ,
         WRITE
      } type;

      core::Process* process;
      ObjectPtr waiter;
      byte buffer[0];
   };

   Buffer* makeBuffer(int size)
   {
      return new(operator new(sizeof(Buffer) + size)) Buffer;
   }

   struct HandleInfo
   {
      String filename;
      HANDLE handle;
      int referenceCount;
   };

   /* Multiple processes might refer to the same open file.  Since
    * there is no way to guarantee that one process will be restored
    * before another when restoring state, we need to have some way to
    * keep track of which files are currently open... */
   std::map<String, HandleInfo*> gOpenHandles;

   /* I really tried to find a way to simply ask Windows what
    * permissions the calling process had on a given file, but after
    * some time searching through the API docs I determined that the
    * process for obtaining this information was ridiculously
    * complicated and not worth learning.  Instead, I just make
    * multiple attepmts at opening the file and take the first one
    * that works.  Yes, I am aware of the _access() function, but I am
    * not confident that it is safe to mix those POSIX-workalikes with
    * true Win32 calls. */
   const DWORD ACCESS_FLAGS[4] =
      {GENERIC_READ | GENERIC_WRITE, GENERIC_READ, GENERIC_WRITE, 0};

   HandleInfo* openHandle(String filename, bool create)
   {
      std::map<String, HandleInfo*>::iterator iter = gOpenHandles.find(filename);
      if(iter != gOpenHandles.end())
      {
         ++iter->second->referenceCount;
         return iter->second;
      }

      FW_MAKE_CSTRING(filenameCString, filename);

      DWORD attributes = GetFileAttributes(filenameCString);
      if(attributes != INVALID_FILE_ATTRIBUTES
      && (attributes & FILE_ATTRIBUTE_DIRECTORY))
      {
         SetLastError(ERROR_IS_DIRECTORY);
         return null;
      }

      for(int i = 0; i < 4; i++)
      {
         HANDLE handle =
            CreateFile(filenameCString, ACCESS_FLAGS[i],
                       FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                       NULL, (create?OPEN_ALWAYS:OPEN_EXISTING),
                       FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
                       NULL);

         if(handle != INVALID_HANDLE_VALUE)
         {
            HandleInfo* result = new HandleInfo;
            result->filename = filename;
            result->handle = handle;
            result->referenceCount = 1;
            gOpenHandles[filename] = result;
            return result;
         }

         DWORD errorCode = GetLastError();

         if(errorCode != ERROR_ACCESS_DENIED
         && errorCode != ERROR_SHARING_VIOLATION)
            return null;
      }

      return null;
   }

   void closeHandle(HandleInfo* info)
   {
      if(--info->referenceCount == 0)
      {
         CloseHandle(info->handle);
         gOpenHandles.erase(info->filename);
         delete info;
      }
   }

   EVLAN_NATIVE_IDENTIFIER(gFileErrorId, fileError);

   Value throwError(String path, DWORD errorCode)
   {
      //for security reasons, don't reveal full file path
      int lastSlash = path.findLastOf('\\');
      if(lastSlash >= 0)
         path = path.getInterval(lastSlash + 1);
      arrays::makeString(path);
      Value result = error::throwOsError(errorCode, gFileErrorId, *gStackTop);
      simPop();
      return result;
   }

   // FileInfo =======================================================
   implementation FileInfo: implements MessageReceiver
   {
      FileInfo(core::Process* process, HandleInfo* handle, EventManagerPtr eventManager);
      FileInfo(core::Process* process, framework::SystemPtr system);
      ~FileInfo();

      // MessageReceiver ---------------------------------------------
      void receive(MessagePtr message);

      // internal ----------------------------------------------------
      void doRead();
      void doWrite();
      static void CALLBACK completionRoutine(DWORD errorCode, DWORD transferCount, OVERLAPPED* win32Overlapped);

      core::Process* mProcess;
      HandleInfo* mHandle;
      Buffer* mBuffer;
      EventManagerPtr mEventManager;

      MessagePtr mStartupMessage;
   };
   typedef SmartPointer<FileInfo> FileInfoPtr;

   FileInfo::FileInfo(core::Process* process, HandleInfo* handle, EventManagerPtr eventManager)
      : mProcess(process), mHandle(handle), mBuffer(null), mEventManager(eventManager)
   {
   }
   FileInfo::FileInfo(core::Process* process, framework::SystemPtr system)
      : mProcess(process), mBuffer(null), mEventManager(system->getEventManager())
   {
      if(mProcess->state.isArray()
      && (mProcess->state.array.size == 2
       || (process->state.array.size == 5
        && mProcess->state.array.elements[3].isScalar()
        && (mProcess->state.array.elements[4].isScalar()
         || mProcess->state.array.elements[4].isPackedArray()))))
      {
         String rootPath = directory::getRootPath(arrays::getString(mProcess->state.array.elements[0]));
         String filename = rootPath + $("/") + arrays::getString(mProcess->state.array.elements[1]);
         mHandle = openHandle(filename, false);

         if(mHandle == null)
            Debug::log(Debug::ERROR) << "File that was open before persistent state restore could not be re-opened:  " << filename;

         if(mProcess->state.array.size == 4)
         {
            mProcess->keepAlive = true;
            mStartupMessage = makeMessage(this, 0, null);
            mStartupMessage->send();
         }
      }
      else if(mProcess->state.isArray()
           && (mProcess->state.array.size == 1
            || (process->state.array.size == 4
             && mProcess->state.array.elements[2].isScalar()
             && (mProcess->state.array.elements[3].isScalar()
              || mProcess->state.array.elements[3].isPackedArray()))))
      {
         Debug::log(Debug::ERROR) << "Encountered file with obsolete state:  " << arrays::getString(mProcess->state.array.elements[0]);
         mHandle = null;

         if(mProcess->state.array.size == 4)
         {
            mProcess->keepAlive = true;
            mStartupMessage = makeMessage(this, 0, null);
            mStartupMessage->send();
         }
      }
      else
         FW_THROW(spnew FatalError($("Type error in file process state.")));
   }
   FileInfo::~FileInfo()
   {
      if(mBuffer != null)
      {
         mBuffer->process = null;
         mBuffer->waiter = null;
      }
      if(mHandle != null)
         closeHandle(mHandle);
   }

   void FileInfo::receive(MessagePtr message)
   {
      if(message == mStartupMessage)
      {
         mStartupMessage->disable();
         mStartupMessage = null;

         if(mHandle == null)
         {
            mProcess->keepAlive = false;
            if(mProcess->state.array.size == 4)
               simPush(mProcess->state.array.elements[1]);
            else
               simPush(mProcess->state.array.elements[2]);
            simPush(simThrow(gFileErrorId, $("File no longer exists.")));
            simQueue();

            simPush(mProcess->state);
            if(mProcess->state.array.size == 4)
               gStackTop->array.size = 1;
            else
               gStackTop->array.size = 2;
            memory::modifyProcessState(mProcess);
            return;
         }
         else if(mProcess->state.array.elements[4].isScalar())
            doRead();
         else
            doWrite();
      }
   }

   void FileInfo::doRead()
   {
      if(mBuffer != null)
         Debug::log(Debug::ERROR) << "Started new I/O on a low-level file before previous I/O completed.";

      mProcess->keepAlive = true;

      uint64 start = (uint64)mProcess->state.array.elements[3].builtin.value.scalar;
      uint64 end = (uint64)mProcess->state.array.elements[4].builtin.value.scalar;
      uint64 size = end - start;

      if(start > end)
         size = 0;
      if(size > 0x0000000080000000ll)
         FW_THROW(spnew FatalError($("Tried to read way too much data from a file at once.")));

      mBuffer = makeBuffer(size);
      mBuffer->Offset = start & 0x00000000FFFFFFFFll;
      mBuffer->OffsetHigh = (start & 0xFFFFFFFF00000000ll) >> 32;
      mBuffer->hEvent = null;
      mBuffer->type = Buffer::READ;
      mBuffer->process = mProcess;
      mBuffer->waiter = mEventManager->getExternalWaiter();

      if(!ReadFileEx(mHandle->handle, mBuffer->buffer, size, mBuffer, completionRoutine))
         completionRoutine(GetLastError(), 0, mBuffer);
   }

   void FileInfo::doWrite()
   {
      if(mBuffer != null)
         Debug::log(Debug::ERROR) << "Started new I/O on a low-level file before previous I/O completed.";

      mProcess->keepAlive = true;

      uint64 start = (uint64)mProcess->state.array.elements[3].builtin.value.scalar;
      int size = mProcess->state.array.elements[4].packedArray.size;
      byte* data = (byte*)mProcess->state.array.elements[4].packedArray.elements
                 + mProcess->state.array.elements[4].packedArray.offset;

      mBuffer = makeBuffer(size);
      mBuffer->Offset = start & 0x00000000FFFFFFFFll;
      mBuffer->OffsetHigh = (start & 0xFFFFFFFF00000000ll) >> 32;
      mBuffer->hEvent = null;
      mBuffer->type = Buffer::WRITE;
      mBuffer->process = mProcess;
      mBuffer->waiter = mEventManager->getExternalWaiter();
      memcpy(mBuffer->buffer, data, size);

      if(!WriteFileEx(mHandle->handle, mBuffer->buffer, size, mBuffer, completionRoutine))
         completionRoutine(GetLastError(), 0, mBuffer);
   }

   void CALLBACK FileInfo::completionRoutine(DWORD errorCode, DWORD transferCount, OVERLAPPED* win32Overlapped)
   {
      Buffer* buffer = (Buffer*)win32Overlapped;

      if(buffer->process == null)
      {
         delete buffer;
         return;
      }

      FileInfoPtr fileInfo =
         buffer->process->nativeInfo.dynamicCast<FileInfoPtr>();
      if(fileInfo->mBuffer != buffer)
      {
         Debug::log(Debug::ERROR) << "On file I/O completion routine, FileInfo's mBuffer did not match the buffer completed.";
         delete buffer;
         return;
      }
      fileInfo->mBuffer = null;

      simPush(buffer->process->state.array.elements[2]);

      simPush(buffer->process->state);
      gStackTop->array.size = 2;
      memory::modifyProcessState(buffer->process);

      if(errorCode == ERROR_SUCCESS || errorCode == ERROR_HANDLE_EOF)
      {
         if(buffer->type == Buffer::READ)
         {
            if(transferCount > 0)
            {
               int wordCount = transferCount / sizeof(word);
               if(transferCount % sizeof(word) > 0)
                  ++wordCount;
               word* target = memory::allocateWords(wordCount);
               memcpy(target, buffer->buffer, transferCount);

               --gStackTop;
               gStackTop->type = Value::PACKED_ARRAY;
               gStackTop->packedArray.packType    = Value::PACKED_DATA;
               gStackTop->packedArray.elementSize = 3;
               gStackTop->packedArray.offset      = 0;
               gStackTop->packedArray.size        = transferCount;
               gStackTop->packedArray.elements    = target;
            }
            else
               simPush(Value::makeArray(0, null));
         }
         else
            simPush(Value::makeScalar((double)transferCount));

         *gStackTop = simReturn(*gStackTop);
      }
      else
         simPush(throwError(fileInfo->mHandle->filename, errorCode));

      simQueue();
      buffer->process->keepAlive = false;
      delete buffer;
   }

   EVLAN_NATIVE_PROCESS(file, FileInfo)

   // interface ======================================================
   EVLAN_NATIVE_FUNCTION_3(fileRead, fileProcess, startPos, endPos)
   {
      if(!fileProcess.isProcess() || !startPos.isScalar() || !endPos.isScalar())
         return simTypeError(fileProcess, startPos, endPos);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      messages::getCurrentContext();

      Value* elements = memory::allocateValues(5);
      elements[0] = fileInfo->mProcess->state.array.elements[0];
      elements[1] = fileInfo->mProcess->state.array.elements[1];
      elements[2] = *gStackTop;
      elements[3] = startPos;
      elements[4] = endPos;
      simPush(Value::makeArray(5, elements));
      memory::modifyProcessState(fileInfo->mProcess);

      fileInfo->doRead();
      return Value::makeTask(Value::STOP);
   }

   EVLAN_NATIVE_FUNCTION_3(fileWrite, fileProcess, startPos, data)
   {
      if(!fileProcess.isProcess() || !startPos.isScalar())
         return simTypeError(fileProcess, startPos, data);
      if((!data.isArray() && !data.isPackedArray())
      || (data.isArray() && data.array.size != 0)
      || (data.isPackedArray()
       && (data.packedArray.packType != Value::PACKED_DATA
        || data.packedArray.elementSize != 3)))
         return simTypeError(data);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      messages::getCurrentContext();

      Value* elements = memory::allocateValues(5);
      elements[0] = fileInfo->mProcess->state.array.elements[0];
      elements[1] = fileInfo->mProcess->state.array.elements[1];
      elements[2] = *gStackTop;
      elements[3] = startPos;
      elements[4] = data;
      simPush(Value::makeArray(5, elements));
      memory::modifyProcessState(fileInfo->mProcess);

      fileInfo->doWrite();
      return Value::makeTask(Value::STOP);
   }

   EVLAN_NATIVE_FUNCTION_1(fileGetSize, fileProcess)
   {
      if(!fileProcess.isProcess())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      uint64 size;
      if(!GetFileSizeEx(fileInfo->mHandle->handle, (LARGE_INTEGER*)&size))
         return throwError(fileInfo->mHandle->filename, GetLastError());

      return simReturn(core::Value::makeScalar((double)size));
   }

   EVLAN_NATIVE_FUNCTION_2(fileSetSize, fileProcess, newSize)
   {
      if(!fileProcess.isProcess() || !newSize.isScalar())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      uint64 size = (uint64)newSize.builtin.value.scalar;

      if(!SetFilePointerEx(fileInfo->mHandle->handle, *(LARGE_INTEGER*)&size, NULL, FILE_BEGIN)
      || !SetEndOfFile(fileInfo->mHandle->handle))
         return throwError(fileInfo->mHandle->filename, GetLastError());

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(fileGetTimestamp, fileProcess)
   {
      if(!fileProcess.isProcess())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      int64 modified;

      if(!GetFileTime(fileInfo->mHandle->handle, null, null, (FILETIME*)&modified))
         return throwError(fileInfo->mHandle->filename, GetLastError());

      modified -= 116444736000000000ll;       //1970

      return simReturn(Value::makeScalar((double)modified / 10000000.0));
   }

   EVLAN_NATIVE_FUNCTION_2(fileSetTimestamp, fileProcess, newTime)
   {
      if(!fileProcess.isProcess() || !newTime.isScalar())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      int64 origin = 116444736000000000ll;    //1970
      int64 modified = (int64)(newTime.builtin.value.scalar * 10000000.0) + origin;

      if(!SetFileTime(fileInfo->mHandle->handle, null, null, (FILETIME*)&modified))
         return throwError(fileInfo->mHandle->filename, GetLastError());

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(fileClone, fileProcess)
   {
      if(!fileProcess.isProcess())
         return simTypeError(fileProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(fileInfo, fileProcess, FileInfoPtr);

      if(fileInfo->mHandle == null)
         return simThrow(gFileErrorId, $("File no longer exists."));

      simPush(Value::makeArray(2, fileInfo->mProcess->state.array.elements));
      memory::makeProcess($("file"));
      gStackTop->process.process->nativeInfo =
         spnew FileInfo(gStackTop->process.process, fileInfo->mHandle, fileInfo->mEventManager);
      ++fileInfo->mHandle->referenceCount;

      return simReturn(*gStackTop);
   }

   DWORD make(String path, bool create, framework::SystemPtr system)
   {
      HandleInfo* handle = openHandle(path, create);
      if(handle == null)
         return GetLastError();

      Value* elements = memory::allocateValues(2);
      elements[0] = gStackTop[1];
      elements[1] = gStackTop[0];
      simPop();
      *gStackTop = Value::makeArray(2, elements);
      memory::makeProcess($("file"));

      gStackTop->process.process->nativeInfo =
         spnew FileInfo(gStackTop->process.process, handle, system->getEventManager());

      return ERROR_SUCCESS;
   }
}

// ===================================================================

} //namespace win32
} //namespace api
} //namespace evlan
