/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/11/28 (KV)
 *  + First Version
 */

// ===================================================================

#include "zlib.h"
#include <zlib.h>
#include <framework/System.h>
#include <evlan/vm/native.h>
#include <evlan/vm/memory.h>
#include <evlan/vm/shortcuts.h>

#ifndef ZLIB_VERNUM
#define ZLIB_VERNUM 0x0000
#endif

namespace evlan {
namespace api {
using namespace evlan::vm;
using namespace evlan::vm::shortcuts;
using ::framework::common::uint;  //I'm told this is needed on OSX Tiger

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::api::zlib
// TODO (KV): Review evlan::api::zlib

namespace zlib
{
   static const int OUT_BUFFER_SIZE = 8192;
   EVLAN_NATIVE_IDENTIFIER(gAlreadyTerminatedId, alreadyTerminated);

   // process info ===================================================
   implementation ZlibInfo: implements Object
   {
      enum Direction {INFLATE, DEFLATE};

      ZlibInfo(Direction direction);
      ZlibInfo(core::Process* process, framework::SystemPtr system);
      ~ZlibInfo();

      Value run(Value& input, int flushFlag);
      Value throwError();

      enum
      {
         NORMAL,
         TERMINATED,
         ERROR,
         RESTORED
      } mStatus;

      Direction mDirection;

      z_stream mState;
      core::Process* mProcess;
   };
   typedef SmartPointer<ZlibInfo> ZlibInfoPtr;

   ZlibInfo::ZlibInfo(Direction direction)
      : mStatus(NORMAL), mDirection(direction)
   {
      memset(&mState, 0, sizeof(z_stream));
   }

   ZlibInfo::ZlibInfo(core::Process* process, framework::SystemPtr system)
      : mStatus(RESTORED), mProcess(process)
   {
   }

   ZlibInfo::~ZlibInfo()
   {
      if(mStatus == NORMAL)
      {
         if(mDirection == INFLATE)
            inflateEnd(&mState);
         else
            deflateEnd(&mState);
      }
   }

   Value ZlibInfo::run(Value& input, int flushFlag)
   {
      if(mStatus == RESTORED)
         return simThrow($("Zlib stream broken after persistent state restore."));
      if(mStatus == TERMINATED)
         return simThrow(gAlreadyTerminatedId, $("Tried to write to zlib stream after it was terminated."));

      int count = 0;
      int offset = input.packedArray.offset;
      int remaining = input.packedArray.size;

      while(true)
      {
         byte* out = (byte*)memory::allocateWords(OUT_BUFFER_SIZE/sizeof(word));
         mState.next_out = (Bytef*)out;
         mState.avail_out = OUT_BUFFER_SIZE;

         byte* pos = (byte*)input.packedArray.elements + offset;
         mState.next_in = (Bytef*)pos;
         mState.avail_in = remaining;

         int result;
         if(mDirection == INFLATE)
            result = inflate(&mState, flushFlag);
         else
            result = deflate(&mState, flushFlag);

         /* Z_BUF_ERROR will apparently occur if the input buffer is
          * empty and there is no data left to output.  This is
          * awfully annoying, as it is in fact impossible to avoid if
          * the previous call exactly filled the output buffer (and
          * thus my code has no way of knowing if there is more data
          * left).  Of course, the documents did not bother to explain
          * this. */
         if(result == Z_BUF_ERROR && remaining == 0)
            result = Z_OK;

         if(result != Z_OK && result != Z_STREAM_END)
         {
            mStatus = ERROR;
            simPop(count);
            if(mDirection == INFLATE)
               inflateEnd(&mState);
            else
               deflateEnd(&mState);
            return throwError();
         }

         offset = (byte*)mState.next_in - (byte*)input.packedArray.elements;
         remaining = mState.avail_in;

         if(mState.avail_out < (uint)OUT_BUFFER_SIZE)
         {
            //Some data was output.  Put it on the stack.
            simPush(Value::makeByteArray(OUT_BUFFER_SIZE - mState.avail_out, out));
            ++count;
         }

         if(mState.avail_out > 0 || result == Z_STREAM_END)
         {
            /* Output buffer is not full.  Therefore, there is no more
             * data to receive for now.  End the loop. */
            int unusedWords = mState.avail_out/sizeof(word);
            word* endpoint = (word*)(out + OUT_BUFFER_SIZE) - unusedWords;
            memory::returnWords(endpoint, unusedWords);

            if(result == Z_STREAM_END)
            {
               mStatus = TERMINATED;
               if(mDirection == INFLATE)
                  inflateEnd(&mState);
               else
                  deflateEnd(&mState);
            }
            break;
         }
      }

      Value* elements = memory::allocateValues(count);
      for(int i = count-1; i >= 0; i--)
      {
         elements[i] = *gStackTop;
         simPop();
      }

      return simReturn(Value::makeArray(count, elements));
   }

   Value ZlibInfo::throwError()
   {
      if(mState.msg == null)
         return simThrow($("Unknown error in zlib process (no error message return by zlib)."));
      else
         return simThrow(String(mState.msg));
   }

   EVLAN_NATIVE_PROCESS(zlibStream, ZlibInfo)

   // glue code ======================================================
   EVLAN_NATIVE_FUNCTION_1(zlibMakeDeflator, level)
   {
      if(!level.isInteger())
         return simTypeError(level);

      ZlibInfoPtr info = spnew ZlibInfo(ZlibInfo::DEFLATE);
      if(deflateInit(&info->mState, (int)level.builtin.value.scalar) != Z_OK)
      {
         info->mStatus = ZlibInfo::ERROR;
         return info->throwError();
      }

      simPush(Value::makeEmpty());
      memory::makeProcess($("zlibStream"));
      gStackTop->process.process->nativeInfo = info;
      info->mProcess = gStackTop->process.process;

      return simReturn(*gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_0(zlibMakeInflator)
   {
      ZlibInfoPtr info = spnew ZlibInfo(ZlibInfo::INFLATE);
      if(inflateInit(&info->mState) != Z_OK)
      {
         info->mStatus = ZlibInfo::ERROR;
         return info->throwError();
      }

      simPush(Value::makeEmpty());
      memory::makeProcess($("zlibStream"));
      gStackTop->process.process->nativeInfo = info;
      info->mProcess = gStackTop->process.process;

      return simReturn(*gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_1(zlibMakeGzipper, level)
   {
      if(!level.isInteger())
         return simTypeError(level);
      if(ZLIB_VERNUM < 0x1200)
         return simThrow($("zlib version is insufficient.  Please upgrade to zlib 1.2.x and recompile Evlan."));

      ZlibInfoPtr info = spnew ZlibInfo(ZlibInfo::DEFLATE);
      if(deflateInit2(&info->mState, (int)level.builtin.value.scalar, Z_DEFLATED,
                      31, 8, Z_DEFAULT_STRATEGY) != Z_OK)
      {
         info->mStatus = ZlibInfo::ERROR;
         return info->throwError();
      }

      simPush(Value::makeEmpty());
      memory::makeProcess($("zlibStream"));
      gStackTop->process.process->nativeInfo = info;
      info->mProcess = gStackTop->process.process;

      return simReturn(*gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_0(zlibMakeGunzipper)
   {
      if(ZLIB_VERNUM < 0x1200)
         return simThrow($("zlib version is insufficient.  Please upgrade to zlib 1.2.x and recompile Evlan."));

      ZlibInfoPtr info = spnew ZlibInfo(ZlibInfo::INFLATE);
      if(inflateInit2(&info->mState, 31) != Z_OK)
      {
         info->mStatus = ZlibInfo::ERROR;
         return info->throwError();
      }

      simPush(Value::makeEmpty());
      memory::makeProcess($("zlibStream"));
      gStackTop->process.process->nativeInfo = info;
      info->mProcess = gStackTop->process.process;

      return simReturn(*gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_2(zlibWrite, zlibProcess, buffer)
   {
      if(!zlibProcess.isProcess() || !buffer.isByteArray())
         return simTypeError(zlibProcess, buffer);
      EVLAN_NATIVE_GET_PROCESS_INFO(zlibInfo, zlibProcess, ZlibInfoPtr);

      return zlibInfo->run(buffer, Z_NO_FLUSH);
   }

   EVLAN_NATIVE_FUNCTION_1(zlibFlush, zlibProcess)
   {
      if(!zlibProcess.isProcess())
         return simTypeError(zlibProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(zlibInfo, zlibProcess, ZlibInfoPtr);

      Value buffer = Value::makeByteArray(0, null);
      return zlibInfo->run(buffer, Z_SYNC_FLUSH);
   }

   EVLAN_NATIVE_FUNCTION_1(zlibTerminate, zlibProcess)
   {
      if(!zlibProcess.isProcess())
         return simTypeError(zlibProcess);
      EVLAN_NATIVE_GET_PROCESS_INFO(zlibInfo, zlibProcess, ZlibInfoPtr);

      Value buffer = Value::makeByteArray(0, null);
      return zlibInfo->run(buffer, Z_FINISH);
   }
}

// ===================================================================

} //namespace api
} //namespace evlan
