/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/01/12 (KV)
 *  + First Version
 */

// ===================================================================

#include "Bytecode.h"

namespace evlan {
namespace compiler {

// ===================================================================
// AUTHOR (KV): Created module.

Bytecode::Bytecode() {}
Bytecode::Bytecode(Array<const Array<const uint32> > dataTable,
                   Array<const String> importTable,
                   Array<const String> memberNameTable,
                   Array<const String> atomTable,
                   Array<const ObjectClass> classes,
                   Array<const Instruction> instructions,
                   int rootInstruction)
   : dataTable(dataTable),
     importTable(importTable),
     memberNameTable(memberNameTable),
     atomTable(atomTable),
     classes(classes),
     instructions(instructions),
     rootInstruction(rootInstruction)
{}
Bytecode::~Bytecode() {}

namespace
{
   template <typename Element>
   Array<Element> copyArray(Array<const Element> input)
   {
      Array<Element> result(input.getSize());
      for(int i = 0; i < input.getSize(); i++)
         result[i] = input[i];
      return result;
   }

   template <typename Element>
   Array<Array<Element> > copyArrayOfArrays(Array<const Array<const Element> > input)
   {
      Array<Array<Element> > result(input.getSize());
      for(int i = 0; i < input.getSize(); i++)
         result[i] = copyArray(input[i]);
      return result;
   }
}

MutableBytecode::MutableBytecode() {}
MutableBytecode::MutableBytecode(Array<Array<const uint32> > dataTable,
                                 Array<String> importTable,
                                 Array<String> memberNameTable,
                                 Array<String> atomTable,
                                 Array<ObjectClass> classes,
                                 Array<Instruction> instructions,
                                 int rootInstruction)
   : dataTable(dataTable),
     importTable(importTable),
     memberNameTable(memberNameTable),
     atomTable(atomTable),
     classes(classes),
     instructions(instructions),
     rootInstruction(rootInstruction)
{}
MutableBytecode::MutableBytecode(BytecodePtr bytecode)
{
   dataTable = copyArray(bytecode->dataTable);
   importTable = copyArray(bytecode->importTable);
   memberNameTable = copyArray(bytecode->memberNameTable);
   atomTable = copyArray(bytecode->atomTable);
   classes = copyArray(bytecode->classes);

   instructions = Array<Instruction>(bytecode->instructions.getSize());
   for(int i = 0; i < bytecode->instructions.getSize(); i++)
   {
      instructions[i] = Instruction(
         bytecode->instructions[i].type,
         copyArray(bytecode->instructions[i].inputs),
         bytecode->instructions[i].function,
         bytecode->instructions[i].branch,
         bytecode->instructions[i].line,
         bytecode->instructions[i].column);
   }

   rootInstruction = bytecode->rootInstruction;
}
MutableBytecode::~MutableBytecode() {}

BytecodePtr MutableBytecode::toConst()
{
   return spnew Bytecode(
      dataTable.toConst(),
      importTable.toConst(),
      memberNameTable.toConst(),
      atomTable.toConst(),
      classes.toConst(),
      instructions.reinterpretCast<Array<Bytecode::Instruction> >().toConst(),
      rootInstruction);
}

#ifdef INCLUDE_WRITER_CODE
void writeInstruction(framework::TextWriterPtr out, Bytecode::Instruction instruction)
{
   FW_MATCH_START(instruction.type)
   // immediates --------------------------------------------------
   FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadTrue)
      out->write($("loadTrue"));
   FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadFalse)
      out->write($("loadFalse"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, i)
      out->write($("loadInteger("));
      out->write(String(i));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, i)
      out->write($("loadInteger64("));
      out->write(String(i));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x)
      out->write($("loadReal("));
      out->write(String(x));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData1, d)
      out->write($("loadData1(0b"));
      out->write(d?$("1"):$("0"));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData2, d)
      out->write($("loadData2(0b"));
      out->write((d&2)?$("1"):$("0"));
      out->write((d&1)?$("1"):$("0"));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData4, d)
      out->write($("loadData4(0x"));
      out->write(String((uint32)(d&0x0000000F), 1, 16));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData8, d)
      out->write($("loadData8(0x"));
      out->write(String((uint32)(d&0x000000FF), 2, 16));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData16, d)
      out->write($("loadData16(0x"));
      out->write(String((uint32)(d&0x0000FFFF), 4, 16));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData32, d)
      out->write($("loadData32(0x"));
      out->write(String(d, 8, 16));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData64, d)
      out->write($("loadData64(0x"));
      out->write(String(d, 16, 16));
      out->write($(")"));
   FW_MATCH_CASE_2(Bytecode::Instruction::Type::loadData, size, index)
      out->write($("loadData("));
      out->write(String(size));
      out->write($(", "));
      out->write(String(index));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadCharacter, c)
      out->write($("loadCharacter("));
      if(c < ' ')
      {
         switch(c)
         {
         case '\a':
            out->write($("\'\\a\'"));
            break;
         case '\b':
            out->write($("\'\\b\'"));
            break;
         case '\f':
            out->write($("\'\\f\'"));
            break;
         case '\n':
            out->write($("\'\\n\'"));
            break;
         case '\r':
            out->write($("\'\\r\'"));
            break;
         case '\t':
            out->write($("\'\\t\'"));
            break;
         case '\v':
            out->write($("\'\\v\'"));
            break;
         case '\0':
            out->write($("\'\\0\'"));
            break;
         default:
            out->write($("0x"));
            out->write(String(c, 2, 16));
            break;
         }
      }
      else if(c <= 0x000000FF)
      {
         out->write($("\'"));
         out->write((char)c, 1);
         out->write($("\'"));
      }
      else
      {
         out->write($("0x"));
         out->write(String(c, 8, 16));
      }
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadAtom, index)
      out->write($("loadAtom("));
      out->write(String(index));
      out->write($(")"));
   FW_MATCH_CASE_3(Bytecode::Instruction::Type::loadArray, type, size, index)
      out->write($("loadArray("));
      out->write(String(type));
      out->write($(", "));
      out->write(String(size));
      out->write($(", "));
      out->write(String(index));
      out->write($(")"));

   // environmentals ----------------------------------------------
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadImport, index)
      out->write($("loadImport("));
      out->write(String(index));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadParameter, index)
      out->write($("loadParameter("));
      out->write(String(index));
      out->write($(")"));

   // computations ------------------------------------------------
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::makeArray, size)
      out->write($("makeArray("));
      out->write(String(size));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::makeObject, index)
      out->write($("makeObject("));
      out->write(String(index));
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::makeFunction, parameters)
      out->write($("makeFunction("));
      for(int i = 0; i < parameters.getSize(); i++)
      {
         if(i > 0)
            out->write($(", "));
         out->write(parameters[i]);
      }
      out->write($(")"));
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadMember, index)
      out->write($("loadMember("));
      out->write(String(index));
      out->write($(")"));
   FW_MATCH_CASE_0(Bytecode::Instruction::Type::apply)
      out->write($("apply"));
   FW_MATCH_CASE_0(Bytecode::Instruction::Type::testCondition)
      out->write($("testCondition"));
   FW_MATCH_END

   out->write($(" f"));
   out->write(String(instruction.function));
   out->write($("b"));
   out->write(String(instruction.branch));

   out->write($(" {"));
   for(int i = 0; i < instruction.inputs.getSize(); i++)
   {
      if(i != 0)
         out->write($(", "));
      out->write(String(instruction.inputs[i]));
   }
   out->write($("}"));
}

void writeBytecode(framework::TextWriterPtr out, BytecodePtr bytecode)
{
   out->write($("data table:\n"));
   for(int i = 0; i < bytecode->dataTable.getSize(); i++)
   {
      out->write(String(i, 4));
      out->write($(": "));
      for(int j = 0; j < bytecode->dataTable[i].getSize(); j++)
      {
         if(j > 0 && j % 8 == 0)
            out->write($("\n      "));
         out->write($(" "));
         out->write(String(bytecode->dataTable[i][j], 8, 16));
      }
      out->write($("\n"));
   }

   out->write($("\nimport table:\n"));
   for(int i = 0; i < bytecode->importTable.getSize(); i++)
   {
      out->write(String(i, 4));
      out->write($(": "));
      out->write(bytecode->importTable[i]);
      out->write($("\n"));
   }

   out->write($("\nmember name table:\n"));
   for(int i = 0; i < bytecode->memberNameTable.getSize(); i++)
   {
      out->write(String(i, 4));
      out->write($(": "));
      out->write(bytecode->memberNameTable[i]);
      out->write($("\n"));
   }

   out->write($("\natom table:\n"));
   for(int i = 0; i < bytecode->atomTable.getSize(); i++)
   {
      out->write(String(i, 4));
      out->write($(": "));
      out->write(bytecode->atomTable[i]);
      out->write($("\n"));
   }

   out->write($("\nclasses:\n"));
   for(int i = 0; i < bytecode->classes.getSize(); i++)
   {
      out->write(String(i, 4));
      out->write($(":"));
      int pos = 5;
      for(int j = 0; j < bytecode->classes[i].memberTable.getSize(); j++)
      {
         String name = bytecode->memberNameTable[bytecode->classes[i].memberTable[j]];
         if(pos + name.getLength() + 1 > 80)
         {
            out->write($("\n     "));
            pos = 0;
         }
         out->write($(" "));
         out->write(name);
         pos += name.getLength() + 1;
      }
      out->write($("\n"));
   }

   out->write($("\ninstructions:\n"));
   for(int i = 0; i < bytecode->instructions.getSize(); i++)
   {
      out->write(String(i, 4));
      out->write($(": "));
      writeInstruction(out, bytecode->instructions[i]);
      out->write($("\n"));
   }

   out->write($("\nroot instruction: "));
   out->write(String(bytecode->rootInstruction));
   out->write($("\n"));
}
#endif

// ===================================================================

} //namespace compiler
} //namespace evlan
