/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/01/12 (KV)
 *  + First Version
 */

// ===================================================================

#ifndef evlan_compiler_Bytecode_h__INCLUDED__
#define evlan_compiler_Bytecode_h__INCLUDED__

#include <framework/util/Pattern.h>
#ifdef INCLUDE_WRITER_CODE
#include <framework/TextWriter.h>
#endif

namespace evlan {
namespace compiler {
using namespace ::framework::common;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Document evlan::compiler::frontend::Bytecode
// TODO (KV): Test evlan::compiler::frontend::Bytecode

struct Bytecode: implements Object
{
   struct Instruction;
   struct ObjectClass;

   Bytecode();
   Bytecode(Array<const Array<const uint32> > dataTable,
            Array<const String> importTable,
            Array<const String> memberNameTable,
            Array<const String> atomTable,
            Array<const ObjectClass> classes,
            Array<const Instruction> instructions,
            int rootInstruction);
   ~Bytecode();

   enum ArrayType
   {
      BOOLEAN,
      INTEGER8,
      INTEGER16,
      INTEGER32,
      INTEGER64,
      REAL,
      CHARACTER8,
      CHARACTER16,
      CHARACTER32,
      DATA        /* for data* types, type = DATA + n, where each
                   * data item has 2^n bits. */
   };

   struct Instruction
   {
      FW_PATTERN_START(Type)
         // immediates -----------------------------------------------
         FW_PATTERN_VALUE_0(loadTrue)
         FW_PATTERN_VALUE_0(loadFalse)

         FW_PATTERN_VALUE_1(loadInteger, int32)
         FW_PATTERN_VALUE_1(loadInteger64, int64)

         FW_PATTERN_VALUE_1(loadReal, double)

         FW_PATTERN_VALUE_1(loadData1, uint8)
         FW_PATTERN_VALUE_1(loadData2, uint8)
         FW_PATTERN_VALUE_1(loadData4, uint8)
         FW_PATTERN_VALUE_1(loadData8, uint8)
         FW_PATTERN_VALUE_1(loadData16, uint16)
         FW_PATTERN_VALUE_1(loadData32, uint32)
         FW_PATTERN_VALUE_1(loadData64, uint64)
         /* This last one must ONLY be used for datas of 64 bits or
          * larger.  (Yes, that means there are two ways to load
          * 64-bit datas.) */
         FW_PATTERN_VALUE_2(loadData, int /* lg(bitCount) */,
                                      int /* data table index */)

         FW_PATTERN_VALUE_1(loadCharacter, uint32)

         FW_PATTERN_VALUE_1(loadAtom,   int /* atom table index */)

         FW_PATTERN_VALUE_3(loadArray, int /* type */, int /* size */,
                                       int /* data table index */)

         // environmentals -------------------------------------------
         //load module from import table
         FW_PATTERN_VALUE_1(loadImport, int /* import table index */)
         //load parameter: input is associated makeFunction instruction
         FW_PATTERN_VALUE_1(loadParameter, int /* parameter number */)

         // computations ---------------------------------------------
         //make array: inputs are elements
         FW_PATTERN_VALUE_1(makeArray, int /* size */)
         //make object: inputs are members
         FW_PATTERN_VALUE_1(makeObject, int /* class index */)
         //make function: input is function's root instruction
         FW_PATTERN_VALUE_2(makeFunction, Array<const String> /* parameters */,
                                          String /* documentation */)

         //load member of input object
         FW_PATTERN_VALUE_1(loadMember, int /* member name index */)

         //apply function: first input is function, rest are parameters
         FW_PATTERN_VALUE_0(apply)

         //if/then/else: inputs are condition, positive result, and negative result
         FW_PATTERN_VALUE_0(testCondition)
      FW_PATTERN_END

      Type type;
      Array<const int> inputs;

      int function;
      int branch;

      int line;
      int column;

      inline Instruction() {}
      inline Instruction(Type type,
                         Array<const int> inputs,
                         int function, int branch,
                         int line, int column)
         : type(type),
           inputs(inputs),
           function(function),
           branch(branch),
           line(line),
           column(column)
      {}
   };

   struct ObjectClass
   {
      //Each integer is an index into the member name table.
      Array<const int> memberTable;
      //Documentation for the object, if given in the source code.
      String documentation;

      inline ObjectClass(Array<const int> memberTable, String documentation)
         : memberTable(memberTable), documentation(documentation) {}
      inline ObjectClass() {}

      inline bool operator<(Bytecode::ObjectClass other) const
      {
         if(Array<const int>::SequenceOrdering()(memberTable, other.memberTable))
            return true;
         else if(Array<const int>::SequenceOrdering()(other.memberTable, memberTable))
            return false;
         else
            return documentation < other.documentation;
      }
   };

   /* The data table contains raw binary data which may be used in
    * arrays or as Data* types larger than 64-bit. */
   Array<const Array<const uint32> > dataTable;
   /* The import table contains the path and name of every module
    * referenced by an import expression in this module. */
   Array<const String> importTable;
   /* The member name table contains every member name referenced
    * within this module. */
   Array<const String> memberNameTable;
   /* The atom table contains every atom name used within this module. */
   Array<const String> atomTable;
   /* This one contains information about each type of object found
    * in the module. */
   Array<const ObjectClass> classes;
   /* This contains the actual instruction graph for the module. */
   Array<const Instruction> instructions;

   /* Index of the module's root instruction. */
   int rootInstruction;
};
typedef SmartPointer<const Bytecode> BytecodePtr;

/* A template for you!
   FW_MATCH_START(instruction.type)
   // immediates -----------------------------------------------------
   FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadTrue)
   FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadFalse)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData1, d)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData2, d)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData4, d)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData8, d)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData16, d)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData32, d)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData64, d)
   FW_MATCH_CASE_2(Bytecode::Instruction::Type::loadData, size, index)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadCharacter, c)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadString, index)
   FW_MATCH_CASE_3(Bytecode::Instruction::Type::loadArray, type, size, index)

   // environmentals -------------------------------------------------
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadImport, index)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadParameter, index)

   // computations ---------------------------------------------------
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::makeArray, size)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::makeObject, index)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::makeFunction, parameters)
   FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadMember, index)
   FW_MATCH_CASE_0(Bytecode::Instruction::Type::apply)
   FW_MATCH_CASE_0(Bytecode::Instruction::Type::testCondition)
   FW_MATCH_END
*/

struct MutableBytecode: implements Object
{
   struct Instruction;
   typedef Bytecode::ObjectClass ObjectClass;

   MutableBytecode();
   MutableBytecode(Array<Array<const uint32> > dataTable,
                   Array<String> importTable,
                   Array<String> memberNameTable,
                   Array<String> atomTable,
                   Array<ObjectClass> classes,
                   Array<Instruction> instructions,
                   int rootInstruction);
   MutableBytecode(BytecodePtr bytecode);
   ~MutableBytecode();

   BytecodePtr toConst();

   struct Instruction
   {
      typedef Bytecode::Instruction::Type Type;

      Type type;
      Array<int> inputs;

      int function;
      int branch;

      int line;
      int column;

      inline Instruction() {}
      inline Instruction(Type type,
                         Array<int> inputs,
                         int function, int branch,
                         int line, int column)
         : type(type),
           inputs(inputs),
           function(function),
           branch(branch),
           line(line),
           column(column)
      {}
   };

   Array<Array<const uint32> > dataTable;
   Array<String> importTable;
   Array<String> memberNameTable;
   Array<String> atomTable;
   Array<ObjectClass> classes;
   Array<Instruction> instructions;

   int rootInstruction;
};
typedef SmartPointer<MutableBytecode> MutableBytecodePtr;

#ifdef INCLUDE_WRITER_CODE
void writeBytecode(framework::TextWriterPtr out, BytecodePtr bytecode);
void writeInstruction(framework::TextWriterPtr out, Bytecode::Instruction instruction);
#endif

// ===================================================================

} //namespace compiler
} //namespace evlan

#endif //included
