/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/01/23 (KV)
 *  + First Version
 */

// ===================================================================

#include "CompilerImpl.h"
#include <evlan/compiler/syntax/Tokenizer.h>
#include <evlan/compiler/syntax/BlockParser.h>
#include <evlan/compiler/syntax/TokenFilter.h>
#include <evlan/compiler/grammar/Parser.h>
#include <evlan/compiler/dialects/basic/BasicDialect.h>
#include <evlan/compiler/backend/makeBytecode.h>
#include <evlan/compiler/backend/cleanupMemberAccess.h>
#include <evlan/compiler/backend/cleanupUnusedInstructions.h>
#include <evlan/compiler/backend/cleanupUnusedTables.h>
#include <evlan/compiler/backend/optimizeArithmetic.h>
#include <evlan/compiler/backend/optimizeArrays.h>
#include <evlan/compiler/backend/optimizeMemberTables.h>
//#include <evlan/compiler/backend/optimizeRedundantExpressions.h>

namespace evlan {
namespace compiler {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::compiler::CompilerImpl
// TODO (KV): Review evlan::compiler::CompilerImpl

implementation CompilerImpl: implements Compiler
{
   CompilerImpl(ErrorCollectorPtr errorCollector,
                dialects::DialectPtr dialect,
                ProcessingLevel processingLevel);
   ~CompilerImpl();

   // Compiler -------------------------------------------------------
   void setDialect(Dialect dialect);
   void setProcessingLevel(ProcessingLevel level);

   Array<const Token> tokenize(String source);

   Expression parse(String source);
   Expression parse(Array<const Token> tokens);

   BytecodePtr compile(String source);
   BytecodePtr compile(Array<const Token> tokens);
   BytecodePtr compile(Expression parseTree);

   // internal -------------------------------------------------------
   ErrorCollectorPtr mErrorCollector;
   dialects::DialectPtr mDialect;
   ProcessingLevel mProcessingLevel;
};

CompilerImpl::CompilerImpl(ErrorCollectorPtr errorCollector,
                           dialects::DialectPtr dialect,
                           ProcessingLevel processingLevel)
   : mErrorCollector(errorCollector),
     mDialect(dialect),
     mProcessingLevel(processingLevel)
{
}

CompilerImpl::~CompilerImpl()
{
}

// Compiler ----------------------------------------------------------
Array<const Token> CompilerImpl::tokenize(String source)
{
   int errorCount = mErrorCollector->getErrorCount();

   syntax::KeywordSetPtr keywords = mDialect->getKeywordSet();
   Array<const Token> tokens;
   tokens = syntax::tokenize(source, mErrorCollector, keywords);
   tokens = syntax::parseBlocks(tokens, mErrorCollector, keywords);
   tokens = syntax::filterTokens(tokens);

   if(mErrorCollector->getErrorCount() > errorCount)
      FW_THROW(spnew HadErrors());
   return tokens;
}

Expression CompilerImpl::parse(String source)
{
   return parse(tokenize(source));
}
Expression CompilerImpl::parse(Array<const Token> tokens)
{
   int errorCount = mErrorCollector->getErrorCount();
   Expression parseTree =
      mDialect->makeParser()
              ->parse(tokens, mErrorCollector)
              ->toExpression(grammar::Environment(), mErrorCollector);
   if(mErrorCollector->getErrorCount() > errorCount)
      FW_THROW(spnew HadErrors());

   return parseTree;
}

BytecodePtr CompilerImpl::compile(String source)
{
   return compile(parse(source));
}
BytecodePtr CompilerImpl::compile(Array<const Token> tokens)
{
   return compile(parse(tokens));
}
BytecodePtr CompilerImpl::compile(Expression parseTree)
{
   //TODO (KV):  Optimize...  more!
   int errorCount = mErrorCollector->getErrorCount();

   MutableBytecodePtr bytecode;
   bytecode = backend::makeBytecode(parseTree, mErrorCollector);
   if(mProcessingLevel >= CLEANUP)
   {
      backend::cleanupMemberAccess(bytecode, mErrorCollector);
      if(mProcessingLevel >= OPTIMIZE)
      {
         backend::optimizeArithmetic(bytecode);
         backend::optimizeArrays(bytecode);
         backend::optimizeMemberTables(bytecode);
//         backend::optimizeRedundantExpressions(bytecode);
      }
      backend::cleanupUnusedInstructions(bytecode);
      backend::cleanupUnusedTables(bytecode);
   }
   else
      mErrorCollector->addWarning(0, 0, $("Bytecode clean-up skipped.  Output may be incorrect."));

   if(mErrorCollector->getErrorCount() > errorCount)
      FW_THROW(spnew HadErrors());
   return bytecode->toConst();
}

// factory function ==================================================
CompilerPtr makeCompilerImpl(ErrorCollectorPtr errorCollector)
{
   return spnew CompilerImpl(errorCollector,
      dialects::basic::makeBasicDialect(), Compiler::OPTIMIZE);
}

CompilerPtr makeCompilerImpl(ErrorCollectorPtr errorCollector,
                             Compiler::Dialect dialect,
                             Compiler::ProcessingLevel processingLevel)
{
   dialects::DialectPtr realDialect;
   realDialect = dialects::basic::makeBasicDialect();
   return spnew CompilerImpl(errorCollector, realDialect, processingLevel);
}

CompilerPtr makeCompilerImpl(ErrorCollectorPtr errorCollector,
                             dialects::DialectPtr dialect,
                             Compiler::ProcessingLevel processingLevel)
{
   return spnew CompilerImpl(errorCollector, dialect, processingLevel);
}

// ===================================================================

} //namespace compiler
} //namespace evlan
