/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Expression.h"

// begin namespace ===================================================
namespace evlan {
namespace compiler {

#ifdef INCLUDE_WRITER_CODE
namespace
{
   void printExpression(framework::TextWriterPtr out, Expression expression, String indent)
   {
      String subIndent = indent + $("   ");
      out->write(String(expression.line));
      out->write($(":"));
      out->write(String(expression.column));
      out->write($(":"));

      FW_MATCH_START(expression.value)
      FW_MATCH_CASE_1(Expression::Value::import, name)
         out->write($("import \""));
         out->write(name);
         out->write($("\"\n"));

      FW_MATCH_CASE_1(Expression::Value::literal, literal)
         out->write($("literal "));
         FW_MATCH_START(literal)
         FW_MATCH_CASE_1(Expression::Literal::boolean, flag)
            if(flag)
               out->write($("true\n"));
            else
               out->write($("false\n"));
         FW_MATCH_CASE_1(Expression::Literal::integer, i)
            out->write($("integer "));
            out->write(String(i));
            out->write($("\n"));
         FW_MATCH_CASE_1(Expression::Literal::integer64, i)
            out->write($("integer64 "));
            out->write(String(i));
            out->write($("\n"));
         FW_MATCH_CASE_1(Expression::Literal::real, x)
            out->write($("real "));
            out->write(String(x));
            out->write($("\n"));
         FW_MATCH_CASE_1(Expression::Literal::data8,  d)
            out->write($("data8 "));
            out->write(String((uint32)d, 2, 16));
            out->write($("\n"));
         FW_MATCH_CASE_1(Expression::Literal::data16, d)
            out->write($("data16 "));
            out->write(String((uint32)d, 4, 16));
            out->write($("\n"));
         FW_MATCH_CASE_1(Expression::Literal::data32, d)
            out->write($("data32 "));
            out->write(String(d, 8, 16));
            out->write($("\n"));
         FW_MATCH_CASE_1(Expression::Literal::data64, d)
            out->write($("data64 "));
            out->write(String(d, 16, 16));
            out->write($("\n"));
         FW_MATCH_CASE_1(Expression::Literal::character, c)
            //FIXME:  Too lazy to do this properly
            if(c == '\'')
               out->write($("character '\\\''\n"));
            else if(c == '\\')
               out->write($("character '\\\\'\n"));
            else if(c >= ' ' && c < 0x7F)
            {
               out->write($("character '"));
               out->write((char)c, 1);
               out->write($("'\n"));
            }
            else
            {
               out->write($("character 0x"));
               out->write(String(c, 8, 16));
               out->write($("\n"));
            }
         FW_MATCH_CASE_1(Expression::Literal::atom, name)
            out->write($("atom "));
            out->write(name);
            out->write($("\n"));
         FW_MATCH_CASE_1(Expression::Literal::string8, chars)
            //FIXME:  Too lazy to do this properly
            out->write($("string\n"));
         FW_MATCH_CASE_1(Expression::Literal::string16, chars)
            //FIXME:  Too lazy to do this properly
            out->write($("string\n"));
         FW_MATCH_CASE_1(Expression::Literal::string32, chars)
            //FIXME:  Too lazy to do this properly
            out->write($("string\n"));
         FW_MATCH_END

      FW_MATCH_CASE_2(Expression::Value::variable, level, item)
         out->write($("variable\n"));
         out->write(indent);
         out->write($("environmentIndex = "));
         out->write(String(level));
         out->write($(", "));
         out->write(String(item));
         out->write($("\n"));

      FW_MATCH_CASE_2(Expression::Value::memberAccess, item, member)
         out->write($("memberAccess\n"));
         out->write(indent);
         out->write($("object = "));
         printExpression(out, item, subIndent);
         out->write(indent);
         out->write($("member = "));
         out->write(member);
         out->write($("\n"));

      FW_MATCH_CASE_1(Expression::Value::arrayConstructor, elements)
         out->write($("array\n"));
         for(int i = 0; i < elements.getSize(); i++)
         {
            out->write(indent);
            out->write($("element"));
            out->write(String(i));
            out->write($(" = "));
            printExpression(out, elements[i], subIndent);
         }

      FW_MATCH_CASE_3(Expression::Value::conditional, input, thenClause, elseClause)
         out->write($("if\n"));
         out->write(indent);
         out->write($("input = "));
         printExpression(out, input, subIndent);
         out->write(indent);
         out->write($("then = "));
         printExpression(out, thenClause, subIndent);
         out->write(indent);
         out->write($("else = "));
         printExpression(out, elseClause, subIndent);

      FW_MATCH_CASE_2(Expression::Value::whereBlock, members, result)
         out->write($("where\n"));
         for(int i = 0; i < members.getSize(); i++)
         {
            out->write(indent);
            out->write($("item"));
            out->write(String(i));
            out->write($(" = "));
            printExpression(out, members[i], subIndent);
         }
         out->write(indent);
         out->write($("result = "));
         printExpression(out, result, subIndent);

      FW_MATCH_CASE_2(Expression::Value::function, parameters, definition)
         out->write($("function ("));
         for(int i = 0; i < parameters.getSize(); i++)
         {
            if(i > 0)
               out->write($(", "));
            out->write(parameters[i]);
         }
         out->write($(")\n"));
         out->write(indent);
         out->write($("definition = "));
         printExpression(out, definition, subIndent);
      FW_MATCH_CASE_2(Expression::Value::application, function, parameters)
         out->write($("application\n"));
         out->write(indent);
         out->write($("function = "));
         printExpression(out, function, subIndent);
         for(int i = 0; i < parameters.getSize(); i++)
         {
            out->write(indent);
            out->write($("parameter"));
            out->write(String(i));
            out->write($(" = "));
            printExpression(out, parameters[i], subIndent);
         }

      FW_MATCH_CASE_1(Expression::Value::objectDefinition, members)
         out->write($("object\n"));
         for(int i = 0; i < members.getSize(); i++)
         {
            out->write(indent);
            out->write(members[i].name);
            out->write($(" = "));
            printExpression(out, members[i].value, subIndent);
         }

      FW_MATCH_END
   }
}

void writeExpression(framework::TextWriterPtr out, Expression expression)
{
   printExpression(out, expression, $("   "));
}
#endif

} //namespace compiler
} //namespace evlan
