/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Token.h"

// begin namespace ===================================================
namespace evlan {
namespace compiler {

#ifdef INCLUDE_WRITER_CODE
namespace
{
   template<typename Character>
   void writeString(framework::TextWriterPtr out, Array<Character> text)
   {
      out->write($("string(\""));
      for(int i = 0; i < text.getSize(); i++)
      {
         if((char32)text[i] < 128)
         {
            switch((char)text[i])
            {
            case '\0':
               out->write($("\\0"));
               break;
            case '\a':
               out->write($("\\a"));
               break;
            case '\b':
               out->write($("\\b"));
               break;
            case '\f':
               out->write($("\\f"));
               break;
            case '\n':
               out->write($("\\n"));
               break;
            case '\r':
               out->write($("\\r"));
               break;
            case '\t':
               out->write($("\\t"));
               break;
            case '\v':
               out->write($("\\v"));
               break;
            case '\"':
               out->write($("\\\""));
               break;
            case '\'':
               out->write($("\\\'"));
               break;
            case '\\':
               out->write($("\\\\"));
               break;
            default:
               out->write((char)text[i], 1);
               break;
            }
         }
         else if((char32)text[i] < 0x00010000)
         {
            out->write($("\\u"));
            out->write(String(text[i], 4, 16));
         }
         else
         {
            out->write($("\\U"));
            out->write(String(text[i], 8, 16));
         }
      }
      out->write($("\")"));
   }
}

void writeToken(framework::TextWriterPtr out, Token::Value token)
{
   FW_MATCH_START(token)
      FW_MATCH_CASE_1(Token::Value::whitespace, amount)
         out->write($("whitespace("));
         out->write(String(amount));
         out->write($(")"));

      FW_MATCH_CASE_0(Token::Value::endOfLine)
         out->write($("endOfLine"));
      FW_MATCH_CASE_0(Token::Value::endOfStatement)
         out->write($("endOfStatement"));
      FW_MATCH_CASE_0(Token::Value::startOfBlock)
         out->write($("startOfBlock"));
      FW_MATCH_CASE_0(Token::Value::endOfBlock)
         out->write($("endOfBlock"));
      FW_MATCH_CASE_0(Token::Value::endOfSource)
         out->write($("endOfSource"));

      FW_MATCH_CASE_1(Token::Value::identifier, text)
         out->write($("identifier(\""));
         out->write(text);
         out->write($("\")"));
      FW_MATCH_CASE_1(Token::Value::keyword, id)
         out->write($("keyword("));
         out->write(id.getName());
         out->write($(")"));

      FW_MATCH_CASE_1(Token::Value::integer, i)
         out->write($("integer("));
         out->write(String(i));
         out->write($(")"));
      FW_MATCH_CASE_1(Token::Value::integer64, i)
         out->write($("integer64("));
         out->write(String(i));
         out->write($(")"));

      FW_MATCH_CASE_1(Token::Value::real, r)
         out->write($("real("));
         out->write(String(r));
         out->write($(")"));

      FW_MATCH_CASE_1(Token::Value::data8, d)
         out->write($("data8(0x"));
         out->write(String((uint32)d, 2, 16));
         out->write($(")"));
      FW_MATCH_CASE_1(Token::Value::data16, d)
         out->write($("data16(0x"));
         out->write(String((uint32)d, 4, 16));
         out->write($(")"));
      FW_MATCH_CASE_1(Token::Value::data32, d)
         out->write($("data32(0x"));
         out->write(String(d, 8, 16));
         out->write($(")"));
      FW_MATCH_CASE_1(Token::Value::data64, d)
         out->write($("data64(0x"));
         out->write(String(d, 16, 16));
         out->write($(")"));

      FW_MATCH_CASE_1(Token::Value::character, c)
         out->write($("character('"));
         if(c < 128)
         {
            switch((char)c)
            {
            case '\0':
               out->write($("\\0"));
               break;
            case '\a':
               out->write($("\\a"));
               break;
            case '\b':
               out->write($("\\b"));
               break;
            case '\f':
               out->write($("\\f"));
               break;
            case '\n':
               out->write($("\\n"));
               break;
            case '\r':
               out->write($("\\r"));
               break;
            case '\t':
               out->write($("\\t"));
               break;
            case '\v':
               out->write($("\\v"));
               break;
            case '\"':
               out->write($("\\\""));
               break;
            case '\'':
               out->write($("\\\'"));
               break;
            case '\\':
               out->write($("\\\\"));
               break;
            default:
               out->write((char)c, 1);
               break;
            }
         }
         else if(c < 0x00010000)
         {
            out->write($("\\u"));
            out->write(String(c, 4, 16));
         }
         else
         {
            out->write($("\\U"));
            out->write(String(c, 8, 16));
         }
         out->write($("')"));

      FW_MATCH_CASE_1(Token::Value::atom, name)
         out->write($("atom("));
         out->write(name);
         out->write($(")"));

      FW_MATCH_CASE_1(Token::Value::string8, text)
         writeString(out, text);

      FW_MATCH_CASE_1(Token::Value::string16, text)
         writeString(out, text);

      FW_MATCH_CASE_1(Token::Value::string32, text)
         writeString(out, text);

      FW_MATCH_CASE_1(Token::Value::comment, text)
         out->write($("comment(\""));
         out->write(text);
         out->write($("\")"));

   FW_MATCH_END
}

void writeToken(framework::TextWriterPtr out, Token token)
{
   out->write(String(token.line));
   out->write($(":"));
   out->write(String(token.column));
   out->write($(":"));
   writeToken(out, token.value);
}
#endif

} //namespace compiler
} //namespace evlan
