/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/01/27 (KV)
 *  + First Version
 */

// ===================================================================

#include "cleanupUnusedTables.h"

namespace evlan {
namespace compiler {
namespace backend {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::compiler::backend::cleanupUnusedTables
// TODO (KV): Review evlan::compiler::backend::cleanupUnusedTables

void cleanupUnusedTables(MutableBytecodePtr bytecode)
{
   Array<bool> dataMarks(bytecode->dataTable.getSize());
   Array<bool> atomMarks(bytecode->atomTable.getSize());
   Array<bool> importMarks(bytecode->importTable.getSize());
   Array<bool> memberNameMarks(bytecode->memberNameTable.getSize());
   Array<bool> classMarks(bytecode->classes.getSize());

   for(int i = 0; i < dataMarks.getSize(); i++)
      dataMarks[i] = false;
   for(int i = 0; i < atomMarks.getSize(); i++)
      atomMarks[i] = false;
   for(int i = 0; i < importMarks.getSize(); i++)
      importMarks[i] = false;
   for(int i = 0; i < memberNameMarks.getSize(); i++)
      memberNameMarks[i] = false;
   for(int i = 0; i < classMarks.getSize(); i++)
      classMarks[i] = false;

   for(int i = 0; i < bytecode->instructions.getSize(); i++)
   {
      FW_MATCH_START(bytecode->instructions[i].type)
      FW_MATCH_CASE_2(MutableBytecode::Instruction::Type::loadData, size, index)
         dataMarks[index] = true;
      FW_MATCH_CASE_1(MutableBytecode::Instruction::Type::loadAtom, index)
         atomMarks[index] = true;
      FW_MATCH_CASE_3(MutableBytecode::Instruction::Type::loadArray, type, size, index)
         dataMarks[index] = true;
      FW_MATCH_CASE_1(MutableBytecode::Instruction::Type::loadImport, index)
         importMarks[index] = true;
      FW_MATCH_CASE_1(MutableBytecode::Instruction::Type::makeObject, index)
         if(classMarks[index] != true)
         {
            classMarks[index] = true;
            for(int i = 0; i < bytecode->classes[index].memberTable.getSize(); i++)
               memberNameMarks[bytecode->classes[index].memberTable[i]] = true;
         }
      FW_MATCH_CASE_1(MutableBytecode::Instruction::Type::loadMember, index)
         memberNameMarks[index] = true;
      FW_MATCH_DEFAULT
         //nothing
      FW_MATCH_END
   }

   int dataCount = 0;
   int atomCount = 0;
   int importCount = 0;
   int memberNameCount = 0;
   int classCount = 0;

   Array<int> dataMap(bytecode->dataTable.getSize());
   Array<int> atomMap(bytecode->atomTable.getSize());
   Array<int> importMap(bytecode->importTable.getSize());
   Array<int> memberNameMap(bytecode->memberNameTable.getSize());
   Array<int> classMap(bytecode->classes.getSize());

   for(int i = 0; i < dataMarks.getSize(); i++)
   {
      if(dataMarks[i])
         dataMap[i] = dataCount++;
      else
         dataMap[i] = -1;
   }
   for(int i = 0; i < atomMarks.getSize(); i++)
   {
      if(atomMarks[i])
         atomMap[i] = atomCount++;
      else
         atomMap[i] = -1;
   }
   for(int i = 0; i < importMarks.getSize(); i++)
   {
      if(importMarks[i])
         importMap[i] = importCount++;
      else
         importMap[i] = -1;
   }
   for(int i = 0; i < memberNameMarks.getSize(); i++)
   {
      if(memberNameMarks[i])
         memberNameMap[i] = memberNameCount++;
      else
         memberNameMap[i] = -1;
   }
   for(int i = 0; i < classMarks.getSize(); i++)
   {
      if(classMarks[i])
         classMap[i] = classCount++;
      else
         classMap[i] = -1;
   }

   Array<Array<const uint32> > newDataTable(dataCount);
   Array<String> newImportTable(importCount);
   Array<String> newMemberNameTable(memberNameCount);
   Array<String> newAtomTable(atomCount);
   Array<Bytecode::ObjectClass> newClasses(classCount);

   for(int i = 0; i < dataMarks.getSize(); i++)
   {
      if(dataMap[i] != -1)
         newDataTable[dataMap[i]] = bytecode->dataTable[i];
   }
   for(int i = 0; i < atomMarks.getSize(); i++)
   {
      if(atomMap[i] != -1)
         newAtomTable[atomMap[i]] = bytecode->atomTable[i];
   }
   for(int i = 0; i < importMarks.getSize(); i++)
   {
      if(importMap[i] != -1)
         newImportTable[importMap[i]] = bytecode->importTable[i];
   }
   for(int i = 0; i < memberNameMarks.getSize(); i++)
   {
      if(memberNameMap[i] != -1)
         newMemberNameTable[memberNameMap[i]] = bytecode->memberNameTable[i];
   }
   for(int i = 0; i < classMarks.getSize(); i++)
   {
      if(classMap[i] != -1)
      {
         Array<int> newMemberTable(bytecode->classes[i].memberTable.getSize());
         for(int j = 0; j < newMemberTable.getSize(); j++)
            newMemberTable[j] = memberNameMap[bytecode->classes[i].memberTable[j]];
         newClasses[classMap[i]] = Bytecode::ObjectClass(newMemberTable.toConst(),
            bytecode->classes[i].documentation);
      }
   }

   for(int i = 0; i < bytecode->instructions.getSize(); i++)
   {
      FW_MATCH_START(bytecode->instructions[i].type)
      FW_MATCH_CASE_2(MutableBytecode::Instruction::Type::loadData, size, index)
         bytecode->instructions[i].type =
            MutableBytecode::Instruction::Type::loadData(size, dataMap[index]);
      FW_MATCH_CASE_1(MutableBytecode::Instruction::Type::loadAtom, index)
         bytecode->instructions[i].type =
            MutableBytecode::Instruction::Type::loadAtom(atomMap[index]);
      FW_MATCH_CASE_3(MutableBytecode::Instruction::Type::loadArray, type, size, index)
         bytecode->instructions[i].type =
            MutableBytecode::Instruction::Type::loadArray(type, size, dataMap[index]);
      FW_MATCH_CASE_1(MutableBytecode::Instruction::Type::loadImport, index)
         bytecode->instructions[i].type =
            MutableBytecode::Instruction::Type::loadImport(importMap[index]);
      FW_MATCH_CASE_1(MutableBytecode::Instruction::Type::makeObject, index)
         bytecode->instructions[i].type =
            MutableBytecode::Instruction::Type::makeObject(classMap[index]);
      FW_MATCH_CASE_1(MutableBytecode::Instruction::Type::loadMember, index)
         bytecode->instructions[i].type =
            MutableBytecode::Instruction::Type::loadMember(memberNameMap[index]);
      FW_MATCH_DEFAULT
         //nothing
      FW_MATCH_END
   }

   bytecode->dataTable = newDataTable;
   bytecode->importTable = newImportTable;
   bytecode->memberNameTable = newMemberNameTable;
   bytecode->atomTable = newAtomTable;
   bytecode->classes = newClasses;
}

// ===================================================================

} //namespace backend
} //namespace compiler
} //namespace evlan
