/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/01/15 (KV)
 *  + First Version
 */

// ===================================================================

#include "makeBytecode.h"
#include <framework/util/Vector.h>
#include <framework/util/TableBuilder.h>

namespace evlan {
namespace compiler {
namespace backend {

//Fuck the C++ standard for disallowing this, and shame on GCC for
//allowing it in previous versions then spontaneously breaking it.
//using ::framework::util;
namespace util {using namespace ::framework::util;}

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::compiler::backend::BytecodeMaker
// TODO (KV): Review evlan::compiler::backend::BytecodeMaker

class BytecodeMaker
{
public:
   BytecodeMaker(ErrorCollectorPtr errorCollector);
   ~BytecodeMaker();

   MutableBytecodePtr make(Expression expression);

private:
   // internal types -------------------------------------------------
   struct Forward
   {
      int target;
      int line;
      int column;

      inline Forward() {}
      inline Forward(int target, int line, int column)
         : target(target), line(line), column(column) {}
   };

   // internal fields ------------------------------------------------
   ErrorCollectorPtr mErrorCollector;

   util::TableBuilder<Array<const uint32>, Array<const uint32>::SequenceOrdering > mStringTable;
   util::TableBuilder<String> mImportTable;
   util::TableBuilder<String> mMemberNameTable;
   util::TableBuilder<String> mAtomTable;
   util::TableBuilder<Bytecode::ObjectClass> mClasses;
   util::Vector<MutableBytecode::Instruction> mInstructions;

   int mRootInstruction;

   /* Since the graph can contain cycles, "chicken-and-egg" problems
    * sometimes show up in its construction.  To get around this,
    * instruction inputs are temporarily allowed to be negative
    * indices, where these indices refer to elements in mForwards
    * rather than mInstructions.  mForwards simply contains the
    * correct indices which should replace these negative indices once
    * the graph has been fully constructed.  This is mainly used for
    * "where" expressions. */
   util::Vector<Forward> mForwards;

   // internal functions ---------------------------------------------
   int addToGraph(Expression expression,
                  util::Vector<Array<const int> >& environment,
                  int function, int branch);

   void resolveForward(int index);
};

BytecodeMaker::BytecodeMaker(ErrorCollectorPtr errorCollector)
   : mErrorCollector(errorCollector)
{
}

BytecodeMaker::~BytecodeMaker()
{
}

MutableBytecodePtr BytecodeMaker::make(Expression expression)
{
   {
      util::Vector<Array<const int> > environment;
      mRootInstruction = addToGraph(expression, environment, -1, -1);
   }

   //fix the forwards
   for(int i = 0; i < mForwards.getSize(); i++)
   {
      if(mForwards[i].target < 0)
         resolveForward(i);
   }

   if(mRootInstruction < 0)
      mRootInstruction = mForwards[~mRootInstruction].target;

   for(int i = 0; i < mInstructions.getSize(); i++)
   {
      Array<int> newInputs(mInstructions[i].inputs.getSize());

      for(int j = 0; j < newInputs.getSize(); j++)
      {
         if(mInstructions[i].inputs[j] < 0)
            newInputs[j] = mForwards[~mInstructions[i].inputs[j]].target;
         else
            newInputs[j] = mInstructions[i].inputs[j];
      }

      mInstructions[i].inputs = newInputs;
   }

   return spnew MutableBytecode(
      mStringTable.makeTable(),
      mImportTable.makeTable(),
      mMemberNameTable.makeTable(),
      mAtomTable.makeTable(),
      mClasses.makeTable(),
      mInstructions.getArray(),
      mRootInstruction);
}

// internal functions ------------------------------------------------
int BytecodeMaker::addToGraph(Expression expression,
                              util::Vector<Array<const int> >& environment,
                              int function, int branch)
{
   MutableBytecode::Instruction::Type instructionType;
   Array<int> inputs;

   FW_MATCH_START(expression.value)
   FW_MATCH_CASE_1(Expression::Value::import, name)
      instructionType = MutableBytecode::Instruction::Type::loadImport(
         mImportTable.lookup(name));

   FW_MATCH_CASE_1(Expression::Value::literal, literal)
      FW_MATCH_START(literal)
      FW_MATCH_CASE_1(Expression::Literal::boolean, flag)
         if(flag)
            instructionType = MutableBytecode::Instruction::Type::loadTrue();
         else
            instructionType = MutableBytecode::Instruction::Type::loadFalse();
      FW_MATCH_CASE_1(Expression::Literal::integer, i)
         instructionType = MutableBytecode::Instruction::Type::loadInteger(i);
      FW_MATCH_CASE_1(Expression::Literal::integer64, i)
         instructionType = MutableBytecode::Instruction::Type::loadInteger64(i);
      FW_MATCH_CASE_1(Expression::Literal::real, x)
         instructionType = MutableBytecode::Instruction::Type::loadReal(x);
      FW_MATCH_CASE_1(Expression::Literal::data8,  d)
         instructionType = MutableBytecode::Instruction::Type::loadData8(d);
      FW_MATCH_CASE_1(Expression::Literal::data16, d)
         instructionType = MutableBytecode::Instruction::Type::loadData16(d);
      FW_MATCH_CASE_1(Expression::Literal::data32, d)
         instructionType = MutableBytecode::Instruction::Type::loadData32(d);
      FW_MATCH_CASE_1(Expression::Literal::data64, d)
         instructionType = MutableBytecode::Instruction::Type::loadData64(d);
      FW_MATCH_CASE_1(Expression::Literal::character, c)
         instructionType = MutableBytecode::Instruction::Type::loadCharacter(c);
      FW_MATCH_CASE_1(Expression::Literal::atom, name)
         instructionType = MutableBytecode::Instruction::Type::loadAtom(
            mAtomTable.lookup(name));
      FW_MATCH_CASE_1(Expression::Literal::string8, chars)
         Array<uint32> data((chars.getSize() + 3) / 4);
#ifdef PLATFORM_BIG_ENDIAN
         if(chars != null)
            memcpy(&data[0], &chars[0], chars.getSize());
#else
         for(int i = 0; i < chars.getSize(); i++)
            ((char8*)&data[0])[i^3] = chars[i];
#endif
         instructionType = MutableBytecode::Instruction::Type::loadArray(
            Bytecode::CHARACTER8, chars.getSize(), mStringTable.lookup(data.toConst()));
      FW_MATCH_CASE_1(Expression::Literal::string16, chars)
         Array<uint32> data((chars.getSize() + 1) / 2);
#ifdef PLATFORM_BIG_ENDIAN
         if(chars != null)
            memcpy(&data[0], &chars[0], chars.getSize()*2);
#else
         for(int i = 0; i < chars.getSize(); i++)
            ((char16*)&data[0])[i^1] = chars[i];
#endif
         instructionType = MutableBytecode::Instruction::Type::loadArray(
            Bytecode::CHARACTER16, chars.getSize(), mStringTable.lookup(data.toConst()));
      FW_MATCH_CASE_1(Expression::Literal::string32, chars)
         instructionType = MutableBytecode::Instruction::Type::loadArray(
            Bytecode::CHARACTER32, chars.getSize(), mStringTable.lookup(chars));
      FW_MATCH_END

   FW_MATCH_CASE_2(Expression::Value::variable, level, item)
      return environment[level][item];

   FW_MATCH_CASE_2(Expression::Value::memberAccess, item, member)
      instructionType = MutableBytecode::Instruction::Type::loadMember(
         mMemberNameTable.lookup(member));
      inputs = Array<int>(1);
      inputs[0] = addToGraph(item, environment, function, branch);

   FW_MATCH_CASE_1(Expression::Value::arrayConstructor, elements)
      instructionType = MutableBytecode::Instruction::Type::makeArray(elements.getSize());
      inputs = Array<int>(elements.getSize());
      for(int i = 0; i < elements.getSize(); i++)
         inputs[i] = addToGraph(elements[i], environment, function, branch);

   FW_MATCH_CASE_3(Expression::Value::conditional, input, thenClause, elseClause)
      int branchInstruction = mInstructions.getSize();
      mInstructions.pushBack(MutableBytecode::Instruction());

      inputs = Array<int>(3);
      inputs[0] = addToGraph(input, environment, function, branch);
      inputs[1] = addToGraph(thenClause, environment, function, branchInstruction*2);
      inputs[2] = addToGraph(elseClause, environment, function, branchInstruction*2+1);

      mInstructions[branchInstruction] = MutableBytecode::Instruction(
            MutableBytecode::Instruction::Type::testCondition(),
            inputs, function, branch, expression.line, expression.column);
      return branchInstruction;

   FW_MATCH_CASE_2(Expression::Value::whereBlock, members, result)
      //Use mForwards here to get around chicken-and-egg problem.
      Array<int> environmentFrame(members.getSize());
      for(int i = 0; i < members.getSize(); i++)
      {
         environmentFrame[i] = ~mForwards.getSize();
         mForwards.pushBack(Forward(0, members[i].line, members[i].column));
      }
      environment.pushBack(environmentFrame.toConst());
      for(int i = 0; i < members.getSize(); i++)
      {
         /* COOLEST BUG EVER!
          * Previously, this code looked like:
          *    mForwards[~environmentFrame[i]].target =
          *       addToGraph(members[i], environment);
          * I found, however, that sometimes the value in "target" was, for
          * some reason, not changed by the assignment.  addToGraph() was
          * returning the correct value, but it simply was not being written
          * to "mForwards[~environmentFrame[i]].target".  It worked sometimes,
          * but not always.  The only thing I could think of was that it was
          * a GCC bug, but that didn't make sense since the same code
          * sometimes worked and sometimes didn't.
          *
          * Well, finally I figured it out:  GCC was computing the address of
          * "mForwards[~environmentFrame[i]].target" *before* calling
          * addToGraph().  Since mForwards is a vector and addToGraph() could
          * add elements to it, the address of "target" could be different
          * once the call returned.  So, it would end up writing the returned
          * value to the wrong location.
          *
          * I am *so* glad that I will soon be able to write my code in a
          * language that doesn't have such idiotic problems! */
         int temp = addToGraph(members[i], environment, function, branch);
         mForwards[~environmentFrame[i]].target = temp;
      }
      int resultInstruction = addToGraph(result, environment, function, branch);
      environment.popBack();
      return resultInstruction;

   FW_MATCH_CASE_3(Expression::Value::function, parameters, definition, documentation)
      int functionInstruction = mInstructions.getSize();
      mInstructions.pushBack(MutableBytecode::Instruction());
      Array<int> paramInputs(1);
      paramInputs[0] = functionInstruction;

      Array<int> environmentFrame(parameters.getSize());
      for(int i = 0; i < parameters.getSize(); i++)
      {
         environmentFrame[i] = mInstructions.getSize();
         mInstructions.pushBack(MutableBytecode::Instruction(
            MutableBytecode::Instruction::Type::loadParameter(i),
            paramInputs, functionInstruction, -1, expression.line, expression.column));
      }
      environment.pushBack(environmentFrame.toConst());
      inputs = Array<int>(1);
      inputs[0] = addToGraph(definition, environment, functionInstruction, -1);
      environment.popBack();

      mInstructions[functionInstruction] = MutableBytecode::Instruction(
            MutableBytecode::Instruction::Type::makeFunction(parameters, documentation),
            inputs, function, branch, expression.line, expression.column);
      return functionInstruction;

   FW_MATCH_CASE_2(Expression::Value::application, functionExpression, parameters)
      instructionType = MutableBytecode::Instruction::Type::apply();
      inputs = Array<int>(parameters.getSize() + 1);
      inputs[0] = addToGraph(functionExpression, environment, function, branch);
      for(int i = 0; i < parameters.getSize(); i++)
         inputs[i+1] = addToGraph(parameters[i], environment, function, branch);

   FW_MATCH_CASE_2(Expression::Value::objectDefinition, members, documentation)
      Array<int> memberTable(members.getSize());
      for(int i = 0; i < members.getSize(); i++)
         memberTable[i] = mMemberNameTable.lookup(members[i].name);
      int classIndex = mClasses.lookup(Bytecode::ObjectClass(memberTable.toConst(), documentation));

      Array<int> environmentFrame(members.getSize());
      for(int i = 0; i < members.getSize(); i++)
      {
         environmentFrame[i] = ~mForwards.getSize();
         mForwards.pushBack(Forward(0, members[i].value.line, members[i].value.column));
      }
      environment.pushBack(environmentFrame.toConst());
      inputs = Array<int>(members.getSize());
      for(int i = 0; i < members.getSize(); i++)
      {
         /* do NOT edit these two lines unless you've read the
          * "COOLEST BUG EVER" comment up above. */
         inputs[i] = addToGraph(members[i].value, environment, function, branch);
         mForwards[~environmentFrame[i]].target = inputs[i];
      }
      environment.popBack();

      instructionType = MutableBytecode::Instruction::Type::makeObject(classIndex);

   FW_MATCH_END

   mInstructions.pushBack(MutableBytecode::Instruction(
      instructionType, inputs,
      function, branch,
      expression.line, expression.column));
   return mInstructions.getSize() - 1;
}

void BytecodeMaker::resolveForward(int index)
{
   int target = mForwards[index].target;

   if(target < 0)
   {
      if(~target == mForwards.getSize())
      {
         mErrorCollector->addError(mForwards[index].line, mForwards[index].column,
            $("Value defined to equal itself."));
         mForwards[index].target = 0;
      }
      else if(~target < mForwards.getSize())
      {
         mForwards[index].target = ~mForwards.getSize();

         if(mForwards[~target].target < 0)
            resolveForward(~target);
         mForwards[index].target = mForwards[~target].target;
      }
   }
}

// interface =========================================================
MutableBytecodePtr makeBytecode(Expression expression,
                                ErrorCollectorPtr errorCollector)
{
   BytecodeMaker maker(errorCollector);

   return maker.make(expression);
}

// ===================================================================

} //namespace backend
} //namespace compiler
} //namespace evlan
