/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/03/16 (KV)
 *  + First Version
 */

// ===================================================================

#include "optimizeArithmetic.h"
#include <stack>
#include <math.h>

namespace evlan {
namespace compiler {
namespace backend {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::compiler::backend::optimizeArithmetic
// TODO (KV): Review evlan::compiler::backend::optimizeArithmetic

namespace
{
   double modulus(double x, double y)
   {
      double result = ::fmod(x, y);
      if(x * y < 0 && result != 0)
         return result + y;
      else
         return result;
   }

   int modulus(int x, int y)
   {
      int result = x % y;
      if(x * y < 0 && result != 0)
         return result + y;
      else
         return result;
   }

   int64 modulus(int64 x, int64 y)
   {
      int64 result = x % y;
      if(x * y < 0 && result != 0)
         return result + y;
      else
         return result;
   }

   int64 modulus(int64 x, int y)
   {
      int64 result = x % y;
      if(x * y < 0 && result != 0)
         return result + y;
      else
         return result;
   }

   int64 modulus(int x, int64 y)
   {
      int64 result = x % y;
      if(x * y < 0 && result != 0)
         return result + y;
      else
         return result;
   }
}

enum {NOT_DONE, DOING, DONE};

void optimizeArithmetic(MutableBytecodePtr bytecode)
{
   std::stack<int> todo;
   //status:  NOT_DONE, DOING, or DONE
   Array<int8> status(bytecode->instructions.getSize());

   for(int i = bytecode->instructions.getSize() - 1; i >= 0; i--)
   {
      status[i] = NOT_DONE;
      todo.push(i);
   }

   while(!todo.empty())
   {
      int instruction = todo.top();
      if(status[instruction] == DONE)
      {
         todo.pop();
         continue;
      }
      status[instruction] = DOING;  //prevent cycles

      FW_MATCH_START(bytecode->instructions[instruction].type)
      FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadMember, index)
         int input = bytecode->instructions[instruction].inputs[0];
         if(status[input] == NOT_DONE)
            todo.push(input);
         else
         {
            if(bytecode->memberNameTable[index] == $("negate"))
            {
               FW_MATCH_START(bytecode->instructions[input].type)
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(-x);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(-x);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(-x);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_END
            }
            else if(bytecode->memberNameTable[index] == $("not"))
            {
               FW_MATCH_START(bytecode->instructions[input].type)
               FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadTrue)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadFalse();
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadFalse)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadTrue();
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData1, d)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData1(~d);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData2, d)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData2(~d);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData4, d)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData4(~d);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData8, d)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData8(~d);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData16, d)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData16(~d);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData32, d)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData32(~d);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData64, d)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData64(~d);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_END
            }
            else if(bytecode->memberNameTable[index] == $("size"))
            {
               FW_MATCH_START(bytecode->instructions[input].type)
               FW_MATCH_CASE_3(Bytecode::Instruction::Type::loadArray, type, size, index)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(size);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_CASE_1(Bytecode::Instruction::Type::makeArray, size)
                  bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(size);
                  bytecode->instructions[instruction].inputs = null;
               FW_MATCH_END
            }

            status[instruction] = DONE;
            todo.pop();
         }

      FW_MATCH_CASE_0(Bytecode::Instruction::Type::apply)
         if(bytecode->instructions[instruction].inputs.getSize() == 2)
         {
            int function = bytecode->instructions[instruction].inputs[0];

            FW_MATCH_START(bytecode->instructions[function].type)
            FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadMember, index)
               int input1 = bytecode->instructions[function].inputs[0];
               int input2 = bytecode->instructions[instruction].inputs[1];

               if(status[input1] != NOT_DONE
               && status[input2] != NOT_DONE)
               {
                  if(bytecode->memberNameTable[index] == $("add"))
                  {
                     FW_MATCH_START(bytecode->instructions[input1].type)
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(x1 + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadCharacter, c)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadCharacter(x1 + c);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadCharacter, c)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadCharacter(x1 + c);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadCharacter, c)
                           if(x1 == (int)x1)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadCharacter((int)x1 + c);
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadCharacter, c)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadCharacter(c + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadCharacter(c + x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           if(x2 == (int)x2)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadCharacter(c + (int)x2);
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_END
                     FW_MATCH_END
                  }
                  else if(bytecode->memberNameTable[index] == $("subtract"))
                  {
                     FW_MATCH_START(bytecode->instructions[input1].type)
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(x1 - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadCharacter, c)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadCharacter(c - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadCharacter(c - x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           if(x2 == (int)x2)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadCharacter(c - (int)x2);
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_END
                     FW_MATCH_END
                  }
                  else if(bytecode->memberNameTable[index] == $("multiply"))
                  {
                     FW_MATCH_START(bytecode->instructions[input1].type)
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(x1 * x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 * x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 * x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 * x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 * x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 * x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 * x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 * x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 * x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_END
                  }
                  else if(bytecode->memberNameTable[index] == $("divide"))
                  {
                     FW_MATCH_START(bytecode->instructions[input1].type)
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           if(x2 != 0 && x1 / x2 == (double)x1 / x2)
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(x1 / x2);
                           else
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal((double)x1 / x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           if(x2 != 0 && x1 / x2 == (double)x1 / x2)
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 / x2);
                           else
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal((double)x1 / x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 / x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           if(x2 != 0 && x1 / x2 == (double)x1 / x2)
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 / x2);
                           else
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal((double)x1 / x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           if(x2 != 0 && x1 / x2 == (double)x1 / x2)
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(x1 / x2);
                           else
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal((double)x1 / x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 / x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 / x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 / x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(x1 / x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_END
                  }
                  else if(bytecode->memberNameTable[index] == $("modulus"))
                  {
                     FW_MATCH_START(bytecode->instructions[input1].type)
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           if(x2 != 0)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(modulus(x1, x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                           else
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(modulus((double)x1, (double)x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           if(x2 != 0)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(modulus(x1, x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                           else
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(modulus((double)x1, (double)x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           if(x2 == (int)x2 && x2 != 0)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(modulus(x1, (int)x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                           else
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(modulus((double)x1, x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           if(x2 != 0)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(modulus(x1, x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                           else
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(modulus((double)x1, (double)x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           if(x2 != 0)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(modulus(x1, x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                           else
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(modulus((double)x1, (double)x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           if(x2 == (int)x2 && x2 != 0)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(modulus(x1, (int)x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                           else
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(modulus((double)x1, x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           if(x1 == (int)x1 && x2 != 0)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(modulus((int)x1, x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                           else
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(modulus(x1, (double)x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           if(x1 == (int)x1 && x2 != 0)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger64(modulus((int)x1, x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                           else
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(modulus(x1, (double)x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           if(x1 == (int)x1 && x2 == (int)x2 && x2 != 0)
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadInteger(modulus((int)x1, (int)x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                           else
                           {
                              bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(modulus(x1, x2));
                              bytecode->instructions[instruction].inputs = null;
                           }
                        FW_MATCH_END
                     FW_MATCH_END
                  }
                  else if(bytecode->memberNameTable[index] == $("raiseToPower"))
                  {
                     FW_MATCH_START(bytecode->instructions[input1].type)
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(::pow((double)x1, (double)x2));
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(::pow((double)x1, (double)x2));
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(::pow((double)x1, x2));
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(::pow((double)x1, (double)x2));
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(::pow((double)x1, (double)x2));
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(::pow((double)x1, x2));
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(::pow(x1, (double)x2));
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadInteger64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(::pow(x1, (double)x2));
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadReal, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadReal(::pow(x1, x2));
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_END
                  }
                  else if(bytecode->memberNameTable[index] == $("and"))
                  {
                     FW_MATCH_START(bytecode->instructions[input1].type)
                     FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadTrue)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadTrue)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadTrue();
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadFalse)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadFalse();
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadFalse)
                        bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadFalse();
                        bytecode->instructions[instruction].inputs = null;
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData1, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData1, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData1(x1 & x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData2, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData2, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData2(x1 & x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData4, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData4, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData4(x1 & x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData8, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData8, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData8(x1 & x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData16, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData16, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData16(x1 & x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData32, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData32, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData32(x1 & x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData64, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData64(x1 & x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_DEFAULT
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadFalse)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadFalse();
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_END
                  }
                  else if(bytecode->memberNameTable[index] == $("or"))
                  {
                     FW_MATCH_START(bytecode->instructions[input1].type)
                     FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadFalse)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadTrue)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadTrue();
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadFalse)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadFalse();
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadTrue)
                        bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadTrue();
                        bytecode->instructions[instruction].inputs = null;
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData1, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData1, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData1(x1 | x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData2, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData2, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData2(x1 | x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData4, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData4, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData4(x1 | x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData8, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData8, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData8(x1 | x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData16, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData16, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData16(x1 | x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData32, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData32, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData32(x1 | x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData64, x1)
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_1(Bytecode::Instruction::Type::loadData64, x2)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadData64(x1 | x2);
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_DEFAULT
                        FW_MATCH_START(bytecode->instructions[input2].type)
                        FW_MATCH_CASE_0(Bytecode::Instruction::Type::loadTrue)
                           bytecode->instructions[instruction].type = Bytecode::Instruction::Type::loadTrue();
                           bytecode->instructions[instruction].inputs = null;
                        FW_MATCH_END
                     FW_MATCH_END
                  }

                  status[instruction] = DONE;
                  todo.pop();
               }
               else
               {
                  if(status[input1] == NOT_DONE)
                     todo.push(input1);
                  if(status[input2] == NOT_DONE)
                     todo.push(input2);
               }

            FW_MATCH_DEFAULT
               status[instruction] = DONE;
               todo.pop();
            FW_MATCH_END
         }
         else
         {
            status[instruction] = DONE;
            todo.pop();
         }

      FW_MATCH_DEFAULT
         status[instruction] = DONE;
         todo.pop();
      FW_MATCH_END
   }
}

// ===================================================================

} //namespace backend
} //namespace compiler
} //namespace evlan
