/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "BasicParser.h"
#include <framework/util/Vector.h>
#include "ParseObjects.h"
#include <map>
#include <stack>

// begin namespace ===================================================
namespace evlan {
namespace compiler {
namespace dialects {
namespace basic {
using namespace grammar;
using namespace syntax;

// implementation definition =================================================
implementation BasicParser: implements Parser
{
   BasicParser(BasicKeywordSetPtr keywordSet);
   ~BasicParser();

   // Parser ---------------------------------------------------------
   ParseTreePtr parse(Array<const Token> input,
                      ErrorCollectorPtr errorCollector);

   // internal -------------------------------------------------------
   BasicKeywordSetPtr mKeywordSet;
   ErrorCollectorPtr mErrorCollector;
   Token mCurrentToken;
   Array<const Token> mInput;
   int mPos;

   inline void shift()
   {
      if(++mPos < mInput.getSize())
         mCurrentToken = mInput[mPos];
   }

   inline util::Atom getKeywordId(Token token)
   {
      FW_MATCH_START(token.value)
      FW_MATCH_CASE_1(Token::Value::keyword, id)
         return id;
      FW_MATCH_DEFAULT
         return null;
      FW_MATCH_END
      return null;  //silly macros...
   }

   ParseTreePtr parseFunction(bool hadOpenParen);
   ParseTreePtr parseArray();
   ParseTreePtr parseBlockArray();
   ParseTreePtr parseConditional();
   ParseTreePtr parseObject();
   ParseTreePtr parseTry();
   ParseTreePtr parseThrow();
   ParseTreePtr parseBracedObject();
   ParseTreePtr parseImport();
   ParseTreePtr parseDo();
   ParseTreePtr parseString();
   ParseTreePtr parseItem();

   ParseTreePtr parseUnaryOps();
   ParseTreePtr parseBinaryOps();
   ParseTreePtr parseCons();
   ParseTreePtr parseWhere();
   ParseTreePtr parseExpression();

   struct Statement
   {
      String documentation;
      String valueName;
      ParseTreePtr value;
   };
   Statement parseStatement();
   bool skipRestOfStatement();
   bool skipRestOfElement();

   std::map<util::Atom, int> mOpPriorities;
   std::map<util::Atom, String> mBinaryOps;
   std::map<util::Atom, String> mUnaryOps;
};

BasicParser::BasicParser(BasicKeywordSetPtr keywordSet)
   : mKeywordSet(keywordSet)
{
   mOpPriorities[mKeywordSet->mOpPlusId         ] = 3;
   mOpPriorities[mKeywordSet->mOpMinusId        ] = 3;
   mOpPriorities[mKeywordSet->mOpTimesId        ] = 4;
   mOpPriorities[mKeywordSet->mOpDivideId       ] = 4;
   mOpPriorities[mKeywordSet->mOpModulusId      ] = 4;
   mOpPriorities[mKeywordSet->mOpPowerId        ] = 5;
   mOpPriorities[mKeywordSet->mOpEqualId        ] = 2;
   mOpPriorities[mKeywordSet->mOpLessId         ] = 2;
   mOpPriorities[mKeywordSet->mOpGreaterId      ] = 2;
   mOpPriorities[mKeywordSet->mOpLessEqualId    ] = 2;
   mOpPriorities[mKeywordSet->mOpGreaterEqualId ] = 2;
   mOpPriorities[mKeywordSet->mOpNotEqualId     ] = 2;
   mOpPriorities[mKeywordSet->mOpAndId          ] = 1;
   mOpPriorities[mKeywordSet->mOpOrId           ] = 1;

   mBinaryOps[mKeywordSet->mOpPlusId         ] = $("add");
   mBinaryOps[mKeywordSet->mOpMinusId        ] = $("subtract");
   mBinaryOps[mKeywordSet->mOpTimesId        ] = $("multiply");
   mBinaryOps[mKeywordSet->mOpDivideId       ] = $("divide");
   mBinaryOps[mKeywordSet->mOpModulusId      ] = $("modulus");
   mBinaryOps[mKeywordSet->mOpPowerId        ] = $("raiseToPower");
   mBinaryOps[mKeywordSet->mOpEqualId        ] = $("equals");
   mBinaryOps[mKeywordSet->mOpLessId         ] = $("isLessThan");
   mBinaryOps[mKeywordSet->mOpGreaterId      ] = null;
   mBinaryOps[mKeywordSet->mOpLessEqualId    ] = null;
   mBinaryOps[mKeywordSet->mOpGreaterEqualId ] = null;
   mBinaryOps[mKeywordSet->mOpNotEqualId     ] = null;
   mBinaryOps[mKeywordSet->mOpAndId          ] = $("and");
   mBinaryOps[mKeywordSet->mOpOrId           ] = $("or");

   mUnaryOps[mKeywordSet->mOpMinusId ] = $("negate");
   mUnaryOps[mKeywordSet->mOpNotId   ] = $("not");
}

BasicParser::~BasicParser()
{
}

ParseTreePtr BasicParser::parse(Array<const Token> input,
                                ErrorCollectorPtr errorCollector)
{
   mErrorCollector = errorCollector;
   mInput = input;
   mPos = 0;
   mCurrentToken = mInput[0];

   ParseTreePtr result = parseExpression();
   if(mCurrentToken.value.getId() != Token::Value::endOfSource_ID)
   {
      mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
         $("Expected end of source."));
   }


   Array<ParseObjects::WhereItem> automatics(17);

   int i = 0;
   automatics[i++].name = $("Integer");
   automatics[i++].name = $("Scalar");
   automatics[i++].name = $("Character");
   automatics[i++].name = $("Data");
   automatics[i++].name = $("Atom");
   automatics[i++].name = $("Array");
   automatics[i++].name = $("String");
   automatics[i++].name = $("Task");
   automatics[i++].name = $("Variable");
   automatics[i++].name = $("Maybe");
   automatics[i++].name = $("List");
   automatics[i++].name = $("Exception");
   automatics[i++].name = $("Tree");
   automatics[i++].name = $("Thread");
   automatics[i++].name = $("Object");

   for(int j = 0; j < i; j++)
   {
      automatics[j].value = ParseObjects::makeImport(-1, -1,
         $("/evlan.org/common/") + automatics[j].name);
   }

   automatics[i].name = $("*cons");   //List.make; not meant to be typed
   automatics[i].value = ParseObjects::makeMemberAccess(-1, -1,
      ParseObjects::makeVariable(-1, -1, $("List"), 0), $("make"));
   ++i;

   automatics[i].name = $("empty");   //List.empty
   automatics[i].value = ParseObjects::makeMemberAccess(-1, -1,
      ParseObjects::makeVariable(-1, -1, $("List"), 0), $("empty"));
   ++i;

   for(int j = 0; j < i; j++)
   {
      automatics[j].line = -1;
      automatics[j].column = -1;
   }

   result = ParseObjects::makeWhereBlock(-1, -1, result, automatics.toConst());

   mInput = null;
   mErrorCollector = null;
   return result;
}

ParseTreePtr BasicParser::parseFunction(bool hadOpenParen)
{
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;
   util::Vector<String> paramNames;

   if(getKeywordId(mCurrentToken) != mKeywordSet->mOpCloseParenId)
   {
      while(true)
      {
         FW_MATCH_START(mCurrentToken.value)
         FW_MATCH_CASE_1(Token::Value::identifier, name)
            paramNames.pushBack(name);
            shift();
         FW_MATCH_DEFAULT
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected identifier."));
         FW_MATCH_END

         if(!hadOpenParen)
            break;
         else if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCloseParenId)
         {
            shift();
            break;
         }
         else if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCommaId)
            shift();
         else
         {
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected '") + mKeywordSet->mOpCommaName + $("'."));

            if(mCurrentToken.value.getId() == Token::Value::endOfSource_ID)
               break;

            shift();
         }
      }
   }
   else
      shift();

   if(getKeywordId(mCurrentToken) == mKeywordSet->mOpAbstractId)
   {
      //'=>'
      shift();
      ParseTreePtr body = parseExpression();
      return ParseObjects::makeFunction(line, column, paramNames.getArray().toConst(), body);
   }
   else
   {
      mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
         $("Expected '") + mKeywordSet->mOpAbstractName + $("'."));
      return ParseObjects::makeFunction(line, column, paramNames.getArray().toConst(),
         makeEmptyParseTree(mCurrentToken.line, mCurrentToken.column));
   }
}

ParseTreePtr BasicParser::parseArray()
{
   //current token is '{'
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;
   util::Vector<ParseTreePtr> elements;

   shift();

   if(getKeywordId(mCurrentToken) != mKeywordSet->mOpCloseBraceId)
   {
      while(true)
      {
         elements.pushBack(parseExpression());
         if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCloseBraceId)
            break;
         else if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCommaId)
            shift();
         else
         {
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected '") + mKeywordSet->mOpCommaName + $("'."));

            if(!skipRestOfElement())
               break;
         }
      }
   }

   return ParseObjects::makeArray(line, column, elements.getArray().toConst());
}

ParseTreePtr BasicParser::parseBlockArray()
{
   //current token is 'array'
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;
   util::Vector<ParseTreePtr> elements;

   shift();

   if(getKeywordId(mCurrentToken) == mKeywordSet->mOfId)
   {
      shift();
      shift();  //startOfBlock (checked by block parser)

      //parse main block...
      while(mCurrentToken.value.getId() != Token::Value::endOfBlock_ID
         && mCurrentToken.value.getId() != Token::Value::endOfSource_ID)
      {
         if(mCurrentToken.value.getId() == Token::Value::comment_ID)
            shift();   //ignore statement documentation
         elements.pushBack(parseExpression());

         if(mCurrentToken.value.getId() == Token::Value::endOfStatement_ID)
            shift();
         else
         {
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected end of statement."));
            if(!skipRestOfStatement())
               break;
         }
      }

      if(mCurrentToken.value.getId() == Token::Value::endOfBlock_ID)
         shift();
      else
      {
         mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
            $("Expected end of block."));
      }
   }

   return ParseObjects::makeArray(line, column, elements.getArray().toConst());
}

ParseTreePtr BasicParser::parseConditional()
{
   //current token is 'if'
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;

   shift();
   ParseTreePtr input = parseExpression();
   ParseTreePtr thenClause, elseClause;

   if(getKeywordId(mCurrentToken) == mKeywordSet->mDoId)
   {
      thenClause = parseDo();
      elseClause = ParseObjects::makeApplication(line, column,
         ParseObjects::makeMemberAccess(line, column,
            ParseObjects::makeVariable(line, column, $("Task")), $("returnNone")),
         Array<ParseTreePtr>(null));
   }
   else
   {
      if(getKeywordId(mCurrentToken) != mKeywordSet->mThenId)
      {
         mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
            $("Expected '") + mKeywordSet->mThenName + $("'."));
      }
      else
         shift();
      thenClause = parseExpression();

      if(getKeywordId(mCurrentToken) != mKeywordSet->mElseId)
      {
         mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
            $("Expected '") + mKeywordSet->mElseName + $("'."));
      }
      else
         shift();
      elseClause = parseExpression();
   }

   return ParseObjects::makeConditional(line, column, input,
                                        thenClause, elseClause);
}

ParseTreePtr BasicParser::parseObject()
{
   //current token is 'object'
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;
   util::Vector<ParseObjects::MemberDefinition> members;

   shift();

   if(getKeywordId(mCurrentToken) == mKeywordSet->mOfId)
   {
      shift();
      shift();  //startOfBlock

      //parse main block...
      while(mCurrentToken.value.getId() != Token::Value::endOfBlock_ID
         && mCurrentToken.value.getId() != Token::Value::endOfSource_ID)
      {
         ParseObjects::MemberDefinition member;
         member.line = mCurrentToken.line;
         member.column = mCurrentToken.column;

         Statement statement = parseStatement();
         member.documentation = statement.documentation;
         member.name = statement.valueName;
         member.value = statement.value;

         members.pushBack(member);
      }
      if(mCurrentToken.value.getId() == Token::Value::endOfBlock_ID)
         shift();
   }

   return ParseObjects::makeObject(line, column, members.getArray().toConst());
}

ParseTreePtr BasicParser::parseTry()
{
   //current token is 'try'
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;

   shift();
   ParseTreePtr body = parseExpression();

   util::Vector<ParseObjects::MemberDefinition> members;

#if 0
   //block-style catch...  I decided I don't like it so much
   if(getKeywordId(mCurrentToken) == mKeywordSet->mCatchId)
   {
      shift();
      shift();  //startOfBlock

      //parse main block...
      while(mCurrentToken.value.getId() != Token::Value::endOfBlock_ID
         && mCurrentToken.value.getId() != Token::Value::endOfSource_ID)
      {
         ParseObjects::MemberDefinition member;
         member.line = mCurrentToken.line;
         member.column = mCurrentToken.column;

         Statement statement = parseStatement();
         member.name = statement.valueName;
         member.value = statement.value;

         members.pushBack(member);
      }
      if(mCurrentToken.value.getId() == Token::Value::endOfBlock_ID)
         shift();
   }
   else
   {
      mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
         $("Expected \"") + mKeywordSet->mCatchName + $("\"."));
   }
#else
   while(getKeywordId(mCurrentToken) == mKeywordSet->mCatchId)
   {
      shift();

      ParseObjects::MemberDefinition member;
      member.line = mCurrentToken.line;
      member.column = mCurrentToken.column;

      FW_MATCH_START(mCurrentToken.value)
      FW_MATCH_CASE_1(Token::Value::identifier, name)
         member.name = name;
         shift();
      FW_MATCH_DEFAULT
         mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
            $("Expected identifier."));
      FW_MATCH_END

      member.value = parseExpression();

      members.pushBack(member);
   }
#endif

   if(members.getSize() > 0)
   {
      ParseTreePtr catchDescriptor = ParseObjects::makeObject(line, column, members.getArray().toConst());

      Array<ParseTreePtr> parameters(2);
      parameters[0] = body;
      parameters[1] = catchDescriptor;

      body = ParseObjects::makeApplication(line, column,
         ParseObjects::makeMemberAccess(line, column,
            ParseObjects::makeVariable(line, column, $("Task")), $("try")),
         parameters);
   }

   if(getKeywordId(mCurrentToken) == mKeywordSet->mFinallyId)
   {
      shift();

      Array<ParseTreePtr> parameters(2);
      parameters[0] = body;
      parameters[1] = parseExpression();

      body = ParseObjects::makeApplication(line, column,
         ParseObjects::makeMemberAccess(line, column,
            ParseObjects::makeVariable(line, column, $("Task")), $("finally")),
         parameters);
   }

   return body;
}

ParseTreePtr BasicParser::parseThrow()
{
   //current token is 'throw'
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;

   shift();

   util::Vector<ParseTreePtr> elements;

   while(true)
   {
      int subLine = mCurrentToken.line;
      int subColumn = mCurrentToken.column;

      String typeName;

      FW_MATCH_START(mCurrentToken.value)
      FW_MATCH_CASE_1(Token::Value::identifier, name)
         typeName = name;
         shift();
      FW_MATCH_DEFAULT
      FW_MATCH_END

      if(typeName == null)
         break;

      elements.pushBack(
         ParseObjects::makeLiteral(subLine, subColumn,
                                   Expression::Literal::atom(typeName)));

      int pos = elements.getSize();
      elements.pushBack(null);  //fill in later

      if(getKeywordId(mCurrentToken) == mKeywordSet->mOpOpenParenId)
      {
         shift();

         if(getKeywordId(mCurrentToken) != mKeywordSet->mOpCloseParenId)
         {
            elements.pushBack(parseExpression());

            while(getKeywordId(mCurrentToken) == mKeywordSet->mOpCommaId)
            {
               shift();
               elements.pushBack(parseExpression());
            }
         }

         if(getKeywordId(mCurrentToken) != mKeywordSet->mOpCloseParenId)
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column, $("Missing close paren at end of parameter list."));
         else
            shift();
      }

      elements[pos] =
         ParseObjects::makeLiteral(subLine, subColumn,
                                   Expression::Literal::integer(elements.getSize() - pos - 1));
   }

   return ParseObjects::makeApplication(line, column,
      ParseObjects::makeMemberAccess(line, column,
         ParseObjects::makeVariable(line, column, $("Task")), $("throw")),
      ParseObjects::makeArray(line, column, elements.getArray().toConst()));
}

ParseTreePtr BasicParser::parseBracedObject()
{
   //current token is '{'
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;
   util::Vector<ParseObjects::MemberDefinition> members;

   shift();

   if(getKeywordId(mCurrentToken) != mKeywordSet->mOpCloseBraceId)
   {
      while(true)
      {
         ParseObjects::MemberDefinition member;

         member.line = mCurrentToken.line;
         member.column = mCurrentToken.column;

         FW_MATCH_START(mCurrentToken.value)
         FW_MATCH_CASE_1(Token::Value::identifier, name)
            member.name = name;
            shift();
         FW_MATCH_DEFAULT
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected identifier."));
         FW_MATCH_END

         if(getKeywordId(mCurrentToken) == mKeywordSet->mOpAssignId)
            shift();
         else
         {
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected '") + mKeywordSet->mOpAssignName + $("'."));
         }

         member.value = parseExpression();
         members.pushBack(member);

         if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCloseBraceId)
            break;
         else if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCommaId)
            shift();
         else
         {
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected '") + mKeywordSet->mOpCommaName + $("'."));

            if(!skipRestOfElement())
               break;
         }
      }
   }

   return ParseObjects::makeObject(line, column, members.getArray().toConst());
}

ParseTreePtr BasicParser::parseImport()
{
   //current token is 'import'
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;

   shift();

   FW_MATCH_START(mCurrentToken.value)
   FW_MATCH_CASE_1(Token::Value::string8, s)
      shift();
      return ParseObjects::makeImport(line, column, s);
   FW_MATCH_CASE_1(Token::Value::string16, s)
      shift();
      return ParseObjects::makeImport(line, column, s);
   FW_MATCH_CASE_1(Token::Value::string32, s)
      shift();
      return ParseObjects::makeImport(line, column, s);
   FW_MATCH_DEFAULT
      mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
         $("Expected string."));
   FW_MATCH_END

   return ParseObjects::makeImport(line, column, String(null));
}

ParseTreePtr BasicParser::parseDo()
{
   //current token is "do"
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;
   shift();

   if(mCurrentToken.value.getId() != Token::Value::startOfBlock_ID)
      return makeEmptyParseTree(mCurrentToken.line, mCurrentToken.column);

   util::Vector<ParseObjects::DoStatement> statements;
   ParseTreePtr returnValue;
   bool hadStatementsAfterReturn = false;

   shift();  //block start

   while(mCurrentToken.value.getId() != Token::Value::endOfBlock_ID
      && mCurrentToken.value.getId() != Token::Value::endOfSource_ID)
   {
      if(mCurrentToken.value.getId() == Token::Value::comment_ID)
         shift();   //ignore statement documentation

      if(returnValue != null && !hadStatementsAfterReturn)
      {
         mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
            $("Return must be last statement in do block."));
         hadStatementsAfterReturn = true;
      }

      String variableName;

      enum
      {
         PROCEDURE,
         VALUE,
         RETURN
      } statementType = PROCEDURE;

      FW_MATCH_START(mCurrentToken.value)
      FW_MATCH_CASE_1(Token::Value::identifier, name)
         if(getKeywordId(mInput[mPos+1]) == mKeywordSet->mOpStoreId)
         {
            variableName = name;
            shift();  //identifier
            shift();  //:=
         }
         else if(getKeywordId(mInput[mPos+1]) == mKeywordSet->mOpAssignId)
         {
            statementType = VALUE;
            variableName = name;
            shift();  //identifier
            shift();  //=
         }
      FW_MATCH_CASE_1(Token::Value::keyword, keywordId)
         if(keywordId == mKeywordSet->mReturnId)
         {
            statementType = RETURN;
            shift();  //return
         }
      FW_MATCH_DEFAULT
         //nothing
      FW_MATCH_END

      if(statementType == RETURN)
      {
         if(returnValue == null)
            returnValue = parseExpression();
         else
            parseExpression();
      }
      else
      {
         ParseObjects::DoStatement statement;
         statement.line = mCurrentToken.line;
         statement.column = mCurrentToken.column;
         statement.variable = variableName;
         statement.body = parseExpression();
         statement.isProcedure = (statementType == PROCEDURE);

         if(returnValue == null)
            statements.pushBack(statement);
      }

      if(mCurrentToken.value.getId() == Token::Value::endOfStatement_ID)
         shift();
      else
      {
         mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
            $("Expected end of statement."));
         if(!skipRestOfStatement())
            break;
      }
   }

   if(mCurrentToken.value.getId() == Token::Value::endOfBlock_ID)
      shift();
   else
   {
      mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
         $("Expected end of block."));
   }

   return ParseObjects::makeDo(line, column, statements.getArray().toConst(), returnValue);
}

ParseTreePtr BasicParser::parseString()
{
   int totalLength = 0;
   bool done = false;
   int endPos = mPos;
   int maxSize = 8;
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;

   while(!done)
   {
      FW_MATCH_START(mInput[endPos].value)
      FW_MATCH_CASE_1(Token::Value::string8, s2)
         totalLength += s2.getSize();
         ++endPos;
      FW_MATCH_CASE_1(Token::Value::string16, s2)
         totalLength += s2.getSize();
         if(maxSize < 16)
            maxSize = 16;
         ++endPos;
      FW_MATCH_CASE_1(Token::Value::string32, s2)
         totalLength += s2.getSize();
         maxSize = 32;
         ++endPos;
      FW_MATCH_DEFAULT
         done = true;
      FW_MATCH_END
   }

   if(endPos == mPos + 1)
   {
      shift();
      return null;
   }

   switch(maxSize)
   {
   default:
   case 8:
      {
         Array<char8> text(totalLength);
         int filled = 0;
         done = false;
         while(!done)
         {
            FW_MATCH_START(mCurrentToken.value)
            FW_MATCH_CASE_1(Token::Value::string8, s)
               for(int i = 0; i < s.getSize(); i++)
                  text[filled++] = s[i];
               shift();
            FW_MATCH_DEFAULT
               done = true;
            FW_MATCH_END
         }
         return ParseObjects::makeLiteral(line, column, Expression::Literal::string8(text.toConst()));
      }
   case 16:
      {
         Array<char16> text(totalLength);
         int filled = 0;
         done = false;
         while(!done)
         {
            FW_MATCH_START(mCurrentToken.value)
            FW_MATCH_CASE_1(Token::Value::string8, s)
               for(int i = 0; i < s.getSize(); i++)
                  text[filled++] = (byte)s[i];
               shift();
            FW_MATCH_CASE_1(Token::Value::string16, s)
               for(int i = 0; i < s.getSize(); i++)
                  text[filled++] = s[i];
               shift();
            FW_MATCH_DEFAULT
               done = true;
            FW_MATCH_END
         }
         return ParseObjects::makeLiteral(line, column, Expression::Literal::string16(text.toConst()));
      }
   case 32:
      {
         Array<char32> text(totalLength);
         int filled = 0;
         done = false;
         while(!done)
         {
            FW_MATCH_START(mCurrentToken.value)
            FW_MATCH_CASE_1(Token::Value::string8, s)
               for(int i = 0; i < s.getSize(); i++)
                  text[filled++] = (byte)s[i];
               shift();
            FW_MATCH_CASE_1(Token::Value::string16, s)
               for(int i = 0; i < s.getSize(); i++)
                  text[filled++] = s[i];
               shift();
            FW_MATCH_CASE_1(Token::Value::string32, s)
               for(int i = 0; i < s.getSize(); i++)
                  text[filled++] = s[i];
               shift();
            FW_MATCH_DEFAULT
               done = true;
            FW_MATCH_END
         }
         return ParseObjects::makeLiteral(line, column, Expression::Literal::string32(text.toConst()));
      }
   }
}

ParseTreePtr BasicParser::parseItem()
{
   int line = mCurrentToken.line;
   int column = mCurrentToken.column;
   ParseTreePtr result;

   FW_MATCH_START(mCurrentToken.value)
   FW_MATCH_CASE_1(Token::Value::identifier, name)
      if(getKeywordId(mInput[mPos+1]) == mKeywordSet->mOpAbstractId)
         result = parseFunction(false);
      else
      {
         result = ParseObjects::makeVariable(line, column, name);
         shift();
      }

   FW_MATCH_CASE_1(Token::Value::integer, i)
      result = ParseObjects::makeLiteral(line, column, Expression::Literal::integer(i));
      shift();
   FW_MATCH_CASE_1(Token::Value::integer64, i)
      result = ParseObjects::makeLiteral(line, column, Expression::Literal::integer64(i));
      shift();
   FW_MATCH_CASE_1(Token::Value::real, r)
      result = ParseObjects::makeLiteral(line, column, Expression::Literal::real(r));
      shift();
   FW_MATCH_CASE_1(Token::Value::data8, d)
      result = ParseObjects::makeLiteral(line, column, Expression::Literal::data8(d));
      shift();
   FW_MATCH_CASE_1(Token::Value::data16, d)
      result = ParseObjects::makeLiteral(line, column, Expression::Literal::data16(d));
      shift();
   FW_MATCH_CASE_1(Token::Value::data32, d)
      result = ParseObjects::makeLiteral(line, column, Expression::Literal::data32(d));
      shift();
   FW_MATCH_CASE_1(Token::Value::data64, d)
      result = ParseObjects::makeLiteral(line, column, Expression::Literal::data64(d));
      shift();
   FW_MATCH_CASE_1(Token::Value::character, c)
      result = ParseObjects::makeLiteral(line, column, Expression::Literal::character(c));
      shift();
   FW_MATCH_CASE_1(Token::Value::atom, s)
      result = ParseObjects::makeLiteral(line, column, Expression::Literal::atom(s));
      shift();
   FW_MATCH_CASE_1(Token::Value::string8, s)
      result = parseString();
      if(result == null)
         result = ParseObjects::makeLiteral(line, column, Expression::Literal::string8(s));
   FW_MATCH_CASE_1(Token::Value::string16, s)
      result = parseString();
      if(result == null)
         result = ParseObjects::makeLiteral(line, column, Expression::Literal::string16(s));
   FW_MATCH_CASE_1(Token::Value::string32, s)
      result = parseString();
      if(result == null)
         result = ParseObjects::makeLiteral(line, column, Expression::Literal::string32(s));

   FW_MATCH_CASE_1(Token::Value::keyword, keywordId)
      if(keywordId == mKeywordSet->mTrueId)
      {
         result = ParseObjects::makeLiteral(line, column, Expression::Literal::boolean(true));
         shift();
      }
      else if(keywordId == mKeywordSet->mFalseId)
      {
         result = ParseObjects::makeLiteral(line, column, Expression::Literal::boolean(false));
         shift();
      }
      else if(keywordId == mKeywordSet->mOpOpenParenId)
      {
         bool isFunction = false;
         shift();

         FW_MATCH_START(mCurrentToken.value)
         FW_MATCH_CASE_1(Token::Value::identifier, name)
            Token next = mInput[mPos+1];
            if(getKeywordId(next) == mKeywordSet->mOpCommaId
            || (getKeywordId(next) == mKeywordSet->mOpCloseParenId
             && getKeywordId(mInput[mPos+2]) == mKeywordSet->mOpAbstractId))
            {
               isFunction = true;
               result = parseFunction(true);
            }
         FW_MATCH_CASE_1(Token::Value::keyword, keywordId2)
            if(keywordId2 == mKeywordSet->mOpCloseParenId)
            {
               isFunction = true;
               result = parseFunction(true);
            }
         FW_MATCH_DEFAULT
            //nothing...
         FW_MATCH_END

         if(!isFunction)
         {
            ParseTreePtr expression = parseExpression();

            if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCloseParenId)
               shift();
            else
            {
               mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
                  $("Expected '") + mKeywordSet->mOpCloseParenName + $("'."));
            }
            result = expression;
         }
      }
      else if(keywordId == mKeywordSet->mOpOpenBraceId)
      {
         bool isObject = false;

         FW_MATCH_START(mInput[mPos+1].value)
         FW_MATCH_CASE_1(Token::Value::identifier, name)
            if(getKeywordId(mInput[mPos+2]) == mKeywordSet->mOpAssignId)
               isObject = true;
         FW_MATCH_DEFAULT
            //nothing...
         FW_MATCH_END

         if(isObject)
            result = parseBracedObject();
         else
            result = parseArray();

         if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCloseBraceId)
            shift();
         else
         {
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected '") + mKeywordSet->mOpCloseBraceName + $("'."));
         }
      }
      else if(keywordId == mKeywordSet->mArrayId)
         result = parseBlockArray();
      else if(keywordId == mKeywordSet->mIfId)
         result = parseConditional();
      else if(keywordId == mKeywordSet->mObjectId)
         result = parseObject();
      else if(keywordId == mKeywordSet->mImportId)
         result = parseImport();
      else if(keywordId == mKeywordSet->mDoId)
         result = parseDo();
      else if(keywordId == mKeywordSet->mTryId)
         result = parseTry();
      else if(keywordId == mKeywordSet->mThrowId)
         result = parseThrow();
      else if(keywordId == mKeywordSet->mOpScopeId)
      {
         int depth = 0;
         do
         {
            ++depth;
            shift();
         } while(getKeywordId(mCurrentToken) == mKeywordSet->mOpScopeId);

         FW_MATCH_START(mCurrentToken.value)
         FW_MATCH_CASE_1(Token::Value::identifier, name)
            result = ParseObjects::makeVariable(
               mCurrentToken.line, mCurrentToken.column, name, depth);
            shift();
         FW_MATCH_DEFAULT
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
                                      $("Expected identifier."));
         FW_MATCH_END
      }
   FW_MATCH_DEFAULT
      //nothing
   FW_MATCH_END

   if(result == null)
   {
      mErrorCollector->addError(line, column, $("Expected item."));
      result = makeEmptyParseTree(line, column);
   }
   else while(true)
   {
      if(getKeywordId(mCurrentToken) == mKeywordSet->mOpDotId)
      {
         shift();
         FW_MATCH_START(mCurrentToken.value)
         FW_MATCH_CASE_1(Token::Value::identifier, name)
            result = ParseObjects::makeMemberAccess(
               mCurrentToken.line, mCurrentToken.column, result, name);
            shift();
         FW_MATCH_DEFAULT
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column, $("Expected identifier."));
         FW_MATCH_END
      }
      else if(getKeywordId(mCurrentToken) == mKeywordSet->mOfId)
      {
         int line = mCurrentToken.line;
         int column = mCurrentToken.column;

         shift();
         shift();   //startOfBlock

         util::Vector<ParseTreePtr> parameters;

         while(mCurrentToken.value.getId() != Token::Value::endOfBlock_ID
            && mCurrentToken.value.getId() != Token::Value::endOfSource_ID)
         {
            if(mCurrentToken.value.getId() == Token::Value::comment_ID)
               shift();   //ignore statement documentation
            parameters.pushBack(parseExpression());

            if(mCurrentToken.value.getId() == Token::Value::endOfStatement_ID)
               shift();
            else
            {
               mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
                  $("Expected end of statement."));
               if(!skipRestOfStatement())
                  break;
            }
         }

         result = ParseObjects::makeApplication(line, column, result, parameters.getArray());

         if(mCurrentToken.value.getId() == Token::Value::endOfBlock_ID)
            shift();
         else
         {
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected end of block."));
         }
      }
      else if(getKeywordId(mCurrentToken) == mKeywordSet->mOpOpenParenId)
      {
         int line = mCurrentToken.line;
         int column = mCurrentToken.column;

         shift();
         util::Vector<ParseTreePtr> parameters;
         if(getKeywordId(mCurrentToken) != mKeywordSet->mOpCloseParenId)
         {
            parameters.pushBack(parseExpression());

            while(getKeywordId(mCurrentToken) == mKeywordSet->mOpCommaId)
            {
               shift();
               parameters.pushBack(parseExpression());
            }
         }

         result = ParseObjects::makeApplication(line, column, result, parameters.getArray());

         if(getKeywordId(mCurrentToken) != mKeywordSet->mOpCloseParenId)
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column, $("Missing close paren at end of parameter list."));
         else
            shift();
      }
      else if(getKeywordId(mCurrentToken) == mKeywordSet->mOpOpenBracketId)
      {
         int line = mCurrentToken.line;
         int column = mCurrentToken.column;

         result = ParseObjects::makeMemberAccess(line, column, result, $("getElement"));

         shift();
         Array<ParseTreePtr> parameters(1);
         parameters[0] = parseExpression();

         result = ParseObjects::makeApplication(line, column, result, parameters);

         if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCloseBracketId)
            shift();
         else
         {
            mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
               $("Expected '") + mKeywordSet->mOpCloseBracketName + $("'."));
         }
      }
      else
         break;
   }

   return result;
}


ParseTreePtr BasicParser::parseUnaryOps()
{
   util::Atom op = getKeywordId(mCurrentToken);
   String funcName;
   if(op != null)
   {
      std::map<util::Atom, String>::iterator iter;
      iter = mUnaryOps.find(op);
      if(iter != mUnaryOps.end())
         funcName = iter->second;
   }

   if(funcName == null)
      return parseItem();
   else
   {
      int line = mCurrentToken.line;
      int column = mCurrentToken.column;
      shift();
      ParseTreePtr item = parseUnaryOps();
      return ParseObjects::makeMemberAccess(line, column, item, funcName);
   }
}

struct OpStackElement
{
   ParseTreePtr item;
   util::Atom op;
   int priority;
   int line;
   int column;

   inline OpStackElement(ParseTreePtr i, util::Atom o, int p, int l, int c)
      : item(i), op(o), priority(p), line(l), column(c) {}
};

ParseTreePtr BasicParser::parseBinaryOps()
{
   std::stack<OpStackElement> opStack;

   while(true)
   {
      ParseTreePtr item = parseUnaryOps();
      util::Atom op = getKeywordId(mCurrentToken);
      int priority = -1;
      if(op != null)
      {
         std::map<util::Atom, int>::iterator iter = mOpPriorities.find(op);
         if(iter != mOpPriorities.end())
            priority = iter->second;
      }

      while(!opStack.empty() && priority <= opStack.top().priority)
      {
         //item = apply opStack.top().op to opStack.top().item and item
         String funcName = mBinaryOps[opStack.top().op];
         if(funcName != null)
         {
            ParseTreePtr func =
               ParseObjects::makeMemberAccess(
                  opStack.top().line, opStack.top().column,
                  opStack.top().item, funcName);
            item =
               ParseObjects::makeApplication(
                  opStack.top().line, opStack.top().column,
                  func, item);
         }
         else if(opStack.top().op == mKeywordSet->mOpGreaterId)
         {
            ParseTreePtr func =
               ParseObjects::makeMemberAccess(
                  opStack.top().line, opStack.top().column,
                  item, $("isLessThan"));
            item =
               ParseObjects::makeApplication(
                  opStack.top().line, opStack.top().column,
                  func, opStack.top().item);
         }
         else if(opStack.top().op == mKeywordSet->mOpLessEqualId)
         {
            ParseTreePtr func =
               ParseObjects::makeMemberAccess(
                  opStack.top().line, opStack.top().column,
                  item, $("isLessThan"));
            item =
               ParseObjects::makeApplication(
                  opStack.top().line, opStack.top().column,
                  func, opStack.top().item);
            item =
               ParseObjects::makeMemberAccess(
                  opStack.top().line, opStack.top().column,
                  item, $("not"));
         }
         else if(opStack.top().op == mKeywordSet->mOpGreaterEqualId)
         {
            ParseTreePtr func =
               ParseObjects::makeMemberAccess(
                  opStack.top().line, opStack.top().column,
                  opStack.top().item, $("isLessThan"));
            item =
               ParseObjects::makeApplication(
                  opStack.top().line, opStack.top().column,
                  func, item);
            item =
               ParseObjects::makeMemberAccess(
                  opStack.top().line, opStack.top().column,
                  item, $("not"));
         }
         else if(opStack.top().op == mKeywordSet->mOpNotEqualId)
         {
            ParseTreePtr func =
               ParseObjects::makeMemberAccess(
                  opStack.top().line, opStack.top().column,
                  opStack.top().item, $("equals"));
            item =
               ParseObjects::makeApplication(
                  opStack.top().line, opStack.top().column,
                  func, item);
            item =
               ParseObjects::makeMemberAccess(
                  opStack.top().line, opStack.top().column,
                  item, $("not"));
         }
         opStack.pop();
      }

      if(priority == -1)
         return item;
      else
      {
         opStack.push(OpStackElement(item, op, priority,
            mCurrentToken.line, mCurrentToken.column));
         shift();
      }
   }
}


struct ConsStackElement
{
   ParseTreePtr item;
   int line;
   int column;

   inline ConsStackElement(ParseTreePtr i, int l, int c)
      : item(i), line(l), column(c) {}
};

ParseTreePtr BasicParser::parseCons()
{
   ParseTreePtr item = parseBinaryOps();

   if(getKeywordId(mCurrentToken) == mKeywordSet->mOpConsId)
   {
      std::stack<ConsStackElement> consStack;
      consStack.push(ConsStackElement(item, mCurrentToken.line, mCurrentToken.column));

      do
      {
         shift();
         item = parseBinaryOps();
         consStack.push(ConsStackElement(item, mCurrentToken.line, mCurrentToken.column));
      } while(getKeywordId(mCurrentToken) == mKeywordSet->mOpConsId);

      item = consStack.top().item;
      consStack.pop();

      while(!consStack.empty())
      {
         Array<ParseTreePtr> parameters(2);
         parameters[0] = consStack.top().item;
         parameters[1] = item;
         item = ParseObjects::makeApplication(
            consStack.top().line, consStack.top().column,
            ParseObjects::makeVariable(
               consStack.top().line, consStack.top().column, $("*cons"), 0),
            parameters);
         consStack.pop();
      }
   }

   return item;
}

ParseTreePtr BasicParser::parseWhere()
{
   ParseTreePtr output = parseCons();

   if(getKeywordId(mCurrentToken) == mKeywordSet->mWhereId)
   {
      int line = mCurrentToken.line;
      int column = mCurrentToken.column;
      util::Vector<ParseObjects::WhereItem> items;

      shift();
      shift();  //startOfBlock

      //parse main block...
      while(mCurrentToken.value.getId() != Token::Value::endOfBlock_ID
         && mCurrentToken.value.getId() != Token::Value::endOfSource_ID)
      {
         ParseObjects::WhereItem item;
         item.line = mCurrentToken.line;
         item.column = mCurrentToken.column;

         Statement statement = parseStatement();
         item.documentation = statement.documentation;
         item.name = statement.valueName;
         item.value = statement.value;

         items.pushBack(item);
      }
      if(mCurrentToken.value.getId() == Token::Value::endOfBlock_ID)
         shift();

      return ParseObjects::makeWhereBlock(line, column, output,
                                          items.getArray().toConst());
   }
   else
      return output;
}

ParseTreePtr BasicParser::parseExpression()
{
   return parseWhere();
}

BasicParser::Statement BasicParser::parseStatement()
{
   Statement result;

   FW_MATCH_START(mCurrentToken.value)
   FW_MATCH_CASE_1(Token::Value::comment, text)
      result.documentation = text;
      shift();
   FW_MATCH_END

   FW_MATCH_START(mCurrentToken.value)
   FW_MATCH_CASE_1(Token::Value::identifier, name)
      result.valueName = name;
      shift();
   FW_MATCH_DEFAULT
      mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
                                $("Expected identifier."));
   FW_MATCH_END

   if(getKeywordId(mCurrentToken) == mKeywordSet->mOpAssignId)
      shift();
   else
   {
      mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
         $("Expected '") + mKeywordSet->mOpAssignName + $("'."));
   }

   result.value = parseExpression();

   if(mCurrentToken.value.getId() == Token::Value::endOfStatement_ID)
      shift();
   else
   {
      mErrorCollector->addError(mCurrentToken.line, mCurrentToken.column,
         $("Expected end of statement."));
      skipRestOfStatement();
   }

   return result;
}

bool BasicParser::skipRestOfStatement()
{
   int blockLevel = 0;
   while(true)
   {
      if(blockLevel == 0 && mCurrentToken.value.getId() == Token::Value::endOfStatement_ID)
      {
         shift();
         return true;
      }
      else if(mCurrentToken.value.getId() == Token::Value::startOfBlock_ID)
         ++blockLevel;
      else if(mCurrentToken.value.getId() == Token::Value::endOfBlock_ID)
      {
         if(blockLevel == 0)
            return false;
         else
            --blockLevel;
      }
      else if(mCurrentToken.value.getId() == Token::Value::endOfSource_ID)
         return false;
      shift();
   }
}

bool BasicParser::skipRestOfElement()
{
   int blockLevel = 0;
   while(true)
   {
      if(blockLevel == 0 && getKeywordId(mCurrentToken) == mKeywordSet->mOpCommaId)
      {
         shift();
         return true;
      }
      else if(getKeywordId(mCurrentToken) == mKeywordSet->mOpOpenBraceId)
         ++blockLevel;
      else if(getKeywordId(mCurrentToken) == mKeywordSet->mOpCloseBraceId)
      {
         if(blockLevel == 0)
            return false;
         else
            --blockLevel;
      }
      else if(mCurrentToken.value.getId() == Token::Value::endOfSource_ID
           || mCurrentToken.value.getId() == Token::Value::endOfBlock_ID
           || mCurrentToken.value.getId() == Token::Value::endOfStatement_ID)
         return false;
      shift();
   }
}

// factory function ==================================================
ParserPtr makeBasicParser(BasicKeywordSetPtr keywordSet)
{
   return spnew BasicParser(keywordSet);
}

} //namespace basic
} //namespace dialects
} //namespace compiler
} //namespace evlan
