/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "ParseObjects.h"

// begin namespace ===================================================
namespace evlan {
namespace compiler {
namespace dialects {
namespace basic {
using namespace grammar;

// members ===========================================================
namespace ParseObjects
{
   // Import =========================================================
   implementation Import: implements ParseTree
   {
      Import(int line, int column, String name)
         : mLine(line), mColumn(column), mName(name)
      {}

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         return Expression(mLine, mColumn, Expression::Value::import(mName));
      }

      void setDocumentation(String documentation) {}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":import \""));
         out->write(mName);
         out->write($("\""));
      }
#endif

      int mLine;
      int mColumn;
      String mName;
   };

   ParseTreePtr makeImport(int line, int column, String name)
   {
      return spnew Import(line, column, name);
   }

   ParseTreePtr makeImport(int line, int column, Array<const char8> name)
   {
      return spnew Import(line, column, String(name));
   }

   ParseTreePtr makeImport(int line, int column, Array<const char16> name)
   {
      Array<char> asciiName(name.getSize());
      for(int i = 0; i < name.getSize(); i++)
         asciiName[i] = name[i];
      return spnew Import(line, column, String(asciiName.toConst()));
   }

   ParseTreePtr makeImport(int line, int column, Array<const uint32> name)
   {
      Array<char> asciiName(name.getSize());
      for(int i = 0; i < name.getSize(); i++)
         asciiName[i] = name[i];
      return spnew Import(line, column, String(asciiName.toConst()));
   }

   // Literal ========================================================
   implementation Literal: implements ParseTree
   {
      Literal(int line, int column, Expression::Literal value)
         : mLine(line), mColumn(column), mValue(value)
      {}

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         return Expression(mLine, mColumn, Expression::Value::literal(mValue));
      }

      void setDocumentation(String documentation) {}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":literal "));

         FW_MATCH_START(mValue)
         FW_MATCH_CASE_1(Expression::Literal::integer, x)
            out->write($("integer "));
            out->write(String(x));
         FW_MATCH_CASE_1(Expression::Literal::integer64, x)
            out->write($("integer64 "));
            out->write(String(x));
         FW_MATCH_CASE_1(Expression::Literal::real, x)
            out->write($("real "));
            out->write(String(x));
         FW_MATCH_CASE_1(Expression::Literal::data8, x)
            out->write($("data8 0x"));
            out->write(String((uint32)x, 2, 16));
         FW_MATCH_CASE_1(Expression::Literal::data16, x)
            out->write($("data8 0x"));
            out->write(String((uint32)x, 4, 16));
         FW_MATCH_CASE_1(Expression::Literal::data32, x)
            out->write($("data8 0x"));
            out->write(String(x, 8, 16));
         FW_MATCH_CASE_1(Expression::Literal::data64, x)
            out->write($("data8 0x"));
            out->write(String(x, 16, 16));
         FW_MATCH_CASE_1(Expression::Literal::character, x)
            out->write($("character "));
            out->write(String(x));
         FW_MATCH_CASE_1(Expression::Literal::string8, x)
            out->write($("string"));
         FW_MATCH_CASE_1(Expression::Literal::string16, x)
            out->write($("string"));
         FW_MATCH_CASE_1(Expression::Literal::string32, x)
            out->write($("string"));
         FW_MATCH_END
      }
#endif

      int mLine;
      int mColumn;
      Expression::Literal mValue;
   };

   ParseTreePtr makeLiteral(int line, int column, Expression::Literal value)
   {
      return spnew Literal(line, column, value);
   }

   // Variable =======================================================
   implementation Variable: implements ParseTree
   {
      Variable(int line, int column, String name, int depth)
         : mLine(line), mColumn(column), mName(name), mDepth(depth)
      {}

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         Environment::Index index = context.getIndex(mName, mDepth);
         if(index.level < 0)
         {
            errorCollector->addError(mLine, mColumn,
               $("Undefined symbol: ") + mName);
         }
         return Expression(mLine, mColumn,
            Expression::Value::variable(index.level, index.item));
      }

      void setDocumentation(String documentation) {}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":variable "));
         out->write(mName);
      }
#endif

      int mLine;
      int mColumn;
      String mName;
      int mDepth;
   };

   ParseTreePtr makeVariable(int line, int column, String name, int depth)
   {
      return spnew Variable(line, column, name, depth);
   }

   // MemberAccess ===================================================
   implementation MemberAccess: implements ParseTree
   {
      MemberAccess(int line, int column,
                   ParseTreePtr object, String member)
         : mLine(line), mColumn(column), mObject(object), mMember(member)
      {}

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         return Expression(mLine, mColumn,
                           Expression::Value::memberAccess(
                              mObject->toExpression(context, errorCollector),
                              mMember));
      }

      void setDocumentation(String documentation) {}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":memberAccess\n"));
         out->write(indent);
         out->write($("object = "));
         mObject->print(indent + $("   "), out);
         out->write($("\n"));
         out->write(indent);
         out->write($("member = "));
         out->write(mMember);
      }
#endif

      int mLine;
      int mColumn;
      ParseTreePtr mObject;
      String mMember;
   };

   ParseTreePtr makeMemberAccess(int line, int column,
                                 ParseTreePtr object, String member)
   {
      return spnew MemberAccess(line, column, object, member);
   }

   // Array ==========================================================
   implementation ArrayConstructor: implements ParseTree
   {
      ArrayConstructor(int line, int column, Array<const ParseTreePtr> elements)
         : mLine(line), mColumn(column), mElements(elements)
      {}

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         Array<Expression> elements(mElements.getSize());
         for(int i = 0; i < elements.getSize(); i++)
            elements[i] = mElements[i]->toExpression(context, errorCollector);
         return Expression(mLine, mColumn,
                           Expression::Value::arrayConstructor(
                              elements.toConst()));
      }

      void setDocumentation(String documentation) {}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":array"));

         String subIndent = indent + $("   ");
         for(int i = 0; i < mElements.getSize(); i++)
         {
            out->write($("\n"));
            out->write(indent);
            out->write($("element"));
            out->write(String(i));
            out->write($(" = "));
            mElements[i]->print(subIndent, out);
         }
      }
#endif

      int mLine;
      int mColumn;
      Array<const ParseTreePtr> mElements;
   };

   ParseTreePtr makeArray(int line, int column,
                          Array<const ParseTreePtr> elements)
   {
      return spnew ArrayConstructor(line, column, elements);
   }

   // Conditional ====================================================
   implementation Conditional: implements ParseTree
   {
      Conditional(int line, int column, ParseTreePtr input,
                  ParseTreePtr thenClause, ParseTreePtr elseClause)
         : mLine(line), mColumn(column), mInput(input), mThenClause(thenClause), mElseClause(elseClause)
      {}

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         Expression input = mInput->toExpression(context, errorCollector);
         Expression thenClause = mThenClause->toExpression(context, errorCollector);
         Expression elseClause = mElseClause->toExpression(context, errorCollector);
         return Expression(mLine, mColumn,
            Expression::Value::conditional(input, thenClause, elseClause));
      }

      void setDocumentation(String documentation) {}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":if\n"));

         String subIndent = indent + $("   ");
         out->write(indent);
         out->write($("input = "));
         mInput->print(subIndent, out);
         out->write($("\n"));
         out->write(indent);
         out->write($("then = "));
         mThenClause->print(subIndent, out);
         out->write($("\n"));
         out->write(indent);
         out->write($("else = "));
         mElseClause->print(subIndent, out);
      }
#endif

      int mLine;
      int mColumn;
      ParseTreePtr mInput;
      ParseTreePtr mThenClause;
      ParseTreePtr mElseClause;
   };

   ParseTreePtr makeConditional(int line, int column, ParseTreePtr input,
                                ParseTreePtr thenClause, ParseTreePtr elseClause)
   {
      return spnew Conditional(line, column, input, thenClause, elseClause);
   }

   // WhereBlock =====================================================
   implementation WhereBlock: implements ParseTree
   {
      WhereBlock(int line, int column, ParseTreePtr output,
                 Array<const WhereItem> items)
         : mLine(line), mColumn(column), mItems(items), mOutput(output)
      {
         for(int i = 0; i < mItems.getSize(); i++)
            mItems[i].value->setDocumentation(mItems[i].documentation);
      }

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         Array<String> names(mItems.getSize());
         for(int i = 0; i < mItems.getSize(); i++)
            names[i] = mItems[i].name;
         Environment newContext = context.push(names.toConst());

         Array<Expression> items(mItems.getSize());
         for(int i = 0; i < items.getSize(); i++)
            items[i] = mItems[i].value->toExpression(newContext, errorCollector);
         Expression output = mOutput->toExpression(newContext, errorCollector);
         return Expression(mLine, mColumn,
                           Expression::Value::whereBlock(items.toConst(), output));
      }

      void setDocumentation(String documentation) {}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":where"));
         String subIndent = indent + $("   ");
         for(int i = 0; i < mItems.getSize(); i++)
         {
            out->write($("\n"));
            out->write(indent);
            out->write(mItems[i].name);
            out->write($(" = "));
            mItems[i].value->print(subIndent, out);
         }
         out->write($("\n"));
         out->write(indent);
         out->write($("result = "));
         mOutput->print(subIndent, out);
      }
#endif

      int mLine;
      int mColumn;
      Array<const WhereItem> mItems;
      ParseTreePtr mOutput;
   };

   ParseTreePtr makeWhereBlock(int line, int column, ParseTreePtr output,
                               Array<const WhereItem> items)
   {
      return spnew WhereBlock(line, column, output, items);
   }

   // Function =======================================================
   implementation Function: implements ParseTree
   {
      Function(int line, int column, Array<const String> inputNames, ParseTreePtr body)
         : mLine(line), mColumn(column), mInputNames(inputNames), mBody(body)
      {}

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         Environment newContext = context.push(mInputNames);
         return Expression(mLine, mColumn,
            Expression::Value::function(mInputNames,
               mBody->toExpression(newContext, errorCollector), mDocumentation));
      }

      void setDocumentation(String documentation) {mDocumentation = documentation;}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":abstraction\n"));

         String subIndent = indent + $("   ");
         out->write(indent);
         out->write($("body = "));
         mBody->print(subIndent, out);
      }
#endif

      int mLine;
      int mColumn;
      String mDocumentation;
      Array<const String> mInputNames;
      ParseTreePtr mBody;
   };

   ParseTreePtr makeFunction(int line, int column,
                             Array<const String> inputNames, ParseTreePtr body)
   {
      return spnew Function(line, column, inputNames, body);
   }

   // FunctionApplication ============================================
   implementation Application: implements ParseTree
   {
      Application(int line, int column,
                  ParseTreePtr function, Array<ParseTreePtr> parameters)
         : mLine(line), mColumn(column), mFunction(function), mParameters(parameters)
      {}

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         Expression function = mFunction->toExpression(context, errorCollector);
         Array<Expression> parameters(mParameters.getSize());
         for(int i = 0; i < parameters.getSize(); i++)
            parameters[i] = mParameters[i]->toExpression(context, errorCollector);
         return Expression(mLine, mColumn,
            Expression::Value::application(function, parameters));
      }

      void setDocumentation(String documentation) {}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         String subIndent = indent + $("   ");

         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":application\n"));
         out->write(indent);
         out->write($("function = "));
         mFunction->print(subIndent, out);
         for(int i = 0; i < mParameters.getSize(); i++)
         {
            out->write($("\n"));
            out->write(indent);
            out->write($("parameter"));
            out->write(String(i));
            out->write($(" = "));
            mParameters[i]->print(subIndent, out);
         }
      }
#endif

      int mLine;
      int mColumn;
      ParseTreePtr mFunction;
      Array<ParseTreePtr> mParameters;
   };

   ParseTreePtr makeApplication(int line, int column,
                                ParseTreePtr function,
                                Array<ParseTreePtr> parameters)
   {
      return spnew Application(line, column, function, parameters);
   }

   ParseTreePtr makeApplication(int line, int column,
                                ParseTreePtr function,
                                ParseTreePtr parameter)
   {
      Array<ParseTreePtr> parameters(1);
      parameters[0] = parameter;
      return spnew Application(line, column, function, parameters);
   }

   // ObjectDefinition ===============================================
   implementation ObjectDefinition: implements ParseTree
   {
      ObjectDefinition(int line, int column,
                       Array<const MemberDefinition> members)
         : mLine(line), mColumn(column), mMembers(members)
      {
         for(int i = 0; i < mMembers.getSize(); i++)
            mMembers[i].value->setDocumentation(mMembers[i].documentation);
      }

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         Array<String> memberNames(mMembers.getSize());
         for(int i = 0; i < memberNames.getSize(); i++)
            memberNames[i] = mMembers[i].name;
         Environment newContext = context.push(memberNames.toConst());

         Array<Expression::MemberDefinition> members(mMembers.getSize());
         for(int i = 0; i < members.getSize(); i++)
         {
            members[i].name = mMembers[i].name;
            members[i].value = mMembers[i].value->toExpression(newContext, errorCollector);
         }

         return Expression(mLine, mColumn,
            Expression::Value::objectDefinition(members.toConst(), mDocumentation));
      }

      void setDocumentation(String documentation) {mDocumentation = documentation;}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         String subIndent = indent + $("   ");

         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":object"));

         for(int i = 0; i < mMembers.getSize(); i++)
         {
            out->write($("\n"));
            out->write(indent);
            out->write($("membername"));
            out->write(String(i));
            out->write($(" = "));
            out->write(mMembers[i].name);

            out->write($("\n"));
            out->write(indent);
            out->write($("membervalue"));
            out->write(String(i));
            out->write($(" = "));
            mMembers[i].value->print(subIndent, out);
         }
      }
#endif

      int mLine;
      int mColumn;
      String mDocumentation;
      Array<const MemberDefinition> mMembers;
   };

   ParseTreePtr makeObject(int line, int column,
                           Array<const MemberDefinition> members)
   {
      return spnew ObjectDefinition(line, column, members);
   }

   // DoBlock ========================================================
   implementation DoBlock: implements ParseTree
   {
      DoBlock(int line, int column, Array<const DoStatement> statements, ParseTreePtr returnValue)
         : mLine(line), mColumn(column), mStatements(statements), mReturnValue(returnValue)
      {}

      Expression toExpression(Environment context,
                              ErrorCollectorPtr errorCollector)
      {
         Array<Environment> environments(mStatements.getSize()+1);

         environments[0] = context;

         for(int i = 1; i <= mStatements.getSize(); i++)
         {
            if(mStatements[i-1].variable == null)
               environments[i] = environments[i-1].push(Array<const String>(null));
            else
               environments[i] = environments[i-1].push(mStatements[i-1].variable);
         }

         Environment::Index taskIndex = environments[0].getIndex($("Task"), 0);
         String returnName;
         Array<Expression> returnParameters;
         if(mReturnValue == null)
            returnName = $("returnNone");
         else
         {
            returnName = $("return");
            returnParameters = Array<Expression>(1);
            returnParameters[0] =
               mReturnValue->toExpression(environments[mStatements.getSize()],
                                          errorCollector);
         }

         Expression next =
            Expression(mLine, mColumn,
               Expression::Value::application(
                  Expression(mLine, mColumn,
                     Expression::Value::memberAccess(
                        Expression(mLine, mColumn,
                           Expression::Value::variable(taskIndex.level, taskIndex.item)),
                        returnName)),
                  returnParameters));

         for(int i = mStatements.getSize() - 1; i >= 0; i--)
         {
            if(mStatements[i].isProcedure)
            {
               Array<String> variables(1);
               variables[0] = mStatements[i].variable;
               next = Expression(mStatements[i].line, mStatements[i].column,
                  Expression::Value::function(variables.toConst(), next, null));

               Expression body =
                  mStatements[i].body->toExpression(
                     environments[i], errorCollector);

               Array<Expression> parameters(2);
               parameters[0] = body;
               parameters[1] = next;

               //apply Task.call to body and next
               next = Expression(body.line, body.column,
                  Expression::Value::application(
                     Expression(body.line, body.column,
                        Expression::Value::memberAccess(
                           Expression(body.line, body.column,
                              Expression::Value::variable(taskIndex.level, taskIndex.item)),
                           $("call"))),
                     parameters));
            }
            else
            {
               Expression body =
                  mStatements[i].body->toExpression(
                     environments[i+1], errorCollector);

               Array<Expression> members(1);
               members[0] = body;
               next = Expression(body.line, body.column,
                                 Expression::Value::whereBlock(members.toConst(), next));
            }
         }

         return next;
      }

      void setDocumentation(String documentation) {}

#ifdef INCLUDE_WRITER_CODE
      void print(String indent, framework::TextWriterPtr out)
      {
         String subIndent = indent + $("   ");

         out->write(String(mLine));
         out->write($(":"));
         out->write(String(mColumn));
         out->write($(":do"));

         for(int i = 0; i < mStatements.getSize(); i++)
         {
            out->write($("\n"));
            out->write(indent);
            out->write($("statement"));
            out->write(String(i));
            out->write($(": "));

            if(mStatements[i].variable != null)
            {
               out->write(mStatements[i].variable);
               if(mStatements[i].isProcedure)
                  out->write($(" := "));
               else
                  out->write($(" = "));
            }

            mStatements[i].body->print(subIndent, out);
         }

         out->write($("\n"));
         out->write(indent);
         out->write($("return = "));
         mReturnValue->print(subIndent, out);
      }
#endif

      int mLine;
      int mColumn;
      Array<const DoStatement> mStatements;
      ParseTreePtr mReturnValue;
   };

   ParseTreePtr makeDo(int line, int column, Array<const DoStatement> statements, ParseTreePtr returnValue)
   {
      return spnew DoBlock(line, column, statements, returnValue);
   }

}

} //namespace dialects
} //namespace basic
} //namespace compiler
} //namespace evlan
