/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "BlockParser.h"
#include <stack>
#include <framework/util/Vector.h>

// begin namespace ===================================================
namespace evlan {
namespace compiler {
namespace syntax {

//Fuck the C++ standard for disallowing this, and shame on GCC for
//allowing it in previous versions then spontaneously breaking it.
//using ::framework::util;
namespace util {using namespace ::framework::util;}

namespace
{
   struct Line
   {
      Array<const Token> tokens;
      int indentLevel;
      int lineNumber;
      int length;
      bool hasCode;
      bool hasComment;
      bool endsWithBlockKeyword;
   };
}

Array<const Token> parseBlocks(Array<const Token> source,
                               ErrorCollectorPtr errorCollector,
                               KeywordSetPtr keywords)
{
   util::Vector<Line> lines;
   int pos = 0;

   while(pos < source.getSize())
   {
      int start = pos;
      bool done = false;
      Line line;
      line.lineNumber = 0;
      line.length = 0;
      line.indentLevel = 0;
      line.hasCode = false;
      line.hasComment = false;
      line.endsWithBlockKeyword = false;

      while(!done)
      {
         FW_MATCH_START(source[pos].value)
         FW_MATCH_CASE_0(Token::Value::endOfSource)
            line.lineNumber = source[pos].line;
            line.length = source[pos].column;
            done = true;
         FW_MATCH_CASE_0(Token::Value::endOfLine)
            line.lineNumber = source[pos].line;
            line.length = source[pos].column;
            done = true;
         FW_MATCH_CASE_1(Token::Value::whitespace, amount)
            if(source[pos].column == 0)
               line.indentLevel = amount;
         FW_MATCH_CASE_1(Token::Value::comment, text)
            line.hasComment = true;
         FW_MATCH_CASE_1(Token::Value::keyword, id)
            if(line.endsWithBlockKeyword)
            {
               errorCollector->addError(source[pos].line, source[pos].column,
                  $("Stuff after block keyword.  A block keyword must always "
                    "be the last item on a line."));
            }
            line.hasCode = true;
            line.endsWithBlockKeyword = keywords->isBlockKeyword(id);
         FW_MATCH_DEFAULT
            if(line.endsWithBlockKeyword)
            {
               errorCollector->addError(source[pos].line, source[pos].column,
                  $("Stuff after block keyword.  A block keyword must always "
                    "be the last item on a line."));
               line.endsWithBlockKeyword = false;
            }
            line.hasCode = true;
         FW_MATCH_END
         ++pos;
      }

      line.tokens = source.getInterval(start, pos);

      lines.pushBack(line);
   }

   int currentLevel = 0;
   for(int i = lines.getSize() - 1; i >= 0; i--)
   {
      if(lines[i].hasCode)
         currentLevel = lines[i].indentLevel;
      else
         lines[i].indentLevel = currentLevel;
   }

   util::Vector<Token> tokens;

   std::stack<int> blockStack;  //stores indentation levels
   blockStack.push(-1);
   bool hadCode = false;

   for(int i = 0; i < lines.getSize(); i++)
   {
      if(!lines[i].hasCode && !lines[i].hasComment)
      {
         //fix whitespace (assist syntax highlighter)
         tokens.pushBack(Token(lines[i].lineNumber, 0, Token::Value::whitespace(lines[i].indentLevel)));
      }
      else
      {
         for(int j = 0; j < lines[i].tokens.getSize() - 1; j++)
            tokens.pushBack(lines[i].tokens[j]);
      }

      if(lines[i].hasCode)
         hadCode = true;

      if(lines[i].endsWithBlockKeyword)
      {
         //start new block
         if((i+1) >= lines.getSize() || lines[i+1].indentLevel <= blockStack.top())
         {
            errorCollector->addError(lines[i].lineNumber, lines[i].length,
               $("Missing block after block keyword."));
         }
         else
         {
            blockStack.push(lines[i+1].indentLevel);
            tokens.pushBack(Token(lines[i].lineNumber, lines[i].length, Token::Value::startOfBlock(lines[i+1].indentLevel)));
            hadCode = false;
         }
      }

      if((i+1) < lines.getSize())
      {
         while(lines[i+1].indentLevel < blockStack.top())
         {
            //end of block
            if(hadCode)
               tokens.pushBack(Token(lines[i].lineNumber, lines[i].length, Token::Value::endOfStatement()));
            tokens.pushBack(Token(lines[i].lineNumber, lines[i].length, Token::Value::endOfBlock()));
            blockStack.pop();
            hadCode = true;
         }

         if(hadCode && lines[i+1].indentLevel == blockStack.top())
         {
            //end of statement
            tokens.pushBack(Token(lines[i].lineNumber, lines[i].length, Token::Value::endOfStatement()));
            hadCode = false;
         }
      }

      tokens.pushBack(lines[i].tokens[lines[i].tokens.getSize() - 1]);
   }

   Token endOfSourceToken = tokens[tokens.getSize() - 1];
   tokens.popBack();

   //end all blocks
   while(blockStack.top() >= 0)
   {
      if(hadCode)
      {
         tokens.pushBack(Token(endOfSourceToken.line, endOfSourceToken.column,
                               Token::Value::endOfStatement()));
      }
      tokens.pushBack(Token(endOfSourceToken.line, endOfSourceToken.column,
                            Token::Value::endOfBlock()));
      blockStack.pop();
      hadCode = true;
   }

   tokens.pushBack(endOfSourceToken);

   return tokens.getArray().toConst();
}

} //namespace syntax
} //namespace compiler
} //namespace evlan
