/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Tokenizer.h"
#include <framework/util/Vector.h>
#include <framework/util/Atom.h>
#include <evlan/utf8.h>

// begin namespace ===================================================
namespace evlan {
namespace compiler {
namespace syntax {

//Fuck the C++ standard for disallowing this, and shame on GCC for
//allowing it in previous versions then spontaneously breaking it.
//using ::framework::util;
namespace util {using namespace ::framework::util;}

// implementation definition =================================================
class Tokenizer
{
public:
   Tokenizer(String source,
             ErrorCollectorPtr errorCollector,
             KeywordSetPtr keywords,
             int startLine);
   ~Tokenizer();

   Token read();

private:
   // internal -------------------------------------------------------
   Token::Value readValue(int tokenLine, int tokenColumn);
   uint32 convertEscapeSequence(char c);

   String mSource;
   ErrorCollectorPtr mErrorCollector;
   KeywordSetPtr mKeywordSet;

   int mPos;
   int mLine;
   int mLineStart;

   enum CharacterType
   {
      TYPE_END_OF_SOURCE,
      TYPE_NEWLINE,
      TYPE_WHITESPACE,
      TYPE_IDENTIFIER_START,
      TYPE_ATOM_START,
      TYPE_LETTER,       //includes underscore
      TYPE_NUMBER,
      TYPE_STRING_DELIMITER,
      TYPE_CHARACTER_DELIMITER,
      TYPE_LINE_COMMENT,
      TYPE_SYMBOL,
      TYPE_SINGLE_SYMBOL,
      TYPE_INVALID
   };

   static const int TAB_STOP = 8;

   static const CharacterType CHARACTER_TYPE_MAP[128];

   inline char peek()
   {
      if(mPos >= mSource.getLength())
         return '\0';
      else
         return mSource[mPos];
   }
};

Tokenizer::Tokenizer(String source,
                     ErrorCollectorPtr errorCollector,
                     KeywordSetPtr keywords,
                     int startLine)
   : mSource(source),
     mErrorCollector(errorCollector),
     mKeywordSet(keywords),

     mPos(0),
     mLine(startLine),
     mLineStart(0)
{
   /* Some stupid editors like to write UTF-8 BOM's, even though they
    * completely defeat the purpose of UTF-8. */
   if((byte)mSource[0] == 0xEF && (byte)mSource[1] == 0xBB && (byte)mSource[2] == 0xBF)
   {
      mErrorCollector->addWarning(0, 0,
         $("UTF-8 byte-order mark (BOM) detected at start of source code.  BOM's in UTF-8 text "
           "completely defeat the purpose of UTF-8.  Please tell your editor not to output "
           "BOM's.  If your editor has no such option, please complain to the author and use "
           "a different editor."));
      mPos = 3;
   }
}

Tokenizer::~Tokenizer()
{
}

Token Tokenizer::read()
{
   int tokenLine = mLine;
   int tokenColumn = mPos - mLineStart;
   return Token(tokenLine, tokenColumn, readValue(tokenLine, tokenColumn));
}

Token::Value Tokenizer::readValue(int tokenLine, int tokenColumn)
{
   char c = peek();
   bool forceIdentifier = false;

   CharacterType type;
   if((int8)c < 0)
      type = TYPE_INVALID;
   else
      type = CHARACTER_TYPE_MAP[c];

   switch(type)
   {
   case TYPE_END_OF_SOURCE:
      {
         return Token::Value::endOfSource();
      }

   case TYPE_NEWLINE:
      {
         //parse newlines

         //Newline formats:
         //  "\n"   - Standard
         //  "\r\n" - Microsoft
         //  "\r"   - Old Macintosh (I think OSX fixes this)
         //Since '\r' and '\n' have no conceivable purpose in a text file other
         //than to represent newlines, we can parse all three newline formats
         //without needing to know which one the file uses before hand.  We can
         //even parse files that use a mixture of the three (to some extent).
         //
         //Even though it would theoretically be more portable, relying on the
         //standard library to translate newlines would be a bad idea.  In that
         //case, a file written on a Windows machine would not compile in Unix,
         //which would be much worse for portability.

         ++mPos;

         if(c == '\r' && peek() == '\n')
         {
            //stupid stupid stupid stupid
            ++mPos;
         }

         ++mLine;
         mLineStart = mPos;
         return Token::Value::endOfLine();
      }

   case TYPE_WHITESPACE:
      {
         //parse whitespace
         int startColumn = mPos - mLineStart;
         int column = startColumn;

         while(CHARACTER_TYPE_MAP[c] == TYPE_WHITESPACE)
         {
            if(c == '\t')
               column = (column / TAB_STOP + 1) * TAB_STOP;
            else
               ++column;

            ++mPos;
            c = peek();
         }
         mLineStart = mPos - column;

         return Token::Value::whitespace(column - startColumn);
      }

   case TYPE_IDENTIFIER_START:
      {
         //parse identifier (forced: can't be keyword)
         ++mPos;
         c = peek();
         if(CHARACTER_TYPE_MAP[c] != TYPE_LETTER)
         {
            mErrorCollector->addError(tokenLine, tokenColumn,
               $("Identifier expected."));
            return Token::Value::identifier(null);
         }
         forceIdentifier = true;
      }
      //no break here!  Continue to the next case.

   case TYPE_LETTER:
      {
         //parse identifier
         int startPos = mPos;

         do
         {
            ++mPos;
            c = peek();
         } while ((CHARACTER_TYPE_MAP[c] == TYPE_LETTER)
               || (CHARACTER_TYPE_MAP[c] == TYPE_NUMBER));

         String name = mSource.getInterval(startPos, mPos);

         if(!forceIdentifier)
         {
            util::Atom keyword = mKeywordSet->getKeywordId(name);
            if(keyword != null)
               return Token::Value::keyword(keyword);
         }
         return Token::Value::identifier(name);
      }

   case TYPE_ATOM_START:
      {
         //parse atom
         ++mPos;
         c = peek();
         if(CHARACTER_TYPE_MAP[c] != TYPE_LETTER)
         {
            mErrorCollector->addError(tokenLine, tokenColumn,
               $("Atom name expected."));
            return Token::Value::atom(null);
         }
         else
         {
            int startPos = mPos;

            do
            {
               ++mPos;
               c = peek();
            } while ((CHARACTER_TYPE_MAP[c] == TYPE_LETTER)
                  || (CHARACTER_TYPE_MAP[c] == TYPE_NUMBER));

            String name = mSource.getInterval(startPos, mPos);

            return Token::Value::atom(name);
         }
      }

   case TYPE_NUMBER:
      {
         //parse number

         //attempt to parse integer or real
         int startPos = mPos;

         do
         {
            ++mPos;
            c = peek();
         } while('0' <= c && c <= '9');

         if(c == 'x' && (mPos - startPos) == 1 && mSource[startPos] == '0')
         {
            //parse hex number (data literal)
            ++mPos;

            int startPos = mPos;
            uint64 data = 0;
            while(true)
            {
               c = peek();
               if('0' <= c && c <= '9')
                  data = (data << 4) | (c - '0');
               else if('A' <= c && c <= 'F')
                  data = (data << 4) | (c - 'A' + 10);
               else if('a' <= c && c <= 'f')
                  data = (data << 4) | (c - 'a' + 10);
               else
                  break;
               ++mPos;
            }

            if(mPos - startPos == 0)
            {
               mErrorCollector->addError(tokenLine, tokenColumn,
                  $("Data literal requires digits."));
               return Token::Value::data8(0);
            }
            else if(mPos - startPos == 2)
               return Token::Value::data8((uint8)data);
            else if(mPos - startPos == 4)
               return Token::Value::data16((uint16)data);
            else if(mPos - startPos == 8)
               return Token::Value::data32((uint32)data);
            else if(mPos - startPos == 16)
               return Token::Value::data64(data);
            else
            {
               mErrorCollector->addError(tokenLine, tokenColumn,
                  $("Data literal must be 8, 16, 32, or 64 bits."));
               return Token::Value::data32(0xFFFFFFFF);
            }
         }
         else if((mPos - startPos) > 1 && mSource[startPos] == '0')
         {
            //does anyone use octal anymore?  I mean, besides for Unix file
            // permissions?  For the moment, I'm making leading zeros illegal
            // so that if we do add octal in the future, it won't break
            // anything.
            mErrorCollector->addError(tokenLine, tokenColumn,
               $("Redundant zeros before integer declaration."));
         }

         ++mPos;
         char c2 = peek();
         --mPos;

         if(c == '.' && '0' <= c2 && c2 <= '9')
         {
            //parse real
            do
            {
               ++mPos;
               c = peek();
            } while('0' <= c && c <= '9');

            if(c == 'e')
            {
               //parse exponent
               ++mPos;
               c = peek();
               if(c != '+' && c != '-')
                  --mPos;

               do
               {
                  ++mPos;
                  c = peek();
               } while('0' <= c && c <= '9');
            }

            String digits = mSource.getInterval(startPos, mPos);

            if(CHARACTER_TYPE_MAP[c] == TYPE_LETTER)
            {
               mErrorCollector->addError(tokenLine, tokenColumn + digits.getLength(),
                  $("Missing space at end of numeric literal."));
            }

            //note that String::toFloat() returns a double.
            return Token::Value::real(digits.toFloat());
         }
         else
         {
            //parse integer
            String digits = mSource.getInterval(startPos, mPos);

            int64 result = 0;
            for(int i = 0; i < digits.getLength(); i++)
            {
               if(result > 922337203685477580ll)
               {
                  result = -1;
                  break;
               }
               result = result * 10 + (digits[i] - '0');
               if(result < 0)
               {
                  //overflow
                  result = -1;
                  break;
               }
            }

            if(c == 'e')
            {
               mErrorCollector->addError(tokenLine, tokenColumn + digits.getLength(),
                  $("Integers cannot have exponents."));
            }
            else if(CHARACTER_TYPE_MAP[c] == TYPE_LETTER)
            {
               mErrorCollector->addError(tokenLine, tokenColumn + digits.getLength(),
                  $("Missing space at end of numeric literal."));
            }

            if(result < 0)
            {
               mErrorCollector->addError(tokenLine, tokenColumn,
                  $("Integer literal larger than 64 bits."));
               return Token::Value::integer64(0x7FFFFFFFFFFFFFFFll);
            }
            else if(result < 0x0000000080000000ll)
               return Token::Value::integer((int32)result);
            else
               return Token::Value::integer64(result);
         }
      }

   case TYPE_STRING_DELIMITER:
      {
//         if(mPos + 2 < mSource.getLength()
//         && mSource[mPos+1] == '\"'
//         && mSource[mPos+2] == '\"')
//         {
//            //parse multi-line string
//            mPos += 3;
//            while(mPos < mSource.getLength() && mSource[mPos] != '\n')
//               ++mPos
//
//            int endOfLastLine = mPos;
//            int indent = -1;
//            util::Vector<uint32> data;
//            int blankLines = 0;
//
//            while(true)
//            {
//               //check indent
//               int lineStartPos = mPos;
//
//               int lineIndent = 0;
//               while(mPos < mSource.getLength()
//                  && (mSource[mPos] == ' ' || mSource[mPos] == '\t'))
//               {
//                  if(mSource[mPos] == ' ')
//                     ++indent;
//                  else
//                     indent += 8 - indent % 8;  //tab
//                  if(indent != -1 && lineIndent >= indent)
//                     break;
//                  ++mPos;
//               }
//
//               //blank lines need not be properly indented
//               if(mSource[mPos] == '\n')
//               {
//                  ++blankLines;
//                  continue;
//               }
//
//               if(indent == -1)
//                  indent = lineIndent;
//               else if(lineIndent < indent)
//               {
//                  //the end!
//                  mPos = endOfLastLine;
//                  break;
//               }
//
//               while(blankLines > 0)
//               {
//                  --blankLines;
//                  data.pushBack('\n');
//               }
//
//               for(int i = indent; i < lineIndent; i++)
//                  data.pushBack(' ');
//
//               //read text line
//               while(true)
//               {
//                  if(mPos == mSource.getLength())
//                  {
//                     endOfLastLine = mPos;
//                     data.pushBack('\n');
//                     break;
//                  }
//
//                  uint32 c = mSource[mPos++];
//                  if(c == '\n')
//                  {
//                     endOfLastLine = mPos-1;
//                     data.pushBack(c);
//                     break;
//                  }
//                  else if(c == '\\')
//                  {
//                     if(mPos == mSource.getLength())
//                     {
//                        endOfLastLine = mPos;
//                        break;
//                     }
//                     else if(mSource[mPos++] == '\n')
//                     {
//                        endOfLastLine = mPos-1;
//                        break;
//                     }
//                     c = convertEscapeSequence(mSource[mPos++]);
//                  }
//                  data.pushBack(c);
//               }
//            }
//         }
//         else
         {
            //parse string

            //find endpoint
            ++mPos;  //open quote
            int startPoint = mPos;
            int unescapedSize = 0;
            while(mPos < mSource.getLength()
               && mSource[mPos] != '\"'
               && mSource[mPos] != '\n')
            {
               if(mSource[mPos] == '\\'
               && mPos+1 < mSource.getLength()
               && mSource[mPos+1] != '\n')
                  mPos += 2;
               else
                  ++mPos;
               ++unescapedSize;
            }

            Array<char> unescaped(unescapedSize);
            for(int i = startPoint, j = 0; i < mPos; i++)
            {
               if(mSource[i] == '\\' && i+1 < mPos)
                  unescaped[j++] = convertEscapeSequence(mSource[++i]);
               else
                  unescaped[j++] = mSource[i];
            }

            if(mPos == mSource.getLength())
               mErrorCollector->addError(mLine, mPos - mLineStart, $("Unexpected end-of-file in string literal."));
            else if(mSource[mPos] == '\n')
               mErrorCollector->addError(mLine, mPos - mLineStart, $("Unexpected end-of-line in string literal."));
            else
               ++mPos;  //end quote

            if(unescapedSize == 0)
               return Token::Value::string8(null);

            utf8::DecodeInfo utf8Info = utf8::getDecodeInfo((byte*)&unescaped[0], unescapedSize);
            switch(utf8Info.bitCount)
            {
            default:
            case 7:
               return Token::Value::string8(unescaped.toConst());
            case 8:
               {
                  Array<char8> decoded(utf8Info.size);
                  utf8::decode(&decoded[0], (byte*)&unescaped[0], unescapedSize);
                  return Token::Value::string8(decoded.toConst());
               }
            case 16:
               {
                  Array<char16> decoded(utf8Info.size);
                  utf8::decode(&decoded[0], (byte*)&unescaped[0], unescapedSize);
                  return Token::Value::string16(decoded.toConst());
               }
            case 32:
               {
                  Array<char32> decoded(utf8Info.size);
                  utf8::decode(&decoded[0], (byte*)&unescaped[0], unescapedSize);
                  return Token::Value::string32(decoded.toConst());
               }
            }
         }
      }

   case TYPE_CHARACTER_DELIMITER:
      {
         //parse character
         char32 unicodeChar = 0;
         ++mPos;
         c = peek();

         if(c == '\'')
         {
            ++mPos;
            mErrorCollector->addError(tokenLine, tokenColumn,
               $("Empty character literal."));
            return Token::Value::character(0);
         }
         else if(c == '\n')
         {
            mErrorCollector->addError(tokenLine, tokenColumn,
               $("Unexpected end-of-line in character literal."));
            return Token::Value::character(0);
         }
         else if(c == '\0')
         {
            mErrorCollector->addError(tokenLine, tokenColumn,
               $("Unexpected end-of-file in character literal."));
            return Token::Value::character(0);
         }
         else if(c == '\\')
         {
            ++mPos;
            c = peek();
            unicodeChar = convertEscapeSequence(c);
            ++mPos;
         }
         else if(c & 0x80)
         {
            //read UTF-8 character
            int startPos = mPos;
            while(mPos < mSource.getLength() && (mSource[mPos] & 0x80))
               ++mPos;
            utf8::DecodeInfo info = utf8::getDecodeInfo((byte*)&mSource.getBuffer()[startPos], mPos - startPos);

            if(info.size == 1)
               utf8::decode(&unicodeChar, (byte*)&mSource.getBuffer()[startPos], mPos - startPos);
            else
            {
               mErrorCollector->addError(tokenLine, tokenColumn,
                  $("Character literal contains more than one character (and they're non-ASCII, BTW)."));
               unicodeChar = 0x0000FFFD;
            }
         }
         else
         {
            unicodeChar = c;
            ++mPos;
         }

         if(peek() == '\'')
            ++mPos;
         else
         {
            mErrorCollector->addError(mLine, mPos - mLineStart,
               $("Unterminated character literal"));
         }
         return Token::Value::character(unicodeChar);
      }

   case TYPE_LINE_COMMENT:
      {
         //parse comment
         int startPos = mPos;

         do
         {
            ++mPos;
            c = peek();
         } while(c != '\n' && c != '\r' && c != '\0');
         String text = mSource.getInterval(startPos+1, mPos);

         return Token::Value::comment(text);
      }

   case TYPE_SYMBOL:
      {
         //parse symbol
         int startPos = mPos;

         do
         {
            ++mPos;
            c = peek();
         } while (CHARACTER_TYPE_MAP[c] == TYPE_SYMBOL);

         String name = mSource.getInterval(startPos, mPos);

         util::Atom keyword = mKeywordSet->getKeywordId(name);
         if(keyword != null)
            return Token::Value::keyword(keyword);
         else
         {
            mErrorCollector->addError(tokenLine, tokenColumn,
               $("Invalid symbol."));
            return Token::Value::whitespace(0);
         }
      }

   case TYPE_SINGLE_SYMBOL:
      {
         //parse one-character symbol
         String name = String(c);
         ++mPos;

         util::Atom keyword = mKeywordSet->getKeywordId(name);
         if(keyword != null)
            return Token::Value::keyword(keyword);
         else
         {
            mErrorCollector->addError(tokenLine, tokenColumn,
               $("Invalid symbol."));
            return Token::Value::whitespace(0);
         }
      }

   default:
   case TYPE_INVALID:
      {
         mErrorCollector->addError(tokenLine, tokenColumn,
            $("Invalid character: ") + String((int)c));
         ++mPos;
         return Token::Value::whitespace(0);
      }
   }
}

uint32 Tokenizer::convertEscapeSequence(char c)
{
   switch(c)
   {
   case '0':
      return 0;
   case ' ':
      return ' ';   //useful after """
   case 'a':
      return '\a';
   case 'b':
      return '\b';
   case 'f':
      return '\f';
   case 'n':
      return '\n';
   case 'r':
      return '\r';
   case 't':
      return '\t';
   case 'v':
      return '\v';
   case '\"':
      return '\"';
   case '\'':
      return '\'';
   case '\\':
      return '\\';
   default:
      mErrorCollector->addError(mLine, mPos - mLineStart, $("Invalid escape sequence."));
      return '?';
   }
}

const Tokenizer::CharacterType Tokenizer::CHARACTER_TYPE_MAP[128] =
{
   TYPE_END_OF_SOURCE       , //NUL
   TYPE_INVALID             , //SOH
   TYPE_INVALID             , //STX
   TYPE_INVALID             , //ETX
   TYPE_INVALID             , //EOT
   TYPE_INVALID             , //ENQ
   TYPE_INVALID             , //ACK
   TYPE_INVALID             , //BEL
   TYPE_INVALID             , //BS
   TYPE_WHITESPACE          , //HT
   TYPE_NEWLINE             , //LF
   TYPE_INVALID             , //VT
   TYPE_INVALID             , //FF
   TYPE_NEWLINE             , //CR
   TYPE_INVALID             , //SO
   TYPE_INVALID             , //SI
   TYPE_INVALID             , //DLE
   TYPE_INVALID             , //DC1
   TYPE_INVALID             , //DC2
   TYPE_INVALID             , //DC3
   TYPE_INVALID             , //DC4
   TYPE_INVALID             , //NAK
   TYPE_INVALID             , //SYN
   TYPE_INVALID             , //ETB
   TYPE_INVALID             , //CAN
   TYPE_INVALID             , //EM
   TYPE_INVALID             , //SUB
   TYPE_INVALID             , //ESC
   TYPE_INVALID             , //FS
   TYPE_INVALID             , //GS
   TYPE_INVALID             , //RS
   TYPE_INVALID             , //US
   TYPE_WHITESPACE          , //SP
   TYPE_SYMBOL              , //!
   TYPE_STRING_DELIMITER    , //"
   TYPE_LINE_COMMENT        , //#
   TYPE_IDENTIFIER_START    , //$
   TYPE_SYMBOL              , //%
   TYPE_SYMBOL              , //&
   TYPE_CHARACTER_DELIMITER , //'
   TYPE_SINGLE_SYMBOL       , //(
   TYPE_SINGLE_SYMBOL       , //)
   TYPE_SYMBOL              , //*
   TYPE_SYMBOL              , //+
   TYPE_SINGLE_SYMBOL       , //,
   TYPE_SYMBOL              , //-
   TYPE_SYMBOL              , //.
   TYPE_SYMBOL              , ///
   TYPE_NUMBER              , //0
   TYPE_NUMBER              , //1
   TYPE_NUMBER              , //2
   TYPE_NUMBER              , //3
   TYPE_NUMBER              , //4
   TYPE_NUMBER              , //5
   TYPE_NUMBER              , //6
   TYPE_NUMBER              , //7
   TYPE_NUMBER              , //8
   TYPE_NUMBER              , //9
   TYPE_SYMBOL              , //:
   TYPE_SYMBOL              , //;
   TYPE_SYMBOL              , //<
   TYPE_SYMBOL              , //=
   TYPE_SYMBOL              , //>
   TYPE_INVALID             , //?
   TYPE_ATOM_START          , //@
   TYPE_LETTER              , //A
   TYPE_LETTER              , //B
   TYPE_LETTER              , //C
   TYPE_LETTER              , //D
   TYPE_LETTER              , //E
   TYPE_LETTER              , //F
   TYPE_LETTER              , //G
   TYPE_LETTER              , //H
   TYPE_LETTER              , //I
   TYPE_LETTER              , //J
   TYPE_LETTER              , //K
   TYPE_LETTER              , //L
   TYPE_LETTER              , //M
   TYPE_LETTER              , //N
   TYPE_LETTER              , //O
   TYPE_LETTER              , //P
   TYPE_LETTER              , //Q
   TYPE_LETTER              , //R
   TYPE_LETTER              , //S
   TYPE_LETTER              , //T
   TYPE_LETTER              , //U
   TYPE_LETTER              , //V
   TYPE_LETTER              , //W
   TYPE_LETTER              , //X
   TYPE_LETTER              , //Y
   TYPE_LETTER              , //Z
   TYPE_SINGLE_SYMBOL       , //[
   TYPE_SINGLE_SYMBOL       , //backslash
   TYPE_SINGLE_SYMBOL       , //]
   TYPE_SYMBOL              , //^
   TYPE_LETTER              , //_
   TYPE_INVALID             , //`
   TYPE_LETTER              , //a
   TYPE_LETTER              , //b
   TYPE_LETTER              , //c
   TYPE_LETTER              , //d
   TYPE_LETTER              , //e
   TYPE_LETTER              , //f
   TYPE_LETTER              , //g
   TYPE_LETTER              , //h
   TYPE_LETTER              , //i
   TYPE_LETTER              , //j
   TYPE_LETTER              , //k
   TYPE_LETTER              , //l
   TYPE_LETTER              , //m
   TYPE_LETTER              , //n
   TYPE_LETTER              , //o
   TYPE_LETTER              , //p
   TYPE_LETTER              , //q
   TYPE_LETTER              , //r
   TYPE_LETTER              , //s
   TYPE_LETTER              , //t
   TYPE_LETTER              , //u
   TYPE_LETTER              , //v
   TYPE_LETTER              , //w
   TYPE_LETTER              , //x
   TYPE_LETTER              , //y
   TYPE_LETTER              , //z
   TYPE_SINGLE_SYMBOL       , //{
   TYPE_SYMBOL              , //|
   TYPE_SINGLE_SYMBOL       , //}
   TYPE_SINGLE_SYMBOL       , //~
   TYPE_INVALID             , //DEL
};

// interface =========================================================
Array<const Token> tokenize(String source,
                            ErrorCollectorPtr errorCollector,
                            KeywordSetPtr keywords,
                            int startLine)
{
   Tokenizer tokenizer(source, errorCollector, keywords, startLine);
   util::Vector<Token> tokens;

   while(true)
   {
      Token token = tokenizer.read();
      tokens.pushBack(token);

      FW_MATCH_START(token.value)
      FW_MATCH_CASE_0(Token::Value::endOfSource)
         return tokens.getArray().toConst();
      FW_MATCH_DEFAULT
         //continue...
      FW_MATCH_END
   }
}

} //namespace syntax
} //namespace compiler
} //namespace evlan
