/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/12/03 (KV)
 *  + First Version
 */

// ===================================================================

#include "utf8.h"

namespace evlan {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::utf8
// TODO (KV): Review evlan::utf8

namespace utf8
{
   struct SequenceInfo
   {
      int size;
      int bitCount;   //7, 8, 16, 32
      char32 bits;

      inline SequenceInfo(int size, int bitCount, char32 bits)
         : size(size), bitCount(bitCount), bits(bits) {}
      inline SequenceInfo() {}
   };

   const SequenceInfo REPLACEMENT_CHARACTER(1, 16, 0x0000FFFD);

   inline SequenceInfo getSequenceInfo(byte b)
   {
      if(b < 0x80)
         return SequenceInfo(1, 7, b);
      else if(b < 0xC0)
         return REPLACEMENT_CHARACTER;  //extraneous tail byte
      else if(b < 0xC4)
         return SequenceInfo(2, 8, b & 0x1F);
      else if(b < 0xE0)
         return SequenceInfo(2, 16, b & 0x1F);
      else if(b < 0xF0)
         return SequenceInfo(3, 16, b & 0x0F);
      else if(b < 0xF8)
         return SequenceInfo(4, 32, b & 0x07);
      else if(b < 0xFC)
         return SequenceInfo(5, 32, b & 0x03);
      else if(b < 0xFE)
         return SequenceInfo(6, 32, b & 0x01);
      else //invalid: 1111111x does not start a UTF-8 sequence.
         return REPLACEMENT_CHARACTER;
   }

   inline bool isValid(char32 c)
   {
      return c < 0x0000D800
          || (0x0000E000 <= c && c < 0x0000FFFE)
          || (0x00010000 <= c && c < 0x80000000);
   }

   inline int getEncodedSize(char32 c)
   {
      if(c < 0x00000080)
         return 1;
      else if(c < 0x00000800)
         return 2;
      else if(c < 0x00010000)
         return 3;  /* Conveniently, invalid ranges 0xD800-0xDFFF and
                     * 0xFFFE-0xFFFF will be encoded as {0xEF, 0xBF,
                     * 0xBD}, which is also 3 bytes, so this return
                     * value is correct even for them. */
      else if(c < 0x00200000)
         return 4;
      else if(c < 0x04000000)
         return 5;
      else if(c < 0x80000000)
         return 6;
      else
         return 3;  //invalid: will be represented as {0xEF, 0xBF, 0xBD}
   }

   DecodeInfo getDecodeInfo(byte* source, int size)
   {
      DecodeInfo result = {0, 7};

      for(int i = 0; i < size;)
      {
         SequenceInfo sequence = getSequenceInfo(source[i++]);

         for(int j = 1; j < sequence.size; j++)
         {
            if(i < size && (source[i] & 0xC0) == 0x80)
               sequence.bits = (sequence.bits << 6) | (source[i++] & 0x3F);
            else
            {
               //sequence too short
               sequence = REPLACEMENT_CHARACTER;
               break;
            }
         }

         //invalidate overlong sequences
         if(getEncodedSize(sequence.bits) != sequence.size)
            sequence = REPLACEMENT_CHARACTER;

         ++result.size;
         if(sequence.bitCount > result.bitCount)
            result.bitCount = sequence.bitCount;
      }

      return result;
   }

   template<typename Character>
   int decodeTemplate(Character target[], byte source[], int size)
   {
      int pos = 0;

      for(int i = 0; i < size;)
      {
         SequenceInfo sequence = getSequenceInfo(source[i++]);

         for(int j = 1; j < sequence.size; j++)
         {
            if(i < size && (source[i] & 0xC0) == 0x80)
               sequence.bits = (sequence.bits << 6) | (source[i++] & 0x3F);
            else
            {
               //sequence too short
               sequence = REPLACEMENT_CHARACTER;
               break;
            }
         }

         /* Invalidate overlong sequences (note that if the sequence
          * is already REPLACEMENT_CHARACTER, this condition will
          * always be true, harmlessly replacing sequence with the
          * same value. */
         if(getEncodedSize(sequence.bits) != sequence.size)
            sequence = REPLACEMENT_CHARACTER;

         target[pos++] = sequence.bits;
      }

      return pos;
   }

   int decode(char8  target[], byte source[], int size)
      {return decodeTemplate((byte*)target, source, size);}
   int decode(char16 target[], byte source[], int size)
      {return decodeTemplate(target, source, size);}
   int decode(char32 target[], byte source[], int size)
      {return decodeTemplate(target, source, size);}



   template<typename Character>
   int getEncodedSizeTemplate(Character source[], int size)
   {
      int result = 0;
      for(int i = 0; i < size; i++)
         result += getEncodedSize(source[i]);
      return result;
   }

   int getEncodedSize(char8  source[], int size)
      {return getEncodedSizeTemplate((byte*)source, size);}
   int getEncodedSize(char16 source[], int size)
      {return getEncodedSizeTemplate(source, size);}
   int getEncodedSize(char32 source[], int size)
      {return getEncodedSizeTemplate(source, size);}

   inline void addReplacementCharacter(byte target[], int& pos)
   {
      target[pos++] = 0xEF;
      target[pos++] = 0xBF;
      target[pos++] = 0xBD;
   }

   template<typename Character>
   int encodeTemplate(byte target[], Character source[], int size)
   {
      int pos = 0;
      for(int i = 0; i < size; i++)
      {
         char32 c = source[i];
         if(c < 0x00000080)
            target[pos++] = c;
         else if(c < 0x00000800)
         {
            target[pos++] = 0xC0 |  (c >>  6);
            target[pos++] = 0x80 | ((c >>  0) & 0x3F);
         }
         else if(c < 0x0000D800)
         {
            target[pos++] = 0xE0 |  (c >> 12);
            target[pos++] = 0x80 | ((c >>  6) & 0x3F);
            target[pos++] = 0x80 | ((c >>  0) & 0x3F);
         }
         else if(c < 0x0000E000)
         {
            /* invalid: Range reserved for surrogates, which are only
             * used in UTF-16. */
            addReplacementCharacter(target, pos);
         }
         else if(c < 0x0000FFFE)
         {
            target[pos++] = 0xE0 |  (c >> 12);
            target[pos++] = 0x80 | ((c >>  6) & 0x3F);
            target[pos++] = 0x80 | ((c >>  0) & 0x3F);
         }
         else if(c < 0x00010000)
         {
            //invalid: 0xFFFE and 0xFFFF
            addReplacementCharacter(target, pos);
         }
         else if(c < 0x00200000)
         {
            target[pos++] = 0xF0 |  (c >> 18);
            target[pos++] = 0x80 | ((c >> 12) & 0x3F);
            target[pos++] = 0x80 | ((c >>  6) & 0x3F);
            target[pos++] = 0x80 | ((c >>  0) & 0x3F);
         }
         else if(c < 0x04000000)
         {
            target[pos++] = 0xF8 |  (c >> 24);
            target[pos++] = 0x80 | ((c >> 18) & 0x3F);
            target[pos++] = 0x80 | ((c >> 12) & 0x3F);
            target[pos++] = 0x80 | ((c >>  6) & 0x3F);
            target[pos++] = 0x80 | ((c >>  0) & 0x3F);
         }
         else if(c < 0x80000000)
         {
            target[pos++] = 0xFC | ((c >> 30) & 0x01);
            target[pos++] = 0x80 | ((c >> 24) & 0x3F);
            target[pos++] = 0x80 | ((c >> 18) & 0x3F);
            target[pos++] = 0x80 | ((c >> 12) & 0x3F);
            target[pos++] = 0x80 | ((c >>  6) & 0x3F);
            target[pos++] = 0x80 | ((c >>  0) & 0x3F);
         }
         else
         {
            //invalid: 0x80000000 and up
            addReplacementCharacter(target, pos);
         }
      }

      return pos;
   }

   int encode(byte target[], char8  source[], int size)
      {return encodeTemplate(target, (byte*)source, size);}
   int encode(byte target[], char16 source[], int size)
      {return encodeTemplate(target, source, size);}
   int encode(byte target[], char32 source[], int size)
      {return encodeTemplate(target, source, size);}
}

// ===================================================================

} //namespace evlan
