/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/12/03 (KV)
 *  + First Version
 */

// ===================================================================

#ifndef evlan_utf8_h__INCLUDED__
#define evlan_utf8_h__INCLUDED__

#include <framework/common/all.h>

namespace evlan {
using namespace ::framework::common;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Document evlan::utf8
// TODO (KV): Test evlan::utf8

namespace utf8
{
   struct DecodeInfo
   {
      int size;       //number of characters in output
      int bitCount;   /* 7, 8, 16, 32 -- maximum character size
                       * needed.  Note that 7-bit data does not need
                       * to be decoded, but 8-bit data does. */
   };

   /* getDecodeInfo does an initial pass over the data to determine
    * information you might want to know before allocating space for
    * the output, such as the size needed and whether or not 8-bit or
    * 16-bit characters will be sufficient.  If memory usage is a
    * concern, you may want to run this first.  However, note that it
    * essentially performs all the same steps as decode(), but
    * discards the results.  If CPU time is more important that memory
    * usage, you might consider simply allocating the output buffer
    * under the assumption that 32-bit characters are needed, but that
    * every byte will decode to one character, since this is
    * guaranteed to be sufficient. */
   DecodeInfo getDecodeInfo(byte* source, int size);
   /* Decode the UTF-8 data pointed to by "source", which is "size"
    * bytes long.  Place the decoded text in the buffer pointed to by
    * "target", which must be large enough to hold the result.
    * Returns the number of characters in the result.  Note that you
    * can either call getDecodeInfo() ahead of time to determine how
    * much space to allocate for "target", or you can allocate the
    * maximum possible size (one character per source byte).
    *
    * You can decode to 8-bit, 16-bit, or 32-bit characters.  Note
    * that if the source contains characters that don't fit in the
    * destination's character size, those characters will be
    * truncated.  You should avoid this.
    *
    * The target buffer is NOT terminated with a zero character.  If
    * you want one, add it yourself.
    *
    * Note on invalid data:  decode() will still work on invalid
    * data, but any illegal sequence in the input will be replaced
    * with the 16-bit character 0xFFFD, defined as the "replacement
    * character" by the Unicode standard.  I don't have the Unicode
    * documentation, but I think this is exactly what character
    * 0xFFFD was meant for.  At least, I know of several programs
    * which operate in this manner.  Note that overlong UTF-8
    * sequences are treated as invalid by this decoder and thus
    * replaced with 0xFFFD, as are sequences that would decode to the
    * ranges [0xD800,0xDFFF] or [0xFFFE,0xFFFF].
    *
    * There are security implications, however.  Namely, this means
    * that comparing two utf-8 byte sequences will not always yield
    * the same result as comparing their decoded values.  For example,
    * you might explicitly disallow access to a particular file, and
    * to insure that the file is not being accessed, you might compare
    * the filename given in a particular request to the filename of
    * the protected file.  If you did your comparison before decoding
    * to UTF-8, and if your file happened to include character 0xFFFD
    * in its name, then an attacker could simply use a malformed
    * sequence in the place of that character to address the file
    * while bypassing your security check.  There are several good
    * ways to avoid this, all of which should be followed for maximum
    * security:
    *  - Never allow the 0xFFFD character to be used in identifiers.
    *  - Perform any string comparisons after converting the string to
    *    its final format.
    *  - Use whitelist-based security instead of blacklist-based.  In
    *    a whitelisting system, only correctly formatted strings could
    *    pass the whitelist because malformed strings obviously
    *    wouldn't be on it in the first place. */
   int decode(char8  target[], byte source[], int size);
   int decode(char16 target[], byte source[], int size);
   int decode(char32 target[], byte source[], int size);

   /* Get the number of bytes which will be necessary to encode the
    * text pointed to by "source", which is "size" characters long. */
   int getEncodedSize(char8  source[], int size);
   int getEncodedSize(char16 source[], int size);
   int getEncodedSize(char32 source[], int size);

   /* Encode "source", which is "size" characters long, and store the
    * result in "target", which must be large enough to store the
    * data.  You can either call getEncodedSize() to determine how
    * much space to allocate, or you can assume that every character
    * in "source" will decode to the maximum possible number of bytes,
    * which is 2, 3, or 6 for 8-bit, 16-bit, and 32-bit, respectively.
    * Calling getEncodedSize() is recommended since it will almost
    * always save a lot of space.  In any case, encode() returns the
    * number of bytes written to "target".
    *
    * Any character in the illegal ranges [0xD800,0xDFFF], [0xFFFE,
    * 0xFFFF], or [0x80000000,0xFFFFFFFF] will be encoded as if it
    * were 0x0000FFFD (i.e., the result will be {0xEF, 0xBF, 0xBD}).
    * The same security implications as with decoding invalid
    * sequences apply. */
   int encode(byte target[], char8  source[], int size);
   int encode(byte target[], char16 source[], int size);
   int encode(byte target[], char32 source[], int size);
};

// ===================================================================

} //namespace evlan

#endif //included
