/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/02/17 (KV)
 *  + First Version
 */

// ===================================================================

#include "SegmentSet.h"

namespace evlan {
namespace vm {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::vm::SegmentSet
// TODO (KV): Review evlan::vm::SegmentSet

SegmentSet::Node* SegmentSet::mFirstFreeNode;
std::vector<SegmentSet::Node*> SegmentSet::mNodeBlocks;

SegmentSet::~SegmentSet()
{
   if(mRoot != null)
      freeNode(mRoot);
}

word* SegmentSet::lookup(int address)
{
//   Debug::log() << "Segment Set " << this << ":  lookup " << address;
   Node* node = mRoot;

   while(true)
   {
#ifdef DEBUG
      //should never try to look up an address not in the set
      if(node == null)
         FW_THROW(spnew FatalError($("Tried to look up address that wasn't in SegmentSet.")));
#endif

      if(address < node->start)
         node = node->left;
      else if(address >= node->end)
         node = node->right;
      else
         return address - node->start + node->newLocation;
   }
}

word* SegmentSet::lookupNoThrow(int address)
{
//   Debug::log() << "Segment Set " << this << ":  lookup " << address;
   Node* node = mRoot;

   while(true)
   {
      if(node == null)
         return null;

      if(address < node->start)
         node = node->left;
      else if(address >= node->end)
         node = node->right;
      else
         return address - node->start + node->newLocation;
   }
}

bool SegmentSet::insert(int start, int end)
{
//   Debug::log() << "Segment Set " << this << ":  insert " << start << " - " << end;

#ifdef DEBUG
   if(end < 0 || start > mRange)
      FW_THROW(spnew FatalError($("Tried to insert out-of-range segment into SegmentSet.")));
   else if(start < 0 || end > mRange)
      FW_THROW(spnew FatalError($("Tried to insert partially out-of-range segment into SegmentSet.")));
#endif

   Node* node = mRoot;
   Node* parent = null;
   Node* insertionPoint = null;
   int rangeStart = 0;
   int rangeEnd = mRange;
   bool tookLeftBranch = false;
   bool needCleanupLeft = false;
   bool needCleanupRight = false;

   while(true)
   {
      if(node == null)
      {
         node = allocateNode();

         node->parent = parent;
         if(parent == null)
            mRoot = node;
         else if(tookLeftBranch)
            parent->left = node;
         else
            parent->right = node;

         node->left = null;
         node->right = null;

         node->start = start;
         node->end = end;

         break;
      }
      else if(end < node->start)
      {
         int midpoint = (rangeStart + rangeEnd) / 2;

         if(end >= midpoint)
         {
            //new segment is closer to the midpoint
            /* Replace existing segment with inserted segment, then
             * continue as if we were inserting the existing
             * segment.  However, remember that we inserted here
             * because some cleanup will be necessary later. */
            if(insertionPoint == null)
            {
               insertionPoint = node;
               if(start < midpoint)
                  needCleanupLeft = true;
            }
            int swap = node->start;
            node->start = start;
            start = swap;
            swap = node->end;
            node->end = end;
            end = swap;

            parent = node;
            node = node->right;
            rangeStart = midpoint;
            tookLeftBranch = false;
         }
         else
         {
            parent = node;
            node = node->left;
            rangeEnd = midpoint;
            tookLeftBranch = true;
         }
      }
      else if(start > node->end)
      {
         int midpoint = (rangeStart + rangeEnd) / 2;

         if(start <= midpoint)
         {
            //new segment is closer to the midpoint
            /* Replace existing segment with inserted segment, then
             * continue as if we were inserting the existing
             * segment.  However, remember that we inserted here
             * because some cleanup will be necessary later. */
            if(insertionPoint == null)
            {
               insertionPoint = node;
               if(end > midpoint)
                  needCleanupRight = true;
            }
            int swap = node->start;
            node->start = start;
            start = swap;
            swap = node->end;
            node->end = end;
            end = swap;

            parent = node;
            node = node->left;
            rangeEnd = midpoint;
            tookLeftBranch = true;
         }
         else
         {
            parent = node;
            node = node->right;
            rangeStart = midpoint;
            tookLeftBranch = false;
         }
      }
      else
      {
         //segments touch
         if(start < node->start)
         {
            needCleanupLeft = true;
            node->start = start;
         }
         if(end > node->end)
         {
            needCleanupRight = true;
            node->end = end;
         }
         if(!needCleanupLeft && !needCleanupRight)
            return false;   //did not insert anything

         insertionPoint = node;
         break;
      }
   }

   if(needCleanupLeft && insertionPoint->left != null)
      cleanupLeft(insertionPoint, insertionPoint->left, null, null);

   if(needCleanupRight && insertionPoint->right != null)
      cleanupRight(insertionPoint, insertionPoint->right, null, null);

   return true;
}

void SegmentSet::reset(int newRange)
{
   if(mRoot != null)
      freeNode(mRoot);
   mRoot = null;
   mRange = newRange;
}

void SegmentSet::getSize(int* segmentCount, int* totalLength)
{
   int count = 0;
   int size = 0;
   if(mRoot != null)
   {
      Node* node = mRoot;

      while(node->left != null)
         node = node->left;

      while(node != null)
      {
         ++count;
         size += node->end - node->start;
         if(node->right == null)
         {
            Node* previous;
            do
            {
               previous = node;
               node = node->parent;
            } while(node != null && node->right == previous);
         }
         else
         {
            node = node->right;
            while(node->left != null)
               node = node->left;
         }
      }
   }

   if(segmentCount != null)
      *segmentCount = count;
   if(totalLength != null)
      *totalLength = size;
}

void SegmentSet::getSegments(Segment segments[])
{
//   Debug::log() << "Segment Set " << this << ":  getSegments:";
   if(mRoot == null)
      return;

   Node* node = mRoot;
   int count = 0;
#ifdef DEBUG
   int depth = 0;
#endif

   while(node->left != null)
   {
      node = node->left;
#ifdef DEBUG
      ++depth;
#endif
   }

   while(node != null)
   {
//      Debug::log() << "   " << node->start << " - " << node->end;
      segments[count].mNode = node;
#ifdef DEBUG
      segments[count].mDepth = depth;
#endif

      ++count;
      if(node->right == null)
      {
         Node* previous;
         do
         {
            previous = node;
            node = node->parent;
#ifdef DEBUG
            --depth;
#endif
         } while(node != null && node->right == previous);
      }
      else
      {
         node = node->right;
#ifdef DEBUG
         ++depth;
#endif
         while(node->left != null)
         {
            node = node->left;
#ifdef DEBUG
            ++depth;
#endif
         }
      }
   }
}

int SegmentSet::countSegmentsInRange(int start, int end)
{
   if(mRoot == null)
      return 0;

   Node* node = mRoot;

   //find segment in range
   while(true)
   {
      Node* next = null;

      if(node->start >= end)
         next = node->left;
      else if(node->end <= start)
         next = node->right;

      if(next == null)
         break;

      node = next;
   }

   //traverse back up until we get the actual first segment in the range
   while(node->end <= start)
   {
      node = node->parent;
      if(node == null)
         return 0;
   }

   //traverse linearly and count
   int count = 0;
   while(node != null && node->start < end)
   {
      ++count;

      if(node->right == null)
      {
         Node* previous;
         do
         {
            previous = node;
            node = node->parent;
         } while(node != null && node->right == previous);
      }
      else
      {
         node = node->right;
         while(node->left != null)
            node = node->left;
      }
   }

   return count;
}

void SegmentSet::getSegmentsInRange(int segments[], int start, int end)
{
   if(mRoot == null)
      return;

   Node* node = mRoot;

   //find segment in range
   while(true)
   {
      Node* next = null;

      if(node->start >= end)
         next = node->left;
      else if(node->end <= start)
         next = node->right;

      if(next == null)
         break;

      node = next;
   }

   //traverse back up until we get the actual first segment in the range
   while(node->end <= start)
   {
      node = node->parent;
      if(node == null)
         return;
   }

   //traverse linearly and count
   int count = 0;
   while(node != null && node->start < end)
   {
      segments[count++] = node->start;
      segments[count++] = node->end;

      if(node->right == null)
      {
         Node* previous;
         do
         {
            previous = node;
            node = node->parent;
         } while(node != null && node->right == previous);
      }
      else
      {
         node = node->right;
         while(node->left != null)
            node = node->left;
      }
   }
}

//int gNodeCount = 0;

SegmentSet::Node* SegmentSet::allocateNode()
{
   if(mFirstFreeNode == null)
   {
      mFirstFreeNode = new Node[NODE_BLOCK_SIZE];
      mNodeBlocks.push_back(mFirstFreeNode);
      for(int i = 1; i < NODE_BLOCK_SIZE; i++)
         mFirstFreeNode[i-1].parent = &mFirstFreeNode[i];
      mFirstFreeNode[NODE_BLOCK_SIZE-1].parent = null;
   }

//   Debug::log() << "node count +: " << ++gNodeCount;

   Node* result = mFirstFreeNode;
   mFirstFreeNode = mFirstFreeNode->parent;
   return result;
}

void SegmentSet::freeNode(Node* node)
{
   if(node->parent != null)
   {
      if(node->parent->left == node)
         node->parent->left = null;
      else if(node->parent->right == node)
         node->parent->right = null;
      node->parent = null;
   }
   while(node != null)
   {
      if(node->left != null)
      {
         Node* child = node->left;
         node->left = null;
         node = child;
      }
      else if(node->right != null)
      {
         Node* child = node->right;
         node->right = null;
         node = child;
      }
      else
      {
//         Debug::log() << "Removed: " << node->start << " - " << node->end;
//         Debug::log() << "node count -: " << --gNodeCount;
         Node* parent = node->parent;
         node->parent = mFirstFreeNode;
         mFirstFreeNode = node;
         node = parent;
      }
   }
}

struct SegmentSet::NodeList
{
   NodeList* next;
   Node* node;
   bool filled;
};

/* As mentioned previously, inserting a segment sometimes requires
 * removing some existing nodes if the segments overlap (and thus are
 * combined into one big segment).  If we insert a segment which
 * overlaps an existing segment, we instead extend the existing
 * segment to include the new segment's range.  However, there may be
 * descendent nodes which also overlap with the newly-inserted
 * segment, so we have to travel down the left and right branches of
 * the updated node to catch and remove these.  cleanupLeft and
 * cleanupRight are for traversing the left and right nodes,
 * respectively.
 *
 * At each node we come across down these branches, any time we
 * encounter a node containing a segment which overlaps the
 * newly-inserted segment, we must go back and extend that original
 * segment.  Then, we remove the node.
 *
 * Removing the node may be trickier than it at first seems.  When a
 * node contains a segment that overlaps with the inserted segment, we
 * automatically know that one whole child branch of that node can be
 * removed, since that branch is entirely between the node and the
 * inserted segment (and, since they overlap, any space between them
 * must overlap as well).  However, if we just replace the node with
 * its other branch, we would be decreasing the depth of that branch
 * by one, which changes the ranges for all the nodes in that branch,
 * which could cause violations of the range constraint.  Instead, we
 * must replace the node with the descendant which contains the
 * segment closest to it (but that doesn't overlap the inserted
 * segment).
 *
 * Complicated?  By the time you read all this, the whole thing has
 * been thoroughly tested, so don't worry.  :P */
SegmentSet::NodeList* SegmentSet::cleanupLeft(Node* insertionPoint,
                                              Node* subtree,
                                              NodeList* firstEmptyNode,
                                              NodeList* lastEmptyNode)
{
   while(true)
   {
      bool removeMe = false;

      if(subtree->end >= insertionPoint->start)
      {
         //segments intersect
         if(subtree->start < insertionPoint->start)
         {
            //extend insertion point to include this segment
            insertionPoint->start = subtree->start;
         }

         //right subtree is completely redundant
         if(subtree->right != null)
            freeNode(subtree->right);

         //this segment is completely subsumed by the inserted segment
         removeMe = true;
      }
      else if(subtree->right != null)
         subtree = subtree->right;
      else if(firstEmptyNode == null)
         return null;  //no more nodes intersect original
      else
      {
         //we are at the rightmost segment of the subtree
         //replace the first empty node with this node's segment
         firstEmptyNode->node->start = subtree->start;
         firstEmptyNode->node->end = subtree->end;
         firstEmptyNode->filled = true;
         firstEmptyNode = firstEmptyNode->next;
         if(firstEmptyNode == null)
            lastEmptyNode = null;

         //no longer need this node
         removeMe = true;
      }

      if(removeMe)
      {
         //note that, if we get here, subtree->right is known to be null

         if(subtree->left != null)
         {
            //add this node to the empty nodes list
            NodeList listNode;

            listNode.node = subtree;
            listNode.next = null;
            listNode.filled = false;

            //continue down left branch
            if(firstEmptyNode == null)
               firstEmptyNode = cleanupLeft(insertionPoint, subtree->left, &listNode, &listNode);
            else
            {
               lastEmptyNode->next = &listNode;
               firstEmptyNode = cleanupLeft(insertionPoint, subtree->left, firstEmptyNode, &listNode);
               lastEmptyNode->next = null;
            }

            if(listNode.filled)
               return firstEmptyNode;  //all done with everything
            if(firstEmptyNode == &listNode)
               firstEmptyNode = null;
            //if the node wasn't filled, it has no more children
         }

         //since this node has no children, just remove it
         Node* dead = subtree;
         subtree = subtree->parent;
         freeNode(dead);

         if(firstEmptyNode == null || subtree == lastEmptyNode->node)
            return firstEmptyNode;
      }
   }
}

SegmentSet::NodeList* SegmentSet::cleanupRight(Node* insertionPoint,
                                               Node* subtree,
                                               NodeList* firstEmptyNode,
                                               NodeList* lastEmptyNode)
{
   while(true)
   {
      bool removeMe = false;

      if(subtree->start <= insertionPoint->end)
      {
         //segments intersect
         if(subtree->end > insertionPoint->end)
         {
            //extend insertion point to include this segment
            insertionPoint->end = subtree->end;
         }

         //left subtree is completely redundant
         if(subtree->left != null)
            freeNode(subtree->left);

         //this segment is completely subsumed by the inserted segment
         removeMe = true;
      }
      else if(subtree->left != null)
         subtree = subtree->left;
      else if(firstEmptyNode == null)
         return null;  //no more nodes intersect original
      else
      {
         //we are at the leftmost segment of the subtree
         //replace the first empty node with this node's segment
         firstEmptyNode->node->start = subtree->start;
         firstEmptyNode->node->end = subtree->end;
         firstEmptyNode->filled = true;
         firstEmptyNode = firstEmptyNode->next;
         if(firstEmptyNode == null)
            lastEmptyNode = null;

         //no longer need this node
         removeMe = true;
      }

      if(removeMe)
      {
         //note that, if we get here, subtree->left is known to be null

         if(subtree->right != null)
         {
            //add this node to the empty nodes list
            NodeList listNode;

            listNode.node = subtree;
            listNode.next = null;
            listNode.filled = false;

            //continue down right branch
            if(firstEmptyNode == null)
               firstEmptyNode = cleanupRight(insertionPoint, subtree->right, &listNode, &listNode);
            else
            {
               lastEmptyNode->next = &listNode;
               firstEmptyNode = cleanupRight(insertionPoint, subtree->right, firstEmptyNode, &listNode);
               lastEmptyNode->next = null;
            }

            if(listNode.filled)
               return firstEmptyNode;  //all done with everything
            if(firstEmptyNode == &listNode)
               firstEmptyNode = null;
            //if the node wasn't filled, it has no more children
         }

         //since this node has no children, just remove it
         Node* dead = subtree;
         subtree = subtree->parent;
         freeNode(dead);

         if(firstEmptyNode == null || subtree == lastEmptyNode->node)
            return firstEmptyNode;
      }
   }
}

// ===================================================================

/* Regression test code!
 *
 * The following randomly fills a SegmentSet with lots and lots of
 * segments then checks if the results are correct.  It uses a flat
 * array of booleans to remember which segments have been marked.
 * When complete, it prints the complete tree possibly using
 * indentation to show nesting (indentation only available for debug
 * builds). */

/*
   Array<bool> memory(32768);
   vm::SegmentSet set(32768);
   for(int i = 0; i < 32768; i++)
      memory[i] = false;

   uint seed = time(NULL);
   srand(seed);
   int start = 0, end = 0;
   for(int i = 0; i < 4096; i++)
   {
#if 1
      //use this code to choose totally random segments
      start = rand() % 32768;
      end = start + (1 << (rand() % 7));
      if(end > 32768)
         end = 32768;
#else
      //use this code to insert segments roughly in ascending order
      start = end + (1 << (rand() % 7));
      if(start > 32767)
         start = 0;
      end = start + (1 << (rand() % 7));
      if(end > 32768)
         end = 32768;
#endif

      bool allTrue = true;
      for(int j = start; j < end; j++)
         allTrue = allTrue && memory[j];

      if(set.insert(start, end))
      {
         if(allTrue)
         {
            cout << "FAILED:  Incorrectly reported that an already-inserted segment was new.  Seed = " << seed << endl;
            return;
         }
      }
      else
      {
         if(!allTrue)
         {
            cout << "FAILED:  Incorrectly reported that a new segment was already inserted.  Seed = " << seed << endl;
            return;
         }
      }

      for(int j = start; j < end; j++)
         memory[j] = true;
   }

   int previous = -1;

   int size, length;
   set.getSize(&size, &length);
   vm::SegmentSet::Segment segments[size];
   set.getSegments(segments);

   for(int i = 0; i < size; i++)
   {
#ifdef DEBUG
      for(int j = 0; j < segments[i].getDepth(); j++)
         cout << "   ";
#endif
      cout << segments[i].getStart() << " - " << segments[i].getEnd() << '\n';

      if(segments[i].getStart() <= previous)
      {
         cout << "FAILED:  Out of order.  Seed = " << seed << endl;
         return;
      }
      previous = segments[i].getEnd();

      for(int j = segments[i].getStart(); j < segments[i].getEnd(); j++)
      {
         if(!memory[j])
         {
            cout << "FAILED:  Included extra segments.  Seed = " << seed << endl;
            return;
         }
         memory[j] = false;
      }
   }

   cout << "total segments: " << size << endl;
   cout << "total length: " << length << endl;

   for(int i = 0; i < 32768; i++)
   {
      if(memory[i])
      {
         cout << "FAILED:  Missed a segment.  Seed = " << seed << endl;
         return;
      }
   }
   cout << "SUCCESS" << endl;
*/

/* This code lets you insert segments into a SegmentSet one by one and
 * watch what happens. */

/*
   cout << "set range: ";
   int range;
   cin >> range;

   vm::SegmentSet set(range);

   while(true)
   {
      cout << '\n';

      int size, length;
      set.getSize(&size, &length);
      vm::SegmentSet::Segment segments[size];
      set.getSegments(segments);

      for(int i = 0; i < size; i++)
      {
#ifdef DEBUG
         for(int j = 0; j < segments[i].getDepth(); j++)
            cout << "   ";
#endif
         cout << segments[i].getStart() << " - " << segments[i].getEnd() << '\n';
      }

      cout << "total length: " << length
           << "\n\nadd segment: ";
      int start, end;
      cin >> start >> end;
      if(set.insert(start, end))
         cout << "added\n";
      else
         cout << "already in set\n";
   }
*/

// ===================================================================

} //namespace vm
} //namespace evlan
