/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/02/17 (KV)
 *  + First Version
 */

// ===================================================================

#ifndef evlan_vm_SegmentSet_h__INCLUDED__
#define evlan_vm_SegmentSet_h__INCLUDED__

#include <framework/common/all.h>
#include <vector>

namespace evlan {
namespace vm {
using namespace ::framework::common;

namespace memory {struct FileSegment;}

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Document evlan::vm::SegmentSet
// TODO (KV): Test evlan::vm::SegmentSet

class SegmentSet;
typedef SmartPointer<SegmentSet> SegmentSetPtr;

/* This structure holds a set of memory segments which have been
 * marked by the collector.  If adjacent or overlapping segments
 * are inserted, they are combined into one big segment.
 *
 * The structure is a binary tree, balanced by taking into account
 * the fact that the range of values within it is limited.  A
 * particular SegmentSet will apply only to one memory block.  So,
 * the left branch of the root node will apply to the first half of
 * that block, and the right branch to the second half.  If a
 * segment happens to cross the half-way point, it is stored in the
 * root node itself.
 *
 * Note that, if a particular subtree only contains one segment,
 * then that segment is stored directly in the root of that
 * subtree, even if it doesn't cross the half-way point.  In
 * particular, the following constraint is always maintained:  No
 * descendant of a node may contain a segment with an endpoint that
 * is between the node's segment and the midpoint of its range.
 *
 * Another important constraint is that no two nodes will ever
 * contain segments which touch or overlap.
 *
 * Why not a red-black tree?  Well, how do you remove an entire
 * branch from a red-black tree in reasonable time?  If a segment
 * is inserted which completely subsumes a branch, such an
 * operation is necessary.  Besides, when the range of values is
 * known, I think this is actually more efficient by quite a bit.
 *
 * This class does not conform to my usual standards of insulation.
 * However, it is only intended for use within the virtual machine,
 * where speed is rather important. */
class SegmentSet
{
private:
   struct Node
   {
      Node* parent;
      Node* left;
      Node* right;

      //Endpoints of segment in this node.
      int start;
      int end;

      /* Location to which this segment has been moved.  It is assumed
       * that this value will not be set or read until after all
       * segments have been inserted into the set.  So, if you set it
       * to something, then insert a new segment, the value might be
       * lost. */
      word* newLocation;
   };
   struct NodeList;

public:
   /* If you don't want to set a range right away, you can call
    * reset() later to set it. */
   inline SegmentSet(): mRoot(null), mRange(0) {}
   inline SegmentSet(int range): mRoot(null), mRange(range) {}
   ~SegmentSet();

   word* lookup(int address);
   word* lookupNoThrow(int address);
   bool insert(int start, int end);

   //remove all segments from the set and change the range
   void reset(int newRange);

   class Segment
   {
   public:
      inline int getStart() {return mNode->start;}
      inline int getEnd  () {return mNode->end  ;}

      inline void  setNewLocation(word* location) {mNode->newLocation = location;}
      inline word* getNewLocation()               {return mNode->newLocation;}

#ifdef DEBUG
      inline int getDepth() {return mDepth;}
#endif

      memory::FileSegment* mFileSegment;  //hack

   private:
      Node* mNode;
#ifdef DEBUG
      int mDepth;
#endif

      friend class SegmentSet;
   };

   /* The memory manager typically makes several passes through a
    * SegmentSet at a time, without inserting or removing segments in
    * between.  Rather than use an iterator which has to climb up and
    * down the tree, it is more efficient to create a big old array on
    * the stack and write the whole list there. */
   void getSize(int* segmentCount, int* totalLength);
   void getSegments(Segment segments[]);

   /* These are used only in persistence.cpp as of this writing.  They
    * find all segments contained in the given range.  The segments
    * array (likely allocated on the stack) is a set of start/end
    * pairs marking said segments. */
   int countSegmentsInRange(int start, int end);
   void getSegmentsInRange(int segments[], int start, int end);

private:
   Node* mRoot;
   //endpoints of segments in the set can be anything from zero to mRange
   int mRange;

   static Node* mFirstFreeNode;
   static std::vector<Node*> mNodeBlocks;
   static const int NODE_BLOCK_SIZE = 4096;

   static SegmentSet::Node* allocateNode();
   static void freeNode(Node* node);
   static NodeList* cleanupLeft(Node* insertionPoint,
                                Node* subtree,
                                NodeList* firstEmptyNode,
                                NodeList* lastEmptyNode);
   static NodeList* cleanupRight(Node* insertionPoint,
                                 Node* subtree,
                                 NodeList* firstEmptyNode,
                                 NodeList* lastEmptyNode);

   //disable copying
   SegmentSet(const SegmentSet&);
   SegmentSet& operator=(const SegmentSet&);
};

// ===================================================================

} //namespace vm
} //namespace evlan

#endif //included
