/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/02/10 (KV)
 *  + First Version
 */

// ===================================================================

#ifndef evlan_vm_core_h__INCLUDED__
#define evlan_vm_core_h__INCLUDED__

#include <framework/common/all.h>
#ifdef INCLUDE_WRITER_CODE
#include <framework/TextWriter.h>
#endif //INCLUDE_WRITER_CODE

namespace framework {namespace util {}}

namespace evlan {
namespace vm {
using namespace ::framework::common;

//Fuck the C++ standard for disallowing this, and shame on GCC for
//allowing it in previous versions then spontaneously breaking it.
//using ::framework::util;
namespace util {using namespace ::framework::util;}

#ifdef INCLUDE_WRITER_CODE
using ::framework::TextWriter;
using ::framework::TextWriterPtr;
#endif //INCLUDE_WRITER_CODE

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Document evlan::vm::Core
// TODO (KV): Test evlan::vm::Core

namespace core
{
   enum Opcode
   {
      NOOP,

      RETURN,
      GOTO,                //parameter = relative address (in instructions)
      TEST_CONDITION,      /* parameter = relative branch address, if false.
                            * If true, next instruction is skipped.
                            * On type error, next instruction is executed. */

      PUSH_EMPTY,          //parameter = number of empty values to push to stack
      POP,                 //parameter = number of values to pop

      LOAD,                //load from stack
      LOAD_ENVIRONMENT,    //load from function environment
      LOAD_CONSTANT,       //load from constant table
      STORE,               //store to stack (does not pop)
      STORE_POP,           //store to stack and pop

      GET_MEMBER,          //of an object
      APPLY,               //parameter = arity, for typechecking
      APPLY_RETURN,        //tail recursion

      MAKE_FUNCTION,       //parameter = function tag index within module
      MAKE_OBJECT,         //parameter = object tag index within module
      MAKE_ARRAY,          //parameter = size

      ALLOCATE_MULTI,      //parameter = total size
      MAKE_FUNCTION_MULTI,
      MAKE_OBJECT_MULTI,
      MAKE_ARRAY_MULTI,
      STORE_MULTI,         //parameter = index
      STORE_MULTI_POP,     //parameter = index

      IMMEDIATE_INTEGER,   //parameter = 24-bit unsinged integer value
      IMMEDIATE_NEGATIVE,  //parameter = 24-bit unsigned integer value (negated)
      IMMEDIATE_BOOLEAN,   //parameter = 1 for true, 0 for false
      IMMEDIATE_CHARACTER, //parameter = 24-bit character code
      IMMEDIATE_ATOM,      //parameter = atom identifier

      IMMEDIATE_INTEGER32,   //value follows in place of next instruction
      IMMEDIATE_INTEGER64,   //value follows in place of next two instructions
      IMMEDIATE_REAL32,      //value follows in place of next instruction
      IMMEDIATE_REAL64,      //value follows in place of next two instructions
      IMMEDIATE_CHARACTER32, //value follows in place of next instruction
      IMMEDIATE_ATOM32,      //value follows in place of next instruction

      IMMEDIATE_DATA1,
      IMMEDIATE_DATA2,
      IMMEDIATE_DATA4,
      IMMEDIATE_DATA8,
      IMMEDIATE_DATA16,
      IMMEDIATE_DATA32,      //value follows in place of next instruction
      IMMEDIATE_DATA64,      //value follows in place of next two instructions

      //builtin functions
      NOT,
      NEGATE,
      SIZE,

      AND,
      OR,
      ADD,
      SUBTRACT,
      MULTIPLY,
      DIVIDE,
      MODULUS,
      POWER,
      EQUALS,
      IS_LESS_THAN,
      IS_GREATER_THAN,
      GET_ELEMENT
   };

   struct Instruction
   {
      unsigned opcode    : 8;
      unsigned parameter : 24;

      inline Instruction() {}
      inline explicit Instruction(int opcode)
         : opcode(opcode) {}
      inline Instruction(int opcode, int parameter)
         : opcode(opcode), parameter(parameter) {}
   };
   static const int PARAMETER_BITS = 24;

   struct SourceLocation
   {
        signed line   : 24;
      unsigned column : 8;

      inline SourceLocation() {}
      inline SourceLocation(int line, int column)
         : line(line), column(column) {}
   };

   struct EnvironmentIndex
   {
      unsigned level : 8;
      unsigned item  : 20;
   };

   struct Module;
   struct Process;

   struct FunctionTag
   {
      Module* module;
      int arity;
      int environmentSize;

      int* parameterNames;
      int line;
      int column;

      int documentationLength;
      char* documentation;

      uint64 profileTimer;
      uint64 profileCounter;

      int instructionCount;
      SourceLocation* sourceLocations;
      Instruction instructions[0];
   };

   struct NativeFunctionTag
   {
      String identifier;
      int arity;
      Array<const String> parameterNames;

      uint64 profileTimer;
      uint64 profileCounter;

      void (*nativeApply)();
   };

   struct ObjectTag
   {
      struct Member
      {
         int nameId;
         int offset;

         //Member lists must be sorted.  This makes the task easy.
         inline bool operator<(const Member& other) const
            {return nameId < other.nameId;}
      };

      Module* module;

      int documentationLength;
      char* documentation;

      int memberCount;
      Member members[0];
   };

   union Value
   {
      /* On 32-bit systems, sizeof(Value) should be 12.
       * On 64-bit systems, sizeof(Value) should be 24.
       * You might consider it wasteful that some types of values,
       * such as integers, only require half of the space allocated
       * for them.  However, if values had variable size, then they
       * would have to be allocated on the heap, which would be far
       * more wasteful. */

      enum Type
      {
         //no pointers in these
         EMPTY,
         BUILTIN_OBJECT,
         NATIVE_CONTAINER,  //an object containing all native functions

         //these have pointers which need not be scanned
         NATIVE_FUNCTION,

         //the rest have pointers to the heap (and possibly tags as well)
         RUNTIME_ERROR,
         FUNCTION,
         OBJECT,
         ARRAY,
         PACKED_ARRAY,
         LARGE_DATA,     //128 bits or more
         TASK,
         CONTEXT,

         MODULE,    //just a direct module link
         PROCESS
      };

      enum BuiltinType
      {
         BOOLEAN,
         SCALAR,
         CHARACTER,
         ATOM,
         DATA1,
         DATA2,
         DATA4,
         DATA8,
         DATA16,
         DATA32,
         DATA64
      };

      enum PackType
      {
         PACKED_BOOLEAN,
         PACKED_SCALAR,
         PACKED_INTEGER,
         PACKED_CHARACTER,
         PACKED_ATOM,
         PACKED_DATA
      };

      enum TaskType
      {
         STOP,
         CALL,
         RETURN,
         TRY,
         THROW,
         FINALLY
      };

      enum ErrorCode
      {
         NOT_A_FUNCTION = 1,
         ARITY_MISMATCH,
         NOT_AN_OBJECT,
         NO_SUCH_MEMBER,
         CONDITION_NOT_BOOLEAN,
         TYPE_ERROR,
         OUT_OF_BOUNDS
      };

      int8 type;

      struct
      {
         int8 type;
         int8 builtinType;
         union
         {
            bool boolean;
            uint32 data32;  //also used for characters and atoms

#ifdef PLATFORM_64BIT
            double scalar;
            uint64 data64;
#else
            double scalar __attribute__((packed, aligned(4)));
            uint64 data64 __attribute__((packed, aligned(4)));
#endif
         } value;
      } builtin;

      struct
      {
         int8 type;
         NativeFunctionTag* tag;
         int32 id;   //identifier number, used when loading persistent state
      } nativeFunction;

      struct
      {
         int8 type;
         uint16 errorCode;
         Instruction* location;
      } runtimeError;

      struct
      {
         int8 type;
         FunctionTag* tag;
         Value* environment;
      } function;
      struct
      {
         int8 type;
         ObjectTag* tag;
         Value* members;
      } object;
      struct
      {
         int8 type;
         int32 size;
         Value* elements;
      } array;
      struct
      {
         int8 type;
         int8 packType;
         int8 elementSize;  //lg(number of bits per element)
         /* The elements pointer must be aligned on a word boundary.
          * If the actual data items are smaller than words, then
          * offset indicates how many elements to skip at the start.
          * This allows getInterval() to be implemented without any
          * need to copy. */
         int8 offset;
         int32 size;        //in elements
         word* elements;
      } packedArray;
      struct
      {
         int8 type;
         int8 size;      //actually, lg(size)
         word* value;
      } largeData;
      struct
      {
         int8 type;
         int8 taskType;
         Value* parameters;
         int32 parameterCount;
      } task;
      struct
      {
         int8 type;
         int8 contextType;   //CALL, TRY, FINALLY, THROW, or STOP (for the root)
         bool throwOnReturn; //used only for THROW
         Value* parameter;   //return function or catch block
         Value* parent;
      } context;

      struct
      {
         int8 type;
         Module* module;
      } module;
      struct
      {
         int8 type;
         Process* process;
      } process;


      static inline Value makeEmpty()
      {
         Value result;
         result.type = EMPTY;
         return result;
      }
      static inline Value makeError(ErrorCode code, Instruction* location = null)
      {
         Value result;
         result.type = RUNTIME_ERROR;
         result.runtimeError.errorCode = code;
         result.runtimeError.location = location;
         return result;
      }
      static inline Value makeBoolean(bool b)
      {
         Value result;
         result.type = BUILTIN_OBJECT;
         result.builtin.builtinType = BOOLEAN;
         result.builtin.value.boolean = b;
         return result;
      }
      static inline Value makeScalar(double x)
      {
         Value result;
         result.type = BUILTIN_OBJECT;
         result.builtin.builtinType = SCALAR;
         result.builtin.value.scalar = x;
         return result;
      }
//      static inline Value makeScalar(int x)
//      {
//         Value result;
//         result.type = BUILTIN_OBJECT;
//         result.builtin.builtinType = SCALAR;
//         result.builtin.value.scalar = x;
//         return result;
//      }
      static inline Value makeCharacter(uint32 c)
      {
         Value result;
         result.type = BUILTIN_OBJECT;
         result.builtin.builtinType = CHARACTER;
         result.builtin.value.data32 = c;
         return result;
      }
      static inline Value makeAtom(int id)
      {
         Value result;
         result.type = BUILTIN_OBJECT;
         result.builtin.builtinType = ATOM;
         result.builtin.value.data32 = id;
         return result;
      }
      static inline Value makeBuiltin(uint32 d, BuiltinType builtinType)
      {
         Value result;
         result.type = BUILTIN_OBJECT;
         result.builtin.builtinType = builtinType;
         result.builtin.value.data32 = d;
         return result;
      }
      static inline Value makeBuiltin(uint64 d, BuiltinType builtinType)
      {
         Value result;
         result.type = BUILTIN_OBJECT;
         result.builtin.builtinType = builtinType;
         result.builtin.value.data64 = d;
         return result;
      }
      static inline Value makeFunction(FunctionTag* tag, Value* environment)
      {
         Value result;
         result.type = FUNCTION;
         result.function.tag = tag;
         result.function.environment = environment;
         return result;
      }
      static inline Value makeTask(TaskType taskType, int parameterCount = 0, Value* parameters = null)
      {
         Value result;
         result.type = TASK;
         result.task.taskType = taskType;
         result.task.parameters = parameters;
         result.task.parameterCount = parameterCount;
         return result;
      }
      static inline Value makeContext(TaskType contextType, Value* parameter = null, Value* parent = null)
      {
         Value result;
         result.type = CONTEXT;
         result.context.contextType = contextType;
         result.context.throwOnReturn = false;
         result.context.parameter = parameter;
         result.context.parent = parent;
         return result;
      }
      static inline Value makeObject(ObjectTag* tag, Value* members)
      {
         Value result;
         result.type = OBJECT;
         result.object.tag = tag;
         result.object.members = members;
         return result;
      }
      static inline Value makeArray(int size, Value* elements)
      {
         Value result;
         result.type = ARRAY;
         result.array.size = size;
         result.array.elements = elements;
         return result;
      }
      static inline Value makePackedArray(int packType, int elementSize,
                                          int size, word* elements)
      {
         Value result;
         result.type = PACKED_ARRAY;
         result.packedArray.packType = packType;
         result.packedArray.elementSize = elementSize;
         result.packedArray.offset = 0;
         result.packedArray.size = size;
         result.packedArray.elements = elements;
         return result;
      }
      static inline Value makeByteArray(int size, byte* elements)
      {
         Value result;
         result.type = PACKED_ARRAY;
         result.packedArray.packType = PACKED_DATA;
         result.packedArray.elementSize = 3;
         result.packedArray.offset = 0;
         result.packedArray.size = size;
         result.packedArray.elements = (word*)elements;
         return result;
      }
      static inline Value makeModule(Module* module)
      {
         Value result;
         result.type = MODULE;
         result.module.module = module;
         return result;
      }
      static inline Value makeProcess(Process* process)
      {
         Value result;
         result.type = PROCESS;
         result.process.process = process;
         return result;
      }

      inline bool hasPointer()    {return type >= RUNTIME_ERROR;}

      inline bool isEmpty()       {return type == EMPTY;}
      inline bool isBuiltin()     {return type == BUILTIN_OBJECT;}
      inline bool isError()       {return type == RUNTIME_ERROR;}
      inline bool isFunction()    {return type == FUNCTION || type == NATIVE_FUNCTION;}
      inline bool isEvlanFunction()  {return type == FUNCTION;}
      inline bool isNativeFunction() {return type == NATIVE_FUNCTION;}
      inline bool isTask()        {return type == TASK;}
      inline bool isContext()     {return type == CONTEXT;}
      inline bool isObject()      {return type == OBJECT;}
      inline bool isArray()       {return type == ARRAY;}
      inline bool isPackedArray() {return type == PACKED_ARRAY;}
      inline bool isLargeData()   {return type == LARGE_DATA;}
      inline bool isProcess()     {return type == PROCESS;}
      inline bool isModule()      {return type == MODULE;}

      inline bool isBoolean()     {return type == BUILTIN_OBJECT && builtin.builtinType == BOOLEAN;}
      inline bool isScalar()      {return type == BUILTIN_OBJECT && builtin.builtinType == SCALAR;}
      inline bool isCharacter()   {return type == BUILTIN_OBJECT && builtin.builtinType == CHARACTER;}
      inline bool isAtom()        {return type == BUILTIN_OBJECT && builtin.builtinType == ATOM;}
      inline bool isData()        {return type == BUILTIN_OBJECT && builtin.builtinType >= DATA1;}

      inline bool isInteger()
      {
         return type == BUILTIN_OBJECT
             && builtin.builtinType == SCALAR
             && builtin.value.scalar == (int)builtin.value.scalar;
      }

      inline bool isString()
      {
         return (type == PACKED_ARRAY && packedArray.packType == core::Value::PACKED_CHARACTER)
             || (type == ARRAY && array.size == 0);
      }

      inline bool hasArity(int arity)
      {
         return (type == FUNCTION && function.tag->arity == arity)
             || (type == NATIVE_FUNCTION && nativeFunction.tag->arity == arity);
      }

      inline bool isByteArray()
      {
         return (type == ARRAY && array.size == 0)
             || (type == PACKED_ARRAY && packedArray.packType == PACKED_DATA
                 && packedArray.elementSize == 3);
      }
   };

   static const int WORDS_PER_VALUE = sizeof(Value) / sizeof(word);
#ifdef PLATFORM_64BIT
   static const int LOG2_WORD_SIZE = 6;
#else
   static const int LOG2_WORD_SIZE = 5;
#endif

   struct Module
   {
      int functionTagCount;
      int objectTagCount;
      int constantCount;
      int extraDataSize;

      Value root;
      int id;
      ObjectPtr nativeInfo;

      FunctionTag** functionTags;
      ObjectTag** objectTags;
      Value* constants;
      word* extraData;
   };

   struct Process
   {
      Value state;
      ObjectPtr nativeInfo; //used for processes with native implementations
      bool keepAlive; /* True if the process should not be garbage-
                       * collected.  This should only be used for
                       * processes that are waiting for system events. */

      /* This is an identifier ID for the process's type.  This is
       * used when persistent states are restarted to figure out what
       * constructor to call to restart the process.  Typically this
       * will be something like "system", "timer", "file", or
       * "listener". */
      int type;
   };

   // globals ========================================================
#ifdef PLATFORM_WIN32
   /* dlltool --export-all seems to skip variables, so we have to
    * manually specify them as exports. */
#ifdef COMPILING_EVLANLIB
#define EVLAN_VM_CORE_EXPORT __attribute__((dllexport))
#else
#define EVLAN_VM_CORE_EXPORT __attribute__((dllimport))
#endif
#else
#define EVLAN_VM_CORE_EXPORT
#endif

   /* The stack grows downwards, so to allocate space on it, you need
    * to subtract the number of Values you want.  The end of the stack
    * space is protected by a guard page.
    *
    * These pointers are maintained by the memory manager. */
   extern Value* gStackTop EVLAN_VM_CORE_EXPORT;
   extern Value* gStackEnd EVLAN_VM_CORE_EXPORT;

   //task queue, maintained by messages.cpp
   struct Task
   {
      Value task;
      Value context;
   };
   extern Task* gTaskQueue EVLAN_VM_CORE_EXPORT;
   extern int gTaskQueueSize EVLAN_VM_CORE_EXPORT;

#ifdef INCLUDE_WRITER_CODE
   void writeValue(TextWriterPtr out, const Value& value);
   void writeModule(TextWriterPtr out, const Module* module);
   void writeFunction(TextWriterPtr out, const FunctionTag* function);
#endif //INCLUDE_WRITER_CODE
};

// ===================================================================

} //namespace vm
} //namespace evlan

#endif //included
