/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/03/12 (KV)
 *  + First Version
 */

// ===================================================================

#include "loader.h"
#include "memory.h"
#include "simulator.h"
#include "arrays.h"
#include "translator.h"
#include "native.h"
#include "shortcuts.h"
#include <framework/util/Vector.h>
#include <framework/messages/Message.h>
#include <evlan/compiler/CompilerImpl.h>
#include <map>
#include <algorithm>
#include <stdio.h>

namespace evlan {
namespace vm {
using namespace shortcuts;
using framework::messages::MessageReceiver;
using framework::messages::MessagePtr;
using framework::messages::makeMessage;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::vm::loader
// TODO (KV): Review evlan::vm::loader

namespace loader
{
   FW_DEFINE_EXCEPTION(LoadFailed, $("Failed to load module."));

   implementation FinalizerInfo;
   struct InternalModuleInfo: public ModuleInfo
   {
      FinalizerInfo* finalizer;
   };

   typedef SmartPointer<InternalModuleInfo> InternalModuleInfoPtr;

   int gNextId = 0;

   std::map<int, core::Module*> gModulesById;
   std::map<String, InternalModuleInfoPtr> gModulesByName;
   std::map<core::Module*, InternalModuleInfoPtr> gModulesByPointer;

   String gModuleDirectory;
   compiler::ErrorCollectorPtr gErrorCollector;

   int makeId()
   {
      while(gModulesById.find(gNextId++) != gModulesById.end()) {}
      return gNextId-1;
   }

   implementation FinalizerInfo: implements MessageReceiver
   {
      FinalizerInfo(core::Process* process, framework::SystemPtr system);
      ~FinalizerInfo();

      void receive(MessagePtr message);

      core::Process* mProcess;
      MessagePtr mMessage;
   };

   FinalizerInfo::FinalizerInfo(core::Process* process, framework::SystemPtr system)
      : mProcess(process)
   {
      mMessage = makeMessage(this, 0, null);
      if(mProcess->state.isArray()
      && mProcess->state.array.size == 2
      && mProcess->state.array.elements[0].isInteger())
      {
         mProcess->keepAlive = true;
         std::map<int, core::Module*>::iterator iter =
            gModulesById.find((int)mProcess->state.array.elements[0].builtin.value.scalar);

         std::map<core::Module*, InternalModuleInfoPtr>::iterator iter2;

         if(iter != gModulesById.end())
            iter2 = gModulesByPointer.find(iter->second);

         if(iter == gModulesById.end()
         || iter2 == gModulesByPointer.end())
            mMessage->send();  //module already gone
         else
            iter2->second->finalizer = this;
      }
   }

   FinalizerInfo::~FinalizerInfo()
   {
   }

   void FinalizerInfo::receive(MessagePtr message)
   {
      simPush(Value::makeContext(Value::STOP));
      simPush(mProcess->state.array.elements[1]);
      simPush(mProcess->state.array.elements[0]);
      simApply(1);
      simQueue();

      simPush(Value::makeEmpty());
      memory::modifyProcessState(mProcess);
      mProcess->keepAlive = false;
   }

   EVLAN_NATIVE_PROCESS(finalizer, FinalizerInfo)

   implementation ModuleRemover: implements Object
   {
      ModuleRemover(core::Module* module)
         : mId(module->id), mModule(module) {}
      ~ModuleRemover();

      int mId;
      core::Module* mModule;
   };

   ModuleRemover::~ModuleRemover()
   {
      std::map<core::Module*, InternalModuleInfoPtr>::iterator iter =
         gModulesByPointer.find(mModule);

      if(iter != gModulesByPointer.end())
      {
         if(iter->second->finalizer != null)
            iter->second->finalizer->mMessage->send();
         if(iter->second->fromFile)
            gModulesByName.erase(iter->second->name);
         gModulesByPointer.erase(iter);
      }

      gModulesById.erase(mId);
   }

   // interface ======================================================
   void init(String moduleDirectory,
             compiler::ErrorCollectorPtr errorCollector)
   {
      gNextId = 0;
      gModuleDirectory = moduleDirectory;
      gErrorCollector = errorCollector;
   }

   void shutdown()
   {
      gModuleDirectory = null;
      gErrorCollector = null;

      gModulesById.clear();
      gModulesByName.clear();
      gModulesByPointer.clear();
   }

   ModuleInfoPtr loadModule(String name, String relativeTo)
   {
      if(relativeTo != null && relativeTo[0] == '/'
      && name       != null && name      [0] != '/')
      {
         int pos = relativeTo.findLastOf('/');
         relativeTo = relativeTo.getInterval(0, pos);

         while(name.getInterval(0, 3) == $("../"))
         {
            pos = relativeTo.findLastOf('/');
            if(pos == -1)
            {
               relativeTo = null;
               break;
            }
            else
            {
               relativeTo = relativeTo.getInterval(0, pos);
               name = name.getInterval(3);
            }
         }

         if(name != null)
            name = relativeTo + $("/") + name;
      }

      std::map<String, InternalModuleInfoPtr>::iterator iter = gModulesByName.find(name);
      if(iter != gModulesByName.end())
      {
         if(iter->second == null)
         {
            String oldModuleName = gErrorCollector->setModule(name);
            gErrorCollector->addError($("Recursive import."));
            gErrorCollector->setModule(oldModuleName);
            FW_THROW(spnew LoadFailed());
         }

         simPush(iter->second->module->root);
         return iter->second;
      }

      FILE* file = null;

      if(name != null && name[0] == '/')
      {
         String filename = gModuleDirectory + $("/") + name.getInterval(1) + $(".evlan");

         FW_MAKE_CSTRING(cfilename, filename);
         file = fopen(cfilename, "rb");
      }

      if(file == null)
      {
         String oldModuleName = gErrorCollector->setModule(name);
         gModulesByName.erase(name);
         gErrorCollector->addError($("Module implementation not found."));
         gErrorCollector->setModule(oldModuleName);
         FW_THROW(spnew LoadFailed());
      }
      else
      {
         gModulesByName[name] = null;
         try
         {
            fseek(file, 0, SEEK_END);
            size_t size = ftell(file);
            Array<char> data(size);
            fseek(file, 0, SEEK_SET);
            if(size > 0)
               fread(&data[0], 1, size, file);
            fclose(file);

            ModuleInfoPtr result = loadModuleFromSource(name, String(data.toConst()));
            result->fromFile = true;
            gModulesByName[name] = result.reinterpretCast<InternalModuleInfoPtr>();
            return result;
         }
         catch(...)
         {
            gModulesByName.erase(name);
            throw;
         }
      }
   }

   ModuleInfoPtr loadModuleFromSource(String name, String source)
   {
      String oldModuleName = gErrorCollector->setModule(name);

      try
      {
         compiler::CompilerPtr compiler = compiler::makeCompilerImpl(gErrorCollector);
         compiler::BytecodePtr bytecode = compiler->compile(source);

         gErrorCollector->setModule(oldModuleName);
         return loadModuleFromBytecode(name, bytecode);
      }
      FW_CATCH_START
         FW_CATCH(compiler::Compiler::HadErrors, e)
         {
            gErrorCollector->setModule(oldModuleName);
            FW_THROW(spnew LoadFailed(e));
         }
         FW_CATCH(Exception, e)
         {
            gErrorCollector->setModule(oldModuleName);
            throw;
         }
      FW_CATCH_END
   }

   ModuleInfoPtr loadModuleFromBytecode(String name, compiler::BytecodePtr bytecode)
   {
      String oldModuleName = gErrorCollector->setModule(name);

      InternalModuleInfoPtr info = spnew InternalModuleInfo;
      info->finalizer = null;
      info->fromFile = false;
      info->name = name;
      info->module = null;

      core::Module* module;
      try
      {
         module = translator::translateBytecode(bytecode);
      }
      FW_CATCH_START
         FW_CATCH(translator::InvalidBytecode, e)
         {
            gErrorCollector->addError($("Bytecode invalid:  ") + e->getDescription());
            gErrorCollector->setModule(oldModuleName);
            FW_THROW(spnew LoadFailed(e));
         }
         FW_CATCH(Exception, e)
         {
            gErrorCollector->addError($("Failed to translate bytecode:  ") + e->getDescription());
            gErrorCollector->setModule(oldModuleName);
            FW_THROW(spnew LoadFailed(e));
         }
      FW_CATCH_END

      module->id = makeId();
      gModulesById[module->id] = module;

      simPush(Value::makeModule(module));
      info->module = module;
      gModulesByPointer[module] = info;

      for(int i = 0; i < bytecode->importTable.getSize(); i++)
      {
         if(bytecode->importTable[i] == $("(native)"))
         {
            --core::gStackTop;
            core::gStackTop->type = core::Value::NATIVE_CONTAINER;
         }
         else
         {
            try
            {
               loadModule(bytecode->importTable[i], name);
            }
            FW_CATCH_START
               FW_CATCH(LoadFailed, e)
               {
                  simPop(i + 1);
                  gErrorCollector->addError($("Failed to load imports."));
                  gErrorCollector->setModule(oldModuleName);
                  FW_THROW(spnew LoadFailed(e));
               }
               FW_CATCH(Exception, e)
               {
                  simPop(i + 1);
                  gErrorCollector->setModule(oldModuleName);
                  throw;
               }
            FW_CATCH_END
         }
      }

      Value* importArray = memory::allocateValues(bytecode->importTable.getSize());
      for(int i = bytecode->importTable.getSize()-1; i >= 0; i--)
      {
         importArray[i] = *gStackTop;
         simPop();
      }

      simPush(Value::makeFunction(
         module->functionTags[module->functionTagCount - 1], null));
      simPush(Value::makeArray(bytecode->importTable.getSize(), importArray));

      simulator::apply(1);

      module->nativeInfo = spnew ModuleRemover(module);
      memory::setModuleRoot(module);
      gStackTop[1] = gStackTop[0];
      simPop();

      gErrorCollector->setModule(oldModuleName);
      return info;
   }

   void initModule(String name, bool fromFile, core::Module* module)
   {
      if(gModulesById.find(module->id) != gModulesById.end())
         FW_THROW(spnew FatalError($("Module id number clash.")));
      if(gNextId <= module->id)
         gNextId = module->id + 1;

      InternalModuleInfoPtr info = spnew InternalModuleInfo;
      info->finalizer = null;
      info->fromFile = fromFile;
      info->name = name;
      info->module = module;

      gModulesByPointer[module] = info;
      if(fromFile)
         gModulesByName[name] = info;

      gModulesById[module->id] = module;
      module->nativeInfo = spnew ModuleRemover(module);
   }

   ModuleInfoPtr getModuleInfo(core::Module* module)
   {
      std::map<core::Module*, InternalModuleInfoPtr>::iterator iter =
         gModulesByPointer.find(module);

      if(iter == gModulesByPointer.end())
         return null;
      else
         return iter->second;
   }

   struct ModuleIdOrdering
   {
      inline bool operator()(ModuleInfoPtr a, ModuleInfoPtr b)
         {return a->module->id < b->module->id;}
   };

   Array<const ModuleInfoPtr> listResidentModules()
   {
      Array<ModuleInfoPtr> result(gModulesByPointer.size());

      int i = 0;
      for(std::map<core::Module*, InternalModuleInfoPtr>::iterator iter = gModulesByPointer.begin();
          iter != gModulesByPointer.end(); ++iter, ++i)
         result[i] = iter->second;

      if(result != null)
         std::sort(&result[0], &result[0] + result.getSize(), ModuleIdOrdering());

      return result.toConst();
   }



   // evlan hooks ====================================================
   implementation TemporaryErrorCollector: implements compiler::ErrorCollector
   {
      TemporaryErrorCollector();
      ~TemporaryErrorCollector();

      String setModule(String name);
      int getErrorCount();

      void addError  (int line, int column, String description);
      void addError  (String description);
      void addWarning(int line, int column, String description);
      void addWarning(String description);

      struct Error
      {
         String module;
         int line;
         int column;
         String description;

         inline Error(String module, int line, int column, String description)
            : module(module), line(line), column(column), description(description) {}
         inline Error() {}
      };

      util::Vector<Error> mErrors;
      util::Vector<Error> mWarnings;
      String mModule;
   };

   TemporaryErrorCollector::TemporaryErrorCollector() {}
   TemporaryErrorCollector::~TemporaryErrorCollector() {}

   String TemporaryErrorCollector::setModule(String name)
   {
      String old = mModule;
      mModule = name;
      return old;
   }
   int TemporaryErrorCollector::getErrorCount() {return mErrors.getSize();}

   void TemporaryErrorCollector::addError(int line, int column, String description)
      {mErrors.pushBack(Error(mModule, line, column, description));}
   void TemporaryErrorCollector::addError(String description)
      {mErrors.pushBack(Error(mModule, -1, -1, description));}
   void TemporaryErrorCollector::addWarning(int line, int column, String description)
      {mWarnings.pushBack(Error(mModule, line, column, description));}
   void TemporaryErrorCollector::addWarning(String description)
      {mWarnings.pushBack(Error(mModule, -1, -1, description));}

   void buildErrors(Array<TemporaryErrorCollector::Error> errors)
   {
      for(int i = 0; i < errors.getSize(); i++)
      {
         arrays::makeString(errors[i].description);

         Value* elements = memory::allocateValues(3);
         elements[0] = Value::makeScalar(errors[i].line);
         elements[1] = Value::makeScalar(errors[i].column);
         elements[2] = gStackTop[0];

         *gStackTop = Value::makeArray(3, elements);
      }

      Value* elements = memory::allocateValues(errors.getSize());
      for(int i = errors.getSize()-1; i >= 0; i--)
      {
         elements[i] = *gStackTop;
         simPop();
      }
      simPush(Value::makeArray(errors.getSize(), elements));
   }

   //one temporary error collector used by all hook functions.
   SmartPointer<TemporaryErrorCollector> gTemporaryErrorCollector =
      spnew TemporaryErrorCollector;

   // glue code ------------------------------------------------------
   EVLAN_NATIVE_FUNCTION_2(loaderCompile, name, code)
   {
      if(!name.isString() || !code.isByteArray())
         return simTypeError(name, code);

      compiler::BytecodePtr bytecode;
      core::Module* module = null;

      try
      {
         compiler::CompilerPtr compiler =
            compiler::makeCompilerImpl(gTemporaryErrorCollector,
                                       compiler::Compiler::BASIC,
                                       compiler::Compiler::OPTIMIZE);

         String codeText;
         if(code.isPackedArray())
         {
            codeText = String(
               ((char*)code.packedArray.elements + code.packedArray.offset),
               code.packedArray.size);
         }
         bytecode = compiler->compile(codeText);
         module = translator::translateBytecode(bytecode);
         module->id = makeId();
         initModule(arrays::getString(name), false, module);
      }
      FW_CATCH_START
         FW_CATCH(compiler::Compiler::HadErrors, e)
         {
            if(gTemporaryErrorCollector->getErrorCount() == 0)
               gTemporaryErrorCollector->addError(e->getDescription());
         }
         FW_CATCH(translator::InvalidBytecode, e)
         {
            gTemporaryErrorCollector->addError(e->getLine(), e->getColumn(), e->getError());
         }
         FW_CATCH(Exception, e)
         {
            gTemporaryErrorCollector->addError($("Caught unexpected exception: ") + e->getDescription());
         }
      FW_CATCH_END

      if(module != null)
      {
         simPush(Value::makeModule(module));

         //make import table
         for(int i = 0; i < bytecode->importTable.getSize(); i++)
            arrays::makeString(bytecode->importTable[i]);
         core::Value* elements = memory::allocateValues(bytecode->importTable.getSize());
         for(int i = bytecode->importTable.getSize()-1; i >= 0; i--)
         {
            elements[i] = *gStackTop;
            simPop();
         }
         simPush(Value::makeArray(bytecode->importTable.getSize(), elements));
      }

      buildErrors(gTemporaryErrorCollector->mErrors.getArray());
      buildErrors(gTemporaryErrorCollector->mWarnings.getArray());

      gTemporaryErrorCollector->mErrors.clear();
      gTemporaryErrorCollector->mWarnings.clear();

      if(module == null)
      {
         Value* elements = memory::allocateValues(2);
         elements[0] = gStackTop[1];
         elements[1] = gStackTop[0];
         simPop(2);
         return Value::makeArray(2, elements);
      }
      else
      {
         Value* elements = memory::allocateValues(5);
         elements[0] = gStackTop[1];
         elements[1] = gStackTop[0];
         elements[2] = gStackTop[3];
         elements[3] = gStackTop[2];
         elements[4] = Value::makeScalar(module->id);
         simPop(4);
         return Value::makeArray(5, elements);
      }
   }

   EVLAN_NATIVE_FUNCTION_0(loaderListResidentModules)
   {
      int i = 0;
      for(std::map<core::Module*, InternalModuleInfoPtr>::iterator iter = gModulesByPointer.begin();
          iter != gModulesByPointer.end(); ++iter, ++i)
      {
         InternalModuleInfoPtr info = iter->second;

         simPush(Value::makeScalar(info->module->id));
         arrays::makeString(info->name);
         simPush(Value::makeBoolean(info->fromFile));
         simPush(Value::makeModule(info->module));
      }

      int count = i;

      Value* elements = memory::allocateValues(count * 5);

      for(i = count-1; i >= 0; i--)
      {
         elements[i*4+0] = gStackTop[3];
         elements[i*4+1] = gStackTop[2];
         elements[i*4+2] = gStackTop[1];
         elements[i*4+3] = gStackTop[0];
         simPop(4);
      }

      for(i = 0; i < count; i++)
         elements[count*4+i] = Value::makeArray(4, elements + i*4);
      return simReturn(Value::makeArray(count, elements + count*4));
   }

   EVLAN_NATIVE_FUNCTION_1(loaderGetModuleByName, name)
   {
      if(!name.isString())
         return simTypeError(name);

      std::map<String, InternalModuleInfoPtr>::iterator iter =
         gModulesByName.find(arrays::getString(name));

      if(iter == gModulesByName.end())
         return Value::makeArray(0, null);
      else
      {
         simPush(iter->second->module->root);
         Value* elements = memory::allocateValues(1);
         elements[0] = *gStackTop;
         return Value::makeArray(1, elements);
      }
   }

   EVLAN_NATIVE_FUNCTION_1(loaderGetModuleById, id)
   {
      if(!id.isInteger())
         return simTypeError(id);

      std::map<int, core::Module*>::iterator iter =
         gModulesById.find((int)id.builtin.value.scalar);

      if(iter == gModulesById.end())
         return Value::makeArray(0, null);
      else
      {
         simPush(iter->second->root);
         Value* elements = memory::allocateValues(1);
         elements[0] = *gStackTop;
         return Value::makeArray(1, elements);
      }
   }

   EVLAN_NATIVE_FUNCTION_1(loaderGetModuleConstructor, module)
   {
      if(!module.isModule())
         return simTypeError(module);

      return Value::makeFunction(
         module.module.module->functionTags[
            module.module.module->functionTagCount-1],
         null);
   }

   EVLAN_NATIVE_FUNCTION_2(loaderActivate, module, root)
   {
      if(!module.isModule())
         return simTypeError(module);

      simPush(root);
      memory::setModuleRoot(module.module.module);

      return *gStackTop;
   }

   EVLAN_NATIVE_FUNCTION_2(loaderSetFinalizer, module, finalizer)
   {
      if(!module.isModule())
         return simTypeError(module);

      Value* elements = memory::allocateValues(2);
      elements[0] = Value::makeScalar(module.module.module->id);
      elements[1] = finalizer;
      simPush(Value::makeArray(2, elements));
      memory::makeProcess($("finalizer"));
      gStackTop->process.process->nativeInfo =
         spnew FinalizerInfo(gStackTop->process.process, null);

      return simReturn();
   }
}

// ===================================================================

} //namespace vm
} //namespace evlan
