/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/02/11 (KV)
 *  + First Version
 */

// ===================================================================

#include "memory.h"
#include "operations.h"
#include "memoryInternal.h"
#include <algorithm>
#include <queue>
#include <stack>

#ifdef PLATFORM_WIN32
#define _WIN32_WINNT 0x0500
#include <windows.h>
#undef ERROR
#elif defined(PLATFORM_POSIX)
#include <sys/mman.h>
#include <signal.h>
#include <errno.h>

#ifndef MAP_ANON
#define MAP_ANON MAP_ANONYMOUS
#endif
#endif

namespace evlan {
namespace vm {
using ::framework::common::byte;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::vm::memory
// TODO (KV): Review evlan::vm::memory

namespace memory
{
   /* If there is a God, He uses generational garbage collection.
    * (I imagine His code is a bit cleaner, though.) */

   bool gSuperSweeping = false;

   namespace sequence
   {
      /* Calling next() repeatedly produces a sequence of numbers.
       * Here is how it starts (line breaks added to make the pattern
       * more obvious) (note that each digit is one element; I think
       * the pattern is easier to see with no delimiters between
       * them):
       *   0
       *   10
       *   1210
       *   12131210
       *   1213121412131210
       *   12131214121312151213121412131210
       *   ...
       * The pattern is used by the collector to decide which
       * generations to sweep.  Each time a sweep is performed, two
       * generations are swept:  Generation zero and one other.  The
       * other is chosen by calling next().  If next() returns zero,
       * then no other generation is swept, but otherwise the number
       * returned by next() indicates the other generation.
       *
       * And what is so special about this sequence, you ask?  Well,
       * you'll notice that each generation is swept exactly twice
       * between any two sweeps of the next older generation, and
       * these sweeps are evenly spaced.  Additionally, between the
       * time that a generation is swept and the next time that the
       * next younger generation is swept, all generations younger
       * than that will be swept.  The collection algorithm takes
       * advantage of these facts. */
      int next()
      {
         uint64 i = counter++;
         i ^= counter;
         int result = 1;
         while(i != 1)
         {
            i >>= 1;
            ++result;
         }

         if(result > max)
         {
            ++max;
            counter = 0;
            return 0;
         }
         else
            return result;
      };
   }

   const Parameters DEFAULT_PARAMETERS =
   {
#if 1
      128*1024, //stack size
      16384,    //heap block size
      16384,    //preferred block size
      8192      //large allocation threshold
#else
      //useful for debugging
      128*1024, //stack size
      2048,     //heap block size
      2048,     //preferred block size
      1024      //large allocation threshold
#endif
   };

   // ================================================================

   /* The garbage collector often allocates in BIG chunks.  It is
    * likely most efficient to allocate these as raw memory pages
    * direct from the OS rather than use the C/C++ standard library
    * memory management.  Maybe I'll get around to implementing it as
    * such eventually.  It would also be fun to stick guard pages at
    * one or both ends, and trigger collection based on page faults. */
   void* allocatePages(int size, bool guardPage)
   {
//      void* result = operator new(size + 2048 * sizeof(uint32));
//      uint32* start = (uint32*)result;
//      result = start + 1024;
//      uint32* end = (uint32*)((byte*)result + size);
//      for(int i = 0; i < 1024; i++)
//      {
//         start[i] = 0xdeadbeef;
//         end[i] = 0xdeadbeef;
//      }

#ifdef PLATFORM_WIN32
      /* VirtualAlloc will reserve a full 64k of the process's address
       * space.  For smaller blocks (such as simple modules) we'd
       * rather not do that. */
      if(size < 32768)
         return calloc(size, 1);      //calloc zeros memory

      void* result = VirtualAlloc(null, size, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
      if(result == null)
         FW_THROW(spnew Win32Error($("VirtualAlloc")));

      if(guardPage)
      {
         DWORD iDontNeedThisButWindowsInsists;
         if(!VirtualProtect(result, 1, PAGE_NOACCESS,
                            &iDontNeedThisButWindowsInsists))
            FW_THROW(spnew Win32Error($("VirtualProtect")));
      }

      return result;

#elif defined(PLATFORM_POSIX)
      if(size < 32768)
         return calloc(size, 1);      //calloc zeros memory

      void* result = mmap(null, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON, -1, 0);
      if(result == null)
         FW_THROW(spnew ErrnoError($("mmap")));

#if 0
      if(guardPage)
      {
         /* TODO:  Would be nice to catch SIGSEGV and detect if it was
          *        an Evlan stack overflow, to match the equivalent
          *        win32 functionality.  However, it's gonna crash
          *        either way, so it would only be a convenience. */
         //FIXME:  Doesn't work on Linux.  What else is new?
         if(mprotect(result, 1, PROT_NONE) < 0)
            FW_THROW(spnew ErrnoError($("mprotect")));
      }
#endif

      return result;

#else
#warning "Using stupid memory allocation."
      return calloc(size, 1);        //calloc zeros memory
#endif
   }
   void freePages(void* memory, int size)
   {
#ifdef PLATFORM_WIN32
      if(size < 32768)
         free(memory);
      else if(!VirtualFree(memory, 0, MEM_RELEASE))
         FW_THROW(spnew Win32Error($("VirtualFree")));

#elif defined(PLATFORM_POSIX)
      if(size < 32768)
         free(memory);
      else if(munmap(memory, size) < 0)
         FW_THROW(spnew ErrnoError($("munmap")));

#else
      free(memory);
#endif

//      uint32* start = (uint32*)memory - 1024;
//      uint32* end = (uint32*)((byte*)memory + size);
//
//      for(int i = 0; i < 1024; i++)
//      {
//         if(start[i] != 0xdeadbeef)
//            Debug::log(Debug::ERROR) << "Memory messed up before block: " << i;
//         if(end[i] != 0xdeadbeef)
//            Debug::log(Debug::ERROR) << "Memory messed up after block: " << i;
//      }
//
//      operator delete(start);
   }

   Block* allocateBlock(int segmentCount)
   {
#ifdef DEBUG
      if(segmentCount == 0)
         FW_THROW(spnew FatalError($("Allocated a block with zero segments.")));
#endif
      Block* result = (Block*)operator new(sizeof(Block) + sizeof(FileSegment) * (segmentCount + 1));
      result = new(result) Block;
      result->fileSegmentCount = segmentCount;
      return result;
   }

   Block* allocateHeap()
   {
      int size = gParameters.heapSize;

      Block* block = allocateBlock(1);
      block->next = null;
      block->generation = 0;
      block->size = size;
      block->data = (word*)allocatePages(size * sizeof(word));
      block->dataSegments.reset(size);
      block->valueSegments.reset(size);

      block->headerLocation = 0;
      block->fileSegments[0].start = 0;
      block->fileSegments[1].start = size;
      block->fileSegments[0].fileLocation = 0;

      gHeapEnd = block->data + size;
      gHeapTop = block->data;

      gBlockMap[BlockKey(block->data, block->data + size)] = block;

      return block;
   }

   void deleteBlocks(Block* blocks)
   {
#ifdef DEBUG
      int i = 0;
#endif
      while(blocks != null)
      {
         Block* next = blocks->next;
         gBlockMap.erase(BlockKey(blocks->data, blocks->data + blocks->size));
#ifdef TRACK_DEADBLOCKS
         gDeadBlockMap[BlockKey(blocks->data, blocks->data + blocks->size)] = blocks;
#else
         freePages(blocks->data, blocks->size * sizeof(word));
         delete blocks;
#endif
         blocks = next;

#ifdef DEBUG
         if(++i > 65536)
         {
            Debug::log() << "Infinite loop in deleteBlocks()?";
         }
#endif
      }
   }

#ifdef PLATFORM_WIN32
   LONG CALLBACK handleException(EXCEPTION_POINTERS* exceptionInfo)
   {
      if(exceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION)
      {
         void* address = (void*)exceptionInfo->ExceptionRecord->ExceptionInformation[1];
         if(address >= core::gStackEnd
         && address < core::gStackEnd + gParameters.stackSize)
            MessageBox(GetConsoleWindow(), "Congratulations; you overflowed the simulator stack.\nI am going to crash now.", "Evlan", MB_OK | MB_ICONERROR | MB_SETFOREGROUND | MB_TOPMOST);
      }
      else if(exceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_STACK_OVERFLOW)
      {
         void** ebp = (void**)exceptionInfo->ContextRecord->Ebp;

         {
            int i = 1;
            Debug::Message message = Debug::log();
            while(ebp != null)
            {
               message << ebp[1];
               if(i % 6 == 0)
                  message << '\n';
               else
                  message << ' ';
               ebp = (void**)ebp[0];
               ++i;
            }
         }
         MessageBox(GetConsoleWindow(), "Execution stack overflow.", "Evlan", MB_OK | MB_ICONERROR | MB_SETFOREGROUND | MB_TOPMOST);
      }

      return EXCEPTION_CONTINUE_SEARCH;
   }
#endif

   void internalInit()
   {
      internalInit(DEFAULT_PARAMETERS);
   }

   void internalInit(Parameters parameters)
   {
#ifdef PLATFORM_WIN32
      SetUnhandledExceptionFilter(handleException);
#endif

      gParameters = parameters;

      if(gParameters.heapSize < gParameters.largeAllocationThreshold * 2)
         gParameters.heapSize = gParameters.largeAllocationThreshold * 2;

      // TODO (KV):  Insure sizes are multiples of page size
   }

   void init()
   {
      init(DEFAULT_PARAMETERS);
   }

   void init(Parameters parameters)
   {
      internalInit(parameters);

      memset(gGenerations, 0, sizeof(gGenerations));

      gGenerations[0].blocks   = allocateHeap();
      gGenerations[0].modules  = null;
      gGenerations[0].phase    = Generation::PASSIVE;
      gGenerations[0].subPhase = -1;

      for(int i = 1; i < 64; i++)
      {
         gGenerations[i].phase = Generation::UPDATE;
         gGenerations[i].subPhase = -1;
      }

      core::gStackEnd = (core::Value*)allocatePages(
         gParameters.stackSize * sizeof(core::Value), true);
      core::gStackTop = core::gStackEnd + gParameters.stackSize;
   }

   void shutdown()
   {
      freePages(core::gStackEnd, gParameters.stackSize * sizeof(core::Value));
      core::gStackTop = null;
      core::gStackEnd = null;

      for(int i = 0; i < 64; i++)
      {
         deleteBlocks(gGenerations[i].blocks);

         InternalModule* modules = gGenerations[i].modules;
         while(modules != null)
         {
            InternalModule* next = modules->next;
            modules->~InternalModule();
            freePages(modules, modules->totalSize);
            modules = next;
         }

         InternalProcess* processes = gGenerations[i].processes;
         while(processes != null)
         {
            InternalProcess* next = processes->next;
            delete processes;
            processes = next;
         }
      }
      memset(gGenerations, 0, sizeof(gGenerations));

      gBlockMap.clear();
      gModuleBlockMap.clear();
#ifdef TRACK_DEADBLOCKS
      for(std::map<BlockKey, Block*>::iterator iter = gDeadBlockMap->begin();
          iter != gDeadBlockMap->end(); ++iter)
      {
         freePages(iter->second->data, iter->second->size * sizeof(word));
         delete iter->second;
      }
      gDeadBlockMap.clear();
#endif

      gHeapTop = null;
      gHeapEnd = null;
   }

   word* allocateWords(int size)
   {
      if(size == 0)
         return null;
      else if(size < gParameters.largeAllocationThreshold)
      {
         if(gHeapEnd - gHeapTop < size)
            collect();
         word* result = gHeapTop;
         gHeapTop += size;
#ifdef DEBUG
         static void* trace1 = null, *trace2 = null;
         for(int i = 0; i < size; i++)
         {
            if(result[i] != 0)
               FW_THROW(spnew FatalError($("Corruption in unallocated heap space.  Last caller:  0x") + String((word)trace1, 8, 16) + $(" 0x") + String((word)trace2, 8, 16)));
         }
         trace1 = __builtin_return_address(0);
         trace2 = __builtin_return_address(1);
#endif
         return result;
      }
      else
      {
         Block* block = allocateBlock(1);

         block->next = gGenerations[0].blocks;
         gGenerations[0].blocks = block;
         block->generation = 0;
         block->size = size;
         block->data = (word*)allocatePages(size * sizeof(word));
         block->dataSegments.reset(size);
         block->valueSegments.reset(size);

         block->headerLocation = 0;
         block->fileSegments[0].start = 0;
         block->fileSegments[1].start = size;
         block->fileSegments[0].fileLocation = 0;

         gBlockMap[BlockKey(block->data, block->data + size)] = block;

         return block->data;
      }
   }

   void returnWords(word* target, int size)
   {
      if(target + size == gHeapTop)
      {
         memset(target, 0, size * sizeof(word));
         gHeapTop = target;
      }
   }

   core::Module* allocateModule(int functionTagCount,
                                int objectTagCount,
                                int constantCount,
                                int extraDataSize,
                                const int arities[],
                                const int instructionCounts[],
                                const int functionDocSizes[],
                                const int memberCounts[],
                                const int objectDocSizes[],
                                bool hasSourceLocations)
   {
      int totalSize = sizeof(InternalModule)
                    + functionTagCount * (sizeof(core::FunctionTag*) + sizeof(core::FunctionTag))
                    + objectTagCount * (sizeof(core::ObjectTag*) + sizeof(core::ObjectTag))
                    + constantCount * sizeof(core::Value);
      int totalInstructionCount = 0;
      for(int i = 0; i < functionTagCount; i++)
      {
         totalInstructionCount += instructionCounts[i];
         totalSize += arities[i] * sizeof(int);
         totalSize += functionDocSizes[i];
      }
      totalSize += totalInstructionCount * sizeof(core::Instruction);
      if(hasSourceLocations)
         totalSize += totalInstructionCount * sizeof(core::SourceLocation);
      for(int i = 0; i < objectTagCount; i++)
      {
         totalSize += memberCounts[i] * sizeof(core::ObjectTag::Member);
         totalSize += objectDocSizes[i];
      }

      InternalModule* module = (InternalModule*)allocatePages(totalSize);
      gModuleBlockMap[BlockKey(module, (byte*)module + totalSize)] = module;

      if(extraDataSize > 0)
      {
         /* This MUST be allocated on the heap because stuff might
          * point to it without pointing to the module. */
         /* This MUST be allocated BEFORE the module is placed into
          * generation 0, to insure that the extra data does not end
          * up in a different generation. */
         module->extraData = allocateWords(extraDataSize);
      }

      module->functionTagCount = functionTagCount;
      module->objectTagCount   = objectTagCount;
      module->constantCount    = constantCount;
      module->extraDataSize    = extraDataSize;

      module->functionTags = (core::FunctionTag**)(module + 1);
      module->objectTags = (core::ObjectTag**)(module->functionTags + functionTagCount);
      module->constants = (core::Value*)(module->objectTags + objectTagCount);
      module->totalSize = totalSize;

      module->next = gGenerations[0].modules;
      gGenerations[0].modules = module;
      module->generation = 0;
      module->rootTrackGeneration = -1;
      module->marked = false;
      module->tracked = true;

      void* pos = module->constants + constantCount;

      for(int i = 0; i < functionTagCount; i++)
      {
         core::FunctionTag* tag = (core::FunctionTag*)pos;
         pos = (byte*)pos + sizeof(core::FunctionTag) + instructionCounts[i] * sizeof(core::Instruction);
         module->functionTags[i] = tag;

         tag->module = module;
         tag->instructionCount = instructionCounts[i];
         tag->arity = arities[i];
         tag->parameterNames = (int*)pos;
         pos = (byte*)pos + arities[i] * sizeof(int);
      }

      for(int i = 0; i < objectTagCount; i++)
      {
         core::ObjectTag* tag = (core::ObjectTag*)pos;
         pos = (byte*)pos + sizeof(core::ObjectTag) + memberCounts[i] * sizeof(core::ObjectTag::Member);
         module->objectTags[i] = tag;

         tag->module = module;
         tag->memberCount = memberCounts[i];
      }

      if(hasSourceLocations)
      {
         for(int i = 0; i < functionTagCount; i++)
         {
            module->functionTags[i]->sourceLocations = (core::SourceLocation*)pos;
            pos = (core::SourceLocation*)pos + instructionCounts[i];
         }
      }

      /* Documents must come last because they are byte-aligned rather
       * than word-aligned. */
      for(int i = 0; i < functionTagCount; i++)
      {
         module->functionTags[i]->documentation = (char*)pos;
         module->functionTags[i]->documentationLength = functionDocSizes[i];
         pos = (byte*)pos + functionDocSizes[i];
      }
      for(int i = 0; i < objectTagCount; i++)
      {
         module->objectTags[i]->documentation = (char*)pos;
         module->objectTags[i]->documentationLength = objectDocSizes[i];
         pos = (byte*)pos + objectDocSizes[i];
      }

      return module;
   }

   void setModuleRoot(core::Module* module)
   {
      InternalModule* internalModule = ((InternalModule*)module);
      internalModule->root = *core::gStackTop;
      internalModule->rootTrackGeneration = 0;
      if(internalModule->generation > 0)
      {
         internalModule->rootTrackNext = gGenerations[0].rootTrackModules;
         if(internalModule->rootTrackNext != null)
            internalModule->rootTrackNext->rootTrackLink = &internalModule->rootTrackNext;
         gGenerations[0].rootTrackModules = internalModule;
         internalModule->rootTrackLink = &gGenerations[0].rootTrackModules;
      }
      else
      {
         internalModule->rootTrackNext = null;
         internalModule->rootTrackLink = null;
      }
   }

   core::FunctionTag* findFunction(core::Instruction* instruction)
   {
      std::map<BlockKey, InternalModule*>::iterator iter =
         gModuleBlockMap.find(BlockKey(instruction, instruction + 1));
      if(iter == gModuleBlockMap.end())
         return null;
      else
      {
         InternalModule* module = iter->second;

         int a = 0;
         int b = module->functionTagCount;

         while(a < b)
         {
            int midpoint = (a + b) / 2;
            if(instruction < module->functionTags[midpoint]->instructions)
               b = midpoint;
            else if(instruction >= module->functionTags[midpoint]->instructions +
                                   module->functionTags[midpoint]->instructionCount)
               a = midpoint + 1;
            else
               return module->functionTags[midpoint];
         }
         return null;
      }
   }

   void makeProcess(String type)
   {
      makeProcess(operations::getIdentifierId(type));
   }
   void makeProcess(int type)
   {
      InternalProcess* process = new InternalProcess;
      process->marked = false;
      process->state = *core::gStackTop;
      process->keepAlive = false;
      process->type = type;

      process->next = gGenerations[0].processes;
      gGenerations[0].processes = process;
      process->generation = 0;

      process->stateTrackNext = null;
      process->stateTrackLink = null;

      process->stateTrackGeneration = 0;

      process->headerLocation = 0;

      *core::gStackTop = core::Value::makeProcess(process);
   }

   void modifyProcessState(core::Process* process)
   {
      InternalProcess* internalProcess = (InternalProcess*)process;

      //remove from old track list
      if(internalProcess->stateTrackNext != null)
         internalProcess->stateTrackNext->stateTrackLink = internalProcess->stateTrackLink;
      if(internalProcess->stateTrackLink != null)
         *internalProcess->stateTrackLink = internalProcess->stateTrackNext;
      internalProcess->stateTrackNext = null;
      internalProcess->stateTrackLink = null;

      //set new state
      internalProcess->state = *(core::gStackTop++);

      //add to generation zero
      internalProcess->stateTrackGeneration = 0;
      if(internalProcess->generation > 0)
      {
         internalProcess->stateTrackNext = gGenerations[0].stateTrackProcesses;
         if(internalProcess->stateTrackNext != null)
            internalProcess->stateTrackNext->stateTrackLink = &internalProcess->stateTrackNext;
         gGenerations[0].stateTrackProcesses = internalProcess;
         internalProcess->stateTrackLink = &gGenerations[0].stateTrackProcesses;
      }
   }

   // ================================================================
   void scanValues(core::Value* start, int size, int generation);

   void markValueSegment(core::Value* start, int size, int generation)
   {
      if(size == 0)
         return;

      std::map<BlockKey, Block*>::iterator iter =
         gBlockMap.find(BlockKey(start, start + size));
      if(iter == gBlockMap.end())
         FW_THROW(spnew FatalError($("Invalid pointer encountered during scan.")));
      Block* block = iter->second;
      int startIndex = (word*)start - block->data;

      if(gGenerations[block->generation].phase == Generation::UPDATE)
      {
         //must make sure to mark the *updated* location
         start = (core::Value*)block->valueSegments.lookup(startIndex);

#ifdef TRACK_DEADBLOCKS
         if(gDeadBlockMap.find(BlockKey(start, start + size))
            != gDeadBlockMap.end())
            FW_THROW(spnew FatalError($("Pointer to dead block discovered.")));
#endif

         iter = gBlockMap.find(BlockKey(start, start + size));
#ifdef DEBUG
         if(iter == gBlockMap.end())
            FW_THROW(spnew FatalError($("Updated block location not in map.")));
#endif
         block = iter->second;
         startIndex = (word*)start - block->data;
      }

      if(block->generation <= generation)
      {
#ifdef DEBUG
         if(!gSuperSweeping && block->generation < generation)
            FW_THROW(spnew FatalError($("Reverse generational link detected.")));
#endif

         if(block->valueSegments.insert(startIndex, startIndex + size * core::WORDS_PER_VALUE))
            scanValues(start, size, generation);
      }
      else if(gGenerations[block->generation].phase == Generation::MARK
           && (gGenerations[block->generation].subPhase < 2
            || gGenerations[block->generation].subPhase <= generation))
         block->valueSegments.insert(startIndex, startIndex + size * core::WORDS_PER_VALUE);
   }

   void markDataSegment(word* start, int size, int generation)
   {
      if(size == 0)
         return;

      std::map<BlockKey, Block*>::iterator iter =
         gBlockMap.find(BlockKey(start, start + size));
      if(iter == gBlockMap.end())
         FW_THROW(spnew FatalError($("Invalid pointer encountered during scan.")));
      Block* block = iter->second;
      int startIndex = start - block->data;

      if(gGenerations[block->generation].phase == Generation::UPDATE)
      {
         //must make sure to mark the *updated* location
         start = block->dataSegments.lookup(startIndex);

#ifdef TRACK_DEADBLOCKS
         if(gDeadBlockMap.find(BlockKey(start, start + size))
            != gDeadBlockMap.end())
            FW_THROW(spnew FatalError($("Pointer to dead block discovered.")));
#endif

         iter = gBlockMap.find(BlockKey(start, start + size));
#ifdef DEBUG
         if(iter == gBlockMap.end())
            FW_THROW(spnew FatalError($("Updated block location not in map.")));
#endif
         block = iter->second;
         startIndex = start - block->data;
      }

#ifdef DEBUG
      if(!gSuperSweeping && block->generation < generation)
         FW_THROW(spnew FatalError($("Reverse generational link detected.")));
#endif

      if(block->generation <= generation
      || (gGenerations[block->generation].phase == Generation::MARK
       && (gGenerations[block->generation].subPhase < 2
        || gGenerations[block->generation].subPhase <= generation)))
         block->dataSegments.insert(startIndex, startIndex + size);
   }

   struct ScanSegment
   {
      core::Value* segment;
      int size;

      inline ScanSegment() {}
      inline ScanSegment(core::Value* segment, int size)
         : segment(segment), size(size) {}
   };

   void scanValues(core::Value* segment, int size, int generation)
   {
      //hack:  recursive scanValues turned out to overflow the stack...
      static bool running = false;
      static std::stack<ScanSegment> segmentStack;

      segmentStack.push(ScanSegment(segment, size));
      if(running)
         return;

      running = true;

      while(!segmentStack.empty())
      {
         segment = segmentStack.top().segment;
         size = segmentStack.top().size;
         segmentStack.pop();

         for(int i = 0; i < size; i++)
         {
            void* tag = null;
            switch(segment[i].type)
            {
            case core::Value::RUNTIME_ERROR:
               if(segment[i].runtimeError.location != null)
                  tag = segment[i].runtimeError.location;
               break;
            case core::Value::TASK:
               markValueSegment(segment[i].task.parameters,
                                segment[i].task.parameterCount,
                                generation);
               break;
            case core::Value::CONTEXT:
               if(segment[i].context.contextType == core::Value::THROW)
                  markValueSegment(segment[i].context.parameter, 2, generation);
               else if(segment[i].context.parameter != null)
                  markValueSegment(segment[i].context.parameter, 1, generation);
               if(segment[i].context.parent != null)
                  markValueSegment(segment[i].context.parent, 1, generation);
               break;
            case core::Value::FUNCTION:
               tag = segment[i].function.tag;
               markValueSegment(segment[i].function.environment,
                                segment[i].function.tag->environmentSize,
                                generation);
               break;
            case core::Value::OBJECT:
               tag = segment[i].object.tag;
               markValueSegment(segment[i].object.members,
                                segment[i].object.tag->memberCount,
                                generation);
               break;
            case core::Value::ARRAY:
               markValueSegment(segment[i].array.elements,
                                segment[i].array.size,
                                generation);
               break;
            case core::Value::PACKED_ARRAY:
               {
                  int wordSize;
                  if(segment[i].packedArray.elementSize < core::LOG2_WORD_SIZE)
                  {
                     int diff = (core::LOG2_WORD_SIZE - segment[i].packedArray.elementSize);
                     int offsetSize = segment[i].packedArray.size + segment[i].packedArray.offset;
                     wordSize = offsetSize >> diff;
                     if((offsetSize & ((1 << diff) - 1)) != 0)
                        ++wordSize;
                  }
                  else if(segment[i].packedArray.elementSize == core::LOG2_WORD_SIZE)
                     wordSize = segment[i].packedArray.size;
                  else
                     wordSize = segment[i].packedArray.size << (segment[i].packedArray.elementSize - core::LOG2_WORD_SIZE);

                  markDataSegment(segment[i].packedArray.elements,
                                  wordSize,
                                  generation);
               }
               break;
            case core::Value::LARGE_DATA:
               markDataSegment(segment[i].largeData.value,
                               1 << (segment[i].largeData.size - core::LOG2_WORD_SIZE),
                               generation);
               break;

            case core::Value::MODULE:
               tag = segment[i].module.module;
               break;

            case core::Value::PROCESS:
               {
                  InternalProcess* process = (InternalProcess*)segment[i].process.process;
                  if(process != null
                  && !process->marked
                  && !process->keepAlive
                  && (process->generation <= generation
                   || (gGenerations[process->generation].phase == Generation::MARK
                    && (gGenerations[process->generation].subPhase < 2
                     || gGenerations[process->generation].subPhase <= generation))))
                  {
#ifdef DEBUG
                     if(!gSuperSweeping && process->generation < generation)
                        FW_THROW(spnew FatalError($("Reverse generational link detected.")));
#endif
                     process->marked = true;

                     if((process->stateTrackGeneration == process->generation || gSuperSweeping)
                     && process->generation <= generation)
                        scanValues(&process->state, 1, generation);
                  }
               }
               break;

            case core::Value::EMPTY:
            case core::Value::BUILTIN_OBJECT:
            case core::Value::NATIVE_CONTAINER:
            case core::Value::NATIVE_FUNCTION:
               //no (scannable) pointers
               break;

            default:
               FW_THROW(spnew FatalError($("Encountered invalid value during memory scan.  Location: 0x")
                                        + String((word)(segment + i), 2 * sizeof(word), 16)));
            }

            if(tag != null)
            {
               std::map<BlockKey, InternalModule*>::iterator iter =
                  gModuleBlockMap.find(BlockKey(tag, (word*)tag + 1));
               if(iter != gModuleBlockMap.end())
               {
                  InternalModule* module = iter->second;

                  if(!module->marked
                  && module->tracked
                  && (module->generation <= generation
                   || (gGenerations[module->generation].phase == Generation::MARK
                    && (gGenerations[module->generation].subPhase < 2
                     || gGenerations[module->generation].subPhase <= generation))))
                  {
#ifdef DEBUG
                     if(!gSuperSweeping && module->generation < generation)
                        FW_THROW(spnew FatalError($("Reverse generational link detected.")));
#endif
                     module->marked = true;
                     markDataSegment(module->extraData, module->extraDataSize, generation);
                     scanValues(module->constants, module->constantCount, generation);

                     if((module->rootTrackGeneration == module->generation || gSuperSweeping)
                     && module->generation <= generation)
                        scanValues(&module->root, 1, generation);
                  }
               }
            }
         }
      }

      running = false;
   }

   // ================================================================
   core::Value* updateValueSegment(core::Value* start, int size, int generation)
   {
      if(size == 0)
         return start;

      std::map<BlockKey, Block*>::iterator iter =
         gBlockMap.find(BlockKey(start, start + size));
      if(iter == gBlockMap.end())
      {
#ifdef TRACK_DEADBLOCKS
         if(gDeadBlockMap.find(BlockKey(start, start + size))
            != gDeadBlockMap.end())
            FW_THROW(spnew FatalError($("Pointer to dead block discovered.")));
#endif
         return start;
      }
      Block* block = iter->second;

      while(gGenerations[block->generation].phase == Generation::UPDATE)
      {
         start = (core::Value*)block->valueSegments.lookup((word*)start - block->data);
         iter = gBlockMap.find(BlockKey(start, start + size));
         if(iter == gBlockMap.end())
            return start;
         block = iter->second;
      }

      if(block->generation <= generation)
         return (core::Value*)block->valueSegments.lookup((word*)start - block->data);
      else
         return start;
   }

   word* updateDataSegment(word* start, int size, int generation)
   {
      if(size == 0)
         return start;

      std::map<BlockKey, Block*>::iterator iter =
         gBlockMap.find(BlockKey(start, start + size));
      if(iter == gBlockMap.end())
      {
#ifdef TRACK_DEADBLOCKS
         if(gDeadBlockMap.find(BlockKey(start, start + size))
            != gDeadBlockMap.end())
            FW_THROW(spnew FatalError($("Pointer to dead block discovered.")));
#endif
         return start;
      }
      Block* block = iter->second;

      while(gGenerations[block->generation].phase == Generation::UPDATE)
      {
         start = block->dataSegments.lookup(start - block->data);
         iter = gBlockMap.find(BlockKey(start, start + size));
         if(iter == gBlockMap.end())
            return start;
         block = iter->second;
      }

      if(block->generation <= generation)
         return block->dataSegments.lookup(start - block->data);
      else
         return start;
   }

   void updateValues(core::Value* segment, int size, int generation)
   {
      for(int i = 0; i < size; i++)
      {
         switch(segment[i].type)
         {
         case core::Value::TASK:
            segment[i].task.parameters =
               updateValueSegment(segment[i].task.parameters,
                                  segment[i].task.parameterCount,
                                  generation);
            break;
         case core::Value::CONTEXT:
            if(segment[i].context.contextType == core::Value::THROW)
               segment[i].context.parameter = updateValueSegment(segment[i].context.parameter, 2, generation);
            else if(segment[i].context.parameter != null)
               segment[i].context.parameter = updateValueSegment(segment[i].context.parameter, 1, generation);
            if(segment[i].context.parent != null)
               segment[i].context.parent = updateValueSegment(segment[i].context.parent, 1, generation);
            break;
         case core::Value::FUNCTION:
            segment[i].function.environment =
               updateValueSegment(segment[i].function.environment,
                                  segment[i].function.tag->environmentSize,
                                  generation);
            break;
         case core::Value::OBJECT:
            segment[i].object.members =
               updateValueSegment(segment[i].object.members,
                                  segment[i].object.tag->memberCount,
                                  generation);
            break;
         case core::Value::ARRAY:
            segment[i].array.elements =
               updateValueSegment(segment[i].array.elements,
                                  segment[i].array.size,
                                  generation);
            break;
         case core::Value::PACKED_ARRAY:
            {
               int wordSize;
               if(segment[i].packedArray.elementSize < core::LOG2_WORD_SIZE)
               {
                  int diff = (core::LOG2_WORD_SIZE - segment[i].packedArray.elementSize);
                  int offsetSize = segment[i].packedArray.size + segment[i].packedArray.offset;
                  wordSize = offsetSize >> diff;
                  if((offsetSize & ((1 << diff) - 1)) != 0)
                     ++wordSize;
               }
               else if(segment[i].packedArray.elementSize == core::LOG2_WORD_SIZE)
                  wordSize = segment[i].packedArray.size;
               else
                  wordSize = segment[i].packedArray.size << (segment[i].packedArray.elementSize - core::LOG2_WORD_SIZE);

               segment[i].packedArray.elements =
                  updateDataSegment(segment[i].packedArray.elements,
                                    wordSize,
                                    generation);
            }
            break;
         case core::Value::LARGE_DATA:
               updateDataSegment(segment[i].largeData.value,
                                 1 << (segment[i].largeData.size - core::LOG2_WORD_SIZE),
                                 generation);
            break;

         case core::Value::EMPTY:
         case core::Value::RUNTIME_ERROR:
         case core::Value::BUILTIN_OBJECT:
         case core::Value::NATIVE_CONTAINER:
         case core::Value::NATIVE_FUNCTION:
         case core::Value::MODULE:
         case core::Value::PROCESS:
            //no (scannable) pointers
            break;

         default:
            FW_THROW(spnew FatalError($("Encountered invalid value during pointer update.")));
            break;
         }
      }
   }

   // ================================================================
   void clearMarks(int generation)
   {
      for(Block* block = gGenerations[generation].blocks; block != null; block = block->next)
      {
         block->valueSegments.reset(block->size);
         block->dataSegments.reset(block->size);
      }

      for(InternalModule* module = gGenerations[generation].modules; module != null; module = module->next)
         module->marked = false;
      for(InternalProcess* process = gGenerations[generation].processes; process != null; process = process->next)
         process->marked = false;
   }

   void scanGeneration(int generation, bool premarked)
   {
      if(premarked)
      {
         //count blocks
         int blockCount = 0;
         for(Block* block = gGenerations[generation].blocks; block != null; block = block->next)
            ++blockCount;

         //count value segments, as a whole and per block
         int segmentCount = 0;
         int blockSizes[blockCount];

         {
            int i = 0;
            for(Block* block = gGenerations[generation].blocks; block != null; block = block->next)
            {
               block->valueSegments.getSize(&blockSizes[i], null);
               segmentCount += blockSizes[i];
               ++i;
            }
         }

         //make a list of all current value segments
         core::Value* valueSegments[segmentCount];
         int valueCounts[segmentCount];

         {
            SegmentSet::Segment segments[segmentCount];
            int pos = 0;
            int i = 0;
            for(Block* block = gGenerations[generation].blocks; block != null; block = block->next)
            {
               block->valueSegments.getSegments(segments + pos);
               for(int j = 0; j < blockSizes[i]; j++)
               {
                  valueSegments[pos + j] = (core::Value*)(block->data
                                         + segments[pos + j].getStart());
                  valueCounts[pos + j] = (segments[pos + j].getEnd()
                                        - segments[pos + j].getStart())
                                       / core::WORDS_PER_VALUE;
               }
               pos += blockSizes[i];
               ++i;
            }
         }

         /* Scan them.  Note that scanning could, of course, create more
          * segments or expand existing segments.  However, scanValues()
          * recursively scans these as well when this happens, so we
          * don't have to worry about them here. */
         for(int i = 0; i < segmentCount; i++)
            scanValues(valueSegments[i], valueCounts[i], generation);
      }
#ifdef DEBUG
      else
      {
         //insure that no marks exist!
         int segmentCount, length;
         for(Block* block = gGenerations[generation].blocks; block != null; block = block->next)
         {
            block->valueSegments.getSize(&segmentCount, &length);
            if(segmentCount > 0)
               FW_THROW(spnew FatalError($("Block in non-premarked generation had marks.")));
            block->dataSegments.getSize(&segmentCount, &length);
            if(segmentCount > 0)
               FW_THROW(spnew FatalError($("Block in non-premarked generation had marks.")));
         }

         for(InternalModule* module = gGenerations[generation].modules; module != null; module = module->next)
         {
            if(module->marked)
               FW_THROW(spnew FatalError($("Module in non-premarked generation had marks.")));
         }
         for(InternalProcess* process = gGenerations[generation].processes; process != null; process = process->next)
         {
            if(process->marked)
               FW_THROW(spnew FatalError($("Process in non-premarked generation had marks.")));
         }
      }
#endif

      for(InternalModule* module = gGenerations[generation].rootTrackModules;
          module != null; module = module->rootTrackNext)
      {
         if(module->generation != generation)
            scanValues(&module->root, 1, generation);
      }

      for(InternalModule* module = gGenerations[generation].modules;
          module != null; module = module->next)
      {
         if(module->rootTrackGeneration == generation
         && (module->marked || !module->tracked))
         {
            markDataSegment(module->extraData, module->extraDataSize, generation);
            scanValues(module->constants, module->constantCount, generation);
            scanValues(&module->root, 1, generation);
         }
      }

      for(InternalProcess* process = gGenerations[generation].stateTrackProcesses;
          process != null; process = process->stateTrackNext)
      {
         if(process->generation != generation)
            scanValues(&process->state, 1, generation);
      }

      for(InternalProcess* process = gGenerations[generation].processes;
          process != null; process = process->next)
      {
         if(process->stateTrackGeneration == generation
         && (process->marked || process->keepAlive))
            scanValues(&process->state, 1, generation);
      }

      for(int i = generation + 1; i < 64; i++)
      {
         if(gGenerations[i].subPhase == generation)
            gGenerations[i].subPhase = generation - 1;
      }
   }

   // ================================================================
   //some helper stuff for sweepGeneration()
   struct BlockInfo
   {
      Block* block;
      int valueSegmentCount;
      int dataSegmentCount;
   };

   struct PotentialBlock
   {
      int id;
      int spaceRemaining;

      inline PotentialBlock(int id)
         : id(id), spaceRemaining(gParameters.blockSize) {}
      inline PotentialBlock(int id, int spaceRemaining)
         : id(id), spaceRemaining(spaceRemaining) {}

      inline bool operator<(const PotentialBlock& other) const
      {
         return spaceRemaining < other.spaceRemaining;
      }
   };

   bool orderSegments(SegmentSet::Segment a, SegmentSet::Segment b)
   {
      //descending order by size
      return a.getEnd() - a.getStart() > b.getEnd() - b.getStart();
   }

   void mapSegmentsToBlocks(SegmentSet::Segment segments[], int count, int targetGeneration, bool addToMap)
   {
      if(count == 0)
         return;

      /*
      sort segments by size
      priority queue of blocks -> block with most space left on top
      for each segment
         if top block has space for segment
            put segment in top block
         else
            create new block and put segment in it
            push block on priority queue
      */

      //sort segments by length in order to help in fitting them together
      std::sort(segments, segments + count, orderSegments);

      int blockCount = 0;
      int segmentIds[count];

      {
         std::priority_queue<PotentialBlock> blockQueue;
         blockQueue.push(PotentialBlock(blockCount++));

         for(int i = 0; i < count; i++)
         {
            int size = segments[i].getEnd() - segments[i].getStart();
            if(size > gParameters.blockSize)
               segmentIds[i] = blockCount++;   //entire segment becomes one block
            else if(size > blockQueue.top().spaceRemaining)
            {
               //not enough room in exist block; create new one
               segmentIds[i] = blockCount;
               blockQueue.push(PotentialBlock(blockCount++));
            }
            else
            {
               PotentialBlock block = blockQueue.top();
               blockQueue.pop();
               segmentIds[i] = block.id;
               block.spaceRemaining -= size;
               blockQueue.push(block);
            }
         }
      }

      struct
      {
         int size;
         int fileSegmentCount;
         Block* block;
         bool lastWasZero;
      } blockInfo[blockCount];

      memset(blockInfo, 0, sizeof(blockInfo));

      for(int i = 0; i < count; i++)
      {
         blockInfo[segmentIds[i]].size += segments[i].getEnd() - segments[i].getStart();

         int j = 0;
         while(segments[i].mFileSegment[j].start < segments[i].getEnd())
         {
            bool isZero = segments[i].mFileSegment[j].fileLocation == 0;
            if(!blockInfo[segmentIds[i]].lastWasZero || !isZero)
               ++blockInfo[segmentIds[i]].fileSegmentCount;
            blockInfo[segmentIds[i]].lastWasZero = isZero;
            ++j;
         }
      }

      for(int i = 0; i < blockCount; i++)
      {
         if(blockInfo[i].fileSegmentCount == 0)
            continue;  /* Rare:  All segments were very large and thus were
                        *        mapped to their own blocks.  Thus, the initial
                        *        block pushed to the queue ended up empty. */

         Block* block = allocateBlock(blockInfo[i].fileSegmentCount);

         block->next = gGenerations[targetGeneration].blocks;
         gGenerations[targetGeneration].blocks = block;
         block->generation = targetGeneration;
         block->size = blockInfo[i].size;
         block->fileSegments[blockInfo[i].fileSegmentCount].start = blockInfo[i].size;

         block->headerLocation = 0;

         block->data = (word*)allocatePages(block->size * sizeof(word));
         if(addToMap)
            gBlockMap[BlockKey(block->data, block->data + block->size)] = block;
         block->valueSegments.reset(block->size);
         block->dataSegments.reset(block->size);

         blockInfo[i].block = block;
         blockInfo[i].fileSegmentCount = 0;
         blockInfo[i].size = 0;
         blockInfo[i].lastWasZero = false;
      }

      for(int i = 0; i < count; i++)
      {
         int blockId = segmentIds[i];
         segments[i].setNewLocation(blockInfo[blockId].block->data + blockInfo[blockId].size);

         int j = 0;
         while(segments[i].mFileSegment[j].start < segments[i].getEnd())
         {
            if(!blockInfo[blockId].lastWasZero || segments[i].mFileSegment[j].fileLocation != 0)
            {
               FileSegment fileSegment = segments[i].mFileSegment[j];
               if(j == 0)
               {
                  if(segments[i].mFileSegment[j].fileLocation != 0)
                     fileSegment.fileLocation += (segments[i].getStart() - fileSegment.start) * sizeof(word);
                  fileSegment.start = blockInfo[blockId].size;
               }
               else
                  fileSegment.start += blockInfo[blockId].size - segments[i].getStart();

               blockInfo[blockId].block->fileSegments[blockInfo[blockId].fileSegmentCount++] = fileSegment;
            }

            blockInfo[blockId].lastWasZero = segments[i].mFileSegment[j].fileLocation == 0;
            ++j;
         }

         blockInfo[blockId].size += segments[i].getEnd() - segments[i].getStart();
      }
   }

   // ================================================================
   /* This returns a list of blocks that should be deleted as soon as
    * generation zero is updated. */
   Block* sweepGeneration(int generation)
   {
      /* First, make a list of all blocks, deciding which are to be
       * crunched and which can just be moved into the next
       * generation. */
      int blockCount = 0;
      for(Block* block = gGenerations[generation].blocks; block != null; block = block->next)
         ++blockCount;

      BlockInfo blockInfo[blockCount];
      int crunchedBlockCount = 0;
      int unchangedBlockCount = 0;
      int valueSegmentCount = 0;
      int dataSegmentCount = 0;
      for(Block* block = gGenerations[generation].blocks; block != null; block = block->next)
      {
         int valueSize, valueLength,
             dataSize, dataLength,
             length;
         block->valueSegments.getSize(&valueSize, &valueLength);
         block->dataSegments.getSize(&dataSize, &dataLength);
         length = valueLength + dataLength;

         /* FIXME (KV):  Consider crunching bigger blocks even when
          * more than half of the block is still in use?  Don't forget
          * that a block has to stick around for awhile after the
          * crunched copy is made, while an uncrunched block is just
          * moved to the next generation with no copying.  So, it
          * actually takes more memory for awhile. */
         if(length < gParameters.largeAllocationThreshold
         || length * 2 <= block->size)
         {
            /* We can gain a significant amount of memory by crunching
             * this block. */
            valueSegmentCount += valueSize;
            dataSegmentCount += dataSize;
            blockInfo[crunchedBlockCount].block             = block;
            blockInfo[crunchedBlockCount].valueSegmentCount = valueSize;
            blockInfo[crunchedBlockCount].dataSegmentCount  = dataSize;
            ++crunchedBlockCount;
         }
         else
         {
            /* We cannot gain a significant amount of memory by
             * crunching this block. */
            ++unchangedBlockCount;
            int index = blockCount - unchangedBlockCount;
            blockInfo[index].block             = block;
            blockInfo[index].valueSegmentCount = valueSize;
            blockInfo[index].dataSegmentCount  = dataSize;
         }
      }

      // -------------------------------------------------------------
      /* Allocate memory in which to copy crunched blocks and map
       * segments to this memory, but don't copy just yet. */

      /* First see if we need to clean up the next generation.
       * We can't actually delete these blocks yet, since the
       * generation being swept may contain pointers to them.
       * However, we need to take them out of the next generation's
       * block list so that we don't get them mixed up with new blocks
       * that we put there.  Once this generation is updated we can
       * delete the blocks. */
      Block* garbageBlocks = null;
      int targetGeneration = generation + 1;
      if(gGenerations[targetGeneration].phase == Generation::UPDATE)
      {
         garbageBlocks = gGenerations[targetGeneration].blocks;
         gGenerations[targetGeneration].blocks = null;
         gGenerations[targetGeneration].phase = Generation::PASSIVE;

         //hack:  move the old blocks into this generation so that they get updated
         for(Block* block = garbageBlocks; block != null; block = block->next)
            block->generation = generation;
      }
      else
      {
         gGenerations[targetGeneration].phase = Generation::MARK;
         gGenerations[targetGeneration].subPhase = generation - 1;
      }

      gGenerations[generation].blocks = null;
      if(valueSegmentCount > 0)
      {
         SegmentSet::Segment segments[valueSegmentCount];
         int pos = 0;

         for(int i = 0; i < crunchedBlockCount; i++)
         {
            blockInfo[i].block->valueSegments.getSegments(segments + pos);

            int k = 0;
            for(int j = 0; j < blockInfo[i].valueSegmentCount; j++)
            {
               int start = segments[pos + j].getStart();
               while(start >= blockInfo[i].block->fileSegments[k].start)
                  ++k;
               --k;
               segments[pos+j].mFileSegment = blockInfo[i].block->fileSegments + k;
            }

            pos += blockInfo[i].valueSegmentCount;
         }

#ifdef DEBUG
         if(pos != valueSegmentCount)
            Debug::log(Debug::ERROR) << "evlan::vm::core::memory:  Assertion failed:  pos == valueSegmentCount";
#endif

         mapSegmentsToBlocks(segments, valueSegmentCount, targetGeneration, true);
      }
      if(dataSegmentCount > 0)
      {
         SegmentSet::Segment segments[dataSegmentCount];
         int pos = 0;

         for(int i = 0; i < crunchedBlockCount; i++)
         {
            blockInfo[i].block->dataSegments.getSegments(segments + pos);

            int k = 0;
            for(int j = 0; j < blockInfo[i].dataSegmentCount; j++)
            {
               int start = segments[pos + j].getStart();
               while(start >= blockInfo[i].block->fileSegments[k].start)
                  ++k;
               --k;
               segments[pos+j].mFileSegment = blockInfo[i].block->fileSegments + k;
            }

            pos += blockInfo[i].dataSegmentCount;
         }

#ifdef DEBUG
         if(pos != dataSegmentCount)
            Debug::log(Debug::ERROR) << "evlan::vm::core::memory:  Assertion failed:  pos != dataSegmentCount";
#endif

         mapSegmentsToBlocks(segments, dataSegmentCount, targetGeneration, true);
      }

      // -------------------------------------------------------------
      //Update pointers and copy segments into the new generation.

      /* Unchanged blocks need to be moved to the new generation
       * before any updates are done.  This way, pointers to these
       * blocks will not be updated, since the blocks will be in a
       * generation which is not in the update phase. */
      for(int i = crunchedBlockCount; i < blockCount; i++)
      {
         //move to the next generation
         blockInfo[i].block->next = gGenerations[targetGeneration].blocks;
         gGenerations[targetGeneration].blocks = blockInfo[i].block;
         blockInfo[i].block->generation = targetGeneration;
      }

      //update unchanged blocks
      for(int i = crunchedBlockCount; i < blockCount; i++)
      {
         //update pointers
         SegmentSet::Segment segments[blockInfo[i].valueSegmentCount];
         blockInfo[i].block->valueSegments.getSegments(segments);
         for(int j = 0; j < blockInfo[i].valueSegmentCount; j++)
         {
            core::Value* data = (core::Value*)(blockInfo[i].block->data + segments[j].getStart());
            int size = (segments[j].getEnd() - segments[j].getStart()) / core::WORDS_PER_VALUE;
            updateValues(data, size, generation);
         }

         blockInfo[i].block->valueSegments.reset(blockInfo[i].block->size);
         blockInfo[i].block->dataSegments.reset(blockInfo[i].block->size);
      }

      //update crunched blocks
      for(int i = 0; i < crunchedBlockCount; i++)
      {
         if(blockInfo[i].valueSegmentCount > 0)
         {
            SegmentSet::Segment segments[blockInfo[i].valueSegmentCount];
            blockInfo[i].block->valueSegments.getSegments(segments);
            for(int j = 0; j < blockInfo[i].valueSegmentCount; j++)
            {
               core::Value* data = (core::Value*)(blockInfo[i].block->data + segments[j].getStart());
               int size = (segments[j].getEnd() - segments[j].getStart()) / core::WORDS_PER_VALUE;
               updateValues(data, size, generation);
               memcpy(segments[j].getNewLocation(), data, size * sizeof(core::Value));
            }
         }

         if(blockInfo[i].dataSegmentCount > 0)
         {
            SegmentSet::Segment segments[blockInfo[i].dataSegmentCount];
            blockInfo[i].block->dataSegments.getSegments(segments);
            for(int j = 0; j < blockInfo[i].dataSegmentCount; j++)
            {
               word* data = blockInfo[i].block->data + segments[j].getStart();
               int size = segments[j].getEnd() - segments[j].getStart();
               memcpy(segments[j].getNewLocation(), data, size * sizeof(word));
            }
         }

         //keep old block in current generation, to be deleted later
         blockInfo[i].block->next = gGenerations[generation].blocks;
         gGenerations[generation].blocks = blockInfo[i].block;
      }

      // -------------------------------------------------------------
      //deal with modules
      InternalModule* module = gGenerations[generation].modules;
      gGenerations[generation].modules = null;
      while(module != null)
      {
         InternalModule* next = module->next;
         if(module->marked || !module->tracked)
         {
            updateValues(module->constants, module->constantCount, generation);
            module->extraData = updateDataSegment(module->extraData, module->extraDataSize, generation);

            module->generation = targetGeneration;
            module->next = gGenerations[targetGeneration].modules;
            gGenerations[targetGeneration].modules = module;
            module->marked = false;

            if(module->rootTrackGeneration == generation)
            {
               updateValues(&module->root, 1, generation);
//               catch(...)
//               {
//                  MessageBox(GetConsoleWindow(), "From module.", "Evlan", MB_OK | MB_ICONERROR | MB_SETFOREGROUND | MB_TOPMOST);
//                  throw;
//               }

               module->rootTrackGeneration = targetGeneration;
            }
         }
         else
         {
            if(module->rootTrackLink != null)
               *module->rootTrackLink = module->rootTrackNext;
            if(module->rootTrackNext != null)
               module->rootTrackNext->rootTrackLink = module->rootTrackLink;
            gModuleBlockMap.erase(BlockKey(module, (byte*)module + module->totalSize));
            module->~InternalModule();
            freePages(module, module->totalSize);
         }
         module = next;
      }

      module = gGenerations[generation].rootTrackModules;
      gGenerations[generation].rootTrackModules = null;
      while(module != null)
      {
         InternalModule* next = module->rootTrackNext;

         updateValues(&module->root, 1, generation);
         module->rootTrackGeneration = targetGeneration;

         if(module->generation > targetGeneration)
         {
            module->rootTrackNext = gGenerations[targetGeneration].rootTrackModules;
            if(module->rootTrackNext != null)
               module->rootTrackNext->rootTrackLink = &module->rootTrackNext;
            gGenerations[targetGeneration].rootTrackModules = module;
            module->rootTrackLink = &gGenerations[targetGeneration].rootTrackModules;
         }
         else
         {
            module->rootTrackNext = null;
            module->rootTrackLink = null;
         }

         module = next;
      }

      // -------------------------------------------------------------
      //deal with processes
      InternalProcess* process = gGenerations[generation].processes;
      gGenerations[generation].processes = null;
      while(process != null)
      {
         InternalProcess* next = process->next;
         if(process->marked || process->keepAlive)
         {
            process->generation = targetGeneration;
            process->next = gGenerations[targetGeneration].processes;
            gGenerations[targetGeneration].processes = process;
            process->marked = false;

            if(process->stateTrackGeneration == generation)
            {
               updateValues(&process->state, 1, generation);
               process->stateTrackGeneration = targetGeneration;
            }
         }
         else
         {
            if(process->stateTrackLink != null)
               *process->stateTrackLink = process->stateTrackNext;
            if(process->stateTrackNext != null)
               process->stateTrackNext->stateTrackLink = process->stateTrackLink;
            delete process;
         }
         process = next;
      }

      process = gGenerations[generation].stateTrackProcesses;
      gGenerations[generation].stateTrackProcesses = null;
      while(process != null)
      {
         InternalProcess* next = process->stateTrackNext;

         updateValues(&process->state, 1, generation);
         process->stateTrackGeneration = targetGeneration;

         if(process->generation > targetGeneration)
         {
            process->stateTrackNext = gGenerations[targetGeneration].stateTrackProcesses;
            if(process->stateTrackNext != null)
               process->stateTrackNext->stateTrackLink = &process->stateTrackNext;
            gGenerations[targetGeneration].stateTrackProcesses = process;
            process->stateTrackLink = &gGenerations[targetGeneration].stateTrackProcesses;
         }
         else
         {
            process->stateTrackNext = null;
            process->stateTrackLink = null;
         }

         process = next;
      }

      // -------------------------------------------------------------
      //finally, deal with phases and such
      gGenerations[generation].phase = Generation::UPDATE;
      gGenerations[generation].subPhase = -1;
      gGenerations[generation].headerLocation = 0;
      gGenerations[generation+1].headerLocation = 0;

      return garbageBlocks;
   }

   // ================================================================
   void collect()
   {
      int otherGeneration = sequence::next();

#if 0
      {
         Debug::Message message = Debug::log();
         message << "collect() " << otherGeneration << " from:";

         if(__builtin_frame_address(0) != 0) {
         message << ' ' << __builtin_return_address(0);
         if(__builtin_frame_address(1) != 0) {
         message << ' ' << __builtin_return_address(1);
         if(__builtin_frame_address(2) != 0) {
         message << ' ' << __builtin_return_address(2);
         if(__builtin_frame_address(3) != 0) {
         message << ' ' << __builtin_return_address(3);
         if(__builtin_frame_address(4) != 0) {
         message << ' ' << __builtin_return_address(4);
         if(__builtin_frame_address(5) != 0) {
         message << ' ' << __builtin_return_address(5);
         if(__builtin_frame_address(6) != 0) {
         message << ' ' << __builtin_return_address(6);
         if(__builtin_frame_address(7) != 0) {
         message << ' ' << __builtin_return_address(7);
         if(__builtin_frame_address(8) != 0) {
         message << ' ' << __builtin_return_address(8);
         if(__builtin_frame_address(9) != 0) {
         message << ' ' << __builtin_return_address(9);
         if(__builtin_frame_address(10) != 0) {
         message << ' ' << __builtin_return_address(10);
         if(__builtin_frame_address(11) != 0) {
         message << ' ' << __builtin_return_address(11);
         if(__builtin_frame_address(12) != 0) {
         message << ' ' << __builtin_return_address(12);
         if(__builtin_frame_address(13) != 0) {
         message << ' ' << __builtin_return_address(13);
         if(__builtin_frame_address(14) != 0) {
         message << ' ' << __builtin_return_address(14);
         if(__builtin_frame_address(15) != 0) {
         message << ' ' << __builtin_return_address(15);
         }}}}}}}}}}}}}}}}
      }
#endif

#ifdef INCLUDE_WRITER_CODE
      if(gDebugOutput != null && gDebugVerbosity >= 2)
      {
         gDebugOutput->write($("======================================================================\n"));
         gDebugOutput->write($("Before sweep:\n"));
         printStatus(gDebugOutput, gDebugVerbosity);
      }
#endif //INCLUDE_WRITER_CODE

      //generation zero tends to be left in the UPDATE state for no particular reason.
      gGenerations[0].phase = Generation::PASSIVE;

      //scan generation zero
      scanGeneration(0, false);
      scanValues(core::gStackTop, core::gStackEnd + gParameters.stackSize - core::gStackTop, 0);
      scanValues((core::Value*)core::gTaskQueue, core::gTaskQueueSize * 2, 0);

      //sweep generation zero
      Block* garbageBlocks = sweepGeneration(0);
      updateValues(core::gStackTop, core::gStackEnd + gParameters.stackSize - core::gStackTop, 0);
      updateValues((core::Value*)core::gTaskQueue, core::gTaskQueueSize * 2, 0);
      deleteBlocks(garbageBlocks);
      deleteBlocks(gGenerations[0].blocks);
      gGenerations[0].blocks = allocateHeap();

#ifdef INCLUDE_WRITER_CODE
      if(gDebugOutput != null && gDebugVerbosity >= 2)
      {
         gDebugOutput->write($("----------------------------------------------------------------------\n"));
         gDebugOutput->write($("After generation 0 sweep:\n"));
         printStatus(gDebugOutput, gDebugVerbosity);
      }
#endif //INCLUDE_WRITER_CODE

      //scan and sweep other generation
      if(otherGeneration != 0)
      {
         if(otherGeneration == 1)
         {
            //scan stack again
            scanGeneration(otherGeneration, false);
            scanValues(core::gStackTop, core::gStackEnd + gParameters.stackSize - core::gStackTop, 1);
            scanValues((core::Value*)core::gTaskQueue, core::gTaskQueueSize * 2, 1);
         }
         else
            scanGeneration(otherGeneration, true);

         Block* otherGarbageBlocks = sweepGeneration(otherGeneration);
         deleteBlocks(otherGarbageBlocks);

         if(otherGeneration == 1)
         {
            //update stack again
            updateValues(core::gStackTop, core::gStackEnd + gParameters.stackSize - core::gStackTop, 1);
            updateValues((core::Value*)core::gTaskQueue, core::gTaskQueueSize * 2, 1);

            //might as well delete these now
            deleteBlocks(gGenerations[1].blocks);
            gGenerations[1].blocks = null;
         }

#ifdef INCLUDE_WRITER_CODE
         if(gDebugOutput != null && gDebugVerbosity >= 2)
         {
            gDebugOutput->write($("----------------------------------------------------------------------\n"));
            gDebugOutput->write($("After generation "));
            gDebugOutput->write(String(otherGeneration));
            gDebugOutput->write($(" sweep:\n"));
            printStatus(gDebugOutput, gDebugVerbosity);
         }
#endif //INCLUDE_WRITER_CODE
      }

#ifdef INCLUDE_WRITER_CODE
      if(gDebugOutput != null && gDebugVerbosity < 2)
         printStatus(gDebugOutput, gDebugVerbosity);
#endif //INCLUDE_WRITER_CODE
   }

   void superCollect(int depth)
   {
      if(depth <= 0)
         return;
      if(depth > 64)
         depth = 64;
      gSuperSweeping = true;

#ifdef INCLUDE_WRITER_CODE
      if(gDebugOutput != null && gDebugVerbosity >= 2)
      {
         gDebugOutput->write($("======================================================================\n"));
         gDebugOutput->write($("Before super sweep:\n"));
         printStatus(gDebugOutput, gDebugVerbosity);
      }
#endif //INCLUDE_WRITER_CODE

      //generation zero tends to be left in the UPDATE state for no particular reason.
      gGenerations[0].phase = Generation::PASSIVE;

      for(int i = 0; i < depth; i++)
         gGenerations[i].headerLocation = 0;

      //figure out how many generations are actually in use
      int generationCount = 0;
      for(int i = 0; i < depth; i++)
      {
         if(gGenerations[i].blocks != null
         || gGenerations[i].modules != null
         || gGenerations[i].processes != null)
            generationCount = i + 1;
      }

      //list of blocks which should be deleted later
      Block* oldBlocks = null;

      //clear all marks
      for(int i = 0; i < generationCount; i++)
      {
         if(gGenerations[i].phase != Generation::UPDATE)
            clearMarks(i);
      }

      //scan all
      scanValues(core::gStackTop, core::gStackEnd + gParameters.stackSize - core::gStackTop, depth - 1);
      scanValues((core::Value*)core::gTaskQueue, core::gTaskQueueSize * 2, depth - 1);
      for(int i = 0; i < generationCount; i++)
      {
         for(InternalModule* module = gGenerations[i].rootTrackModules;
             module != null; module = module->rootTrackNext)
         {
            if(module->generation >= depth)
               scanValues(&module->root, 1, depth - 1);
         }

         for(InternalProcess* process = gGenerations[i].stateTrackProcesses;
             process != null; process = process->stateTrackNext)
         {
            if(process->generation >= depth)
               scanValues(&process->state, 1, depth - 1);
         }

         for(InternalModule* module = gGenerations[i].modules;
             module != null; module = module->next)
         {
            if(!module->tracked)
            {
               markDataSegment(module->extraData, module->extraDataSize, depth - 1);
               scanValues(module->constants, module->constantCount, depth - 1);
               //no need to scan root as it is not yet set
            }
         }

         for(InternalProcess* process = gGenerations[i].processes;
             process != null; process = process->next)
         {
            if(process->keepAlive)
               scanValues(&process->state, 1, depth - 1);
         }
      }

      //map all, freeing old block data as we go
      Block* valueBlocks[generationCount];  //keep track of these so we can update later
      for(int i = 0; i < generationCount; i++)
      {
         if(gGenerations[i].phase != Generation::UPDATE)
         {
            //make an array of the blocks
            int blockCount = 0;
            for(Block* block = gGenerations[i].blocks; block != null; block = block->next)
               ++blockCount;

            if(blockCount > 0)
            {
               Block* blocks[blockCount];
               blockCount = 0;
               for(Block* block = gGenerations[i].blocks; block != null; block = block->next)
                  blocks[blockCount++] = block;

               //move them to the old blocks list
               gGenerations[i].blocks = null;
               blocks[blockCount - 1]->next = oldBlocks;
               oldBlocks = blocks[0];

               //map value segments
               {
                  int segmentCounts[blockCount];
                  int segmentCount = 0;
                  for(int j = 0; j < blockCount; j++)
                  {
                     blocks[j]->valueSegments.getSize(segmentCounts + j, null);
                     segmentCount += segmentCounts[j];
                  }

                  SegmentSet::Segment segments[segmentCount];
                  segmentCount = 0;
                  for(int j = 0; j < blockCount; j++)
                  {
                     blocks[j]->valueSegments.getSegments(segments + segmentCount);

                     int l = 0;
                     for(int k = 0; k < segmentCounts[j]; k++)
                     {
                        int start = segments[segmentCount + k].getStart();
                        while(start >= blocks[j]->fileSegments[l].start)
                           ++l;
                        --l;
                        segments[segmentCount+k].mFileSegment = blocks[j]->fileSegments + l;
                     }

                     segmentCount += segmentCounts[j];
                  }

                  mapSegmentsToBlocks(segments, segmentCount, i, false);
                  valueBlocks[i] = gGenerations[i].blocks;

                  /* I totally forgot that mapSegmentsToBlocks sorts the segment
                   * list!  It took me forever to figure out why the segments
                   * were showing up in the wrong order.  Yet another reason why
                   * values should be immutable. */
                  segmentCount = 0;
                  for(int j = 0; j < blockCount; j++)
                  {
                     blocks[j]->valueSegments.getSegments(segments + segmentCount);
                     segmentCount += segmentCounts[j];
                  }

                  segmentCount = 0;
                  for(int j = 0; j < blockCount; j++)
                  {
                     for(int k = 0; k < segmentCounts[j]; k++)
                     {
                        memcpy(segments[segmentCount].getNewLocation(),
                               blocks[j]->data + segments[segmentCount].getStart(),
                               (segments[segmentCount].getEnd() -
                                segments[segmentCount].getStart()) * sizeof(word));
                        ++segmentCount;
                     }
                  }
               }

               //map data segments
               {
                  int segmentCounts[blockCount];
                  int segmentCount = 0;
                  for(int j = 0; j < blockCount; j++)
                  {
                     blocks[j]->dataSegments.getSize(segmentCounts + j, null);
                     segmentCount += segmentCounts[j];
                  }

                  SegmentSet::Segment segments[segmentCount];
                  segmentCount = 0;
                  for(int j = 0; j < blockCount; j++)
                  {
                     blocks[j]->dataSegments.getSegments(segments + segmentCount);

                     int l = 0;
                     for(int k = 0; k < segmentCounts[j]; k++)
                     {
                        int start = segments[segmentCount + k].getStart();
                        while(start >= blocks[j]->fileSegments[l].start)
                           ++l;
                        --l;
                        segments[segmentCount+k].mFileSegment = blocks[j]->fileSegments + l;
                     }

                     segmentCount += segmentCounts[j];
                  }

                  mapSegmentsToBlocks(segments, segmentCount, i, false);

                  /* I totally forgot that mapSegmentsToBlocks sorts the segment
                   * list!  It took me forever to figure out why the segments
                   * were showing up in the wrong order.  Yet another reason why
                   * values should be immutable. */
                  segmentCount = 0;
                  for(int j = 0; j < blockCount; j++)
                  {
                     blocks[j]->dataSegments.getSegments(segments + segmentCount);
                     segmentCount += segmentCounts[j];
                  }

                  segmentCount = 0;
                  for(int j = 0; j < blockCount; j++)
                  {
                     for(int k = 0; k < segmentCounts[j]; k++)
                     {
                        memcpy(segments[segmentCount].getNewLocation(),
                               blocks[j]->data + segments[segmentCount].getStart(),
                               (segments[segmentCount].getEnd() -
                                segments[segmentCount].getStart()) * sizeof(word));
                        ++segmentCount;
                     }
                  }
               }

               //delete old data
               for(int j = 0; j < blockCount; j++)
                  freePages(blocks[j]->data, blocks[j]->size * sizeof(word));
            }
            else
               valueBlocks[i] = null;
         }
         else
         {
            /* Update-phase generation.  No need to map this data to
             * somewhere new since it is already mapped.  Just need to
             * keep its segment maps around. */
            valueBlocks[i] = null;

            if(gGenerations[i].blocks != null)
            {
               Block* block = gGenerations[i].blocks;
               while(true)
               {
                  freePages(block->data, block->size * sizeof(word));
                  if(block->next == null)
                     break;
                  block = block->next;
               }

               block->next = oldBlocks;
               oldBlocks = gGenerations[i].blocks;

               gGenerations[i].blocks = null;
            }
         }
      }

      //update everything...
      updateValues(core::gStackTop, core::gStackEnd + gParameters.stackSize - core::gStackTop, depth - 1);
      updateValues((core::Value*)core::gTaskQueue, core::gTaskQueueSize * 2, depth - 1);
      for(int i = 0; i < generationCount; i++)
      {
         if(gGenerations[i].phase != Generation::UPDATE)
         {
            for(Block* block = valueBlocks[i]; block != null; block = block->next)
            {
               updateValues((core::Value*)block->data,
                            block->size / core::WORDS_PER_VALUE,
                            depth - 1);
            }

            if(gGenerations[i].phase == Generation::MARK)
            {
               for(Block* block = valueBlocks[i]; block != null; block = block->next)
                  block->valueSegments.insert(0, block->size);
               for(Block* block = gGenerations[i].blocks; block != valueBlocks[i]; block = block->next)
                  block->dataSegments.insert(0, block->size);
            }
         }
      }

      for(int i = 0; i < generationCount; i++)
      {
         for(InternalModule** prev = &gGenerations[i].modules; *prev != null;)
         {
            InternalModule* module = *prev;

            if(module->marked || !module->tracked)
            {
               updateValues(module->constants, module->constantCount, depth - 1);
               module->extraData = updateDataSegment(module->extraData, module->extraDataSize, depth - 1);
               prev = &module->next;
               module->marked = module->tracked && (gGenerations[i].phase == Generation::MARK);

               updateValues(&module->root, 1, depth - 1);
            }
            else
            {
               *prev = module->next;

               if(module->rootTrackLink != null)
                  *module->rootTrackLink = module->rootTrackNext;
               if(module->rootTrackNext != null)
                  module->rootTrackNext->rootTrackLink = module->rootTrackLink;
               gModuleBlockMap.erase(BlockKey(module, (byte*)module + module->totalSize));
               module->~InternalModule();
               freePages(module, module->totalSize);
            }
         }

         for(InternalProcess** prev = &gGenerations[i].processes; *prev != null;)
         {
            InternalProcess* process = *prev;

            if(process->marked || process->keepAlive)
            {
               prev = &process->next;
               process->marked = !process->keepAlive && (gGenerations[i].phase == Generation::MARK);

               updateValues(&process->state, 1, depth - 1);
            }
            else
            {
               *prev = process->next;

               if(process->stateTrackLink != null)
                  *process->stateTrackLink = process->stateTrackNext;
               if(process->stateTrackNext != null)
                  process->stateTrackNext->stateTrackLink = process->stateTrackLink;
               delete process;
            }
         }
      }

      for(int i = 0; i < generationCount; i++)
      {
         for(InternalModule* module = gGenerations[i].rootTrackModules;
             module != null; module = module->rootTrackNext)
         {
            if(module->generation >= depth)
               updateValues(&module->root, 1, depth - 1);
         }

         for(InternalProcess* process = gGenerations[i].stateTrackProcesses;
             process != null; process = process->stateTrackNext)
         {
            if(process->generation >= depth)
               updateValues(&process->state, 1, depth - 1);
         }
      }

      //delete old blocks
      while(oldBlocks != null)
      {
         Block* next = oldBlocks->next;
         gBlockMap.erase(BlockKey(oldBlocks->data, (word*)oldBlocks->data + oldBlocks->size));
         delete oldBlocks;
         oldBlocks = next;
      }

#ifdef DEBUG
      if(!gBlockMap.empty())
      {
         std::map<BlockKey, Block*>::iterator iter = gBlockMap.begin();

         while(iter != gBlockMap.end())
         {
            if(iter->second->generation < depth)
               Debug::log(Debug::WARNING) << "Undeleted block in generation: " << iter->second->generation;
            ++iter;
         }
      }
#endif

      //update block map
      for(int i = 0; i < generationCount; i++)
      {
         if(gGenerations[i].phase != Generation::UPDATE)
         {
            for(Block* block = gGenerations[i].blocks; block != null; block = block->next)
               gBlockMap[BlockKey(block->data, block->data + block->size)] = block;
         }
#ifdef DEBUG
         else if(gGenerations[i].blocks != null)
            Debug::log(Debug::WARNING) << "For some reason an UPDATE generation still has blocks.";
#endif
      }

      //allocate new heap
      Block* newHeap = allocateHeap();
      newHeap->next = gGenerations[0].blocks;
      gGenerations[0].blocks = newHeap;

#ifdef INCLUDE_WRITER_CODE
      if(gDebugOutput != null && gDebugVerbosity >= 2)
      {
         gDebugOutput->write($("----------------------------------------------------------------------\n"));
         gDebugOutput->write($("After super sweep:\n"));
         printStatus(gDebugOutput, gDebugVerbosity);
      }

      if(gDebugOutput != null && gDebugVerbosity < 2)
         printStatus(gDebugOutput, gDebugVerbosity);
#endif //INCLUDE_WRITER_CODE

      gSuperSweeping = false;
   }

   // ----------------------------------------------------------------
   //debugging stuff

#ifdef INCLUDE_WRITER_CODE
   void enableDebugMessages(TextWriterPtr out, int verbosity)
   {
      gDebugOutput = out;
      gDebugVerbosity = verbosity;
   }

   void disableDebugMessages()
   {
      gDebugOutput = null;
      gDebugVerbosity = -1;
   }

   void printStatus(TextWriterPtr out, int verbosity)
   {
      std::map<BlockKey, Block*> blockMapCopy(gBlockMap);

      int blockCounts[64];
      int wordCounts[64];
      int maxGeneration = 0;

      for(int i = 0; i < 64; i++)
      {
         blockCounts[i] = 0;
         wordCounts[i] = 0;

         for(Block* block = gGenerations[i].blocks; block != null; block = block->next)
         {
            if(block->generation != i)
            {
               out->write($("ERROR:  Block claims wrong generation:  "));
               out->write(String(i));
               out->write($("-"));
               out->write(String(blockCounts[i]));
               out->write($(" (claims to be "));
               out->write(String(block->generation));
               out->write($(")\n"));
            }

            std::map<BlockKey, Block*>::iterator iter =
               blockMapCopy.find(BlockKey(block->data, block->data + block->size));
            if(iter == gBlockMap.end())
            {
               out->write($("ERROR:  Block not in block map:  "));
               out->write(String(i));
               out->write($("-"));
               out->write(String(blockCounts[i]));
               out->write($("\n"));
            }
            else if(iter->second != block)
            {
               out->write($("ERROR:  Block map lookup returns wrong block:  "));
               out->write(String(i));
               out->write($("-"));
               out->write(String(blockCounts[i]));
               out->write($("\n"));
            }
            else
               blockMapCopy.erase(iter);

            ++blockCounts[i];
            wordCounts[i] += block->size;
         }

         if(gGenerations[i].blocks != null
         || gGenerations[i].modules != null
         || gGenerations[i].processes != null)
            maxGeneration = i;
      }

      if(!blockMapCopy.empty())
      {
         out->write($("ERROR:  Memory leak:"));
         for(std::map<BlockKey, Block*>::iterator i = blockMapCopy.begin(); i != blockMapCopy.end(); ++i)
         {
            Block* block = i->second;
            out->write($("   Block not deleted:\n"));
            out->write($("      Location:   0x"));
            out->write(String((word)block->data, sizeof(word)*2, 16));
            out->write($(" - 0x"));
            out->write(String((word)(block->data + block->size), sizeof(word)*2, 16));
            out->write($("\n"));
            out->write($("      Size:       "));
            out->write(String(block->size));
            out->write($("\n"));
            out->write($("      Generation: "));
            out->write(String(block->generation));
            out->write($("\n"));
         }
      }

      if(verbosity <= 1)
      {
         out->write($("GC:  "));
         for(int i = 0; i <= maxGeneration; i++)
         {
            if(i > 0)
               out->write($(", "));
            out->write(String(blockCounts[i]));
            if(verbosity >= 1)
            {
               out->write($(" ("));
               out->write(String(wordCounts[i]));
               out->write($(")"));
            }
         }
         out->write($("\n"));
      }
      else
      {
         core::Value* stackStart = core::gStackEnd + gParameters.stackSize;
         int stackSize = stackStart - core::gStackTop;
         out->write($("Stack:\n"));
         out->write($("   Location:  0x"));
         out->write(String((word)core::gStackEnd, sizeof(word)*2, 16));
         out->write($(" - 0x"));
         out->write(String((word)stackStart, sizeof(word)*2, 16));
         out->write($("\n   Top:       0x"));
         out->write(String((word)core::gStackTop, sizeof(word)*2, 16));
         out->write($("\n   Size:      "));
         out->write(String(stackSize));
         out->write($("\n"));

         if(verbosity >= 3)
         {
            out->write($("   Data:\n"));
            for(int k = 0; k < stackSize; k++)
            {
               out->write($("      "));
               out->write(String(k, 8, 16));
               out->write($(": "));
               writeValue(out, core::gStackTop[k]);
               out->write($("\n"));
            }
         }

         out->write($("\n"));

         for(int i = 0; i <= maxGeneration; i++)
         {
            out->write($("Generation "));
            out->write(String(i));
            out->write($(":\n"));
            out->write($("   Block count: "));
            out->write(String(blockCounts[i]));
            out->write($("\n"));
            out->write($("   Word count:  "));
            out->write(String(wordCounts[i]));
            out->write($("\n"));
            out->write($("   Phase:       "));
            if(gGenerations[i].phase == Generation::PASSIVE)
               out->write($("passive\n"));
            else if(gGenerations[i].phase == Generation::MARK)
            {
               out->write($("mark "));
               out->write(String(gGenerations[i].subPhase));
               out->write($("\n"));
            }
            else
               out->write($("update\n"));

            int j = 0;
            for(InternalModule* module = gGenerations[i].modules; module != null; module = module->next)
            {
               out->write($("   Module "));
               out->write(String(j));
               out->write($(":\n"));
               out->write($("      Location:  0x"));
               out->write(String((word)(void*)module, sizeof(word)*2, 16));
               out->write($(" - 0x"));
               out->write(String((word)(void*)((byte*)module + module->totalSize), sizeof(word)*2, 16));
               out->write($("\n      ID:        "));
               out->write(String(module->id));
               out->write($("\n      RootGen:   "));
               out->write(String(module->rootTrackGeneration));
               out->write($("\n      Marked:    "));
               out->write(module->marked?$("yes"):$("no"));
               out->write($("\n      Tracked:   "));
               out->write(module->tracked?$("yes"):$("no"));
               out->write($("\n      Root:      "));
               writeValue(out, module->root);

               out->write($("\n      Constants: "));
               for(int k = 0; k < module->constantCount; k++)
               {
                  out->write($("\n         "));
                  out->write(String(k, 8, 16));
                  out->write($(": "));
                  writeValue(out, module->constants[k]);
               }

               out->write($("\n"));
               ++j;
            }

            j = 0;
            for(InternalProcess* process = gGenerations[i].processes; process != null; process = process->next)
            {
               out->write($("   Process "));
               out->write(String(j));
               out->write($(":"));
               out->write($("\n      Type:      "));
               out->write(operations::getIdentifierName(process->type));
               out->write($("\n      Location:  0x"));
               out->write(String((word)(void*)process, sizeof(word)*2, 16));
               out->write($("\n      StateGen:  "));
               out->write(String(process->stateTrackGeneration));
               out->write($("\n      Marked:    "));
               out->write(process->marked?$("yes"):$("no"));
               out->write($("\n      KeepAlive: "));
               out->write(process->keepAlive?$("yes"):$("no"));
               out->write($("\n      State:     "));
               writeValue(out, process->state);
               out->write($("\n"));
               ++j;
            }

            j = 0;
            for(Block* block = gGenerations[i].blocks; block != null; block = block->next)
            {
               out->write($("   Block "));
               out->write(String(j));
               out->write($(":\n"));
               out->write($("      Location:  0x"));
               out->write(String((word)block->data, sizeof(word)*2, 16));
               out->write($(" - 0x"));
               out->write(String((word)(block->data + block->size), sizeof(word)*2, 16));
               out->write($("\n      Size:      "));
               out->write(String(block->size));
               out->write($("\n"));

               word* data = block->data;
               int size = block->size;
               if(data + size == gHeapEnd)
               {
                  size = gHeapTop - data;

                  out->write($("      Heap top:  0x"));
                  out->write(String((word)data, sizeof(word)*2, 16));
                  out->write($("\n      Heap size: "));
                  out->write(String(size));
                  out->write($("\n"));
               }

               if(verbosity >= 3)
               {
                  out->write($("      Data:"));
                  for(int k = 0; k < size; k++)
                  {
                     if(k % 3 == 0)
                     {
                        out->write($("\n         0x"));
                        out->write(String((word)(data + k), 8, 16));
                        out->write($(": "));
                        if(k + 3 <= size)
                        {
                           writeValue(out, *((core::Value*)(data + k)));
                           out->write($("\n                     "));
                        }
                     }
                     out->write(String(data[k], 8, 16));
                     out->write($(" "));
                  }

                  out->write($("\n      Value segments:"));
                  int segmentCount;
                  block->valueSegments.getSize(&segmentCount, null);

                  {
                     SegmentSet::Segment segments[segmentCount];
                     block->valueSegments.getSegments(segments);
                     for(int k = 0; k < segmentCount; k++)
                     {
                        out->write($("\n         0x"));
                        out->write(String((word)(block->data+segments[k].getStart()), sizeof(word)*2, 16));
                        out->write($(" - 0x"));
                        out->write(String((word)(block->data+segments[k].getEnd()), sizeof(word)*2, 16));
                        out->write($(" => 0x"));
                        out->write(String((word)(segments[k].getNewLocation()), sizeof(word)*2, 16));
                     }
                  }

                  out->write($("\n      Data segments:"));
                  block->dataSegments.getSize(&segmentCount, null);

                  {
                     SegmentSet::Segment segments[segmentCount];
                     block->dataSegments.getSegments(segments);
                     for(int k = 0; k < segmentCount; k++)
                     {
                        out->write($("\n         0x"));
                        out->write(String((word)(block->data+segments[k].getStart()), sizeof(word)*2, 16));
                        out->write($(" - 0x"));
                        out->write(String((word)(block->data+segments[k].getEnd()), sizeof(word)*2, 16));
                        out->write($(" => 0x"));
                        out->write(String((word)(segments[k].getNewLocation()), sizeof(word)*2, 16));
                     }
                  }

                  out->write($("\n      File segments:"));

                  for(int k = 0; k < block->fileSegmentCount; k++)
                  {
                     out->write($("\n         0x"));
                     out->write(String((word)(block->data+block->fileSegments[k].start), sizeof(word)*2, 16));
                     out->write($(" - 0x"));
                     out->write(String((word)(block->data+block->fileSegments[k+1].start), sizeof(word)*2, 16));
                     out->write($(" => 0x"));
                     out->write(String(block->fileSegments[k].fileLocation, sizeof(word)*2, 16));
                  }

                  out->write($("\n"));
               }
               ++j;
            }

            out->write($("\n"));
         }
      }

      out->flush();
   }
#endif //INCLUDE_WRITER_CODE
};

// ===================================================================

} //namespace vm
} //namespace evlan
