/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/02/11 (KV)
 *  + First Version
 */

// ===================================================================

#ifndef evlan_vm_memory_h__INCLUDED__
#define evlan_vm_memory_h__INCLUDED__

#include "core.h"

namespace evlan {
namespace vm {
using namespace ::framework::common;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Document evlan::vm::memory
// TODO (KV): Test evlan::vm::memory

namespace memory
{
   struct Parameters
   {
      /* These values are used as hints affecting the performance of
       * the garbage collector.  They are only hints; if the value you
       * give won't work or will be horribly inefficient, it will be
       * adjusted to something better. */

      //The size of the stack, in Values
      int stackSize;
      /* The size allocated for new main heap blocks, in words.  The
       * garbage collector runs once for every time this is filled.
       * Set larger to make the collector run less often, or smaller
       * to make it run more often. */
      int heapSize;
      /* The prefered size of memory blocks allocated when crunching
       * existing blocks, in words.  Should be at least a few pages. */
      int blockSize;

      /* If this many words or more are allocated at once, they are
       * placed in a separate block rather than put on the heap. */
      int largeAllocationThreshold;
   };

   /* Initialize the garbage collector with the given parameters.
    * This will additionally set up the stack and heap pointers in the
    * core namespace. */
   void init();                      //use defaults
   void init(Parameters parameters);
   void shutdown();

   //Allocate heap space.
   word* allocateWords(int size);
   inline core::Value* allocateValues(int size)
      {return (core::Value*)allocateWords(size * core::WORDS_PER_VALUE);}
   template<typename Element>
   inline Element* allocate(int size)
   {
      if(sizeof(Element) < sizeof(word))
         return (Element*)allocateWords((size*sizeof(Element) + (sizeof(word)-1)) / sizeof(word));
      else
         return (Element*)allocateWords(size * (sizeof(Element)/sizeof(word)));
   }

   /* Return memory to the heap.  This only works if the returned
    * memory is at the very top of the heap.  That is, it must include
    * the last word of the block returned by the last call to
    * allocateWords() (or allocateValues()).  It happens that this
    * happens fairly often (in particular when packing freshly created
    * arrays).  This is basically an optimization hack. */
   void returnWords(word* target, int size);

   /* Allocate a module, including space for all tags and constants.
    * All known values (including all pointers) are set as appropriate
    * and other space is zeroed.
    *
    * After allocateModule() is returned, the caller is expected to do
    * the following:
    *  1. Produce the value representing this module, possibly by
    *     running some root function using the simulator.
    *  2. Call setModuleRoot(). */
   core::Module* allocateModule(int functionTagCount,
                                int objectTagCount,
                                int constantCount,
                                int extraDataSize,        //in words
                                const int arities[],
                                const int instructionCounts[],
                                const int functionDocSizes[],
                                const int memberCounts[],
                                const int objectDocSizes[],
                                bool hasSourceLocations = false);
   /* When setModuleRoot is called, the top of the stack should
    * contain the root value of the module.  The garbage collector
    * will copy this into the module structure (without popping it),
    * and it will insure that this root value exists as long as the
    * module is reachable.
    *
    * Do not call this multiple times for the same module.  Crashes
    * will result. */
   void setModuleRoot(core::Module* module);

   /* Find the function tag containing a given instruction.  This is
    * mainly useful for finding the location where a runtime error
    * occured. */
   core::FunctionTag* findFunction(core::Instruction* instruction);

   /* Creates a new process, using the value on the top of the stack
    * as the initial state.  That value is popped and the process is
    * pushed in its place. */
   void makeProcess(int type);
   void makeProcess(String type);
   /* Sets thet state of the given process to the value at the top of
    * the stack, which is popped. */
   void modifyProcessState(core::Process* process);

   //Run the generational collector.
   void collect();

   /* Sweep the [depth] newest generations, removing absolutely
    * everything that is not reachable. */
   void superCollect(int depth = 64);

   // ----------------------------------------------------------------
   //debugging stuff

#ifdef INCLUDE_WRITER_CODE
   /* When debug messages are enabled, printStatus() will be called
    * before and after every sweep with the given parameters. */
   void enableDebugMessages(TextWriterPtr out, int verbosity);
   void disableDebugMessages();

   /* Dumps info about the garbage collector's internal state to the
    * given output stream, in human-readable format. */
   void printStatus(TextWriterPtr out, int verbosity);
#endif //INCLUDE_WRITER_CODE
};

// ===================================================================

} //namespace vm
} //namespace evlan

#endif //included
