/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/06/28 (KV)
 *  + First Version
 */

// ===================================================================

#ifndef evlan_vm_memoryInternal_h__INCLUDED__
#define evlan_vm_memoryInternal_h__INCLUDED__

#include "SegmentSet.h"
#include "core.h"
#include "memory.h"
#include <map>

namespace evlan {
namespace vm {
using namespace ::framework::common;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Document evlan::vm::memoryInternal
// TODO (KV): Test evlan::vm::memoryInternal

//#define TRACK_DEADBLOCKS
#ifdef TRACK_DEADBLOCKS
#warning "Tracking dead blocks, which means no memory is ever reclaimed."
#endif

/* This header defines some data structures which are used internally
 * by the garbage collector.  Most code need not -- and *should not*
 * -- know about these, but the persistence code needs to use them and
 * I didn't want to put that into memory.cpp, which is big enough
 * already. */
namespace memory
{
   namespace sequence
   {
      extern uint64 counter;
      extern int max;

      int next();
   }

   struct InternalProcess: core::Process
   {
      InternalProcess* next;
      InternalProcess* stateTrackNext;
      InternalProcess** stateTrackLink;

      int generation;
      int stateTrackGeneration;

      int headerLocation;        //location of process in file; 0 = not written

      bool marked;
   };

   struct InternalModule: core::Module
   {
      ~InternalModule()
      {
         for(int i = 0; i < functionTagCount; i++)
            functionTags[i]->parameterNames = null;
      }

      InternalModule* next;
      InternalModule* rootTrackNext;
      InternalModule** rootTrackLink;

      int generation;
      int rootTrackGeneration;
      int totalSize;

      int headerLocation;        //location of module in file; 0 = not written

      bool marked;
      bool tracked;
   };

   struct FileSegment
   {
      int start;         /* Start point of the segment in the block in words.
                          * The end point is the start of the next segment.
                          * Any array of FileSegments must include a sentinel
                          * FileSegment with start = the size of the block, but
                          * this last segment should not be counted in the block's
                          * fileSegmentCount. */
      int fileLocation;  //0 = not in file
   };

   struct Block
   {
      Block* next;
      int generation;
      int size;        //in words
      word* data;
      SegmentSet dataSegments;
      SegmentSet valueSegments;

      int headerLocation;        //location of header in file; 0 = not written
      int fileSegmentCount;
      FileSegment fileSegments[]; //location of values in file, sorted

      int getFileLocation(word* pointer);
   };

   struct Generation
   {
      Block* blocks;
      InternalModule* modules;
      InternalModule* rootTrackModules;
      InternalProcess* processes;
      InternalProcess* stateTrackProcesses;

      int headerLocation;        //location of header in file; 0 = not written

      /* Each generation goes through three phases, which determine
       * how pointers to this generation's data -- encountered while
       * sweeping younger generations -- should be treated.
       *
       * When the generation is initially populated with data by a
       * sweep of the next younger generation, it enters the "passive"
       * phase.  In this phase, pointers to its data can be completely
       * ignored, as it is known that they will be encountered again
       * sometime before this generation is swept (if they still
       * exist).
       *
       * When the generation is expanded by a second sweep of the next
       * younger generation, it enters the "mark" phase.  During this
       * phase, if a pointer is encountered pointing to this
       * generation's data, the data must be marked by updating this
       * generation's segment sets.  It can be shown that *all*
       * existing pointers to this generation will be scanned between
       * the time it enters the "mark" phase and the time at which it
       * is swept, given the sweep sequence described earlier.
       *
       * After the generation is swept, it enters the "update" phase.
       * The generation's data blocks are not freed yet, as there may
       * still be pointers to them in younger generations.  During the
       * update phase, when such pointers are encountered, they must
       * be updated to point at the new location of their respective
       * data.  The "update" phase lasts just as long as the "mark"
       * phase, so it can again be shown that all pointers will be
       * encountered before this phase ends.
       *
       * The generation's data blocks are finally freed once the next
       * younger generation is swept again.  At this point, the loop
       * starts over with the "passive" phase again. */
      enum Phase
      {
         UPDATE,
         PASSIVE,
         MARK
      } phase;

      /* subPhase is an index of another generation.  When a
       * generation is in the "mark" phase, its sub-phase indicates
       * what other generation should be paying attention to it (by
       * marking its pointers).  When the "mark" phase begins, the
       * subPhase is initially set to the generation's index minus
       * two (because the generation minus one has just been swept and
       * is therefore empty anyway).  When that generation is next
       * swept, pointers in it pointing to this generation are marked
       * accordingly, and then this generation's sub-phase is
       * decremented.
       *
       * For example, when generation 6 enters the "mark" phase, it
       * sets subPhase to 4.  Then, when generation 4 is next swept,
       * it marks any data within 6 to which it encounters pointers,
       * then decrements 6's subPhase to 3.  Then, when generation 3
       * is swept, the same thing happens, and subPhase becomes 2.
       * This continues all the way down to zero.
       *
       * It can be shown that this pattern will not allow any pointers
       * to "slip through" without being marked.  At the same time,
       * this pattern delays marks as long as possible.  This is good
       * because if pointers disappear in the meantime, then they will
       * not be marked, allowing the data to which they point to be
       * swept sooner. */
      int subPhase;
   };

   /* This is a bit of a hack.  We want to create a map of memory
    * blocks to their corresponding generations.  Overlapping blocks
    * will never be inserted into this map.  When we look up a block,
    * though, we will only use a smaller segment of the block as our
    * key.  So, operator< on two BlockKeys acts like they are equal
    * any time that the two overlap.  This is not really a strict
    * ordering, but it will work for our usage patterns.
    *
    * (The alternative would be to write *yet another* custom search
    * tree structure.  I'll pass, thanks.) */
   struct BlockKey
   {
      void* start;
      void* end;

      inline bool operator<(const BlockKey& other) const
      {
         return end <= other.start;
      }

      inline BlockKey(void* start, void* end)
         : start(start), end(end) {}
   };

   /* You'll need quite an uptime to have more than 64 generations.  I
    * don't think even FreeBSD has broken a million years yet. */
   extern Generation gGenerations[64];
   extern std::map<BlockKey, Block*> gBlockMap;
   extern std::map<BlockKey, InternalModule*> gModuleBlockMap;
#ifdef TRACK_DEADBLOCKS
   extern std::map<BlockKey, Block*> gDeadBlockMap;
#endif
   /* The heap now grows upwards!  It used to grow downwards.  There
    * is really no reason for me to tell you about this change since
    * this code has still never left my computer and thus no one else
    * is actually aware of which direction the Evlan heap grows, and
    * I myself certainly am not going to forget, but for some reason
    * I'm typing this anyway.  ...  I'm going to code that way now
    * ::points::... */
   extern word* gHeapTop;
   extern word* gHeapEnd;

#ifdef INCLUDE_WRITER_CODE
   extern TextWriterPtr gDebugOutput;
   extern int gDebugVerbosity;
#endif //INCLUDE_WRITER_CODE

   void* allocatePages(int size, bool guardPage = false);
   Block* allocateBlock(int segmentCount);
   void freePages(void* memory, int size);

   void internalInit();
   void internalInit(Parameters parameters);

   extern Parameters gParameters;
}

// ===================================================================

} //namespace vm
} //namespace evlan

#endif //included
