/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/04/21 (KV)
 *  + First Version
 */

// ===================================================================

#include "messages.h"
#include <framework/messages/Message.h>
#include "memory.h"
#include "native.h"
#include "arrays.h"
#include "operations.h"
#include "simulator.h"
#include "loader.h"
#include "shortcuts.h"

namespace evlan {
namespace vm {
using framework::messages::MessageReceiver;
using framework::messages::MessagePtr;
using framework::messages::makeMessage;
using namespace shortcuts;

// ===================================================================
// AUTHOR (KV): Created module.

namespace messages
{
   EVLAN_NATIVE_IDENTIFIER(gNotATaskId, notATask);
   EVLAN_NATIVE_IDENTIFIER(gErrorId,    error);

   int gRealTaskQueueSize = 0;

   implementation TaskManager: implements MessageReceiver
   {
      TaskManager();
      ~TaskManager();

      void receive(MessagePtr message);

      MessagePtr mMessage;
      int mCurrentTask;
      bool mIsRunning;
   };

   SmartPointer<TaskManager> gTaskManager;

   TaskManager::TaskManager()
      : mCurrentTask(0), mIsRunning(false)
   {
      mMessage = makeMessage(this, 0, null);
   }
   TaskManager::~TaskManager()
   {}

   void TaskManager::receive(MessagePtr message)
   {
      if(core::gTaskQueueSize == 0)
         return;

      mIsRunning = true;

      /* Count mCurrentTask backwards because when forks occur, new
       * tasks are added to the end of the queue.  To avoid any
       * possibility of starvation, we must make sure that we
       * handle all tasks that were already in the queue first. */
      --mCurrentTask;
      if(mCurrentTask < 0 || mCurrentTask >= core::gTaskQueueSize)
         mCurrentTask = core::gTaskQueueSize - 1;

      bool deleteMe = false;
      Value& task = core::gTaskQueue[mCurrentTask].task;
      Value& context = core::gTaskQueue[mCurrentTask].context;

   repeat:
      if(task.type != Value::TASK)
      {
         arrays::makeString($("Tried to do a non-task."));

         Value* elements = memory::allocateValues(6);
         elements[0] = Value::makeAtom(gNotATaskId);
         elements[1] = Value::makeScalar(1);
         elements[2] = task;
         elements[3] = Value::makeAtom(gErrorId);
         elements[4] = Value::makeScalar(1);
         elements[5] = *gStackTop;
         simPop();

         task = Value::makeTask(Value::THROW, 6, elements);
      }

      Value* oldContext;

      switch(task.task.taskType)
      {
      default:
      case Value::STOP:
         deleteMe = true;
         break;

      case Value::CALL:
      case Value::TRY:
      case Value::FINALLY:
         oldContext = memory::allocateValues(1);
         *oldContext = context;

         context.context.contextType = task.task.taskType;
         context.context.parameter = task.task.parameters + 1;
         context.context.parent = oldContext;

         task = task.task.parameters[0];
         goto repeat;  //efficiency hack

      case Value::RETURN:
         while(context.context.contextType == Value::TRY
            || (context.context.contextType == Value::THROW
             && !context.context.throwOnReturn))
            context = *context.context.parent;

         switch(context.context.contextType)
         {
         default:
         case Value::STOP:
            //nothing to return to
            deleteMe = true;
            break;

         case Value::CALL:
            simPush(*context.context.parameter);
            context = *context.context.parent;

            if(gStackTop->hasArity(1))
            {
               if(task.task.parameters == null)
                  simPush(Value::makeEmpty());
               else
                  simPush(*task.task.parameters);

               simApply(1);
            }
            else
               simApply(0);

            core::gTaskQueue[mCurrentTask].task = *gStackTop;
            simPop();
            break;

         case Value::RETURN:
            if(context.context.parameter == null)
               task = Value::makeTask(Value::RETURN);
            else
               task = Value::makeTask(Value::RETURN, 1, context.context.parameter);
            context = *context.context.parent;
            goto repeat;  //efficiency hack

         case Value::FINALLY:
            simPush(*context.context.parameter);
            context.context.contextType = Value::RETURN;
            if(task.task.parameterCount == 0)
               context.context.parameter = null;
            else
               context.context.parameter = task.task.parameters;
            task = *gStackTop;
            simPop();
            goto repeat;  //efficiency hack

         case Value::THROW:
            //throwOnReturn is true, so...
            task = Value::makeTask(Value::THROW, 0, null);
            goto repeat;  //efficiency hack
         }
         break;

      case Value::THROW:
         simPush(context);  //save stack trace...

         {
            while(task.task.parameterCount == 0)
            {
               //want to re-throw...
               while(context.context.contextType != Value::THROW
                  && context.context.contextType != Value::FINALLY
                  && context.context.contextType != Value::TRY
                  && context.context.contextType != Value::STOP)
                  context = *context.context.parent;

               if(context.context.contextType == Value::THROW)
               {
                  *gStackTop = context.context.parameter[1];
                  task.task.parameters = context.context.parameter[0].array.elements;
                  task.task.parameterCount = context.context.parameter[0].array.size;
                  context = *context.context.parent;
               }
               else
                  break;
            }

            int pos = 0;
            int typeCount = 0;
            while(pos < task.task.parameterCount)
            {
               pos += 2 + (int)task.task.parameters[pos+1].builtin.value.scalar;
               ++typeCount;
            }

            int ids[typeCount];
            int parameterCounts[typeCount];
            int positions[typeCount];
            pos = 0;
            typeCount = 0;

            while(pos < task.task.parameterCount)
            {
               ids[typeCount] = task.task.parameters[pos].builtin.value.data32;
               parameterCounts[typeCount] = (int)task.task.parameters[pos+1].builtin.value.scalar;
               positions[typeCount] = pos + 2;
               pos += 2 + parameterCounts[typeCount];
               ++typeCount;
            }

            int typeIndex = -1;

            while(typeIndex == -1)
            {
               while(context.context.contextType != Value::TRY
                  && context.context.contextType != Value::FINALLY
                  && context.context.contextType != Value::STOP)
                  context = *context.context.parent;

               if(context.context.contextType == Value::STOP)
                  break;

               if(context.context.contextType == Value::FINALLY)
               {
                  simPush(*context.context.parameter);
                  typeIndex = -3;
                  break;
               }

               core::ObjectTag* tag = context.context.parameter->object.tag;
               for(int i = 0; i < typeCount; i++)
               {
                  int index = operations::getMemberIndex(tag, ids[i]);
                  if(index != -1)
                  {
                     simPush(context.context.parameter->object.members[index]);
                     typeIndex = i;
                     break;
                  }
               }

               if(typeIndex == -1)
               {
                  int index = operations::getMemberIndex(tag, operations::BUILTIN_ALL);
                  if(index != -1)
                  {
                     simPush(context.context.parameter->object.members[index]);
                     typeIndex = -2;
                  }
               }

               if(typeIndex == -1)
                  context = *context.context.parent;
            }

            if(typeIndex == -1)
            {
               Debug::log(Debug::ERROR) << "Uncaught Evlan exception.";
               deleteMe = true;
            }
            else
            {
               Value* newParameters = memory::allocateValues(2);
               newParameters[0] = Value::makeArray(task.task.parameterCount, task.task.parameters);
               newParameters[1] = gStackTop[1];
               context.context.contextType = Value::THROW;
               context.context.throwOnReturn = false;
               context.context.parameter = newParameters;

               if(typeIndex == -2)
                  simApply(0);
               else if(typeIndex == -3)
                  context.context.throwOnReturn = true;
               else
               {
                  for(int i = 0; i < parameterCounts[typeIndex]; i++)
                     simPush(task.task.parameters[positions[typeIndex] + i]);

                  simApply(parameterCounts[typeIndex]);
               }

               core::gTaskQueue[mCurrentTask].task = *gStackTop;
               simPop();
            }
         }

         simPop();
         break;
      }

      if(deleteMe)
      {
         --core::gTaskQueueSize;
         if(mCurrentTask < core::gTaskQueueSize)
            core::gTaskQueue[mCurrentTask] = core::gTaskQueue[core::gTaskQueueSize];
      }

      if(core::gTaskQueueSize > 0)
         mMessage->send();
      mIsRunning = false;
   }

   void init(int taskCount)
   {
      if(gTaskManager == null)
         gTaskManager = spnew TaskManager;

      gRealTaskQueueSize = 0;
      core::gTaskQueueSize = 0;

      if(taskCount > 0)
      {
         gRealTaskQueueSize = 1;
         while(gRealTaskQueueSize < taskCount)
            gRealTaskQueueSize <<= 1;
         core::gTaskQueue = new core::Task[gRealTaskQueueSize];
         core::gTaskQueueSize = taskCount;
         gTaskManager->mMessage->send();
      }
   }
   void shutdown()
   {
      gTaskManager = null;
      if(core::gTaskQueue != null)
      {
         delete [] core::gTaskQueue;
         core::gTaskQueue = null;
      }
   }

   void queueTask()
   {
      if(gStackTop[1].type != Value::CONTEXT)
         FW_THROW(spnew FatalError($("Called queueTask() with something other than a context in stack[1].")));

      if(core::gTaskQueueSize == gRealTaskQueueSize)
      {
         if(gRealTaskQueueSize == 0)
         {
            gRealTaskQueueSize = 1;
            core::gTaskQueue = new core::Task[1];
         }
         else
         {
            gRealTaskQueueSize *= 2;
            core::Task* newQueue = new core::Task[gRealTaskQueueSize];
            memcpy(newQueue, core::gTaskQueue, sizeof(core::Task) * core::gTaskQueueSize);
            delete [] core::gTaskQueue;
            core::gTaskQueue = newQueue;
         }
      }

      if(core::gTaskQueueSize == 0)
         gTaskManager->mMessage->send();
      core::gTaskQueue[core::gTaskQueueSize].task = gStackTop[0];
      core::gTaskQueue[core::gTaskQueueSize].context = gStackTop[1];
      ++core::gTaskQueueSize;
      simPop(2);
   }

   void getCurrentContext()
   {
      if(gTaskManager->mIsRunning)
         simPush(core::gTaskQueue[gTaskManager->mCurrentTask].context);
      else
         simPush(Value::makeContext(Value::STOP, null, null));
   }

   EVLAN_NATIVE_FUNCTION_2(taskCall, body, next)
   {
      if(!next.hasArity(0) && !next.hasArity(1))
         return simTypeError(next);

      Value* elements = memory::allocateValues(2);
      elements[0] = body;
      elements[1] = next;
      return Value::makeTask(Value::CALL, 2, elements);
   }

   EVLAN_NATIVE_FUNCTION_2(taskFinally, body, handler)
   {
      Value* elements = memory::allocateValues(2);
      elements[0] = body;
      elements[1] = handler;
      return Value::makeTask(Value::FINALLY, 2, elements);
   }

   EVLAN_NATIVE_FUNCTION_1(taskReturn, parameter)
   {
      Value* elements = memory::allocateValues(1);
      elements[0] = parameter;
      return Value::makeTask(Value::RETURN, 1, elements);
   }

   EVLAN_NATIVE_FUNCTION_0(taskReturnNone)
   {
      return Value::makeTask(Value::RETURN);
   }

   EVLAN_NATIVE_FUNCTION_2(taskTry, body, handlers)
   {
      if(!handlers.isObject())
         return simTypeError(handlers);

      for(int i = 0; i < handlers.object.tag->memberCount; i++)
      {
         if(!handlers.object.members[i].isFunction())
            return simTypeError(handlers.object.members[i]);
      }

      Value* elements = memory::allocateValues(2);
      elements[0] = body;
      elements[1] = handlers;
      return Value::makeTask(Value::TRY, 2, elements);
   }

   EVLAN_NATIVE_FUNCTION_1(taskThrow, descriptor)
   {
      if(!descriptor.isArray())
         return simTypeError(descriptor);

      int pos = 0;
      while(pos < descriptor.array.size)
      {
         if((pos + 1 == descriptor.array.size)
         || !descriptor.array.elements[pos].isAtom()
         || !descriptor.array.elements[pos+1].isInteger())
            return simTypeError();
         int parameterCount = (int)descriptor.array.elements[pos+1].builtin.value.scalar;
         pos += 2 + parameterCount;
         if(pos > descriptor.array.size)
            return simTypeError();
      }

      return Value::makeTask(Value::THROW, descriptor.array.size, descriptor.array.elements);
   }

   EVLAN_NATIVE_FUNCTION_0(taskStop)
   {
      return Value::makeTask(Value::STOP);
   }

   EVLAN_NATIVE_FUNCTION_1(taskGetContext, result)
   {
      getCurrentContext();
      Value* elements = memory::allocateValues(2);
      elements[0] = *gStackTop;
      elements[1] = result;
      return simReturn(Value::makeArray(2, elements));
   }

   EVLAN_NATIVE_FUNCTION_2(taskRunContext, context, result)
   {
      if(!context.isContext())
         return simTypeError(context);

      Value* elements = memory::allocateValues(2);
      elements[0] = context;
      elements[1] = result;

      simPush(context);
      simPush(simReturn(Value::makeArray(2, elements)));
      queueTask();

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(taskGetException, context)
   {
      if(!context.isContext())
         return simTypeError(context);

      while(context.context.contextType != Value::THROW
         && context.context.contextType != Value::STOP)
         context = *context.context.parent;

      if(context.context.contextType == Value::THROW)
         return Value::makeArray(2, context.context.parameter);
      else
         return Value::makeArray(0, null);
   }

   EVLAN_NATIVE_IDENTIFIER(gCallId,    call);
   EVLAN_NATIVE_IDENTIFIER(gTryId,     try);
   EVLAN_NATIVE_IDENTIFIER(gThrowId,   throw);
   EVLAN_NATIVE_IDENTIFIER(gFinallyId, finally);
   EVLAN_NATIVE_IDENTIFIER(gReturnId,  return);
   EVLAN_NATIVE_IDENTIFIER(gStopId,    stop);

   EVLAN_NATIVE_FUNCTION_1(taskDescribeContext, context)
   {
      if(!context.isContext())
         return simTypeError(context);

      Value* elements;

      switch(context.context.contextType)
      {
      case Value::CALL:
         elements = memory::allocateValues(3);
         elements[0] = Value::makeAtom(gCallId);
         elements[1] = *context.context.parameter;
         elements[2] = *context.context.parent;
         return Value::makeArray(3, elements);
      case Value::TRY:
         elements = memory::allocateValues(3);
         elements[0] = Value::makeAtom(gTryId);
         elements[1] = *context.context.parameter;
         elements[2] = *context.context.parent;
         return Value::makeArray(3, elements);
      case Value::THROW:
         elements = memory::allocateValues(3);
         elements[0] = Value::makeAtom(gThrowId);
         elements[1] = Value::makeArray(2, context.context.parameter);
         elements[2] = *context.context.parent;
         return Value::makeArray(3, elements);
      case Value::FINALLY:
         elements = memory::allocateValues(3);
         elements[0] = Value::makeAtom(gFinallyId);
         elements[1] = *context.context.parameter;
         elements[2] = *context.context.parent;
         return Value::makeArray(3, elements);
      case Value::RETURN:
         elements = memory::allocateValues(3);
         elements[0] = Value::makeAtom(gReturnId);
         if(context.context.parameter == null)
            elements[1] = Value::makeEmpty();
         else
            elements[1] = *context.context.parameter;
         elements[2] = *context.context.parent;
         return Value::makeArray(3, elements);
      default:
      case Value::STOP:
         elements = memory::allocateValues(1);
         elements[0] = Value::makeAtom(gStopId);
         return Value::makeArray(1, elements);
      }
   }
};

// ===================================================================

} //namespace vm
} //namespace evlan
