/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/03/18 (KV)
 *  + First Version
 */

// ===================================================================

#include "native.h"
#include "memory.h"
#include "operations.h"
#include <algorithm>
#include <map>

namespace evlan {
namespace vm {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::vm::native
// TODO (KV): Review evlan::vm::native

namespace native
{
   std::map<int, core::NativeFunctionTag*> gNativeFunctions;
   std::map<int, ProcessConstructor> gNativeProcesses;

   bool gIsInitialized;
   NativeIdentifier* gNativeIdentifierList;
   NativeFunction* gNativeFunctionList;
   NativeProcess* gNativeProcessList;

   ObjectPtr restoreSystemProcess(core::Process* process, framework::SystemPtr system)
   {
      return system;
   }

   //Global initialization order in C++ is such a pain to get right...
   void internalInit()
   {
      static bool isFirstCall = true;
      if(isFirstCall)
      {
         gIsInitialized = false;
         gNativeIdentifierList = null;
         gNativeFunctionList = null;
         gNativeProcessList = null;
         isFirstCall = false;
      }
   }

   void init()
   {
      internalInit();

      {
         NativeIdentifier* list = gNativeIdentifierList;
         while(list != null)
         {
            *list->mVariable = operations::getIdentifierId(list->mIdentifier);
            list = list->mNext;
         }
      }


      {
         NativeFunction* list = gNativeFunctionList;
         while(list != null)
         {
            core::NativeFunctionTag* tag = new core::NativeFunctionTag;
            tag->identifier = list->mIdentifier;
            tag->arity = list->mArity;
            tag->parameterNames = list->mParameterNames;
            tag->nativeApply = list->mNativeApply;
            gNativeFunctions[operations::getIdentifierId(list->mIdentifier)] = tag;

            list = list->mNext;
         }
      }

      {
         gNativeProcesses[operations::getIdentifierId($("system"))] = &restoreSystemProcess;

         NativeProcess* list = gNativeProcessList;
         while(list != null)
         {
            gNativeProcesses[operations::getIdentifierId(list->mIdentifier)] = list->mMake;
            list = list->mNext;
         }
      }

      gIsInitialized = true;
   }

   void shutdown()
   {
      std::map<int, core::NativeFunctionTag*>::iterator iter = gNativeFunctions.begin();
      for(; iter != gNativeFunctions.end(); ++iter)
         delete iter->second;
      gNativeFunctions.clear();
      gNativeProcesses.clear();
      gIsInitialized = false;
   }

   Array<const String> getList()
   {
      Array<String> result(gNativeFunctions.size());

      std::map<int, core::NativeFunctionTag*>::iterator iter = gNativeFunctions.begin();
      for(int i = 0; iter != gNativeFunctions.end(); ++iter, ++i)
         result[i] = operations::getIdentifierName(iter->first);

      return result.toConst();
   }

   Array<const int> getIdList()
   {
      Array<int> result(gNativeFunctions.size());

      std::map<int, core::NativeFunctionTag*>::iterator iter = gNativeFunctions.begin();
      for(int i = 0; iter != gNativeFunctions.end(); ++iter, ++i)
         result[i] = iter->first;

      return result.toConst();
   }

   NativeIdentifier::NativeIdentifier(String identifier, uint32* variable)
      : mIdentifier(identifier),
        mVariable(variable)
   {
      internalInit();

      mNext = gNativeIdentifierList;
      mPrevious = &gNativeIdentifierList;

      gNativeIdentifierList = this;
      if(mNext != null)
         mNext->mPrevious = &mNext;
   }

   NativeIdentifier::~NativeIdentifier()
   {
      *mPrevious = mNext;
      if(mNext != null)
         mNext->mPrevious = mPrevious;
   }

   NativeFunction::NativeFunction(String identifier, int arity,
                                  void (*nativeApply)(),
                                  Array<const String> parameterNames)
      : mIdentifier(identifier),
        mArity(arity),
        mNativeApply(nativeApply),
        mParameterNames(parameterNames)
   {
      internalInit();

      mNext = gNativeFunctionList;
      mPrevious = &gNativeFunctionList;

      gNativeFunctionList = this;
      if(mNext != null)
         mNext->mPrevious = &mNext;
   }

   NativeFunction::~NativeFunction()
   {
      *mPrevious = mNext;
      if(mNext != null)
         mNext->mPrevious = mPrevious;
   }

   core::NativeFunctionTag* getFunction(int id)
   {
      std::map<int, core::NativeFunctionTag*>::iterator iter =
         gNativeFunctions.find(id);
      if(iter == gNativeFunctions.end())
         return null;
      else
         return iter->second;
   }


   NativeProcess::NativeProcess(String identifier, ProcessConstructor make)
      : mIdentifier(identifier),
        mMake(make)
   {
      internalInit();

      mNext = gNativeProcessList;
      mPrevious = &gNativeProcessList;

      gNativeProcessList = this;
      if(mNext != null)
         mNext->mPrevious = &mNext;
   }

   NativeProcess::~NativeProcess()
   {
      *mPrevious = mNext;
      if(mNext != null)
         mNext->mPrevious = mPrevious;
   }

   ProcessConstructor getProcess(int id)
   {
      std::map<int, ProcessConstructor>::iterator iter =
         gNativeProcesses.find(id);
      if(iter == gNativeProcesses.end())
         return null;
      else
         return iter->second;
   }


   void makeSystemProcess(framework::SystemPtr system)
   {
      --core::gStackTop;
      core::gStackTop->type = core::Value::EMPTY;
      memory::makeProcess($("system"));
      core::gStackTop->process.process->nativeInfo = system;
   }
}

// ===================================================================

} //namespace vm
} //namespace evlan
