/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/04/01 (KV)
 *  + First Version
 */

// ===================================================================

#include "operations.h"
#include <framework/util/Vector.h>
#include <map>
#include <math.h>
#include "native.h"

namespace evlan {
namespace vm {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::vm::operations
// TODO (KV): Review evlan::vm::operations

namespace operations
{
   util::Vector<String> gIdentifierNames;
   std::map<String, int> gIdentifierMap;

   String BUILTIN_NAMES[] =
   {
      $("not"),
      $("negate"),
      $("size"),
      $("arity"),

      $("and"),
      $("or"),
      $("add"),
      $("subtract"),
      $("multiply"),
      $("divide"),
      $("modulus"),
      $("raiseToPower"),
      $("equals"),
      $("isLessThan"),
      $("isGreaterThan"),
      $("getElement"),

      $("all"),

      /* These are just here because I wanted them to appear in the
       * right order when displaying the contents of an object. */
      $("key"),
      $("value")
   };

   void init()
   {
      if(gIdentifierNames.getSize() == 0)
      {
         for(int i = 0; i < (int)(sizeof(BUILTIN_NAMES) / sizeof(String)); i++)
         {
            gIdentifierNames.pushBack(BUILTIN_NAMES[i]);
            gIdentifierMap[BUILTIN_NAMES[i]] = i;
         }
      }
   }

   void init(Array<const String> identifiers)
   {
      for(int i = 0; i < identifiers.getSize(); i++)
      {
         gIdentifierNames.pushBack(identifiers[i]);
         gIdentifierMap[identifiers[i]] = i;
      }
   }

   void shutdown()
   {
      gIdentifierNames.clear();
      gIdentifierMap.clear();
   }

   int getIdentifierId(String name)
   {
      std::map<String, int>::iterator iter = gIdentifierMap.find(name);
      if(iter == gIdentifierMap.end())
      {
         int result = gIdentifierNames.getSize();
         gIdentifierNames.pushBack(name);
         gIdentifierMap[name] = result;
         return result;
      }
      else
         return iter->second;
   }

   String getIdentifierName(int id)
   {
      return gIdentifierNames[id];
   }

   Array<const String> getIdentifierNames()
   {
      return gIdentifierNames.getArray().toConst();
   }

   void getMember(String memberName)
   {
      getMember(getMemberNameId(memberName));
   }

   void getMember(int id)
   {
      switch(core::gStackTop->type)
      {
      case core::Value::RUNTIME_ERROR:
         return;

      case core::Value::NATIVE_CONTAINER:
         core::gStackTop->nativeFunction.tag = native::getFunction(id);
         if(core::gStackTop->nativeFunction.tag != null)
         {
            core::gStackTop->nativeFunction.id = id;
            core::gStackTop->type = core::Value::NATIVE_FUNCTION;
         }
         else
            *core::gStackTop = core::Value::makeError(core::Value::NO_SUCH_MEMBER, null);
         return;

      case core::Value::BUILTIN_OBJECT:
      case core::Value::ARRAY:
      case core::Value::PACKED_ARRAY:
         //FIXME:  Do I really care to allow things like "increment = 1.add"?
         *core::gStackTop = core::Value::makeError(core::Value::NO_SUCH_MEMBER, null);
         return;

      case core::Value::NATIVE_FUNCTION:
         if(id == BUILTIN_ARITY)
            *core::gStackTop = core::Value::makeScalar(core::gStackTop->nativeFunction.tag->arity);
         else
            *core::gStackTop = core::Value::makeError(core::Value::NOT_AN_OBJECT, null);
         return;

      case core::Value::FUNCTION:
         if(id == BUILTIN_ARITY)
            *core::gStackTop = core::Value::makeScalar(core::gStackTop->function.tag->arity);
         else
            *core::gStackTop = core::Value::makeError(core::Value::NOT_AN_OBJECT, null);
         return;

      case core::Value::OBJECT:
         {
            core::ObjectTag* tag = core::gStackTop->object.tag;

            int a = 0, b = tag->memberCount;

            while(a < b)
            {
               int midpoint = (a + b) / 2;
               if(id < tag->members[midpoint].nameId)
                  b = midpoint;
               else if(id > tag->members[midpoint].nameId)
                  a = midpoint + 1;
               else
               {
                  //found it
                  *core::gStackTop = core::gStackTop->object.members[tag->members[midpoint].offset];
                  return;
               }
            }

            *core::gStackTop = core::Value::makeError(core::Value::NO_SUCH_MEMBER, null);
            return;
         }

      default:
         *core::gStackTop = core::Value::makeError(core::Value::NOT_AN_OBJECT, null);
         return;
      }
   }

   int getMemberIndex(core::ObjectTag* tag, int id)
   {
      int a = 0, b = tag->memberCount;

      while(a < b)
      {
         int midpoint = (a + b) / 2;
         if(id < tag->members[midpoint].nameId)
            b = midpoint;
         else if(id > tag->members[midpoint].nameId)
            a = midpoint + 1;
         else
         {
            //found it
            return tag->members[midpoint].offset;
         }
      }

      return -1;
   }


   void doNot()
   {
      if(core::gStackTop->type == core::Value::BUILTIN_OBJECT
      && core::gStackTop->builtin.builtinType == core::Value::BOOLEAN)
            core::gStackTop->builtin.value.boolean = !core::gStackTop->builtin.value.boolean;
      else if(!core::gStackTop->isError())
         *core::gStackTop = core::Value::makeError(core::Value::TYPE_ERROR, null);
   }

   void doNegate()
   {
      if(core::gStackTop->isScalar())
         core::gStackTop->builtin.value.scalar = -core::gStackTop->builtin.value.scalar;
      else if(!core::gStackTop->isError())
         *core::gStackTop = core::Value::makeError(core::Value::TYPE_ERROR, null);
   }

   void binaryTypeError()
   {
      if(!core::gStackTop[1].isError())
      {
         if(core::gStackTop[0].isError())
            core::gStackTop[1] = core::gStackTop[0];
         else
            core::gStackTop[1] = core::Value::makeError(core::Value::TYPE_ERROR, null);
      }
      ++core::gStackTop;
   }

   void doAnd()
   {
      if(core::gStackTop[1].isBoolean())
      {
         if(!core::gStackTop[1].builtin.value.boolean)
            ++core::gStackTop;
         else if(core::gStackTop[0].isBoolean())
         {
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[0].builtin.value.boolean;
            ++core::gStackTop;
         }
         else
            binaryTypeError();
      }
      else if(core::gStackTop[0].isBoolean()
           && !core::gStackTop[0].builtin.value.boolean)
      {
         core::gStackTop[1].type = core::Value::BUILTIN_OBJECT;
         core::gStackTop[1].builtin.builtinType = core::Value::BOOLEAN;
         core::gStackTop[1].builtin.value.boolean = false;
         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }

   void doOr()
   {
      if(core::gStackTop[1].isBoolean())
      {
         if(core::gStackTop[1].builtin.value.boolean)
            ++core::gStackTop;
         else if(core::gStackTop[0].isBoolean())
         {
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[0].builtin.value.boolean;
            ++core::gStackTop;
         }
         else
            binaryTypeError();
      }
      else if(core::gStackTop[0].isBoolean()
           && core::gStackTop[0].builtin.value.boolean)
      {
         core::gStackTop[1].type = core::Value::BUILTIN_OBJECT;
         core::gStackTop[1].builtin.builtinType = core::Value::BOOLEAN;
         core::gStackTop[1].builtin.value.boolean = true;
         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }

   void doAdd()
   {
      if(core::gStackTop[1].isScalar()
      && core::gStackTop[0].isScalar())
      {
         core::gStackTop[1].builtin.value.scalar =
            core::gStackTop[1].builtin.value.scalar +
            core::gStackTop[0].builtin.value.scalar;
         ++core::gStackTop;
      }
      else if(core::gStackTop[1].isCharacter()
           && core::gStackTop[0].isScalar())
      {
         int intval = (int)core::gStackTop[0].builtin.value.scalar;
         if(core::gStackTop[0].builtin.value.scalar == (double)intval)
         {
            core::gStackTop[1].builtin.value.data32 =
               core::gStackTop[1].builtin.value.data32 + intval;
            ++core::gStackTop;
         }
         else
            binaryTypeError();
      }
      else if(core::gStackTop[1].isScalar()
           && core::gStackTop[0].isCharacter())
      {
         int intval = (int)core::gStackTop[1].builtin.value.scalar;
         if(core::gStackTop[1].builtin.value.scalar == (double)intval)
         {
            core::gStackTop[1].builtin.builtinType = core::Value::CHARACTER;
            core::gStackTop[1].builtin.value.data32 =
               intval + core::gStackTop[0].builtin.value.data32;
            ++core::gStackTop;
         }
         else
            binaryTypeError();
      }
      else
         binaryTypeError();
   }

   void doSubtract()
   {
      if(core::gStackTop[1].isScalar()
      && core::gStackTop[0].isScalar())
      {
         core::gStackTop[1].builtin.value.scalar =
            core::gStackTop[1].builtin.value.scalar -
            core::gStackTop[0].builtin.value.scalar;
         ++core::gStackTop;
      }
      else if(core::gStackTop[1].isCharacter()
           && core::gStackTop[0].isScalar())
      {
         int intval = (int)core::gStackTop[0].builtin.value.scalar;
         if(core::gStackTop[0].builtin.value.scalar == (double)intval)
         {
            core::gStackTop[1].builtin.value.data32 =
               core::gStackTop[1].builtin.value.data32 - intval;
            ++core::gStackTop;
         }
         else
            binaryTypeError();
      }
      else if(core::gStackTop[1].isCharacter()
           && core::gStackTop[0].isCharacter())
      {
         core::gStackTop[1].builtin.value.scalar =
            core::gStackTop[1].builtin.value.data32 -
            core::gStackTop[0].builtin.value.data32;
         core::gStackTop[1].builtin.builtinType = core::Value::SCALAR;
         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }

   void doMultiply()
   {
      if(core::gStackTop[1].isScalar()
      && core::gStackTop[0].isScalar())
      {
         core::gStackTop[1].builtin.value.scalar =
            core::gStackTop[1].builtin.value.scalar *
            core::gStackTop[0].builtin.value.scalar;
         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }

   void doDivide()
   {
      if(core::gStackTop[1].isScalar()
      && core::gStackTop[0].isScalar())
      {
         core::gStackTop[1].builtin.value.scalar =
            core::gStackTop[1].builtin.value.scalar /
            core::gStackTop[0].builtin.value.scalar;
         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }

   void doModulus()
   {
      if(core::gStackTop[1].isScalar()
      && core::gStackTop[0].isScalar())
      {
         /* if the denominator is infinite...  shouldn't the result
          * always be just the numerator?  Why does it return nan? */
         if(isfinite(core::gStackTop[0].builtin.value.scalar))
         {
            /* Modulus is defined differently in Evlan than in C!
             *    modulus = (x, y) => x - Scalar.floor(x/y) * y
             * This form tends to be far more useful in practice. */
            double result = ::fmod(
               core::gStackTop[1].builtin.value.scalar,
               core::gStackTop[0].builtin.value.scalar);

            if(core::gStackTop[1].builtin.value.scalar
             * core::gStackTop[0].builtin.value.scalar < 0
            && result != 0)
               result += core::gStackTop[0].builtin.value.scalar;

            core::gStackTop[1].builtin.value.scalar = result;
         }

         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }

   void doPower()
   {
      if(core::gStackTop[1].isScalar()
      && core::gStackTop[0].isScalar())
      {
         core::gStackTop[1].builtin.value.scalar = ::pow(
            core::gStackTop[1].builtin.value.scalar,
            core::gStackTop[0].builtin.value.scalar);
         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }

   void doEquals()
   {
      if(core::gStackTop[1].isBuiltin()
      && core::gStackTop[0].isBuiltin()
      && core::gStackTop[1].builtin.builtinType == core::gStackTop[0].builtin.builtinType)
      {
         switch(core::gStackTop[1].builtin.builtinType)
         {
         case core::Value::BOOLEAN:
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[1].builtin.value.boolean ==
               core::gStackTop[0].builtin.value.boolean;
            break;
         case core::Value::SCALAR:
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[1].builtin.value.scalar ==
               core::gStackTop[0].builtin.value.scalar;
            break;
         case core::Value::CHARACTER:
         case core::Value::ATOM:
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[1].builtin.value.data32 ==
               core::gStackTop[0].builtin.value.data32;
            break;
         default:
            {
               int size = core::gStackTop[1].builtin.builtinType - core::Value::DATA1;
               if(size == 6)
               {
                  core::gStackTop[1].builtin.value.boolean =
                     core::gStackTop[1].builtin.value.data64 ==
                     core::gStackTop[0].builtin.value.data64;
               }
               else if(size == 5)
               {
                  core::gStackTop[1].builtin.value.boolean =
                     core::gStackTop[1].builtin.value.data32 ==
                     core::gStackTop[0].builtin.value.data32;
               }
               else
               {
                  uint32 mask = (1 << (1 << size)) - 1;
                  core::gStackTop[1].builtin.value.boolean =
                     (core::gStackTop[1].builtin.value.data32 & mask) ==
                     (core::gStackTop[0].builtin.value.data32 & mask);
               }
            }
            break;
         }

         core::gStackTop[1].builtin.builtinType = core::Value::BOOLEAN;
         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }

   void doIsLessThan()
   {
      if(core::gStackTop[1].isBuiltin()
      && core::gStackTop[0].isBuiltin()
      && core::gStackTop[1].builtin.builtinType == core::gStackTop[0].builtin.builtinType)
      {
         switch(core::gStackTop[1].builtin.builtinType)
         {
         case core::Value::BOOLEAN:
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[1].builtin.value.boolean <
               core::gStackTop[0].builtin.value.boolean;
            break;
         case core::Value::SCALAR:
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[1].builtin.value.scalar <
               core::gStackTop[0].builtin.value.scalar;
            break;
         case core::Value::CHARACTER:
         case core::Value::ATOM:
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[1].builtin.value.data32 <
               core::gStackTop[0].builtin.value.data32;
            break;
         default:
            {
               int size = core::gStackTop[1].builtin.builtinType - core::Value::DATA1;
               if(size == 6)
               {
                  core::gStackTop[1].builtin.value.boolean =
                     core::gStackTop[1].builtin.value.data64 <
                     core::gStackTop[0].builtin.value.data64;
               }
               else if(size == 5)
               {
                  core::gStackTop[1].builtin.value.boolean =
                     core::gStackTop[1].builtin.value.data32 <
                     core::gStackTop[0].builtin.value.data32;
               }
               else
               {
                  uint32 mask = (1 << (1 << size)) - 1;
                  core::gStackTop[1].builtin.value.boolean =
                     (core::gStackTop[1].builtin.value.data32 & mask) <
                     (core::gStackTop[0].builtin.value.data32 & mask);
               }
            }
            break;
         }

         core::gStackTop[1].builtin.builtinType = core::Value::BOOLEAN;
         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }

   void doIsGreaterThan()
   {
      if(core::gStackTop[1].isBuiltin()
      && core::gStackTop[0].isBuiltin()
      && core::gStackTop[1].builtin.builtinType == core::gStackTop[0].builtin.builtinType)
      {
         switch(core::gStackTop[1].builtin.builtinType)
         {
         case core::Value::BOOLEAN:
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[1].builtin.value.boolean >
               core::gStackTop[0].builtin.value.boolean;
            break;
         case core::Value::SCALAR:
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[1].builtin.value.scalar >
               core::gStackTop[0].builtin.value.scalar;
            break;
         case core::Value::CHARACTER:
         case core::Value::ATOM:
            core::gStackTop[1].builtin.value.boolean =
               core::gStackTop[1].builtin.value.data32 >
               core::gStackTop[0].builtin.value.data32;
            break;
         default:
            {
               int size = core::gStackTop[1].builtin.builtinType - core::Value::DATA1;
               if(size == 6)
               {
                  core::gStackTop[1].builtin.value.boolean =
                     core::gStackTop[1].builtin.value.data64 >
                     core::gStackTop[0].builtin.value.data64;
               }
               else if(size == 5)
               {
                  core::gStackTop[1].builtin.value.boolean =
                     core::gStackTop[1].builtin.value.data32 >
                     core::gStackTop[0].builtin.value.data32;
               }
               else
               {
                  uint32 mask = (1 << (1 << size)) - 1;
                  core::gStackTop[1].builtin.value.boolean =
                     (core::gStackTop[1].builtin.value.data32 & mask) >
                     (core::gStackTop[0].builtin.value.data32 & mask);
               }
            }
            break;
         }

         core::gStackTop[1].builtin.builtinType = core::Value::BOOLEAN;
         ++core::gStackTop;
      }
      else
         binaryTypeError();
   }
}

// ===================================================================

} //namespace vm
} //namespace evlan
