/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/06/28 (KV)
 *  + First Version
 */

// ===================================================================

#include "persistence.h"
#include "memoryInternal.h"
#include "memory.h"
#include "operations.h"
#include "messages.h"
#include "loader.h"
#include "native.h"
#include "shortcuts.h"
#include <algorithm>
#include <stdio.h>
#include <framework/util/Vector.h>
#include <framework/messages/Message.h>

namespace evlan {
namespace vm {
using framework::messages::MessageReceiver;
using framework::messages::MessagePtr;
using framework::messages::makeMessage;
using namespace shortcuts;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::vm::persistence
// TODO (KV): Review evlan::vm::persistence

namespace persistence
{
#define DEBUG_PRINT(BLAH)
//#define DEBUG_PRINT(BLAH) Debug::log() << BLAH

   using namespace memory;

   struct GenerationHeader
   {
      int phase;
      int subPhase;
      int blockCount;
      int moduleCount;
      int rootTrackCount;
      int processCount;
      int stateTrackCount;
   };

   struct ModuleHeader
   {
      InternalModule* module;
      int headerLocation;
      int totalSize;
   };
   struct ProcessHeader
   {
      InternalProcess* process;
      int headerLocation;
      int type;
   };
   struct RootTrackHeader
   {
      int moduleLocation;
      core::Value root;
   };
   struct StateTrackHeader
   {
      int processLocation;
      core::Value state;
   };

   struct SegmentMapping
   {
      word start;
      word end;
      word newStart;

      inline SegmentMapping() {}
      inline SegmentMapping(word location): start(location) {}
      inline SegmentMapping(word start, word end, word newStart)
         : start(start), end(end), newStart(newStart) {}

      inline bool operator<(const SegmentMapping& other) const
         {return start < other.start;}
   };

   String gStateDirectory;
   String gStateName;
   FILE* gStateOutput;
   FILE* gIndexOutput;
   FILE* gIdentifierOutput;
   int gPosition;
   int gIdentifierCount;

   FILE* gStateInput;
   Array<int> gFrameLocations;

   Array<util::Vector<SegmentMapping> > gOldToNewMap;
   util::Vector<SegmentMapping> gFileToNewMap;

   std::map<int, InternalModule*> gModuleMap;
   std::map<int, InternalProcess*> gProcessMap;

   int writeFrame();
   int writeGeneration(int index);
   void writeModule(InternalModule* module);
   int writeBlock(Block* block);

   void readFrame(int location);
   void readGeneration(int location, int index);
   InternalModule* readModule(ModuleHeader header);
   Block* readBlock(int location);

   void readFrameMapping(int location, int frameNumber);
   void readGenerationMapping(int location, int frameNumber);
   void readBlockMapping(int location, int frameNumber);

   void remapAll(int frame);
   void remapModule(InternalModule* module);
   void remapValues(core::Value* segment, int size, int frame);
   void remapValueSegment(core::Value*& start, int size, int referringFrame);
   void remapDataSegment(word*& start, int size, int referringFrame);

   //write data to the state file, and return the start position of that data
   int write(const void* data, int size)
   {
      fwrite(data, 1, size, gStateOutput);
      int position = gPosition;
      gPosition += size;
      return position;
   }

   template <typename Type>
   inline int write(Type data)
      {return write(&data, sizeof(Type));}

   FILE* openFile(String name, char* mode)
   {
      FW_MAKE_CSTRING(cname, name);
      FILE* file = fopen(cname, mode);
      if(file == null)
         FW_THROW(spnew FatalError($("File not found: ") + name));
      return file;
   }

   void closeFile(FILE*& file)
   {
      fclose(file);
      file = null;
   }

   Array<const byte> readCompleteFile(String filename)
   {
      FILE* file = openFile(filename, "rb");
      fseek(file, 0, SEEK_END);
      int size = ftell(file);
      if(size == 0)
         return null;
      Array<byte> buffer(size);
      fseek(file, 0, SEEK_SET);
      fread(&buffer[0], 1, size, file);
      return buffer.toConst();
   }

   template<typename Type> inline int max(Type a, Type b) {return (a>b)?a:b;}
   template<typename Type> inline int min(Type a, Type b) {return (a<b)?a:b;}



   void init(String stateDirectory)
   {
      gStateDirectory = stateDirectory;
      gStateName = null;
   }

   void create(String name)
   {
      //erase old header locations
      for(int i = 0; i < 64; i++)
      {
         gGenerations[i].headerLocation = 0;

         for(Block* block = gGenerations[i].blocks; block != null; block = block->next)
         {
            block->headerLocation = 0;
            block->fileSegmentCount = 1;
            block->fileSegments[0].start = 0;
            block->fileSegments[0].fileLocation = 0;
            block->fileSegments[1].start = block->size;
            block->fileSegments[1].fileLocation = 0;
         }

         for(InternalModule* module = gGenerations[i].modules; module != null; module = module->next)
            module->headerLocation = 0;
         for(InternalProcess* process = gGenerations[i].processes; process != null; process = process->next)
            process->headerLocation = 0;
      }

      gStateName = gStateDirectory + $("/") + name;

      if(name == null)
         return;

      String stateName = gStateName + $(".state");
      String indexName = gStateName + $(".index");
      String identifierName = gStateName + $(".idents");

      gPosition = 0;
      gIdentifierCount = 0;

      gStateOutput = openFile(stateName, "wb");
      write("Evlan State File", strlen("Evlan State File"));
      closeFile(gStateOutput);

      gIndexOutput = openFile(indexName, "wb");
      closeFile(gIndexOutput);
      gIdentifierOutput = openFile(identifierName, "wb");
      closeFile(gIdentifierOutput);
   }

   void restore(framework::SystemPtr system, String name, int frame)
   {
      internalInit();

      gStateName = gStateDirectory + $("/") + name;

      String stateName = gStateName + $(".state");
      String indexName = gStateName + $(".index");
      String identifierName = gStateName + $(".idents");

      {
         Array<const byte> identifierData = readCompleteFile(identifierName);
         String identifierText = String(identifierData.reinterpretCast<Array<const char> >());

         int count = 0;
         for(int i = 0; i < identifierText.getLength(); i++)
         {
            if(identifierText[i] == '\n')
               ++count;
         }

         int pos = 0;
         Array<String> identifiers(count);
         count = 0;
         for(int i = 0; i < identifierText.getLength(); i++)
         {
            if(identifierText[i] == '\n')
            {
               identifiers[count++] = identifierText.getInterval(pos, i);
               pos = i + 1;
            }
         }

         gIdentifierCount = count;

         operations::init(identifiers.toConst());
         native::init();
      }

      {
         Array<const byte> indexData = readCompleteFile(indexName);
         gFrameLocations = Array<int>(indexData.getSize() / sizeof(int));
         if(indexData.getSize() > 0)
            memcpy(&gFrameLocations[0], &indexData[0], indexData.getSize());
      }

      if(frame == -1)
         frame = gFrameLocations.getSize() - 1;

      gStateInput = openFile(stateName, "rb");
      fseek(gStateInput, 0, SEEK_END);
      gPosition = ftell(gStateInput);

      //read frame
      DEBUG_PRINT("Reading frame...");
      readFrame(gFrameLocations[frame]);

      //build old->new mapping
      DEBUG_PRINT("Reading mapping...");
      for(int i = 0; i <= frame; i++)
         readFrameMapping(gFrameLocations[i], i);

      //done with these
      closeFile(gStateInput);
      gFileToNewMap.clear();
      gModuleMap.clear();
      gProcessMap.clear();

      //remap
      DEBUG_PRINT("Remapping...");
      remapAll(frame);
      gOldToNewMap = null;
      gFrameLocations = null;

      //clear segment sets
      DEBUG_PRINT("Clearing segments...");
      for(int i = 0; i < 64; i++)
      {
         for(Block* block = gGenerations[i].blocks; block != null; block = block->next)
            block->valueSegments.reset(block->size);
      }

      //restore processes
      DEBUG_PRINT("Restoring processes...");
      for(int i = 63; i >= 0; i--)
      {
         DEBUG_PRINT("generation " << i);
         for(InternalProcess* process = gGenerations[i].processes; process != null; process = process->next)
         {
            if(process->next == process)
               FW_THROW(spnew FatalError($("Process links to itself???")));
            if(process->stateTrackGeneration == i)
            {
               DEBUG_PRINT(process << " " << operations::getIdentifierName(process->type));
               native::ProcessConstructor constructor = native::getProcess(process->type);
               if(constructor == null)
                  FW_THROW(spnew FatalError($("Persistent state included native process which no longer exists:  ") + operations::getIdentifierName(process->type)));
               process->nativeInfo = constructor(process, system);
            }
         }

         DEBUG_PRINT("state track processes...");
         for(InternalProcess* process = gGenerations[i].stateTrackProcesses; process != null; process = process->stateTrackNext)
         {
            DEBUG_PRINT(process << " " << operations::getIdentifierName(process->type));
            if(process->stateTrackNext == process)
               FW_THROW(spnew FatalError($("Process links to itself???")));
            native::ProcessConstructor constructor = native::getProcess(process->type);
            if(constructor == null)
               FW_THROW(spnew FatalError($("Persistent state included native process which no longer exists:  ") + operations::getIdentifierName(process->type)));
            process->nativeInfo = constructor(process, system);
         }
      }

      /* Touch generations containing modules or processes, as their
       * memory locations have changed, requiring that the generations
       * be rewritten on next save. */
      for(int i = 0; i < 64; i++)
      {
         if(gGenerations[i].modules             != null
         || gGenerations[i].processes           != null
         || gGenerations[i].rootTrackModules    != null
         || gGenerations[i].stateTrackProcesses != null)
            gGenerations[i].headerLocation = 0;
      }

      //super-sweep to clean up
      DEBUG_PRINT("Cleaning up...");
      superCollect();

      DEBUG_PRINT("Done!");
   }

   void shutdown()
   {
      gStateName = null;
      gStateDirectory = null;
   }

   // ================================================================
   void update()
   {
      if(gStateName == null)
         return;  //no state file set

      if(core::gStackEnd + gParameters.stackSize - core::gStackTop != 0)
         FW_THROW(spnew FatalError($("Stuff was on the stack when persistent state was updated.")));

      String stateName = gStateName + $(".state");
      String indexName = gStateName + $(".index");
      String identifierName = gStateName + $(".idents");

      gStateOutput = openFile(stateName, "ab");
      fseek(gStateOutput, 0, SEEK_END);  //make ftell() not return zero
      gPosition = ftell(gStateOutput);
      gIndexOutput = openFile(indexName, "ab");
      gIdentifierOutput = openFile(identifierName, "ab");

      /* need to at least collect one generation here, in order to get
       * rid of the heap block.  Probably a good idea to collect at
       * least two generations.  Collecting more is a trade-off. */
      /* FIXME:  I previously used superCollect(4) here, but later
       * realized that generations in the "update" phase which contain
       * blocks need special handling.  In particular, rather than
       * writing the contents of the blocks, the segment mappings need
       * to be written.  They will then need to be applied when
       * remapping pointers on restore.  superCollect'ing all 64
       * generations obviates this problem because all generations in
       * the update phase will be left empty.  However, it will take
       * longer to perform the sweep, and will cause all generation
       * and block headers (but not all actual data) to be rewritten
       * when saved, even if they did not change.  I'm feeling lazy at
       * the moment so I'll implement the better solution (writing
       * maps) if I feel like it.  Fortunately, when I do get around
       * to implementing that, old save states will still work... */
      superCollect();

      int frameLocation = writeFrame();
      fwrite(&frameLocation, sizeof(int), 1, gIndexOutput);

      Array<const String> identifiers = operations::getIdentifierNames();

      while(gIdentifierCount < identifiers.getSize())
      {
         if(identifiers[gIdentifierCount] != null)
         {
            fwrite(&identifiers[gIdentifierCount].getBuffer()[0], 1,
                   identifiers[gIdentifierCount].getLength(), gIdentifierOutput);
         }
         fwrite("\n", 1, 1, gIdentifierOutput);
         ++gIdentifierCount;
      }

      closeFile(gStateOutput);
      closeFile(gIndexOutput);
      closeFile(gIdentifierOutput);
   }


   int writeFrame()
   {
      int generationHeaders[64];
      for(int i = 63; i >= 0; i--)
         generationHeaders[i] = writeGeneration(i);

      int frameLocation = write(sequence::counter);
      write(sequence::max);
      write(generationHeaders, sizeof(int) * 64);

      write(core::gTaskQueueSize);
      write(core::gTaskQueue, core::gTaskQueueSize * 2 * sizeof(core::Value));

      return frameLocation;
   }

   int writeGeneration(int index)
   {
      if(gGenerations[index].headerLocation != 0)
         return gGenerations[index].headerLocation;

      GenerationHeader header = {0, 0, 0, 0, 0, 0, 0};

      header.phase = gGenerations[index].phase;
      header.subPhase = gGenerations[index].subPhase;

      //write all blocks
      for(Block* block = gGenerations[index].blocks; block != null; block = block->next)
      {
         if(block->data != gHeapTop)
         {
            ++header.blockCount;
            writeBlock(block);
         }
      }

      //write all modules
      for(InternalModule* module = gGenerations[index].modules; module != null; module = module->next)
      {
         ++header.moduleCount;
         if(module->rootTrackGeneration == index)
            ++header.rootTrackCount;
         writeModule(module);
      }

      //count some stuff
      for(InternalModule* module = gGenerations[index].rootTrackModules; module != null; module = module->rootTrackNext)
         ++header.rootTrackCount;
      for(InternalProcess* process = gGenerations[index].processes; process != null; process = process->next)
      {
         ++header.processCount;
         if(process->stateTrackGeneration == index)
            ++header.stateTrackCount;
      }
      for(InternalProcess* process = gGenerations[index].stateTrackProcesses; process != null; process = process->stateTrackNext)
         ++header.stateTrackCount;

      //write header
      gGenerations[index].headerLocation = write(header);

      //   blocks
      {
         int blockLocations[header.blockCount];
         int i = 0;
         for(Block* block = gGenerations[index].blocks; block != null; block = block->next)
         {
            if(block->data != gHeapTop)
               blockLocations[i++] = block->headerLocation;
         }
         write(blockLocations, header.blockCount * sizeof(int));
      }

      //   modules
      {
         ModuleHeader moduleHeaders[header.moduleCount];
         int i = 0;
         for(InternalModule* module = gGenerations[index].modules; module != null; module = module->next)
         {
            moduleHeaders[i].module = module;
            moduleHeaders[i].totalSize = module->totalSize;
            moduleHeaders[i++].headerLocation = module->headerLocation;
         }
         write(moduleHeaders, header.moduleCount * sizeof(ModuleHeader));
      }

      //   processes
      {
         ProcessHeader processHeaders[header.processCount];
         int i = 0;
         for(InternalProcess* process = gGenerations[index].processes; process != null; process = process->next)
         {
            if(process->headerLocation == 0)
               process->headerLocation = gPosition + (i * sizeof(ProcessHeader));

            processHeaders[i].process = process;
            processHeaders[i].type = process->type;
            processHeaders[i++].headerLocation = process->headerLocation;
         }
         write(processHeaders, header.processCount * sizeof(ProcessHeader));
      }

      //   root track modules
      {
         RootTrackHeader trackHeaders[header.rootTrackCount];
         int i = 0;
         for(InternalModule* module = gGenerations[index].modules; module != null; module = module->next)
         {
            if(module->rootTrackGeneration == index)
            {
               trackHeaders[i].moduleLocation = module->headerLocation;
               trackHeaders[i++].root = module->root;
            }
         }

         for(InternalModule* module = gGenerations[index].rootTrackModules; module != null; module = module->rootTrackNext)
         {
            trackHeaders[i].moduleLocation = module->headerLocation;
            trackHeaders[i++].root = module->root;
         }

         write(trackHeaders, header.rootTrackCount * sizeof(RootTrackHeader));
      }

      //   state track processes
      {
         StateTrackHeader trackHeaders[header.stateTrackCount];
         int i = 0;
         for(InternalProcess* process = gGenerations[index].processes; process != null; process = process->next)
         {
            if(process->stateTrackGeneration == index)
            {
               trackHeaders[i].processLocation = process->headerLocation;
               trackHeaders[i++].state = process->state;
            }
         }

         for(InternalProcess* process = gGenerations[index].stateTrackProcesses; process != null; process = process->stateTrackNext)
         {
            trackHeaders[i].processLocation = process->headerLocation;
            trackHeaders[i++].state = process->state;
         }

         write(trackHeaders, header.stateTrackCount * sizeof(StateTrackHeader));
      }

      return gGenerations[index].headerLocation;
   }

   void writeModule(InternalModule* module)
   {
      if(module->headerLocation == 0)
      {
         module->headerLocation = write(module);

         loader::ModuleInfoPtr info = loader::getModuleInfo(module);
         write(info->name.getLength());
         if(info->name != null)
            write(&info->name.getBuffer()[0], info->name.getLength());
         write(info->fromFile);

         write(module, module->totalSize);  //heh heh
      }
   }

   int writeBlock(Block* block)
   {
      if(block->headerLocation != 0)
         return block->headerLocation;

      //write segments which are not yet in file
      for(int i = 0; i < block->fileSegmentCount; i++)
      {
         if(block->fileSegments[i].fileLocation == 0)
         {
            block->fileSegments[i].fileLocation =
               write(block->data + block->fileSegments[i].start,
                     (block->fileSegments[i+1].start -
                      block->fileSegments[i].start) * sizeof(word));
         }
      }

      //write header
      //   memory location
      block->headerLocation = write(block->data);
      write(block->size);

      //   segments
      write(block->fileSegmentCount);
      write(block->fileSegments, block->fileSegmentCount * sizeof(FileSegment));

      return block->headerLocation;
   }

   // ================================================================
   void seek(int location)
   {
      fseek(gStateInput, location, SEEK_SET);
   }

   void read(void* buffer, int byteCount)
   {
      if(byteCount == 0)
         return;
      if(fread(buffer, 1, byteCount, gStateInput) != (size_t)byteCount)
         FW_THROW(spnew FatalError($("Encountered unexpected end of file while loading state.")));
   }

   template <typename Type>
   inline void read(Type& item)
      {read(&item, sizeof(item));}

   /* Returns index of last element in array which is less than or
    * equal to the element given. */
   template <typename Array, typename Element>
   int findInSorted(Array array, Element element)
   {
      if(array.getSize() == 0)
         FW_THROW(spnew FatalError($("findInSorted() called on empty array while loading state.")));

      int start = 0, end = array.getSize();

      while(end - start > 1)
      {
         int midpoint = (start + end) / 2;
         if(element < array[midpoint])
            end = midpoint;
         else
            start = midpoint;
      }

      return start;
   }

   template <typename Array>
   void sort(Array array)
   {
      if(array.getSize() > 0)
         std::sort(&array[0], &array[0] + array.getSize());
   }

   word mapPointer(util::Vector<SegmentMapping> map, word pointer)
   {
      if(map.getSize() == 0)
         FW_THROW(spnew FatalError($("Invalid pointer encountered when restoring state.")));
      int index = findInSorted(map, SegmentMapping(pointer));
      if(pointer < map[index].start || pointer >= map[index].end)
      {
         DEBUG_PRINT((void*)pointer << "?  "
                  << (void*)map[index].start << " - "
                  << (void*)map[index].end);
         FW_THROW(spnew FatalError($("Invalid pointer encountered when restoring state.")));
      }
      return pointer - map[index].start + map[index].newStart;
   }

   int fileLocationToFrame(int location)
   {
      if(location < gFrameLocations[0])
         return 0;
      else
         return findInSorted(gFrameLocations, location) + 1;
   }

   template<typename Type>
   inline void fixPointer(Type*& pointer, int frame)
   {
      pointer = (Type*)mapPointer(gOldToNewMap[frame], (word)pointer);
   }

   void readFrame(int location)
   {
      DEBUG_PRINT("Reading frame: " << location);

      int generationHeaders[64];
      seek(location);
      read(sequence::counter);
      read(sequence::max);
      read(generationHeaders, sizeof(int)*64);

      int taskQueueSize;

      read(taskQueueSize);

      DEBUG_PRINT("Task queue size: " << taskQueueSize);

      core::gStackEnd = (core::Value*)allocatePages(
         gParameters.stackSize * sizeof(core::Value), true);
      core::gStackTop = core::gStackEnd + gParameters.stackSize;

      messages::init(taskQueueSize);
      read(core::gTaskQueue, taskQueueSize * 2 * sizeof(core::Value));

      for(int i = 63; i >= 0; i--)
         readGeneration(generationHeaders[i], i);

      sort(gFileToNewMap);
   }

   void readGeneration(int location, int index)
   {
      DEBUG_PRINT("Reading generation " << index << ": " << location);

      //read header
      GenerationHeader header;
      seek(location);
      read(header);

      gGenerations[index].headerLocation = location;
      gGenerations[index].phase = (Generation::Phase)header.phase;
      gGenerations[index].subPhase = header.subPhase;

      DEBUG_PRINT("   phase: "
         << ((header.phase == Generation::PASSIVE)?"passive":
            ((header.phase == Generation::MARK)?"mark":
            ((header.phase == Generation::UPDATE)?"update":"UNKNOWN"))));

      DEBUG_PRINT("   subPhase: " << header.subPhase);

      int blockLocations[header.blockCount];
      ModuleHeader moduleHeaders[header.moduleCount];
      ProcessHeader processHeaders[header.processCount];
      RootTrackHeader rootTrackHeaders[header.rootTrackCount];
      StateTrackHeader stateTrackHeaders[header.stateTrackCount];

      read(blockLocations, header.blockCount * sizeof(int));
      read(moduleHeaders, header.moduleCount * sizeof(ModuleHeader));
      read(processHeaders, header.processCount * sizeof(ProcessHeader));
      read(rootTrackHeaders, header.rootTrackCount * sizeof(RootTrackHeader));
      read(stateTrackHeaders, header.stateTrackCount * sizeof(StateTrackHeader));

      //   blocks
      gGenerations[index].blocks = null;
      for(int i = 0; i < header.blockCount; i++)
      {
         Block* block = readBlock(blockLocations[i]);
         block->generation = index;
         block->next = gGenerations[index].blocks;
         gGenerations[index].blocks = block;
      }

      //   modules
      gGenerations[index].modules = null;
      for(int i = 0; i < header.moduleCount; i++)
      {
         InternalModule* module = readModule(moduleHeaders[i]);
         gModuleMap[moduleHeaders[i].headerLocation] = module;
         module->generation = index;
         module->next = gGenerations[index].modules;
         gGenerations[index].modules = module;
      }

      //   processes
      gGenerations[index].processes = null;
      for(int i = 0; i < header.processCount; i++)
      {
         InternalProcess* process = new InternalProcess;
         process->marked = false;
         process->state.type = core::Value::EMPTY; //filled in later
         process->keepAlive = false;
         process->type = processHeaders[i].type;

         DEBUG_PRINT("process type: " << operations::getIdentifierName(processHeaders[i].type));

         process->next = gGenerations[index].processes;
         gGenerations[index].processes = process;
         process->generation = index;

         process->stateTrackNext = null;
         process->stateTrackLink = null;
         process->stateTrackGeneration = -1;

         process->headerLocation = processHeaders[i].headerLocation;
         gProcessMap[processHeaders[i].headerLocation] = process;
      }

      //   root track modules
      gGenerations[index].rootTrackModules = null;
      for(int i = 0; i < header.rootTrackCount; i++)
      {
         std::map<int, InternalModule*>::iterator iter;
         iter = gModuleMap.find(rootTrackHeaders[i].moduleLocation);
         if(iter == gModuleMap.end())
            FW_THROW(spnew FatalError($("Encountered root track header for unknown module while loading state.")));

         InternalModule* module = iter->second;

         module->root = rootTrackHeaders[i].root;
         module->rootTrackGeneration = index;
         if(module->generation != index)
         {
            module->rootTrackNext = gGenerations[index].rootTrackModules;
            if(module->rootTrackNext != null)
               module->rootTrackNext->rootTrackLink = &module->rootTrackNext;
            gGenerations[index].rootTrackModules = module;
            module->rootTrackLink = &gGenerations[index].rootTrackModules;
         }
      }

      //   state track processes
      gGenerations[index].stateTrackProcesses = null;
      for(int i = 0; i < header.stateTrackCount; i++)
      {
         std::map<int, InternalProcess*>::iterator iter;
         iter = gProcessMap.find(stateTrackHeaders[i].processLocation);
         if(iter == gProcessMap.end())
            FW_THROW(spnew FatalError($("Encountered state track header for unknown process while loading state.")));

         InternalProcess* process = iter->second;
         DEBUG_PRINT("process state: " << stateTrackHeaders[i].processLocation);
         if(process->stateTrackGeneration != -1)
            FW_THROW(spnew FatalError($("Process state encountered twice.")));

         process->state = stateTrackHeaders[i].state;
         process->stateTrackGeneration = index;
         if(process->generation != index)
         {
            process->stateTrackNext = gGenerations[index].stateTrackProcesses;
            if(process->stateTrackNext != null)
               process->stateTrackNext->stateTrackLink = &process->stateTrackNext;
            gGenerations[index].stateTrackProcesses = process;
            process->stateTrackLink = &gGenerations[index].stateTrackProcesses;
         }
      }
   }

   InternalModule* readModule(ModuleHeader header)
   {
      DEBUG_PRINT("Reading module: " << header.headerLocation);

      seek(header.headerLocation);

      InternalModule* oldLocation;

      read(oldLocation);

      int nameLength;
      read(nameLength);
      char name[nameLength];
      read(name, nameLength);
      bool fromFile;
      read(fromFile);

      InternalModule* module = (InternalModule*)allocatePages(header.totalSize);
      read(module, header.totalSize);

      DEBUG_PRINT("Module at: " << module << " - " << ((byte*)module + header.totalSize));

      memset(&module->nativeInfo, 0, sizeof(ObjectPtr));

      module->next = null;
      module->rootTrackNext = null;
      module->rootTrackLink = null;
      module->headerLocation = header.headerLocation;

      //root may or may not be filled in later
      memset(&module->root, 0, sizeof(core::Value));

      gModuleBlockMap[BlockKey(module, (byte*)module + header.totalSize)] = module;

      //fix pointers
      word offset = (word)module - (word)oldLocation;
      module->functionTags = (core::FunctionTag**)((word)module->functionTags + offset);
      module->objectTags   = (core::ObjectTag**  )((word)module->objectTags   + offset);
      module->constants    = (core::Value*       )((word)module->constants    + offset);

      for(int i = 0; i < module->functionTagCount; i++)
      {
         module->functionTags[i] = (core::FunctionTag*)((word)module->functionTags[i] + offset);
         module->functionTags[i]->module = module;
         module->functionTags[i]->documentation =
            (char*)((word)module->functionTags[i]->documentation + offset);
         if(module->functionTags[i]->parameterNames != null)
         {
            module->functionTags[i]->parameterNames =
               (int*)((word)module->functionTags[i]->parameterNames + offset);
         }
         if(module->functionTags[i]->sourceLocations != null)
         {
            module->functionTags[i]->sourceLocations =
               (core::SourceLocation*)((word)module->functionTags[i]->sourceLocations + offset);
         }
      }

      for(int i = 0; i < module->objectTagCount; i++)
      {
         module->objectTags[i] = (core::ObjectTag*)((word)module->objectTags[i] + offset);
         module->objectTags[i]->module = module;
         module->objectTags[i]->documentation =
            (char*)((word)module->objectTags[i]->documentation + offset);
      }

      loader::initModule(String(name, nameLength), fromFile, module);

      return module;
   }

   Block* readBlock(int location)
   {
      DEBUG_PRINT("Reading block: " << location);

      word* oldLocation;
      int size;
      int fileSegmentCount;

      seek(location);
      read(oldLocation);
      read(size);
      read(fileSegmentCount);

      DEBUG_PRINT("   size: " << size);
      DEBUG_PRINT("   segment count: " << fileSegmentCount);

      Block* block = allocateBlock(fileSegmentCount);

      read(block->fileSegments, fileSegmentCount * sizeof(FileSegment));
      block->fileSegments[fileSegmentCount].start = size;

      block->next = null;
      block->size = size;
      block->dataSegments.reset(size);
      block->valueSegments.reset(size);
      block->headerLocation = location;

      block->data = (word*)allocatePages(size * sizeof(word));

      DEBUG_PRINT("   location: " << block->data << " - " << (block->data + size));
      DEBUG_PRINT("   old location: " << oldLocation << " - " << (oldLocation + size));

      for(int i = 0; i < fileSegmentCount; i++)
      {
         int size = block->fileSegments[i+1].start - block->fileSegments[i].start;
         seek(block->fileSegments[i].fileLocation);
         read(block->data + block->fileSegments[i].start, size * sizeof(word));

         DEBUG_PRINT("   segment: " << (block->fileSegments[i].fileLocation)
                           << " - " << (block->fileSegments[i].fileLocation + size * sizeof(word)));
         gFileToNewMap.pushBack(
            SegmentMapping((word)(block->fileSegments[i].fileLocation),
                           (word)(block->fileSegments[i].fileLocation + size * sizeof(word)),
                           (word)(block->data + block->fileSegments[i].start)));
      }

      gBlockMap[BlockKey(block->data, block->data + size)] = block;

      return block;
   }



   void readFrameMapping(int location, int frameNumber)
   {
      DEBUG_PRINT("mapping frame: " << frameNumber);

      if(gOldToNewMap == null)
         gOldToNewMap = Array<util::Vector<SegmentMapping> >(gFrameLocations.getSize());

      int generationHeaders[64];
      seek(location + sizeof(uint64) + sizeof(int));
      read(generationHeaders, sizeof(int)*64);
      for(int i = 0; i < 64; i++)
         readGenerationMapping(generationHeaders[i], frameNumber);

      sort(gOldToNewMap[frameNumber]);
   }

   void readGenerationMapping(int location, int frameNumber)
   {
      //read header
      GenerationHeader header;
      seek(location);
      read(header);

      int blockLocations[header.blockCount];
      ModuleHeader moduleHeaders[header.moduleCount];
      ProcessHeader processHeaders[header.processCount];

      read(blockLocations, header.blockCount * sizeof(int));
      read(moduleHeaders, header.moduleCount * sizeof(ModuleHeader));
      read(processHeaders, header.processCount * sizeof(ProcessHeader));

      for(int i = 0; i < header.blockCount; i++)
         readBlockMapping(blockLocations[i], frameNumber);

      for(int i = 0; i < header.moduleCount; i++)
      {
         std::map<int, InternalModule*>::iterator iter =
            gModuleMap.find(moduleHeaders[i].headerLocation);
         if(iter != gModuleMap.end())
         {
            //module is currently active
            gOldToNewMap[frameNumber].pushBack(
               SegmentMapping((word)moduleHeaders[i].module,
                              (word)((byte*)moduleHeaders[i].module + moduleHeaders[i].totalSize),
                              (word)iter->second));
         }
      }

      for(int i = 0; i < header.processCount; i++)
      {
         std::map<int, InternalProcess*>::iterator iter =
            gProcessMap.find(processHeaders[i].headerLocation);
         if(iter != gProcessMap.end())
         {
            //process is curretnly active
            gOldToNewMap[frameNumber].pushBack(
               SegmentMapping((word)processHeaders[i].process,
                              (word)(processHeaders[i].process + 1),
                              (word)iter->second));
         }
      }
   }

   void readBlockMapping(int location, int frameNumber)
   {
      word* oldLocation;
      int size;
      int fileSegmentCount;

      seek(location);
      read(oldLocation);
      read(size);
      read(fileSegmentCount);

      FileSegment fileSegments[fileSegmentCount + 1];
      read(fileSegments, fileSegmentCount * sizeof(FileSegment));
      fileSegments[fileSegmentCount].start = size;

      for(int i = 0; i < fileSegmentCount; i++)
      {
         int index = findInSorted(gFileToNewMap, SegmentMapping(fileSegments[i].fileLocation));
         word* oldSegmentLocation = oldLocation + fileSegments[i].start;

         //start and end points of the segment in the file
         word segmentStart = fileSegments[i].fileLocation;
         word segmentEnd = fileSegments[i].fileLocation +
            (fileSegments[i+1].start - fileSegments[i].start) * sizeof(word);

         for(; index < gFileToNewMap.getSize(); ++index)
         {
            if(gFileToNewMap[index].end <= segmentStart)
               continue;
            if(gFileToNewMap[index].start >= segmentEnd)
               break;

            word pieceStart = max(segmentStart, gFileToNewMap[index].start);
            word pieceEnd   = min(segmentEnd  , gFileToNewMap[index].end  );

            word target = gFileToNewMap[index].newStart
                        + pieceStart - gFileToNewMap[index].start;

            gOldToNewMap[frameNumber].pushBack(
               SegmentMapping(
                  (word)oldSegmentLocation + (pieceStart - segmentStart),
                  (word)oldSegmentLocation + (pieceEnd   - segmentStart),
                  target));

            DEBUG_PRINT("   " << (void*)((word)oldSegmentLocation + (pieceStart - segmentStart))
                     << " - " << (void*)((word)oldSegmentLocation + (pieceEnd   - segmentStart))
                     << " => " << (void*)target
                     << " - " << (void*)(target + (pieceEnd - pieceStart)));
         }
      }
   }


   // ================================================================
   void remapAll(int frame)
   {
      //stack should be empty
//      remapValues(core::gStackTop,
//                  core::gStackEnd + gParameters.stackSize - core::gStackTop,
//                  frame);
      remapValues((core::Value*)core::gTaskQueue, core::gTaskQueueSize * 2, frame);

      for(int i = 0; i < 64; i++)
      {
         int frame = fileLocationToFrame(gGenerations[i].headerLocation);

         for(InternalModule* module = gGenerations[i].modules; module != null; module = module->next)
            remapModule(module);
         for(InternalProcess* process = gGenerations[i].processes; process != null; process = process->next)
         {
            DEBUG_PRINT("Remapping process: " << process);
            remapValues(&process->state, 1, frame);
         }
      }
   }

   void remapModule(InternalModule* module)
   {
      int frame = fileLocationToFrame(module->headerLocation);
      int rootFrame = fileLocationToFrame(gGenerations[module->rootTrackGeneration].headerLocation);

      DEBUG_PRINT("generation header location: " << gGenerations[module->rootTrackGeneration].headerLocation);

      remapValues(&module->root, 1, rootFrame);
      remapValues(module->constants, module->constantCount, frame);

      remapDataSegment(module->extraData, module->extraDataSize, frame);
   }

   void remapValues(core::Value* segment, int size, int frame)
   {
      DEBUG_PRINT("remapValues: " << segment << " - " << (segment + size));

      for(int i = 0; i < size; i++)
      {
         switch(segment[i].type)
         {
         case core::Value::RUNTIME_ERROR:
            if(segment[i].runtimeError.location != null)
               fixPointer(segment[i].runtimeError.location, frame);
            break;
         case core::Value::NATIVE_FUNCTION:
            segment[i].nativeFunction.tag =
               native::getFunction(segment[i].nativeFunction.id);
            if(segment[i].nativeFunction.tag == null)
               FW_THROW(spnew FatalError($("Persistent state contained a reference to a native function that no longer exists:  ") + operations::getIdentifierName(segment[i].nativeFunction.id)));
            break;
         case core::Value::TASK:
            remapValueSegment(segment[i].task.parameters,
                              segment[i].task.parameterCount,
                              frame);
            break;
         case core::Value::CONTEXT:
            if(segment[i].context.contextType == core::Value::THROW)
               remapValueSegment(segment[i].context.parameter, 2, frame);
            else if(segment[i].context.parameter != null)
               remapValueSegment(segment[i].context.parameter, 1, frame);
            if(segment[i].context.parent != null)
               remapValueSegment(segment[i].context.parent, 1, frame);
            break;
         case core::Value::FUNCTION:
            fixPointer(segment[i].function.tag, frame);
            remapValueSegment(segment[i].function.environment,
                              segment[i].function.tag->environmentSize,
                              frame);
            break;
         case core::Value::OBJECT:
            fixPointer(segment[i].object.tag, frame);
            remapValueSegment(segment[i].object.members,
                              segment[i].object.tag->memberCount,
                              frame);
            break;
         case core::Value::ARRAY:
            remapValueSegment(segment[i].array.elements,
                              segment[i].array.size,
                              frame);
            break;
         case core::Value::PACKED_ARRAY:
            {
               int wordSize;
               if(segment[i].packedArray.elementSize < core::LOG2_WORD_SIZE)
               {
                  int diff = (core::LOG2_WORD_SIZE - segment[i].packedArray.elementSize);
                  int offsetSize = segment[i].packedArray.size + segment[i].packedArray.offset;
                  wordSize = offsetSize >> diff;
                  if((offsetSize & ((1 << diff) - 1)) != 0)
                     ++wordSize;
               }
               else if(segment[i].packedArray.elementSize == core::LOG2_WORD_SIZE)
                  wordSize = segment[i].packedArray.size;
               else
                  wordSize = segment[i].packedArray.size << (segment[i].packedArray.elementSize - core::LOG2_WORD_SIZE);

               remapDataSegment(segment[i].packedArray.elements,
                                wordSize,
                                frame);
            }
            break;
         case core::Value::LARGE_DATA:
            remapDataSegment(segment[i].largeData.value,
                             1 << (segment[i].largeData.size - core::LOG2_WORD_SIZE),
                             frame);
            break;

         case core::Value::MODULE:
            fixPointer(segment[i].module.module, frame);
            break;

         case core::Value::PROCESS:
            fixPointer(segment[i].process.process, frame);
            break;

         case core::Value::EMPTY:
         case core::Value::BUILTIN_OBJECT:
         case core::Value::NATIVE_CONTAINER:
            //no pointers
            break;

         default:
            FW_THROW(spnew FatalError($("Encountered invalid value while remaping pointers from loaded state.  Location: 0x")
                                     + String((word)(segment + i), 2 * sizeof(word), 16)));
         }
      }
   }

   void remapValueSegment(core::Value*& start, int size, int referringFrame)
   {
      if(size == 0)
         return;

      DEBUG_PRINT("remapValueSegment: " << start << " - " << (start + size));
      DEBUG_PRINT("   ref frame: " << referringFrame);

      start = (core::Value*)mapPointer(gOldToNewMap[referringFrame], (word)start);

      DEBUG_PRINT("   => " << start << " - " << (start + size));

      std::map<BlockKey, Block*>::iterator iter =
         gBlockMap.find(BlockKey(start, start + size));
      if(iter == gBlockMap.end())
         FW_THROW(spnew FatalError($("Couldn't find new location for pointer when loading state.")));
      Block* block = iter->second;
      int startIndex = (word*)start - block->data;
      int endIndex = startIndex + size * core::WORDS_PER_VALUE;

      /* Unlike in markValueSegment(), we need to be extra careful not
       * to hit any pointer twice.  So, we find out exactly which
       * segments have already been covered and only do the rest. */
      int segmentCount = block->valueSegments.countSegmentsInRange(startIndex, endIndex);
      int segments[segmentCount*2];
      block->valueSegments.getSegmentsInRange(segments, startIndex, endIndex);

      block->valueSegments.insert(startIndex, startIndex + size * core::WORDS_PER_VALUE);

      for(int i = 0; i <= segmentCount; i++)
      {
         core::Value* segmentStart;
         core::Value* segmentEnd;

         if(i == 0)
         {
            if(segmentCount > 0 && segments[0] <= startIndex)
               continue;
            segmentStart = start;
         }
         else
            segmentStart = (core::Value*)(block->data + segments[i*2-1]);

         if(i == segmentCount)
         {
            if(segmentCount > 0 && segments[segmentCount*2-1] >= endIndex)
               continue;
            segmentEnd = start + size;
         }
         else
            segmentEnd = (core::Value*)(block->data + segments[i*2]);

         DEBUG_PRINT("   sub-segment:  " << segmentStart << " - " << segmentEnd);

         int frame = fileLocationToFrame(block->getFileLocation((word*)segmentStart));

         remapValues(segmentStart, segmentEnd - segmentStart, frame);
      }
   }

   void remapDataSegment(word*& start, int size, int referringFrame)
   {
      if(size == 0)
         return;

      start = (word*)mapPointer(gOldToNewMap[referringFrame], (word)start);
   }



   EVLAN_NATIVE_FUNCTION_1(memorySweep, depth)
   {
      if(!depth.isInteger())
         return simTypeError(depth);
      superCollect((int)depth.builtin.value.scalar);
      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(memoryFindProcesses, type)
   {
      if(!type.isAtom())
         return simTypeError(type);

      int count = 0;

      for(int i = 0; i < 64; i++)
      {
         for(InternalProcess* process = gGenerations[i].processes; process != null; process = process->next)
         {
            if(process->type == (int)type.builtin.value.data32)
            {
               ++count;
               simPush(Value::makeProcess(process));
            }
         }
      }

      Value* elements = allocateValues(count);
      memcpy(elements, gStackTop, count * sizeof(Value));
      simPop(count);

      return simReturn(Value::makeArray(count, elements));
   }

   EVLAN_NATIVE_FUNCTION_0(memoryFindKeepaliveProcesses)
   {
      int count = 0;

      for(int i = 0; i < 64; i++)
      {
         for(InternalProcess* process = gGenerations[i].processes; process != null; process = process->next)
         {
            if(process->keepAlive)
            {
               ++count;
               simPush(Value::makeProcess(process));
            }
         }
      }

      Value* elements = allocateValues(count);
      memcpy(elements, gStackTop, count * sizeof(Value));
      simPop(count);

      return simReturn(Value::makeArray(count, elements));
   }










   implementation ProcessFinalizerInfo: implements MessageReceiver
   {
      ProcessFinalizerInfo(core::Process* process, framework::SystemPtr system);
      ProcessFinalizerInfo(core::Process* process);
      ~ProcessFinalizerInfo();

      void receive(MessagePtr message);

      core::Process* mProcess;
      MessagePtr mMessage;

      bool mLinkExists;
   };
   typedef SmartPointer<ProcessFinalizerInfo> ProcessFinalizerInfoPtr;

   ProcessFinalizerInfo::ProcessFinalizerInfo(core::Process* process, framework::SystemPtr system)
      : mProcess(process), mLinkExists(false)
   {
      mMessage = makeMessage(this, 0, null);
      if(mProcess->state.isTask())
      {
         if(mProcess->keepAlive)
         {
            //finalizationNotifier has been initialized
            mLinkExists = true;
         }
         else
         {
            /* finalizationNotifier has not been initialized.  If it
             * is initialized after this process, it will set
             * mLinkExists to true before this message is received,
             * aborting the finalization. */
            mMessage->send();
            mProcess->keepAlive = true;
         }
      }
   }

   ProcessFinalizerInfo::ProcessFinalizerInfo(core::Process* process)
      : mProcess(process), mLinkExists(true)
   {
      mMessage = makeMessage(this, 0, null);
      if(mProcess->state.isTask())
         mProcess->keepAlive = true;
   }

   ProcessFinalizerInfo::~ProcessFinalizerInfo()
   {
   }

   void ProcessFinalizerInfo::receive(MessagePtr message)
   {
      if(mLinkExists)
         return;

      simPush(Value::makeContext(Value::STOP));
      simPush(mProcess->state);
      simQueue();

      simPush(Value::makeEmpty());
      memory::modifyProcessState(mProcess);
      mProcess->keepAlive = false;
   }

   EVLAN_NATIVE_PROCESS(processFinalizer, ProcessFinalizerInfo)

   implementation FinalizationNotifierInfo: implements Object
   {
      FinalizationNotifierInfo(core::Process* process, framework::SystemPtr system);
      FinalizationNotifierInfo(core::Process* process);
      ~FinalizationNotifierInfo();

      core::Process* mTargetProcess;
   };

   FinalizationNotifierInfo::FinalizationNotifierInfo(core::Process* process, framework::SystemPtr system)
   {
      mTargetProcess = process->state.process.process;

#ifdef DEBUG
      if(((InternalProcess*)mTargetProcess)->generation < ((InternalProcess*)process)->generation)
         FW_THROW(spnew FatalError($("Reverse generational link in finalizer state.")));
#endif

      if(mTargetProcess->keepAlive)
      {
         ProcessFinalizerInfoPtr finalizerInfo =
            mTargetProcess->nativeInfo.dynamicCast<ProcessFinalizerInfoPtr>();
         finalizerInfo->mLinkExists = true;
      }
      else
         mTargetProcess->keepAlive = true;
   }

   FinalizationNotifierInfo::FinalizationNotifierInfo(core::Process* process)
   {
      mTargetProcess = process->state.process.process;
   }

   FinalizationNotifierInfo::~FinalizationNotifierInfo()
   {
      ProcessFinalizerInfoPtr finalizerInfo =
         mTargetProcess->nativeInfo.dynamicCast<ProcessFinalizerInfoPtr>();
      finalizerInfo->mLinkExists = false;
      finalizerInfo->mMessage->send();
   }

   EVLAN_NATIVE_PROCESS(finalizationNotifier, FinalizationNotifierInfo)

   EVLAN_NATIVE_FUNCTION_1(memoryMakeFinalizer, finalize)
   {
      if(!finalize.isTask())
         return simTypeError(finalize);

      simPush(finalize);
      memory::makeProcess($("processFinalizer"));
      gStackTop->process.process->nativeInfo =
         spnew ProcessFinalizerInfo(gStackTop->process.process);

      memory::makeProcess($("finalizationNotifier"));
      gStackTop->process.process->nativeInfo =
         spnew FinalizationNotifierInfo(gStackTop->process.process);

      return simReturn(*gStackTop);
   }

   EVLAN_NATIVE_FUNCTION_0(memoryStartProfiling)
   {
      for(int i = 0; i < 64; i++)
      {
         for(InternalModule* module = gGenerations[i].modules; module != null; module = module->next)
         {
            for(int j = 0; j < module->functionTagCount; j++)
            {
               module->functionTags[j]->profileTimer = 0;
               module->functionTags[j]->profileCounter = 0;
            }
         }
      }

      Array<const int> ids = native::getIdList();

      for(int i = 0; i < ids.getSize(); i++)
      {
         core::NativeFunctionTag* tag = native::getFunction(ids[i]);
         tag->profileTimer = 0;
         tag->profileCounter = 0;
      }

      simulator::enableProfiling(true);

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_0(memoryStopProfiling)
   {
      simulator::enableProfiling(false);

      return simReturn();
   }

   EVLAN_NATIVE_FUNCTION_1(memoryGetProfileData, module)
   {
      if(!module.isModule())
         return simTypeError(module);

      Value* info = allocateValues(module.module.module->functionTagCount * 4);
      for(int i = 0; i < module.module.module->functionTagCount; i++)
      {
         info[i*4+0] = Value::makeScalar(module.module.module->functionTags[i]->line);
         info[i*4+1] = Value::makeScalar(module.module.module->functionTags[i]->column);
         info[i*4+2] = Value::makeScalar(module.module.module->functionTags[i]->profileTimer);
         info[i*4+3] = Value::makeScalar(module.module.module->functionTags[i]->profileCounter);
      }

      return Value::makeArray(module.module.module->functionTagCount * 4, info);
   }

   EVLAN_NATIVE_FUNCTION_0(memoryGetNativeProfileData)
   {
      Array<const int> ids = native::getIdList();

      Value* info = allocateValues(ids.getSize() * 3);
      for(int i = 0; i < ids.getSize(); i++)
      {
         core::NativeFunctionTag* tag = native::getFunction(ids[i]);
         info[i*3+0] = Value::makeAtom(ids[i]);
         info[i*3+1] = Value::makeScalar(tag->profileTimer);
         info[i*3+2] = Value::makeScalar(tag->profileCounter);
      }

      return Value::makeArray(ids.getSize() * 3, info);
   }
};

// ===================================================================

} //namespace vm
} //namespace evlan
