/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/03/01 (KV)
 *  + First Version
 */

// ===================================================================

#include "simulator.h"
#include "memory.h"
#include "operations.h"
#include "arrays.h"
#include <framework/util/Vector.h>

namespace evlan {
namespace vm {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::vm::simulator
// TODO (KV): Review evlan::vm::simulator

namespace simulator
{
   static const int TICK_PERIOD = 1 << 22;
   int gTickCount = 0;
   void (*gTickCallback)();

   bool gProfilingEnabled = false;
   bool gCurrentFunctionIsNative = false;
   void* gCurrentFunction = null;
   uint64 gInstructionCount = 0;
   uint64 gProfileTime = 0;

   void enableProfiling(bool enable) {gProfilingEnabled = enable;}

   inline void updateProfileTime() {gProfileTime = gInstructionCount;}

   void profileEnter(void* function, bool isNative, bool start)
   {
      uint64 oldTime = gProfileTime;
      updateProfileTime();
      if(gCurrentFunction != null)
      {
         if(gCurrentFunctionIsNative)
            ((core::NativeFunctionTag*)gCurrentFunction)->profileTimer += gProfileTime - oldTime;
         else
            ((core::FunctionTag*)gCurrentFunction)->profileTimer += gProfileTime - oldTime;
      }

      gCurrentFunctionIsNative = isNative;
      gCurrentFunction = function;

      if(start && function != null)
      {
         if(isNative)
            ++((core::NativeFunctionTag*)function)->profileCounter;
         else
            ++((core::FunctionTag*)function)->profileCounter;
      }
   }



   void applyMember(int memberId)
   {
      --core::gStackTop;
      *core::gStackTop = core::gStackTop[2];
      operations::getMember(memberId);
      core::gStackTop[2] = *core::gStackTop;
      ++core::gStackTop;
      simulator::apply(1);
   }

   void apply(int arity)
   {
      core::Value* function = core::gStackTop + arity;
      core::Instruction* instruction = null;
      core::FunctionTag* tag = null;

      void* oldFunction = gCurrentFunction;
      bool oldFunctionIsNative = gCurrentFunctionIsNative;

   tailRecurse:
//      {
//         static int count = 0;
//         if(++count > 600000)
//         {
//            if(function->type == core::Value::FUNCTION)
//            {
//               Debug::log() << loader::getModuleInfo(function->function.tag->module)->name
//                  << ":" << function->function.tag->line
//                  << ":" << function->function.tag->column
//                  << " " << function << " " << function->function.tag;
//
//               if(function->function.tag->environmentSize > 1)
//               {
//                  Debug::log() << function->function.environment[1].builtin.value.scalar;
//               }
//            }
//         }
//
//         if(function->function.tag == (core::FunctionTag*)0x0171424c)
//         {
//            if(function->function.environment[1].builtin.value.scalar < 0)
//               function->function.environment[1].builtin.value.scalar = 0;
//         }
//      }

      switch(function->type)
      {
      case core::Value::RUNTIME_ERROR:
         core::gStackTop += arity;
         return;
      case core::Value::NATIVE_FUNCTION:
         if(function->nativeFunction.tag->arity != arity)
         {
            core::gStackTop += arity;
            *core::gStackTop = core::Value::makeError(core::Value::ARITY_MISMATCH, instruction);
         }
         else
         {
            if(gProfilingEnabled) profileEnter(function->nativeFunction.tag, true, true);
            function->nativeFunction.tag->nativeApply();
            if(gProfilingEnabled) profileEnter(oldFunction, oldFunctionIsNative, false);
         }
         return;
      case core::Value::FUNCTION:
         if(function->function.tag->arity != arity)
         {
            core::gStackTop += arity;
            *core::gStackTop = core::Value::makeError(core::Value::ARITY_MISMATCH, instruction);
            return;
         }
         if(gProfilingEnabled) profileEnter(function->function.tag, false, true);
         break;
      default:
         core::gStackTop += arity;
         *core::gStackTop = core::Value::makeError(core::Value::NOT_A_FUNCTION, instruction);
         return;
      }

      // =============================================================
      tag = function->function.tag;
      instruction = tag->instructions;

      while(true)
      {
         if(gProfilingEnabled) ++gInstructionCount;

         if(gTickCallback != null && ++gTickCount > TICK_PERIOD)
         {
            gTickCount = 0;
            gTickCallback();
         }

         switch(instruction->opcode)
         {
         case core::NOOP:
            break;
         case core::RETURN:
            *function = *core::gStackTop;
            core::gStackTop = function;
            if(gProfilingEnabled) profileEnter(oldFunction, oldFunctionIsNative, false);
            return;
         case core::GOTO:
            instruction += instruction->parameter;
            break;
         case core::TEST_CONDITION:
            if(core::gStackTop->type == core::Value::BUILTIN_OBJECT
            && core::gStackTop->builtin.builtinType == core::Value::BOOLEAN)
            {
               if(core::gStackTop->builtin.value.boolean)
                  ++instruction;
               else
                  instruction += instruction->parameter;
               ++core::gStackTop;
            }
            else if(core::gStackTop->type != core::Value::RUNTIME_ERROR)
               *core::gStackTop = core::Value::makeError(core::Value::CONDITION_NOT_BOOLEAN, instruction);
            break;
         case core::PUSH_EMPTY:
            for(uint i = 0; i < instruction->parameter; i++)
               (--core::gStackTop)->type = core::Value::EMPTY;
            break;
         case core::POP:
            core::gStackTop += instruction->parameter;
            break;
         case core::LOAD:
            --core::gStackTop;
            *core::gStackTop = core::gStackTop[instruction->parameter];
            break;
         case core::LOAD_ENVIRONMENT:
            *(--core::gStackTop) = function->function.environment[instruction->parameter];
            break;
         case core::LOAD_CONSTANT:
            *(--core::gStackTop) = function->function.tag->module->constants[instruction->parameter];
            break;
         case core::STORE:
            core::gStackTop[instruction->parameter] = *core::gStackTop;
            break;
         case core::STORE_POP:
            core::gStackTop[instruction->parameter] = *core::gStackTop;
            ++core::gStackTop;
            break;
         case core::GET_MEMBER:
            operations::getMember(instruction->parameter);
            goto check_error;
         case core::APPLY:
            apply(instruction->parameter);
            goto check_error;
         case core::APPLY_RETURN:
            arity = instruction->parameter;
            for(int i = arity; i >= 0; i--)
               function[i-arity] = core::gStackTop[i];
            core::gStackTop = function - arity;
            goto tailRecurse;  //OH MY GOD!  A GOTO!  IT'S THE END OF THE WORLD!
         case core::MAKE_FUNCTION:
            {
               int size = tag->module->functionTags[instruction->parameter]->environmentSize;
               core::Value* environment = null;
               if(size > 0)
               {
                  environment = memory::allocateValues(size);
                  for(int i = 0; i < size; i++)
                     environment[i] = core::gStackTop[size-1-i];
               }

               core::gStackTop += size - 1;
               core::gStackTop->type = core::Value::FUNCTION;
               core::gStackTop->function.tag = tag->module->functionTags[instruction->parameter];
               core::gStackTop->function.environment = environment;
            }
            break;
         case core::MAKE_OBJECT:
            {
               int size = tag->module->objectTags[instruction->parameter]->memberCount;
               core::Value* members = null;
               if(size > 0)
               {
                  members = memory::allocateValues(size);
                  for(int i = 0; i < size; i++)
                     members[i] = core::gStackTop[size-1-i];
               }

               core::gStackTop += size - 1;
               core::gStackTop->type = core::Value::OBJECT;
               core::gStackTop->object.tag = tag->module->objectTags[instruction->parameter];
               core::gStackTop->object.members = members;
            }
            break;
         case core::MAKE_ARRAY:
            {
               int size = instruction->parameter;
               core::Value* elements = null;
               if(size > 0)
               {
                  elements = memory::allocateValues(size);
                  for(int i = 0; i < size; i++)
                     elements[i] = core::gStackTop[size-1-i];
               }

               core::gStackTop += size - 1;
               core::gStackTop->type = core::Value::ARRAY;
               core::gStackTop->array.size = instruction->parameter;
               core::gStackTop->array.elements = elements;
               arrays::tryPackArray();
            }
            break;
         case core::ALLOCATE_MULTI:
            //looks the same as MAKE_ARRAY, yes, but semantically it's different
            {
               int size = instruction->parameter;
               core::Value* elements = null;
               if(size > 0)
               {
                  elements = memory::allocateValues(size);
                  for(int i = 0; i < size; i++)
                     elements[i] = core::gStackTop[size-1-i];
               }

               core::gStackTop += size - 1;
               core::gStackTop->type = core::Value::ARRAY;
               core::gStackTop->array.size = 0;
               core::gStackTop->array.elements = elements;
            }
            break;
         case core::MAKE_FUNCTION_MULTI:
            --core::gStackTop;
            core::gStackTop->type = core::Value::FUNCTION;
            core::gStackTop->function.tag = tag->module->functionTags[instruction->parameter];
            core::gStackTop->function.environment = core::gStackTop[1].array.elements + core::gStackTop[1].array.size;
            core::gStackTop[1].array.size += core::gStackTop->function.tag->environmentSize;
            break;
         case core::MAKE_OBJECT_MULTI:
            --core::gStackTop;
            core::gStackTop->type = core::Value::OBJECT;
            core::gStackTop->object.tag = tag->module->objectTags[instruction->parameter];
            core::gStackTop->object.members = core::gStackTop[1].array.elements + core::gStackTop[1].array.size;
            core::gStackTop[1].array.size += core::gStackTop->object.tag->memberCount;
            break;
         case core::MAKE_ARRAY_MULTI:
            --core::gStackTop;
            core::gStackTop->type = core::Value::ARRAY;
            core::gStackTop->array.size = instruction->parameter;
            core::gStackTop->array.elements = core::gStackTop[1].array.elements + core::gStackTop[1].array.size;
            core::gStackTop[1].array.size += instruction->parameter;
            break;
         case core::STORE_MULTI:
            core::gStackTop[1].array.elements[instruction->parameter] = *core::gStackTop;
            break;
         case core::STORE_MULTI_POP:
            core::gStackTop[1].array.elements[instruction->parameter] = *core::gStackTop;
            ++core::gStackTop;
            break;

         case core::IMMEDIATE_INTEGER:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::SCALAR;
            core::gStackTop->builtin.value.scalar = instruction->parameter;
            break;
         case core::IMMEDIATE_NEGATIVE:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::SCALAR;
            /* GCC bug!  If you remove the explicit int cast here, it
             * doesn't extend the sign correctly.  This only appears
             * to be a problem when casting to a double. */
            core::gStackTop->builtin.value.scalar = -(int)instruction->parameter;
            break;
         case core::IMMEDIATE_BOOLEAN:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::BOOLEAN;
            core::gStackTop->builtin.value.boolean = instruction->parameter?true:false;
            break;
         case core::IMMEDIATE_CHARACTER:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::CHARACTER;
            core::gStackTop->builtin.value.data32 = instruction->parameter;
            break;
         case core::IMMEDIATE_ATOM:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::ATOM;
            core::gStackTop->builtin.value.data32 = instruction->parameter;
            break;
         case core::IMMEDIATE_INTEGER32:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::SCALAR;
            core::gStackTop->builtin.value.scalar = *(int32*)(++instruction);
            break;
         case core::IMMEDIATE_INTEGER64:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::SCALAR;
            core::gStackTop->builtin.value.scalar = *(int64*)(++instruction);
            ++instruction;
            break;
         case core::IMMEDIATE_REAL32:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::SCALAR;
            core::gStackTop->builtin.value.scalar = *(float*)(++instruction);
            break;
         case core::IMMEDIATE_REAL64:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::SCALAR;
            core::gStackTop->builtin.value.scalar = *(double*)(++instruction);
            ++instruction;
            break;
         case core::IMMEDIATE_CHARACTER32:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::CHARACTER;
            core::gStackTop->builtin.value.data32 = *(uint32*)(++instruction);
            break;
         case core::IMMEDIATE_ATOM32:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::ATOM;
            core::gStackTop->builtin.value.data32 = *(uint32*)(++instruction);
            break;
         case core::IMMEDIATE_DATA1:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::DATA1;
            core::gStackTop->builtin.value.data32 = instruction->parameter?1:0;
            break;
         case core::IMMEDIATE_DATA2:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::DATA2;
            core::gStackTop->builtin.value.data32 = instruction->parameter & 0x00000003;
            break;
         case core::IMMEDIATE_DATA4:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::DATA4;
            core::gStackTop->builtin.value.data32 = instruction->parameter & 0x0000000F;
            break;
         case core::IMMEDIATE_DATA8:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::DATA8;
            core::gStackTop->builtin.value.data32 = instruction->parameter & 0x000000FF;
            break;
         case core::IMMEDIATE_DATA16:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::DATA16;
            core::gStackTop->builtin.value.data32 = instruction->parameter & 0x0000FFFF;
            break;
         case core::IMMEDIATE_DATA32:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::DATA32;
            core::gStackTop->builtin.value.data32 = *(uint32*)(++instruction);
            break;
         case core::IMMEDIATE_DATA64:
            --core::gStackTop;
            core::gStackTop->type = core::Value::BUILTIN_OBJECT;
            core::gStackTop->builtin.builtinType = core::Value::DATA64;
            core::gStackTop->builtin.value.data64 = *(uint64*)(++instruction);
            ++instruction;
            break;

         case core::NOT:
            if(core::gStackTop->isBuiltin())
               operations::doNot();
            else
               getMember(operations::BUILTIN_NOT);
            goto check_error;

         case core::NEGATE:
            if(core::gStackTop->isBuiltin())
               operations::doNegate();
            else
               getMember(operations::BUILTIN_NEGATE);
            goto check_error;

         case core::SIZE:
            if(core::gStackTop->isArray())
               *core::gStackTop = core::Value::makeScalar(core::gStackTop->array.size);
            else if(core::gStackTop->isPackedArray())
               *core::gStackTop = core::Value::makeScalar(core::gStackTop->packedArray.size);
            else if(core::gStackTop->isData())
            {
               *core::gStackTop = core::Value::makeScalar(
                  1 << (core::gStackTop->builtin.builtinType - core::Value::DATA1));
            }
            else if(core::gStackTop->isLargeData())
            {
               *core::gStackTop = core::Value::makeScalar(
                  1 << core::gStackTop->largeData.size);
            }
            else
            {
               operations::getMember(operations::BUILTIN_SIZE);
               goto check_error;
            }
            break;

         case core::AND:
            if(core::gStackTop[1].isBuiltin() || core::gStackTop[1].isError())
               operations::doAnd();
            else
               applyMember(operations::BUILTIN_AND);
            goto check_error;

         case core::OR:
            if(core::gStackTop[1].isBuiltin() || core::gStackTop[1].isError())
               operations::doOr();
            else
               applyMember(operations::BUILTIN_OR);
            goto check_error;

         case core::ADD:
            if(core::gStackTop[1].isBuiltin())
               operations::doAdd();
            else
               applyMember(operations::BUILTIN_ADD);
            goto check_error;

         case core::SUBTRACT:
            if(core::gStackTop[1].isBuiltin())
               operations::doSubtract();
            else
               applyMember(operations::BUILTIN_SUBTRACT);
            goto check_error;

         case core::MULTIPLY:
            if(core::gStackTop[1].isBuiltin())
               operations::doMultiply();
            else
               applyMember(operations::BUILTIN_MULTIPLY);
            goto check_error;

         case core::DIVIDE:
            if(core::gStackTop[1].isBuiltin())
               operations::doDivide();
            else
               applyMember(operations::BUILTIN_DIVIDE);
            goto check_error;

         case core::MODULUS:
            if(core::gStackTop[1].isBuiltin())
               operations::doModulus();
            else
               applyMember(operations::BUILTIN_MODULUS);
            goto check_error;

         case core::POWER:
            if(core::gStackTop[1].isBuiltin())
               operations::doPower();
            else
               applyMember(operations::BUILTIN_POWER);
            goto check_error;

         case core::EQUALS:
            if(core::gStackTop[1].isBuiltin())
               operations::doEquals();
            else
               applyMember(operations::BUILTIN_EQUALS);
            goto check_error;

         case core::IS_LESS_THAN:
            if(core::gStackTop[1].isBuiltin())
               operations::doIsLessThan();
            else
               applyMember(operations::BUILTIN_IS_LESS_THAN);
            goto check_error;

         case core::IS_GREATER_THAN:
            if(core::gStackTop[1].isBuiltin())
               operations::doIsGreaterThan();
            else
               applyMember(operations::BUILTIN_IS_GREATER_THAN);
            goto check_error;

         case core::GET_ELEMENT:
            if(core::gStackTop[1].isArray() || core::gStackTop[1].isPackedArray())
               arrays::getElement();
            else
               applyMember(operations::BUILTIN_GET_ELEMENT);
            goto check_error;

         default:
            FW_THROW(spnew FatalError($("Invalid opcode.")));
            break;
         }

         goto end_loop;

      check_error:
         if(core::gStackTop->isError()
         && core::gStackTop->runtimeError.location == null)
            core::gStackTop->runtimeError.location = instruction;

      end_loop:
#ifdef DEBUG
         if(instruction - tag->instructions >= tag->instructionCount)
            FW_THROW(spnew FatalError($("Instruction pointer left bounds of function.")));
#endif
         ++instruction;
      }
   }

   void setTickCallback(void (*tick)())
   {
      gTickCallback = tick;
      gTickCount = 0;
   }
}

// ===================================================================

} //namespace vm
} //namespace evlan
