/* Evlan - the Evlan programming language.
 * Copyright (C) 2002-2005 Kenton Varda
 * <http://www.evlan.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/03/01 (KV)
 *  + First Version
 */

// ===================================================================

#include "translator.h"
#include "memory.h"
#include "operations.h"
#include "arrays.h"
#include <framework/util/Vector.h>
#include <stack>
#include <algorithm>

namespace evlan {
namespace vm {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement evlan::vm::translator
// TODO (KV): Review evlan::vm::translator

namespace translator
{
/* You'll need to #define INCLUDE_WRITER_CODE on compile for
 * WRITE_DEBUG_STUFF to work. */
//#define WRITE_DEBUG_STUFF

#ifdef WRITE_DEBUG_STUFF
#define FAIL(MESSAGE) gDebugWriter->write($("FAILURE: " MESSAGE "\n"))

   using framework::io;

   implementation DebugWriter: implements TextWriter
   {
      DebugWriter(): mMessage(Debug::log()) {}
      ~DebugWriter() {}

      // TextWriter --------------------------------------------------
      void write(String text) {mMessage << text;}
      void write(char c, int count) {mMessage << String(c, count);}

      void flush() {}
      void close() {}

      // internal ----------------------------------------------------
      Debug::Message mMessage;
   };
   TextWriterPtr gDebugWriter;

#else
#define FAIL(MESSAGE) FW_THROW(spnew FatalError($(MESSAGE)))
#endif

   FW_PATTERN_START(InternalInstruction)
      // immediates -----------------------------------------------
      FW_PATTERN_VALUE_0(loadTrue)
      FW_PATTERN_VALUE_0(loadFalse)

      FW_PATTERN_VALUE_1(loadInteger, int32)
      FW_PATTERN_VALUE_1(loadInteger64, int64)

      FW_PATTERN_VALUE_1(loadReal, double)

      FW_PATTERN_VALUE_1(loadData1, uint8)
      FW_PATTERN_VALUE_1(loadData2, uint8)
      FW_PATTERN_VALUE_1(loadData4, uint8)
      FW_PATTERN_VALUE_1(loadData8, uint8)
      FW_PATTERN_VALUE_1(loadData16, uint16)
      FW_PATTERN_VALUE_1(loadData32, uint32)
      FW_PATTERN_VALUE_1(loadData64, uint64)
      FW_PATTERN_VALUE_2(loadData, int /* lg(bitCount) */,
                                   int /* data table index */)

      FW_PATTERN_VALUE_1(loadCharacter, uint32)

      FW_PATTERN_VALUE_1(loadAtom, int /* atom table index */)

      FW_PATTERN_VALUE_3(loadArray, int /* type */, int /* size */,
                                    int /* data table index */)

      // environmentals -------------------------------------------
      FW_PATTERN_VALUE_1(loadImport, int /* import table index */)
      FW_PATTERN_VALUE_1(loadParameter, int /* parameter number */)

      // computations ---------------------------------------------
      FW_PATTERN_VALUE_1(makeArray, int /* size */)
      FW_PATTERN_VALUE_1(makeObject, int /* member table index */)
      FW_PATTERN_VALUE_2(makeFunction, Array<const String> /* parameters */,
                                       String /* documentation */)
      FW_PATTERN_VALUE_1(loadMember, int /* member name index */)
      FW_PATTERN_VALUE_0(apply)
      FW_PATTERN_VALUE_0(testCondition)

      // optimized instructions -----------------------------------
      FW_PATTERN_VALUE_0(add)
      FW_PATTERN_VALUE_0(subtract)
      FW_PATTERN_VALUE_0(multiply)
      FW_PATTERN_VALUE_0(divide)
      FW_PATTERN_VALUE_0(modulus)
      FW_PATTERN_VALUE_0(power)
      FW_PATTERN_VALUE_0(negate)
      FW_PATTERN_VALUE_0(and_)
      FW_PATTERN_VALUE_0(or_)
      FW_PATTERN_VALUE_0(not_)
      FW_PATTERN_VALUE_0(equals)
      FW_PATTERN_VALUE_0(isLessThan)
      FW_PATTERN_VALUE_0(isGreaterThan)
      FW_PATTERN_VALUE_0(size)
      FW_PATTERN_VALUE_0(getElement)
   FW_PATTERN_END

   struct InstructionInfo
   {
      InternalInstruction instruction;
      Array<const int> inputs;
      Array<int> outputs;

      int functionId;
      int depth;

      /* If this instruction is a makeFunction, functionId is the id
       * of the function it is within, and subFunctionId is the id of
       * the function being made. */
      int subFunctionId;

      int tableIndex;
      int stackIndex;
      int usageCount;   /* Once this reaches outputs.getSize(), the
                         * value can be removed from the stack frame. */

      int cycleSet;

      int branch;
      int branchLevel;

      inline InstructionInfo()
         : functionId(-1), depth(-1), subFunctionId(-1),
           tableIndex(-1), stackIndex(-1), usageCount(-1),
           cycleSet(-1), branch(-1), branchLevel(-1) {}
   };

   struct FunctionInfo
   {
      Array<int> parameters;       //index to loadParameter instructions
      Array<int> environmentTable;
      Array<core::Instruction> code;
      Array<core::SourceLocation> sourceLocations;

      int instruction;     //index of makeFunction instruction
      int arity;

      Array<const String> parameterNames;
      int line;
      int column;

      String documentation;

      inline FunctionInfo()
         : instruction(-1), arity(-1) {}
   };

   struct CycleSet
   {
      Array<int> members;
      Array<int> inputs;
   };

   struct OrderedInstruction
   {
      enum Type
      {
         PUSH_EMPTY         = 1,
         LOAD_ENVIRONMENT   = 2,
         LOAD_STACK         = 3,

         COMPUTE            = 1 << 3,
         POP                = 1 << 4,
         STORE              = 1 << 5,
         MULTI              = 1 << 6
      };

      int type;
      int index;

      inline OrderedInstruction() {}
      inline OrderedInstruction(int type, int index)
         : type(type), index(index) {}
   };

   struct ConditionalBranch
   {
      int parent;
      int level;
      int instruction;
      bool side;
      int frameSize;
      int frameOffset;

      util::Vector<OrderedInstruction> orderedInstructions;
      util::Vector<int> frameContents;

      inline ConditionalBranch() {}
      inline ConditionalBranch(int parent, int level, int instruction, bool side)
         : parent(parent), level(level), instruction(instruction), side(side),
           frameSize(0), frameOffset(0) {}
   };

   inline bool isLoadParameter(const InternalInstruction& instruction)
   {
      return instruction.getId() ==
         InternalInstruction::loadParameter_ID;
   }

   inline bool isFunction(const InternalInstruction& instruction)
   {
      return instruction.getId() ==
         InternalInstruction::makeFunction_ID;
   }

   inline bool isConditional(const InternalInstruction& instruction)
   {
      return instruction.getId() ==
         InternalInstruction::testCondition_ID;
   }

   inline bool isTrivial(const InternalInstruction& instruction)
   {
      switch(instruction.getId())
      {
      case InternalInstruction::loadTrue_ID:
      case InternalInstruction::loadFalse_ID:
      case InternalInstruction::loadInteger_ID:
      case InternalInstruction::loadInteger64_ID:
      case InternalInstruction::loadReal_ID:
      case InternalInstruction::loadData1_ID:
      case InternalInstruction::loadData2_ID:
      case InternalInstruction::loadData4_ID:
      case InternalInstruction::loadData8_ID:
      case InternalInstruction::loadData16_ID:
      case InternalInstruction::loadData32_ID:
      case InternalInstruction::loadData64_ID:
      case InternalInstruction::loadData_ID:
      case InternalInstruction::loadCharacter_ID:
      case InternalInstruction::loadAtom_ID:
      case InternalInstruction::loadArray_ID:
         return true;
      default:
         return false;
      }
   }

   class Recompiler
   {
   public:
      Recompiler(compiler::BytecodePtr bytecode);

      core::Module* recompile();

   private:
      //returns environment table for the function
      Array<int> separateFunctions(int root, int functionId, int depth);
      void findCycles();
      void groupBranches(int instruction, int currentBranch);
      void order(int instruction, bool pop);
      void determineFrameSize(int branch);
      void linearize(int branch, int offset, bool canTailRecurse);

      compiler::BytecodePtr mBytecode;
      Array<InstructionInfo> mInstructionInfo;
      Array<FunctionInfo> mFunctionInfo;
      Array<CycleSet> mCycleSets;
      Array<core::Instruction> mRootCode;
      Array<core::SourceLocation> mRootSourceLocations;
      Array<int> mMemberNameMap;
      Array<int> mAtomMap;

      util::Vector<core::Instruction> mCode;
      util::Vector<core::SourceLocation> mSourceLocations;
      util::Vector<ConditionalBranch> mCurrentBranches;
      int mCurrentFunction;
   };

   Recompiler::Recompiler(compiler::BytecodePtr bytecode)
      : mBytecode(bytecode) {}

   core::Module* Recompiler::recompile()
   {
      mInstructionInfo = Array<InstructionInfo>(mBytecode->instructions.getSize());
      int constantCount = 0;

      mMemberNameMap = Array<int>(mBytecode->memberNameTable.getSize());
      for(int i = 0; i < mMemberNameMap.getSize(); i++)
         mMemberNameMap[i] = operations::getMemberNameId(mBytecode->memberNameTable[i]);

      mAtomMap = Array<int>(mBytecode->atomTable.getSize());
      for(int i = 0; i < mAtomMap.getSize(); i++)
         mAtomMap[i] = operations::getAtomId(mBytecode->atomTable[i]);

      //convert instructions to internal instructions
      for(int i = 0; i < mBytecode->instructions.getSize(); i++)
      {
         mInstructionInfo[i].inputs = mBytecode->instructions[i].inputs;
         FW_MATCH_START(mBytecode->instructions[i].type)
         // immediates -----------------------------------------------------
         FW_MATCH_CASE_0(compiler::Bytecode::Instruction::Type::loadTrue)
            mInstructionInfo[i].instruction = InternalInstruction::loadTrue();
         FW_MATCH_CASE_0(compiler::Bytecode::Instruction::Type::loadFalse)
            mInstructionInfo[i].instruction = InternalInstruction::loadFalse();
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadInteger, n)
            mInstructionInfo[i].instruction = InternalInstruction::loadInteger(n);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadInteger64, n)
            mInstructionInfo[i].instruction = InternalInstruction::loadInteger64(n);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadReal, x)
            mInstructionInfo[i].instruction = InternalInstruction::loadReal(x);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadData1, d)
            mInstructionInfo[i].instruction = InternalInstruction::loadData1(d);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadData2, d)
            mInstructionInfo[i].instruction = InternalInstruction::loadData2(d);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadData4, d)
            mInstructionInfo[i].instruction = InternalInstruction::loadData4(d);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadData8, d)
            mInstructionInfo[i].instruction = InternalInstruction::loadData8(d);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadData16, d)
            mInstructionInfo[i].instruction = InternalInstruction::loadData16(d);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadData32, d)
            mInstructionInfo[i].instruction = InternalInstruction::loadData32(d);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadData64, d)
            mInstructionInfo[i].instruction = InternalInstruction::loadData64(d);
         FW_MATCH_CASE_2(compiler::Bytecode::Instruction::Type::loadData, size, index)
            mInstructionInfo[i].instruction = InternalInstruction::loadData(size, index);
            mInstructionInfo[i].tableIndex = constantCount++;
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadCharacter, c)
            mInstructionInfo[i].instruction = InternalInstruction::loadCharacter(c);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadAtom, index)
            mInstructionInfo[i].instruction = InternalInstruction::loadAtom(index);
         FW_MATCH_CASE_3(compiler::Bytecode::Instruction::Type::loadArray, type, size, index)
            mInstructionInfo[i].instruction = InternalInstruction::loadArray(type, size, index);
            mInstructionInfo[i].tableIndex = constantCount++;

         // environmentals -------------------------------------------------
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadImport, index)
            mInstructionInfo[i].instruction = InternalInstruction::loadImport(index);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadParameter, index)
            mInstructionInfo[i].instruction = InternalInstruction::loadParameter(index);

         // computations ---------------------------------------------------
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::makeArray, size)
            mInstructionInfo[i].instruction = InternalInstruction::makeArray(size);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::makeObject, index)
            mInstructionInfo[i].instruction = InternalInstruction::makeObject(index);
         FW_MATCH_CASE_2(compiler::Bytecode::Instruction::Type::makeFunction, parameters, doc)
            mInstructionInfo[i].instruction = InternalInstruction::makeFunction(parameters, doc);
         FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadMember, index)
            mInstructionInfo[i].instruction = InternalInstruction::loadMember(index);

            if(mBytecode->memberNameTable[index] == $("negate"))
               mInstructionInfo[i].instruction = InternalInstruction::negate();
            else if(mBytecode->memberNameTable[index] == $("not"))
               mInstructionInfo[i].instruction = InternalInstruction::not_();
            else if(mBytecode->memberNameTable[index] == $("size"))
               mInstructionInfo[i].instruction = InternalInstruction::size();

         FW_MATCH_CASE_0(compiler::Bytecode::Instruction::Type::apply)
            mInstructionInfo[i].instruction = InternalInstruction::apply();

            if(mBytecode->instructions[i].inputs.getSize() == 2)
            {
               int function = mBytecode->instructions[i].inputs[0];
               FW_MATCH_START(mBytecode->instructions[function].type)
               FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadMember, index)
                  bool optimized = true;
                  if(mBytecode->memberNameTable[index] == $("add"))
                     mInstructionInfo[i].instruction = InternalInstruction::add();
                  else if(mBytecode->memberNameTable[index] == $("subtract"))
                     mInstructionInfo[i].instruction = InternalInstruction::subtract();
                  else if(mBytecode->memberNameTable[index] == $("multiply"))
                     mInstructionInfo[i].instruction = InternalInstruction::multiply();
                  else if(mBytecode->memberNameTable[index] == $("divide"))
                     mInstructionInfo[i].instruction = InternalInstruction::divide();
                  else if(mBytecode->memberNameTable[index] == $("modulus"))
                     mInstructionInfo[i].instruction = InternalInstruction::modulus();
                  else if(mBytecode->memberNameTable[index] == $("raiseToPower"))
                     mInstructionInfo[i].instruction = InternalInstruction::power();
                  else if(mBytecode->memberNameTable[index] == $("and"))
                     mInstructionInfo[i].instruction = InternalInstruction::and_();
                  else if(mBytecode->memberNameTable[index] == $("or"))
                     mInstructionInfo[i].instruction = InternalInstruction::or_();
                  else if(mBytecode->memberNameTable[index] == $("equals"))
                     mInstructionInfo[i].instruction = InternalInstruction::equals();
                  else if(mBytecode->memberNameTable[index] == $("isLessThan"))
                     mInstructionInfo[i].instruction = InternalInstruction::isLessThan();
                  else if(mBytecode->memberNameTable[index] == $("isGreaterThan"))
                     mInstructionInfo[i].instruction = InternalInstruction::isGreaterThan();
                  else if(mBytecode->memberNameTable[index] == $("getElement"))
                     mInstructionInfo[i].instruction = InternalInstruction::getElement();
                  else
                     optimized = false;
                  if(optimized)
                  {
                     Array<int> newInputs(2);
                     newInputs[0] = mBytecode->instructions[function].inputs[0];
                     newInputs[1] = mBytecode->instructions[i].inputs[1];
                     mInstructionInfo[i].inputs = newInputs.toConst();
                  }
               FW_MATCH_END
            }

         FW_MATCH_CASE_0(compiler::Bytecode::Instruction::Type::testCondition)
            mInstructionInfo[i].instruction = InternalInstruction::testCondition();
         FW_MATCH_END
      }

      // =============================================================
      //make output lists for each instruction
      {
         /* This code looks ugly because I wanted to allocate space
          * for the output lists for *all* instructions at once,
          * rather than do an allocation for each instruction.  Of
          * course, this is really pretty silly of me, since an
          * allocation has to be done for each instruction anyway
          * because the instructions are patterns.  Thus, this is an
          * excellent example of what C coders do all too often:
          * over-optimize one function, producing ugly code, even
          * though the gains are irrelevant in the end!  (doh) */
         int inputCount = 0;
         int functionCount = 0;
         Array<int> outputCounts(mInstructionInfo.getSize());

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            outputCounts[i] = 0;

            if(isFunction(mInstructionInfo[i].instruction))
               mInstructionInfo[i].subFunctionId = functionCount++;
         }
         mFunctionInfo = Array<FunctionInfo>(functionCount);

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            inputCount += mInstructionInfo[i].inputs.getSize();
            for(int j = 0; j < mInstructionInfo[i].inputs.getSize(); j++)
               ++outputCounts[mInstructionInfo[i].inputs[j]];

            FW_MATCH_START(mInstructionInfo[i].instruction)
            FW_MATCH_CASE_2(InternalInstruction::makeFunction, parameterNames, documentation)
               mFunctionInfo[mInstructionInfo[i].subFunctionId].instruction = i;
               mFunctionInfo[mInstructionInfo[i].subFunctionId].arity = parameterNames.getSize();
               mFunctionInfo[mInstructionInfo[i].subFunctionId].parameterNames = parameterNames;
               mFunctionInfo[mInstructionInfo[i].subFunctionId].line = mBytecode->instructions[i].line;
               mFunctionInfo[mInstructionInfo[i].subFunctionId].column = mBytecode->instructions[i].column;
               mFunctionInfo[mInstructionInfo[i].subFunctionId].documentation = documentation;
            FW_MATCH_DEFAULT
               //ignore
            FW_MATCH_END

            FW_MATCH_START(mInstructionInfo[i].instruction)
            FW_MATCH_CASE_1(InternalInstruction::loadImport, id)
               mInstructionInfo[i].functionId = -1;
            FW_MATCH_DEFAULT
               if(mBytecode->instructions[i].function == -1)
                  mInstructionInfo[i].functionId = -1;
               else
               {
                  mInstructionInfo[i].functionId =
                     mInstructionInfo[mBytecode->instructions[i].function].subFunctionId;
               }
            FW_MATCH_END
         }

         int pos = 0;
         Array<int> outputs(inputCount);
         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            mInstructionInfo[i].outputs = outputs.getInterval(pos, pos + outputCounts[i]);
            pos += outputCounts[i];
            outputCounts[i] = 0;
         }

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            for(int j = 0; j < mInstructionInfo[i].inputs.getSize(); j++)
            {
               int input = mInstructionInfo[i].inputs[j];
               mInstructionInfo[input].outputs[outputCounts[input]++] = i;
            }
         }

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            FW_MATCH_START(mInstructionInfo[i].instruction)
            FW_MATCH_CASE_1(InternalInstruction::loadParameter, index)
               mInstructionInfo[i].inputs = null;

            FW_MATCH_CASE_1(compiler::Bytecode::Instruction::Type::loadMember, index)
               /* hackity hack hack:  this instruction was optimized out
                * because it is a member access for a builtin funciton,
                * and thus has been converted to that builtin instruction.
                * (i.e. x.add(y) becomes add(x,y).)  Remove its inputs to,
                * uh, cut it off. */
               if(mInstructionInfo[i].outputs.getSize() == 0
               && mBytecode->rootInstruction != i)
                  mInstructionInfo[i].inputs = null;

            FW_MATCH_CASE_1(InternalInstruction::makeFunction, parameterNames)
               if(mInstructionInfo[i].outputs.getSize() != 0)
               {
                  /* Yet another example of needless optimization:  The
                   * parameters are currently mixed in with the outputs.
                   * We want to separate these into two arrays, one for
                   * parameters and one for outputs.  Of course, we can
                   * use the same space to store both of these resulting
                   * arrays since the original will no longer be needed
                   * afterwards.  So, we do an in-place pivot, kinda like
                   * you'd see as part of a quicksort algorithm, in order
                   * to make the outputs appear at the beginning of the
                   * array and the parameters appear at the end.
                   *
                   * (Why did I do this extra optimization if it was
                   * needless?  For fun!  Duh.) */
                  int outputPos = 0;
                  int parameterPos = mInstructionInfo[i].outputs.getSize();

                  int currentOutput = mInstructionInfo[i].outputs[0];
                  while(true)
                  {
                     if(isLoadParameter(mInstructionInfo[currentOutput].instruction))
                     {
                        --parameterPos;
                        int nextOutput = mInstructionInfo[i].outputs[parameterPos];
                        mInstructionInfo[i].outputs[parameterPos] = currentOutput;
                        if(outputPos == parameterPos)
                           break;
                        currentOutput = nextOutput;
                     }
                     else
                     {
                        mInstructionInfo[i].outputs[outputPos] = currentOutput;
                        ++outputPos;
                        if(outputPos == parameterPos)
                           break;
                        currentOutput = mInstructionInfo[i].outputs[outputPos];
                     }
                  }

                  mFunctionInfo[mInstructionInfo[i].subFunctionId].parameters =
                     mInstructionInfo[i].outputs.getInterval(parameterPos);
                  mInstructionInfo[i].outputs = mInstructionInfo[i].outputs.getInterval(0, parameterPos);
               }
            FW_MATCH_END
         }
      }

      // =============================================================
      //separate functions
      //this should return null
//      if(separateFunctions(mBytecode->rootInstruction, -1, 0) != null)
//         FAIL("separateFunctions() on root returned non-null.");

      //find environment tables
      {
         Array<util::Vector<int> > environmentTables(mFunctionInfo.getSize());

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            FW_MATCH_START(mInstructionInfo[i].instruction)
            FW_MATCH_CASE_1(InternalInstruction::makeFunction, parameterNames)
               int input = mBytecode->instructions[i].inputs[0];
               if(mInstructionInfo[input].functionId != mInstructionInfo[i].subFunctionId)
                  environmentTables[mInstructionInfo[i].subFunctionId].pushBack(input);
            FW_MATCH_DEFAULT
               for(int j = 0; j < mInstructionInfo[i].inputs.getSize(); j++)
               {
                  int input = mInstructionInfo[i].inputs[j];
                  if(mInstructionInfo[input].functionId != mInstructionInfo[i].functionId)
                     environmentTables[mInstructionInfo[i].functionId].pushBack(input);
               }
            FW_MATCH_END
         }

         for(int i = 0; i < mFunctionInfo.getSize(); i++)
         {
            int parent = mInstructionInfo[mFunctionInfo[i].instruction].functionId;

            for(int j = 0; j < environmentTables[i].getSize(); j++)
            {
               int instruction = environmentTables[i][j];
               int functionId = mInstructionInfo[instruction].functionId;
               int currentParent = parent;

               while(currentParent != functionId)
               {
                  environmentTables[currentParent].pushBack(instruction);

                  if(currentParent > i)
                     break;  //we'll get to it later

                  currentParent = mInstructionInfo[
                     mFunctionInfo[currentParent].instruction].functionId;
               }
            }
         }

         for(int i = 0; i < mFunctionInfo.getSize(); i++)
         {
            if(environmentTables[i].getSize() > 1)
            {
               std::sort(&environmentTables[i][0],
                         &environmentTables[i][0] + environmentTables[i].getSize());

               int pos = 1;
               for(int j = 1; j < environmentTables[i].getSize(); j++)
               {
                  if(environmentTables[i][j] != environmentTables[i][j-1])
                  {
                     if(j > pos)
                        environmentTables[i][pos] = environmentTables[i][j];
                     ++pos;
                  }
               }

               mFunctionInfo[i].environmentTable = environmentTables[i].getArray().getInterval(0, pos);
            }
            else
               mFunctionInfo[i].environmentTable = environmentTables[i].getArray();
         }
      }

      // =============================================================
      //some housekeeping

      //remove trivial instructions from environment tables
      for(int i = 0; i < mFunctionInfo.getSize(); i++)
      {
         int pos = 0;
         for(int j = 0; j < mFunctionInfo[i].environmentTable.getSize(); j++)
         {
            int instruction = mFunctionInfo[i].environmentTable[j];
            if(isTrivial(mInstructionInfo[instruction].instruction))
            {
               //we want to remove this one from the list
            }
            else if(pos != j)
               mFunctionInfo[i].environmentTable[pos++] = instruction;
            else
               ++pos;
         }
         mFunctionInfo[i].environmentTable =
            mFunctionInfo[i].environmentTable.getInterval(0, pos);
         mInstructionInfo[mFunctionInfo[i].instruction].inputs =
            mFunctionInfo[i].environmentTable.toConst();
      }

      /* now we have to redo the stupid output lists so that they only
       * point to instructions within the same function.  In
       * particular, if a makeFunction instruction uses some other
       * instruction on its environment table, and the instructions
       * are part of the same function, the later instruction should
       * list the makeFunction as an output. */
      {
         for(int i = 0; i < mInstructionInfo.getSize(); i++)
            mInstructionInfo[i].outputs = null;

         int outputCount = 0;
         Array<int> outputCounts(mInstructionInfo.getSize());

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
            outputCounts[i] = 0;

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            Array<const int> inputs = mInstructionInfo[i].inputs;

            for(int j = 0; j < inputs.getSize(); j++)
            {
               if(mInstructionInfo[inputs[j]].functionId == mInstructionInfo[i].functionId)
               {
                  ++outputCounts[inputs[j]];
                  ++outputCount;
               }
            }
         }

         int pos = 0;
         Array<int> outputs(outputCount);
         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            mInstructionInfo[i].outputs = outputs.getInterval(pos, pos + outputCounts[i]);
            pos += outputCounts[i];
            outputCounts[i] = 0;
         }

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            Array<const int> inputs = mInstructionInfo[i].inputs;

            for(int j = 0; j < inputs.getSize(); j++)
            {
               if(mInstructionInfo[inputs[j]].functionId == mInstructionInfo[i].functionId)
                  mInstructionInfo[inputs[j]].outputs[outputCounts[inputs[j]]++] = i;
            }
         }
      }

      // =============================================================
      //compile to linear instructions!

      findCycles();

#ifdef WRITE_DEBUG_STUFF
      {
         for(int i = 0; i < mCycleSets.getSize(); i++)
         {
            gDebugWriter->write($("cycle set "));
            gDebugWriter->write(String(i));
            gDebugWriter->write($(": {"));
            for(int j = 0; j < mCycleSets[i].members.getSize(); j++)
            {
               if(j > 0)
                  gDebugWriter->write($(", "));
               gDebugWriter->write(String(mCycleSets[i].members[j]));
            }
            gDebugWriter->write($("} inputs = {"));
            for(int j = 0; j < mCycleSets[i].inputs.getSize(); j++)
            {
               if(j > 0)
                  gDebugWriter->write($(", "));
               if(mCycleSets[i].inputs[j] >= 0)
                  gDebugWriter->write(String(mCycleSets[i].inputs[j]));
               else
               {
                  gDebugWriter->write($("("));
                  gDebugWriter->write(String(~mCycleSets[i].inputs[j]));
                  gDebugWriter->write($(")"));
               }
            }
            gDebugWriter->write($("}\n"));
         }
      }
#endif

      for(int i = -1; i < mFunctionInfo.getSize(); i++)
      {
         mCurrentFunction = i;
         int rootInstruction;
         if(i == -1)
            rootInstruction = mBytecode->rootInstruction;
         else
            rootInstruction = mBytecode->instructions[mFunctionInfo[i].instruction].inputs[0];

         mCurrentBranches.pushBack(ConditionalBranch(-1, 0, -1, true));

         groupBranches(rootInstruction, 0);
         order(rootInstruction, false);

         //set table indices for environment
         if(i >= 0)
         {
            for(int j = 0; j < mFunctionInfo[i].environmentTable.getSize(); j++)
               mInstructionInfo[mFunctionInfo[i].environmentTable[j]].tableIndex = j;
         }

         determineFrameSize(0);

#ifdef WRITE_DEBUG_STUFF
         {
            Debug::Message message = Debug::log();
            gDebugWriter->write($("function "));
            gDebugWriter->write(String(i));
            gDebugWriter->write($(":\n"));
            for(int j = 0; j < mCurrentBranches.getSize(); j++)
            {
               gDebugWriter->write($("   branch "));
               gDebugWriter->write(String(j));
               gDebugWriter->write($(" (frame size: "));
               gDebugWriter->write(String(mCurrentBranches[j].frameSize));
               gDebugWriter->write($("):\n"));
               for(int k = 0; k < mCurrentBranches[j].orderedInstructions.getSize(); k++)
               {
                  int instruction = mCurrentBranches[j].orderedInstructions[k].index;
                  int type = mCurrentBranches[j].orderedInstructions[k].type;
                  if(type == OrderedInstruction::PUSH_EMPTY)
                  {
                     gDebugWriter->write($("      pe   "));
                     instruction = ~instruction;
                  }
                  else if(type == OrderedInstruction::LOAD_ENVIRONMENT)
                     gDebugWriter->write($("      le   "));
                  else if(type == OrderedInstruction::LOAD_STACK)
                     gDebugWriter->write($("      ls   "));
                  else
                  {
                     int count = 0;
                     gDebugWriter->write($("      "));
                     if(type & OrderedInstruction::COMPUTE)
                     {
                        gDebugWriter->write($("c"));
                        ++count;
                     }
                     if(type & OrderedInstruction::POP)
                     {
                        gDebugWriter->write($("p"));
                        ++count;
                     }
                     if(type & OrderedInstruction::STORE)
                     {
                        gDebugWriter->write($("s"));
                        ++count;
                     }
                     if(type & OrderedInstruction::MULTI)
                     {
                        gDebugWriter->write($("m"));
                        ++count;
                     }
                     while(count++ < 5)
                        gDebugWriter->write($(" "));
                  }

                  compiler::Bytecode::Instruction temp = mBytecode->instructions[instruction];
                  temp.inputs = mInstructionInfo[instruction].inputs;
                  gDebugWriter->write(String(instruction));
                  gDebugWriter->write($(": "));
                  compiler::writeInstruction(gDebugWriter, temp);
                  gDebugWriter->write($(" -> {"));
                  for(int l = 0; l < mInstructionInfo[instruction].outputs.getSize(); l++)
                  {
                     if(l > 0)
                        gDebugWriter->write($(", "));
                     gDebugWriter->write(String(mInstructionInfo[instruction].outputs[l]));
                  }
                  gDebugWriter->write($("}, fid "));
                  gDebugWriter->write(String(mInstructionInfo[instruction].functionId));
                  gDebugWriter->write($(", branch "));
                  gDebugWriter->write(String(mInstructionInfo[instruction].branch));
                  gDebugWriter->write($("\n"));
               }
            }
         }
#endif

         linearize(0, 0, true);

         if(mCode.getSize() != mSourceLocations.getSize())
            FW_THROW(spnew FatalError($("Not the same number of instructions and source locations.")));

         if(i == -1)
         {
            mRootCode = mCode.getArray();
            mRootSourceLocations = mSourceLocations.getArray();
         }
         else
         {
            mFunctionInfo[i].code = mCode.getArray();
            mFunctionInfo[i].sourceLocations = mSourceLocations.getArray();
         }

         mCurrentBranches.clear();
         mCode.clear();
         mSourceLocations.clear();
      }

      // =============================================================
      //allocate and fill in the module
      {
         int instructionCounts[mFunctionInfo.getSize() + 1];
         int arities[mFunctionInfo.getSize() + 1];
         int functionDocSizes[mFunctionInfo.getSize() + 1];
         for(int i = 0; i < mFunctionInfo.getSize(); i++)
         {
            instructionCounts[i] = mFunctionInfo[i].code.getSize();
            arities[i] = mFunctionInfo[i].arity;
            functionDocSizes[i] = mFunctionInfo[i].documentation.getLength();
         }
         instructionCounts[mFunctionInfo.getSize()] = mRootCode.getSize();
         arities[mFunctionInfo.getSize()] = 1;
         functionDocSizes[mFunctionInfo.getSize()] = 0;

         int memberCounts[mBytecode->classes.getSize()];
         int objectDocSizes[mBytecode->classes.getSize()];
         for(int i = 0; i < mBytecode->classes.getSize(); i++)
         {
            memberCounts[i] = mBytecode->classes[i].memberTable.getSize();
            objectDocSizes[i] = mBytecode->classes[i].documentation.getLength();
         }

         Array<int> dataOffsets(mBytecode->dataTable.getSize());
         int extraDataSize = 0;
         for(int i = 0; i < mBytecode->dataTable.getSize(); i++)
         {
            dataOffsets[i] = extraDataSize;
            int size = mBytecode->dataTable[i].getSize();
#ifdef PLATFORM_64BIT
            if(size % 2 == 1)
               ++size;
            size /= 2;
#endif
            extraDataSize += size;
         }

         core::Module* module = memory::allocateModule(
            mFunctionInfo.getSize() + 1,
            mBytecode->classes.getSize(),
            constantCount,
            extraDataSize,
            arities,
            instructionCounts,
            functionDocSizes,
            memberCounts,
            objectDocSizes,
            true);

         for(int i = 0; i < mFunctionInfo.getSize(); i++)
         {
            core::FunctionTag* tag = module->functionTags[i];
            for(int j = 0; j < mFunctionInfo[i].arity; j++)
               tag->parameterNames[j] = operations::getIdentifierId(mFunctionInfo[i].parameterNames[j]);
            tag->line = mFunctionInfo[i].line;
            tag->column = mFunctionInfo[i].column;
            tag->environmentSize = mFunctionInfo[i].environmentTable.getSize();
            memcpy(tag->instructions, &mFunctionInfo[i].code[0],
                   tag->instructionCount * sizeof(core::Instruction));
            memcpy(tag->sourceLocations, &mFunctionInfo[i].sourceLocations[0],
                   tag->instructionCount * sizeof(core::SourceLocation));

            mFunctionInfo[i].documentation.copyToBuffer(
               tag->documentation, 0, tag->documentationLength);
         }
         {
            core::FunctionTag* tag = module->functionTags[mFunctionInfo.getSize()];
            tag->parameterNames[0] = operations::getIdentifierId($("imports"));
            tag->line = 0;
            tag->column = 0;
            tag->environmentSize = 0;
            memcpy(tag->instructions, &mRootCode[0],
                   tag->instructionCount * sizeof(core::Instruction));
            memcpy(tag->sourceLocations, &mRootSourceLocations[0],
                   tag->instructionCount * sizeof(core::SourceLocation));
         }

         for(int i = 0; i < mBytecode->classes.getSize(); i++)
         {
            core::ObjectTag* tag = module->objectTags[i];
            for(int j = 0; j < tag->memberCount; j++)
            {
               tag->members[j].nameId = mMemberNameMap[mBytecode->classes[i].memberTable[j]];
               tag->members[j].offset = j;
            }
            std::sort(tag->members, tag->members + tag->memberCount);

            mBytecode->classes[i].documentation.copyToBuffer(
               tag->documentation, 0, tag->documentationLength);
         }

         Array<int8> dataElementSizes(dataOffsets.getSize());
         for(int i = 0; i < dataOffsets.getSize(); i++)
            dataElementSizes[i] = -1;

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            FW_MATCH_START(mInstructionInfo[i].instruction)
            FW_MATCH_CASE_2(InternalInstruction::loadData, size, index)
               int j = mInstructionInfo[i].tableIndex;
               module->constants[j].type = core::Value::LARGE_DATA;
               module->constants[j].largeData.size = size;
               module->constants[j].largeData.value = module->extraData + dataOffsets[index];

            FW_MATCH_CASE_3(InternalInstruction::loadArray, type, size, index)
               int j = mInstructionInfo[i].tableIndex;
               module->constants[j].type = core::Value::PACKED_ARRAY;
               switch(type)
               {
               case compiler::Bytecode::BOOLEAN:
                  module->constants[j].packedArray.packType = core::Value::PACKED_BOOLEAN;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = 0;
                  break;
               case compiler::Bytecode::INTEGER8:
                  module->constants[j].packedArray.packType = core::Value::PACKED_INTEGER;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = 3;
                  break;
               case compiler::Bytecode::INTEGER16:
                  module->constants[j].packedArray.packType = core::Value::PACKED_INTEGER;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = 4;
                  break;
               case compiler::Bytecode::INTEGER32:
                  module->constants[j].packedArray.packType = core::Value::PACKED_INTEGER;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = 5;
                  break;
               case compiler::Bytecode::INTEGER64:
                  module->constants[j].packedArray.packType = core::Value::PACKED_INTEGER;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = 6;
                  break;
               case compiler::Bytecode::REAL:
                  module->constants[j].packedArray.packType = core::Value::PACKED_SCALAR;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = 6;
                  break;
               case compiler::Bytecode::CHARACTER8:
                  module->constants[j].packedArray.packType = core::Value::PACKED_CHARACTER;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = 3;
                  break;
               case compiler::Bytecode::CHARACTER16:
                  module->constants[j].packedArray.packType = core::Value::PACKED_CHARACTER;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = 4;
                  break;
               case compiler::Bytecode::CHARACTER32:
                  module->constants[j].packedArray.packType = core::Value::PACKED_CHARACTER;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = 5;
                  break;
               default:
                  module->constants[j].packedArray.packType = core::Value::PACKED_DATA;
                  dataElementSizes[index] = module->constants[j].packedArray.elementSize = type - compiler::Bytecode::DATA;
                  break;
               }

               module->constants[j].packedArray.size = size;

               module->constants[j].packedArray.offset = 0;

               module->constants[j].packedArray.elements = module->extraData + dataOffsets[index];

            FW_MATCH_DEFAULT
               //ignore
            FW_MATCH_END
         }

         for(int i = 0; i < mBytecode->dataTable.getSize(); i++)
         {
            if(mBytecode->dataTable[i].getSize() > 0)
            {
               if(dataElementSizes[i] != -1)
               {
                  arrays::repackData(
                     module->extraData + dataOffsets[i],
                     &mBytecode->dataTable[i][0],
                     mBytecode->dataTable[i].getSize(),
                     dataElementSizes[i]);
               }
               else
               {
                  memcpy(module->extraData + dataOffsets[i],
                         &mBytecode->dataTable[i][0],
                         sizeof(uint32) * mBytecode->dataTable[i].getSize());
               }
            }
         }

         return module;
      }
   }

   // ================================================================
   Array<int> Recompiler::separateFunctions(int root, int functionId, int depth)
   {
      /*
      push root
      while stack is not empty
         pop instruction
         if its functionId is not equal to the one being scanned
            add to environment table
         else if its depth has not been set
            set depth
            if it is a function
               create secondary stack
               push parameters
               while stack is not empty
                  pop instruction
                  if instruction is not this function
                     if instruction's function is not this one
                        set it so that it is
                        push all outputs
               recursively separate this function
               push all members of the resulting environment table
            else
               push all inputs
      */

      std::stack<int> todo;
      todo.push(root);
      util::Vector<int> environmentTable;

      while(!todo.empty())
      {
         int instruction = todo.top();
         todo.pop();

         if(mInstructionInfo[instruction].depth > depth)
         {
            FW_THROW(spnew InvalidBytecode(
               $("Instruction inputs value computed in deeper function."),
               mBytecode->instructions[instruction].line,
               mBytecode->instructions[instruction].column));
         }
         else if(mInstructionInfo[instruction].functionId != functionId)
         {
            //add to environment table
            environmentTable.pushBack(instruction);
         }
         else if(mInstructionInfo[instruction].depth == -1)
         {
            mInstructionInfo[instruction].depth = depth;

            FW_MATCH_START(mInstructionInfo[instruction].instruction)
            FW_MATCH_CASE_1(InternalInstruction::makeFunction, parameterNames)
               /* Mark all instructions that depend on this function's
                * parameters as being part of the function. */
               int subFunctionId = mInstructionInfo[instruction].subFunctionId;

               {
                  std::stack<int> todo2;
                  for(int i = 0; i < mFunctionInfo[subFunctionId].parameters.getSize(); i++)
                     todo2.push(mFunctionInfo[subFunctionId].parameters[i]);
                  while(!todo2.empty())
                  {
                     int instruction2 = todo2.top();
                     todo2.pop();
                     if(instruction2 != instruction)
                     {
                        if(mInstructionInfo[instruction2].functionId != subFunctionId)
                        {
                           mInstructionInfo[instruction2].functionId = subFunctionId;
                           for(int i = 0; i < mInstructionInfo[instruction2].outputs.getSize(); i++)
                              todo2.push(mInstructionInfo[instruction2].outputs[i]);
                        }
                     }
                  }
               }

               //further separate the function
               Array<int> subEnvironment =
                  separateFunctions(mBytecode->instructions[instruction].inputs[0],
                                    subFunctionId, depth + 1);
               mFunctionInfo[subFunctionId].environmentTable = subEnvironment;
               mInstructionInfo[instruction].inputs = subEnvironment.toConst();
               for(int i = 0; i < subEnvironment.getSize(); i++)
                  todo.push(subEnvironment[i]);

            FW_MATCH_DEFAULT
               for(int i = 0; i < mInstructionInfo[instruction].inputs.getSize(); i++)
                  todo.push(mInstructionInfo[instruction].inputs[i]);
            FW_MATCH_END
         }
      }

      Array<int> result = environmentTable.getArray();

      if(result.getSize() > 0)
      {
         std::sort(&result[0], &result[0] + result.getSize());

         //remove duplicates
         int pos = 0;
         for(int i = 1; i < result.getSize(); i++)
         {
            if(result[i] != result[pos])
               result[++pos] = result[i];
         }
         result = result.getInterval(0, pos + 1);
      }

      return result;
   }

   // ================================================================
   struct CycleTask
   {
      int instruction;
      int input;

      inline CycleTask() {}
      inline CycleTask(int instruction)
         : instruction(instruction), input(-1) {}
   };

   struct CycleInfo
   {
      int setId;
      bool marked;
      bool done;

      inline CycleInfo(): setId(-1), marked(false), done(false) {}
   };

   void Recompiler::findCycles()
   {
      /* *sigh* I had to completely rewrite this function nearly a
       * year after writing it the first time.  This turned out to be
       * hard, considering how much all of this code sucks.
       *
       * Notes to self:
       * At this point, the input and output sets for each instruction
       * are not exactly opposites of each other.  They are only true
       * opposites if you ignore inputs which point at instructions in
       * different functions.  (The "outputs" ignore cross-function
       * uses.)
       *
       * Otherwise, though, if the inputs represent edges of the
       * instruction graph, the outputs represent the transpose of
       * that graph.  This is exactly what we need to use the cycle-
       * finding algorithm on page 489 of Introduction to Algorithms
       * (Cormen et al.).  Unlike my original algorithm, their
       * algorithm should actually work in all cases.
       *
       * In the future, I will check Intro to Algorithms before
       * inventing my own. */

      /* Step 1:  Perform depth-first seach.  When backtracking over
       * an element, add it to the end of an array, such that we end
       * up with an array containing all instructions in reverse
       * order of finish time.
       * Step 2:  Perform a depth-first search on the transpose,
       * preferring instructions that appear later in the previous
       * list over ones appearing earlier.  Add each instruction found
       * to yet another array, in the order found.  Note which
       * instructions are roots of a DFS tree.
       * Step 3:  Every root that has more than one non-root element
       * following is then part of a cycle with those non-roots. */

      Array<int> instructionsBySet(mInstructionInfo.getSize());
      Array<bool> isRoot(mInstructionInfo.getSize());

      {
         //step 1
         Array<int> instructionsByFinishTime(mInstructionInfo.getSize());
         Array<bool> marked(mInstructionInfo.getSize());
         int pos = 0;
         for(int i = 0; i < mInstructionInfo.getSize(); i++)
            marked[i] = false;

         util::Vector<CycleTask> todo;

         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            if(marked[i])
               continue;

            marked[i] = true;
            todo.pushBack(CycleTask(i));

            while(todo.getSize() > 0)
            {
               int instruction = todo.getBack().instruction;
               int input = ++todo.getBack().input;

               Array<const int> inputs = mInstructionInfo[instruction].inputs;
               if(input == inputs.getSize())
               {
                  instructionsByFinishTime[pos++] = instruction;
                  todo.popBack();
               }
               else if(!marked[inputs[input]]
                    && mInstructionInfo[inputs[input]].functionId == mInstructionInfo[instruction].functionId)
               {
                  marked[inputs[input]] = true;
                  todo.pushBack(CycleTask(inputs[input]));
               }
            }
         }

         if(pos != mInstructionInfo.getSize())
            FW_THROW(spnew FatalError($("Not all instructions accounted for in step 1.")));

         //step 2
         pos = 0;
         for(int i = 0; i < mInstructionInfo.getSize(); i++)
         {
            marked[i] = false;
            isRoot[i] = false;
         }

         for(int i = mInstructionInfo.getSize() - 1; i >= 0; i--)
         {
            if(marked[instructionsByFinishTime[i]])
               continue;

            marked[instructionsByFinishTime[i]] = true;
            todo.pushBack(CycleTask(instructionsByFinishTime[i]));
            isRoot[pos] = true;
            instructionsBySet[pos++] = instructionsByFinishTime[i];

            while(todo.getSize() > 0)
            {
               int instruction = todo.getBack().instruction;
               int input = ++todo.getBack().input;

               Array<int> outputs = mInstructionInfo[instruction].outputs;
               if(input == outputs.getSize())
                  todo.popBack();
               else if(!marked[outputs[input]])
               {
                  marked[outputs[input]] = true;
                  todo.pushBack(CycleTask(outputs[input]));
                  instructionsBySet[pos++] = outputs[input];
               }
            }

            if(isRoot[pos-1])
            {
               /* Didn't find any other instructions in the cycle.  Does this
                * instruction refer to itself, though? */

               --pos;
               int instruction = instructionsByFinishTime[i];
               for(int j = 0; j < mInstructionInfo[instruction].inputs.getSize(); j++)
               {
                  if(mInstructionInfo[instruction].inputs[j] == instruction)
                  {
                     //yep
                     ++pos;
                     break;
                  }
               }
            }
         }

         instructionsBySet = instructionsBySet.getInterval(0, pos);
         isRoot = isRoot.getInterval(0, pos);

#if 0
         {
            Debug::Message message = Debug::log();
            for(int i = 0; i < mInstructionInfo.getSize(); i++)
            {
               message << i;
               for(int j = 0; j < mInstructionInfo[i].inputs.getSize(); j++)
                  message << " " << mInstructionInfo[i].inputs[j];
               message << "\n";
            }

            message << "\n";

            for(int i = 0; i < instructionsByFinishTime.getSize(); i++)
               message << instructionsByFinishTime[i] << " ";

            message << "\n";

            for(int i = 0; i < instructionsBySet.getSize(); i++)
            {
               if(isRoot[i])
                  message << "root: ";
               else
                  message << "      ";
               message << instructionsBySet[i] << " "
                  << (mBytecode->instructions[instructionsBySet[i]].line+1) << ":"
                  << (mBytecode->instructions[instructionsBySet[i]].column+1);
               message << "\n";
            }
         }
#endif
      }


      {
         int pos = 0;

         for(int i = 0; i < instructionsBySet.getSize(); i++)
         {
            if(isRoot[i])
               ++pos;
         }

         Array<int> setRoots(pos);
         Array<int> setSizes(pos);
         pos = 0;

         for(int i = 0; i < instructionsBySet.getSize(); i++)
         {
            if(isRoot[i])
               setRoots[pos++] = i;
         }

         int totalSize = 0;
         for(int i = 0; i < setRoots.getSize(); i++)
         {
            int j = setRoots[i] + 1;
            while(j < instructionsBySet.getSize() && !isRoot[j])
               ++j;
            setSizes[i] = j - setRoots[i];
            totalSize += setSizes[i];
         }

         mCycleSets = Array<CycleSet>(setRoots.getSize());
         Array<int> allMembers(totalSize);
         pos = 0;

         for(int i = 0; i < setRoots.getSize(); i++)
         {
            mCycleSets[i].members = allMembers.getInterval(pos, pos + setSizes[i]);
            pos += setSizes[i];
         }

         for(int i = 0; i < setRoots.getSize(); i++)
         {
            for(int j = 0; j < setSizes[i]; j++)
            {
               int instruction = instructionsBySet[setRoots[i] + j];

               switch(mInstructionInfo[instruction].instruction.getId())
               {
               case InternalInstruction::makeArray_ID:
               case InternalInstruction::makeObject_ID:
               case InternalInstruction::makeFunction_ID:
                  //legal
                  break;
               default:
                  FW_THROW(spnew InvalidBytecode(
                     $("Illegal cycle (instruction depends on its own result)."),
                     mBytecode->instructions[instruction].line,
                     mBytecode->instructions[instruction].column));
                  break;
               }

               mCycleSets[i].members[j] = instruction;
               mInstructionInfo[instruction].cycleSet = i;
            }
         }

         instructionsBySet = null;
         isRoot = null;
      }

      //remove intra-set outputs
      for(int i = 0; i < mCycleSets.getSize(); i++)
      {
         for(int j = 0; j < mCycleSets[i].members.getSize(); j++)
         {
            int instruction = mCycleSets[i].members[j];
            Array<int> outputs = mInstructionInfo[instruction].outputs;

            int pos = 0;
            for(int k = 0; k < outputs.getSize(); k++)
            {
               if(mInstructionInfo[outputs[k]].cycleSet != i)
               {
                  if(pos < k)
                     outputs[pos] = outputs[k];
                  ++pos;
               }
            }
            mInstructionInfo[instruction].outputs = outputs.getInterval(0, pos);
         }
      }

      //create combined input sets
      for(int i = 0; i < mCycleSets.getSize(); i++)
      {
         int inputCount = 0;

         for(int j = 0; j < mCycleSets[i].members.getSize(); j++)
            inputCount += mInstructionInfo[mCycleSets[i].members[j]].inputs.getSize();

         mCycleSets[i].inputs = Array<int>(inputCount);
         int pos = 0;
         for(int j = 0; j < mCycleSets[i].members.getSize(); j++)
         {
            Array<const int> other = mInstructionInfo[mCycleSets[i].members[j]].inputs;
            for(int k = 0; k < other.getSize(); k++)
            {
               if(mInstructionInfo[other[k]].cycleSet == i)
                  mCycleSets[i].inputs[pos++] = ~other[k];
               else
                  mCycleSets[i].inputs[pos++] = other[k];
            }
         }
      }
   }

   // ================================================================
   void Recompiler::groupBranches(int instruction, int currentBranch)
   {
      /*
      if it already has a branch set
         if that branch is the same as this one
            return
         set current branch to deepest common branch

      set instruction's branch to current branch

      if it's a conditional
         allocate two branches
         set their parent to the current branch
         set one as the true side and one as the false side
         set their level to parent's level + 1
         set their instruction to this instruction

         if condition is from same function
            group condition with original branch
         if true branch is from same function
            group true branch with true branch
         if false branch is from same function
            group false branch with false branch
      else
         for all inputs
            if input is from same function
               group it with current branch
      */

      if(mInstructionInfo[instruction].functionId != mCurrentFunction)
         return;
      else if(mInstructionInfo[instruction].branch == currentBranch
           && mInstructionInfo[instruction].branchLevel == mCurrentBranches[currentBranch].level)
         return;
      else if(mInstructionInfo[instruction].branch != -1)
      {
         //find deepest common branch
         int a = mInstructionInfo[instruction].branch;
         int b = currentBranch;
         while(mCurrentBranches[a].level > mCurrentBranches[b].level)
            a = mCurrentBranches[a].parent;
         while(mCurrentBranches[a].level < mCurrentBranches[b].level)
            b = mCurrentBranches[b].parent;

         while(a != b)
         {
            a = mCurrentBranches[a].parent;
            b = mCurrentBranches[b].parent;
         }

         currentBranch = a;
      }

      mInstructionInfo[instruction].branch = currentBranch;
      mInstructionInfo[instruction].branchLevel = mCurrentBranches[currentBranch].level;

      if(isConditional(mInstructionInfo[instruction].instruction))
      {
         int trueBranch, falseBranch;
         if(mInstructionInfo[instruction].subFunctionId != -1)
         {
            trueBranch = mInstructionInfo[instruction].subFunctionId;
            falseBranch = trueBranch + 1;

            mCurrentBranches[trueBranch].parent = currentBranch;
            mCurrentBranches[trueBranch].level = mCurrentBranches[currentBranch].level + 1;
            mCurrentBranches[falseBranch].parent = currentBranch;
            mCurrentBranches[falseBranch].level = mCurrentBranches[currentBranch].level + 1;
         }
         else
         {
            trueBranch = mCurrentBranches.getSize();
            mCurrentBranches.pushBack(
               ConditionalBranch(
                  currentBranch,
                  mCurrentBranches[currentBranch].level + 1,
                  instruction,
                  true));

            falseBranch = mCurrentBranches.getSize();
            mCurrentBranches.pushBack(
               ConditionalBranch(
                  currentBranch,
                  mCurrentBranches[currentBranch].level + 1,
                  instruction,
                  false));

            mInstructionInfo[instruction].subFunctionId = trueBranch;
         }

         Array<const int> inputs = mInstructionInfo[instruction].inputs;
         groupBranches(inputs[0], currentBranch);
         groupBranches(inputs[1], trueBranch);
         groupBranches(inputs[2], falseBranch);
      }
      else
      {
         Array<const int> inputs = mInstructionInfo[instruction].inputs;

         for(int i = 0; i < inputs.getSize(); i++)
            groupBranches(inputs[i], currentBranch);
      }
   }

   // ================================================================
   struct OrderTask
   {
      int instruction;
      int input;
      int lastInputBranch;
      bool pop;
      bool needLoadStack;

      inline OrderTask() {}
      inline OrderTask(int instruction, bool pop)
         : instruction(instruction), input(-1), lastInputBranch(-1), pop(pop) {}
   };

   void Recompiler::order(int instruction, bool pop)
   {
      /* I appologize for the mess here, but at some point along the
       * line I converted this function from a recursive
       * implementation to a stack-based one, and I had to do some
       * ugly things to make it work.  I figured that recursively
       * searching the execution graph (which could be quite deep in
       * some cases) wasn't such a good idea. */
      std::stack<OrderTask> todo;
      todo.push(OrderTask(instruction, pop));

      //hack fix...
      if(isTrivial(mInstructionInfo[instruction].instruction))
      {
         mCurrentBranches[0].orderedInstructions.pushBack(
            OrderedInstruction(OrderedInstruction::COMPUTE, instruction));
         return;
      }
      else if(mInstructionInfo[instruction].functionId != mCurrentFunction)
      {
         mCurrentBranches[0].orderedInstructions.pushBack(
            OrderedInstruction(OrderedInstruction::LOAD_ENVIRONMENT, instruction));
         return;
      }
      else if(isLoadParameter(mInstructionInfo[instruction].instruction))
      {
         mCurrentBranches[0].orderedInstructions.pushBack(
            OrderedInstruction(OrderedInstruction::COMPUTE, instruction));
         return;
      }
      //end hack fix

      while(!todo.empty())
      {
         instruction = todo.top().instruction;
         pop = todo.top().pop;
         int i = ++todo.top().input;
         int branch = mInstructionInfo[instruction].branch;
         Array<const int> inputs;
         if(mInstructionInfo[instruction].cycleSet == -1)
            inputs = mInstructionInfo[instruction].inputs;
         else
            inputs = mCycleSets[mInstructionInfo[instruction].cycleSet].inputs.toConst();

#if 0
         {
            Debug::Message message = Debug::log();
            message << mBytecode->instructions[instruction].line << ":" << mBytecode->instructions[instruction].column << ": ";
            message << instruction << " " << mInstructionInfo[instruction].cycleSet;
         }
#endif

         if(todo.top().lastInputBranch != -1)
         {
            //input was created on a different branch, then stored to stack frame
            //load it on the branch where it was expected
            mCurrentBranches[todo.top().lastInputBranch].orderedInstructions.pushBack(
               OrderedInstruction(OrderedInstruction::LOAD_STACK, inputs[i-1]));
            todo.top().lastInputBranch = -1;
         }

         if(i < inputs.getSize())
         {
            int targetBranch = branch;

            if(i > 0 && isConditional(mInstructionInfo[instruction].instruction))
               targetBranch = mInstructionInfo[instruction].subFunctionId + i - 1;

            if(inputs[i] < 0)
            {
               mCurrentBranches[targetBranch].orderedInstructions.pushBack(
                  OrderedInstruction(OrderedInstruction::PUSH_EMPTY, inputs[i]));
            }
            else if(isTrivial(mInstructionInfo[inputs[i]].instruction))
            {
               mCurrentBranches[targetBranch].orderedInstructions.pushBack(
                  OrderedInstruction(OrderedInstruction::COMPUTE, inputs[i]));
            }
            else if(mInstructionInfo[inputs[i]].functionId != mInstructionInfo[instruction].functionId)
            {
               mCurrentBranches[targetBranch].orderedInstructions.pushBack(
                  OrderedInstruction(OrderedInstruction::LOAD_ENVIRONMENT, inputs[i]));
            }
            else if(isLoadParameter(mInstructionInfo[inputs[i]].instruction))
            {
               mCurrentBranches[targetBranch].orderedInstructions.pushBack(
                  OrderedInstruction(OrderedInstruction::COMPUTE, inputs[i]));
            }
            else
            {
               if(mInstructionInfo[inputs[i]].usageCount < 0)
               {
                  bool sameBranch = mInstructionInfo[inputs[i]].branch == targetBranch;
                  if(!sameBranch)
                     todo.top().lastInputBranch = targetBranch;
                  todo.push(OrderTask(inputs[i], !sameBranch));
               }
               else
               {
                  //load on this branch
                  mCurrentBranches[targetBranch].orderedInstructions.pushBack(
                     OrderedInstruction(OrderedInstruction::LOAD_STACK, inputs[i]));
               }
            }
         }
         else
         {
            if(isTrivial(mInstructionInfo[instruction].instruction))
            {
               mCurrentBranches[branch].orderedInstructions.pushBack(
                  OrderedInstruction(OrderedInstruction::COMPUTE, instruction));
            }
            else if(mInstructionInfo[instruction].functionId != mCurrentFunction)
            {
               mCurrentBranches[branch].orderedInstructions.pushBack(
                  OrderedInstruction(OrderedInstruction::LOAD_ENVIRONMENT, instruction));
            }
            else if(isLoadParameter(mInstructionInfo[instruction].instruction))
            {
               mCurrentBranches[branch].orderedInstructions.pushBack(
                  OrderedInstruction(OrderedInstruction::COMPUTE, instruction));
            }
            else
            {
               if(mInstructionInfo[instruction].cycleSet == -1)
                  mInstructionInfo[instruction].usageCount = 0;
               else
               {
                  Array<int> members =
                     mCycleSets[mInstructionInfo[instruction].cycleSet].members;
                  for(int j = 0; j < members.getSize(); j++)
                     mInstructionInfo[members[j]].usageCount = 0;
               }

               //add this instruction to the branch's instruction list
               int type = OrderedInstruction::COMPUTE;
               if(pop)
               {
                  type |= OrderedInstruction::POP;
                  if(mInstructionInfo[instruction].outputs.getSize() > 0)
                     type |= OrderedInstruction::STORE;
               }
               else if(mInstructionInfo[instruction].outputs.getSize() > 1)
                  type |= OrderedInstruction::STORE;
               if(mInstructionInfo[instruction].cycleSet != -1)
                  type |= OrderedInstruction::MULTI;
               mCurrentBranches[branch].orderedInstructions.pushBack(
                  OrderedInstruction(type, instruction));
            }

            todo.pop();
         }
      }
   }

   // ================================================================
   void Recompiler::determineFrameSize(int branch)
   {
      for(int i = 0; i < mCurrentBranches[branch].orderedInstructions.getSize(); i++)
      {
         OrderedInstruction instruction = mCurrentBranches[branch].orderedInstructions[i];

         if(instruction.type == OrderedInstruction::LOAD_STACK)
         {
            //need to load from stack frame

            //find it on the stack frame
            int other = mInstructionInfo[instruction.index].branch;
            int j;
            for(j = 0; j < mCurrentBranches[other].frameContents.getSize(); j++)
            {
               if(mCurrentBranches[other].frameContents[j] == instruction.index)
                  break;
            }

            //check to make sure the thing is on the frame
            if(j == mCurrentBranches[other].frameContents.getSize())
            {
               Debug::log(Debug::ERROR) << "Load from stack before store: " << mCurrentFunction << " - " << branch;
//               continue;
               FAIL("Load from stack before store.");
            }

            if(++mInstructionInfo[instruction.index].usageCount ==
               mInstructionInfo[instruction.index].outputs.getSize())
            {
               //final usage...  remove from frame
               mInstructionInfo[instruction.index].usageCount = 0;
               mCurrentBranches[other].frameContents[j] = -1;
            }
         }
         else if(instruction.type & OrderedInstruction::COMPUTE)
         {
            if(isConditional(mInstructionInfo[instruction.index].instruction))
            {
               determineFrameSize(mInstructionInfo[instruction.index].subFunctionId);
               determineFrameSize(mInstructionInfo[instruction.index].subFunctionId + 1);
            }

            if(instruction.type & OrderedInstruction::MULTI)
            {
               Array<int> members =
                  mCycleSets[mInstructionInfo[instruction.index].cycleSet].members;
               int k = 0;
               for(int j = 0; j < members.getSize(); j++)
               {
                  if((members[j] != instruction.index && mInstructionInfo[members[j]].outputs.getSize() > 0)
                  || (members[j] == instruction.index
                   && (j != members.getSize() - 1
                    || (instruction.type & OrderedInstruction::STORE))))
                  {
                     //find a place in the frame for this value
                     for(; k < mCurrentBranches[branch].frameContents.getSize(); k++)
                     {
                        if(mCurrentBranches[branch].frameContents[k] == -1)
                           break;
                     }

                     if(k == mCurrentBranches[branch].frameContents.getSize())
                        mCurrentBranches[branch].frameContents.pushBack(members[j]);
                     else
                        mCurrentBranches[branch].frameContents[k] = members[j];

                     mInstructionInfo[members[j]].usageCount = 0;

                     if(members[j] == instruction.index
                     && !(instruction.type & OrderedInstruction::STORE))
                        mCurrentBranches[branch].frameContents[k] = -1;
                     ++k;
                  }
               }

               if(!(instruction.type & OrderedInstruction::POP))
                  mInstructionInfo[instruction.index].usageCount = 1;
            }
            else if(instruction.type & OrderedInstruction::STORE)
            {
               if(instruction.type & OrderedInstruction::POP)
                  mInstructionInfo[instruction.index].usageCount = 0;
               else
                  mInstructionInfo[instruction.index].usageCount = 1;

               //find a place in the frame for this value
               int j;
               for(j = 0; j < mCurrentBranches[branch].frameContents.getSize(); j++)
               {
                  if(mCurrentBranches[branch].frameContents[j] == -1)
                     break;
               }

               if(j == mCurrentBranches[branch].frameContents.getSize())
                  mCurrentBranches[branch].frameContents.pushBack(instruction.index);
               else
                  mCurrentBranches[branch].frameContents[j] = instruction.index;
            }
         }
      }

      for(int i = 0; i < mCurrentBranches[branch].frameContents.getSize(); i++)
      {
         if(mCurrentBranches[branch].frameContents[i] != -1)
         {
//            Debug::log(Debug::ERROR) << "Stuff left over in frame: " << mCurrentFunction << " - " << branch;
            FAIL("Stuff left over in frame.");
         }
      }

      mCurrentBranches[branch].frameSize = mCurrentBranches[branch].frameContents.getSize();
      mCurrentBranches[branch].frameContents.clear();
   }

   // ================================================================
   void Recompiler::linearize(int branch, int offset, bool canTailRecurse)
   {
      mCurrentBranches[branch].frameOffset = offset;
      offset += mCurrentBranches[branch].frameSize;
      if(mCurrentBranches[branch].frameSize > 0)
      {
         mCode.pushBack(core::Instruction(
            core::PUSH_EMPTY, mCurrentBranches[branch].frameSize));
         mSourceLocations.pushBack(core::SourceLocation(-1, -1));
      }

#define ADD_INSTRUCTION(OPCODE, PARAMETER)                   \
      {mCode.pushBack(core::Instruction(OPCODE, PARAMETER)); \
       mSourceLocations.pushBack(sourceLocation);}

      for(int i = 0; i < mCurrentBranches[branch].orderedInstructions.getSize(); i++)
      {
         //stuff
         OrderedInstruction instruction = mCurrentBranches[branch].orderedInstructions[i];
         core::SourceLocation sourceLocation;
         if(instruction.index >= 0)
         {
            sourceLocation = core::SourceLocation(
               mBytecode->instructions[instruction.index].line,
               mBytecode->instructions[instruction.index].column);
         }
         else
         {
            sourceLocation = core::SourceLocation(
               mBytecode->instructions[~instruction.index].line,
               mBytecode->instructions[~instruction.index].column);
         }

         if(instruction.type == OrderedInstruction::LOAD_STACK)
         {
            //need to load from stack frame

            //find it on the stack frame
            int other = mInstructionInfo[instruction.index].branch;
            int j;
            for(j = 0; j < mCurrentBranches[other].frameContents.getSize(); j++)
            {
               if(mCurrentBranches[other].frameContents[j] == instruction.index)
                  break;
            }

            ADD_INSTRUCTION(core::LOAD, offset - mCurrentBranches[other].frameOffset - j);
            ++offset;

            if(++mInstructionInfo[instruction.index].usageCount ==
               mInstructionInfo[instruction.index].outputs.getSize())
            {
               //final usage...  remove from frame
               mInstructionInfo[instruction.index].usageCount = 0;
               mCurrentBranches[other].frameContents[j] = -1;
            }
         }
         else if(instruction.type == OrderedInstruction::LOAD_ENVIRONMENT)
         {
            ADD_INSTRUCTION(core::LOAD_ENVIRONMENT, mInstructionInfo[instruction.index].tableIndex);
            ++offset;
         }
         else if(instruction.type == OrderedInstruction::PUSH_EMPTY)
         {
            if(mCode.getSize() > 0 && mCode.getBack().opcode == core::PUSH_EMPTY)
               ++mCode.getBack().parameter;
            else
               ADD_INSTRUCTION(core::PUSH_EMPTY, 1);
            ++offset;
         }
         else if(instruction.type & OrderedInstruction::COMPUTE)
         {
            int location = -1;
            if(instruction.type & OrderedInstruction::MULTI)
            {
               Array<int> members =
                  mCycleSets[mInstructionInfo[instruction.index].cycleSet].members;
               Array<int> inputs =
                  mCycleSets[mInstructionInfo[instruction.index].cycleSet].inputs;

               ADD_INSTRUCTION(core::ALLOCATE_MULTI, inputs.getSize());
               offset -= inputs.getSize() - 1;

               int k = 0;
               for(int j = 0; j < members.getSize(); j++)
               {
                  FW_MATCH_START(mInstructionInfo[members[j]].instruction)
                  FW_MATCH_CASE_1(InternalInstruction::makeArray, size)
                     ADD_INSTRUCTION(core::MAKE_ARRAY_MULTI, size);
                  FW_MATCH_CASE_1(InternalInstruction::makeObject, index)
                     ADD_INSTRUCTION(core::MAKE_OBJECT_MULTI, index);
                  FW_MATCH_CASE_1(InternalInstruction::makeFunction, parameterNames)
                     int subFunctionId = mInstructionInfo[members[j]].subFunctionId;
                     ADD_INSTRUCTION(core::MAKE_FUNCTION_MULTI, subFunctionId);
                  FW_MATCH_END
                  ++offset;

                  for(int l = 0; l < inputs.getSize(); l++)
                  {
                     if(inputs[l] == ~members[j])
                        ADD_INSTRUCTION(core::STORE_MULTI, l);
                  }

                  if(members[j] != instruction.index
                  && mInstructionInfo[members[j]].outputs.getSize() == 0)
                  {
                     mCode.getBack().opcode = core::STORE_MULTI_POP;
                     --offset;
                  }
                  else if(members[j] != instruction.index
                       || j != members.getSize() - 1
                       || (instruction.type & OrderedInstruction::STORE))
                  {
                     //find a place in the frame for this value
                     for(; k < mCurrentBranches[branch].frameContents.getSize(); k++)
                     {
                        if(mCurrentBranches[branch].frameContents[k] == -1)
                           break;
                     }

                     if(members[j] == instruction.index)
                        location = k;

                     if(k == mCurrentBranches[branch].frameContents.getSize())
                        mCurrentBranches[branch].frameContents.pushBack(members[j]);
                     else
                        mCurrentBranches[branch].frameContents[k] = members[j];

                     if(members[j] == instruction.index && j == members.getSize() - 1)
                     {
                        ADD_INSTRUCTION(core::STORE,
                           offset - mCurrentBranches[branch].frameOffset - k - 1);
                     }
                     else
                     {
                        ADD_INSTRUCTION(core::STORE_POP,
                           offset - mCurrentBranches[branch].frameOffset - k - 1);
                        --offset;
                     }
                     ++k;

                     mInstructionInfo[members[j]].usageCount = 0;
                  }
               }

               //remove the multi-allocate array
               if(members[members.getSize() - 1] == instruction.index)
               {
                  if(instruction.type & OrderedInstruction::POP)
                  {
                     ADD_INSTRUCTION(core::POP, 2)
                     offset -= 2;
                  }
                  else
                  {
                     ADD_INSTRUCTION(core::STORE_POP, 1)
                     --offset;
                  }
               }
               else
               {
                  ADD_INSTRUCTION(core::POP, 1);

                  if(!(instruction.type & OrderedInstruction::POP))
                  {
                     ADD_INSTRUCTION(core::LOAD,
                        offset - mCurrentBranches[branch].frameOffset - location - 1);

                     if(!(instruction.type & OrderedInstruction::STORE))
                        mCurrentBranches[branch].frameContents[location] = -1;
                  }
                  else
                     --offset;
               }

               if(!(instruction.type & OrderedInstruction::POP))
                  mInstructionInfo[instruction.index].usageCount = 1;
            }
            else
            {
               FW_MATCH_START(mInstructionInfo[instruction.index].instruction)
               // immediates -----------------------------------------------------
               FW_MATCH_CASE_0(InternalInstruction::loadTrue)
                  ADD_INSTRUCTION(core::IMMEDIATE_BOOLEAN, 1);
                  ++offset;
               FW_MATCH_CASE_0(InternalInstruction::loadFalse)
                  ADD_INSTRUCTION(core::IMMEDIATE_BOOLEAN, 0);
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadInteger, i)
                  if(i >= 0 && i < (1 << core::PARAMETER_BITS))
                     ADD_INSTRUCTION(core::IMMEDIATE_INTEGER, i)
                  else if(-i >= 0 && -i < (1 << core::PARAMETER_BITS))
                     ADD_INSTRUCTION(core::IMMEDIATE_NEGATIVE, -i)
                  else
                  {
                     ADD_INSTRUCTION(core::IMMEDIATE_INTEGER32, 0);
                     ADD_INSTRUCTION(0, 0);
                     *((int32*)&mCode[mCode.getSize() - 1]) = i;
                  }
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadInteger64, i)
                  ADD_INSTRUCTION(core::IMMEDIATE_INTEGER64, 0);
                  ADD_INSTRUCTION(0,0);
                  ADD_INSTRUCTION(0,0);
                  *((int64*)&mCode[mCode.getSize() - 2]) = i;
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadReal, x)
                  float f = (float)x;
                  if((double)f == x)
                  {
                     ADD_INSTRUCTION(core::IMMEDIATE_REAL32, 0);
                     ADD_INSTRUCTION(0,0);
                     *((float*)&mCode[mCode.getSize() - 1]) = f;
                  }
                  else
                  {
                     ADD_INSTRUCTION(core::IMMEDIATE_REAL64, 0);
                     ADD_INSTRUCTION(0,0);
                     ADD_INSTRUCTION(0,0);
                     *((double*)&mCode[mCode.getSize() - 2]) = x;
                  }
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadData1, d)
                  ADD_INSTRUCTION(core::IMMEDIATE_DATA1, d);
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadData2, d)
                  ADD_INSTRUCTION(core::IMMEDIATE_DATA2, d);
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadData4, d)
                  ADD_INSTRUCTION(core::IMMEDIATE_DATA4, d);
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadData8, d)
                  ADD_INSTRUCTION(core::IMMEDIATE_DATA8, d);
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadData16, d)
                  ADD_INSTRUCTION(core::IMMEDIATE_DATA16, d);
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadData32, d)
                  ADD_INSTRUCTION(core::IMMEDIATE_DATA32, d);
                  ADD_INSTRUCTION(0,0);
                  *((uint32*)&mCode[mCode.getSize() - 1]) = d;
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadData64, d)
                  ADD_INSTRUCTION(core::IMMEDIATE_DATA64, d);
                  ADD_INSTRUCTION(0,0);
                  ADD_INSTRUCTION(0,0);
                  *((uint64*)&mCode[mCode.getSize() - 2]) = d;
                  ++offset;
               FW_MATCH_CASE_2(InternalInstruction::loadData, size, index)
                  ADD_INSTRUCTION(core::LOAD_CONSTANT,
                     mInstructionInfo[instruction.index].tableIndex);
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadCharacter, c)
                  if(c >= 0 && c < (1 << core::PARAMETER_BITS))
                     ADD_INSTRUCTION(core::IMMEDIATE_CHARACTER, c)
                  else
                  {
                     ADD_INSTRUCTION(core::IMMEDIATE_CHARACTER32, 0);
                     ADD_INSTRUCTION(0,0);
                     *((uint32*)&mCode[mCode.getSize() - 1]) = c;
                  }
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadAtom, index)
                  index = mAtomMap[index];
                  if(index >= 0 && index < (1 << core::PARAMETER_BITS))
                     ADD_INSTRUCTION(core::IMMEDIATE_ATOM, index)
                  else
                  {
                     ADD_INSTRUCTION(core::IMMEDIATE_ATOM32, 0);
                     ADD_INSTRUCTION(0,0);
                     *((uint32*)&mCode[mCode.getSize() - 1]) = index;
                  }
                  ++offset;
               FW_MATCH_CASE_3(InternalInstruction::loadArray, type, size, index)
                  ADD_INSTRUCTION(core::LOAD_CONSTANT,
                     mInstructionInfo[instruction.index].tableIndex);
                  ++offset;

               // environmentals -------------------------------------------------
               FW_MATCH_CASE_1(InternalInstruction::loadImport, index)
//                  ADD_INSTRUCTION(core::LOAD,
//                     offset + mBytecode->importTable.getSize() - index);
//                  ++offset;

                  ADD_INSTRUCTION(core::LOAD, offset + 1);
                  ADD_INSTRUCTION(core::IMMEDIATE_INTEGER, index);
                  ADD_INSTRUCTION(core::GET_ELEMENT, 0);
                  ++offset;
               FW_MATCH_CASE_1(InternalInstruction::loadParameter, index)
                  ADD_INSTRUCTION(core::LOAD,
                     offset + mFunctionInfo[mCurrentFunction].arity - index);
                  ++offset;

               // computations ---------------------------------------------------
               FW_MATCH_CASE_1(InternalInstruction::makeArray, size)
                  ADD_INSTRUCTION(core::MAKE_ARRAY, size);
                  offset -= size - 1;

               FW_MATCH_CASE_1(InternalInstruction::makeObject, index)
                  ADD_INSTRUCTION(core::MAKE_OBJECT, index);
                  offset -= mBytecode->classes[index].memberTable.getSize() - 1;

               FW_MATCH_CASE_1(InternalInstruction::makeFunction, parameterNames)
                  int subFunctionId = mInstructionInfo[instruction.index].subFunctionId;
                  ADD_INSTRUCTION(core::MAKE_FUNCTION, subFunctionId);
                  offset -= mFunctionInfo[subFunctionId].environmentTable.getSize() - 1;

               FW_MATCH_CASE_1(InternalInstruction::loadMember, index)
                  ADD_INSTRUCTION(core::GET_MEMBER, mMemberNameMap[index]);

               FW_MATCH_CASE_0(InternalInstruction::apply)
                  int parameterCount =
                     mInstructionInfo[instruction.index].inputs.getSize() - 1;

                  offset -= parameterCount;

                  //check for tail recursion
                  if(canTailRecurse
                  && i == mCurrentBranches[branch].orderedInstructions.getSize() - 1)
                  {
                     ADD_INSTRUCTION(core::APPLY_RETURN, parameterCount);
                     if(offset != mCurrentBranches[branch].frameOffset + mCurrentBranches[branch].frameSize + 1)
                        FAIL("Final offset does not match starting offset.");
                     return;
                  }
                  else
                     ADD_INSTRUCTION(core::APPLY, parameterCount);

               FW_MATCH_CASE_0(InternalInstruction::testCondition)
                  if(canTailRecurse
                  && i == mCurrentBranches[branch].orderedInstructions.getSize() - 1)
                  {
                     if(offset != mCurrentBranches[branch].frameOffset + mCurrentBranches[branch].frameSize + 1)
                        FAIL("Final offset does not match starting offset.");

                     int branchIndex = mCode.getSize();
                     ADD_INSTRUCTION(core::TEST_CONDITION, 0);
                     ADD_INSTRUCTION(core::RETURN, 0);
                     --offset;
                     linearize(mInstructionInfo[instruction.index].subFunctionId, offset, true);
                     mCode[branchIndex].parameter = mCode.getSize() - branchIndex - 1;

                     linearize(mInstructionInfo[instruction.index].subFunctionId + 1, offset, true);
                     return;
                  }
                  else
                  {
                     int branchIndex = mCode.getSize();
                     ADD_INSTRUCTION(core::TEST_CONDITION, 0);
                     int errorGotoIndex = mCode.getSize();
                     ADD_INSTRUCTION(core::GOTO, 0);
                     --offset;
                     linearize(mInstructionInfo[instruction.index].subFunctionId, offset, false);
                     mCode[branchIndex].parameter = mCode.getSize() - branchIndex;

                     int gotoIndex = mCode.getSize();
                     ADD_INSTRUCTION(core::GOTO, 0);
                     linearize(mInstructionInfo[instruction.index].subFunctionId + 1, offset, false);
                     mCode[gotoIndex].parameter = mCode.getSize() - gotoIndex - 1;
                     mCode[errorGotoIndex].parameter = mCode.getSize() - errorGotoIndex - 1;
                     ++offset;
                  }

               // optimized instructions -----------------------------------
               FW_MATCH_CASE_0(InternalInstruction::add)
                  ADD_INSTRUCTION(core::ADD, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::subtract)
                  ADD_INSTRUCTION(core::SUBTRACT, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::multiply)
                  ADD_INSTRUCTION(core::MULTIPLY, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::divide)
                  ADD_INSTRUCTION(core::DIVIDE, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::modulus)
                  ADD_INSTRUCTION(core::MODULUS, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::power)
                  ADD_INSTRUCTION(core::POWER, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::negate)
                  ADD_INSTRUCTION(core::NEGATE, 0);
               FW_MATCH_CASE_0(InternalInstruction::and_)
                  ADD_INSTRUCTION(core::AND, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::or_)
                  ADD_INSTRUCTION(core::OR, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::not_)
                  ADD_INSTRUCTION(core::NOT, 0);
               FW_MATCH_CASE_0(InternalInstruction::equals)
                  ADD_INSTRUCTION(core::EQUALS, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::isLessThan)
                  ADD_INSTRUCTION(core::IS_LESS_THAN, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::isGreaterThan)
                  ADD_INSTRUCTION(core::IS_GREATER_THAN, 0); --offset;
               FW_MATCH_CASE_0(InternalInstruction::size)
                  ADD_INSTRUCTION(core::SIZE, 0);
               FW_MATCH_CASE_0(InternalInstruction::getElement)
                  ADD_INSTRUCTION(core::GET_ELEMENT, 0); --offset;
               FW_MATCH_END

               if(instruction.type & OrderedInstruction::STORE)
               {
                  //find a place in the frame for this value
                  int j;
                  for(j = 0; j < mCurrentBranches[branch].frameContents.getSize(); j++)
                  {
                     if(mCurrentBranches[branch].frameContents[j] == -1)
                        break;
                  }

                  if(j == mCurrentBranches[branch].frameContents.getSize())
                     mCurrentBranches[branch].frameContents.pushBack(instruction.index);
                  else
                     mCurrentBranches[branch].frameContents[j] = instruction.index;

                  j = offset - mCurrentBranches[branch].frameOffset - j - 1;

                  if(instruction.type & OrderedInstruction::POP)
                  {
                     ADD_INSTRUCTION(core::STORE_POP, j);
                     mInstructionInfo[instruction.index].usageCount = 0;
                     --offset;
                  }
                  else
                  {
                     ADD_INSTRUCTION(core::STORE, j);
                     mInstructionInfo[instruction.index].usageCount = 1;
                  }
               }
               else if(instruction.type & OrderedInstruction::POP)
               {
                  ADD_INSTRUCTION(core::POP, 1);
                  --offset;
               }
            }
         }

         if(offset < mCurrentBranches[branch].frameOffset)
            FAIL("Offset dropped below frame.");
      }

      if(offset != mCurrentBranches[branch].frameOffset + mCurrentBranches[branch].frameSize + 1)
         FAIL("Final offset does not match starting offset.");

      core::SourceLocation sourceLocation(0,0);

      if(canTailRecurse)
         ADD_INSTRUCTION(core::RETURN, 0)
      else if(mCurrentBranches[branch].frameSize > 0)
      {
         ADD_INSTRUCTION(core::STORE_POP,
            offset - mCurrentBranches[branch].frameOffset - 1);
         --offset;
         if(mCurrentBranches[branch].frameSize > 1)
         {
            ADD_INSTRUCTION(core::POP, mCurrentBranches[branch].frameSize - 1);
            offset -= mCurrentBranches[branch].frameSize - 1;
         }
      }

      for(int i = 0; i < mCurrentBranches[branch].frameContents.getSize(); i++)
      {
         if(mCurrentBranches[branch].frameContents[i] != -1)
            FAIL("Stuff left over in frame (in linearize()).");
      }
      if(mCurrentBranches[branch].frameContents.getSize() != mCurrentBranches[branch].frameSize)
         FAIL("Frame size different after linearize().");
   }

   // actual interface ===============================================
   core::Module* translateBytecode(compiler::BytecodePtr bytecode)
   {
      core::Module* result;
#ifdef WRITE_DEBUG_STUFF
      gDebugWriter = spnew DebugWriter;
      gDebugWriter->write($("\n\n======================================================================"));
      try
      {
#endif

      Recompiler recompiler(bytecode);
      result = recompiler.recompile();

#ifdef WRITE_DEBUG_STUFF
      }
      catch(...)
      {
         gDebugWriter = null;
         throw;
      }
#endif

      return result;
   }

   // exceptions -----------------------------------------------------
   String InvalidBytecode::getDescription()
   {
      return $("Bytecode invalid:  ") + String(mLine+1) + $(":") + String(mColumn+1) + $(":  ") + mError;
   }
}

// ===================================================================

} //namespace vm
} //namespace evlan
