/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2003/04/11 (KV)
 *  + Imported code from Evlan.
 */

// ===================================================================

#ifndef framework_common_Array_h__INCLUDED
#define framework_common_Array_h__INCLUDED

#include "SmartPointer.h"

namespace framework {
namespace common {

//Pay no attention to the global functions behind the curtain.
void ThrowOutOfBoundsException(int size, int index);

// class definition ==================================================
//This class adds reference counting to arrays.  It's usage is fairly
// simple:
//   Array<int> myArray;        //declare array of ints
//   myArray = Array<int>(16);  //allocate array of 16 ints
//   myArray[5] = 2;            //set element 5 to value 2
//   myArray = null;            //"null" is an array with zero elements.
//Note that a null array is defined as an array of size zero.  There is
// no difference between the two.
template <typename Element>
class Array
{
public:
   typedef SmartPointer<Element> ElementPtr;

   inline Array(): mSize(0) {}
   inline Array(NullType): mSize(0) {}
   explicit Array(int size)
      : mSize(size)
   {
      if(size > 0)
         mArray = SPInitializer::makeSmartArray<Element>(mSize);
   }

   inline Array(SmartPointer<Element> array, int size)
      : mArray(array), mSize(size) {if(size == 0) mArray = null;}
   inline Array(const Array& other)
      : mArray(other.mArray), mSize(other.mSize) {}

   inline Array<const Element> toConst() const
   {
      return Array<const Element>(mArray, mSize);
   }

   static inline Array makeStatic(Element* data, int size)
      {return Array(SmartPointer<Element>::makeStatic(data), size);}

   inline ~Array() {}

   inline int getSize() const
      {return mSize;}
   inline Element& operator[](int index) const
   {
#ifdef DEBUG
      if(index < 0 || index >= mSize)
         ThrowOutOfBoundsException(mSize, index);
#endif
      return (&(*mArray))[index];
   }
   inline Array getInterval(int start) const
   {
#ifdef DEBUG
      if(start < 0 || start > mSize)
         ThrowOutOfBoundsException(mSize, start);
#endif
      if(mArray == null)
         return *this;
      else
         return Array(mArray + start, mSize - start);
   }
   inline Array getInterval(int start, int end) const
   {
#ifdef DEBUG
      if(start < 0 || start > mSize)
         ThrowOutOfBoundsException(mSize, start);
      if(end   < 0 || end   > mSize)
         ThrowOutOfBoundsException(mSize, end);
      if(end   < start)
         ThrowOutOfBoundsException(mSize, end - start);
#endif
      if(mArray == null)
         return *this;
      else
         return Array(mArray + start, end - start);
   }
   inline SmartPointer<Element> getSmartPointer() const
      {return mArray;}

   inline Array& operator=(NullType)
      {mArray = null; mSize = 0; return *this;}

   inline bool operator==(const Array& other) const {return mArray == other.mArray && mSize == other.mSize;}
   inline bool operator!=(const Array& other) const {return mArray != other.mArray || mSize != other.mSize;}
   inline bool operator<=(const Array& other) const {return mArray <  other.mArray || (mArray == other.mArray && mSize <= other.mSize);}
   inline bool operator>=(const Array& other) const {return mArray >  other.mArray || (mArray == other.mArray && mSize >= other.mSize);}
   inline bool operator< (const Array& other) const {return mArray <  other.mArray || (mArray == other.mArray && mSize <  other.mSize);}
   inline bool operator> (const Array& other) const {return mArray >  other.mArray || (mArray == other.mArray && mSize >  other.mSize);}
   inline bool operator==(NullType) const {return mSize == 0;}
   inline bool operator!=(NullType) const {return mSize != 0;}

   template <typename OtherArrayType>
   inline OtherArrayType reinterpretCast() const
   {
      return OtherArrayType(
         mArray.reinterpretCast<typename OtherArrayType::ElementPtr >(),
         mSize);
   }

   //for use with STL
   struct SequenceOrdering
   {
      bool operator()(Array<Element> a, Array<Element> b)
      {
         int size = (a.getSize()<b.getSize())?a.getSize():b.getSize();
         for(int i = 0; i < size; i++)
         {
            if(a[i] < b[i])
               return true;
            else if(a[i] > b[i])
               return false;
         }
         return a.getSize() < b.getSize();
      }
   };

private:
   ElementPtr mArray;
   int mSize;
};

#define $ARRAY(TYPE, DATA) Array<TYPE>::makeStatic(DATA, sizeof(DATA) / sizeof(TYPE))

} //namespace common
} //namespace framework

#endif //included
