/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2003/05/20 (KV)
 *  + Added ErrnoError
 *  + Added Win32Error
 * 2003/04/11 (KV)
 *  + Imported code from Evlan.
 */

// ===================================================================

#ifndef framework_common_Exception_h__INCLUDED
#define framework_common_Exception_h__INCLUDED

#include "String.h"

namespace framework {
namespace common {

// class definition ==================================================
class Exception;
typedef SmartPointer<Exception> ExceptionPtr;

/* The base class for all throwable exceptions.  Note that you should
 * not throw this using the usual C++ "throw" keyword.  Check the
 * macros later in this file for more info. */
class Exception
{
public:
   /* If an exception is thrown as a response to some other exception, then
    * "cause" should be set to that other exception.  This forms sort of
    * an error chain.  All subclasses of Exception should have "cause" as
    * the last parameter to the constructor, with a default value of null. */
   Exception(ExceptionPtr cause = null);
   virtual ~Exception();

   /* Returns a human-readable description of the error.  If an exception
    * is not caught, this will be printed before the program exits. */
   virtual String getDescription();

   //Gets the exception which caused this one, or null if none.
   inline ExceptionPtr getCause() {return mCause;}

   /* If the framework was compiled with debugging on, getStackTrace()
    * will produce file names and line numbers of each stack frame.
    * This will be done automatically if an exception is not caught.
    * Note that you must have "addr2line" in your path for this to
    * work. */
   String getStackTrace();

   //This is sometimes useful.
   static String getStackTrace(void** stack, int size);

   /* This is just a hook, called from the internal main function.  Don't
    * mind it. */
   static inline void setExecutablePath(String path)
      {mExecutablePath = path;}

private:
   static const int STACK_TRACE_SIZE = 16;
   ExceptionPtr mCause;
   void* mStackTrace[STACK_TRACE_SIZE];

   static String mExecutablePath;

   //disable copying
   Exception(const Exception&);
   Exception& operator=(const Exception&);
};

/* Use ErrnoError to represent errors in C or POSIX library routines which
 * use errno to report error conditions.  This is useful use as the cause
 * for some other exception.  For example, an InputStream which encounters
 * a strange error when calling the OS's read() function could throw a
 * ReadError with this as the cause. */
class ErrnoError: extends Exception
{
public:
   /* functionName is the name of the C or POSIX library function which failed.
    * errorNumber should be the value of errno after that function returned. */
   inline ErrnoError(String functionName,
                     int errorNumber,
                     ExceptionPtr cause = null)
      : Exception(cause),
        mFunctionName(functionName),
        mErrorNumber(errorNumber)
   {}
   //This constructor consults errno automatically.
   ErrnoError(String functionName, ExceptionPtr cause = null);

   virtual ~ErrnoError();

   //The description will be generated using strerror().
   virtual String getDescription();

   inline int getErrorNumber() {return mErrorNumber;}

private:
   String mFunctionName;
   int mErrorNumber;

   ErrnoError(const ErrnoError&);
   ErrnoError& operator=(const ErrnoError&);
};
typedef SmartPointer<ErrnoError> ErrnoErrorPtr;

#ifdef PLATFORM_WIN32
/* Win32Error is just like ErrnoError, except used for Win32 errors (as
 * returned by Win32's GetLastError()). */
class Win32Error: extends Exception
{
public:
   /* functionName is the name of the C or POSIX library function which failed.
    * errorNumber should be the value returned by GetLastError() after that
    * function finished. */
   inline Win32Error(String functionName,
                     unsigned long errorNumber,  //DWORD == unsigned long
                     ExceptionPtr cause = null)
      : Exception(cause),
        mFunctionName(functionName),
        mErrorNumber(errorNumber)
   {}
   //This constructor consults GetLastError() automatically.
   Win32Error(String functionName, ExceptionPtr cause = null);

   virtual ~Win32Error();

   /* The description will be generated using the horrible convoluted pile of
    * crap known as FormatMessage().  Be happy that you don't have to do this
    * yourself. */
   virtual String getDescription();

   inline int getErrorNumber() {return mErrorNumber;}

private:
   String mFunctionName;
   unsigned long mErrorNumber;

   Win32Error(const Win32Error&);
   Win32Error& operator=(const Win32Error&);
};
typedef SmartPointer<Win32Error> Win32ErrorPtr;
#endif

/* FatalError is just a quick exception that you could use for debugging
 * or when an error occurs that couldn't possibly be handled anyway.
 * This saves you the time of defining a custom exception class.  However,
 * I strongly recommend against using this if there is any chance whatsoever
 * that the program could recover from the exception. */
class FatalError: extends Exception
{
public:
   FatalError(String description, ExceptionPtr cause = null);
   virtual ~FatalError();
   virtual String getDescription() {return mDescription;}

private:
   String mDescription;

   FatalError(const FatalError&);
   FatalError& operator=(const FatalError&);
};
typedef SmartPointer<FatalError> FatalErrorPtr;

/* There is really no good way to throw a smart pointer to an exception.
 * Since smart pointers are not built in to the language, C++ doesn't know
 * how smart pointer types relate through inheritance.  So, if you throw
 * a smart pointer to a type derived from Exception, it will not be caught
 * by a catch block set up to catch ExceptionPtr.  Instead, we just have to
 * throw plain ExceptionPtr's everywhere and determine the types manually.
 * Slow, maybe, but I doubt that it's significantly slower, and exceptions
 * are only thrown when there are critical errors anyway.  (Hell, as far
 * as I know, this is essentially how catch blocks work internally anyway.) */

/* Use FW_THROW() to throw exceptions just to make sure they are thrown
 * as ExceptionPtr's and not anything else. */
#define FW_THROW(EXCEPTION)                        \
   throw ExceptionPtr(EXCEPTION)

/* Rather than use catch(), you have to use the FW_CATCH macros.  This
 * allows the usual type checking on exceptions even though we are
 * actually throwing plain old ExceptionPtr's.  Here is an example of
 * the usage:
 *   try
 *   {
 *      doSomething();
 *   }
 *   FW_CATCH_START
 *      FW_CATCH(MyException, e)
 *      {
 *         cerr << "Caught a MyException!" << endl;
 *      }
 *      FW_CATCH(Exception, e)
 *      {
 *         cerr << "Caught some other exception:" << endl
 *         cerr << e->getDescription() << endl;
 *      }
 *   FW_CATCH_END
 *   catch(int i)
 *   {
 *      cerr << "I pitty da foo' who throws integers!" << endl;
 *   }
 *   catch(...)
 *   {
 *      cerr << "I have no idea what I just caught, but it's warm and" << endl
 *      cerr << "squishy.  Ewww." << endl;
 *   }
 * Note how you can catch foreign exception types at the same time.  It's
 * ugly, I know, but there was really no better way to do it short of creating
 * my own language.  Oh, wait... */

#define FW_CATCH_START                                   \
   catch(::framework::common::ExceptionPtr _e) \
   {                                                     \
      if(false)                                          \
         {}

#define FW_CATCH_END                                     \
      else                                               \
         throw;                                          \
   }

#define FW_CATCH(TYPE, VARNAME)                          \
      else if(SmartPointer<TYPE> VARNAME = _e.dynamicCast<SmartPointer<TYPE> >())


/* FW_DECLARE_EXCEPTION is a shortcut macro for creating exceptions.
 * The two parameters are the name of the new exception class and its
 * parent class.  This declares a simple exception class and its
 * corresponding smart pointer type.  Use FW_DECLARE_EXCEPTION in a
 * header file and FW_DEFINE_EXCEPTION in the corresponding source
 * file to create a complete exception class. */
#define FW_DECLARE_EXCEPTION(NAME, BASETYPE)                          \
   class NAME: public BASETYPE                                        \
   {                                                                  \
   public:                                                            \
      inline NAME(ExceptionPtr cause = null)                          \
         : BASETYPE(cause) {}                                         \
      virtual ::framework::common::String getDescription(); \
      virtual ~NAME() {}                                              \
   private:                                                           \
      NAME(const NAME&);                                              \
      NAME& operator=(const NAME&);                                   \
   };                                                                 \
   typedef SmartPointer<NAME> NAME##Ptr

/* FW_DEFINE_EXCEPTION implements an exception class that has been
 * declared using FW_DECLARE_EXCEPTION.  The first parameter is the
 * class name and the second is a human-readable string describing the
 * error.  This string will be returned by the exception's
 * getDescription() method. */
#define FW_DEFINE_EXCEPTION(NAME, DESCRIPTION)                        \
   ::framework::common::String NAME::getDescription()       \
      {return DESCRIPTION;}

} //namespace common
} //namespace framework

#endif //included
