/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2003/04/11 (KV)
 *  + Imported code from Evlan.
 */

// ===================================================================

#ifndef framework_common_SmartPointer_h__INCLUDED
#define framework_common_SmartPointer_h__INCLUDED

#include "MemoryManager.h"

namespace framework {
namespace common {

// class definition ==================================================
#ifdef DEBUG
void ThrowNullPointerException();
#endif

template <typename Type>
class SmartPointer
{
public:  //private!  Pretend it's private!
   //Unfortunately, different instances of the same template class are not
   // able to access each other's private data, and, as far as I know, there
   // is no way to have them declare each other as friends.  This means there
   // is absolutely NO WAY for me to write this class without making these
   // public.  ARG!  But soon I will have my own language, and I won't have to
   // deal with this crap...  soon, soon...
   //In summary, DO NOT ACCESS THESE DIRECTLY OR YOU WILL BE PUNISHED!
   MemoryManager::ObjectInfo* mInfo;
   Type* mObject;

private:
   inline void increment()
   {
      if(mInfo)
         ++(mInfo->mRefCount);
   }
   inline void decrement()
   {
      if(mInfo)
         if(--(mInfo->mRefCount) == 0)
            mInfo->mDestructor(mInfo + 1);
   }

public:
   typedef Type* DumbPointer;

   inline SmartPointer()
      : mInfo(null), mObject(null) {}
   inline SmartPointer(NullType)
      : mInfo(null), mObject(null) {}
   inline SmartPointer(const SmartPointer& other)
      : mInfo(other.mInfo),
        mObject(other.mObject)
   {
      increment();
   }
   template <typename Other>
   inline SmartPointer(const SmartPointer<Other>& other)
      : mInfo(other.mInfo),
        mObject(other.mObject)
   {
      increment();
   }
   //this last constructor allows you to create a smart pointer to a member
   // of an object.  The pointer will act like it points to the [object]
   // parameter, but will use the refcount of the object pointed to by [other]
   // and will delete [other] when the refcount reaches zero.  [object] should
   // be a member of [other], such that when [other] is deleted, [object] will
   // go with it.
   template <typename Other>
   inline SmartPointer(const SmartPointer<Other>& other, Type* object)
      : mInfo(other.mInfo),
        mObject(object)
   {
      increment();
   }

   //this handy function allows you to create smart pointers which do not keep
   // track of references and never attempt to delete the object pointed at.
   // This is mainly intended for pointing at static data which will never be
   // deallocated as long as the program is running.  String::literal() takes
   // advantage of this.
   inline static SmartPointer makeStatic(Type* object)
   {
      SmartPointer result;
      result.mObject = object;
      return result;
   }

   inline ~SmartPointer()
   {
      decrement();
   }


   inline SmartPointer& operator=(const SmartPointer& other)
   {
      mObject = other.mObject;
      if(mInfo != other.mInfo)
      {
         decrement();
         mInfo = other.mInfo;
         increment();
      }
      return *this;
   }
   template <typename Other>
   inline SmartPointer& operator=(const SmartPointer<Other>& other)
   {
      mObject = other.mObject;
      if(mInfo != other.mInfo)
      {
         decrement();
         mInfo = other.mInfo;
         increment();
      }
      return *this;
   }
   inline SmartPointer& operator=(NullType)
   {
      decrement();
      mObject = null;
      mInfo = null;
      return *this;
   }


   inline Type* operator->() const
   {
#ifdef DEBUG
      if(mObject == null)
         ThrowNullPointerException();
#endif
      return mObject;
   }
   inline Type& operator*() const
   {
#ifdef DEBUG
      if(mObject == null)
         ThrowNullPointerException();
#endif
      return *mObject;
   }

   inline bool operator==(const SmartPointer& other) const {return mObject == other.mObject;}
   inline bool operator!=(const SmartPointer& other) const {return mObject != other.mObject;}
   inline bool operator<=(const SmartPointer& other) const {return mObject <= other.mObject;}
   inline bool operator>=(const SmartPointer& other) const {return mObject >= other.mObject;}
   inline bool operator< (const SmartPointer& other) const {return mObject <  other.mObject;}
   inline bool operator> (const SmartPointer& other) const {return mObject >  other.mObject;}
   inline bool operator==(NullType) const {return mObject == null;}
   inline bool operator!=(NullType) const {return mObject != null;}

   inline operator bool()  const {return mObject != null;}
   inline bool operator!() const {return mObject == null;}

   template <typename OtherPtrType>
   inline OtherPtrType dynamicCast() const
   {
      return OtherPtrType(*this,
         dynamic_cast<typename OtherPtrType::DumbPointer >(mObject));
   }
   //this is used to work around a bug where GCC doesn't understand the
   // explicit template parameter in certain places...
   template <typename OtherType>
   inline SmartPointer<OtherType > dynamicCastGCCBugWorkaround(OtherType* blah)
   {
      return dynamicCast<SmartPointer<OtherType > >();
   }

   template <typename OtherPtrType>
   inline OtherPtrType reinterpretCast() const
   {
      return OtherPtrType(*this,
         reinterpret_cast<typename OtherPtrType::DumbPointer >(mObject));
   }

   inline SmartPointer operator+(int offset) const
   {
#ifdef DEBUG
      if(mObject == null)
         ThrowNullPointerException();
#endif
      SmartPointer result(*this);
      result.mObject += offset;
      return result;
   }
   inline SmartPointer& operator++()
   {
#ifdef DEBUG
      if(mObject == null)
         ThrowNullPointerException();
#endif
      ++mObject;
      return *this;
   }
   inline SmartPointer& operator+=(int offset)
   {
#ifdef DEBUG
      if(mObject == null)
         ThrowNullPointerException();
#endif
      mObject += offset;
      return *this;
   }

   inline SmartPointer operator-(int offset) const
   {
#ifdef DEBUG
      if(mObject == null)
         ThrowNullPointerException();
#endif
      SmartPointer result(*this);
      result.mObject -= offset;
      return result;
   }
   inline SmartPointer& operator--()
   {
#ifdef DEBUG
      if(mObject == null)
         ThrowNullPointerException();
#endif
      --mObject;
      return *this;
   }
   inline SmartPointer& operator-=(int offset)
   {
#ifdef DEBUG
      if(mObject == null)
         ThrowNullPointerException();
#endif
      mObject -= offset;
      return *this;
   }


   inline SmartPointer release()
   {
      SmartPointer result;
      result.mInfo = mInfo;
      result.mObject = mObject;
      mInfo = null;
      mObject = null;
      return result;
   }

private:
   inline SmartPointer(MemoryManager::ObjectInfo* info, Type* object)
      : mInfo(info), mObject(object)
   {}
   friend struct SPInitializer;
};

//chop, chop, chop
struct SPInitializer
{
   template <typename Type>
   static void objectDestructor(void* object)
   {
      delete reinterpret_cast<Type*>(object);
   };

   template <typename Type>
   static void arrayDestructor(void* object)
   {
      int* block = (int*)object;
      int size = *block;
      Type* array = reinterpret_cast<Type*>(block + 1);
      for(int i = 0; i < size; i++)
         array[i].~Type();
      MemoryManager::free(block);
   };

   template <typename Type>
   SmartPointer<Type> operator=(Type* object) const
   {
      MemoryManager::ObjectInfo* info =
         reinterpret_cast<MemoryManager::ObjectInfo*>(object) - 1;
      info->mDestructor = objectDestructor<Type>;
      return SmartPointer<Type>(info, object);
   }

   template <typename Type>
   static SmartPointer<Type> makeSmartArray(int size)
   {
      int byteCount = sizeof(int) + sizeof(Type) * size;
      int* block = (int*)MemoryManager::allocate(byteCount);
      *block = size;
      Type* array = reinterpret_cast<Type*>(block + 1);
      for(int i = 0; i < size; i++)
         new((void*)(array + i)) Type;
      MemoryManager::ObjectInfo* info =
         reinterpret_cast<MemoryManager::ObjectInfo*>(block) - 1;
      info->mDestructor = arrayDestructor<Type>;
      return SmartPointer<Type>(info, array);
   }

   template <typename Type>
   static SmartPointer<Type> getThis(Type* object)
   {
      MemoryManager::ObjectInfo* info =
         reinterpret_cast<MemoryManager::ObjectInfo*>(object) - 1;
      ++info->mRefCount;
      return SmartPointer<Type>(info, object);
   }
};
static const SPInitializer SPInitializerHack = SPInitializer();

//mmmm.  olives.
#define spnew ::framework::common::SPInitializerHack = new
//spthis can only be used within objects which were allocated using spnew.
// Using it anywhere else may lead to hard-to-track bugs.
#define spthis (::framework::common::SPInitializer::getThis(this))

//just don't ask, ok?

} //namespace common
} //namespace framework

#endif //included
