/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2003/04/11 (KV)
 *  + Imported code from Evlan.
 */

// ===================================================================

#include "String.h"
#include <ctype.h>
#include <stdlib.h>
#include <math.h>

namespace framework {
namespace common {

// String member functions ==========================================
// constructors ------------------------------------------------------
String::String(const char other[])
{
   int length = strlen(other);
   if(length == 0)
      return;

   Array<char> buf(length);
   memcpy(&buf[0], other, length);
   mBuffer = buf.toConst();
}

String::String(const char other[], int length)
{
   if(length <= 0)
      return;

   Array<char> buf(length);
   memcpy(&buf[0], other, length);
   mBuffer = buf.toConst();
}

String::String(char value)
{
   Array<char> buf(1);
   buf[0] = value;
   mBuffer = buf.toConst();
}

String::String(char value, int count)
{
   Array<char> buf(count);
   if(count > 0)
      memset(&buf[0], value, count);
   mBuffer = buf.toConst();
}

namespace
{
   inline char getDigit(double& x, int base, int flags)
   {
      int digit = (int)x;
      x = (x - digit) * base;
      if(digit < 10)
         return '0' + digit;
      else
         return ((flags & String::CAPITAL_DIGITS)?'A'-10:'a'-10) + digit;
   }

   template <typename Type>
   int writeInt(char* buffer, Type n, int width, int base, int flags)
   {
      if(n == 0 && width <= 1)
      {
         buffer[0] = '0';
         return 1;
      }
      else
      {
         bool isNegative = n < 0;
         if(isNegative)
            n = -n;

         char chars[128];
         int pos = 0;
         while(n > 0)
         {
            int digit = n % base;
            n /= base;
            if(digit < 10)
               chars[pos++] = '0' + digit;
            else
               chars[pos++] = ((flags & String::CAPITAL_DIGITS)?'A'-10:'a'-10) + digit;
         }

         while(pos < width)
            chars[pos++] = '0';

         if(isNegative)
            chars[pos++] = '-';
         else if(flags & String::ALWAYS_SHOW_SIGN)
            chars[pos++] = '+';
         else if(flags & String::BLANK_POSITIVE_SIGN)
            chars[pos++] = ' ';

         for(int i = 0; i <= pos; i++)
            buffer[i] = chars[pos - 1 - i];
         return pos;
      }
   }
}

String::String(double x, int precision, int maxLeadingZeros, int exponentWidth, int base, int flags)
{
   if(isnan(x))
   {
      *this = $("nan");
      return;
   }
   else if(!isfinite(x))
   {
      if(x < 0)
         *this = $("-infinity");
      else
         *this = $("infinity");
      return;
   }

   --precision;  //magic fix for my off-by-one logic!

   if(precision < 0)
      precision = 0;
   if(base < 2)
      base = 2;

   char buffer[precision + maxLeadingZeros + exponentWidth + 64];
   int pos = 0;
   if(x < 0)
   {
      buffer[pos++] = '-';
      x = -x;
   }
   else if(flags & ALWAYS_SHOW_SIGN)
      buffer[pos++] = '+';
   else if(flags & BLANK_POSITIVE_SIGN)
      buffer[pos++] = ' ';

   int exponent;
   if(x == 0.0)
      exponent = 0;
   else switch(base)
   {
   case 2:
      frexp(x, &exponent);
      --exponent;
      break;
   case 4:
      frexp(x, &exponent);
      exponent = (exponent - 1) / 2;
      break;
   case 8:
      frexp(x, &exponent);
      exponent = (exponent - 1) / 3;
      break;
   case 16:
      frexp(x, &exponent);
      exponent = (exponent - 1) / 4;
      break;
   case 10:
      exponent = (int)::floor(::log10(x));
      break;
   default:
      exponent = (int)::floor(::log(x) / ::log((double)base));
      break;
   }

   /* FIXME:  Really small numbers are printed wrong because this
    *         results in infinity. */
   x = x * pow((double)base, (double)(precision - exponent));
   int digitsStart = pos;
   if(x < (double)0xFFFFFFFE)
   {
      uint32 i = (uint32)x;
      x -= i;
      if(x > 0.5)
      {
         ++i;
         if(i >= pow((double)base, (double)(precision+1)))
         {
            i /= base;
            ++exponent;
         }
      }

      pos += writeInt(buffer + pos, i, precision, base, flags);
   }
   else
   {
//      while(x > (double)0xFFFFFFFFFFFFFFFE)
//      {
//         --precision;
//         x /= base;
//      }
      uint64 i = (uint64)x;
      x -= i;
      if(x > 0.5)
      {
         ++i;
         if(i >= pow((double)base, (double)(precision+1)))
         {
            i /= base;
            ++exponent;
         }
      }

      pos += writeInt(buffer + pos, i, precision, base, flags);
   }

   if((flags & SCIENTIFIC) || exponent > precision || exponent < -maxLeadingZeros)
   {
      //print in scientific mode
      if(precision > 0)
      {
         memmove(buffer + digitsStart + 2,
                 buffer + digitsStart + 1,
                 pos - digitsStart - 1);
         ++pos;
         buffer[digitsStart + 1] = '.';

         if((flags & ALWAYS_TRUNCATE_ZEROS) || (!(flags & NEVER_TRUNCATE_ZEROS) && x == 0.0))
         {
            while(buffer[pos-1] == '0')
               --pos;
            if(!(flags & ALWAYS_SHOW_POINT) && buffer[pos-1] == '.')
               --pos;
         }
      }
      else if(flags & ALWAYS_SHOW_POINT)
         buffer[pos++] = '.';

      if(exponent != 0)
      {
         buffer[pos++] = '*';
         buffer[pos++] = '1';
         buffer[pos++] = '0';
         buffer[pos++] = '^';
         pos += writeInt(buffer + pos, exponent, exponentWidth, base, flags);
      }
   }
   else if(exponent == precision)
   {
      if(flags & ALWAYS_SHOW_POINT)
         buffer[pos++] = '.';
   }
   else if(exponent >= 0)
   {
      memmove(buffer + digitsStart + exponent + 2,
              buffer + digitsStart + exponent + 1,
              pos - digitsStart - (exponent + 1));
      ++pos;
      buffer[digitsStart + exponent + 1] = '.';

      if((flags & ALWAYS_TRUNCATE_ZEROS) || (!(flags & NEVER_TRUNCATE_ZEROS) && x == 0.0))
      {
         while(buffer[pos-1] == '0')
            --pos;
         if(!(flags & ALWAYS_SHOW_POINT) && buffer[pos-1] == '.')
            --pos;
      }
   }
   else
   {
      memmove(buffer + digitsStart + (1 - exponent),
              buffer + digitsStart,
              pos - digitsStart);
      pos += 1 - exponent;

      buffer[digitsStart++] = '0';
      buffer[digitsStart++] = '.';
      for(int i = -1; i > exponent; i--)
         buffer[digitsStart++] = '0';

      if((flags & ALWAYS_TRUNCATE_ZEROS) || (!(flags & NEVER_TRUNCATE_ZEROS) && x == 0.0))
      {
         while(buffer[pos-1] == '0')
            --pos;
         if(!(flags & ALWAYS_SHOW_POINT) && buffer[pos-1] == '.')
            --pos;
      }
   }

   Array<char> chars(pos);
   memcpy(&chars[0], buffer, pos);
   mBuffer = chars.toConst();
}

String::String(int32 n, int width, int base, int flags)
{
   if(base < 2)
      base = 2;
   char buffer[width + 33];
   int length = writeInt(buffer, n, width, base, flags);
   Array<char> chars(length);
   memcpy(&chars[0], buffer, length);
   mBuffer = chars.toConst();
}
String::String(uint32 n, int width, int base, int flags)
{
   if(base < 2)
      base = 2;
   char buffer[width + 33];
   int length = writeInt(buffer, n, width, base, flags);
   Array<char> chars(length);
   memcpy(&chars[0], buffer, length);
   mBuffer = chars.toConst();
}
String::String(int64 n, int width, int base, int flags)
{
   if(base < 2)
      base = 2;
   char buffer[width + 65];
   int length = writeInt(buffer, n, width, base, flags);
   Array<char> chars(length);
   memcpy(&chars[0], buffer, length);
   mBuffer = chars.toConst();
}
String::String(uint64 n, int width, int base, int flags)
{
   if(base < 2)
      base = 2;
   char buffer[width + 65];
   int length = writeInt(buffer, n, width, base, flags);
   Array<char> chars(length);
   memcpy(&chars[0], buffer, length);
   mBuffer = chars.toConst();
}


// copyToBuffer() ----------------------------------------------------
void String::copyToBuffer(char buffer[], int startPos, int numChars, bool nullTerminate) const
{
   if(startPos < 0 || numChars <= 0 || startPos >= mBuffer.getSize())
   {
      if(nullTerminate)
         buffer[0] = '\0';
      return;
   }
   if(startPos + numChars > mBuffer.getSize())
      numChars = mBuffer.getSize() - startPos;
   memcpy(buffer, &mBuffer[0], numChars);
   if(nullTerminate)
      buffer[numChars] = '\0';
}


// compare() ---------------------------------------------------------
int String::compare(String other) const
{
   if(mBuffer == other.mBuffer)
      return 0;  //same string!
   if(mBuffer == null)
      return -1;
   if(other.mBuffer == null)
      return 1;

   const char* data = &mBuffer[0];
   const char* otherData = &other.mBuffer[0];

   int minSize = 0;
   if(mBuffer.getSize() <= other.mBuffer.getSize())
      minSize = mBuffer.getSize();
   else
      minSize = other.mBuffer.getSize();

   //lexographical (aka alphabetical) ordering.
   for(int i = 0; i < minSize; i++)
   {
      if(data[i] < otherData[i])
         return -1;
      else if(data[i] > otherData[i])
         return 1;
   }

   if(mBuffer.getSize() > other.mBuffer.getSize())
      return 1;
   else if(mBuffer.getSize() < other.mBuffer.getSize())
      return -1;
   else
      return 0;
}

int String::compare(const char other[]) const
{
   return compare(other, strlen(other));
}

int String::compare(const char other[], int length) const
{
   const char* data = &mBuffer[0];

   int minSize = 0;
   if(mBuffer.getSize() <= length)
      minSize = mBuffer.getSize();
   else
      minSize = length;

   //lexographical (aka alphabetical) ordering.
   for(int i = 0; i < minSize; i++)
   {
      if(data[i] < other[i])
         return -1;
      else if(data[i] > other[i])
         return 1;
   }

   if(mBuffer.getSize() > length)
      return 1;
   else if(mBuffer.getSize() < length)
      return -1;
   else
      return 0;
}

int String::compareCaseInsensitive(String other) const
{
   if(mBuffer == other.mBuffer)
      return 0;  //same string!

   const char* data = &mBuffer[0];
   const char* otherData = &other.mBuffer[0];

   int minSize = 0;
   if(mBuffer.getSize() <= other.mBuffer.getSize())
      minSize = mBuffer.getSize();
   else
      minSize = other.mBuffer.getSize();

   //lexographical (aka alphabetical) ordering.
   for(int i = 0; i < minSize; i++)
   {
      if(::tolower(data[i]) < ::tolower(otherData[i]))
         return -1;
      else if(::tolower(data[i]) > ::tolower(otherData[i]))
         return 1;
   }

   if(mBuffer.getSize() > other.mBuffer.getSize())
      return 1;
   else if(mBuffer.getSize() < other.mBuffer.getSize())
      return -1;
   else
      return 0;
}

int String::compareCaseInsensitive(const char other[]) const
{
   return compare(other, strlen(other));
}

int String::compareCaseInsensitive(const char other[], int length) const
{
   const char* data = &mBuffer[0];

   int minSize = 0;
   if(mBuffer.getSize() <= length)
      minSize = mBuffer.getSize();
   else
      minSize = length;

   //lexographical (aka alphabetical) ordering.
   for(int i = 0; i < minSize; i++)
   {
      if(tolower(data[i]) < tolower(other[i]))
         return -1;
      else if(tolower(data[i]) > tolower(other[i]))
         return 1;
   }

   if(mBuffer.getSize() > length)
      return 1;
   else if(mBuffer.getSize() < length)
      return -1;
   else
      return 0;
}


// isEqualTo ---------------------------------------------------------
bool String::isEqualTo(String other) const
{
   if(mBuffer == other.mBuffer)
      return true;

   if(mBuffer.getSize() != other.mBuffer.getSize())
      return false;
   const char* data = &mBuffer[0];
   const char* otherData = &other.mBuffer[0];
   for(int i = 0; i < mBuffer.getSize(); i++)
   {
      if(data[i] != otherData[i])
         return false;
   }
   return true;
}

bool String::isEqualTo(const char other[]) const
{
   return isEqualTo(other, strlen(other));
}

bool String::isEqualTo(const char other[], int length) const
{
   if(mBuffer.getSize() != length)
      return false;
   const char* data = &mBuffer[0];
   for(int i = 0; i < length; i++)
   {
      if(data[i] != other[i])
         return false;
   }
   return true;
}


// find*Of() ---------------------------------------------------------
int String::findFirstOf(char c, int start) const
{
   if(start < 0 || start >= mBuffer.getSize())
      return -1;
   for(int i = start; i < mBuffer.getSize(); i++)
   {
      if(mBuffer[i] == c)
         return i;
   }
   return -1;
}

int String::findFirstOf(String str, int start) const
{
   if(start < 0 || start >= mBuffer.getSize() - str.getLength())
      return -1;
   int stop = mBuffer.getSize() - str.getLength();
   int j;
   for(int i = start; i < stop; i++)
   {
      for(j = 0; j < str.getLength(); j++)
         if(mBuffer[i+j] != str.mBuffer[j])
            break;
      if(j == str.getLength())
         return i;
   }
   return -1;
}

int String::findLastOf(char c, int start) const
{
   if(start < 0 || start >= mBuffer.getSize())
      start = mBuffer.getSize();
   for(int i = start - 1; i >= 0; i--)
   {
      if(mBuffer[i] == c)
         return i;
   }
   return -1;
}

int String::findLastOf(String str, int start) const
{
   if(start < 0 || start >= mBuffer.getSize() - str.getLength())
      start = mBuffer.getSize() - str.getLength();
   int j;
   for(int i = start - 1; i >= 0; i--)
   {
      for(j = 0; j < str.getLength(); j++)
         if(mBuffer[i+j] != str.mBuffer[j])
            break;
      if(j == str.getLength())
         return i;
   }
   return -1;
}


// getInterval() -----------------------------------------------------
String String::getInterval(int start) const
{
   if(start >= mBuffer.getSize() || start < 0)
      return null;
   //no need to copy the buffer
   return String(mBuffer.getInterval(start));
}

String String::getInterval(int start, int end) const
{
   if(end < 0 || end > mBuffer.getSize())
      end = mBuffer.getSize();
   if(start >= mBuffer.getSize() || start < 0 || end <= start)
      return null;
   //no need to copy the buffer
   return String(mBuffer.getInterval(start, end));
}


// to*() -------------------------------------------------------------
int String::toInt(int* endPoint) const
{
   if(mBuffer == null || isspace(getChar(0)))
   {
      //strtol accepts whitespace before the number.  We don't.
      if(endPoint != null)
         *endPoint = 0;
      return 0;
   }

   //sadly, strtol wants null-termination, so we have to make a copy.
   char buffer[mBuffer.getSize() + 1];
   char* endPtr;
   copyToBuffer(buffer, 0, mBuffer.getSize(), true);
   int result = ::strtol(buffer, &endPtr, 0);
   if(endPoint != null)
      *endPoint = endPtr - buffer;
   return result;
}

uint String::toUInt(int* endPoint) const
{
   if(mBuffer == null || isspace(getChar(0)))
   {
      //strtol accepts whitespace before the number.  We don't.
      if(endPoint != null)
         *endPoint = 0;
      return 0;
   }

   //sadly, strtol wants null-termination, so we have to make a copy.
   char buffer[mBuffer.getSize() + 1];
   char* endPtr;
   copyToBuffer(buffer, 0, mBuffer.getSize(), true);
   uint result = ::strtoul(buffer, &endPtr, 0);
   if(endPoint != null)
      *endPoint = endPtr - buffer;
   return result;
}

double String::toFloat(int* endPoint) const
{
   if(mBuffer == null || isspace(getChar(0)))
   {
      //strtol accepts whitespace before the number.  We don't.
      if(endPoint != null)
         *endPoint = 0;
      return 0.0;
   }

   //sadly, strtod wants null-termination, so we have to make a copy.
   char buffer[mBuffer.getSize() + 1];
   char* endPtr;
   copyToBuffer(buffer, 0, mBuffer.getSize(), true);
   double result = ::strtod(buffer, &endPtr);
   if(endPoint != null)
      *endPoint = endPtr - buffer;
   return result;
}


// append() ----------------------------------------------------------
void String::append(String other)
{
   if(other.mBuffer.getSize() == 0)
      return;
   if(mBuffer.getSize() == 0)
   {
      *this = other;
      return;
   }

   int newLength = mBuffer.getSize() + other.mBuffer.getSize();
   Array<char> buf(newLength);
   memcpy(&buf[0], &mBuffer[0], mBuffer.getSize());
   memcpy(&buf[0] + mBuffer.getSize(), &other.mBuffer[0], other.mBuffer.getSize());
   mBuffer = buf.reinterpretCast<CharArray>();
}

void String::append(char other)
{
   int newLength = mBuffer.getSize() + 1;
   Array<char> buf(newLength);
   if(mBuffer.getSize() > 0)
      memcpy(&buf[0], &mBuffer[0], mBuffer.getSize());
   buf[mBuffer.getSize()] = other;
   mBuffer = buf.reinterpretCast<CharArray>();
}

void String::append(const char other[])
{
   append(other, strlen(other));
}

void String::append(const char other[], int length)
{
   if(length <= 0)
      return;
   if(mBuffer.getSize() == 0)
   {
      *this = String(other, length);
      return;
   }

   int newLength = mBuffer.getSize() + length;
   Array<char> buf(newLength);
   memcpy(&buf[0], &mBuffer[0], mBuffer.getSize());
   memcpy(&buf[0] + mBuffer.getSize(), other, length);
   mBuffer = buf.reinterpretCast<CharArray>();
}


// prepend() ---------------------------------------------------------
void String::prepend(String other)
{
   if(other.mBuffer.getSize() == 0)
      return;
   if(mBuffer.getSize() == 0)
   {
      *this = other;
      return;
   }

   int newLength = mBuffer.getSize() + other.mBuffer.getSize();
   Array<char> buf(newLength);
   memcpy(&buf[0], &other.mBuffer[0], other.mBuffer.getSize());
   memcpy(&buf[0] + other.mBuffer.getSize(), &mBuffer[0], mBuffer.getSize());
   mBuffer = buf.reinterpretCast<CharArray>();
}

void String::prepend(char other)
{
   int newLength = mBuffer.getSize() + 1;
   Array<char> buf(newLength);
   if(mBuffer.getSize() > 0)
      memcpy(&buf[0] + 1, &mBuffer[0], mBuffer.getSize());
   buf[0] = other;
   mBuffer = buf.reinterpretCast<CharArray>();
}

void String::prepend(const char other[])
{
   prepend(other, strlen(other));
}

void String::prepend(const char other[], int length)
{
   if(length <= 0)
      return;
   if(mBuffer.getSize() == 0)
   {
      *this = String(other, length);
      return;
   }

   int newLength = mBuffer.getSize() + length;
   Array<char> buf(newLength);
   memcpy(&buf[0], other, length);
   memcpy(&buf[0] + length, &mBuffer[0], mBuffer.getSize());
   mBuffer = buf.reinterpretCast<CharArray>();
}


// insert() ----------------------------------------------------------
String String::insert(int pos, String other)
{
   if(other == null)
      return *this;
   if(pos < 0)
      pos = mBuffer.getSize();

   if(pos == mBuffer.getSize())
      return *this + other;
   else if(pos == 0)
      return other + *this;

   Array<char> buffer(mBuffer.getSize() + other.getLength());
   memcpy(&buffer[0], &mBuffer[0], pos);
   memcpy(&buffer[pos], &other.mBuffer[0], other.mBuffer.getSize());
   memcpy(&buffer[pos + other.mBuffer.getSize()], &mBuffer[pos], mBuffer.getSize() - pos);
   return String(buffer.toConst());
}

String String::erase(int startPos, int endPos)
{
   if(endPos < 0)
      endPos = mBuffer.getSize();
   if(startPos < 0)
      startPos = mBuffer.getSize();
   if(endPos <= startPos)
      return *this;
   if(startPos >= mBuffer.getSize() || endPos == 0)
      return null;

   if(startPos == 0 && endPos == mBuffer.getSize())
      return null;
   else if(startPos == 0)
      return getInterval(endPos);
   else if(endPos == mBuffer.getSize())
      return getInterval(0, startPos);

   Array<char> buffer(mBuffer.getSize() - endPos + startPos);
   memcpy(&buffer[0], &mBuffer[0], startPos);
   memcpy(&buffer[startPos], &mBuffer[endPos], mBuffer.getSize() - endPos);
   return String(buffer.toConst());
}

String String::replace(int startPos, int endPos, String other)
{
   if(other == null)
      return erase(startPos, endPos);
   if(endPos < 0)
      endPos = mBuffer.getSize();
   if(startPos < 0)
      startPos = mBuffer.getSize();
   if(endPos <= startPos)
      return insert(startPos, other);
   if(startPos >= mBuffer.getSize() || endPos == 0)
      return other;

   if(startPos == 0 && endPos == mBuffer.getSize())
      return other;
   else if(startPos == 0)
      return other + getInterval(endPos);
   else if(endPos == mBuffer.getSize())
      return getInterval(0, startPos) + other;

   Array<char> buffer(mBuffer.getSize() - endPos + startPos + other.getLength());
   memcpy(&buffer[0], &mBuffer[0], startPos);
   memcpy(&buffer[startPos], &other.mBuffer[0], other.mBuffer.getSize());
   memcpy(&buffer[startPos + other.mBuffer.getSize()], &mBuffer[endPos], mBuffer.getSize() - endPos);
   return String(buffer.toConst());
}

} //namespace common
} //namespace framework
