/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2003/04/11 (KV)
 *  + Imported code from Evlan.
 */

// ===================================================================

#ifndef framework_common_String_h__INCLUDED
#define framework_common_String_h__INCLUDED

#include "Array.h"
#include <cstring>

namespace framework {
namespace common {

// class definition ==================================================
/* class String
 *
 * Represents a text string in Unicode-32 format.
 *
 * Pass by:  Value (String)
 *
 * String represents an array of characters.
 *
 * On the outside, String works like any other string class.  Use it like you
 * would any other.
 *
 * String uses reference counting internally.  So, there is no performance
 * concern with passing a String by value.  Operations which change the string
 * automatically make a copy of the data.  So, you can pretend that the string
 * data is actually being passed by value.
 *
 * Also note that String stores a length along with its buffer rather than use
 * a null terminator character, so sub-string operations are trivial and never
 * require memory management or copying.  Gee, why didn't *you* think of
 * that?
 *
 * This code and it's documentation is mostly copied from another project of
 * mine (which has since been abandoned).  In that project, strings always
 * used 32-bit Unicode characters.  Here, we have little use for Unicode
 * (Evlan source code certainly doesn't use it), so I converted it to 8-bit
 * ASCII.  However, I may have missed some comments, so if you see some strange
 * comments about Unicode or 32-bitness, please disregard. */
class String
{
public:
   typedef Array<const char> CharArray;
   // constructors ---------------------------------------------------

   //Construct an empty string.
   inline String() {}
   //Construct an empty string.
   inline String(NullType) {}

   //Initialize the string to a null-terminated ASCII character array.
   explicit String(const char other[]);
   //Initialize the string to an ASCII character array of the given size.
   String(const char other[], int length);
   //Initialize the string to a single ASCII character.
   explicit String(char value);
   //Initialize the string to multiple copies of an ASCII character.
   explicit String(char value, int count);

   enum NumberFormatFlags
   {
      ALWAYS_SHOW_SIGN      = 1 << 0,
      BLANK_POSITIVE_SIGN   = 1 << 1,
      CAPITAL_DIGITS        = 1 << 2,

      SCIENTIFIC            = 1 << 3,
      ALWAYS_SHOW_POINT     = 1 << 4,
      NEVER_TRUNCATE_ZEROS  = 1 << 5,
      ALWAYS_TRUNCATE_ZEROS = 1 << 6     //truncate even when non-exact
   };

   //Initialize the string to a human-readable representation of a real number.
   explicit String(double x, int precision = 8, int maxLeadingZeros = 4,
                   int exponentWidth = 0, int base = 10, int flags = 0);
   //Initialize the string to a human-readable representation of an integer.
   explicit String(int32  n, int width = 0, int base = 10, int flags = 0);
   explicit String(uint32 n, int width = 0, int base = 10, int flags = 0);
   explicit String(int64  n, int width = 0, int base = 10, int flags = 0);
   explicit String(uint64 n, int width = 0, int base = 10, int flags = 0);

   /* Manually initialize the string.
    * Parameters:
    *   buffer  A pointer to a buffer which the string should use. */
   inline explicit String(CharArray buffer)
      : mBuffer(buffer) {}
   inline explicit String(Array<char> buffer)
      : mBuffer(buffer.toConst()) {}

   static inline String makeStatic(const char other[], int length)
      {return String(CharArray(SmartPointer<const char>::makeStatic(other), length));}
   static inline String makeStatic(const char other[])
      {return String(CharArray(SmartPointer<const char>::makeStatic(other), strlen(other)));}

   //Destructor not needed.  Cool, eh?
   //inline ~String() {}

   // comparison -----------------------------------------------------
   /* String comparison.  Works like strcmp.
    *
    * Returns: Negative if this < other, zero if this == other, or positive if
    *          this > other. */
   int compare(String other) const;
   //Compare to a null-terminated ASCII character array.
   int compare(const char other[]) const;
   //Compare to an ASCII character array of the given length.
   int compare(const char other[], int length) const;
   //Compare to null (empty string).
   inline int compare(NullType) const
      {return mBuffer == null ? 0 : 1;}

   int compareCaseInsensitive(String other) const;
   int compareCaseInsensitive(const char other[]) const;
   int compareCaseInsensitive(const char other[], int length) const;

   //Test if two strings are equal (faster than compare()).
   bool isEqualTo(String other) const;
   bool isEqualTo(const char other[]) const;
   bool isEqualTo(const char other[], int length) const;
   //Test if a string is null (empty).
   inline bool isEqualTo(NullType) const
      {return mBuffer == null ? true : false;}


   // concatination --------------------------------------------------
   //Concatinate two strings.
   void append(String other);
   /* Append a character to the end of the string.
    *
    * Remarks:  If you find that you are appending single characters to strings
    *           a lot, you may want to re-think your algorithms.  This, like all
    *           the append() functions, is an O(n) operation, and requires a
    *           memory allocation (slow). */
   void append(char other);
   //Append a null-terminated ASCII character array to the end of the string.
   void append(const char other[]);
   /* Append a null-terminated ASCII character array of the given length to the
    * end of the string. */
   void append(const char other[], int length);

   //Prepend a string to the beginning of the string.
   void prepend(String other);
   //Prepend a character to the beginning of the string.
   void prepend(char other);
   /* Prepend a null-terminated ASCII character array to the beginning of the
    * string. */
   void prepend(const char other[]);
   /* Prepend a null-terminated ASCII character array of the given length to
    * the beginning of the string. */
   void prepend(const char other[], int length);


   // other operations -----------------------------------------------
   //Get length of string.
   inline int getLength() const
      {return mBuffer.getSize();}


   //Gets a smart pointer to the character array which holds the string data.
   inline CharArray getBuffer() const
      {return mBuffer;}

   /* Copy the string data into the given buffer.
    *
    * Parameters:
    *  buffer        The buffer to which to copy the string.
    *  startPos      The index of the first character in the string which
    *                should be copied.
    *  numChars      The number of characters to copy into the buffer.  If this
    *                is larger than the length of the string, the entire string
    *                will be copied.  (Generally, you want to give the length
    *                of the buffer to which you are copying, or that length
    *                minus one if you set nullTerminate to true.)
    *  nullTerminate Whether or not to add a null-terminator to the end of the
    *                string.  (The null terminator is NOT accounted for in
    *                numChars.  Thus, if you want a null terminator, your
    *                buffer must be at least numChars + 1 bytes in size.)
    *
    * Remarks:
    *  This is mainly useful for two purposes:  Performing extensive editing on
    *  the string data, and interfacing with external code that uses a
    *  different representation for strings.  Note that if you just want
    *  to quickly allocate a null-terminated version of a string on the
    *  stack, you can use the FW_MAKE_CSTRING() macro, defined later in
    *  this file. */
   void copyToBuffer(char buffer[], int startPos, int numChars, bool nullTerminate = false) const;
   inline void copyToBuffer(char buffer[], bool nullTerminate = false) const
      {copyToBuffer(buffer, 0, mBuffer.getSize(), nullTerminate);}

   //Get the given character.
   inline char getChar(int pos) const
      {return mBuffer[pos];}


   /* Returns the position of the first instance of c at or after the start
    * position in the string, or -1 if not found. */
   int findFirstOf(char c, int start = 0) const;
   int findFirstOf(String str, int start = 0) const;

   /* Returns the position of the last instance of c before the start position
    * in the string, or -1 if not found. */
   int findLastOf(char c, int start = -1) const;
   int findLastOf(String str, int start = -1) const;


   /* Returns all text at and after the given start position. (fast)
    *
    * Remarks: This function is implemented using copy-on-write techniques,
    *          just like string assignment.  This means that no dynamic memory
    *          operations are used to perform this operation.  However, this
    *          also means that if you take a substring of a very large string,
    *          the substring will count as a reference on the large string, so
    *          the large string will remain in memory until the substring is
    *          changed or deleted.  To force a substring to allocate new memory
    *          for itself rather than refer to its parent, make a copy by
    *          calling copy(). */
   String getInterval(int start) const;
   /*! Returns all text from the string interval [start, end). (fast)
    *
    * Remarks: This function is implemented using copy-on-write techniques,
    *          just like string assignment.  This means that no dynamic memory
    *          operations are used to perform this operation.  However, this
    *          also means that if you take a substring of a very large string,
    *          the substring will count as a reference on the large string, so
    *          the large string will remain in memory until the substring is
    *          changed or deleted.  To force a substring to allocate new memory
    *          for itself rather than refer to its parent, make a copy by
    *          calling copy(). */
   String getInterval(int start, int end) const;

   /* Force allocation of a separate copy of the string data.
    *
    * Sometimes a string (particularily one returned by GetInterval) may be
    * taking up more memory than it really needs.  Calling copy() will
    * allocate and copy the string data, returning a string that uses the
    * minimal amount of memory possible to represent the data.  You can then
    * discard the original string to free up the excess memory used.
    *
    * Note: This will not always decrease total memory usage of your program.
    *       If other strings hold references to the same buffer as the one
    *       being copied, then that buffer will not be deleted.  However, a
    *       new buffer will still be allocated, using *more* memory.  So,
    *       don't call this function unless you know what you are doing. */
   inline String copy()
   {
      if(mBuffer == null)
         return null;
      else
         return String(&mBuffer[0], mBuffer.getSize());
   }


   /* Converts a text representation of an integer into an int.
    *
    * Parameters:
    *  endPoint  If this is not null, the integer pointed to will be filled in
    *            with the index of the first non-numerical character found.  If
    *            no non-numerical characters were found, *endPoint will be set
    *            to the length of the string.  If the string did not start with
    *            a number, *endPoint will be zero (and the function returns
    *            zero). */
   int toInt(int* endPoint = null) const;
   uint toUInt(int* endPoint = null) const;

   /* Converts a text representation of an real number into a floating point
    * value.
    *
    * Parameters:
    *  endPoint  If this is not null, the integer pointed to will be filled in
    *            with the index of the first non-numerical character found.  If
    *            no non-numerical characters were found, *endPoint will be set
    *            to the length of the string.  If the string did not start with
    *            a number, *endPoint will be zero (and the function returns
    *            zero). */
   double toFloat(int* endPoint = null) const;


   //Insert the given string before the given position.
   String insert(int pos, String other);

   //Erase all characters between startPos and endPos (including startPos but not endPos).
   String erase(int startPos, int endPos);

   //Replace the interval [startPos, endPos) with the given string.
   String replace(int startPos, int endPos, String other);


   // operators ------------------------------------------------------
   //Assign to null...
   inline String& operator=(NullType)
      {mBuffer = null; return *this;}

   //Comparison operators.
   template<typename Type>
   inline bool operator<=(Type other) const
      {return compare(other) <= 0;}
   template<typename Type>
   inline bool operator>=(Type other) const
      {return compare(other) >= 0;}
   template<typename Type>
   inline bool operator< (Type other) const
      {return compare(other) <  0;}
   template<typename Type>
   inline bool operator> (Type other) const
      {return compare(other) >  0;}

   //Equality operators.
   template<typename Type>
   inline bool operator==(Type other) const
      {return  isEqualTo(other);}
   template<typename Type>
   inline bool operator!=(Type other) const
      {return !isEqualTo(other);}

   //Concatination operators.
   template<typename Type>
   inline String  operator+ (Type other) const
      { String ret = *this; ret.append(other); return ret; }
   //make GCC not spew warnings.
   inline String  operator+ (String other) const
      { String ret = *this; ret.append(other); return ret; }
   template<typename Type>
   inline String& operator+=(Type other)
      { append(other); return *this; }

   /* Get given character in string.  Not mutable.
    *
    * Remarks: You may wonder why you can't modify the character returned by
    *          operator[].  It seems intuitive to be able to do so, since you
    *          can with regular character arrays.  Well, the problem is, if you
    *          could modify the character, then operator[] would be a write
    *          operation.  Since String is copy-on-write, this would make the
    *          operator much more expensive to use.  Furthermore, operator[] is
    *          used for reading far more often than writing, so making
    *          operator[] a write operation would needlessly decrease read
    *          performance!  If you want to manipulate individual characters
    *          in a string, use CopyToBuffer() to copy the string to a char
    *          array first, and play with that.  Remember that GCC lets you
    *          create dynamicly-sized arrays on the stack!  If I catch ANYONE
    *          writing code with buffer overflow vounerabilities, I WILL hurt
    *          them! */
   inline char operator[](int pos) const
      {return getChar(pos);}

private:
   CharArray mBuffer;   //the reference counted data
};

/* This macro is useful for interfacing with system calls and other
 * libraries that want null-terminated character array representations
 * of strings.  It will create a null-terminated copy of the string on
 * the stack. */
#define FW_MAKE_CSTRING(CSTRING_NAME, STRING)                          \
   char CSTRING_NAME[(STRING).getLength() + 1];                        \
   (STRING).copyToBuffer(CSTRING_NAME, 0, (STRING).getLength(), true);

/* This macro creates a String literal from a quoted character array
 * literal.  This ONLY works with literals.  It is useful because it
 * skips the memory allocation and copy normally needed when creating
 * a String from a character array. */
#define $(LITERAL)                                                    \
   (::framework::common::String::makeStatic(LITERAL, sizeof(LITERAL) - 1))

/* Concatination operator.  First parameter can be any of those accepted by
 * String::Prepend(). */
template<typename Type>
String operator+(Type other, String str)
{
   str.prepend(other);
   return str;
}

} //namespace common
} //namespace framework

#endif //included
