/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2005/03/06 (KV)
 *  + First Version (copied from win32 version)
 */

// ===================================================================

#include "SystemImpl.h"
#include <framework/messages/EventManager.h>
#include <framework/messages/MessageListImpl.h>
#include <framework/internal/MessageImpl.h>
#include <queue>
#include <map>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>
#include <aio.h>
#include <signal.h>

namespace framework {
namespace internal {
namespace freebsd {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement framework::internal::freebsd::SystemImpl
// TODO (KV): Review framework::internal::freebsd::SystemImpl

implementation SystemImpl: implements System
{
   SystemImpl(String programName, String executablePath);
   ~SystemImpl();

   // System ---------------------------------------------------------
   ClockPtr getClock();

   String getCurrentDirectory();
   String getSystemConfigDirectory();
   String getUserConfigDirectory();
   String getDataFileDirectory();

   void createDirectory(String path);
   String createDirectory(String path, String subPath);

   messages::EventManagerPtr getEventManager();

   // internal -------------------------------------------------------
   bool wait(const struct timespec* timeout);
   void mainLoop();

   implementation ClockImpl: implements Clock
   {
      ClockImpl();
      ~ClockImpl();

      // Clock -------------------------------------------------------
      double getTime();
      double getRealTime();

      void notifyAtTime(double time, messages::MessagePtr message);

      // internal ----------------------------------------------------
      struct TimerEvent
      {
         double time;
         messages::MessagePtr message;

         inline bool operator<(const TimerEvent& other) const
            {return time > other.time;}
      };

      double mCurrentTime;
      static const int ORIGIN_SECOND = 946684800;  //2000, in seconds from 1970
      std::priority_queue<TimerEvent> mTimerEvents;
   };
   typedef SmartPointer<ClockImpl> ClockImplPtr;

   implementation EventManagerImpl: implements messages::EventManager
   {
      EventManagerImpl();
      ~EventManagerImpl();

      // EventManager ------------------------------------------------
      StreamWatcherPtr makeStreamWatcher(int socket);
      AioStreamPtr makeAioStream(int fileDiscriptor);
      KqueueWatcherPtr makeKqueueWatcher(word ident, short filter,
                                         uint fflags, word data);

      // internal ----------------------------------------------------
      int mKqueue;

      static const int MAX_EVENTS = 4;

      //Number of events returned last time kevent() was called.
      int mEventCount;

      //The events returned last time kevent() was called.
      struct kevent mEvents[MAX_EVENTS];
      /* Whether or not to disable events in mEvents the next time
       * kevent() is called.  This is set true for all events when
       * kevent() returns, but will be set false if notifyOnEvent()
       * is called on a particular event before the next kevent(). */
      bool mNeedDisable[MAX_EVENTS];

      //The udata field of struct kevent points to a KeventInfo.
      struct KeventInfo
      {
         /* Index of this event in mEvents/mNeedDisable, or -1 if it
          * isn't currently listed there.  If it is listed, then the
          * event can be re-enabled by simply setting mNeedDisable[mIndex]
          * to false (because it hasn't been disabled yet).
          * Otherwise, a kevent() call must be made with EV_ENABLE to
          * enable the event again. */
         int index;
         struct kevent kevent;
         messages::MessageListPtr messages;
      };
   };
   typedef SmartPointer<EventManagerImpl> EventManagerImplPtr;

   implementation StreamWatcherImpl: implements messages::EventManager::StreamWatcher
   {
      StreamWatcherImpl(int socket, EventManagerImplPtr eventManager);
      ~StreamWatcherImpl();

      // StreamWatcher -----------------------------------------------
      void notifyWhenReadable(messages::MessagePtr message);
      void notifyWhenWritable(messages::MessagePtr message);

      // internal ----------------------------------------------------
      messages::EventManager::KqueueWatcherPtr mReadWatcher;
      messages::EventManager::KqueueWatcherPtr mWriteWatcher;
   };

   implementation AioStreamImpl: implements messages::EventManager::AioStream
   {
      AioStreamImpl(int fileDescriptor, EventManagerImplPtr eventManager);
      ~AioStreamImpl();

      // AioStream ---------------------------------------------------
      bool read(void* buffer, uint64 offset, int size, messages::MessagePtr message);
      bool write(void* buffer, uint64 offset, int size, messages::MessagePtr message);
      bool getResult(int* result);

      // internal ----------------------------------------------------
      struct aiocb mAiocb;
      EventManagerImpl::KeventInfo mInfo;

      EventManagerImplPtr mEventManager;
   };

   implementation KqueueWatcherImpl: implements messages::EventManager::KqueueWatcher
   {
      KqueueWatcherImpl(intptr_t ident, short filter, u_int fflags, intptr_t data,
                        EventManagerImplPtr eventManager);
      ~KqueueWatcherImpl();

      // KqueueWatcher -----------------------------------------------
      void getInfo(uint* fflags, word* data);
      void setInfo(uint fflags, word data);
      void clear();
      void notifyOnEvent(messages::MessagePtr message);

      // internal ----------------------------------------------------
      EventManagerImpl::KeventInfo mInfo;
      EventManagerImplPtr mEventManager;
   };

   ClockImplPtr mClock;
   EventManagerImplPtr mEventManager;

   String mCurrentDirectory;
   String mPrefix;
   bool mUnixDirectoryStyle;

   String mProgramName;
};

const struct timespec ZERO_TIMEOUT = {0, 0};

SystemImpl::SystemImpl(String programName, String executablePath)
{
   mClock = spnew ClockImpl;
   mEventManager = spnew EventManagerImpl;

   char* path = getcwd(null, 0);
   if(path == null)
      FW_THROW(spnew ErrnoError($("getcwd")));
   mCurrentDirectory = String(path);
   free(path);

   executablePath = executablePath.getInterval(0, executablePath.findLastOf('/'));
   if(executablePath.getLength() > 4
   && executablePath.getInterval(executablePath.getLength() - 4).compareCaseInsensitive($("/bin")) == 0)
   {
      mPrefix = executablePath.getInterval(0, executablePath.getLength() - 4);
      mUnixDirectoryStyle = true;
   }
   else
   {
      mPrefix = executablePath;
      mUnixDirectoryStyle = false;
   }

   mProgramName = programName;

   //Signals suck.  We will limit our signals to the following fatal list...
   sigset_t signalSet;
   sigfillset(&signalSet);
   sigdelset(&signalSet, SIGHUP);
   sigdelset(&signalSet, SIGINT);
   sigdelset(&signalSet, SIGQUIT);
   sigdelset(&signalSet, SIGILL);
   sigdelset(&signalSet, SIGTRAP);
   sigdelset(&signalSet, SIGABRT);
   sigdelset(&signalSet, SIGEMT);
   sigdelset(&signalSet, SIGFPE);
   sigdelset(&signalSet, SIGKILL);
   sigdelset(&signalSet, SIGBUS);
   sigdelset(&signalSet, SIGSEGV);
   sigdelset(&signalSet, SIGSYS);
   sigdelset(&signalSet, SIGTERM);
   sigprocmask(SIG_BLOCK, &signalSet, null);
}

SystemImpl::~SystemImpl()
{
}

// System ------------------------------------------------------------
ClockPtr SystemImpl::getClock()
{
   return mClock;
}

String SystemImpl::getCurrentDirectory()
{
   return mCurrentDirectory;
}
String SystemImpl::getSystemConfigDirectory()
{
   if(mUnixDirectoryStyle)
   {
      String dir = mPrefix + $("/etc/") + mProgramName;
      createDirectory(dir);
      return dir;
   }
   else
      return mPrefix;
}
String SystemImpl::getUserConfigDirectory()
{
   char* homedir = getenv("HOME");

   if(homedir != null)
   {
      String dir = String(homedir) + $("/.") + mProgramName;
      createDirectory(dir);
      return dir;
   }
   else
      FW_THROW(spnew FatalError($("$HOME is not set.")));
}

String SystemImpl::getDataFileDirectory()
{
   if(mUnixDirectoryStyle)
   {
      String dir = mPrefix + $("/share/") + mProgramName;
      createDirectory(dir);
      return dir;
   }
   else
      return mPrefix;
}

void SystemImpl::createDirectory(String dir)
{
   bool result;

   {
      FW_MAKE_CSTRING(cdir, dir);
      result = mkdir(cdir, 0777) == 0;
   }

   if(result)
      return;

   int error = errno;

   if(error == EEXIST)
      return;
   else if(error == ENOENT)
   {
      int pos = dir.findLastOf('/');
      if(pos == -1)
         FW_THROW(spnew ErrnoError($("mkdir"), error));

      createDirectory(dir.getInterval(0, pos));
   }
   else
      FW_THROW(spnew ErrnoError($("mkdir"), error));

   FW_MAKE_CSTRING(cdir, dir);
   if(mkdir(cdir, 0777) < 0)
      FW_THROW(spnew ErrnoError($("mkdir")));
}

String SystemImpl::createDirectory(String path, String subPath)
{
   String result = path + $("/") + subPath;
   createDirectory(result);
   return result;
}

messages::EventManagerPtr SystemImpl::getEventManager()
{
   return mEventManager;
}

// internal ----------------------------------------------------------
bool SystemImpl::wait(const struct timespec* timeout)
{
   int changeCount = 0;
   for(int i = 0; i < mEventManager->mEventCount; i++)
   {
      if(mEventManager->mNeedDisable[i])
      {
         EventManagerImpl::KeventInfo* info = (EventManagerImpl::KeventInfo*)mEventManager->mEvents[i].udata;

         mEventManager->mEvents[i].flags  = EV_DISABLE;
         mEventManager->mEvents[i].fflags = 0;
         mEventManager->mEvents[i].data   = 0;

         if(i > changeCount)
            mEventManager->mEvents[changeCount] = mEventManager->mEvents[i];
         ++changeCount;
         info->index = -2;
      }
   }

   int result =
      kevent(mEventManager->mKqueue,
             mEventManager->mEvents, changeCount,
             mEventManager->mEvents, EventManagerImpl::MAX_EVENTS,
             timeout);

   if(result > 0)
   {
      mEventManager->mEventCount = result;
      for(int i = 0; i < result; i++)
      {
         EventManagerImpl::KeventInfo* info =
            (EventManagerImpl::KeventInfo*)mEventManager->mEvents[i].udata;

         if(mEventManager->mEvents[i].filter == EVFILT_AIO)
         {
            mEventManager->mEvents[i].flags = EV_DELETE;
            mEventManager->mEvents[i].fflags = 0;
            mEventManager->mEvents[i].data = 0;

            if(kevent(mEventManager->mKqueue,
                      &mEventManager->mEvents[i], 1,
                      null, 0, &ZERO_TIMEOUT) < 0)
               FW_THROW(spnew ErrnoError($("kevent")));
            mEventManager->mNeedDisable[i] = false;
         }
         else
            mEventManager->mNeedDisable[i] = true;

         info->index = i;
         info->kevent = mEventManager->mEvents[i];
         info->messages->send();
         info->messages->clear();
      }

      return false;
   }
   else if(result == 0)
   {
      mEventManager->mEventCount = 0;
      return true;
   }
   else
   {
      mEventManager->mEventCount = 0;
      FW_THROW(spnew ErrnoError($("kevent")));
   }
}

void SystemImpl::mainLoop()
{
   while(true)
   {
      int dispatchCount = pumpMessageQueue(512);

      double time = mClock->getRealTime();

      //get rid of any disabled timer events
      while(!mClock->mTimerEvents.empty()
          && mClock->mTimerEvents.top().message->isDisabled())
         mClock->mTimerEvents.pop();

      //wait for OS events
      timespec* timeout;
      timespec waitTime;
      if(dispatchCount == 512)
      {
         //The message queue may not be empty...
         timeout = &waitTime;
         waitTime.tv_sec  = 0;
         waitTime.tv_nsec = 0;
      }
      else if(mClock->mTimerEvents.empty())
         timeout = null;
      else if(time < mClock->mTimerEvents.top().time)
      {
         timeout = &waitTime;
         double waitTimeD = mClock->mTimerEvents.top().time - time;
         waitTime.tv_nsec = (long)(modf(waitTimeD, &waitTimeD) * 1000000000.0);
         waitTime.tv_sec  = (time_t)waitTimeD;
      }
      else
      {
         timeout = &waitTime;
         waitTime.tv_sec  = 0;
         waitTime.tv_nsec = 0;
      }

      if(!wait(timeout))
         time = mClock->getRealTime();  //wait returned before timeout
      else if(waitTime.tv_sec > 0 || waitTime.tv_nsec > 0)
         time = mClock->mTimerEvents.top().time;  //wait timed out at clock event

      if(!mClock->mTimerEvents.empty()
      && time >= mClock->mTimerEvents.top().time)
      {
         //send timer events

         //back up timer to the time the event was supposed to be sent
         time = mClock->mTimerEvents.top().time;

         do
         {
            mClock->mTimerEvents.top().message->send();
            mClock->mTimerEvents.pop();
         }
         while(!mClock->mTimerEvents.empty()
            && time == mClock->mTimerEvents.top().time);
      }

      mClock->mCurrentTime = time;
   }
}

// ClockImpl =========================================================
SystemImpl::ClockImpl::ClockImpl()
{
   struct timespec startTime;
   clock_gettime(CLOCK_REALTIME, &startTime);

   startTime.tv_sec -= ORIGIN_SECOND;
   mCurrentTime = (double)startTime.tv_sec + (double)startTime.tv_nsec / 1000000000.0;
}

SystemImpl::ClockImpl::~ClockImpl()
{
}

// Clock -------------------------------------------------------------
double SystemImpl::ClockImpl::getTime()
{
   return mCurrentTime;
}

double SystemImpl::ClockImpl::getRealTime()
{
   struct timespec newTime;
   clock_gettime(CLOCK_REALTIME, &newTime);
   newTime.tv_sec -= ORIGIN_SECOND;
   return (double)newTime.tv_sec + (double)newTime.tv_nsec / 1000000000.0;
}

void SystemImpl::ClockImpl::notifyAtTime(double time, messages::MessagePtr message)
{
   if(time < mCurrentTime)
      message->send();
   else
   {
      TimerEvent event;
      event.time = time;
      event.message = message;
      mTimerEvents.push(event);
   }
}

// EventManagerImpl ==================================================
SystemImpl::EventManagerImpl::EventManagerImpl()
   : mEventCount(0)
{
   mKqueue = kqueue();

   if(mKqueue < 0)
      FW_THROW(spnew ErrnoError($("kqueue")));
}
SystemImpl::EventManagerImpl::~EventManagerImpl()
{
   close(mKqueue);
}

// EventManager ------------------------------------------------------
messages::EventManager::StreamWatcherPtr
SystemImpl::EventManagerImpl::makeStreamWatcher(int socket)
{
   return spnew StreamWatcherImpl(socket, spthis);
}

messages::EventManager::AioStreamPtr
SystemImpl::EventManagerImpl::makeAioStream(int fileDiscriptor)
{
   return spnew AioStreamImpl(fileDiscriptor, spthis);
}

messages::EventManager::KqueueWatcherPtr
SystemImpl::EventManagerImpl::makeKqueueWatcher(
   word ident, short filter, uint fflags, word data)
{
   return spnew KqueueWatcherImpl(ident, filter, fflags, data, spthis);
}

// StreamWatcher =====================================================
SystemImpl::StreamWatcherImpl::StreamWatcherImpl(int socket, EventManagerImplPtr eventManager)
{
   u_long arg = 1;
   ioctl(socket, FIONBIO, &arg);

   mReadWatcher = eventManager->makeKqueueWatcher(socket, EVFILT_READ, 0, 0);
   mWriteWatcher = eventManager->makeKqueueWatcher(socket, EVFILT_WRITE, 0, 0);
}
SystemImpl::StreamWatcherImpl::~StreamWatcherImpl()
{
}

// StreamWatcher -----------------------------------------------------
void SystemImpl::StreamWatcherImpl::notifyWhenReadable(messages::MessagePtr message)
{
   mReadWatcher->notifyOnEvent(message);
}
void SystemImpl::StreamWatcherImpl::notifyWhenWritable(messages::MessagePtr message)
{
   mWriteWatcher->notifyOnEvent(message);
}

// AioStream =========================================================
SystemImpl::AioStreamImpl::AioStreamImpl(int fileDescriptor, EventManagerImplPtr eventManager)
   : mEventManager(eventManager)
{
   mInfo.index = -1;
   mInfo.messages = messages::makeMessageListImpl();

   mAiocb.aio_fildes = fileDescriptor;
   mAiocb.aio_sigevent.sigev_notify = SIGEV_KEVENT;
   mAiocb.aio_sigevent.sigev_notify_kqueue = eventManager->mKqueue;
   mAiocb.aio_sigevent.sigev_value.sigval_ptr = &mInfo;
}
SystemImpl::AioStreamImpl::~AioStreamImpl()
{
   /* FIXME:  If an op is in progress, what do we do here?  In current
    *         practice (in Evlan) this will never happen, unless you
    *         try to "shutdown" the VM, but that already crashes for
    *         other reasons. */
}

// AioStream ---------------------------------------------------------
bool SystemImpl::AioStreamImpl::read(void* buffer, uint64 offset, int size, messages::MessagePtr message)
{
   int fd = mAiocb.aio_fildes;
   struct sigevent event = mAiocb.aio_sigevent;

   memset(&mAiocb, 0, sizeof(mAiocb));

   mAiocb.aio_fildes = fd;
   mAiocb.aio_sigevent = event;

   mAiocb.aio_offset = offset;
   mAiocb.aio_nbytes = size;
   mAiocb.aio_buf = buffer;

   bool result = aio_read(&mAiocb) == 0;
   if(result) mInfo.messages->add(message);
   return result;
}
bool SystemImpl::AioStreamImpl::write(void* buffer, uint64 offset, int size, messages::MessagePtr message)
{
   int fd = mAiocb.aio_fildes;
   struct sigevent event = mAiocb.aio_sigevent;

   memset(&mAiocb, 0, sizeof(mAiocb));

   mAiocb.aio_fildes = fd;
   mAiocb.aio_sigevent = event;

   mAiocb.aio_offset = offset;
   mAiocb.aio_nbytes = size;
   mAiocb.aio_buf = buffer;

   bool result = aio_write(&mAiocb) == 0;
   if(result) mInfo.messages->add(message);
   return result;
}
bool SystemImpl::AioStreamImpl::getResult(int* result)
{
   int error = aio_error(&mAiocb);
   if(result != null)
   {
      if(error == 0)
         *result = aio_return(&mAiocb);
      else
         *result = error;
   }
   return error == 0;
}

// KqueueWatcher =====================================================
SystemImpl::KqueueWatcherImpl::KqueueWatcherImpl(intptr_t ident, short filter,
                        u_int fflags, intptr_t data, EventManagerImplPtr eventManager)
   : mEventManager(eventManager)
{
   mInfo.index = -1;

   EV_SET(&mInfo.kevent, ident, filter, EV_ADD, fflags, data, &mInfo);
   if(kevent(mEventManager->mKqueue, &mInfo.kevent, 1, null, 0, &ZERO_TIMEOUT) < 0)
      FW_THROW(spnew ErrnoError($("kevent")));

   mInfo.messages = messages::makeMessageListImpl();
}
SystemImpl::KqueueWatcherImpl::~KqueueWatcherImpl()
{
   if(mInfo.index >= 0)
   {
      mEventManager->mNeedDisable[mInfo.index] = false;
      mInfo.index = -1;
   }

   mInfo.kevent.flags = EV_DELETE;
   mInfo.kevent.fflags = 0;
   mInfo.kevent.data = 0;

   if(kevent(mEventManager->mKqueue, &mInfo.kevent, 1, null, 0, &ZERO_TIMEOUT) < 0)
      FW_THROW(spnew ErrnoError($("kevent")));
}

// KqueueWatcher -----------------------------------------------------
void SystemImpl::KqueueWatcherImpl::getInfo(uint* fflags, word* data)
{
   if(fflags != null)
      *fflags = mInfo.kevent.fflags;
   if(data != null)
      *data = mInfo.kevent.data;
}

void SystemImpl::KqueueWatcherImpl::setInfo(uint fflags, word data)
{
   mInfo.kevent.flags = EV_ADD;
   mInfo.kevent.fflags = fflags;
   mInfo.kevent.data = data;

   if(kevent(mEventManager->mKqueue, &mInfo.kevent, 1, null, 0, &ZERO_TIMEOUT) < 0)
      FW_THROW(spnew ErrnoError($("kevent")));
}

void SystemImpl::KqueueWatcherImpl::clear()
{
   mInfo.kevent.flags = EV_CLEAR;
   mInfo.kevent.fflags = 0;
   mInfo.kevent.data = 0;

   if(kevent(mEventManager->mKqueue, &mInfo.kevent, 1, null, 0, &ZERO_TIMEOUT) < 0)
      FW_THROW(spnew ErrnoError($("kevent")));
}

void SystemImpl::KqueueWatcherImpl::notifyOnEvent(messages::MessagePtr message)
{
   if(mInfo.index >= 0)
   {
      mEventManager->mNeedDisable[mInfo.index] = false;
      mInfo.index = -1;
   }
   else if(mInfo.index == -2)
   {
      mInfo.kevent.flags = EV_ENABLE;
      mInfo.kevent.fflags = 0;
      mInfo.kevent.data = 0;

      if(kevent(mEventManager->mKqueue, &mInfo.kevent, 1, null, 0, &ZERO_TIMEOUT) < 0)
         FW_THROW(spnew ErrnoError($("kevent")));
      mInfo.index = -1;
   }

   mInfo.messages->add(message);
}

// factory function ==================================================
void startSystem(FrontEndPtr frontEnd,
                 FrontEnd::StartupInfo startupInfo,
                 String programName, String executablePath)
{
   SmartPointer<SystemImpl> system = spnew SystemImpl(programName, executablePath);
   startupInfo.system = system;
   frontEnd->run(startupInfo);
   system->mainLoop();
}

// ===================================================================

} //namespace freebsd
} //namespace internal
} //namespace framework
