/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2005/05/02 (KV)
 *  + First Version (copied from freebsd version)
 */

// ===================================================================

#include "SystemImpl.h"
#include <framework/messages/EventManager.h>
#include <framework/messages/MessageListImpl.h>
#include <framework/internal/MessageImpl.h>
#include <queue>
#include <map>
#include <list>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/poll.h>

namespace framework {
namespace internal {
namespace posix {

// ===================================================================
// AUTHOR (KV): Created module.

implementation SystemImpl: implements System
{
   SystemImpl(String programName, String executablePath);
   ~SystemImpl();

   // System ---------------------------------------------------------
   ClockPtr getClock();

   String getCurrentDirectory();
   String getSystemConfigDirectory();
   String getUserConfigDirectory();
   String getDataFileDirectory();

   void createDirectory(String path);
   String createDirectory(String path, String subPath);

   messages::EventManagerPtr getEventManager();

   // internal -------------------------------------------------------
   bool wait(int timeout);
   void mainLoop();

   implementation ClockImpl: implements Clock
   {
      ClockImpl();
      ~ClockImpl();

      // Clock -------------------------------------------------------
      double getTime();
      double getRealTime();

      void notifyAtTime(double time, messages::MessagePtr message);

      // internal ----------------------------------------------------
      struct TimerEvent
      {
         double time;
         messages::MessagePtr message;

         inline bool operator<(const TimerEvent& other) const
            {return time > other.time;}
      };

      double mCurrentTime;
      static const int ORIGIN_SECOND = 946684800;  //2000, in seconds from 1970
      std::priority_queue<TimerEvent> mTimerEvents;
   };
   typedef SmartPointer<ClockImpl> ClockImplPtr;

   implementation EventManagerImpl: implements messages::EventManager
   {
      EventManagerImpl();
      ~EventManagerImpl();

      // EventManager ------------------------------------------------
      StreamWatcherPtr makeStreamWatcher(int socket);
      AioStreamPtr makeAioStream(int fileDiscriptor);

      // internal ----------------------------------------------------
      struct StreamInfo
      {
         int fd;
         bool wantRead;
         bool wantWrite;
         messages::MessageListPtr readMessages;
         messages::MessageListPtr writeMessages;

         //only valid if wantRead or wantWrite is true
         std::list<StreamInfo*>::iterator position;
      };
      std::list<StreamInfo*> mCurrentStreams;
   };
   typedef SmartPointer<EventManagerImpl> EventManagerImplPtr;

   implementation StreamWatcherImpl: implements messages::EventManager::StreamWatcher
   {
      StreamWatcherImpl(int socket, EventManagerImplPtr eventManager);
      ~StreamWatcherImpl();

      // StreamWatcher -----------------------------------------------
      void notifyWhenReadable(messages::MessagePtr message);
      void notifyWhenWritable(messages::MessagePtr message);

      // internal ----------------------------------------------------
      EventManagerImplPtr mEventManager;
      EventManagerImpl::StreamInfo mInfo;
   };

   implementation AioStreamImpl: implements messages::EventManager::AioStream
   {
      AioStreamImpl(int fileDescriptor, EventManagerImplPtr eventManager);
      ~AioStreamImpl();

      // AioStream ---------------------------------------------------
      bool read(void* buffer, uint64 offset, int size, messages::MessagePtr message);
      bool write(void* buffer, uint64 offset, int size, messages::MessagePtr message);
      bool getResult(int* result);

      // internal ----------------------------------------------------
      int mFileDescriptor;
      int mResult;
      bool mResultIsError;
   };

   ClockImplPtr mClock;
   EventManagerImplPtr mEventManager;

   String mCurrentDirectory;
   String mPrefix;
   bool mUnixDirectoryStyle;

   String mProgramName;
};

SystemImpl::SystemImpl(String programName, String executablePath)
{
   mClock = spnew ClockImpl;
   mEventManager = spnew EventManagerImpl;

   char* path = getcwd(null, 0);
   if(path == null)
      FW_THROW(spnew ErrnoError($("getcwd")));
   mCurrentDirectory = String(path);
   free(path);

   executablePath = executablePath.getInterval(0, executablePath.findLastOf('/'));
   if(executablePath.getLength() > 4
   && executablePath.getInterval(executablePath.getLength() - 4).compareCaseInsensitive($("/bin")) == 0)
   {
      mPrefix = executablePath.getInterval(0, executablePath.getLength() - 4);
      mUnixDirectoryStyle = true;
   }
   else
   {
      mPrefix = executablePath;
      mUnixDirectoryStyle = false;
   }

   mProgramName = programName;

   //Signals suck.  We will limit our signals to the following fatal list...
   sigset_t signalSet;
   sigfillset(&signalSet);
   sigdelset(&signalSet, SIGHUP);
   sigdelset(&signalSet, SIGINT);
   sigdelset(&signalSet, SIGQUIT);
   sigdelset(&signalSet, SIGILL);
   sigdelset(&signalSet, SIGTRAP);
   sigdelset(&signalSet, SIGABRT);
//   sigdelset(&signalSet, SIGEMT);
   sigdelset(&signalSet, SIGFPE);
   sigdelset(&signalSet, SIGKILL);
   sigdelset(&signalSet, SIGBUS);
   sigdelset(&signalSet, SIGSEGV);
   sigdelset(&signalSet, SIGSYS);
   sigdelset(&signalSet, SIGTERM);
   sigprocmask(SIG_BLOCK, &signalSet, null);
}

SystemImpl::~SystemImpl()
{
}

// System ------------------------------------------------------------
ClockPtr SystemImpl::getClock()
{
   return mClock;
}

String SystemImpl::getCurrentDirectory()
{
   return mCurrentDirectory;
}
String SystemImpl::getSystemConfigDirectory()
{
   if(mUnixDirectoryStyle)
   {
      String dir = mPrefix + $("/etc/") + mProgramName;
      createDirectory(dir);
      return dir;
   }
   else
      return mPrefix;
}
String SystemImpl::getUserConfigDirectory()
{
   char* homedir = getenv("HOME");

   if(homedir != null)
   {
      String dir = String(homedir) + $("/.") + mProgramName;
      createDirectory(dir);
      return dir;
   }
   else
      FW_THROW(spnew FatalError($("$HOME is not set.")));
}

String SystemImpl::getDataFileDirectory()
{
   if(mUnixDirectoryStyle)
   {
      String dir = mPrefix + $("/share/") + mProgramName;
      createDirectory(dir);
      return dir;
   }
   else
      return mPrefix;
}

void SystemImpl::createDirectory(String dir)
{
   bool result;

   {
      FW_MAKE_CSTRING(cdir, dir);
      result = mkdir(cdir, 0777) == 0;
   }

   if(result)
      return;

   int error = errno;

   if(error == EEXIST)
      return;
   else if(error == ENOENT)
   {
      int pos = dir.findLastOf('/');
      if(pos == -1)
         FW_THROW(spnew ErrnoError($("mkdir"), error));

      createDirectory(dir.getInterval(0, pos));
   }
   else
      FW_THROW(spnew ErrnoError($("mkdir"), error));

   FW_MAKE_CSTRING(cdir, dir);
   if(mkdir(cdir, 0777) < 0)
      FW_THROW(spnew ErrnoError($("mkdir")));
}

String SystemImpl::createDirectory(String path, String subPath)
{
   String result = path + $("/") + subPath;
   createDirectory(result);
   return result;
}

messages::EventManagerPtr SystemImpl::getEventManager()
{
   return mEventManager;
}

// internal ----------------------------------------------------------
bool SystemImpl::wait(int timeout)
{
   int count = mEventManager->mCurrentStreams.size();

   struct pollfd pollfds[count];
   EventManagerImpl::StreamInfo* streamInfos[count];

   memset(pollfds, 0, count * sizeof(struct pollfd));

   {
      int i = 0;
      std::list<EventManagerImpl::StreamInfo*>::iterator iter =
         mEventManager->mCurrentStreams.begin();
      while(iter != mEventManager->mCurrentStreams.end())
      {
         streamInfos[i] = *iter;

         pollfds[i].fd = streamInfos[i]->fd;
         if(streamInfos[i]->wantRead)
            pollfds[i].events |= POLLIN;
         if(streamInfos[i]->wantWrite)
            pollfds[i].events |= POLLOUT;

         ++iter;
         ++i;
      }
   }

   int result = poll(pollfds, count, timeout);

   if(result > 0)
   {
      for(int i = 0; i < count; i++)
      {
         if(streamInfos[i]->wantRead
         && (pollfds[i].revents & (POLLIN | POLLERR | POLLHUP | POLLNVAL)))
         {
            streamInfos[i]->readMessages->send();
            streamInfos[i]->readMessages->clear();
            streamInfos[i]->wantRead = false;
         }

         if(streamInfos[i]->wantWrite
         && (pollfds[i].revents & (POLLOUT | POLLERR | POLLHUP | POLLNVAL)))
         {
            streamInfos[i]->writeMessages->send();
            streamInfos[i]->writeMessages->clear();
            streamInfos[i]->wantWrite = false;
         }

         if(!streamInfos[i]->wantRead && !streamInfos[i]->wantWrite)
            mEventManager->mCurrentStreams.erase(streamInfos[i]->position);
      }

      return false;
   }
   else if(result == 0)
      return true;
   else
      FW_THROW(spnew ErrnoError($("poll")));
}

void SystemImpl::mainLoop()
{
   while(true)
   {
      int dispatchCount = pumpMessageQueue(512);

      double time = mClock->getRealTime();

      //get rid of any disabled timer events
      while(!mClock->mTimerEvents.empty()
          && mClock->mTimerEvents.top().message->isDisabled())
         mClock->mTimerEvents.pop();

      //wait for OS events
      int timeout;
      if(dispatchCount == 512)
      {
         //The message queue may not be empty...
         timeout = 0;
      }
      else if(mClock->mTimerEvents.empty())
         timeout = -1;
      else if(time < mClock->mTimerEvents.top().time)
         timeout = (int)((mClock->mTimerEvents.top().time - time) * 1000.0);
      else
         timeout = 0;

      if(!wait(timeout))
         time = mClock->getRealTime();  //wait returned before timeout
      else if(timeout > 0)
         time = mClock->mTimerEvents.top().time;  //wait timed out at clock event

      if(!mClock->mTimerEvents.empty()
      && time >= mClock->mTimerEvents.top().time)
      {
         //send timer events

         //back up timer to the time the event was supposed to be sent
         time = mClock->mTimerEvents.top().time;

         do
         {
            mClock->mTimerEvents.top().message->send();
            mClock->mTimerEvents.pop();
         }
         while(!mClock->mTimerEvents.empty()
            && time == mClock->mTimerEvents.top().time);
      }

      mClock->mCurrentTime = time;
   }
}

// ClockImpl =========================================================
SystemImpl::ClockImpl::ClockImpl()
{
   struct timeval startTime;
   gettimeofday(&startTime, null);

   startTime.tv_sec -= ORIGIN_SECOND;
   mCurrentTime = (double)startTime.tv_sec + (double)startTime.tv_usec / 1000000.0;
}

SystemImpl::ClockImpl::~ClockImpl()
{
}

// Clock -------------------------------------------------------------
double SystemImpl::ClockImpl::getTime()
{
   return mCurrentTime;
}

double SystemImpl::ClockImpl::getRealTime()
{
   struct timeval newTime;
   gettimeofday(&newTime, null);
   newTime.tv_sec -= ORIGIN_SECOND;
   return (double)newTime.tv_sec + (double)newTime.tv_usec / 1000000.0;
}

void SystemImpl::ClockImpl::notifyAtTime(double time, messages::MessagePtr message)
{
   if(time < mCurrentTime)
      message->send();
   else
   {
      TimerEvent event;
      event.time = time;
      event.message = message;
      mTimerEvents.push(event);
   }
}

// EventManagerImpl ==================================================
SystemImpl::EventManagerImpl::EventManagerImpl()
{
}
SystemImpl::EventManagerImpl::~EventManagerImpl()
{
}

// EventManager ------------------------------------------------------
messages::EventManager::StreamWatcherPtr
SystemImpl::EventManagerImpl::makeStreamWatcher(int socket)
{
   return spnew StreamWatcherImpl(socket, spthis);
}

messages::EventManager::AioStreamPtr
SystemImpl::EventManagerImpl::makeAioStream(int fileDiscriptor)
{
   return spnew AioStreamImpl(fileDiscriptor, spthis);
}

// StreamWatcher =====================================================
SystemImpl::StreamWatcherImpl::StreamWatcherImpl(int socket, EventManagerImplPtr eventManager)
   : mEventManager(eventManager)
{
   u_long arg = 1;
   ioctl(socket, FIONBIO, &arg);

   mInfo.fd = socket;
   mInfo.wantRead = false;
   mInfo.wantWrite = false;
   mInfo.readMessages = messages::makeMessageListImpl();
   mInfo.writeMessages = messages::makeMessageListImpl();
}
SystemImpl::StreamWatcherImpl::~StreamWatcherImpl()
{
   if(mInfo.wantRead || mInfo.wantWrite)
      mEventManager->mCurrentStreams.erase(mInfo.position);
}

// StreamWatcher -----------------------------------------------------
void SystemImpl::StreamWatcherImpl::notifyWhenReadable(messages::MessagePtr message)
{
   if(!mInfo.wantRead && !mInfo.wantWrite)
   {
      mInfo.position = mEventManager->mCurrentStreams.insert(
         mEventManager->mCurrentStreams.end(), &mInfo);
   }

   mInfo.wantRead = true;
   mInfo.readMessages->add(message);
}
void SystemImpl::StreamWatcherImpl::notifyWhenWritable(messages::MessagePtr message)
{
   if(!mInfo.wantRead && !mInfo.wantWrite)
   {
      mInfo.position = mEventManager->mCurrentStreams.insert(
         mEventManager->mCurrentStreams.end(), &mInfo);
   }

   mInfo.wantWrite = true;
   mInfo.writeMessages->add(message);
}

// AioStream =========================================================
SystemImpl::AioStreamImpl::AioStreamImpl(int fileDescriptor, EventManagerImplPtr eventManager)
   : mFileDescriptor(fileDescriptor), mResult(0), mResultIsError(false)
{
}
SystemImpl::AioStreamImpl::~AioStreamImpl()
{
}

// AioStream ---------------------------------------------------------
bool SystemImpl::AioStreamImpl::read(void* buffer, uint64 offset, int size, messages::MessagePtr message)
{
   lseek(mFileDescriptor, offset, SEEK_SET);
   mResult = ::read(mFileDescriptor, buffer, size);

   if(mResult >= 0)
   {
      mResultIsError = false;
      message->send();
   }
   else
   {
      mResultIsError = true;
      mResult = errno;
   }

   return !mResultIsError;
}
bool SystemImpl::AioStreamImpl::write(void* buffer, uint64 offset, int size, messages::MessagePtr message)
{
   lseek(mFileDescriptor, offset, SEEK_SET);
   mResult = ::write(mFileDescriptor, buffer, size);

   if(mResult >= 0)
   {
      mResultIsError = false;
      message->send();
   }
   else
   {
      mResultIsError = true;
      mResult = errno;
   }

   return !mResultIsError;
}
bool SystemImpl::AioStreamImpl::getResult(int* result)
{
   if(result != null)
      *result = mResult;
   return !mResultIsError;
}

// factory function ==================================================
void startSystem(FrontEndPtr frontEnd,
                 FrontEnd::StartupInfo startupInfo,
                 String programName, String executablePath)
{
   SmartPointer<SystemImpl> system = spnew SystemImpl(programName, executablePath);
   startupInfo.system = system;
   frontEnd->run(startupInfo);
   system->mainLoop();
}

// ===================================================================

} //namespace posix
} //namespace internal
} //namespace framework
