/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/04/10 (KV)
 *  + First Version
 */

// ===================================================================

#include "SystemImpl.h"
#include <framework/messages/EventManager.h>
#include <framework/messages/MessageListImpl.h>
#include <framework/internal/MessageImpl.h>
#include <queue>
#include <map>

#include <windows.h>
#include <winsock.h>
#include <shlobj.h>
#undef ERROR

namespace framework {
namespace internal {
namespace win32 {

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Implement framework::internal::win32::SystemImpl
// TODO (KV): Review framework::internal::win32::SystemImpl

implementation SystemImpl: implements System
{
   SystemImpl(String programName, String executablePath);
   ~SystemImpl();

   // System ---------------------------------------------------------
   ClockPtr getClock();

   String getCurrentDirectory();
   String getSystemConfigDirectory();
   String getUserConfigDirectory();
   String getDataFileDirectory();

   void createDirectory(String path);
   String createDirectory(String path, String subPath);

   messages::EventManagerPtr getEventManager();

   // internal -------------------------------------------------------
//   bool haveStuffToWaitFor();
   bool wait(DWORD milliseconds);
   void mainLoop();

   implementation ClockImpl: implements Clock
   {
      ClockImpl();
      ~ClockImpl();

      // Clock -------------------------------------------------------
      double getTime();
      double getRealTime();

      void notifyAtTime(double time, messages::MessagePtr message);

      // internal ----------------------------------------------------
      struct TimerEvent
      {
         double time;
         messages::MessagePtr message;

         inline bool operator<(const TimerEvent& other) const
            {return time > other.time;}
      };

      double mCurrentTime;
      static const int64 ORIGIN_FILETIME = 125911584000000000ll;
      std::priority_queue<TimerEvent> mTimerEvents;
   };
   typedef SmartPointer<ClockImpl> ClockImplPtr;

   implementation EventManagerImpl: implements messages::EventManager
   {
      EventManagerImpl();
      ~EventManagerImpl();

      // EventManager ------------------------------------------------
      StreamWatcherPtr makeStreamWatcher(int socket);
      HandleWatcherPtr makeHandleWatcher(HANDLE handle);
      ObjectPtr getExternalWaiter();

      // internal ----------------------------------------------------
      struct HandleInfo
      {
         messages::MessageListPtr oneTimeMessages;
         messages::MessageListPtr persistentMessages;
      };

      struct SocketInfo
      {
         messages::MessageListPtr readableMessages;
         messages::MessageListPtr writableMessages;
         int waitMask;

         enum WaitFlags
         {
            READ  = 1 << 0,
            WRITE = 1 << 1
         };
      };

      std::map<HANDLE, HandleInfo*> mWatchedHandles;
      std::map<SOCKET, SocketInfo*> mWatchedSockets;
//      util::WeakLink<Object> mExternalWaiter;
//      int mSocketWaitCount;

      HWND mSocketMessageWindow;
   };
   typedef SmartPointer<EventManagerImpl> EventManagerImplPtr;

   implementation StreamWatcherImpl: implements messages::EventManager::StreamWatcher
   {
      StreamWatcherImpl(SOCKET socket, EventManagerImplPtr eventManager);
      ~StreamWatcherImpl();

      // StreamWatcher -----------------------------------------------
      void notifyWhenReadable(messages::MessagePtr message);
      void notifyWhenWritable(messages::MessagePtr message);

      // internal ----------------------------------------------------
      SOCKET mSocket;
      EventManagerImpl::SocketInfo mInfo;
      EventManagerImplPtr mEventManager;

      static LRESULT CALLBACK windowProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam);
   };

   implementation HandleWatcherImpl: implements messages::EventManager::HandleWatcher
   {
      HandleWatcherImpl(HANDLE handle, EventManagerImplPtr eventManager);
      ~HandleWatcherImpl();

      // HandleWatcher -----------------------------------------------
      void notifyOnAllEvents(messages::MessagePtr message);
      void notifyOnNextEvent(messages::MessagePtr message);

      // internal ----------------------------------------------------
      HANDLE mHandle;
      EventManagerImpl::HandleInfo mInfo;
      EventManagerImplPtr mEventManager;
   };


   ClockImplPtr mClock;
   EventManagerImplPtr mEventManager;

   String mCurrentDirectory;
   String mPrefix;
   bool mUnixDirectoryStyle;

   String mProgramName;
};

SystemImpl::SystemImpl(String programName, String executablePath)
{
   mClock = spnew ClockImpl;
   mEventManager = spnew EventManagerImpl;

   char path[MAX_PATH];
   int length = GetCurrentDirectory(MAX_PATH, path);
   if(length == 0)
      FW_THROW(spnew Win32Error($("GetCurrentDirectory")));
   if(length > MAX_PATH)
      FW_THROW(spnew FatalError($("Current directory path too big.")));
   mCurrentDirectory = String(path, length);

   executablePath = executablePath.getInterval(0, executablePath.findLastOf('\\'));
   if(executablePath.getLength() > 4
   && executablePath.getInterval(executablePath.getLength() - 4).compareCaseInsensitive($("\\bin")) == 0)
   {
      mPrefix = executablePath.getInterval(0, executablePath.getLength() - 4);
      mUnixDirectoryStyle = true;
   }
   else
   {
      mPrefix = executablePath;
      mUnixDirectoryStyle = false;
   }

   mProgramName = programName;
}

SystemImpl::~SystemImpl()
{
}

// System ------------------------------------------------------------
ClockPtr SystemImpl::getClock()
{
   return mClock;
}

String SystemImpl::getCurrentDirectory()
{
   return mCurrentDirectory;
}
String SystemImpl::getSystemConfigDirectory()
{
   if(mUnixDirectoryStyle)
   {
      String dir = mPrefix + $("\\etc\\") + mProgramName;
      createDirectory(dir);
      return dir;
   }
   else
      return mPrefix;
}
String SystemImpl::getUserConfigDirectory()
{
   char path[MAX_PATH];
   if(SUCCEEDED(SHGetFolderPath(NULL, CSIDL_APPDATA, NULL, 0, path)))
   {
      String dir = String(path) + $("\\") + mProgramName;
      createDirectory(dir);
      return dir;
   }
   else
      FW_THROW(spnew FatalError($("SHGetFolderPath failed to find application data folder.")));
}

String SystemImpl::getDataFileDirectory()
{
   if(mUnixDirectoryStyle)
   {
      String dir = mPrefix + $("\\share\\") + mProgramName;
      createDirectory(dir);
      return dir;
   }
   else
      return mPrefix;
}

void SystemImpl::createDirectory(String dir)
{
   BOOL result;

   {
      FW_MAKE_CSTRING(cdir, dir);
      result = CreateDirectory(cdir, NULL);
   }

   if(result)
      return;

   DWORD error = GetLastError();

   if(error == ERROR_ALREADY_EXISTS)
      return;
   else if(error == ERROR_PATH_NOT_FOUND)
   {
      int pos = dir.findLastOf('\\');
      if(pos == -1)
         FW_THROW(spnew Win32Error($("CreateDirectory"), error));

      createDirectory(dir.getInterval(0, pos));
   }
   else
      FW_THROW(spnew Win32Error($("CreateDirectory"), error));

   FW_MAKE_CSTRING(cdir, dir);
   if(!CreateDirectory(cdir, NULL))
      FW_THROW(spnew Win32Error($("CreateDirectory")));
}

String SystemImpl::createDirectory(String path, String subPath)
{
   Array<char> buffer(path.getLength() + subPath.getLength() + 1);
   path.copyToBuffer(&buffer[0], false);
   buffer[path.getLength()] = '\\';
   for(int i = 0; i < subPath.getLength(); i++)
   {
      if(subPath[i] == '/')
         buffer[path.getLength() + i + 1] = '\\';
      else
         buffer[path.getLength() + i + 1] = subPath[i];
   }

   String result(buffer.toConst());
   createDirectory(result);
   return result;
}

messages::EventManagerPtr SystemImpl::getEventManager()
{
   return mEventManager;
}

// internal ----------------------------------------------------------
//bool SystemImpl::haveStuffToWaitFor()
//{
//   return !mEventManager->mWatchedHandles.empty()
//        || mEventManager->mSocketWaitCount > 0
//        || (ObjectPtr)mEventManager->mExternalWaiter != null;
//}

bool SystemImpl::wait(DWORD milliseconds)
{
   DWORD count = mEventManager->mWatchedHandles.size();
   HANDLE handles[count];

   std::map<HANDLE, EventManagerImpl::HandleInfo*>::iterator iter =
      mEventManager->mWatchedHandles.begin();
   for(int i = 0; iter != mEventManager->mWatchedHandles.end(); ++iter, ++i)
      handles[i] = iter->first;

   DWORD result = MsgWaitForMultipleObjectsEx(count, handles, milliseconds,
                                              QS_ALLEVENTS, MWMO_ALERTABLE);

   if(result >= WAIT_OBJECT_0 && result < WAIT_OBJECT_0 + count)
   {
      iter = mEventManager->mWatchedHandles.find(handles[result - WAIT_OBJECT_0]);
      iter->second->oneTimeMessages->send();
      iter->second->oneTimeMessages->clear();
      if(iter->second->persistentMessages->send() == 0)
         mEventManager->mWatchedHandles.erase(iter);
      return false;
   }
   else if(result == WAIT_OBJECT_0 + count)
   {
      //window messages on queue
      MSG message;
      while(PeekMessage(&message, NULL, 0, 0, PM_REMOVE))
      {
         if(message.message != WM_QUIT)
         {
            TranslateMessage(&message);
            DispatchMessage(&message);
         }
      }
      return false;
   }
   else if(result == WAIT_IO_COMPLETION)
   {
      //APC's called.  No need to do anything.
      return false;
   }
   else if(result == WAIT_TIMEOUT)
      return true;
   else if(result >= WAIT_ABANDONED_0 && result < WAIT_ABANDONED_0 + count)
   {
      //why does this exist?
      iter = mEventManager->mWatchedHandles.find(handles[result - WAIT_ABANDONED_0]);
      iter->second->oneTimeMessages->send();
      iter->second->oneTimeMessages->clear();
      if(iter->second->persistentMessages->send() == 0)
         mEventManager->mWatchedHandles.erase(iter);
      return false;
   }
   else
      FW_THROW(spnew Win32Error($("MsgWaitForMultipleObjectsEx")));
}

void SystemImpl::mainLoop()
{
   while(true)
   {
      int dispatchCount = pumpMessageQueue(512);

      double time = mClock->getRealTime();

      //get rid of any disabled timer events
      while(!mClock->mTimerEvents.empty()
          && mClock->mTimerEvents.top().message->isDisabled())
         mClock->mTimerEvents.pop();

      //wait for OS events
      DWORD waitTime;
      if(dispatchCount == 512)
         waitTime = 0;
      else if(mClock->mTimerEvents.empty())
      {
         waitTime = INFINITE;
//         if(!haveStuffToWaitFor())
//            return;
      }
      else if(time < mClock->mTimerEvents.top().time)
         waitTime = (DWORD)((mClock->mTimerEvents.top().time - time) * 1000.0);
      else
         waitTime = 0;

      if(!wait(waitTime))
         time = mClock->getRealTime();
      else if(waitTime > 0)
         time = mClock->mTimerEvents.top().time;

      if(!mClock->mTimerEvents.empty()
      && time >= mClock->mTimerEvents.top().time)
      {
         //send timer events

         //back up timer to the time the event was supposed to be sent
         time = mClock->mTimerEvents.top().time;

         do
         {
            mClock->mTimerEvents.top().message->send();
            mClock->mTimerEvents.pop();
         }
         while(!mClock->mTimerEvents.empty()
            && time == mClock->mTimerEvents.top().time);
      }

      mClock->mCurrentTime = time;
   }
}

// ClockImpl =========================================================
SystemImpl::ClockImpl::ClockImpl()
{
   int64 startTime;

   GetSystemTimeAsFileTime((FILETIME*)&startTime);
   mCurrentTime = (double)(startTime - ORIGIN_FILETIME) / 10000000.0;
}

SystemImpl::ClockImpl::~ClockImpl()
{
}

// Clock -------------------------------------------------------------
double SystemImpl::ClockImpl::getTime()
{
   return mCurrentTime;
}

double SystemImpl::ClockImpl::getRealTime()
{
   int64 newTime;
   GetSystemTimeAsFileTime((FILETIME*)&newTime);
   return (double)(newTime - ORIGIN_FILETIME) / 10000000.0;
}

void SystemImpl::ClockImpl::notifyAtTime(double time, messages::MessagePtr message)
{
   if(time < mCurrentTime)
      message->send();
   else
   {
      TimerEvent event;
      event.time = time;
      event.message = message;
      mTimerEvents.push(event);
   }
}

// EventManagerImpl ==================================================
SystemImpl::EventManagerImpl::EventManagerImpl()
//   : mSocketWaitCount(0)
{
   WNDCLASS messageOnlyWindowClass =
      {0, StreamWatcherImpl::windowProc, 0, 0, GetModuleHandle(NULL), NULL,
       NULL, NULL, NULL, "FrameworkSocketMessageWindow"};
   if(RegisterClass(&messageOnlyWindowClass) == 0)
      FW_THROW(spnew Win32Error($("RegisterClass")));

   mSocketMessageWindow = CreateWindow("FrameworkSocketMessageWindow", "",
                                       0, 0, 0, 0, 0,
                                       HWND_MESSAGE, NULL, NULL, this);
   if(mSocketMessageWindow == NULL)
      FW_THROW(spnew Win32Error($("CreateWindow")));
}
SystemImpl::EventManagerImpl::~EventManagerImpl()
{
}

// EventManager ------------------------------------------------------
messages::EventManager::StreamWatcherPtr
SystemImpl::EventManagerImpl::makeStreamWatcher(int socket)
{
   return spnew StreamWatcherImpl(socket, spthis);
}

messages::EventManager::HandleWatcherPtr
SystemImpl::EventManagerImpl::makeHandleWatcher(HANDLE handle)
{
   return spnew HandleWatcherImpl(handle, spthis);
}
ObjectPtr SystemImpl::EventManagerImpl::getExternalWaiter()
{
//   ObjectPtr result = mExternalWaiter;
//   if(result == null)
//   {
//      result = spnew Object;
//      mExternalWaiter = result;
//   }
//   return result;
   return null;
}

// StreamWatcher =====================================================
SystemImpl::StreamWatcherImpl::StreamWatcherImpl(SOCKET socket, EventManagerImplPtr eventManager)
   : mSocket(socket), mEventManager(eventManager)
{
   WSAAsyncSelect(socket, mEventManager->mSocketMessageWindow,
                  WM_USER, FD_CONNECT | FD_ACCEPT | FD_READ | FD_WRITE | FD_CLOSE);
   mInfo.readableMessages = messages::makeMessageListImpl();
   mInfo.writableMessages = messages::makeMessageListImpl();
   mInfo.waitMask = 0;

   mEventManager->mWatchedSockets[mSocket] = &mInfo;
}
SystemImpl::StreamWatcherImpl::~StreamWatcherImpl()
{
//   if(mInfo.waitMask != 0)
//      --mEventManager->mSocketWaitCount;
   mEventManager->mWatchedSockets.erase(mSocket);
   WSAAsyncSelect(mSocket, mEventManager->mSocketMessageWindow, 0, 0);
}

// StreamWatcher -----------------------------------------------------
void SystemImpl::StreamWatcherImpl::notifyWhenReadable(messages::MessagePtr message)
{
   mInfo.readableMessages->add(message);
//   if(mInfo.waitMask == 0)
//      ++mEventManager->mSocketWaitCount;
   mInfo.waitMask |= EventManagerImpl::SocketInfo::READ;
}
void SystemImpl::StreamWatcherImpl::notifyWhenWritable(messages::MessagePtr message)
{
   mInfo.writableMessages->add(message);
//   if(mInfo.waitMask == 0)
//      ++mEventManager->mSocketWaitCount;
   mInfo.waitMask |= EventManagerImpl::SocketInfo::WRITE;
}

// internal ----------------------------------------------------------
LRESULT CALLBACK SystemImpl::StreamWatcherImpl::windowProc(HWND window, UINT message, WPARAM wparam, LPARAM lparam)
{
   SystemImpl::EventManagerImpl* eventManager =
      (SystemImpl::EventManagerImpl*)GetWindowLongPtr(window, GWLP_USERDATA);
   if(eventManager != null)
   {
      switch(message)
      {
      case WM_USER:
         {
//            int error = WSAGETSELECTERROR(lparam);

            std::map<SOCKET, EventManagerImpl::SocketInfo*>::iterator iter =
               eventManager->mWatchedSockets.find((SOCKET)wparam);

            if(iter == eventManager->mWatchedSockets.end())
               return 0;

#if 0
            switch(WSAGETSELECTEVENT(lparam))
            {
            case FD_READ:
               Debug::log() << "Got FD_READ";
               break;
            case FD_ACCEPT:
               Debug::log() << "Got FD_ACCEPT";
               break;
            case FD_CLOSE:
               Debug::log() << "Got FD_CLOSE";
               break;
            case FD_WRITE:
               Debug::log() << "Got FD_WRITE";
               break;
            case FD_CONNECT:
               Debug::log() << "Got FD_CONNECT";
               break;
            default:
               return 0;
            }
#endif

            switch(WSAGETSELECTEVENT(lparam))
            {
            case FD_READ:
            case FD_ACCEPT:
            case FD_CLOSE:
               if(!(iter->second->waitMask & EventManagerImpl::SocketInfo::READ))
                  return 0;
               iter->second->readableMessages->send();
               iter->second->readableMessages->clear();
               iter->second->waitMask &= ~EventManagerImpl::SocketInfo::READ;
               break;
            case FD_WRITE:
            case FD_CONNECT:
               if(!(iter->second->waitMask & EventManagerImpl::SocketInfo::WRITE))
                  return 0;
               iter->second->writableMessages->send();
               iter->second->writableMessages->clear();
               iter->second->waitMask &= ~EventManagerImpl::SocketInfo::WRITE;
               break;
            default:
               return 0;
            }

//            if(iter->second->waitMask == 0)
//               --eventManager->mSocketWaitCount;
         }
         return 0;
      default:
         return DefWindowProc(window, message, wparam, lparam);
      }
   }
   else if(message == WM_CREATE)
   {
      SetWindowLongPtr(window, GWLP_USERDATA, (LONG_PTR)((CREATESTRUCT*)lparam)->lpCreateParams);
      return DefWindowProc(window, message, wparam, lparam);
   }
   else
      return DefWindowProc(window, message, wparam, lparam);
}

// HandleWatcher =====================================================
SystemImpl::HandleWatcherImpl::HandleWatcherImpl(HANDLE handle, EventManagerImplPtr eventManager)
   : mHandle(handle), mEventManager(eventManager)
{
   mInfo.oneTimeMessages = messages::makeMessageListImpl();
   mInfo.persistentMessages = messages::makeMessageListImpl();
}
SystemImpl::HandleWatcherImpl::~HandleWatcherImpl()
{
   mEventManager->mWatchedHandles.erase(mHandle);
}

// HandleWatcher -----------------------------------------------------
void SystemImpl::HandleWatcherImpl::notifyOnAllEvents(messages::MessagePtr message)
{
   mInfo.persistentMessages->add(message);
   mEventManager->mWatchedHandles[mHandle] = &mInfo;
}

void SystemImpl::HandleWatcherImpl::notifyOnNextEvent(messages::MessagePtr message)
{
   mInfo.oneTimeMessages->add(message);
   mEventManager->mWatchedHandles[mHandle] = &mInfo;
}

// factory function ==================================================
void startSystem(FrontEndPtr frontEnd,
                 FrontEnd::StartupInfo startupInfo,
                 String programName, String executablePath)
{
   SmartPointer<SystemImpl> system = spnew SystemImpl(programName, executablePath);
   startupInfo.system = system;
   frontEnd->run(startupInfo);
   system->mainLoop();
}

// ===================================================================

} //namespace win32
} //namespace internal
} //namespace framework
