/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2004/04/10 (KV)
 *  + First Version
 */

// ===================================================================

#ifndef framework_messages_EventManager_h__INCLUDED__
#define framework_messages_EventManager_h__INCLUDED__

#include "Message.h"

namespace framework {
namespace messages {
using namespace ::framework::common;

// ===================================================================
// AUTHOR (KV): Created module.
// TODO (KV): Document framework::messages::EventManager
// TODO (KV): Test framework::messages::EventManager

interface EventManager;
typedef SmartPointer<EventManager> EventManagerPtr;

interface EventManager: implements Object
{
   virtual ~EventManager();

   /* On Windows, stream watchers should only be used on sockets.  On
    * most other OS's, they can be used on any type of file
    * descriptor.
    *
    * After a message is sent by the StreamWatcher, it will not be
    * sent again unless you call the associated notify method again to
    * re-register the message.  On the other hand, you must not call
    * any of the notify methods if the associated condition is already
    * true.  These two restrictions are necessary because some OS's
    * use edge-triggered events and some use level-triggered.
    *
    * Note that false positives will sometimes occur.  This can be the
    * result of various race conditions, as well as due to silly
    * inconsistencies in the way different OS's report events.  In
    * any case, it shouldn't be a problem if you write clean code.
    *
    * Please insure that the descructor for this object is called
    * before the associated file descriptor is closed. */
   interface StreamWatcher: implements Object
   {
      virtual ~StreamWatcher();

      //Readable includes ability to accept(), as well as EOF/disconnect.
      virtual void notifyWhenReadable(MessagePtr message) = 0;
      //Writable includes completion/failure of connect().
      virtual void notifyWhenWritable(MessagePtr message) = 0;
   };
   typedef SmartPointer<StreamWatcher> StreamWatcherPtr;

   virtual StreamWatcherPtr makeStreamWatcher(int socket) = 0;

#ifdef PLATFORM_WIN32
   interface HandleWatcher: implements Object
   {
      virtual ~HandleWatcher();

      virtual void notifyOnAllEvents(MessagePtr message) = 0;
      virtual void notifyOnNextEvent(MessagePtr message) = 0;
   };
   typedef SmartPointer<HandleWatcher> HandleWatcherPtr;

   virtual HandleWatcherPtr makeHandleWatcher(void* handle) = 0;

   /* When using APC's or WindowProc windows to wait for events, make
    * sure to hold a reference to this object.  This lets the system
    * know that you are still expecting messages, so that it will not
    * decide to exit. */
   virtual ObjectPtr getExternalWaiter() = 0;
#endif

#ifdef PLATFORM_POSIX
   /* Happy interface around POSIX AIO!  Should be pretty self-
    * explanatory.  Note, however, that a single AioStream must only
    * have one outstanding AIO request on it at a time (read or
    * write).  If you want to perform multiple operations, you must
    * create multiple AioStreams with the same file descriptor. */
   interface AioStream: implements Object
   {
      virtual ~AioStream();

      /* Returns true if the operation was successfully queued.
       * Otherwise, returns false, and errno contains the error. */
      virtual bool read(void* buffer, uint64 offset, int size, MessagePtr message) = 0;
      virtual bool write(void* buffer, uint64 offset, int size, MessagePtr message) = 0;

      /* Do not call getResult() until you receive a message
       * indicating that the operation completed.  Then, only call it
       * once.  If getResult() returns true, the operation completed,
       * and the transfer size is stored in the value pointed to by
       * "result".  If it returns false, the operation failed, and
       * the error code is stored to "result". */
      virtual bool getResult(int* result) = 0;
   };
   typedef SmartPointer<AioStream> AioStreamPtr;

   virtual AioStreamPtr makeAioStream(int fileDiscriptor) = 0;

#endif

#if defined(PLATFORM_FREEBSD) || defined(PLATFORM_DARWIN)
   //All bow before kqueue()!

   /* This interface is suitable for kevents which do not
    * automatically set EV_CLEAR.  That is, it works well for
    * level-triggered events.  It does not work well with auto-clear
    * events like EVFILT_SIGNAL or EVFILT_TIMER as there is a
    * possibility that an event will be missed if it occurs before the
    * previous event was handled.  There should really be a separate
    * interface for dealing with those filters, but as I am not
    * actually interested in using them at this time, I'm not going to
    * bother writing one yet.  :P
    *
    * Please insure that the descructor for this object is called
    * before the associated ident is closed. */
   interface KqueueWatcher: implements Object
   {
      virtual ~KqueueWatcher();

      /* Gets the fflags and data returned with the kevent last time
       * it fired.  You should call this after receiving a message
       * notifying you of this event, but before calling
       * notifyOnEvent() again and before any calls to setInfo() or
       * clear().  Basically you should call this first thing on
       * receiving notification. */
      virtual void getInfo(uint* fflags, word* data) = 0;

      /* Update the fflags and data members of the kevent.  The next
       * time kevent() is called, EV_ADD will be used to re-add the
       * event with the new flags. */
      virtual void setInfo(uint fflags, word data) = 0;

      //Causes EV_CLEAR to be applied to the kevent.
      virtual void clear() = 0;

      /* Causes the given message to be fired the next time the kevent
       * is returned.  The message will be fired only once.  Normally
       * you would call notifyOnEvent() again after handling the event
       * in order to receive notification of the next event.  If
       * kevent() happens to be called again before you have done
       * this, then this kevent will be disabled with EV_DISABLE, and
       * enabled again after you get around to calling
       * notifyOnEvent(). */
      virtual void notifyOnEvent(MessagePtr message) = 0;
   };
   typedef SmartPointer<KqueueWatcher> KqueueWatcherPtr;

   /* Create a new watcher.  The next time kevent() is called, this
    * ident/filter combination will be added to the kqueue with
    * EV_ADD. */
   virtual KqueueWatcherPtr makeKqueueWatcher(
      word ident, short filter, uint fflags, word data) = 0;

#endif
};

// ===================================================================

} //namespace messages
} //namespace framework

#endif //included
