/* Kenton Varda's C++ framework code.
 * Copyright (C) 2002-2005 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/* Revision History:
 * 2003/05/21 (KV)
 *  + First Version
 */

// ===================================================================

#ifndef framework_messages_Message_h__INCLUDED__
#define framework_messages_Message_h__INCLUDED__

#include "MessageReceiver.h"

namespace framework {
namespace messages {
using namespace ::framework::common;

// ===================================================================
// AUTHOR (KV): Created module.

interface Message;
typedef SmartPointer<Message> MessagePtr;

/* interface Message
 *
 * Represents a message which can be sent to a MessageReceiver.  Each
 * Message is owned by a particular MessageReceiver, and will be sent
 * to that receiver when send() is called.  The Message must go
 * through a queue first, though.  See package.kake for more info on
 * the message-passing system.
 *
 * The only implementation of Message is MessageImpl.  Because it
 * relies on various hooks in the framework, no other implementation
 * is possible. */
interface Message: implements Object
{
   virtual ~Message();

   /* Place the message in the message queue.  Once it reaches the top
    * of the queue, the message's receiver's receive() method is
    * called.
    *
    * A Message does not store a reference to its receiver.  So, it is
    * possible that the receiver will no longer exist when the Message
    * is sent.  In this case, send() returns false.  It is also
    * possible that the receiver will be destroyed while the message
    * is still in the queue.  No warning is given in this case.  So,
    * you must design your interfaces such that nothing particularily
    * bad will happen if a message is sent but not received. */
   virtual bool send() = 0;

   /* Disables the message.  After this has been called, all further
    * calls to send() will do nothing but return false.  Typically
    * this is only called by the owner of the message, to prevent it
    * from being sent again. */
   virtual void disable() = 0;

   /* Returns true if the message has been disabled, or if its
    * receiver no longer exists. */
   virtual bool isDisabled() = 0;

   /* Gets an integer which was assigned to this message by its owner
    * on creation.  This allows the owner to quickly identify the
    * message via a switch statement. */
   virtual int getIdentifier() = 0;

   /* Gets some data which was assigned to this message by its owner
    * on creation.  This data should *only* be used by the owner.  If
    * the message is disabled or the owner is found to no longer
    * exist, the data will be released (and getData() will
    * subsequently return null). */
   virtual ObjectPtr getData() = 0;
};

// factory function ==================================================
/* Creates an implementation of Message with the given receiver.  The
 * id and data arguments will be returned by the Message's
 * getIdentifier() and getData() methods, respectively.  Note that
 * this function should only be called by the receiver itself.  So,
 * the first argument should always be "this". */
MessagePtr makeMessage(MessageReceiver* receiver, int id, ObjectPtr data);

// ===================================================================

} //namespace messages
} //namespace framework

#endif //included
