/* kake - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Binary.h"
#include "Compiler.h"
#include "Filesystem.h"
#include <iostream>
#include <fstream>

namespace kake {
using namespace std;

// class member functions ============================================
Binary::Binary(string name, Type type, string sourcefile, int linenum)
   : mName(name),
     mType(type),
     mMainObject(NULL),
     mSourceFile(sourcefile),
     mLineNum(linenum),
     mBuilt(false),
     mCleaned(false),
     mChecked(false),
     mDocumented(false)
{
}

Binary::~Binary()
{
}

Buildable::BuildResult Binary::Build()
{
   if(mBuilt)
      return mBuildResult;
   mBuilt = true;

   //TODO: build required libs

   //setup cflags and ldflags
   Compiler::ClearLibraries();
   Compiler::ClearParameters();
   Compiler::ClearLinkerParameters();
   Compiler::ClearDefines();

   Compiler::AddParameters(mCFlags);
   Compiler::AddLinkerParameters(mLdFlags);
   Compiler::AddDefines(mDefines);

   mBuildResult = SKIPPED; //assume skipped until we decide otherwise

   //build all parts.
   BuildResult childResult;
   for(unsigned int i = 0; i < mObjects.size(); i++)
   {
      childResult = mObjects[i]->Build();

      switch(childResult)
      {
      case FAILED:
         return mBuildResult = FAILED;
      case UPDATED:
         mBuildResult = UPDATED;
         break;
      case SKIPPED:
         break;
      }
   }

   string outputFile;
   vector<string > objects;
   GetObjects(objects);

   switch(mType)
   {
   case EXECUTABLE:
      outputFile = Compiler::GetExecutableFileForName(mName);
      break;
   case LIBRARY:
      outputFile = Compiler::GetLibraryFileForName(mName);
      break;
   case STATIC_LIBRARY:
      outputFile = Compiler::GetStaticLibraryFileForName(mName);
      break;
   }

   if(mBuildResult == SKIPPED && Filesystem::Exists(outputFile))
   {
      //double-check whether we need to re-link or not
      bool needUpdate = false;
      unsigned int timestamp = Filesystem::GetTimestamp(outputFile);

      //check to see if any of the objects are newer than the output file.
      for(unsigned int i = 0; i < objects.size(); i++)
      {
         if(Filesystem::GetTimestamp(objects[i]) > timestamp)
         {
            //object file newer than linked output
            needUpdate = true;
            break;
         }
      }

      if(!needUpdate)
         return mBuildResult = SKIPPED;
   }

   string type;
   switch(mType)
   {
   case EXECUTABLE:
      type = "executable ";
      break;
   case LIBRARY:
      type = "library ";
      break;
   case STATIC_LIBRARY:
      type = "static library ";
      break;
   }

   cout << "Linking " << type << mName << "...";
   if(Filesystem::IsStderrRedirected())
   {
      cerr << "Linking " << type << mName << "..." << endl;
      cout.flush();
   }
   else
      cout << endl;

   bool success = false;

   switch(mType)
   {
   case EXECUTABLE:
      {
         string objectName = mMainObject->GetName();
         string objectNameWithUnderscores = mMainObject->GetNameWithUnderscores();
         string objectFilename = mMainObject->GetFilename();

         unsigned int pos = objectName.find_last_of(':');
         string classname;
         string package;
         if(pos == string::npos)
         {
            throw "Binary::Build():  This should never happen.";
         }
         else
         {
            classname = objectName.substr(pos + 1);
            package = objectName.substr(0, pos - 1);
         }

         //have to output *more* autogen source for the main function...
         string autogenFile = "build/" + mName + "__main.cpp";
         string autogenObject = "build/" + mName + "__main.o";
         fstream autogen(autogenFile.c_str(), ios::out);
         autogen << "#include <" + objectFilename + ".h>\n";
         autogen << "KAKE_WRITE_MAIN(" << package << ", " << classname << ")\n";
         autogen.close();

         if(!Compiler::Compile(autogenFile, autogenObject, false))
         {
            if(Filesystem::IsStderrRedirected())
               cout << "(errors logged)" << endl << endl;

            cerr << "Errors occurred when compiling code generated by kake to support" << endl;
            cerr << "the executable and/or the testable flag.  Please make sure that the" << endl;
            cerr << "module is properly written to be executable." << endl;
            return mBuildResult = FAILED;
         }

//         objects.push_back(autogenObject);  //already added by GetObjects()
         success = Compiler::LinkExecutable(objects, mName);
      }
      break;
   case LIBRARY:
      success = Compiler::LinkLibrary(objects, mName);
      break;
   case STATIC_LIBRARY:
      success = Compiler::LinkStaticLibrary(objects, mName);
      break;
   }

   if(Filesystem::IsStderrRedirected())
   {
      if(success)
         cout << endl;
      else
         cout << " (errors logged)" << endl;
      cerr << endl;
   }

   if(success)
      return mBuildResult = UPDATED;
   else
      return mBuildResult = FAILED;
}

bool Binary::Check()
{
   if(mChecked)
      return mCheckResult;
   mChecked = true;

   mCheckResult = true;

   if(mType == EXECUTABLE && mMainObject == NULL)
   {
      cout << mSourceFile << ":" << mLineNum << ": ";
      cout << "Executable " << mName << " has no main package or module." << endl;
      mCheckResult = false;
   }

   for(unsigned int i = 0; i < mObjects.size(); i++)
   {
      if(!mObjects[i]->Check())
         mCheckResult = false;
   }

   //TODO: check headers and libs

   return mCheckResult;
}

bool Binary::Clean()
{
   if(mCleaned)
      return mCleanResult;
   mCleaned = true;

   cout << "Cleaning ";
   switch(mType)
   {
   case EXECUTABLE:
      cout << "executable ";
      break;
   case LIBRARY:
      cout << "library ";
      break;
   case STATIC_LIBRARY:
      cout << "static library ";
      break;
   }
   cout << mName << "..." << endl;

   string outputFile;

   switch(mType)
   {
   case EXECUTABLE:
      outputFile = Compiler::GetExecutableFileForName(mName);
      break;
   case LIBRARY:
      {
         outputFile = Compiler::GetLibraryFileForName(mName);
         string dllFile = Compiler::GetDllFileForName(mName);
         if(dllFile != outputFile)
            Filesystem::Delete(dllFile);
      }
      break;
   case STATIC_LIBRARY:
      outputFile = Compiler::GetStaticLibraryFileForName(mName);
      break;
   }

   Filesystem::Delete(outputFile);

   if(mType == EXECUTABLE)
   {
      Filesystem::Delete("build/" + mName + "__main.cpp");
      Filesystem::Delete("build/" + mName + "__main.o");
   }

   for(unsigned int i = 0; i < mObjects.size(); i++)
   {
      if(!mObjects[i]->Clean())
         return mCleanResult = false;
   }

   return mCleanResult = true;
}

bool Binary::Document()
{
   if(mDocumented)
      return mDocumentResult;
   mDocumented = true;

   //TODO

   return mDocumentResult = false;
}


void Binary::GetRequiredHeaders(vector<string >& list)
{
   for(unsigned int i = 0; i < mObjects.size(); i++)
      mObjects[i]->GetRequiredHeaders(list);
}

void Binary::GetRequiredLibraries(vector<string >& list)
{
   for(unsigned int i = 0; i < mObjects.size(); i++)
      mObjects[i]->GetRequiredLibraries(list);
}

void Binary::GetObjects(vector<string >& objects)
{
   for(unsigned int i = 0; i < mObjects.size(); i++)
      mObjects[i]->GetObjects(objects);
   if(mType == EXECUTABLE)
      objects.push_back("build/" + mName + "__main.o");
}

string Binary::GetName() {return mName;}
string Binary::GetNameWithUnderscores() {return mName;}

void Binary::AddObject(Buildable* object)
{
   mObjects.push_back(object);
}

void Binary::AddCFlag(string cflag)
{
   mCFlags.push_back(cflag);
}

void Binary::AddLdFlag(string ldflag)
{
   mLdFlags.push_back(ldflag);
}

void Binary::AddDefine(string name, string value)
{
   mDefines.push_back(pair<string, string>(name, value));
}

} //namespace kake
