/* kake - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "BuildfileElement.h"
#include "Compiler.h"
#include "Filesystem.h"
#include "Module.h"
#include "Binary.h"
#include "Parser.h"
#include <iostream>

namespace kake {

// class member functions ============================================
BuildfileElement::BuildfileElement()
   : mRootPackage("", true, "", 0)
{
}

BuildfileElement::~BuildfileElement()
{
}

Binary* BuildfileElement::GetBinary(string name)
{
   map<string, Binary* >::iterator i = mBinaries.find(name);
   if(i != mBinaries.end())
      return i->second;
   else
      return NULL;
}

void BuildfileElement::GetAllBinaries(vector<Binary* >& binaries)
{
   map<string, Binary* >::iterator i;

   for(i = mBinaries.begin(); i != mBinaries.end(); ++i)
      binaries.push_back(i->second);
}

bool BuildfileElement::Clean(CleanLevel level)
{
   bool ok = true;

   for(unsigned int i = 0; i < mCleanList.size(); i++)
   {
      if(Filesystem::Exists(mCleanList[i]))
      {
         cout << "Deleting '" << mCleanList[i] << "'" << endl;
         Filesystem::Delete(mCleanList[i]);
         if(Filesystem::Exists(mCleanList[i]))
         {
            cerr << "Could not delete file: " << mCleanList[i] << endl;
            ok = false;
         }
      }
   }

   if(level == DISTCLEAN || level == CVSCLEAN)
   {
      for(unsigned int i = 0; i < mDistcleanList.size(); i++)
      {
         if(Filesystem::Exists(mDistcleanList[i]))
         {
            cout << "Deleting '" << mDistcleanList[i] << "'" << endl;
            Filesystem::Delete(mDistcleanList[i]);
            if(Filesystem::Exists(mDistcleanList[i]))
            {
               cerr << "Could not delete file: " << mDistcleanList[i] << endl;
               ok = false;
            }
         }
      }
   }

   if(level == CVSCLEAN)
   {
      for(unsigned int i = 0; i < mCvscleanList.size(); i++)
      {
         if(Filesystem::Exists(mCvscleanList[i]))
         {
            cout << "Deleting '" << mCvscleanList[i] << "'" << endl;
            Filesystem::Delete(mCvscleanList[i]);
            if(Filesystem::Exists(mCvscleanList[i]))
            {
               cerr << "Could not delete file: " << mCvscleanList[i] << endl;
               ok = false;
            }
         }
      }
   }

   return ok;
}

class BuildfileElement::PackageElement: public ParseElement
{
   Package* mPackage;
   bool mInPublic;
   bool mHasMainModule;
public:
   PackageElement(Package* package): mPackage(package), mInPublic(false), mHasMainModule(false) {}
   ~PackageElement() {}

   void InPublic(bool flag) throw(ParseException)
   {
      mInPublic = flag;
   }

   ParseElement* StartSubElement(string type, string name, const vector<Token >& flags,
                                 string sourceFile, int linenum) throw(ParseException)
   {
      if(type == "package")
      {
         if(flags.size() > 0)
            throw ParseException("Parse error before \"" + flags[0].mData + "\".");

         return new PackageElement(mPackage->AddSubPackage(name, mInPublic, sourceFile, linenum));
      }
      else if(type == "module")
         throw ParseException("Modules should not have bodies.");
      else
         throw ParseException("Bad object type.");
   }

   void EndSubElement(ParseElement* element) throw(ParseException)
   {
      delete element;
   }

   void ProcessCodeLine(const vector<Token >& tokens,
                        string sourceFile, int linenum) throw(ParseException)
   {
      if(tokens.size() == 2
           && tokens[0].mType == Token::SYMBOL
           && tokens[1].mType == Token::SYMBOL)
      {
         if(tokens[0].mData == "package")
         {
            Package* subPackage = mPackage->AddSubPackage(
               tokens[1].mData, mInPublic, sourceFile, linenum);
            ParsePackage(subPackage);
            return;
         }
         else if(tokens[0].mData == "module")
         {
            mPackage->AddModule(tokens[1].mData, mInPublic, sourceFile, linenum);
            return;
         }
      }
      throw ParseException("Parse error before \"" + tokens[0].mData + "\"");
   }

   static void ParsePackage(Package* subPackage)
   {
      ParseElement* subElement = new PackageElement(subPackage);
      string filename = "src/" + subPackage->GetFilename()
                      + "/package.kake";

      if(Filesystem::Exists(filename))
      {
         Parser parser;
         parser.Start(subElement);
         parser.Parse(filename);
         parser.Finish();
         delete subElement;
         if(parser.HadErrors())
            throw ParseException("(failed package included from here)");
      }
      else
         throw ParseException("Could not find package file: " + filename);
   }
};

class BuildfileElement::BinaryElement: public ParseElement
{
   Binary* mBinary;
   Package* mRootPackage;
   bool mHasMain;

public:
   BinaryElement(Binary* binary, Package* rootPackage)
      : mBinary(binary),
        mRootPackage(rootPackage),
        mHasMain(false)
   {}
   ~BinaryElement() {}

   void InPublic(bool flag) throw(ParseException)
   {
      throw ParseException("Public and private not valid here.");
   }

   ParseElement* StartSubElement(string type, string name, const vector<Token >& flags,
                                 string sourceFile, int linenum) throw(ParseException)
   {
      throw ParseException("Executables and libraries cannot contain other objects.");
   }

   void EndSubElement(ParseElement* element) throw(ParseException)
   {
      //will never happen
   }

   void ProcessCodeLine(const vector<Token >& tokens,
                        string sourceFile, int linenum) throw(ParseException)
   {
      if(tokens[0].mType == Token::SYMBOL)
      {
         if(tokens[0].mData == "cflags")
         {
            if(tokens.size() != 2 || tokens[1].mType != Token::STRING)
               throw ParseException("Parameter to cflags must be a single string.");

            vector<string > cflags;
            SeparateAtSpaces(cflags, tokens[1].mData);
            for(unsigned int i = 0; i < cflags.size(); i++)
               mBinary->AddCFlag(cflags[i]);
         }
         else if(tokens[0].mData == "ldflags")
         {
            if(tokens.size() != 2 || tokens[1].mType != Token::STRING)
               throw ParseException("Parameter to ldflags must be a single string.");

            vector<string > ldflags;
            SeparateAtSpaces(ldflags, tokens[1].mData);
            for(unsigned int i = 0; i < ldflags.size(); i++)
               mBinary->AddLdFlag(ldflags[i]);
         }
         else if(tokens[0].mData == "define")
         {
            if(tokens.size() < 2 || tokens.size() > 3
            || tokens[1].mType != Token::SYMBOL
            || (tokens.size() == 3 && tokens[2].mType != Token::STRING))
               throw ParseException("Wrong parameters to define (should be symbol optionally followed by string).");

            if(tokens.size() == 3)
               mBinary->AddDefine(tokens[1].mData, tokens[2].mData);
            else
               mBinary->AddDefine(tokens[1].mData, "1");
         }
         else
         {
            unsigned int pos = 0;
            bool isMain = false;
            if(tokens[0].mData == "main")
            {
               if(mHasMain)
                  throw ParseException("Multiple main declarations found in one executable or library.");
               if(tokens.size() < 2 || tokens[1].mType != Token::SYMBOL)
                  throw ParseException("Missing symbol name after main.");
               mHasMain = isMain = true;
            }
            else if(tokens[0].mData != "include")
               throw ParseException("Parse error before \"" + tokens[0].mData + "\"");

            pos = 1;  //skip "main" or "include"

            //have to parse a scoped symbol name... (like "guf::process::blah")
            string symbolName;

            if(tokens[pos].mType != Token::SYMBOL)
               throw ParseException("Parse error before \"" + tokens[pos].mData + "\"");
            symbolName = tokens[pos].mData;

            for(++pos; pos < tokens.size(); pos += 2)
            {
               if(tokens[pos].mType != Token::DOUBLE_COLON)
                  throw ParseException("Parse error before \"" + tokens[pos].mData + "\"");

               symbolName += tokens[pos].mData;

               if(pos + 1 > tokens.size())
                  throw ParseException("Incomplete symbol name given to main.");

               if(tokens[pos+1].mType != Token::SYMBOL)
                  throw ParseException("Parse error before \"" + tokens[pos+1].mData + "\"");

               symbolName += tokens[pos+1].mData;
            }

            Buildable* buildable = mRootPackage->GetBuildable(symbolName);
            if(buildable == NULL)
               throw ParseException("No package or module \"" + symbolName + "\" defined.");
            if(isMain)
            {
               if(Module* module = dynamic_cast<Module*>(buildable))
                  mBinary->SetMainModule(module);
               else
                  throw ParseException("Main object must be a module.");
            }
            else
               mBinary->AddObject(buildable);
         }
      }
      else
         throw ParseException("Parse error before \"" + tokens[0].mData + "\"");
   }
};

class BuildfileElement::CleanElement: public ParseElement
{
   vector<string >& mCleanList;

public:
   CleanElement(vector<string >& cleanList)
      : mCleanList(cleanList)
   {}
   ~CleanElement() {}

   void InPublic(bool flag) throw(ParseException)
   {
      throw ParseException("Public and private not valid here.");
   }

   ParseElement* StartSubElement(string type, string name, const vector<Token >& flags,
                                 string sourceFile, int linenum) throw(ParseException)
   {
      throw ParseException("Clean block cannot contain other objects.");
   }

   void EndSubElement(ParseElement* element) throw(ParseException)
   {
      //will never happen
   }

   void ProcessCodeLine(const vector<Token >& tokens,
                        string sourceFile, int linenum) throw(ParseException)
   {
      if(tokens.size() != 1 || tokens[0].mType != Token::STRING)
         throw ParseException("Syntax error in clean declaration.");

      mCleanList.push_back(tokens[0].mData);
   }
};

void BuildfileElement::InPublic(bool flag) throw(ParseException)
{
   //ignore
}

ParseElement* BuildfileElement::StartSubElement(string type, string name, const vector<Token >& flags,
                                                string sourceFile, int linenum) throw(ParseException)
{
   if(type == "package")
   {
      if(flags.size() > 0)  //no flags allowed
         throw ParseException("Parse error before \"" + flags[0].mData + "\".");

      return new PackageElement(mRootPackage.AddSubPackage(name, true, sourceFile, linenum));
   }
   else if(type == "module")
      throw ParseException("Modules must be within packages.");
   else if(type == "executable")
   {
      if(flags.size() > 0)  //no flags allowed
         throw ParseException("Parse error before \"" + flags[0].mData + "\".");

      Binary* binary = new Binary(name, Binary::EXECUTABLE, sourceFile, linenum);
      mBinaries[name] = binary;
      return new BinaryElement(binary, &mRootPackage);
   }
   else if(type == "library")
   {
      Binary::Type binType = Binary::LIBRARY;
      if(flags.size() > 1)
         throw ParseException("Parse error before \"" + flags[0].mData + "\".");
      else if(flags.size() == 1)
      {
         if(flags[0].mType != Token::SYMBOL || flags[0].mData != "static")
            throw ParseException("Parse error before \"" + flags[0].mData + "\".");
         binType = Binary::STATIC_LIBRARY;
      }

      Binary* binary = new Binary(name, binType, sourceFile, linenum);
      mBinaries[name] = binary;
      return new BinaryElement(binary, &mRootPackage);
   }
   else if(type == "clean")
   {
      if(flags.size() > 0)  //no flags allowed
         throw ParseException("Parse error before \"" + flags[0].mData + "\".");

      if(name == "normal")
         return new CleanElement(mCleanList);
      else if(name == "dist")
         return new CleanElement(mDistcleanList);
      else if(name == "cvs")
         return new CleanElement(mCvscleanList);
      else
         throw ParseException("\"clean\" must be followed by \"normal\", \"dist\", or \"cvs\".");
   }
   else
      throw ParseException("Unknown object type.");
}

void BuildfileElement::EndSubElement(ParseElement* element) throw(ParseException)
{
   delete element;   //no longer needed
}

void BuildfileElement::ProcessCodeLine(const vector<Token >& tokens,
                                       string sourceFile, int linenum) throw(ParseException)
{
   if(tokens.size() == 2
        && tokens[0].mType == Token::SYMBOL
        && tokens[1].mType == Token::SYMBOL)
   {
      if(tokens[0].mData == "package")
      {
         Package* subPackage = mRootPackage.AddSubPackage(
            tokens[1].mData, true, sourceFile, linenum);
         PackageElement::ParsePackage(subPackage);
         return;
      }
      else if(tokens[0].mData == "module")
         throw ParseException("Modules must be within packages.");
   }
   throw ParseException("Parse error before \"" + tokens[0].mData + "\"");
}

void BuildfileElement::SeparateAtSpaces(vector<string >& output, string input)
{
   unsigned int wordStart = 0;

   for(unsigned int i = 0; i < input.length(); i++)
   {
      if(isspace(input[i]))
      {
         if(wordStart < i)
            output.push_back(input.substr(wordStart, i - wordStart));

         wordStart = i + 1;
      }
   }

   if(wordStart != input.length())
      output.push_back(input.substr(wordStart));
}

} //namespace kake
