/* kake - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Compiler.h"
#include "Filesystem.h"
#include "Process.h"
#include <iostream>
#include <fstream>
#include <stdio.h>

namespace kake {

static const string COMPILER_PROGRAM = "c++";
static const string LINKER_PROGRAM   = "c++";
static const string AR_PROGRAM       = "ar";
static const string RANLIB_PROGRAM   = "ranlib";

#ifdef PLATFORM_WIN32
static const string DLLWRAP_PROGRAM = "dllwrap";
static const string DLLTOOL_PROGRAM = "dlltool";
#endif

namespace //anonymous namespace
{
   vector<pair<string, string> > gDefines;
//   vector<string > gIncludeDirs;
//   vector<string > gLibraryDirs;
   vector<string > gLibraries;
   vector<string > gParams;
   vector<string > gLinkerParams;

   bool gDebug = false;
}

// class member functions ============================================
bool Compiler::Compile(string sourceFile, string objectFile, bool outputDeps)
{
   vector<string > params = gParams;

   for(unsigned int i = 0; i < gDefines.size(); i++)
      params.push_back("-D" + gDefines[i].first + "=" + gDefines[i].second);

   params.push_back("-Wall");  //I pity the foo' who doesn't use -Wall!
   params.push_back("-fmessage-length=0");

   if(gDebug)
   {
      params.push_back("-g");
      params.push_back("-DDEBUG");
   }
   else
   {
      params.push_back("-O2");
      params.push_back("-DNDEBUG");
   }

   params.push_back("-I./src");

#if !defined(PLATFORM_WIN32)
   //we need to hard-code -fPIC on platforms where the flag exists.
   //we *always* want position-independent code!
   params.push_back("-fPIC");
#endif

   if(outputDeps)
   {
      string depFilename;

      int extensionPos = objectFile.find_last_of('.');
      if(extensionPos == -1)
      {
         cerr << "Invalid object file: " << objectFile << endl;
         throw "Compiler::Compile(): invalid object file specified";
      }
      depFilename = objectFile.substr(0, extensionPos) + ".d";

      params.push_back("-Wp,-MD," + depFilename);    //output dependencies
   }

   params.push_back("-c");
   params.push_back(sourceFile);

   params.push_back("-o");
   params.push_back(objectFile);

   Process proc;
   proc.Run(COMPILER_PROGRAM, params, false);
   if(proc.Wait() != 0)
      return false;
   return true;
}

void Compiler::GetDependencies(vector<string >& target, string objectFile)
{
   string depFilename;

   int extensionPos = objectFile.find_last_of('.');
   if(extensionPos == -1)
   {
      cerr << "Invalid object file: " << objectFile << endl;
      throw "Compiler::GetDependencies(): invalid object file specified";
   }
   depFilename = objectFile.substr(0, extensionPos) + ".d";

   fstream depfile(depFilename.c_str(), ios::in);
   string token;

   depfile >> token;   //"objectfile.o:"

   //read in all dependencies
   while(depfile >> token)
   {
      if(token != "\\")
         target.push_back(token);
   }
}

bool Compiler::LinkExecutable(const vector<string >& objects, string execname)
{
   vector<string > params;

   if(gDebug)
      params.push_back("-g");
   else
      params.push_back("-O2");

#if !defined(PLATFORM_WIN32)
   params.push_back("-rdynamic");
#endif
   params.push_back("-o");
   params.push_back("bin/" + execname);

   params.insert(params.end(), objects.begin(), objects.end());

   params.push_back("-L./lib");
   for(unsigned int i = 0; i < gLibraries.size(); i++)
      params.push_back("-l" + gLibraries[i]);

   params.insert(params.end(), gLinkerParams.begin(), gLinkerParams.end());

   Process proc;
   proc.Run(LINKER_PROGRAM, params, false);
   if(proc.Wait() != 0)
      return false;
   return true;
}

bool Compiler::LinkLibrary(const vector<string >& objects, string libname)
{
   vector<string > params;

#ifdef PLATFORM_WIN32
   string dllname = libname + ".dll";
   string deffile = "build/" + libname  + ".def";
   string dllfile = "bin/" + libname  + ".dll";
   string arfile = "lib/lib" + libname  + ".a";

   //dlltool --output-def $defname --export-all-symbols --kill-at $objects
   params.push_back("--output-def");
   params.push_back(deffile);
   params.push_back("--export-all-symbols");
   params.push_back("--kill-at");
   params.insert(params.end(), objects.begin(), objects.end());

   Process proc;
   proc.Run(DLLTOOL_PROGRAM, params, false);
   if(proc.Wait() != 0)
      return false;

   params.clear();

   //dllwrap --driver-name $CC --def $defname -o $dllname $objects $libs $ldflags
   params.push_back("--driver-name");
   params.push_back(LINKER_PROGRAM);
   params.push_back("--def");
   params.push_back(deffile);
   params.push_back("-o");
   params.push_back(dllfile);
   params.insert(params.end(), objects.begin(), objects.end());
   params.push_back("-L./lib");
   for(unsigned int i = 0; i < gLibraries.size(); i++)
      params.push_back("-l" + gLibraries[i]);

   if(gDebug)
      params.push_back("-g");
   else
      params.push_back("-O2");
   params.insert(params.end(), gLinkerParams.begin(), gLinkerParams.end());

   proc.Run(DLLWRAP_PROGRAM, params, false);
   if(proc.Wait() != 0)
   {
      Filesystem::Delete(deffile); //delete def file
      return false;
   }

   params.clear();

   //dlltool --dllname $dllname --def $defname --output-lib $arname
   params.push_back("--dllname");
   params.push_back(dllname);
   params.push_back("--def");
   params.push_back(deffile);
   params.push_back("--output-lib");
   params.push_back(arfile);

   proc.Run(DLLTOOL_PROGRAM, params, false);
   if(proc.Wait() != 0)
   {
      Filesystem::Delete(deffile); //delete def file
      return false;
   }

   Filesystem::Delete(deffile); //delete def file

   return true;

#else
   if(gDebug)
      params.push_back("-g");
   else
      params.push_back("-O2");

   params.push_back("-shared");
   params.push_back("-Wl,-soname,lib" + libname + ".so");
   params.push_back("-o");
   params.push_back("lib/lib" + libname + ".so");

   params.insert(params.end(), objects.begin(), objects.end());

   params.push_back("-L./lib");
   for(unsigned int i = 0; i < gLibraries.size(); i++)
      params.push_back("-l" + gLibraries[i]);

   params.insert(params.end(), gLinkerParams.begin(), gLinkerParams.end());

   Process proc;
   proc.Run(LINKER_PROGRAM, params, false);
   if(proc.Wait() != 0)
      return false;
   return true;
#endif
}

bool Compiler::LinkStaticLibrary(const vector<string >& objects, string libname)
{
   vector<string > params;

   string arfile = "lib/lib" + libname  + ".a";

   params.push_back("cru");
   params.push_back(arfile);
   params.insert(params.end(), gLinkerParams.begin(), gLinkerParams.end());
   params.insert(params.end(), objects.begin(), objects.end());

   Process proc;
   proc.Run(AR_PROGRAM, params, false);
   if(proc.Wait() != 0)
      return false;

   params.clear();
   params.push_back(arfile);
   proc.Run(RANLIB_PROGRAM, params, false);
   if(proc.Wait() != 0)
      return false;

   return true;
}

string Compiler::GetExecutableFileForName(string execname)
{
#ifdef PLATFORM_WIN32
   return "bin/" + execname + ".exe";
#else
   return "bin/" + execname;
#endif
}

string Compiler::GetDllFileForName(string libname)
{
#ifdef PLATFORM_WIN32
   return "bin/" + libname + ".dll";
#else
   return "lib/lib" + libname + ".so";
#endif
}

string Compiler::GetLibraryFileForName(string libname)
{
#ifdef PLATFORM_WIN32
   return "lib/lib" + libname + ".a";
#else
   return "lib/lib" + libname + ".so";
#endif
}

string Compiler::GetStaticLibraryFileForName(string libname)
{
   return "lib/lib" + libname + ".a";
}

void Compiler::SetDebug(bool debug)
{
   gDebug = debug;
}

void Compiler::AddDefine(string name, string value)
   { gDefines.push_back(pair<string, string >(name, value)); }
void Compiler::AddDefine(string name)
   { AddDefine(name, "1"); }
void Compiler::AddDefines(const vector<pair<string, string > >& defines)
   { gDefines.insert(gDefines.end(), defines.begin(), defines.end()); }
void Compiler::ClearDefines()
   { gDefines.clear(); }

#if 0
void Compiler::AddIncludeDir(string dir)
   { gIncludeDirs.push_back(dir); }
void Compiler::AddIncludeDirs(const vector<string >& includeDirs)
   { gIncludeDirs.insert(gIncludeDirs.end(), includeDirs.begin(), includeDirs.end()); }
void Compiler::ClearIncludeDirs()
   { gIncludeDirs.clear(); }

void Compiler::AddLibraryDir(string dir)
   { gLibraryDirs.push_back(dir); }
void Compiler::AddLibraryDirs(const vector<string >& dirs)
   { gLibraryDirs.insert(gLibraryDirs.end(), dirs.begin(), dirs.end()); }
void Compiler::ClearLibraryDirs()
   { gLibraryDirs.clear(); }
#endif

void Compiler::AddLibrary(string library)
   { gLibraries.push_back(library); }
void Compiler::AddLibraries(const vector<string >& libraries)
   { gLibraries.insert(gLibraries.end(), libraries.begin(), libraries.end()); }
void Compiler::ClearLibraries()
   { gLibraries.clear(); }

void Compiler::AddParameter(string param)
   { gParams.push_back(param); }
void Compiler::AddParameters(const vector<string >& params)
   { gParams.insert(gParams.end(), params.begin(), params.end()); }
void Compiler::ClearParameters()
   { gParams.clear(); }

void Compiler::AddLinkerParameter(string param)
   { gLinkerParams.push_back(param); }
void Compiler::AddLinkerParameters(const vector<string >& params)
   { gLinkerParams.insert(gLinkerParams.end(), params.begin(), params.end()); }
void Compiler::ClearLinkerParameters()
   { gLinkerParams.clear(); }

} //namespace kake
