/* kake - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Filesystem.h"
#ifdef PLATFORM_POSIX
# include <unistd.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <dirent.h>
# include <fcntl.h>
#elif defined(PLATFORM_WIN32)
# include <io.h>
# include <direct.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <dirent.h>
# define open _open
# define lseek _lseek
# define close _close
# define read _read
# define write _write
# define access _access
# define mkdir(NAME, MODE) _mkdir(NAME)
# define stat _stat
# define O_RDONLY _O_RDONLY
# define O_WRONLY _O_WRONLY
# define O_CREAT  _O_CREAT
# define O_TRUNC  _O_TRUNC
#else
# error "Port Me!"
#endif

#include <stdio.h>

namespace kake {

// class member functions ============================================
bool Filesystem::Exists(string filename)
{
#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   if(access(filename.c_str(), F_OK) < 0)
      return false;
   return true;
#endif
}

void Filesystem::MakeDirectory(string dirname)
{
   if(Exists(dirname))
      return;

   //remove trailing slash
   if(dirname[dirname.length() - 1] == '/')
      dirname = dirname.substr(0, dirname.length() - 1);

   //make the parent directory if it doesn't exist
   int pos = dirname.find_last_of('/');
   if(pos > 0)
   {
      string parent = dirname.substr(0, pos);
      MakeDirectory(parent);
   }

   //make our directory
#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   mkdir(dirname.c_str(), 0777);
#endif
}

void Filesystem::MakeContainingDirectory(string filename)
{
   int pos = filename.find_last_of('/');
   if(pos > 0)
   {
      string parent = filename.substr(0, pos);
      MakeDirectory(parent);
   }
}

void Filesystem::Delete(string filename)
{
   remove(filename.c_str());

   //if file was the last file in the directory, remove the directory
   unsigned int pos;
   pos = filename.find_last_of('/');
   if(pos != string::npos)
      RemoveDirectory(filename.substr(0, pos));
}

void Filesystem::RemoveDirectory(string dirname)
{
   if(rmdir(dirname.c_str()) >= 0)
   {
      //if directory was the last subdir of its parent, remove the parent
      unsigned int pos;
      pos = dirname.find_last_of('/');
      if(pos != string::npos)
         RemoveDirectory(dirname.substr(0, pos));
   }
}

void Filesystem::GetDirectoryTree(vector<string >& dirs, string basedir)
{
   DIR* dir = opendir(basedir.c_str());
   struct dirent* entry;
   if(dir == NULL)
      return;  //not a directory

   dirs.push_back(basedir);
   basedir += "/";

   while((entry = readdir(dir)) != NULL)
   {
      //ignore "." and ".."
      if(strcmp(entry->d_name, ".") == 0 || strcmp(entry->d_name, "..") == 0)
         continue;
      //add all sub-directories
      GetDirectoryTree(dirs, basedir + entry->d_name);
   }

   closedir(dir);
}

void Filesystem::Copy(string source, string destination)
{
#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   int infd = open(source.c_str(), O_RDONLY);
   if(infd < 0)
   {
      perror("open input");
      throw "Filesystem::Copy(): couldn't open input";
   }
   int outfd = open(destination.c_str(), O_WRONLY | O_CREAT, 0666);
   if(outfd < 0)
   {
      perror("open output");
      throw "Filesystem::Copy(): couldn't open output";
   }

   int numread;
   char buffer[BUFSIZ];
   while((numread = read(infd, buffer, BUFSIZ)) > 0)
      write(outfd, buffer, numread);
   close(infd);
   close(outfd);
#endif
}

unsigned int Filesystem::GetTimestamp(string filename)
{
#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   struct stat stats;
   if(stat(filename.c_str(), &stats) < 0)
      return 0;
   return stats.st_mtime;
#endif
}

int  Filesystem::mStderrBackup = -1;

void Filesystem::RedirectStderrToFile(string filename)
{
#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   if(mStderrBackup != -1)
      RestoreStderr();

   int fd = open(filename.c_str(), O_WRONLY | O_CREAT | O_TRUNC, 0666);
   if(fd > 0)
   {
      mStderrBackup = dup(2);
      dup2(fd, 2);   //direct stderr to file
      close(fd);
   }
#endif
}

bool Filesystem::CheckStderr()
{
#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   if(mStderrBackup == -1)
      return false;   //not re-directing.

   static int lastPos = 0;
   int newPos = lseek(2, 0, SEEK_CUR);
   if(newPos != lastPos)
   {
      lastPos = newPos;
      return true;
   }
   else
      return false;
#endif
}

void Filesystem::RestoreStderr()
{
#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   if(mStderrBackup == -1)
      return;
   dup2(mStderrBackup, 2);
   close(mStderrBackup);
   mStderrBackup = -1;
#endif
}

} //namespace kake
