/* kake - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Kake.h"
#include <iostream>
#include <stdio.h>
#include "Filesystem.h"
#include "Process.h"
#include "Preprocessor.h"
#include "Parser.h"
#include "Binary.h"
#include "Compiler.h"

namespace kake {

// class member functions ============================================
BuildfileElement* Kake::mBuildfile;
vector<string >   Kake::mCommandParams;

int Kake::Main(int argc, char* argv[])
{
   enum
   {
      BUILD,
      REBUILD,
      CLEAN,
      DISTCLEAN,
      CVSCLEAN,
      INSTALLSCRIPT,
      SOURCE_DIST,
      BINARY_DIST
   } command = BUILD;
   string commandName = "build";
   bool foundCommandName = false;
   bool log = true;

   for(int i = 1; i < argc; i++)
   {
      if(argv[i][0] == '-')
      {
         if(argv[i][1] == '-')
         {
            if(!strcmp(argv[1], "--verbose"))
            {
               Process::SetEcho(true);
               log = false;
            }
            else if(!strcmp(argv[1], "--debug"))
            {
               Compiler::SetDebug(true);
            }
            else
               PrintUsage(string("Unrecognized parameter: ") + argv[1]);
         }
         else
         {
            int len = strlen(argv[i]);

            for(int j = 1; j < len; j++)
            {
               if(argv[i][j] == 'v')
               {
                  Process::SetEcho(true);
                  log = false;
               }
               else if(argv[i][j] == 'd')
               {
                  Compiler::SetDebug(true);
               }
               else
                  PrintUsage(string("Unrecognized parameter: -") + argv[i][j]);
            }
         }
      }
      else if(!foundCommandName)
      {
         if(!strcmp(argv[i], "build"))
         {
            foundCommandName = true;
            command = BUILD;
            commandName = "build";
         }
         else if(!strcmp(argv[i], "rebuild"))
         {
            foundCommandName = true;
            command = REBUILD;
            commandName = "rebuild";
         }
         else if(!strcmp(argv[i], "clean"))
         {
            foundCommandName = true;
            command = CLEAN;
            commandName = "clean";
         }
         else if(!strcmp(argv[i], "distclean"))
         {
            foundCommandName = true;
            command = DISTCLEAN;
            commandName = "clean";
         }
         else if(!strcmp(argv[i], "cvsclean"))
         {
            foundCommandName = true;
            command = CVSCLEAN;
            commandName = "clean";
         }
         else if(!strcmp(argv[i], "installscript"))
         {
            foundCommandName = true;
            command = INSTALLSCRIPT;
            commandName = "installscript";
         }
         else if(!strcmp(argv[i], "sourcedist"))
         {
            foundCommandName = true;
            command = SOURCE_DIST;
            commandName = "sourcedist";
         }
         else if(!strcmp(argv[i], "bindist"))
         {
            foundCommandName = true;
            command = BINARY_DIST;
            commandName = "bindist";
         }
         else
         {
            foundCommandName = true;
            command = BUILD;
            commandName = "build";
            mCommandParams.push_back(argv[i]);
         }
      }
      else
         mCommandParams.push_back(argv[i]);
   }

   //find all buildfiles
   vector<string > dirtree;
   vector<string > buildfiles;
   string buildfile;

   //check "src" and all sub-directories
   Filesystem::GetDirectoryTree(dirtree, "src");
   for(unsigned int i = 0; i < dirtree.size(); i++)
   {
      buildfile = dirtree[i] + "/Kakefile";
      if(Filesystem::Exists(buildfile))
         buildfiles.push_back(buildfile);
   }

   buildfile = "Kakefile"; //check current directory
   if(Filesystem::Exists(buildfile))
      buildfiles.push_back(buildfile);

   if(buildfiles.size() <= 0)
   {
      cerr << "No Kakefile's found." << endl;
      return 1;
   }

   Parser parser;
   mBuildfile = new BuildfileElement;
   parser.Start(mBuildfile);

   for(unsigned int i = 0; i < buildfiles.size(); i++)
      parser.Parse(buildfiles[i]);

   parser.Finish();
   if(parser.HadErrors())
   {
      cerr << "Errors in build files." << endl;
      return 1;
   }

   bool result = true;

   switch(command)
   {
   case BUILD:
      result = Build(log);
      break;
   case REBUILD:
      if(Clean(BuildfileElement::CLEAN))
         result = Build(log);
      else
         result = false;
      break;
   case CLEAN:
      result = Clean(BuildfileElement::CLEAN);
      break;
   case DISTCLEAN:
      result = Clean(BuildfileElement::DISTCLEAN);
      break;
   case CVSCLEAN:
      result = Clean(BuildfileElement::CVSCLEAN);
      break;
   case INSTALLSCRIPT:
      result = MakeInstallScript();
      break;
   case SOURCE_DIST:
      result = MakeSourceDistribution();
      break;
   case BINARY_DIST:
      if(Build(log))
         result = MakeBinaryDistribution();
      else
         result = false;
      break;
   default:
      //should not happen
      break;
   }

   if(result)
   {
      cout << commandName << " succeeded." << endl;
      return 0;
   }
//   else
//   {
      cout << commandName << " failed." << endl;
      return 1;
//   }
}

void Kake::PrintUsage(string message)
{
   cerr << message << endl;
   cerr << "usage:\n"
           "kake [build|rebuild|clean|installscript|sourcedist|bindist]\n"
           "The parameter specifies the action to perform.  If none is specified,\n"
           "the default of 'build' will be used.  The result of each action is\n"
           "as follows:\n"
           "    build          Builds all packages and modules.\n"
           "    clean          Deletes all files and executables output by a\n"
           "                   build.\n"
           "    rebuild        Equivalent to 'clean' followed by 'build'.\n"
           "    installscript  Generates an installation script and an uninstall\n"
           "                   script.  These can be used to install and uninstall\n"
           "                   the binaries once they have been built.\n"
           "    sourcedist     Builds a source code distribution of the program and\n"
           "                   places it in the directory 'sourcedist'.\n"
           "    bindist        Builds a binary distribution of the pragram and places\n"
           "                   it in the directoly 'bindist'.  This will include all\n"
           "                   binaries, an install script, and an uninstall script,\n"
           "                   but no source code.\n"
           "NOTE:  At this point, only 'build', 'clean', and 'rebuild' are implemented." << endl;
   exit(1);
}

bool Kake::Build(bool log)
{
   if(!Filesystem::Exists("bin"))
      Filesystem::MakeDirectory("bin");
   if(!Filesystem::Exists("lib"))
      Filesystem::MakeDirectory("lib");
   if(!Filesystem::Exists("build"))
      Filesystem::MakeDirectory("build");

   bool ok = true;

   vector<Binary* > binaries;

   if(mCommandParams.size() <= 0)
   {
      mBuildfile->GetAllBinaries(binaries);
      if(binaries.size() == 0)
      {
         cerr << "No executables or libraries defined in build file." << endl;
         return false;
      }
   }
   else
   {
      Binary* binary;
      for(unsigned int i = 0; i < mCommandParams.size(); i++)
      {
         binary = mBuildfile->GetBinary(mCommandParams[i]);
         if(binary == NULL)
         {
            cerr << "Could not find any executable or library called \""
                 << mCommandParams[i] << "\"." << endl;
            ok = false;
         }
         binaries.push_back(binary);
      }
   }

   if(!ok)
      return false;

   for(unsigned int i = 0; i < binaries.size(); i++)
   {
      if(!binaries[i]->Check())
         ok = false;
   }

   if(!ok)
      return false;

   if(log)
   {
      cout << "Logging build to file kake.log" << endl;
      Filesystem::RedirectStderrToFile("kake.log");
   }

   Buildable::BuildResult result = Buildable::SKIPPED;
   for(unsigned int i = 0; i < binaries.size(); i++)
   {
      switch(binaries[i]->Build())
      {
      case Buildable::UPDATED:
         result = Buildable::UPDATED;
         break;
      case Buildable::FAILED:
         result = Buildable::FAILED;
         break;
      case Buildable::SKIPPED:
         break;
      }

      if(result == Buildable::FAILED)
         break;
   }

   if(log)
   {
      if(result == Buildable::UPDATED)
         cerr << "build succeeded." << endl;
      else if(result == Buildable::FAILED)
         cerr << "\nbuild failed." << endl;
      else
         cerr << "Nothing needed to be built.\nbuild succeeded." << endl;
      Filesystem::RestoreStderr();
   }

   if(result == Buildable::SKIPPED)
      cout << "Nothing needed to be built." << endl;
   else if(result == Buildable::FAILED)
      ok = false;

   return ok;
}

bool Kake::Clean(BuildfileElement::CleanLevel level)
{
   bool ok = true;

   vector<Binary* > binaries;
   mBuildfile->GetAllBinaries(binaries);
   for(unsigned int i = 0; i < binaries.size(); i++)
   {
      if(!binaries[i]->Clean())
         ok = false;
   }

   if(!ok)
      return false;

   if(!mBuildfile->Clean(level))
      return false;

   if(Filesystem::Exists("bin"))
      Filesystem::RemoveDirectory("bin");
   if(Filesystem::Exists("lib"))
      Filesystem::RemoveDirectory("lib");
   if(Filesystem::Exists("build"))
      Filesystem::RemoveDirectory("build");

   return true;
}

bool Kake::MakeInstallScript()
{
   cerr << "Sorry, not implemented." << endl;
   return false;
}

bool Kake::MakeSourceDistribution()
{
   cerr << "Sorry, not implemented." << endl;
   return false;
}

bool Kake::MakeBinaryDistribution()
{
   cerr << "Sorry, not implemented." << endl;
   return false;
}

} //namespace kake

#ifndef BUILDING_WITH_KAKE
   KAKE_WRITE_MAIN(kake, Kake)
#endif
