/* kake - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Module.h"
#include <iostream>
#include <fstream>
#include "Filesystem.h"
#include "Compiler.h"

namespace kake {
using namespace std;

// class member functions ============================================
Module::Module(string name, bool isPublic, string sourcefile, int linenum)
   : mIsPublic(isPublic),
     mSourceFile(sourcefile),
     mLineNum(linenum),
     mBuilt(false),
     mCleaned(false),
     mChecked(false),
     mDocumented(false)
{
   mName = mFilename = mNameWithUnderscores = name;

   unsigned int pos;
   while((pos = mFilename.find_first_of(':')) != string::npos)
   {
      mFilename.replace(pos, 2, "/");
      mNameWithUnderscores.replace(pos, 2, "_");
   }
}

Module::~Module()
{
}

Buildable::BuildResult Module::Build()
{
   if(mBuilt)
      return mBuildResult;  //already built!

   mBuilt = true;
   //assume skipped untill we actually build something...
   mBuildResult = SKIPPED;

   //whether or not we printed the "Compiling [whatever]..." message.
   bool alreadyPrintedMessage = false;


   //may need to compile some source code...
   string sourceFile = "src/" + mFilename + ".cpp";
   string objectFile = "build/" + mNameWithUnderscores + ".o";

   bool needUpdate = false;

   //check if we need to re-compile
   if(!Filesystem::Exists(objectFile))
      needUpdate = true;    //object file doesn't exist; need to compile it
   else
   {
      //check all dependencies (files which are #included by the source file)
      unsigned int timestamp = Filesystem::GetTimestamp(objectFile);
      vector<string > dependencies;

      Compiler::GetDependencies(dependencies, objectFile);
      for(unsigned int i = 0; i < dependencies.size(); i++)
      {
         if(!Filesystem::Exists(dependencies[i]))
         {
            needUpdate = true;   //one of the dependencies no longer exists
            break;
         }
         if(Filesystem::GetTimestamp(dependencies[i]) > timestamp)
         {
            needUpdate = true;   //one of the dependencies is newer than the object
            break;
         }
      }
   }

   if(needUpdate)
   {
      cout << "Compiling " << mName << "...";
      if(Filesystem::IsStderrRedirected())
      {
         cerr << "Compiling " << mName << "..." << endl;
         cout.flush();
      }
      else
         cout << endl;

      alreadyPrintedMessage = true;

      if(Filesystem::IsStderrRedirected())
         Filesystem::CheckStderr();  //so we can detect warnings

      if(Compiler::Compile(sourceFile, objectFile, true))
      {
         if(Filesystem::IsStderrRedirected() && Filesystem::CheckStderr())
            cout << " (warnings logged)";

         mBuildResult = UPDATED;
      }
      else
      {
         if(Filesystem::IsStderrRedirected())
            cout << " (errors logged)" << endl;
         mBuildResult = FAILED;
      }
   }


   if(Filesystem::IsStderrRedirected() && alreadyPrintedMessage)
   {
      cout << endl;
      cerr << endl;
   }

   if(mIsPublic && mBuildResult != FAILED)
   {
      //copy includes
      string headerFile = "src/" + mFilename + ".h";
      string target = "include/" + mFilename + ".h";

      Filesystem::MakeContainingDirectory(target);
      Filesystem::Copy(headerFile, target);
   }

   return mBuildResult;
}

bool Module::Check()
{
   if(mChecked)
      return mCheckResult;
   mChecked = true;

   mCheckResult = true;

   string sourceFile = "src/" + mFilename + ".cpp";
   string headerFile = "src/" + mFilename + ".h";

   if(!Filesystem::Exists(sourceFile))
   {
      cerr << mSourceFile << ":" << mLineNum
           << ": Source file missing for module " << mName << "." << endl;
      mCheckResult = false;
   }
   if(!Filesystem::Exists(headerFile))
   {
      cerr << mSourceFile << ":" << mLineNum
           << ": Header file missing for module " << mName << "." << endl;
      mCheckResult = false;
   }

   return mCheckResult;
}

bool Module::Clean()
{
   if(mCleaned)
      return mCleanResult;
   mCleaned = true;

   cout << "Cleaning " << mName << "..." << endl;
   Filesystem::Delete("build/" + mNameWithUnderscores + ".o");
   Filesystem::Delete("build/" + mNameWithUnderscores + ".d");
   if(mIsPublic)
      Filesystem::Delete("include/" + mFilename + ".h");

   return mCleanResult = true;
}

bool Module::Document()
{
   if(mDocumented)
      return mDocumentResult;
   mDocumented = true;

   //TODO

   return mDocumentResult = false;
}

void Module::AddRequiredHeader(string header)
   { mRequiredHeaders.push_back(header); }

void Module::AddRequiredLibrary(string library)
   { mRequiredLibraries.push_back(library); }

void Module::GetRequiredHeaders(vector<string >& list)
   { list.insert(list.end(), mRequiredHeaders.begin(), mRequiredHeaders.end()); }

void Module::GetRequiredLibraries(vector<string >& list)
   { list.insert(list.end(), mRequiredLibraries.begin(), mRequiredLibraries.end()); }

void Module::GetObjects(vector<string >& objects)
{
   objects.push_back("build/" + mNameWithUnderscores + ".o");
}

string Module::GetName() {return mName;}
string Module::GetNameWithUnderscores() {return mNameWithUnderscores;}
string Module::GetFilename() {return mFilename;}

} //namespace kake
