/* kake - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef kake_ParseElement_H__INCLUDED__
#define kake_ParseElement_H__INCLUDED__

// includes ==========================================================
#include "Common.h"
#include <string>
#include <vector>

namespace kake {
using namespace std;

// class definition ==================================================
//ParseElement is the base class for elements that can be parsed, such as modules,
// packages, etc.  The idea is that the parser itself doesn't know what types of elements
// it contains.  It just sends the parsed tokens to abstract ParseElement classes.  This
// is very similar to GUF's XML parser system.
class ParseElement
{
public:
   virtual ~ParseElement();

   struct Token
   {
      enum
      {
         SYMBOL,
         STRING,
         NUMBER,

         OPEN_BRACE,
         CLOSE_BRACE,
         DOUBLE_COLON,
         COLON,
         SEMICOLON,
         EQUALS,
         HASH,
         NEWLINE,
      } mType;
      string mData;
      int mNumber;
   };

   //ParseExeption is thrown by ParseElement's functions when the input code is
   // erroneous somehow.
   class ParseException
   {
      string mErrorMessage;
   public:
      ParseException(string errorMessage);
      string GetErrorMessage() const;
   };

   //When the parser comes across the line "public:" or "private:", it calles the current
   // element's InPublic() function and passes true or false, respectively, as the argument.
   virtual void InPublic(bool flag) throw(ParseException) = 0;

   //StartSubElement is called when the parser encounters a sub-element within the current element.
   // [type] is the type of element, such as "module" or "package", and [name] is the symbol name.
   // [flags] contains anything that was found before the type name, such as "main" in "main module".
   // The return value is the object representing the new element.  All data parsed while within the
   // element will be sent to this object.
   virtual ParseElement* StartSubElement(string type, string name, const vector<Token >& flags,
                                         string sourceFile, int linenum) throw(ParseException) = 0;

   //EndSubElement is called when a sub-element has been fully parsed.  The parameter is the object
   // representing the sub-element.  Of course, this will be the same object which was returned by
   // the last call to StartSubElement, so the element does not
   virtual void EndSubElement(ParseElement* element) throw(ParseException) = 0;

   //ProcessCodeLine passes a line of code to the element.
   virtual void ProcessCodeLine(const vector<Token >& tokens,
                                string sourceFile, int linenum) throw(ParseException) = 0;
};

} //namespace kake

#endif //included
