/* kake - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Parser.h"
#include <iostream>
#include "Preprocessor.h"

namespace kake {

// class member functions ============================================
Parser::Parser()
   : mJustParsedNewline(true),
     mJustParsedColon(false),
     mEscapedChar(false),
     mInString(false),
     mHadErrors(false),
     mInDirective(false),
     mLineNum(0)
{
}

Parser::~Parser()
{
}

void Parser::Parse(string filename)
{
   Preprocessor preproc;

   char buffer[BUFSIZ];
   int numread;

   preproc.Preprocess(filename);
   while((numread = preproc.Read(buffer, BUFSIZ)) > 0)
      Parse(buffer, numread);
   if(preproc.HadErrors())
      mHadErrors = true;
}

void Parser::Start(ParseElement* rootElement)
{
   mElementStack.push(rootElement);
}

void Parser::Parse(const char* data, int size)
{
   char curchar;

   for(int i = 0; i < size; i++)
   {
      curchar = data[i];

      if(mInString)
      {
         if(mEscapedChar)
         {
            switch(curchar)
            {
            case 'a':
               mTokenBuffer += '\a';
               break;
            case 'b':
               mTokenBuffer += '\b';
               break;
            case 'f':
               mTokenBuffer += '\f';
               break;
            case 'n':
               mTokenBuffer += '\n';
               break;
            case 'r':
               mTokenBuffer += '\r';
               break;
            case 't':
               mTokenBuffer += '\t';
               break;
            case 'v':
               mTokenBuffer += '\v';
               break;
            case '\'':
               mTokenBuffer += '\'';
               break;
            case '\"':
               mTokenBuffer += '\"';
               break;
            case '\\':
               mTokenBuffer += '\\';
               break;
            default:
               break;
            }
            mEscapedChar = false;
         }
         else if(curchar == '\"')
         {
            Token token;
            token.mType = Token::STRING;
            token.mData = mTokenBuffer;
            ProcessToken(token);
            mTokenBuffer = "";
            mInString = false;
         }
         else if(curchar == '\\')
            mEscapedChar = true;
         else
            mTokenBuffer += curchar;
         continue;
      }

      if(mJustParsedColon && curchar != ':')
      {
         mJustParsedColon = false;
         Token token;
         token.mType = Token::COLON;
         token.mData = ":";
         ProcessToken(token);
      }

      if(isalnum(curchar) || curchar == '_')
         mTokenBuffer += curchar;
      else
      {
         if(mTokenBuffer.length() > 0)
         {
            if(isdigit(mTokenBuffer[0]))
            {
               bool isNumber = true;
               for(unsigned int j = 0; j < mTokenBuffer.length(); j++)
               {
                  if(!isdigit(mTokenBuffer[j]))
                  {
                     PrintError("Non-digits in number");
                     isNumber = false;
                     break;
                  }
               }

               if(isNumber)
               {
                  Token token;
                  token.mType = Token::NUMBER;
                  token.mNumber = strtol(mTokenBuffer.c_str(), NULL, 10);
                  token.mData = mTokenBuffer;
                  ProcessToken(token);
               }
            }
            else
            {
               Token token;

               token.mType = Token::SYMBOL;
               token.mData = mTokenBuffer;

               ProcessToken(token);
            }

            mTokenBuffer = "";
         }

         if(!isspace(curchar) || curchar == '\n')
         {
            Token token;

            token.mData += curchar;

            switch(curchar)
            {
            case '{':
               token.mType = Token::OPEN_BRACE;
               ProcessToken(token);
               break;
            case '}':
               token.mType = Token::CLOSE_BRACE;
               ProcessToken(token);
               break;
            case ':':
               if(mJustParsedColon)
               {
                  token.mType = Token::DOUBLE_COLON;
                  token.mData = "::";
                  ProcessToken(token);
                  mJustParsedColon = false;
               }
               else
                  mJustParsedColon = true;
               break;
            case ';':
               token.mType = Token::SEMICOLON;
               ProcessToken(token);
               break;
            case '=':
               token.mType = Token::EQUALS;
               ProcessToken(token);
               break;
            case '#':
               token.mType = Token::HASH;
               ProcessToken(token);
               break;
            case '\n':
               token.mType = Token::NEWLINE;
               ProcessToken(token);
               break;
            case '\"':
               mInString = true;
               break;
            default:
               PrintError(string("Bad character: ") + curchar);
               break;
            }
         }
      }
   }
}

void Parser::Finish()
{
   if(mCodeLine.size() != 0)
      PrintError("Unexpected EOF in the middle of a command.");
   else if(mElementStack.size() > 1)
      PrintError("Unexpected EOF in the middle of an element.");
}

void Parser::PrintError(string description)
{
   cerr << mFilename << ":" << mLineNum << ": " << description << endl;
   mHadErrors = true;
}

void Parser::ProcessToken(Token token)
{
   if(token.mType == Token::NEWLINE)
   {
      if(mInDirective)
         mInDirective = false;
      else
         ++mLineNum;
      mJustParsedNewline = true;
      return;
   }
   else if(mJustParsedNewline)
   {
      mJustParsedNewline = false;
      if(token.mType == Token::HASH)
      {
         mInDirective = true;
         mDirectiveState = START_DIRECTIVE;
         return;
      }
   }

   if(mInDirective)
   {
      switch(mDirectiveState)
      {
      case START_DIRECTIVE:
         if(token.mType == Token::NUMBER)
         {
            mLineNum = token.mNumber;
            mDirectiveState = GOT_LINENUM;
         }
         break;
      case GOT_LINENUM:
         if(token.mType == Token::STRING)
         {
            mFilename = token.mData;
            mDirectiveState = GOT_FILENAME;
         }
         break;
      default:
         //ignore
         break;
      }
      return;
   }

   switch(token.mType)
   {
   case Token::SEMICOLON:
      if(mCodeLine.size() > 0)
      {
         try
         {
            mElementStack.top()->ProcessCodeLine(mCodeLine, mFilename, mLineNum);
         }
         catch(const ParseElement::ParseException& e)
         {
            PrintError(e.GetErrorMessage());
         }
         mCodeLine.clear();
      }
      break;

   case Token::COLON:
      if(mCodeLine.size() == 1 && mCodeLine[0].mType == Token::SYMBOL)
      {
         if(mCodeLine[0].mData == "public")
         {
            mElementStack.top()->InPublic(true);
            mCodeLine.clear();
         }
         else if(mCodeLine[0].mData == "private")
         {
            mElementStack.top()->InPublic(false);
            mCodeLine.clear();
         }
      }
      else if(mCodeLine.size() > 0)
      {
         PrintError("Parse error before \":\"");
         mCodeLine.clear();
      }
      break;

   case Token::OPEN_BRACE:
      if(mCodeLine.size() < 1)
      {
         PrintError("Open brace with no preceeding declaration.");
         mElementStack.push(new IgnoreElement);
      }
      else if(mCodeLine.size() == 1)
      {
         PrintError("Missing object type.");
         mElementStack.push(new IgnoreElement);
      }
      else
      {
         vector<Token > flags;
         if(mCodeLine.size() > 2)
            flags.insert(flags.end(), mCodeLine.begin(), mCodeLine.end() - 2);

         if(mCodeLine[mCodeLine.size() - 1].mType != Token::SYMBOL)
            PrintError("Parse error before \"" + mCodeLine[mCodeLine.size() - 1].mData + "\"");
         else if(mCodeLine[mCodeLine.size() - 2].mType != Token::SYMBOL)
            PrintError("Parse error before \"" + mCodeLine[mCodeLine.size() - 2].mData + "\"");
         else
         {
            string type = mCodeLine[mCodeLine.size() - 2].mData;
            string name = mCodeLine[mCodeLine.size() - 1].mData;

            try
            {
               ParseElement* newElement =
                  mElementStack.top()->StartSubElement(type, name, flags, mFilename, mLineNum);
               if(newElement != NULL)
                  mElementStack.push(newElement);
               else
                  mElementStack.push(new IgnoreElement);
            }
            catch(const ParseElement::ParseException& e)
            {
               PrintError(e.GetErrorMessage());
               mElementStack.push(new IgnoreElement);
            }
         }

         mCodeLine.clear();
      }
      break;

   case Token::CLOSE_BRACE:
      if(mCodeLine.size() > 0)
      {
         mCodeLine.clear();
         PrintError("Missing semicolon before close brace");
      }

      {   //scope for topElement
         ParseElement* topElement = mElementStack.top();
         mElementStack.pop();
         if(IgnoreElement* ignore = dynamic_cast<IgnoreElement*>(topElement))
            delete ignore;
         else
            mElementStack.top()->EndSubElement(topElement);
      }
      break;

   default:
      mCodeLine.push_back(token);
      break;
   }
}

Parser::IgnoreElement::IgnoreElement()
{
}

Parser::IgnoreElement::~IgnoreElement()
{
}

void Parser::IgnoreElement::InPublic(bool flag) throw(ParseException)
{
}

ParseElement* Parser::IgnoreElement::StartSubElement(string type, string name, const vector<Token >& flags,
                                                     string sourceFile, int linenum) throw(ParseException)
{
   return new IgnoreElement;
}

void Parser::IgnoreElement::EndSubElement(ParseElement* element) throw(ParseException)
{
}

void Parser::IgnoreElement::ProcessCodeLine(const vector<Token >& tokens,
                                            string sourceFile, int linenum) throw(ParseException)
{
}

} //namespace kake
