/* kake - A Semi-Replacement for Make
 * Copyright (C) 2001 Kenton Varda
 * <temporal@gauge3d.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
#include "Process.h"

#include <stdio.h>
#include <errno.h>
#include <iostream>

#ifdef PLATFORM_POSIX
# include <sys/types.h>
# include <sys/wait.h>
# include <unistd.h>
#elif defined(PLATFORM_WIN32)
# include <io.h>
# include <process.h>
# include <fcntl.h>
# define open     _open
# define lseek    _lseek
# define read     _read
# define write    _write
# define close    _close
# define dup      _dup
# define dup2     _dup2
# define spawnvp  _spawnvp
# define P_NOWAIT _P_NOWAIT
# define O_WRONLY _O_WRONLY
# define O_CREAT  _O_CREAT
# define O_TRUNC  _O_TRUNC
# define O_BINARY _O_BINARY
# define O_NOINHERIT _O_NOINHERIT
# define pipe(PARAM) _pipe(PARAM, 4096, O_BINARY | O_NOINHERIT)
# define waitpid(PID, STATUS, ACTION) _cwait(STATUS, PID, ACTION)
#else
# error "Host OS not supported!"
#endif

namespace kake {

enum {INPUT, OUTPUT};  //for pipe

bool Process::mEcho = false;

// class member functions ============================================
Process::Process()
   : mReadHandle(-1),
     mWriteHandle(-1),
     mChildPID(-1)
{
}

Process::~Process()
{
   if(mChildPID != -1)
      Wait();    //prevent child from developing a taste for brains
}

void Process::Run(string progname, const vector<string >& params, bool usePipe)
{
   if(mChildPID != -1)
      return;  //one per object, please...

   const char* paramArray[params.size() + 2];

   paramArray[0] = progname.c_str();
   if(mEcho) cout << progname;
   for(unsigned int i = 0; i < params.size(); i++)
   {
      paramArray[i+1] = params[i].c_str();
      if(mEcho) cout << ' ' << params[i];
   }
   paramArray[params.size()+1] = NULL;
   if(mEcho) cout << endl;

   int parentToChildPipe[2], childToParentPipe[2];

#ifdef PLATFORM_POSIX
   if(usePipe)
   {
      if(pipe(parentToChildPipe) < 0)
      {
         perror("pipe");
         throw "Process::Run(): pipe() failed";
      }
      if(pipe(childToParentPipe) < 0)
      {
         perror("pipe");
         throw "Process::Run(): pipe() failed";
      }
   }

   if((mChildPID = fork()) == 0)
   {
      //in child

      if(usePipe)
      {
         //copy pipes to stdin and stdout
         close(0);
         close(1);
         dup2(parentToChildPipe[INPUT], 0);
         dup2(childToParentPipe[OUTPUT], 1);
         close(parentToChildPipe[INPUT]);
         close(parentToChildPipe[OUTPUT]);
         close(childToParentPipe[INPUT]);
         close(childToParentPipe[OUTPUT]);
      }

      //run the proggie
      execvp(progname.c_str(), (char*const*)paramArray);

      //execvp should not return...
      perror("execvp");
      exit(-1);
   }

   if(mChildPID < 0)
   {
      perror("fork");
      throw "Process::Run(): fork() failed";
   }

   if(usePipe)
   {
      //close child's end of the pipes
      close(parentToChildPipe[INPUT]);
      close(childToParentPipe[OUTPUT]);

      //remember our ends
      mReadHandle = childToParentPipe[INPUT];
      mWriteHandle = parentToChildPipe[OUTPUT];
   }

#elif defined(PLATFORM_WIN32)
   int backupStdin = 0;
   int backupStdout = 0;

   if(usePipe)
   {
      backupStdin  = dup(0);
      backupStdout = dup(1);

      if(pipe(parentToChildPipe) < 0)
      {
         perror("pipe");
         throw "Process::Run(): pipe() failed";
      }
      if(pipe(childToParentPipe) < 0)
      {
         perror("pipe");
         throw "Process::Run(): pipe() failed";
      }

      dup2(parentToChildPipe[INPUT], 0);
      dup2(childToParentPipe[OUTPUT], 1);
      close(parentToChildPipe[INPUT]);
      close(childToParentPipe[OUTPUT]);
   }

   //spawn.  It's not crappy like fork and exec.  Instead, it's crappy
   // in a totally new and different way.  (Specifically, you can't make
   // the callee process close the unneeded file descriptors.)
   mChildPID = spawnvp(P_NOWAIT, progname.c_str(), (char*const*)paramArray);

   if(usePipe)
   {
      //restore stdin and stdout
      dup2(backupStdin, 0);
      dup2(backupStdout, 1);
      close(backupStdin);
      close(backupStdout);

      //remember our ends of the pipes
      mReadHandle = childToParentPipe[INPUT];
      mWriteHandle = parentToChildPipe[OUTPUT];
   }

   if(mChildPID < 0)
   {
      perror("spawnvp");
      throw "Process::Run(): fork() failed";
   }

#endif //PLATFORM
}

int Process::Read(void* buf, int size)
{
   if(mReadHandle == -1)
      return -1;

#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   return read(mReadHandle, buf, size);
#endif //PLATFORM_POSIX
}

int Process::Write(const void* buf, int size)
{
   if(mWriteHandle == -1)
      return -1;

#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   return write(mWriteHandle, buf, size);
#endif //PLATFORM_POSIX
}

void Process::CloseOutput()
{
   if(mWriteHandle != -1)
      close(mWriteHandle);
   mWriteHandle = -1;
}

int Process::Wait()
{
   if(mChildPID == -1)
      return -1;

#if defined(PLATFORM_POSIX) || defined(PLATFORM_WIN32)
   if(mWriteHandle != -1)
      close(mWriteHandle);
   mWriteHandle = -1;

   int status = -1;
   if(waitpid(mChildPID, &status, 0) < 0)
      perror("waitpid");
   mChildPID = -1;

   if(mReadHandle != -1)
      close(mReadHandle);
   mReadHandle = -1;

   return status;
#endif //PLATFORM_POSIX
}

bool Process::Test()
{
   SetEcho(true);

   Process proc;
   vector<string > params;

   proc.Run("cat", params, true);

   string output = "Hello World!\n";
   proc.Write(output.c_str(), output.length());
   proc.CloseOutput();

   char buf[1025];
   int size;

   while((size = proc.Read(buf, 1024)) > 0)
   {
      buf[size] = 0;
      cout << buf;
//      cout.write(buf, size);   //mysteriously missing in mingw32
      cout.flush();
   }

   cout << "exit status = " << proc.Wait() << endl;
   return true;
}

} //namespace kake
